/*
 * Copyright (c) 2007, 2011, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package com.sun.max.asm.ia32.complete;

import com.sun.max.asm.*;
import com.sun.max.asm.ia32.*;
import com.sun.max.asm.x86.*;

public abstract class IA32RawAssembler extends AbstractIA32Assembler {

    public IA32RawAssembler(int startAddress) {
        super(startAddress);
    }

    public IA32RawAssembler() {
        super();
    }

    @Override
    protected void emitPadding(int numberOfBytes) throws AssemblyException {
        for (int i = 0; i < numberOfBytes; i++) {
            nop();
        }
    }

// START GENERATED RAW ASSEMBLER METHODS
    /**
     * Pseudo-external assembler syntax: {@code aaa  }
     * Example disassembly syntax: {@code aaa     }
     */
    // Template#: 1, Serial#: 224
    public void aaa() {
        assemble0001((byte) 0x37);
    }

    /**
     * Pseudo-external assembler syntax: {@code aad  }<i>imm8</i>
     * Example disassembly syntax: {@code aad       0x12}
     */
    // Template#: 2, Serial#: 1329
    public void aad(byte imm8) {
        assemble0002((byte) 0xD5, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code aam  }<i>imm8</i>
     * Example disassembly syntax: {@code aam       0x12}
     */
    // Template#: 3, Serial#: 1328
    public void aam(byte imm8) {
        assemble0002((byte) 0xD4, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code aas  }
     * Example disassembly syntax: {@code aas     }
     */
    // Template#: 4, Serial#: 1758
    public void aas() {
        assemble0001((byte) 0x3F);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>imm8</i>
     * Example disassembly syntax: {@code adc       al, 0x12}
     */
    // Template#: 5, Serial#: 108
    public void adc_AL(byte imm8) {
        assemble0002((byte) 0x14, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcb      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 6, Serial#: 307
    public void adcb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0x80, (byte) 0x02, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcl      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 7, Serial#: 523
    public void adcl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0x83, (byte) 0x02, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcw      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 8, Serial#: 595
    public void adcw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0004((byte) 0x83, (byte) 0x02, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       ebx[esi * 4 + 18], ax}
     */
    // Template#: 9, Serial#: 80
    public void adc(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0005((byte) 0x11, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       ebx[esi * 4 + 18], eax}
     */
    // Template#: 10, Serial#: 71
    public void adc(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0006((byte) 0x11, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       ebx[esi * 4 + 18], al}
     */
    // Template#: 11, Serial#: 62
    public void adc(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0007((byte) 0x10, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code adcl      ebx[esi * 4 + 18], 0x12345678}
     */
    // Template#: 12, Serial#: 379
    public void adcl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0008((byte) 0x81, (byte) 0x02, disp8, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code adcw      ebx[esi * 4 + 18], 0x1234}
     */
    // Template#: 13, Serial#: 451
    public void adcw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0009((byte) 0x81, (byte) 0x02, disp8, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcb  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcb      [ebx + 18], 0x12}
     */
    // Template#: 14, Serial#: 306
    public void adcb(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0x80, (byte) 0x02, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcl  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcl      [ebx + 18], 0x12}
     */
    // Template#: 15, Serial#: 522
    public void adcl(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0x83, (byte) 0x02, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcw  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcw      [ebx + 18], 0x12}
     */
    // Template#: 16, Serial#: 594
    public void adcw(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0011((byte) 0x83, (byte) 0x02, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       [ebx + 18], ax}
     */
    // Template#: 17, Serial#: 79
    public void adc(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0012((byte) 0x11, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       [ebx + 18], eax}
     */
    // Template#: 18, Serial#: 70
    public void adc(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0013((byte) 0x11, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       [ebx + 18], al}
     */
    // Template#: 19, Serial#: 61
    public void adc(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0014((byte) 0x10, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcl  }<i>disp8</i>, <i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code adcl      [ebx + 18], 0x12345678}
     */
    // Template#: 20, Serial#: 378
    public void adcl(byte disp8, IA32IndirectRegister32 destination, int imm32) {
        assemble0015((byte) 0x81, (byte) 0x02, disp8, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcw  }<i>disp8</i>, <i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code adcw      [ebx + 18], 0x1234}
     */
    // Template#: 21, Serial#: 450
    public void adcw(byte disp8, IA32IndirectRegister32 destination, short imm16) {
        assemble0016((byte) 0x81, (byte) 0x02, disp8, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcb  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcb      ebx[esi * 4], 0x12}
     */
    // Template#: 22, Serial#: 279
    public void adcb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0x80, (byte) 0x02, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcl  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcl      ebx[esi * 4], 0x12}
     */
    // Template#: 23, Serial#: 495
    public void adcl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0x83, (byte) 0x02, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcw      ebx[esi * 4], 0x12}
     */
    // Template#: 24, Serial#: 567
    public void adcw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0018((byte) 0x83, (byte) 0x02, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       ebx[esi * 4], ax}
     */
    // Template#: 25, Serial#: 76
    public void adc(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0019((byte) 0x11, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       ebx[esi * 4], eax}
     */
    // Template#: 26, Serial#: 67
    public void adc(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0020((byte) 0x11, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       ebx[esi * 4], al}
     */
    // Template#: 27, Serial#: 58
    public void adc(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0021((byte) 0x10, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcl  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code adcl      ebx[esi * 4], 0x12345678}
     */
    // Template#: 28, Serial#: 351
    public void adcl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0022((byte) 0x81, (byte) 0x02, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code adcw      ebx[esi * 4], 0x1234}
     */
    // Template#: 29, Serial#: 423
    public void adcw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0023((byte) 0x81, (byte) 0x02, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcw      ax, 0x12}
     */
    // Template#: 30, Serial#: 624
    public void adcw(IA32GeneralRegister16 destination, byte imm8) {
        assemble0024((byte) 0x83, (byte) 0x02, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code adc       ax, ebx[esi * 4 + 18]}
     */
    // Template#: 31, Serial#: 105
    public void adc(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0025((byte) 0x13, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       ax, [ebx + 18]}
     */
    // Template#: 32, Serial#: 104
    public void adc(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0026((byte) 0x13, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code adc       ax, ebx[esi * 4]}
     */
    // Template#: 33, Serial#: 101
    public void adc(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0027((byte) 0x13, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       ax, ax}
     */
    // Template#: 34, Serial#: 83
    public void adc(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0028((byte) 0x11, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       ax, [ebx]}
     */
    // Template#: 35, Serial#: 100
    public void adc(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0029((byte) 0x13, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code adc       ax, [L1: 0x12345678]}
     */
    // Template#: 36, Serial#: 103
    public void m_adc(IA32GeneralRegister16 destination, int m32) {
        assemble0030((byte) 0x13, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code adc       ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 37, Serial#: 107
    public void adc(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0031((byte) 0x13, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code adc       ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 38, Serial#: 102
    public void m_adc(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0032((byte) 0x13, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       ax, [ebx + 305419896]}
     */
    // Template#: 39, Serial#: 106
    public void adc(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0033((byte) 0x13, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcw  }<i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code adcw      ax, 0x1234}
     */
    // Template#: 40, Serial#: 480
    public void adcw(IA32GeneralRegister16 destination, short imm16) {
        assemble0034((byte) 0x81, (byte) 0x02, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcl  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcl      eax, 0x12}
     */
    // Template#: 41, Serial#: 552
    public void adcl(IA32GeneralRegister32 destination, byte imm8) {
        assemble0035((byte) 0x83, (byte) 0x02, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code adc       eax, ebx[esi * 4 + 18]}
     */
    // Template#: 42, Serial#: 97
    public void adc(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0036((byte) 0x13, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       eax, [ebx + 18]}
     */
    // Template#: 43, Serial#: 96
    public void adc(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0037((byte) 0x13, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code adc       eax, ebx[esi * 4]}
     */
    // Template#: 44, Serial#: 93
    public void adc(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0038((byte) 0x13, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       eax, eax}
     */
    // Template#: 45, Serial#: 74
    public void adc(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0039((byte) 0x11, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       eax, [ebx]}
     */
    // Template#: 46, Serial#: 92
    public void adc(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0040((byte) 0x13, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code adc       eax, [L1: 0x12345678]}
     */
    // Template#: 47, Serial#: 95
    public void m_adc(IA32GeneralRegister32 destination, int m32) {
        assemble0041((byte) 0x13, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcl  }<i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code adcl      eax, 0x12345678}
     */
    // Template#: 48, Serial#: 408
    public void adcl(IA32GeneralRegister32 destination, int imm32) {
        assemble0042((byte) 0x81, (byte) 0x02, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code adc       eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 49, Serial#: 99
    public void adc(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0043((byte) 0x13, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code adc       eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 50, Serial#: 94
    public void m_adc(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0044((byte) 0x13, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       eax, [ebx + 305419896]}
     */
    // Template#: 51, Serial#: 98
    public void adc(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0045((byte) 0x13, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcb      al, 0x12}
     */
    // Template#: 52, Serial#: 336
    public void adcb(IA32GeneralRegister8 destination, byte imm8) {
        assemble0046((byte) 0x80, (byte) 0x02, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code adc       al, ebx[esi * 4 + 18]}
     */
    // Template#: 53, Serial#: 89
    public void adc(IA32GeneralRegister8 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0047((byte) 0x12, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       al, [ebx + 18]}
     */
    // Template#: 54, Serial#: 88
    public void adc(IA32GeneralRegister8 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0048((byte) 0x12, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code adc       al, ebx[esi * 4]}
     */
    // Template#: 55, Serial#: 85
    public void adc(IA32GeneralRegister8 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0049((byte) 0x12, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       al, al}
     */
    // Template#: 56, Serial#: 65
    public void adc(IA32GeneralRegister8 destination, IA32GeneralRegister8 source) {
        assemble0050((byte) 0x10, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       al, [ebx]}
     */
    // Template#: 57, Serial#: 84
    public void adc(IA32GeneralRegister8 destination, IA32IndirectRegister32 source) {
        assemble0051((byte) 0x12, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code adc       al, [L1: 0x12345678]}
     */
    // Template#: 58, Serial#: 87
    public void m_adc(IA32GeneralRegister8 destination, int m32) {
        assemble0052((byte) 0x12, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code adc       al, ebx[esi * 4 + 305419896]}
     */
    // Template#: 59, Serial#: 91
    public void adc(IA32GeneralRegister8 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0053((byte) 0x12, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code adc       al, L1: 0x12345678[esi * 4]}
     */
    // Template#: 60, Serial#: 86
    public void m_adc(IA32GeneralRegister8 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0054((byte) 0x12, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       al, [ebx + 305419896]}
     */
    // Template#: 61, Serial#: 90
    public void adc(IA32GeneralRegister8 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0055((byte) 0x12, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcb      [ebx], 0x12}
     */
    // Template#: 62, Serial#: 278
    public void adcb(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0x80, (byte) 0x02, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcl  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcl      [ebx], 0x12}
     */
    // Template#: 63, Serial#: 494
    public void adcl(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0x83, (byte) 0x02, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcw      [ebx], 0x12}
     */
    // Template#: 64, Serial#: 566
    public void adcw(IA32IndirectRegister32 destination, byte imm8) {
        assemble0057((byte) 0x83, (byte) 0x02, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       [ebx], ax}
     */
    // Template#: 65, Serial#: 75
    public void adc(IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0058((byte) 0x11, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       [ebx], eax}
     */
    // Template#: 66, Serial#: 66
    public void adc(IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0059((byte) 0x11, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       [ebx], al}
     */
    // Template#: 67, Serial#: 57
    public void adc(IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0060((byte) 0x10, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcl  }<i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code adcl      [ebx], 0x12345678}
     */
    // Template#: 68, Serial#: 350
    public void adcl(IA32IndirectRegister32 destination, int imm32) {
        assemble0061((byte) 0x81, (byte) 0x02, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcw  }<i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code adcw      [ebx], 0x1234}
     */
    // Template#: 69, Serial#: 422
    public void adcw(IA32IndirectRegister32 destination, short imm16) {
        assemble0062((byte) 0x81, (byte) 0x02, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcl  }<i>imm32</i>
     * Example disassembly syntax: {@code adcl      eax, 0x12345678}
     */
    // Template#: 70, Serial#: 109
    public void adc_EAX(int imm32) {
        assemble0063((byte) 0x15, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcb  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcb      [L1: 0x12345678], 0x12}
     */
    // Template#: 71, Serial#: 281
    public void m_adcb(int m32, byte imm8) {
        assemble0064((byte) 0x80, (byte) 0x02, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcl  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcl      [L1: 0x12345678], 0x12}
     */
    // Template#: 72, Serial#: 497
    public void m_adcl(int m32, byte imm8) {
        assemble0064((byte) 0x83, (byte) 0x02, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcw  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcw      [L1: 0x12345678], 0x12}
     */
    // Template#: 73, Serial#: 569
    public void m_adcw(int m32, byte imm8) {
        assemble0065((byte) 0x83, (byte) 0x02, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcb      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 74, Serial#: 323
    public void adcb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0x80, (byte) 0x02, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcl      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 75, Serial#: 539
    public void adcl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0x83, (byte) 0x02, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcw      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 76, Serial#: 611
    public void adcw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0067((byte) 0x83, (byte) 0x02, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       ebx[esi * 4 + 305419896], ax}
     */
    // Template#: 77, Serial#: 82
    public void adc(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0068((byte) 0x11, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       ebx[esi * 4 + 305419896], eax}
     */
    // Template#: 78, Serial#: 73
    public void adc(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0069((byte) 0x11, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       ebx[esi * 4 + 305419896], al}
     */
    // Template#: 79, Serial#: 64
    public void adc(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0070((byte) 0x10, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code adcl      ebx[esi * 4 + 305419896], 0x12345678}
     */
    // Template#: 80, Serial#: 395
    public void adcl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0071((byte) 0x81, (byte) 0x02, disp32, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code adcw      ebx[esi * 4 + 305419896], 0x1234}
     */
    // Template#: 81, Serial#: 467
    public void adcw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0072((byte) 0x81, (byte) 0x02, disp32, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       [L1: 0x12345678], ax}
     */
    // Template#: 82, Serial#: 78
    public void m_adc(int m32, IA32GeneralRegister16 source) {
        assemble0073((byte) 0x11, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       [L1: 0x12345678], eax}
     */
    // Template#: 83, Serial#: 69
    public void m_adc(int m32, IA32GeneralRegister32 source) {
        assemble0074((byte) 0x11, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       [L1: 0x12345678], al}
     */
    // Template#: 84, Serial#: 60
    public void m_adc(int m32, IA32GeneralRegister8 source) {
        assemble0075((byte) 0x10, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcb  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcb      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 85, Serial#: 280
    public void m_adcb(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0x80, (byte) 0x02, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcl  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcl      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 86, Serial#: 496
    public void m_adcl(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0x83, (byte) 0x02, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcw      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 87, Serial#: 568
    public void m_adcw(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0077((byte) 0x83, (byte) 0x02, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       L1: 0x12345678[esi * 4], ax}
     */
    // Template#: 88, Serial#: 77
    public void m_adc(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0078((byte) 0x11, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       L1: 0x12345678[esi * 4], eax}
     */
    // Template#: 89, Serial#: 68
    public void m_adc(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0079((byte) 0x11, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       L1: 0x12345678[esi * 4], al}
     */
    // Template#: 90, Serial#: 59
    public void m_adc(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0080((byte) 0x10, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcl  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code adcl      L1: 0x12345678[esi * 4], 0x12345678}
     */
    // Template#: 91, Serial#: 352
    public void m_adcl(int m32, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0081((byte) 0x81, (byte) 0x02, m32, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code adcw      L1: 0x12345678[esi * 4], 0x1234}
     */
    // Template#: 92, Serial#: 424
    public void m_adcw(int m32, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0082((byte) 0x81, (byte) 0x02, m32, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcb  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcb      [ebx + 305419896], 0x12}
     */
    // Template#: 93, Serial#: 322
    public void adcb(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0x80, (byte) 0x02, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcl  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcl      [ebx + 305419896], 0x12}
     */
    // Template#: 94, Serial#: 538
    public void adcl(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0x83, (byte) 0x02, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcw  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code adcw      [ebx + 305419896], 0x12}
     */
    // Template#: 95, Serial#: 610
    public void adcw(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0084((byte) 0x83, (byte) 0x02, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       [ebx + 305419896], ax}
     */
    // Template#: 96, Serial#: 81
    public void adc(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0085((byte) 0x11, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       [ebx + 305419896], eax}
     */
    // Template#: 97, Serial#: 72
    public void adc(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0086((byte) 0x11, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adc  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code adc       [ebx + 305419896], al}
     */
    // Template#: 98, Serial#: 63
    public void adc(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0087((byte) 0x10, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcl  }<i>disp32</i>, <i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code adcl      [ebx + 305419896], 0x12345678}
     */
    // Template#: 99, Serial#: 394
    public void adcl(int disp32, IA32IndirectRegister32 destination, int imm32) {
        assemble0088((byte) 0x81, (byte) 0x02, disp32, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcw  }<i>disp32</i>, <i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code adcw      [ebx + 305419896], 0x1234}
     */
    // Template#: 100, Serial#: 466
    public void adcw(int disp32, IA32IndirectRegister32 destination, short imm16) {
        assemble0089((byte) 0x81, (byte) 0x02, disp32, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcl  }<i>m32</i>, <i>imm32</i>
     * Example disassembly syntax: {@code adcl      [L1: 0x12345678], 0x12345678}
     */
    // Template#: 101, Serial#: 353
    public void m_adcl(int m32, int imm32) {
        assemble0090((byte) 0x81, (byte) 0x02, m32, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcw  }<i>m32</i>, <i>imm16</i>
     * Example disassembly syntax: {@code adcw      [L1: 0x12345678], 0x1234}
     */
    // Template#: 102, Serial#: 425
    public void m_adcw(int m32, short imm16) {
        assemble0091((byte) 0x81, (byte) 0x02, m32, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code adcw  }<i>imm16</i>
     * Example disassembly syntax: {@code adcw      ax, 0x1234}
     */
    // Template#: 103, Serial#: 110
    public void adc_AX(short imm16) {
        assemble0092((byte) 0x15, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>imm8</i>
     * Example disassembly syntax: {@code add       al, 0x12}
     */
    // Template#: 104, Serial#: 52
    public void add_AL(byte imm8) {
        assemble0002((byte) 0x04, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code addb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addb      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 105, Serial#: 303
    public void addb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0x80, (byte) 0x00, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code addl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addl      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 106, Serial#: 519
    public void addl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0x83, (byte) 0x00, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code addw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addw      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 107, Serial#: 591
    public void addw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0004((byte) 0x83, (byte) 0x00, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code add       ebx[esi * 4 + 18], ax}
     */
    // Template#: 108, Serial#: 24
    public void add(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0005((byte) 0x01, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code add       ebx[esi * 4 + 18], eax}
     */
    // Template#: 109, Serial#: 15
    public void add(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0006((byte) 0x01, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code add       ebx[esi * 4 + 18], al}
     */
    // Template#: 110, Serial#: 6
    public void add(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0007((byte) 0x00, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code addl      ebx[esi * 4 + 18], 0x12345678}
     */
    // Template#: 111, Serial#: 375
    public void addl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0008((byte) 0x81, (byte) 0x00, disp8, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code addw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code addw      ebx[esi * 4 + 18], 0x1234}
     */
    // Template#: 112, Serial#: 447
    public void addw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0009((byte) 0x81, (byte) 0x00, disp8, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code addb  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addb      [ebx + 18], 0x12}
     */
    // Template#: 113, Serial#: 302
    public void addb(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0x80, (byte) 0x00, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code addl  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addl      [ebx + 18], 0x12}
     */
    // Template#: 114, Serial#: 518
    public void addl(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0x83, (byte) 0x00, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code addw  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addw      [ebx + 18], 0x12}
     */
    // Template#: 115, Serial#: 590
    public void addw(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0011((byte) 0x83, (byte) 0x00, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code add       [ebx + 18], ax}
     */
    // Template#: 116, Serial#: 23
    public void add(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0012((byte) 0x01, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code add       [ebx + 18], eax}
     */
    // Template#: 117, Serial#: 14
    public void add(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0013((byte) 0x01, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code add       [ebx + 18], al}
     */
    // Template#: 118, Serial#: 5
    public void add(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0014((byte) 0x00, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addl  }<i>disp8</i>, <i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code addl      [ebx + 18], 0x12345678}
     */
    // Template#: 119, Serial#: 374
    public void addl(byte disp8, IA32IndirectRegister32 destination, int imm32) {
        assemble0015((byte) 0x81, (byte) 0x00, disp8, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code addw  }<i>disp8</i>, <i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code addw      [ebx + 18], 0x1234}
     */
    // Template#: 120, Serial#: 446
    public void addw(byte disp8, IA32IndirectRegister32 destination, short imm16) {
        assemble0016((byte) 0x81, (byte) 0x00, disp8, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code addb  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addb      ebx[esi * 4], 0x12}
     */
    // Template#: 121, Serial#: 271
    public void addb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0x80, (byte) 0x00, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code addl  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addl      ebx[esi * 4], 0x12}
     */
    // Template#: 122, Serial#: 487
    public void addl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0x83, (byte) 0x00, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code addw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addw      ebx[esi * 4], 0x12}
     */
    // Template#: 123, Serial#: 559
    public void addw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0018((byte) 0x83, (byte) 0x00, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code add       ebx[esi * 4], ax}
     */
    // Template#: 124, Serial#: 20
    public void add(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0019((byte) 0x01, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code add       ebx[esi * 4], eax}
     */
    // Template#: 125, Serial#: 11
    public void add(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0020((byte) 0x01, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code add       ebx[esi * 4], al}
     */
    // Template#: 126, Serial#: 2
    public void add(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0021((byte) 0x00, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addl  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code addl      ebx[esi * 4], 0x12345678}
     */
    // Template#: 127, Serial#: 343
    public void addl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0022((byte) 0x81, (byte) 0x00, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code addw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code addw      ebx[esi * 4], 0x1234}
     */
    // Template#: 128, Serial#: 415
    public void addw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0023((byte) 0x81, (byte) 0x00, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code addw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addw      ax, 0x12}
     */
    // Template#: 129, Serial#: 622
    public void addw(IA32GeneralRegister16 destination, byte imm8) {
        assemble0024((byte) 0x83, (byte) 0x00, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code add       ax, ebx[esi * 4 + 18]}
     */
    // Template#: 130, Serial#: 49
    public void add(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0025((byte) 0x03, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code add       ax, [ebx + 18]}
     */
    // Template#: 131, Serial#: 48
    public void add(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0026((byte) 0x03, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code add       ax, ebx[esi * 4]}
     */
    // Template#: 132, Serial#: 45
    public void add(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0027((byte) 0x03, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code add       ax, ax}
     */
    // Template#: 133, Serial#: 27
    public void add(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0028((byte) 0x01, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code add       ax, [ebx]}
     */
    // Template#: 134, Serial#: 44
    public void add(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0029((byte) 0x03, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code add       ax, [L1: 0x12345678]}
     */
    // Template#: 135, Serial#: 47
    public void m_add(IA32GeneralRegister16 destination, int m32) {
        assemble0030((byte) 0x03, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code add       ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 136, Serial#: 51
    public void add(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0031((byte) 0x03, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code add       ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 137, Serial#: 46
    public void m_add(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0032((byte) 0x03, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code add       ax, [ebx + 305419896]}
     */
    // Template#: 138, Serial#: 50
    public void add(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0033((byte) 0x03, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addw  }<i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code addw      ax, 0x1234}
     */
    // Template#: 139, Serial#: 478
    public void addw(IA32GeneralRegister16 destination, short imm16) {
        assemble0034((byte) 0x81, (byte) 0x00, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code addl  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addl      eax, 0x12}
     */
    // Template#: 140, Serial#: 550
    public void addl(IA32GeneralRegister32 destination, byte imm8) {
        assemble0035((byte) 0x83, (byte) 0x00, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code add       eax, ebx[esi * 4 + 18]}
     */
    // Template#: 141, Serial#: 41
    public void add(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0036((byte) 0x03, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code add       eax, [ebx + 18]}
     */
    // Template#: 142, Serial#: 40
    public void add(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0037((byte) 0x03, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code add       eax, ebx[esi * 4]}
     */
    // Template#: 143, Serial#: 37
    public void add(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0038((byte) 0x03, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code add       eax, eax}
     */
    // Template#: 144, Serial#: 18
    public void add(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0039((byte) 0x01, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code add       eax, [ebx]}
     */
    // Template#: 145, Serial#: 36
    public void add(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0040((byte) 0x03, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code add       eax, [L1: 0x12345678]}
     */
    // Template#: 146, Serial#: 39
    public void m_add(IA32GeneralRegister32 destination, int m32) {
        assemble0041((byte) 0x03, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code addl  }<i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code addl      eax, 0x12345678}
     */
    // Template#: 147, Serial#: 406
    public void addl(IA32GeneralRegister32 destination, int imm32) {
        assemble0042((byte) 0x81, (byte) 0x00, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code add       eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 148, Serial#: 43
    public void add(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0043((byte) 0x03, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code add       eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 149, Serial#: 38
    public void m_add(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0044((byte) 0x03, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code add       eax, [ebx + 305419896]}
     */
    // Template#: 150, Serial#: 42
    public void add(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0045((byte) 0x03, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addb      al, 0x12}
     */
    // Template#: 151, Serial#: 334
    public void addb(IA32GeneralRegister8 destination, byte imm8) {
        assemble0046((byte) 0x80, (byte) 0x00, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code add       al, ebx[esi * 4 + 18]}
     */
    // Template#: 152, Serial#: 33
    public void add(IA32GeneralRegister8 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0047((byte) 0x02, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code add       al, [ebx + 18]}
     */
    // Template#: 153, Serial#: 32
    public void add(IA32GeneralRegister8 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0048((byte) 0x02, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code add       al, ebx[esi * 4]}
     */
    // Template#: 154, Serial#: 29
    public void add(IA32GeneralRegister8 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0049((byte) 0x02, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code add       al, al}
     */
    // Template#: 155, Serial#: 9
    public void add(IA32GeneralRegister8 destination, IA32GeneralRegister8 source) {
        assemble0050((byte) 0x00, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code add       al, [ebx]}
     */
    // Template#: 156, Serial#: 28
    public void add(IA32GeneralRegister8 destination, IA32IndirectRegister32 source) {
        assemble0051((byte) 0x02, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code add       al, [L1: 0x12345678]}
     */
    // Template#: 157, Serial#: 31
    public void m_add(IA32GeneralRegister8 destination, int m32) {
        assemble0052((byte) 0x02, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code add       al, ebx[esi * 4 + 305419896]}
     */
    // Template#: 158, Serial#: 35
    public void add(IA32GeneralRegister8 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0053((byte) 0x02, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code add       al, L1: 0x12345678[esi * 4]}
     */
    // Template#: 159, Serial#: 30
    public void m_add(IA32GeneralRegister8 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0054((byte) 0x02, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code add       al, [ebx + 305419896]}
     */
    // Template#: 160, Serial#: 34
    public void add(IA32GeneralRegister8 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0055((byte) 0x02, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addb      [ebx], 0x12}
     */
    // Template#: 161, Serial#: 270
    public void addb(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0x80, (byte) 0x00, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code addl  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addl      [ebx], 0x12}
     */
    // Template#: 162, Serial#: 486
    public void addl(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0x83, (byte) 0x00, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code addw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addw      [ebx], 0x12}
     */
    // Template#: 163, Serial#: 558
    public void addw(IA32IndirectRegister32 destination, byte imm8) {
        assemble0057((byte) 0x83, (byte) 0x00, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code add       [ebx], ax}
     */
    // Template#: 164, Serial#: 19
    public void add(IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0058((byte) 0x01, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code add       [ebx], eax}
     */
    // Template#: 165, Serial#: 10
    public void add(IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0059((byte) 0x01, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code add       [ebx], al}
     */
    // Template#: 166, Serial#: 1
    public void add(IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0060((byte) 0x00, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addl  }<i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code addl      [ebx], 0x12345678}
     */
    // Template#: 167, Serial#: 342
    public void addl(IA32IndirectRegister32 destination, int imm32) {
        assemble0061((byte) 0x81, (byte) 0x00, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code addw  }<i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code addw      [ebx], 0x1234}
     */
    // Template#: 168, Serial#: 414
    public void addw(IA32IndirectRegister32 destination, short imm16) {
        assemble0062((byte) 0x81, (byte) 0x00, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code addl  }<i>imm32</i>
     * Example disassembly syntax: {@code addl      eax, 0x12345678}
     */
    // Template#: 169, Serial#: 53
    public void add_EAX(int imm32) {
        assemble0063((byte) 0x05, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code addb  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addb      [L1: 0x12345678], 0x12}
     */
    // Template#: 170, Serial#: 273
    public void m_addb(int m32, byte imm8) {
        assemble0064((byte) 0x80, (byte) 0x00, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code addl  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addl      [L1: 0x12345678], 0x12}
     */
    // Template#: 171, Serial#: 489
    public void m_addl(int m32, byte imm8) {
        assemble0064((byte) 0x83, (byte) 0x00, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code addw  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addw      [L1: 0x12345678], 0x12}
     */
    // Template#: 172, Serial#: 561
    public void m_addw(int m32, byte imm8) {
        assemble0065((byte) 0x83, (byte) 0x00, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code addb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addb      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 173, Serial#: 319
    public void addb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0x80, (byte) 0x00, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code addl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addl      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 174, Serial#: 535
    public void addl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0x83, (byte) 0x00, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code addw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addw      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 175, Serial#: 607
    public void addw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0067((byte) 0x83, (byte) 0x00, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code add       ebx[esi * 4 + 305419896], ax}
     */
    // Template#: 176, Serial#: 26
    public void add(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0068((byte) 0x01, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code add       ebx[esi * 4 + 305419896], eax}
     */
    // Template#: 177, Serial#: 17
    public void add(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0069((byte) 0x01, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code add       ebx[esi * 4 + 305419896], al}
     */
    // Template#: 178, Serial#: 8
    public void add(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0070((byte) 0x00, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code addl      ebx[esi * 4 + 305419896], 0x12345678}
     */
    // Template#: 179, Serial#: 391
    public void addl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0071((byte) 0x81, (byte) 0x00, disp32, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code addw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code addw      ebx[esi * 4 + 305419896], 0x1234}
     */
    // Template#: 180, Serial#: 463
    public void addw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0072((byte) 0x81, (byte) 0x00, disp32, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code add       [L1: 0x12345678], ax}
     */
    // Template#: 181, Serial#: 22
    public void m_add(int m32, IA32GeneralRegister16 source) {
        assemble0073((byte) 0x01, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code add       [L1: 0x12345678], eax}
     */
    // Template#: 182, Serial#: 13
    public void m_add(int m32, IA32GeneralRegister32 source) {
        assemble0074((byte) 0x01, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code add       [L1: 0x12345678], al}
     */
    // Template#: 183, Serial#: 4
    public void m_add(int m32, IA32GeneralRegister8 source) {
        assemble0075((byte) 0x00, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addb  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addb      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 184, Serial#: 272
    public void m_addb(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0x80, (byte) 0x00, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code addl  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addl      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 185, Serial#: 488
    public void m_addl(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0x83, (byte) 0x00, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code addw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addw      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 186, Serial#: 560
    public void m_addw(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0077((byte) 0x83, (byte) 0x00, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code add       L1: 0x12345678[esi * 4], ax}
     */
    // Template#: 187, Serial#: 21
    public void m_add(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0078((byte) 0x01, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code add       L1: 0x12345678[esi * 4], eax}
     */
    // Template#: 188, Serial#: 12
    public void m_add(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0079((byte) 0x01, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code add       L1: 0x12345678[esi * 4], al}
     */
    // Template#: 189, Serial#: 3
    public void m_add(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0080((byte) 0x00, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addl  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code addl      L1: 0x12345678[esi * 4], 0x12345678}
     */
    // Template#: 190, Serial#: 344
    public void m_addl(int m32, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0081((byte) 0x81, (byte) 0x00, m32, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code addw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code addw      L1: 0x12345678[esi * 4], 0x1234}
     */
    // Template#: 191, Serial#: 416
    public void m_addw(int m32, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0082((byte) 0x81, (byte) 0x00, m32, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code addb  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addb      [ebx + 305419896], 0x12}
     */
    // Template#: 192, Serial#: 318
    public void addb(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0x80, (byte) 0x00, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code addl  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addl      [ebx + 305419896], 0x12}
     */
    // Template#: 193, Serial#: 534
    public void addl(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0x83, (byte) 0x00, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code addw  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code addw      [ebx + 305419896], 0x12}
     */
    // Template#: 194, Serial#: 606
    public void addw(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0084((byte) 0x83, (byte) 0x00, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code add       [ebx + 305419896], ax}
     */
    // Template#: 195, Serial#: 25
    public void add(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0085((byte) 0x01, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code add       [ebx + 305419896], eax}
     */
    // Template#: 196, Serial#: 16
    public void add(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0086((byte) 0x01, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code add  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code add       [ebx + 305419896], al}
     */
    // Template#: 197, Serial#: 7
    public void add(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0087((byte) 0x00, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addl  }<i>disp32</i>, <i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code addl      [ebx + 305419896], 0x12345678}
     */
    // Template#: 198, Serial#: 390
    public void addl(int disp32, IA32IndirectRegister32 destination, int imm32) {
        assemble0088((byte) 0x81, (byte) 0x00, disp32, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code addw  }<i>disp32</i>, <i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code addw      [ebx + 305419896], 0x1234}
     */
    // Template#: 199, Serial#: 462
    public void addw(int disp32, IA32IndirectRegister32 destination, short imm16) {
        assemble0089((byte) 0x81, (byte) 0x00, disp32, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code addl  }<i>m32</i>, <i>imm32</i>
     * Example disassembly syntax: {@code addl      [L1: 0x12345678], 0x12345678}
     */
    // Template#: 200, Serial#: 345
    public void m_addl(int m32, int imm32) {
        assemble0090((byte) 0x81, (byte) 0x00, m32, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code addw  }<i>m32</i>, <i>imm16</i>
     * Example disassembly syntax: {@code addw      [L1: 0x12345678], 0x1234}
     */
    // Template#: 201, Serial#: 417
    public void m_addw(int m32, short imm16) {
        assemble0091((byte) 0x81, (byte) 0x00, m32, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code addw  }<i>imm16</i>
     * Example disassembly syntax: {@code addw      ax, 0x1234}
     */
    // Template#: 202, Serial#: 54
    public void add_AX(short imm16) {
        assemble0092((byte) 0x05, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code addpd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code addpd     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 203, Serial#: 4447
    public void addpd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x58, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code addpd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code addpd     xmm0, [ebx + 18]}
     */
    // Template#: 204, Serial#: 4446
    public void addpd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x58, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addpd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code addpd     xmm0, ebx[esi * 4]}
     */
    // Template#: 205, Serial#: 4443
    public void addpd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x58, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code addpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code addpd     xmm0, [ebx]}
     */
    // Template#: 206, Serial#: 4442
    public void addpd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x58, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code addpd     xmm0, xmm0}
     */
    // Template#: 207, Serial#: 4450
    public void addpd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x58, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addpd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code addpd     xmm0, [L1: 0x12345678]}
     */
    // Template#: 208, Serial#: 4445
    public void m_addpd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x58, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code addpd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code addpd     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 209, Serial#: 4449
    public void addpd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x58, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code addpd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code addpd     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 210, Serial#: 4444
    public void m_addpd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x58, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code addpd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code addpd     xmm0, [ebx + 305419896]}
     */
    // Template#: 211, Serial#: 4448
    public void addpd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x58, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addps  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code addps     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 212, Serial#: 4375
    public void addps(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0102((byte) 0x58, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code addps  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code addps     xmm0, [ebx + 18]}
     */
    // Template#: 213, Serial#: 4374
    public void addps(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0103((byte) 0x58, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addps  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code addps     xmm0, ebx[esi * 4]}
     */
    // Template#: 214, Serial#: 4371
    public void addps(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0104((byte) 0x58, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code addps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code addps     xmm0, [ebx]}
     */
    // Template#: 215, Serial#: 4370
    public void addps(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0105((byte) 0x58, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code addps     xmm0, xmm0}
     */
    // Template#: 216, Serial#: 4378
    public void addps(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0106((byte) 0x58, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addps  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code addps     xmm0, [L1: 0x12345678]}
     */
    // Template#: 217, Serial#: 4373
    public void m_addps(IA32XMMRegister destination, int m32) {
        assemble0107((byte) 0x58, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code addps  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code addps     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 218, Serial#: 4377
    public void addps(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0108((byte) 0x58, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code addps  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code addps     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 219, Serial#: 4372
    public void m_addps(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0109((byte) 0x58, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code addps  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code addps     xmm0, [ebx + 305419896]}
     */
    // Template#: 220, Serial#: 4376
    public void addps(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0110((byte) 0x58, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code address_size  }
     * Example disassembly syntax: {@code address_size}
     */
    // Template#: 221, Serial#: 261
    public void address_size() {
        assemble0001((byte) 0x67);
    }

    /**
     * Pseudo-external assembler syntax: {@code addsd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code addsd     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 222, Serial#: 4519
    public void addsd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0111((byte) 0x58, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code addsd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code addsd     xmm0, [ebx + 18]}
     */
    // Template#: 223, Serial#: 4518
    public void addsd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0112((byte) 0x58, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addsd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code addsd     xmm0, ebx[esi * 4]}
     */
    // Template#: 224, Serial#: 4515
    public void addsd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0113((byte) 0x58, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code addsd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code addsd     xmm0, [ebx]}
     */
    // Template#: 225, Serial#: 4514
    public void addsd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0114((byte) 0x58, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addsd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code addsd     xmm0, xmm0}
     */
    // Template#: 226, Serial#: 4522
    public void addsd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0115((byte) 0x58, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addsd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code addsd     xmm0, [L1: 0x12345678]}
     */
    // Template#: 227, Serial#: 4517
    public void m_addsd(IA32XMMRegister destination, int m32) {
        assemble0116((byte) 0x58, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code addsd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code addsd     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 228, Serial#: 4521
    public void addsd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0117((byte) 0x58, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code addsd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code addsd     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 229, Serial#: 4516
    public void m_addsd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0118((byte) 0x58, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code addsd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code addsd     xmm0, [ebx + 305419896]}
     */
    // Template#: 230, Serial#: 4520
    public void addsd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0119((byte) 0x58, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addss  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code addss     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 231, Serial#: 4582
    public void addss(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0120((byte) 0x58, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code addss  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code addss     xmm0, [ebx + 18]}
     */
    // Template#: 232, Serial#: 4581
    public void addss(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0121((byte) 0x58, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addss  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code addss     xmm0, ebx[esi * 4]}
     */
    // Template#: 233, Serial#: 4578
    public void addss(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0122((byte) 0x58, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code addss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code addss     xmm0, [ebx]}
     */
    // Template#: 234, Serial#: 4577
    public void addss(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0123((byte) 0x58, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code addss     xmm0, xmm0}
     */
    // Template#: 235, Serial#: 4585
    public void addss(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0124((byte) 0x58, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addss  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code addss     xmm0, [L1: 0x12345678]}
     */
    // Template#: 236, Serial#: 4580
    public void m_addss(IA32XMMRegister destination, int m32) {
        assemble0125((byte) 0x58, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code addss  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code addss     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 237, Serial#: 4584
    public void addss(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0126((byte) 0x58, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code addss  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code addss     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 238, Serial#: 4579
    public void m_addss(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0127((byte) 0x58, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code addss  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code addss     xmm0, [ebx + 305419896]}
     */
    // Template#: 239, Serial#: 4583
    public void addss(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0128((byte) 0x58, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addsubpd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code addsubpd  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 240, Serial#: 3753
    public void addsubpd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xD0, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code addsubpd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code addsubpd  xmm0, [ebx + 18]}
     */
    // Template#: 241, Serial#: 3752
    public void addsubpd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xD0, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addsubpd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code addsubpd  xmm0, ebx[esi * 4]}
     */
    // Template#: 242, Serial#: 3749
    public void addsubpd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xD0, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code addsubpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code addsubpd  xmm0, [ebx]}
     */
    // Template#: 243, Serial#: 3748
    public void addsubpd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xD0, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addsubpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code addsubpd  xmm0, xmm0}
     */
    // Template#: 244, Serial#: 3756
    public void addsubpd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xD0, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code addsubpd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code addsubpd  xmm0, [L1: 0x12345678]}
     */
    // Template#: 245, Serial#: 3751
    public void m_addsubpd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xD0, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code addsubpd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code addsubpd  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 246, Serial#: 3755
    public void addsubpd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xD0, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code addsubpd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code addsubpd  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 247, Serial#: 3750
    public void m_addsubpd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xD0, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code addsubpd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code addsubpd  xmm0, [ebx + 305419896]}
     */
    // Template#: 248, Serial#: 3754
    public void addsubpd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xD0, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>imm8</i>
     * Example disassembly syntax: {@code and       al, 0x12}
     */
    // Template#: 249, Serial#: 164
    public void and_AL(byte imm8) {
        assemble0002((byte) 0x24, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code andb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andb      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 250, Serial#: 311
    public void andb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0x80, (byte) 0x04, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code andl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andl      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 251, Serial#: 527
    public void andl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0x83, (byte) 0x04, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code andw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andw      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 252, Serial#: 599
    public void andw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0004((byte) 0x83, (byte) 0x04, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code and       ebx[esi * 4 + 18], ax}
     */
    // Template#: 253, Serial#: 136
    public void and(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0005((byte) 0x21, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code and       ebx[esi * 4 + 18], eax}
     */
    // Template#: 254, Serial#: 127
    public void and(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0006((byte) 0x21, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code and       ebx[esi * 4 + 18], al}
     */
    // Template#: 255, Serial#: 118
    public void and(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0007((byte) 0x20, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code andl      ebx[esi * 4 + 18], 0x12345678}
     */
    // Template#: 256, Serial#: 383
    public void andl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0008((byte) 0x81, (byte) 0x04, disp8, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code andw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code andw      ebx[esi * 4 + 18], 0x1234}
     */
    // Template#: 257, Serial#: 455
    public void andw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0009((byte) 0x81, (byte) 0x04, disp8, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code andb  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andb      [ebx + 18], 0x12}
     */
    // Template#: 258, Serial#: 310
    public void andb(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0x80, (byte) 0x04, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code andl  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andl      [ebx + 18], 0x12}
     */
    // Template#: 259, Serial#: 526
    public void andl(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0x83, (byte) 0x04, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code andw  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andw      [ebx + 18], 0x12}
     */
    // Template#: 260, Serial#: 598
    public void andw(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0011((byte) 0x83, (byte) 0x04, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code and       [ebx + 18], ax}
     */
    // Template#: 261, Serial#: 135
    public void and(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0012((byte) 0x21, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code and       [ebx + 18], eax}
     */
    // Template#: 262, Serial#: 126
    public void and(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0013((byte) 0x21, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code and       [ebx + 18], al}
     */
    // Template#: 263, Serial#: 117
    public void and(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0014((byte) 0x20, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andl  }<i>disp8</i>, <i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code andl      [ebx + 18], 0x12345678}
     */
    // Template#: 264, Serial#: 382
    public void andl(byte disp8, IA32IndirectRegister32 destination, int imm32) {
        assemble0015((byte) 0x81, (byte) 0x04, disp8, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code andw  }<i>disp8</i>, <i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code andw      [ebx + 18], 0x1234}
     */
    // Template#: 265, Serial#: 454
    public void andw(byte disp8, IA32IndirectRegister32 destination, short imm16) {
        assemble0016((byte) 0x81, (byte) 0x04, disp8, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code andb  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andb      ebx[esi * 4], 0x12}
     */
    // Template#: 266, Serial#: 287
    public void andb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0x80, (byte) 0x04, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code andl  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andl      ebx[esi * 4], 0x12}
     */
    // Template#: 267, Serial#: 503
    public void andl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0x83, (byte) 0x04, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code andw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andw      ebx[esi * 4], 0x12}
     */
    // Template#: 268, Serial#: 575
    public void andw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0018((byte) 0x83, (byte) 0x04, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code and       ebx[esi * 4], ax}
     */
    // Template#: 269, Serial#: 132
    public void and(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0019((byte) 0x21, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code and       ebx[esi * 4], eax}
     */
    // Template#: 270, Serial#: 123
    public void and(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0020((byte) 0x21, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code and       ebx[esi * 4], al}
     */
    // Template#: 271, Serial#: 114
    public void and(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0021((byte) 0x20, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andl  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code andl      ebx[esi * 4], 0x12345678}
     */
    // Template#: 272, Serial#: 359
    public void andl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0022((byte) 0x81, (byte) 0x04, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code andw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code andw      ebx[esi * 4], 0x1234}
     */
    // Template#: 273, Serial#: 431
    public void andw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0023((byte) 0x81, (byte) 0x04, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code andw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andw      ax, 0x12}
     */
    // Template#: 274, Serial#: 626
    public void andw(IA32GeneralRegister16 destination, byte imm8) {
        assemble0024((byte) 0x83, (byte) 0x04, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code and       ax, ebx[esi * 4 + 18]}
     */
    // Template#: 275, Serial#: 161
    public void and(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0025((byte) 0x23, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code and       ax, [ebx + 18]}
     */
    // Template#: 276, Serial#: 160
    public void and(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0026((byte) 0x23, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code and       ax, ebx[esi * 4]}
     */
    // Template#: 277, Serial#: 157
    public void and(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0027((byte) 0x23, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code and       ax, ax}
     */
    // Template#: 278, Serial#: 139
    public void and(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0028((byte) 0x21, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code and       ax, [ebx]}
     */
    // Template#: 279, Serial#: 156
    public void and(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0029((byte) 0x23, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code and       ax, [L1: 0x12345678]}
     */
    // Template#: 280, Serial#: 159
    public void m_and(IA32GeneralRegister16 destination, int m32) {
        assemble0030((byte) 0x23, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code and       ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 281, Serial#: 163
    public void and(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0031((byte) 0x23, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code and       ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 282, Serial#: 158
    public void m_and(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0032((byte) 0x23, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code and       ax, [ebx + 305419896]}
     */
    // Template#: 283, Serial#: 162
    public void and(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0033((byte) 0x23, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andw  }<i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code andw      ax, 0x1234}
     */
    // Template#: 284, Serial#: 482
    public void andw(IA32GeneralRegister16 destination, short imm16) {
        assemble0034((byte) 0x81, (byte) 0x04, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code andl  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andl      eax, 0x12}
     */
    // Template#: 285, Serial#: 554
    public void andl(IA32GeneralRegister32 destination, byte imm8) {
        assemble0035((byte) 0x83, (byte) 0x04, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code and       eax, ebx[esi * 4 + 18]}
     */
    // Template#: 286, Serial#: 153
    public void and(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0036((byte) 0x23, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code and       eax, [ebx + 18]}
     */
    // Template#: 287, Serial#: 152
    public void and(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0037((byte) 0x23, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code and       eax, ebx[esi * 4]}
     */
    // Template#: 288, Serial#: 149
    public void and(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0038((byte) 0x23, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code and       eax, eax}
     */
    // Template#: 289, Serial#: 130
    public void and(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0039((byte) 0x21, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code and       eax, [ebx]}
     */
    // Template#: 290, Serial#: 148
    public void and(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0040((byte) 0x23, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code and       eax, [L1: 0x12345678]}
     */
    // Template#: 291, Serial#: 151
    public void m_and(IA32GeneralRegister32 destination, int m32) {
        assemble0041((byte) 0x23, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code andl  }<i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code andl      eax, 0x12345678}
     */
    // Template#: 292, Serial#: 410
    public void andl(IA32GeneralRegister32 destination, int imm32) {
        assemble0042((byte) 0x81, (byte) 0x04, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code and       eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 293, Serial#: 155
    public void and(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0043((byte) 0x23, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code and       eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 294, Serial#: 150
    public void m_and(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0044((byte) 0x23, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code and       eax, [ebx + 305419896]}
     */
    // Template#: 295, Serial#: 154
    public void and(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0045((byte) 0x23, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andb      al, 0x12}
     */
    // Template#: 296, Serial#: 338
    public void andb(IA32GeneralRegister8 destination, byte imm8) {
        assemble0046((byte) 0x80, (byte) 0x04, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code and       al, ebx[esi * 4 + 18]}
     */
    // Template#: 297, Serial#: 145
    public void and(IA32GeneralRegister8 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0047((byte) 0x22, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code and       al, [ebx + 18]}
     */
    // Template#: 298, Serial#: 144
    public void and(IA32GeneralRegister8 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0048((byte) 0x22, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code and       al, ebx[esi * 4]}
     */
    // Template#: 299, Serial#: 141
    public void and(IA32GeneralRegister8 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0049((byte) 0x22, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code and       al, al}
     */
    // Template#: 300, Serial#: 121
    public void and(IA32GeneralRegister8 destination, IA32GeneralRegister8 source) {
        assemble0050((byte) 0x20, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code and       al, [ebx]}
     */
    // Template#: 301, Serial#: 140
    public void and(IA32GeneralRegister8 destination, IA32IndirectRegister32 source) {
        assemble0051((byte) 0x22, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code and       al, [L1: 0x12345678]}
     */
    // Template#: 302, Serial#: 143
    public void m_and(IA32GeneralRegister8 destination, int m32) {
        assemble0052((byte) 0x22, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code and       al, ebx[esi * 4 + 305419896]}
     */
    // Template#: 303, Serial#: 147
    public void and(IA32GeneralRegister8 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0053((byte) 0x22, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code and       al, L1: 0x12345678[esi * 4]}
     */
    // Template#: 304, Serial#: 142
    public void m_and(IA32GeneralRegister8 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0054((byte) 0x22, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code and       al, [ebx + 305419896]}
     */
    // Template#: 305, Serial#: 146
    public void and(IA32GeneralRegister8 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0055((byte) 0x22, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andb      [ebx], 0x12}
     */
    // Template#: 306, Serial#: 286
    public void andb(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0x80, (byte) 0x04, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code andl  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andl      [ebx], 0x12}
     */
    // Template#: 307, Serial#: 502
    public void andl(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0x83, (byte) 0x04, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code andw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andw      [ebx], 0x12}
     */
    // Template#: 308, Serial#: 574
    public void andw(IA32IndirectRegister32 destination, byte imm8) {
        assemble0057((byte) 0x83, (byte) 0x04, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code and       [ebx], ax}
     */
    // Template#: 309, Serial#: 131
    public void and(IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0058((byte) 0x21, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code and       [ebx], eax}
     */
    // Template#: 310, Serial#: 122
    public void and(IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0059((byte) 0x21, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code and       [ebx], al}
     */
    // Template#: 311, Serial#: 113
    public void and(IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0060((byte) 0x20, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andl  }<i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code andl      [ebx], 0x12345678}
     */
    // Template#: 312, Serial#: 358
    public void andl(IA32IndirectRegister32 destination, int imm32) {
        assemble0061((byte) 0x81, (byte) 0x04, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code andw  }<i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code andw      [ebx], 0x1234}
     */
    // Template#: 313, Serial#: 430
    public void andw(IA32IndirectRegister32 destination, short imm16) {
        assemble0062((byte) 0x81, (byte) 0x04, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code andl  }<i>imm32</i>
     * Example disassembly syntax: {@code andl      eax, 0x12345678}
     */
    // Template#: 314, Serial#: 165
    public void and_EAX(int imm32) {
        assemble0063((byte) 0x25, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code andb  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andb      [L1: 0x12345678], 0x12}
     */
    // Template#: 315, Serial#: 289
    public void m_andb(int m32, byte imm8) {
        assemble0064((byte) 0x80, (byte) 0x04, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code andl  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andl      [L1: 0x12345678], 0x12}
     */
    // Template#: 316, Serial#: 505
    public void m_andl(int m32, byte imm8) {
        assemble0064((byte) 0x83, (byte) 0x04, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code andw  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andw      [L1: 0x12345678], 0x12}
     */
    // Template#: 317, Serial#: 577
    public void m_andw(int m32, byte imm8) {
        assemble0065((byte) 0x83, (byte) 0x04, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code andb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andb      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 318, Serial#: 327
    public void andb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0x80, (byte) 0x04, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code andl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andl      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 319, Serial#: 543
    public void andl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0x83, (byte) 0x04, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code andw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andw      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 320, Serial#: 615
    public void andw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0067((byte) 0x83, (byte) 0x04, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code and       ebx[esi * 4 + 305419896], ax}
     */
    // Template#: 321, Serial#: 138
    public void and(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0068((byte) 0x21, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code and       ebx[esi * 4 + 305419896], eax}
     */
    // Template#: 322, Serial#: 129
    public void and(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0069((byte) 0x21, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code and       ebx[esi * 4 + 305419896], al}
     */
    // Template#: 323, Serial#: 120
    public void and(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0070((byte) 0x20, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code andl      ebx[esi * 4 + 305419896], 0x12345678}
     */
    // Template#: 324, Serial#: 399
    public void andl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0071((byte) 0x81, (byte) 0x04, disp32, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code andw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code andw      ebx[esi * 4 + 305419896], 0x1234}
     */
    // Template#: 325, Serial#: 471
    public void andw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0072((byte) 0x81, (byte) 0x04, disp32, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code and       [L1: 0x12345678], ax}
     */
    // Template#: 326, Serial#: 134
    public void m_and(int m32, IA32GeneralRegister16 source) {
        assemble0073((byte) 0x21, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code and       [L1: 0x12345678], eax}
     */
    // Template#: 327, Serial#: 125
    public void m_and(int m32, IA32GeneralRegister32 source) {
        assemble0074((byte) 0x21, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code and       [L1: 0x12345678], al}
     */
    // Template#: 328, Serial#: 116
    public void m_and(int m32, IA32GeneralRegister8 source) {
        assemble0075((byte) 0x20, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andb  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andb      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 329, Serial#: 288
    public void m_andb(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0x80, (byte) 0x04, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code andl  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andl      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 330, Serial#: 504
    public void m_andl(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0x83, (byte) 0x04, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code andw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andw      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 331, Serial#: 576
    public void m_andw(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0077((byte) 0x83, (byte) 0x04, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code and       L1: 0x12345678[esi * 4], ax}
     */
    // Template#: 332, Serial#: 133
    public void m_and(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0078((byte) 0x21, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code and       L1: 0x12345678[esi * 4], eax}
     */
    // Template#: 333, Serial#: 124
    public void m_and(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0079((byte) 0x21, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code and       L1: 0x12345678[esi * 4], al}
     */
    // Template#: 334, Serial#: 115
    public void m_and(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0080((byte) 0x20, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andl  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code andl      L1: 0x12345678[esi * 4], 0x12345678}
     */
    // Template#: 335, Serial#: 360
    public void m_andl(int m32, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0081((byte) 0x81, (byte) 0x04, m32, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code andw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code andw      L1: 0x12345678[esi * 4], 0x1234}
     */
    // Template#: 336, Serial#: 432
    public void m_andw(int m32, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0082((byte) 0x81, (byte) 0x04, m32, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code andb  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andb      [ebx + 305419896], 0x12}
     */
    // Template#: 337, Serial#: 326
    public void andb(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0x80, (byte) 0x04, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code andl  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andl      [ebx + 305419896], 0x12}
     */
    // Template#: 338, Serial#: 542
    public void andl(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0x83, (byte) 0x04, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code andw  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code andw      [ebx + 305419896], 0x12}
     */
    // Template#: 339, Serial#: 614
    public void andw(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0084((byte) 0x83, (byte) 0x04, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code and       [ebx + 305419896], ax}
     */
    // Template#: 340, Serial#: 137
    public void and(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0085((byte) 0x21, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code and       [ebx + 305419896], eax}
     */
    // Template#: 341, Serial#: 128
    public void and(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0086((byte) 0x21, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code and  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code and       [ebx + 305419896], al}
     */
    // Template#: 342, Serial#: 119
    public void and(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0087((byte) 0x20, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andl  }<i>disp32</i>, <i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code andl      [ebx + 305419896], 0x12345678}
     */
    // Template#: 343, Serial#: 398
    public void andl(int disp32, IA32IndirectRegister32 destination, int imm32) {
        assemble0088((byte) 0x81, (byte) 0x04, disp32, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code andw  }<i>disp32</i>, <i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code andw      [ebx + 305419896], 0x1234}
     */
    // Template#: 344, Serial#: 470
    public void andw(int disp32, IA32IndirectRegister32 destination, short imm16) {
        assemble0089((byte) 0x81, (byte) 0x04, disp32, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code andl  }<i>m32</i>, <i>imm32</i>
     * Example disassembly syntax: {@code andl      [L1: 0x12345678], 0x12345678}
     */
    // Template#: 345, Serial#: 361
    public void m_andl(int m32, int imm32) {
        assemble0090((byte) 0x81, (byte) 0x04, m32, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code andw  }<i>m32</i>, <i>imm16</i>
     * Example disassembly syntax: {@code andw      [L1: 0x12345678], 0x1234}
     */
    // Template#: 346, Serial#: 433
    public void m_andw(int m32, short imm16) {
        assemble0091((byte) 0x81, (byte) 0x04, m32, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code andw  }<i>imm16</i>
     * Example disassembly syntax: {@code andw      ax, 0x1234}
     */
    // Template#: 347, Serial#: 166
    public void and_AX(short imm16) {
        assemble0092((byte) 0x25, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code andnpd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code andnpd    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 348, Serial#: 3069
    public void andnpd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x55, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code andnpd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code andnpd    xmm0, [ebx + 18]}
     */
    // Template#: 349, Serial#: 3068
    public void andnpd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x55, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andnpd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code andnpd    xmm0, ebx[esi * 4]}
     */
    // Template#: 350, Serial#: 3065
    public void andnpd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x55, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code andnpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code andnpd    xmm0, [ebx]}
     */
    // Template#: 351, Serial#: 3064
    public void andnpd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x55, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andnpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code andnpd    xmm0, xmm0}
     */
    // Template#: 352, Serial#: 3072
    public void andnpd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x55, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andnpd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code andnpd    xmm0, [L1: 0x12345678]}
     */
    // Template#: 353, Serial#: 3067
    public void m_andnpd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x55, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code andnpd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code andnpd    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 354, Serial#: 3071
    public void andnpd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x55, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code andnpd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code andnpd    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 355, Serial#: 3066
    public void m_andnpd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x55, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code andnpd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code andnpd    xmm0, [ebx + 305419896]}
     */
    // Template#: 356, Serial#: 3070
    public void andnpd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x55, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andnps  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code andnps    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 357, Serial#: 3023
    public void andnps(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0102((byte) 0x55, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code andnps  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code andnps    xmm0, [ebx + 18]}
     */
    // Template#: 358, Serial#: 3022
    public void andnps(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0103((byte) 0x55, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andnps  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code andnps    xmm0, ebx[esi * 4]}
     */
    // Template#: 359, Serial#: 3019
    public void andnps(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0104((byte) 0x55, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code andnps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code andnps    xmm0, [ebx]}
     */
    // Template#: 360, Serial#: 3018
    public void andnps(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0105((byte) 0x55, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andnps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code andnps    xmm0, xmm0}
     */
    // Template#: 361, Serial#: 3026
    public void andnps(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0106((byte) 0x55, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andnps  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code andnps    xmm0, [L1: 0x12345678]}
     */
    // Template#: 362, Serial#: 3021
    public void m_andnps(IA32XMMRegister destination, int m32) {
        assemble0107((byte) 0x55, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code andnps  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code andnps    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 363, Serial#: 3025
    public void andnps(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0108((byte) 0x55, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code andnps  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code andnps    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 364, Serial#: 3020
    public void m_andnps(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0109((byte) 0x55, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code andnps  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code andnps    xmm0, [ebx + 305419896]}
     */
    // Template#: 365, Serial#: 3024
    public void andnps(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0110((byte) 0x55, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andpd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code andpd     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 366, Serial#: 3060
    public void andpd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x54, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code andpd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code andpd     xmm0, [ebx + 18]}
     */
    // Template#: 367, Serial#: 3059
    public void andpd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x54, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andpd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code andpd     xmm0, ebx[esi * 4]}
     */
    // Template#: 368, Serial#: 3056
    public void andpd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x54, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code andpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code andpd     xmm0, [ebx]}
     */
    // Template#: 369, Serial#: 3055
    public void andpd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x54, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code andpd     xmm0, xmm0}
     */
    // Template#: 370, Serial#: 3063
    public void andpd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x54, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andpd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code andpd     xmm0, [L1: 0x12345678]}
     */
    // Template#: 371, Serial#: 3058
    public void m_andpd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x54, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code andpd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code andpd     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 372, Serial#: 3062
    public void andpd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x54, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code andpd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code andpd     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 373, Serial#: 3057
    public void m_andpd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x54, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code andpd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code andpd     xmm0, [ebx + 305419896]}
     */
    // Template#: 374, Serial#: 3061
    public void andpd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x54, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andps  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code andps     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 375, Serial#: 3014
    public void andps(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0102((byte) 0x54, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code andps  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code andps     xmm0, [ebx + 18]}
     */
    // Template#: 376, Serial#: 3013
    public void andps(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0103((byte) 0x54, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andps  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code andps     xmm0, ebx[esi * 4]}
     */
    // Template#: 377, Serial#: 3010
    public void andps(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0104((byte) 0x54, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code andps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code andps     xmm0, [ebx]}
     */
    // Template#: 378, Serial#: 3009
    public void andps(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0105((byte) 0x54, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code andps     xmm0, xmm0}
     */
    // Template#: 379, Serial#: 3017
    public void andps(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0106((byte) 0x54, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code andps  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code andps     xmm0, [L1: 0x12345678]}
     */
    // Template#: 380, Serial#: 3012
    public void m_andps(IA32XMMRegister destination, int m32) {
        assemble0107((byte) 0x54, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code andps  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code andps     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 381, Serial#: 3016
    public void andps(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0108((byte) 0x54, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code andps  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code andps     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 382, Serial#: 3011
    public void m_andps(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0109((byte) 0x54, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code andps  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code andps     xmm0, [ebx + 305419896]}
     */
    // Template#: 383, Serial#: 3015
    public void andps(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0110((byte) 0x54, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code arpl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code arpl      ebx[esi * 4 + 18], ax}
     */
    // Template#: 384, Serial#: 254
    public void arpl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0129((byte) 0x63, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code arpl  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code arpl      [ebx + 18], ax}
     */
    // Template#: 385, Serial#: 253
    public void arpl(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0130((byte) 0x63, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code arpl  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code arpl      ebx[esi * 4], ax}
     */
    // Template#: 386, Serial#: 250
    public void arpl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0131((byte) 0x63, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code arpl  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code arpl      ax, ax}
     */
    // Template#: 387, Serial#: 257
    public void arpl(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0132((byte) 0x63, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code arpl  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code arpl      [ebx], ax}
     */
    // Template#: 388, Serial#: 249
    public void arpl(IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0133((byte) 0x63, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code arpl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code arpl      ebx[esi * 4 + 305419896], ax}
     */
    // Template#: 389, Serial#: 256
    public void arpl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0134((byte) 0x63, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code arpl  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code arpl      [L1: 0x12345678], ax}
     */
    // Template#: 390, Serial#: 252
    public void m_arpl(int m32, IA32GeneralRegister16 source) {
        assemble0135((byte) 0x63, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code arpl  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code arpl      L1: 0x12345678[esi * 4], ax}
     */
    // Template#: 391, Serial#: 251
    public void m_arpl(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0136((byte) 0x63, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code arpl  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code arpl      [ebx + 305419896], ax}
     */
    // Template#: 392, Serial#: 255
    public void arpl(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0137((byte) 0x63, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bound  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code bound     ax, ebx[esi * 4 + 18]}
     */
    // Template#: 393, Serial#: 246
    public void bound(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0025((byte) 0x62, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code bound  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code bound     ax, [ebx + 18]}
     */
    // Template#: 394, Serial#: 245
    public void bound(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0026((byte) 0x62, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bound  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code bound     ax, ebx[esi * 4]}
     */
    // Template#: 395, Serial#: 242
    public void bound(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0027((byte) 0x62, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code bound  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bound     ax, [ebx]}
     */
    // Template#: 396, Serial#: 241
    public void bound(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0029((byte) 0x62, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bound  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code bound     ax, [L1: 0x12345678]}
     */
    // Template#: 397, Serial#: 244
    public void m_bound(IA32GeneralRegister16 destination, int m32) {
        assemble0030((byte) 0x62, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code bound  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code bound     ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 398, Serial#: 248
    public void bound(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0031((byte) 0x62, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code bound  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code bound     ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 399, Serial#: 243
    public void m_bound(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0032((byte) 0x62, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code bound  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code bound     ax, [ebx + 305419896]}
     */
    // Template#: 400, Serial#: 247
    public void bound(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0033((byte) 0x62, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bound  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code bound     eax, ebx[esi * 4 + 18]}
     */
    // Template#: 401, Serial#: 238
    public void bound(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0036((byte) 0x62, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code bound  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code bound     eax, [ebx + 18]}
     */
    // Template#: 402, Serial#: 237
    public void bound(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0037((byte) 0x62, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bound  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code bound     eax, ebx[esi * 4]}
     */
    // Template#: 403, Serial#: 234
    public void bound(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0038((byte) 0x62, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code bound  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bound     eax, [ebx]}
     */
    // Template#: 404, Serial#: 233
    public void bound(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0040((byte) 0x62, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bound  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code bound     eax, [L1: 0x12345678]}
     */
    // Template#: 405, Serial#: 236
    public void m_bound(IA32GeneralRegister32 destination, int m32) {
        assemble0041((byte) 0x62, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code bound  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code bound     eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 406, Serial#: 240
    public void bound(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0043((byte) 0x62, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code bound  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code bound     eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 407, Serial#: 235
    public void m_bound(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0044((byte) 0x62, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code bound  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code bound     eax, [ebx + 305419896]}
     */
    // Template#: 408, Serial#: 239
    public void bound(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0045((byte) 0x62, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsf  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code bsf       ax, ebx[esi * 4 + 18]}
     */
    // Template#: 409, Serial#: 5100
    public void bsf(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0xBC, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsf  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code bsf       ax, [ebx + 18]}
     */
    // Template#: 410, Serial#: 5099
    public void bsf(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0xBC, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsf  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code bsf       ax, ebx[esi * 4]}
     */
    // Template#: 411, Serial#: 5096
    public void bsf(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0xBC, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsf  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bsf       ax, ax}
     */
    // Template#: 412, Serial#: 5103
    public void bsf(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0141((byte) 0xBC, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsf  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bsf       ax, [ebx]}
     */
    // Template#: 413, Serial#: 5095
    public void bsf(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0xBC, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsf  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code bsf       ax, [L1: 0x12345678]}
     */
    // Template#: 414, Serial#: 5098
    public void m_bsf(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0xBC, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsf  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code bsf       ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 415, Serial#: 5102
    public void bsf(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0xBC, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsf  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code bsf       ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 416, Serial#: 5097
    public void m_bsf(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0xBC, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsf  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code bsf       ax, [ebx + 305419896]}
     */
    // Template#: 417, Serial#: 5101
    public void bsf(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0xBC, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsf  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code bsf       eax, ebx[esi * 4 + 18]}
     */
    // Template#: 418, Serial#: 5091
    public void bsf(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0xBC, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsf  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code bsf       eax, [ebx + 18]}
     */
    // Template#: 419, Serial#: 5090
    public void bsf(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0xBC, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsf  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code bsf       eax, ebx[esi * 4]}
     */
    // Template#: 420, Serial#: 5087
    public void bsf(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0xBC, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsf  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bsf       eax, eax}
     */
    // Template#: 421, Serial#: 5094
    public void bsf(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0150((byte) 0xBC, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsf  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bsf       eax, [ebx]}
     */
    // Template#: 422, Serial#: 5086
    public void bsf(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0xBC, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsf  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code bsf       eax, [L1: 0x12345678]}
     */
    // Template#: 423, Serial#: 5089
    public void m_bsf(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0xBC, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsf  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code bsf       eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 424, Serial#: 5093
    public void bsf(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0xBC, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsf  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code bsf       eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 425, Serial#: 5088
    public void m_bsf(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0xBC, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsf  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code bsf       eax, [ebx + 305419896]}
     */
    // Template#: 426, Serial#: 5092
    public void bsf(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0xBC, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsr  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code bsr       ax, ebx[esi * 4 + 18]}
     */
    // Template#: 427, Serial#: 5118
    public void bsr(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0xBD, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsr  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code bsr       ax, [ebx + 18]}
     */
    // Template#: 428, Serial#: 5117
    public void bsr(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0xBD, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsr  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code bsr       ax, ebx[esi * 4]}
     */
    // Template#: 429, Serial#: 5114
    public void bsr(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0xBD, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsr  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bsr       ax, ax}
     */
    // Template#: 430, Serial#: 5121
    public void bsr(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0141((byte) 0xBD, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsr  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bsr       ax, [ebx]}
     */
    // Template#: 431, Serial#: 5113
    public void bsr(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0xBD, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsr  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code bsr       ax, [L1: 0x12345678]}
     */
    // Template#: 432, Serial#: 5116
    public void m_bsr(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0xBD, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsr  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code bsr       ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 433, Serial#: 5120
    public void bsr(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0xBD, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsr  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code bsr       ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 434, Serial#: 5115
    public void m_bsr(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0xBD, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsr  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code bsr       ax, [ebx + 305419896]}
     */
    // Template#: 435, Serial#: 5119
    public void bsr(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0xBD, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsr  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code bsr       eax, ebx[esi * 4 + 18]}
     */
    // Template#: 436, Serial#: 5109
    public void bsr(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0xBD, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsr  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code bsr       eax, [ebx + 18]}
     */
    // Template#: 437, Serial#: 5108
    public void bsr(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0xBD, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsr  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code bsr       eax, ebx[esi * 4]}
     */
    // Template#: 438, Serial#: 5105
    public void bsr(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0xBD, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsr  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bsr       eax, eax}
     */
    // Template#: 439, Serial#: 5112
    public void bsr(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0150((byte) 0xBD, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsr  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bsr       eax, [ebx]}
     */
    // Template#: 440, Serial#: 5104
    public void bsr(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0xBD, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsr  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code bsr       eax, [L1: 0x12345678]}
     */
    // Template#: 441, Serial#: 5107
    public void m_bsr(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0xBD, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsr  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code bsr       eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 442, Serial#: 5111
    public void bsr(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0xBD, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsr  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code bsr       eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 443, Serial#: 5106
    public void m_bsr(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0xBD, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code bsr  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code bsr       eax, [ebx + 305419896]}
     */
    // Template#: 444, Serial#: 5110
    public void bsr(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0xBD, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bswap  }<i>st_i</i>
     * Example disassembly syntax: {@code bswap     eax}
     */
    // Template#: 445, Serial#: 5158
    public void bswap(IA32GeneralRegister32 st_i) {
        assemble0156((byte) 0xC8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code bt        ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 446, Serial#: 5045
    public void bt(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0157((byte) 0xBA, (byte) 0x04, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code bt        ebx[esi * 4 + 18], ax}
     */
    // Template#: 447, Serial#: 3426
    public void bt(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0158((byte) 0xA3, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code bt        ebx[esi * 4 + 18], eax}
     */
    // Template#: 448, Serial#: 3417
    public void bt(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0159((byte) 0xA3, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code bt        [ebx + 18], 0x12}
     */
    // Template#: 449, Serial#: 5044
    public void bt(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0160((byte) 0xBA, (byte) 0x04, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bt        [ebx + 18], ax}
     */
    // Template#: 450, Serial#: 3425
    public void bt(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0161((byte) 0xA3, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bt        [ebx + 18], eax}
     */
    // Template#: 451, Serial#: 3416
    public void bt(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0162((byte) 0xA3, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code bt        ebx[esi * 4], 0x12}
     */
    // Template#: 452, Serial#: 5029
    public void bt(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0163((byte) 0xBA, (byte) 0x04, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code bt        ebx[esi * 4], ax}
     */
    // Template#: 453, Serial#: 3422
    public void bt(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0164((byte) 0xA3, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code bt        ebx[esi * 4], eax}
     */
    // Template#: 454, Serial#: 3413
    public void bt(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0165((byte) 0xA3, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code bt        ax, 0x12}
     */
    // Template#: 455, Serial#: 5064
    public void bt(IA32GeneralRegister16 destination, byte imm8) {
        assemble0166((byte) 0xBA, (byte) 0x04, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bt        ax, ax}
     */
    // Template#: 456, Serial#: 3429
    public void bt(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0167((byte) 0xA3, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code bt        eax, 0x12}
     */
    // Template#: 457, Serial#: 5060
    public void bt(IA32GeneralRegister32 destination, byte imm8) {
        assemble0168((byte) 0xBA, (byte) 0x04, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bt        eax, eax}
     */
    // Template#: 458, Serial#: 3420
    public void bt(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0169((byte) 0xA3, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code bt        [ebx], 0x12}
     */
    // Template#: 459, Serial#: 5028
    public void bt(IA32IndirectRegister32 destination, byte imm8) {
        assemble0170((byte) 0xBA, (byte) 0x04, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bt        [ebx], ax}
     */
    // Template#: 460, Serial#: 3421
    public void bt(IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0171((byte) 0xA3, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bt        [ebx], eax}
     */
    // Template#: 461, Serial#: 3412
    public void bt(IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0172((byte) 0xA3, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code bt        [L1: 0x12345678], 0x12}
     */
    // Template#: 462, Serial#: 5031
    public void m_bt(int m32, byte imm8) {
        assemble0173((byte) 0xBA, (byte) 0x04, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code bt        ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 463, Serial#: 5053
    public void bt(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0174((byte) 0xBA, (byte) 0x04, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code bt        ebx[esi * 4 + 305419896], ax}
     */
    // Template#: 464, Serial#: 3428
    public void bt(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0175((byte) 0xA3, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code bt        ebx[esi * 4 + 305419896], eax}
     */
    // Template#: 465, Serial#: 3419
    public void bt(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0176((byte) 0xA3, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code bt        [L1: 0x12345678], ax}
     */
    // Template#: 466, Serial#: 3424
    public void m_bt(int m32, IA32GeneralRegister16 source) {
        assemble0177((byte) 0xA3, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code bt        [L1: 0x12345678], eax}
     */
    // Template#: 467, Serial#: 3415
    public void m_bt(int m32, IA32GeneralRegister32 source) {
        assemble0178((byte) 0xA3, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code bt        L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 468, Serial#: 5030
    public void m_bt(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0179((byte) 0xBA, (byte) 0x04, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code bt        L1: 0x12345678[esi * 4], ax}
     */
    // Template#: 469, Serial#: 3423
    public void m_bt(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0180((byte) 0xA3, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code bt        L1: 0x12345678[esi * 4], eax}
     */
    // Template#: 470, Serial#: 3414
    public void m_bt(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0181((byte) 0xA3, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code bt        [ebx + 305419896], 0x12}
     */
    // Template#: 471, Serial#: 5052
    public void bt(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0182((byte) 0xBA, (byte) 0x04, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bt        [ebx + 305419896], ax}
     */
    // Template#: 472, Serial#: 3427
    public void bt(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0183((byte) 0xA3, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bt  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bt        [ebx + 305419896], eax}
     */
    // Template#: 473, Serial#: 3418
    public void bt(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0184((byte) 0xA3, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code btc       ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 474, Serial#: 5051
    public void btc(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0157((byte) 0xBA, (byte) 0x07, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code btc       ebx[esi * 4 + 18], ax}
     */
    // Template#: 475, Serial#: 5082
    public void btc(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0158((byte) 0xBB, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code btc       ebx[esi * 4 + 18], eax}
     */
    // Template#: 476, Serial#: 5073
    public void btc(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0159((byte) 0xBB, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code btc       [ebx + 18], 0x12}
     */
    // Template#: 477, Serial#: 5050
    public void btc(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0160((byte) 0xBA, (byte) 0x07, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code btc       [ebx + 18], ax}
     */
    // Template#: 478, Serial#: 5081
    public void btc(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0161((byte) 0xBB, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code btc       [ebx + 18], eax}
     */
    // Template#: 479, Serial#: 5072
    public void btc(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0162((byte) 0xBB, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code btc       ebx[esi * 4], 0x12}
     */
    // Template#: 480, Serial#: 5041
    public void btc(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0163((byte) 0xBA, (byte) 0x07, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code btc       ebx[esi * 4], ax}
     */
    // Template#: 481, Serial#: 5078
    public void btc(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0164((byte) 0xBB, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code btc       ebx[esi * 4], eax}
     */
    // Template#: 482, Serial#: 5069
    public void btc(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0165((byte) 0xBB, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code btc       ax, 0x12}
     */
    // Template#: 483, Serial#: 5067
    public void btc(IA32GeneralRegister16 destination, byte imm8) {
        assemble0166((byte) 0xBA, (byte) 0x07, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code btc       ax, ax}
     */
    // Template#: 484, Serial#: 5085
    public void btc(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0167((byte) 0xBB, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code btc       eax, 0x12}
     */
    // Template#: 485, Serial#: 5063
    public void btc(IA32GeneralRegister32 destination, byte imm8) {
        assemble0168((byte) 0xBA, (byte) 0x07, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code btc       eax, eax}
     */
    // Template#: 486, Serial#: 5076
    public void btc(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0169((byte) 0xBB, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code btc       [ebx], 0x12}
     */
    // Template#: 487, Serial#: 5040
    public void btc(IA32IndirectRegister32 destination, byte imm8) {
        assemble0170((byte) 0xBA, (byte) 0x07, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code btc       [ebx], ax}
     */
    // Template#: 488, Serial#: 5077
    public void btc(IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0171((byte) 0xBB, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code btc       [ebx], eax}
     */
    // Template#: 489, Serial#: 5068
    public void btc(IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0172((byte) 0xBB, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code btc       [L1: 0x12345678], 0x12}
     */
    // Template#: 490, Serial#: 5043
    public void m_btc(int m32, byte imm8) {
        assemble0173((byte) 0xBA, (byte) 0x07, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code btc       ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 491, Serial#: 5059
    public void btc(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0174((byte) 0xBA, (byte) 0x07, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code btc       ebx[esi * 4 + 305419896], ax}
     */
    // Template#: 492, Serial#: 5084
    public void btc(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0175((byte) 0xBB, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code btc       ebx[esi * 4 + 305419896], eax}
     */
    // Template#: 493, Serial#: 5075
    public void btc(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0176((byte) 0xBB, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code btc       [L1: 0x12345678], ax}
     */
    // Template#: 494, Serial#: 5080
    public void m_btc(int m32, IA32GeneralRegister16 source) {
        assemble0177((byte) 0xBB, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code btc       [L1: 0x12345678], eax}
     */
    // Template#: 495, Serial#: 5071
    public void m_btc(int m32, IA32GeneralRegister32 source) {
        assemble0178((byte) 0xBB, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code btc       L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 496, Serial#: 5042
    public void m_btc(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0179((byte) 0xBA, (byte) 0x07, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code btc       L1: 0x12345678[esi * 4], ax}
     */
    // Template#: 497, Serial#: 5079
    public void m_btc(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0180((byte) 0xBB, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code btc       L1: 0x12345678[esi * 4], eax}
     */
    // Template#: 498, Serial#: 5070
    public void m_btc(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0181((byte) 0xBB, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code btc       [ebx + 305419896], 0x12}
     */
    // Template#: 499, Serial#: 5058
    public void btc(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0182((byte) 0xBA, (byte) 0x07, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code btc       [ebx + 305419896], ax}
     */
    // Template#: 500, Serial#: 5083
    public void btc(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0183((byte) 0xBB, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btc  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code btc       [ebx + 305419896], eax}
     */
    // Template#: 501, Serial#: 5074
    public void btc(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0184((byte) 0xBB, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code btr       ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 502, Serial#: 5049
    public void btr(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0157((byte) 0xBA, (byte) 0x06, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code btr       ebx[esi * 4 + 18], ax}
     */
    // Template#: 503, Serial#: 3523
    public void btr(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0158((byte) 0xB3, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code btr       ebx[esi * 4 + 18], eax}
     */
    // Template#: 504, Serial#: 3514
    public void btr(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0159((byte) 0xB3, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code btr       [ebx + 18], 0x12}
     */
    // Template#: 505, Serial#: 5048
    public void btr(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0160((byte) 0xBA, (byte) 0x06, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code btr       [ebx + 18], ax}
     */
    // Template#: 506, Serial#: 3522
    public void btr(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0161((byte) 0xB3, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code btr       [ebx + 18], eax}
     */
    // Template#: 507, Serial#: 3513
    public void btr(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0162((byte) 0xB3, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code btr       ebx[esi * 4], 0x12}
     */
    // Template#: 508, Serial#: 5037
    public void btr(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0163((byte) 0xBA, (byte) 0x06, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code btr       ebx[esi * 4], ax}
     */
    // Template#: 509, Serial#: 3519
    public void btr(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0164((byte) 0xB3, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code btr       ebx[esi * 4], eax}
     */
    // Template#: 510, Serial#: 3510
    public void btr(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0165((byte) 0xB3, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code btr       ax, 0x12}
     */
    // Template#: 511, Serial#: 5066
    public void btr(IA32GeneralRegister16 destination, byte imm8) {
        assemble0166((byte) 0xBA, (byte) 0x06, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code btr       ax, ax}
     */
    // Template#: 512, Serial#: 3526
    public void btr(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0167((byte) 0xB3, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code btr       eax, 0x12}
     */
    // Template#: 513, Serial#: 5062
    public void btr(IA32GeneralRegister32 destination, byte imm8) {
        assemble0168((byte) 0xBA, (byte) 0x06, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code btr       eax, eax}
     */
    // Template#: 514, Serial#: 3517
    public void btr(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0169((byte) 0xB3, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code btr       [ebx], 0x12}
     */
    // Template#: 515, Serial#: 5036
    public void btr(IA32IndirectRegister32 destination, byte imm8) {
        assemble0170((byte) 0xBA, (byte) 0x06, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code btr       [ebx], ax}
     */
    // Template#: 516, Serial#: 3518
    public void btr(IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0171((byte) 0xB3, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code btr       [ebx], eax}
     */
    // Template#: 517, Serial#: 3509
    public void btr(IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0172((byte) 0xB3, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code btr       [L1: 0x12345678], 0x12}
     */
    // Template#: 518, Serial#: 5039
    public void m_btr(int m32, byte imm8) {
        assemble0173((byte) 0xBA, (byte) 0x06, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code btr       ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 519, Serial#: 5057
    public void btr(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0174((byte) 0xBA, (byte) 0x06, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code btr       ebx[esi * 4 + 305419896], ax}
     */
    // Template#: 520, Serial#: 3525
    public void btr(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0175((byte) 0xB3, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code btr       ebx[esi * 4 + 305419896], eax}
     */
    // Template#: 521, Serial#: 3516
    public void btr(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0176((byte) 0xB3, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code btr       [L1: 0x12345678], ax}
     */
    // Template#: 522, Serial#: 3521
    public void m_btr(int m32, IA32GeneralRegister16 source) {
        assemble0177((byte) 0xB3, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code btr       [L1: 0x12345678], eax}
     */
    // Template#: 523, Serial#: 3512
    public void m_btr(int m32, IA32GeneralRegister32 source) {
        assemble0178((byte) 0xB3, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code btr       L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 524, Serial#: 5038
    public void m_btr(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0179((byte) 0xBA, (byte) 0x06, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code btr       L1: 0x12345678[esi * 4], ax}
     */
    // Template#: 525, Serial#: 3520
    public void m_btr(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0180((byte) 0xB3, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code btr       L1: 0x12345678[esi * 4], eax}
     */
    // Template#: 526, Serial#: 3511
    public void m_btr(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0181((byte) 0xB3, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code btr       [ebx + 305419896], 0x12}
     */
    // Template#: 527, Serial#: 5056
    public void btr(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0182((byte) 0xBA, (byte) 0x06, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code btr       [ebx + 305419896], ax}
     */
    // Template#: 528, Serial#: 3524
    public void btr(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0183((byte) 0xB3, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code btr  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code btr       [ebx + 305419896], eax}
     */
    // Template#: 529, Serial#: 3515
    public void btr(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0184((byte) 0xB3, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code bts       ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 530, Serial#: 5047
    public void bts(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0157((byte) 0xBA, (byte) 0x05, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code bts       ebx[esi * 4 + 18], ax}
     */
    // Template#: 531, Serial#: 4967
    public void bts(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0158((byte) 0xAB, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code bts       ebx[esi * 4 + 18], eax}
     */
    // Template#: 532, Serial#: 4958
    public void bts(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0159((byte) 0xAB, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code bts       [ebx + 18], 0x12}
     */
    // Template#: 533, Serial#: 5046
    public void bts(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0160((byte) 0xBA, (byte) 0x05, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bts       [ebx + 18], ax}
     */
    // Template#: 534, Serial#: 4966
    public void bts(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0161((byte) 0xAB, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bts       [ebx + 18], eax}
     */
    // Template#: 535, Serial#: 4957
    public void bts(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0162((byte) 0xAB, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code bts       ebx[esi * 4], 0x12}
     */
    // Template#: 536, Serial#: 5033
    public void bts(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0163((byte) 0xBA, (byte) 0x05, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code bts       ebx[esi * 4], ax}
     */
    // Template#: 537, Serial#: 4963
    public void bts(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0164((byte) 0xAB, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code bts       ebx[esi * 4], eax}
     */
    // Template#: 538, Serial#: 4954
    public void bts(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0165((byte) 0xAB, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code bts       ax, 0x12}
     */
    // Template#: 539, Serial#: 5065
    public void bts(IA32GeneralRegister16 destination, byte imm8) {
        assemble0166((byte) 0xBA, (byte) 0x05, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bts       ax, ax}
     */
    // Template#: 540, Serial#: 4970
    public void bts(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0167((byte) 0xAB, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code bts       eax, 0x12}
     */
    // Template#: 541, Serial#: 5061
    public void bts(IA32GeneralRegister32 destination, byte imm8) {
        assemble0168((byte) 0xBA, (byte) 0x05, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bts       eax, eax}
     */
    // Template#: 542, Serial#: 4961
    public void bts(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0169((byte) 0xAB, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code bts       [ebx], 0x12}
     */
    // Template#: 543, Serial#: 5032
    public void bts(IA32IndirectRegister32 destination, byte imm8) {
        assemble0170((byte) 0xBA, (byte) 0x05, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bts       [ebx], ax}
     */
    // Template#: 544, Serial#: 4962
    public void bts(IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0171((byte) 0xAB, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bts       [ebx], eax}
     */
    // Template#: 545, Serial#: 4953
    public void bts(IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0172((byte) 0xAB, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code bts       [L1: 0x12345678], 0x12}
     */
    // Template#: 546, Serial#: 5035
    public void m_bts(int m32, byte imm8) {
        assemble0173((byte) 0xBA, (byte) 0x05, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code bts       ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 547, Serial#: 5055
    public void bts(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0174((byte) 0xBA, (byte) 0x05, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code bts       ebx[esi * 4 + 305419896], ax}
     */
    // Template#: 548, Serial#: 4969
    public void bts(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0175((byte) 0xAB, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code bts       ebx[esi * 4 + 305419896], eax}
     */
    // Template#: 549, Serial#: 4960
    public void bts(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0176((byte) 0xAB, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code bts       [L1: 0x12345678], ax}
     */
    // Template#: 550, Serial#: 4965
    public void m_bts(int m32, IA32GeneralRegister16 source) {
        assemble0177((byte) 0xAB, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code bts       [L1: 0x12345678], eax}
     */
    // Template#: 551, Serial#: 4956
    public void m_bts(int m32, IA32GeneralRegister32 source) {
        assemble0178((byte) 0xAB, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code bts       L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 552, Serial#: 5034
    public void m_bts(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0179((byte) 0xBA, (byte) 0x05, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code bts       L1: 0x12345678[esi * 4], ax}
     */
    // Template#: 553, Serial#: 4964
    public void m_bts(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0180((byte) 0xAB, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code bts       L1: 0x12345678[esi * 4], eax}
     */
    // Template#: 554, Serial#: 4955
    public void m_bts(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0181((byte) 0xAB, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code bts       [ebx + 305419896], 0x12}
     */
    // Template#: 555, Serial#: 5054
    public void bts(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0182((byte) 0xBA, (byte) 0x05, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bts       [ebx + 305419896], ax}
     */
    // Template#: 556, Serial#: 4968
    public void bts(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0183((byte) 0xAB, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code bts  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code bts       [ebx + 305419896], eax}
     */
    // Template#: 557, Serial#: 4959
    public void bts(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0184((byte) 0xAB, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code call  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code call      ebx[esi * 4 + 18]}
     */
    // Template#: 558, Serial#: 2457
    public void call(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xFF, (byte) 0x02, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code call  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code call      [ebx + 18]}
     */
    // Template#: 559, Serial#: 2456
    public void call(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xFF, (byte) 0x02, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code call  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code call      ebx[esi * 4]}
     */
    // Template#: 560, Serial#: 2443
    public void call(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xFF, (byte) 0x02, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code call  }<i>destination</i>
     * Example disassembly syntax: {@code call      ax}
     */
    // Template#: 561, Serial#: 2494
    public void call(IA32GeneralRegister16 destination) {
        assemble0188((byte) 0xFF, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code call  }<i>destination</i>
     * Example disassembly syntax: {@code call      eax}
     */
    // Template#: 562, Serial#: 2474
    public void call(IA32GeneralRegister32 destination) {
        assemble0189((byte) 0xFF, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code call  }<i>destination</i>
     * Example disassembly syntax: {@code call      [ebx]}
     */
    // Template#: 563, Serial#: 2442
    public void call(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xFF, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code call  }<i>m32</i>
     * Example disassembly syntax: {@code call      [L1: 0x12345678]}
     */
    // Template#: 564, Serial#: 1913
    public void m_call(int m32) {
        assemble0191((byte) 0x9A, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code call  }<i>rel32</i>
     * Example disassembly syntax: {@code call      L1: +305419896}
     */
    // Template#: 565, Serial#: 2399
    public void call(int rel32) {
        assemble0192((byte) 0xE8, rel32);
    }

    /**
     * Pseudo-external assembler syntax: {@code call  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code call      ebx[esi * 4 + 305419896]}
     */
    // Template#: 566, Serial#: 2467
    public void call(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xFF, (byte) 0x02, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code call  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code call      L1: 0x12345678[esi * 4]}
     */
    // Template#: 567, Serial#: 2444
    public void m_call(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xFF, (byte) 0x02, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code call  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code call      [ebx + 305419896]}
     */
    // Template#: 568, Serial#: 2466
    public void call(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xFF, (byte) 0x02, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code call  }<i>rel16</i>
     * Example disassembly syntax: {@code call      L1: +4660}
     */
    // Template#: 569, Serial#: 2400
    public void call(short rel16) {
        assemble0196((byte) 0xE8, rel16);
    }

    /**
     * Pseudo-external assembler syntax: {@code cbw  }
     * Example disassembly syntax: {@code cbw     }
     */
    // Template#: 570, Serial#: 1909
    public void cbw() {
        assemble0197((byte) 0x98);
    }

    /**
     * Pseudo-external assembler syntax: {@code cdq  }
     * Example disassembly syntax: {@code cdq     }
     */
    // Template#: 571, Serial#: 1912
    public void cdq() {
        assemble0001((byte) 0x99);
    }

    /**
     * Pseudo-external assembler syntax: {@code clc  }
     * Example disassembly syntax: {@code clc     }
     */
    // Template#: 572, Serial#: 2410
    public void clc() {
        assemble0001((byte) 0xF8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cld  }
     * Example disassembly syntax: {@code cld     }
     */
    // Template#: 573, Serial#: 2414
    public void cld() {
        assemble0001((byte) 0xFC);
    }

    /**
     * Pseudo-external assembler syntax: {@code cli  }
     * Example disassembly syntax: {@code cli     }
     */
    // Template#: 574, Serial#: 2412
    public void cli() {
        assemble0001((byte) 0xFA);
    }

    /**
     * Pseudo-external assembler syntax: {@code clts  }
     * Example disassembly syntax: {@code clts    }
     */
    // Template#: 575, Serial#: 2647
    public void clts() {
        assemble0198((byte) 0x06);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmc  }
     * Example disassembly syntax: {@code cmc     }
     */
    // Template#: 576, Serial#: 1346
    public void cmc() {
        assemble0001((byte) 0xF5);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmova  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmova     ax, ebx[esi * 4 + 18]}
     */
    // Template#: 577, Serial#: 2977
    public void cmova(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0x47, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmova  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmova     ax, [ebx + 18]}
     */
    // Template#: 578, Serial#: 2976
    public void cmova(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0x47, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmova  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmova     ax, ebx[esi * 4]}
     */
    // Template#: 579, Serial#: 2973
    public void cmova(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0x47, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmova  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmova     ax, ax}
     */
    // Template#: 580, Serial#: 2980
    public void cmova(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0141((byte) 0x47, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmova  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmova     ax, [ebx]}
     */
    // Template#: 581, Serial#: 2972
    public void cmova(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0x47, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmova  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmova     ax, [L1: 0x12345678]}
     */
    // Template#: 582, Serial#: 2975
    public void m_cmova(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0x47, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmova  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmova     ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 583, Serial#: 2979
    public void cmova(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0x47, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmova  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmova     ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 584, Serial#: 2974
    public void m_cmova(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0x47, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmova  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmova     ax, [ebx + 305419896]}
     */
    // Template#: 585, Serial#: 2978
    public void cmova(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0x47, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmova  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmova     eax, ebx[esi * 4 + 18]}
     */
    // Template#: 586, Serial#: 2968
    public void cmova(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0x47, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmova  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmova     eax, [ebx + 18]}
     */
    // Template#: 587, Serial#: 2967
    public void cmova(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0x47, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmova  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmova     eax, ebx[esi * 4]}
     */
    // Template#: 588, Serial#: 2964
    public void cmova(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0x47, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmova  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmova     eax, eax}
     */
    // Template#: 589, Serial#: 2971
    public void cmova(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0150((byte) 0x47, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmova  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmova     eax, [ebx]}
     */
    // Template#: 590, Serial#: 2963
    public void cmova(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0x47, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmova  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmova     eax, [L1: 0x12345678]}
     */
    // Template#: 591, Serial#: 2966
    public void m_cmova(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0x47, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmova  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmova     eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 592, Serial#: 2970
    public void cmova(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0x47, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmova  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmova     eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 593, Serial#: 2965
    public void m_cmova(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0x47, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmova  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmova     eax, [ebx + 305419896]}
     */
    // Template#: 594, Serial#: 2969
    public void cmova(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0x47, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovae  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovae    ax, ebx[esi * 4 + 18]}
     */
    // Template#: 595, Serial#: 2905
    public void cmovae(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0x43, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovae  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovae    ax, [ebx + 18]}
     */
    // Template#: 596, Serial#: 2904
    public void cmovae(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0x43, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovae  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovae    ax, ebx[esi * 4]}
     */
    // Template#: 597, Serial#: 2901
    public void cmovae(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0x43, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovae  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovae    ax, ax}
     */
    // Template#: 598, Serial#: 2908
    public void cmovae(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0141((byte) 0x43, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovae  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovae    ax, [ebx]}
     */
    // Template#: 599, Serial#: 2900
    public void cmovae(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0x43, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovae  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovae    ax, [L1: 0x12345678]}
     */
    // Template#: 600, Serial#: 2903
    public void m_cmovae(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0x43, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovae  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovae    ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 601, Serial#: 2907
    public void cmovae(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0x43, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovae  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovae    ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 602, Serial#: 2902
    public void m_cmovae(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0x43, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovae  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovae    ax, [ebx + 305419896]}
     */
    // Template#: 603, Serial#: 2906
    public void cmovae(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0x43, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovae  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovae    eax, ebx[esi * 4 + 18]}
     */
    // Template#: 604, Serial#: 2896
    public void cmovae(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0x43, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovae  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovae    eax, [ebx + 18]}
     */
    // Template#: 605, Serial#: 2895
    public void cmovae(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0x43, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovae  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovae    eax, ebx[esi * 4]}
     */
    // Template#: 606, Serial#: 2892
    public void cmovae(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0x43, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovae  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovae    eax, eax}
     */
    // Template#: 607, Serial#: 2899
    public void cmovae(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0150((byte) 0x43, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovae  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovae    eax, [ebx]}
     */
    // Template#: 608, Serial#: 2891
    public void cmovae(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0x43, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovae  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovae    eax, [L1: 0x12345678]}
     */
    // Template#: 609, Serial#: 2894
    public void m_cmovae(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0x43, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovae  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovae    eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 610, Serial#: 2898
    public void cmovae(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0x43, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovae  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovae    eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 611, Serial#: 2893
    public void m_cmovae(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0x43, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovae  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovae    eax, [ebx + 305419896]}
     */
    // Template#: 612, Serial#: 2897
    public void cmovae(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0x43, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovb     ax, ebx[esi * 4 + 18]}
     */
    // Template#: 613, Serial#: 2887
    public void cmovb(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0x42, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovb     ax, [ebx + 18]}
     */
    // Template#: 614, Serial#: 2886
    public void cmovb(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0x42, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovb     ax, ebx[esi * 4]}
     */
    // Template#: 615, Serial#: 2883
    public void cmovb(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0x42, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovb     ax, ax}
     */
    // Template#: 616, Serial#: 2890
    public void cmovb(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0141((byte) 0x42, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovb     ax, [ebx]}
     */
    // Template#: 617, Serial#: 2882
    public void cmovb(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0x42, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovb     ax, [L1: 0x12345678]}
     */
    // Template#: 618, Serial#: 2885
    public void m_cmovb(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0x42, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovb     ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 619, Serial#: 2889
    public void cmovb(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0x42, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovb     ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 620, Serial#: 2884
    public void m_cmovb(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0x42, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovb     ax, [ebx + 305419896]}
     */
    // Template#: 621, Serial#: 2888
    public void cmovb(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0x42, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovb     eax, ebx[esi * 4 + 18]}
     */
    // Template#: 622, Serial#: 2878
    public void cmovb(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0x42, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovb     eax, [ebx + 18]}
     */
    // Template#: 623, Serial#: 2877
    public void cmovb(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0x42, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovb     eax, ebx[esi * 4]}
     */
    // Template#: 624, Serial#: 2874
    public void cmovb(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0x42, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovb     eax, eax}
     */
    // Template#: 625, Serial#: 2881
    public void cmovb(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0150((byte) 0x42, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovb     eax, [ebx]}
     */
    // Template#: 626, Serial#: 2873
    public void cmovb(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0x42, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovb     eax, [L1: 0x12345678]}
     */
    // Template#: 627, Serial#: 2876
    public void m_cmovb(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0x42, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovb     eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 628, Serial#: 2880
    public void cmovb(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0x42, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovb     eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 629, Serial#: 2875
    public void m_cmovb(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0x42, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovb     eax, [ebx + 305419896]}
     */
    // Template#: 630, Serial#: 2879
    public void cmovb(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0x42, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovbe  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovbe    ax, ebx[esi * 4 + 18]}
     */
    // Template#: 631, Serial#: 2959
    public void cmovbe(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0x46, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovbe  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovbe    ax, [ebx + 18]}
     */
    // Template#: 632, Serial#: 2958
    public void cmovbe(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0x46, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovbe  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovbe    ax, ebx[esi * 4]}
     */
    // Template#: 633, Serial#: 2955
    public void cmovbe(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0x46, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovbe  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovbe    ax, ax}
     */
    // Template#: 634, Serial#: 2962
    public void cmovbe(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0141((byte) 0x46, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovbe  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovbe    ax, [ebx]}
     */
    // Template#: 635, Serial#: 2954
    public void cmovbe(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0x46, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovbe  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovbe    ax, [L1: 0x12345678]}
     */
    // Template#: 636, Serial#: 2957
    public void m_cmovbe(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0x46, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovbe  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovbe    ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 637, Serial#: 2961
    public void cmovbe(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0x46, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovbe  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovbe    ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 638, Serial#: 2956
    public void m_cmovbe(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0x46, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovbe  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovbe    ax, [ebx + 305419896]}
     */
    // Template#: 639, Serial#: 2960
    public void cmovbe(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0x46, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovbe  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovbe    eax, ebx[esi * 4 + 18]}
     */
    // Template#: 640, Serial#: 2950
    public void cmovbe(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0x46, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovbe  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovbe    eax, [ebx + 18]}
     */
    // Template#: 641, Serial#: 2949
    public void cmovbe(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0x46, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovbe  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovbe    eax, ebx[esi * 4]}
     */
    // Template#: 642, Serial#: 2946
    public void cmovbe(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0x46, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovbe  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovbe    eax, eax}
     */
    // Template#: 643, Serial#: 2953
    public void cmovbe(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0150((byte) 0x46, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovbe  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovbe    eax, [ebx]}
     */
    // Template#: 644, Serial#: 2945
    public void cmovbe(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0x46, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovbe  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovbe    eax, [L1: 0x12345678]}
     */
    // Template#: 645, Serial#: 2948
    public void m_cmovbe(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0x46, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovbe  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovbe    eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 646, Serial#: 2952
    public void cmovbe(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0x46, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovbe  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovbe    eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 647, Serial#: 2947
    public void m_cmovbe(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0x46, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovbe  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovbe    eax, [ebx + 305419896]}
     */
    // Template#: 648, Serial#: 2951
    public void cmovbe(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0x46, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmove  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmove     ax, ebx[esi * 4 + 18]}
     */
    // Template#: 649, Serial#: 2923
    public void cmove(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0x44, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmove  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmove     ax, [ebx + 18]}
     */
    // Template#: 650, Serial#: 2922
    public void cmove(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0x44, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmove  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmove     ax, ebx[esi * 4]}
     */
    // Template#: 651, Serial#: 2919
    public void cmove(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0x44, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmove  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmove     ax, ax}
     */
    // Template#: 652, Serial#: 2926
    public void cmove(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0141((byte) 0x44, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmove  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmove     ax, [ebx]}
     */
    // Template#: 653, Serial#: 2918
    public void cmove(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0x44, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmove  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmove     ax, [L1: 0x12345678]}
     */
    // Template#: 654, Serial#: 2921
    public void m_cmove(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0x44, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmove  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmove     ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 655, Serial#: 2925
    public void cmove(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0x44, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmove  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmove     ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 656, Serial#: 2920
    public void m_cmove(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0x44, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmove  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmove     ax, [ebx + 305419896]}
     */
    // Template#: 657, Serial#: 2924
    public void cmove(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0x44, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmove  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmove     eax, ebx[esi * 4 + 18]}
     */
    // Template#: 658, Serial#: 2914
    public void cmove(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0x44, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmove  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmove     eax, [ebx + 18]}
     */
    // Template#: 659, Serial#: 2913
    public void cmove(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0x44, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmove  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmove     eax, ebx[esi * 4]}
     */
    // Template#: 660, Serial#: 2910
    public void cmove(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0x44, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmove  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmove     eax, eax}
     */
    // Template#: 661, Serial#: 2917
    public void cmove(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0150((byte) 0x44, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmove  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmove     eax, [ebx]}
     */
    // Template#: 662, Serial#: 2909
    public void cmove(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0x44, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmove  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmove     eax, [L1: 0x12345678]}
     */
    // Template#: 663, Serial#: 2912
    public void m_cmove(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0x44, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmove  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmove     eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 664, Serial#: 2916
    public void cmove(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0x44, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmove  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmove     eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 665, Serial#: 2911
    public void m_cmove(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0x44, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmove  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmove     eax, [ebx + 305419896]}
     */
    // Template#: 666, Serial#: 2915
    public void cmove(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0x44, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovg  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovg     ax, ebx[esi * 4 + 18]}
     */
    // Template#: 667, Serial#: 4366
    public void cmovg(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0x4F, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovg  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovg     ax, [ebx + 18]}
     */
    // Template#: 668, Serial#: 4365
    public void cmovg(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0x4F, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovg  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovg     ax, ebx[esi * 4]}
     */
    // Template#: 669, Serial#: 4362
    public void cmovg(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0x4F, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovg  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovg     ax, ax}
     */
    // Template#: 670, Serial#: 4369
    public void cmovg(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0141((byte) 0x4F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovg  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovg     ax, [ebx]}
     */
    // Template#: 671, Serial#: 4361
    public void cmovg(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0x4F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovg  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovg     ax, [L1: 0x12345678]}
     */
    // Template#: 672, Serial#: 4364
    public void m_cmovg(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0x4F, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovg  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovg     ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 673, Serial#: 4368
    public void cmovg(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0x4F, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovg  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovg     ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 674, Serial#: 4363
    public void m_cmovg(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0x4F, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovg  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovg     ax, [ebx + 305419896]}
     */
    // Template#: 675, Serial#: 4367
    public void cmovg(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0x4F, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovg  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovg     eax, ebx[esi * 4 + 18]}
     */
    // Template#: 676, Serial#: 4357
    public void cmovg(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0x4F, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovg  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovg     eax, [ebx + 18]}
     */
    // Template#: 677, Serial#: 4356
    public void cmovg(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0x4F, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovg  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovg     eax, ebx[esi * 4]}
     */
    // Template#: 678, Serial#: 4353
    public void cmovg(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0x4F, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovg  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovg     eax, eax}
     */
    // Template#: 679, Serial#: 4360
    public void cmovg(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0150((byte) 0x4F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovg  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovg     eax, [ebx]}
     */
    // Template#: 680, Serial#: 4352
    public void cmovg(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0x4F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovg  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovg     eax, [L1: 0x12345678]}
     */
    // Template#: 681, Serial#: 4355
    public void m_cmovg(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0x4F, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovg  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovg     eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 682, Serial#: 4359
    public void cmovg(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0x4F, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovg  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovg     eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 683, Serial#: 4354
    public void m_cmovg(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0x4F, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovg  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovg     eax, [ebx + 305419896]}
     */
    // Template#: 684, Serial#: 4358
    public void cmovg(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0x4F, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovge  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovge    ax, ebx[esi * 4 + 18]}
     */
    // Template#: 685, Serial#: 4330
    public void cmovge(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0x4D, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovge  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovge    ax, [ebx + 18]}
     */
    // Template#: 686, Serial#: 4329
    public void cmovge(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0x4D, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovge  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovge    ax, ebx[esi * 4]}
     */
    // Template#: 687, Serial#: 4326
    public void cmovge(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0x4D, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovge  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovge    ax, ax}
     */
    // Template#: 688, Serial#: 4333
    public void cmovge(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0141((byte) 0x4D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovge  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovge    ax, [ebx]}
     */
    // Template#: 689, Serial#: 4325
    public void cmovge(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0x4D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovge  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovge    ax, [L1: 0x12345678]}
     */
    // Template#: 690, Serial#: 4328
    public void m_cmovge(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0x4D, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovge  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovge    ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 691, Serial#: 4332
    public void cmovge(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0x4D, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovge  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovge    ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 692, Serial#: 4327
    public void m_cmovge(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0x4D, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovge  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovge    ax, [ebx + 305419896]}
     */
    // Template#: 693, Serial#: 4331
    public void cmovge(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0x4D, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovge  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovge    eax, ebx[esi * 4 + 18]}
     */
    // Template#: 694, Serial#: 4321
    public void cmovge(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0x4D, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovge  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovge    eax, [ebx + 18]}
     */
    // Template#: 695, Serial#: 4320
    public void cmovge(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0x4D, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovge  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovge    eax, ebx[esi * 4]}
     */
    // Template#: 696, Serial#: 4317
    public void cmovge(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0x4D, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovge  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovge    eax, eax}
     */
    // Template#: 697, Serial#: 4324
    public void cmovge(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0150((byte) 0x4D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovge  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovge    eax, [ebx]}
     */
    // Template#: 698, Serial#: 4316
    public void cmovge(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0x4D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovge  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovge    eax, [L1: 0x12345678]}
     */
    // Template#: 699, Serial#: 4319
    public void m_cmovge(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0x4D, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovge  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovge    eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 700, Serial#: 4323
    public void cmovge(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0x4D, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovge  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovge    eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 701, Serial#: 4318
    public void m_cmovge(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0x4D, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovge  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovge    eax, [ebx + 305419896]}
     */
    // Template#: 702, Serial#: 4322
    public void cmovge(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0x4D, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovl  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovl     ax, ebx[esi * 4 + 18]}
     */
    // Template#: 703, Serial#: 4312
    public void cmovl(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0x4C, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovl  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovl     ax, [ebx + 18]}
     */
    // Template#: 704, Serial#: 4311
    public void cmovl(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0x4C, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovl  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovl     ax, ebx[esi * 4]}
     */
    // Template#: 705, Serial#: 4308
    public void cmovl(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0x4C, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovl  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovl     ax, ax}
     */
    // Template#: 706, Serial#: 4315
    public void cmovl(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0141((byte) 0x4C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovl  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovl     ax, [ebx]}
     */
    // Template#: 707, Serial#: 4307
    public void cmovl(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0x4C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovl  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovl     ax, [L1: 0x12345678]}
     */
    // Template#: 708, Serial#: 4310
    public void m_cmovl(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0x4C, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovl  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovl     ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 709, Serial#: 4314
    public void cmovl(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0x4C, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovl  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovl     ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 710, Serial#: 4309
    public void m_cmovl(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0x4C, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovl  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovl     ax, [ebx + 305419896]}
     */
    // Template#: 711, Serial#: 4313
    public void cmovl(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0x4C, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovl  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovl     eax, ebx[esi * 4 + 18]}
     */
    // Template#: 712, Serial#: 4303
    public void cmovl(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0x4C, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovl  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovl     eax, [ebx + 18]}
     */
    // Template#: 713, Serial#: 4302
    public void cmovl(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0x4C, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovl  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovl     eax, ebx[esi * 4]}
     */
    // Template#: 714, Serial#: 4299
    public void cmovl(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0x4C, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovl  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovl     eax, eax}
     */
    // Template#: 715, Serial#: 4306
    public void cmovl(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0150((byte) 0x4C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovl  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovl     eax, [ebx]}
     */
    // Template#: 716, Serial#: 4298
    public void cmovl(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0x4C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovl  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovl     eax, [L1: 0x12345678]}
     */
    // Template#: 717, Serial#: 4301
    public void m_cmovl(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0x4C, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovl  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovl     eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 718, Serial#: 4305
    public void cmovl(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0x4C, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovl  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovl     eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 719, Serial#: 4300
    public void m_cmovl(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0x4C, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovl  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovl     eax, [ebx + 305419896]}
     */
    // Template#: 720, Serial#: 4304
    public void cmovl(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0x4C, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovle  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovle    ax, ebx[esi * 4 + 18]}
     */
    // Template#: 721, Serial#: 4348
    public void cmovle(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0x4E, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovle  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovle    ax, [ebx + 18]}
     */
    // Template#: 722, Serial#: 4347
    public void cmovle(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0x4E, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovle  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovle    ax, ebx[esi * 4]}
     */
    // Template#: 723, Serial#: 4344
    public void cmovle(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0x4E, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovle  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovle    ax, ax}
     */
    // Template#: 724, Serial#: 4351
    public void cmovle(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0141((byte) 0x4E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovle  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovle    ax, [ebx]}
     */
    // Template#: 725, Serial#: 4343
    public void cmovle(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0x4E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovle  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovle    ax, [L1: 0x12345678]}
     */
    // Template#: 726, Serial#: 4346
    public void m_cmovle(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0x4E, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovle  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovle    ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 727, Serial#: 4350
    public void cmovle(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0x4E, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovle  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovle    ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 728, Serial#: 4345
    public void m_cmovle(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0x4E, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovle  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovle    ax, [ebx + 305419896]}
     */
    // Template#: 729, Serial#: 4349
    public void cmovle(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0x4E, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovle  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovle    eax, ebx[esi * 4 + 18]}
     */
    // Template#: 730, Serial#: 4339
    public void cmovle(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0x4E, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovle  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovle    eax, [ebx + 18]}
     */
    // Template#: 731, Serial#: 4338
    public void cmovle(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0x4E, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovle  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovle    eax, ebx[esi * 4]}
     */
    // Template#: 732, Serial#: 4335
    public void cmovle(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0x4E, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovle  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovle    eax, eax}
     */
    // Template#: 733, Serial#: 4342
    public void cmovle(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0150((byte) 0x4E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovle  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovle    eax, [ebx]}
     */
    // Template#: 734, Serial#: 4334
    public void cmovle(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0x4E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovle  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovle    eax, [L1: 0x12345678]}
     */
    // Template#: 735, Serial#: 4337
    public void m_cmovle(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0x4E, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovle  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovle    eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 736, Serial#: 4341
    public void cmovle(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0x4E, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovle  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovle    eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 737, Serial#: 4336
    public void m_cmovle(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0x4E, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovle  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovle    eax, [ebx + 305419896]}
     */
    // Template#: 738, Serial#: 4340
    public void cmovle(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0x4E, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovne  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovne    ax, ebx[esi * 4 + 18]}
     */
    // Template#: 739, Serial#: 2941
    public void cmovne(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0x45, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovne  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovne    ax, [ebx + 18]}
     */
    // Template#: 740, Serial#: 2940
    public void cmovne(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0x45, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovne  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovne    ax, ebx[esi * 4]}
     */
    // Template#: 741, Serial#: 2937
    public void cmovne(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0x45, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovne  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovne    ax, ax}
     */
    // Template#: 742, Serial#: 2944
    public void cmovne(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0141((byte) 0x45, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovne  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovne    ax, [ebx]}
     */
    // Template#: 743, Serial#: 2936
    public void cmovne(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0x45, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovne  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovne    ax, [L1: 0x12345678]}
     */
    // Template#: 744, Serial#: 2939
    public void m_cmovne(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0x45, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovne  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovne    ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 745, Serial#: 2943
    public void cmovne(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0x45, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovne  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovne    ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 746, Serial#: 2938
    public void m_cmovne(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0x45, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovne  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovne    ax, [ebx + 305419896]}
     */
    // Template#: 747, Serial#: 2942
    public void cmovne(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0x45, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovne  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovne    eax, ebx[esi * 4 + 18]}
     */
    // Template#: 748, Serial#: 2932
    public void cmovne(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0x45, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovne  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovne    eax, [ebx + 18]}
     */
    // Template#: 749, Serial#: 2931
    public void cmovne(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0x45, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovne  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovne    eax, ebx[esi * 4]}
     */
    // Template#: 750, Serial#: 2928
    public void cmovne(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0x45, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovne  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovne    eax, eax}
     */
    // Template#: 751, Serial#: 2935
    public void cmovne(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0150((byte) 0x45, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovne  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovne    eax, [ebx]}
     */
    // Template#: 752, Serial#: 2927
    public void cmovne(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0x45, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovne  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovne    eax, [L1: 0x12345678]}
     */
    // Template#: 753, Serial#: 2930
    public void m_cmovne(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0x45, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovne  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovne    eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 754, Serial#: 2934
    public void cmovne(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0x45, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovne  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovne    eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 755, Serial#: 2929
    public void m_cmovne(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0x45, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovne  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovne    eax, [ebx + 305419896]}
     */
    // Template#: 756, Serial#: 2933
    public void cmovne(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0x45, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovno  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovno    ax, ebx[esi * 4 + 18]}
     */
    // Template#: 757, Serial#: 2869
    public void cmovno(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0x41, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovno  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovno    ax, [ebx + 18]}
     */
    // Template#: 758, Serial#: 2868
    public void cmovno(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0x41, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovno  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovno    ax, ebx[esi * 4]}
     */
    // Template#: 759, Serial#: 2865
    public void cmovno(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0x41, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovno  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovno    ax, ax}
     */
    // Template#: 760, Serial#: 2872
    public void cmovno(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0141((byte) 0x41, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovno  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovno    ax, [ebx]}
     */
    // Template#: 761, Serial#: 2864
    public void cmovno(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0x41, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovno  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovno    ax, [L1: 0x12345678]}
     */
    // Template#: 762, Serial#: 2867
    public void m_cmovno(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0x41, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovno  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovno    ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 763, Serial#: 2871
    public void cmovno(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0x41, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovno  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovno    ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 764, Serial#: 2866
    public void m_cmovno(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0x41, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovno  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovno    ax, [ebx + 305419896]}
     */
    // Template#: 765, Serial#: 2870
    public void cmovno(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0x41, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovno  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovno    eax, ebx[esi * 4 + 18]}
     */
    // Template#: 766, Serial#: 2860
    public void cmovno(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0x41, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovno  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovno    eax, [ebx + 18]}
     */
    // Template#: 767, Serial#: 2859
    public void cmovno(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0x41, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovno  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovno    eax, ebx[esi * 4]}
     */
    // Template#: 768, Serial#: 2856
    public void cmovno(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0x41, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovno  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovno    eax, eax}
     */
    // Template#: 769, Serial#: 2863
    public void cmovno(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0150((byte) 0x41, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovno  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovno    eax, [ebx]}
     */
    // Template#: 770, Serial#: 2855
    public void cmovno(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0x41, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovno  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovno    eax, [L1: 0x12345678]}
     */
    // Template#: 771, Serial#: 2858
    public void m_cmovno(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0x41, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovno  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovno    eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 772, Serial#: 2862
    public void cmovno(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0x41, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovno  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovno    eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 773, Serial#: 2857
    public void m_cmovno(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0x41, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovno  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovno    eax, [ebx + 305419896]}
     */
    // Template#: 774, Serial#: 2861
    public void cmovno(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0x41, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovnp  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovnp    ax, ebx[esi * 4 + 18]}
     */
    // Template#: 775, Serial#: 4294
    public void cmovnp(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0x4B, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovnp  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovnp    ax, [ebx + 18]}
     */
    // Template#: 776, Serial#: 4293
    public void cmovnp(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0x4B, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovnp  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovnp    ax, ebx[esi * 4]}
     */
    // Template#: 777, Serial#: 4290
    public void cmovnp(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0x4B, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovnp  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovnp    ax, ax}
     */
    // Template#: 778, Serial#: 4297
    public void cmovnp(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0141((byte) 0x4B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovnp  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovnp    ax, [ebx]}
     */
    // Template#: 779, Serial#: 4289
    public void cmovnp(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0x4B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovnp  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovnp    ax, [L1: 0x12345678]}
     */
    // Template#: 780, Serial#: 4292
    public void m_cmovnp(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0x4B, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovnp  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovnp    ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 781, Serial#: 4296
    public void cmovnp(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0x4B, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovnp  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovnp    ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 782, Serial#: 4291
    public void m_cmovnp(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0x4B, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovnp  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovnp    ax, [ebx + 305419896]}
     */
    // Template#: 783, Serial#: 4295
    public void cmovnp(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0x4B, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovnp  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovnp    eax, ebx[esi * 4 + 18]}
     */
    // Template#: 784, Serial#: 4285
    public void cmovnp(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0x4B, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovnp  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovnp    eax, [ebx + 18]}
     */
    // Template#: 785, Serial#: 4284
    public void cmovnp(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0x4B, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovnp  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovnp    eax, ebx[esi * 4]}
     */
    // Template#: 786, Serial#: 4281
    public void cmovnp(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0x4B, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovnp  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovnp    eax, eax}
     */
    // Template#: 787, Serial#: 4288
    public void cmovnp(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0150((byte) 0x4B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovnp  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovnp    eax, [ebx]}
     */
    // Template#: 788, Serial#: 4280
    public void cmovnp(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0x4B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovnp  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovnp    eax, [L1: 0x12345678]}
     */
    // Template#: 789, Serial#: 4283
    public void m_cmovnp(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0x4B, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovnp  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovnp    eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 790, Serial#: 4287
    public void cmovnp(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0x4B, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovnp  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovnp    eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 791, Serial#: 4282
    public void m_cmovnp(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0x4B, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovnp  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovnp    eax, [ebx + 305419896]}
     */
    // Template#: 792, Serial#: 4286
    public void cmovnp(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0x4B, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovns  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovns    ax, ebx[esi * 4 + 18]}
     */
    // Template#: 793, Serial#: 4258
    public void cmovns(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0x49, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovns  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovns    ax, [ebx + 18]}
     */
    // Template#: 794, Serial#: 4257
    public void cmovns(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0x49, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovns  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovns    ax, ebx[esi * 4]}
     */
    // Template#: 795, Serial#: 4254
    public void cmovns(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0x49, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovns  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovns    ax, ax}
     */
    // Template#: 796, Serial#: 4261
    public void cmovns(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0141((byte) 0x49, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovns  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovns    ax, [ebx]}
     */
    // Template#: 797, Serial#: 4253
    public void cmovns(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0x49, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovns  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovns    ax, [L1: 0x12345678]}
     */
    // Template#: 798, Serial#: 4256
    public void m_cmovns(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0x49, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovns  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovns    ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 799, Serial#: 4260
    public void cmovns(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0x49, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovns  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovns    ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 800, Serial#: 4255
    public void m_cmovns(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0x49, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovns  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovns    ax, [ebx + 305419896]}
     */
    // Template#: 801, Serial#: 4259
    public void cmovns(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0x49, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovns  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovns    eax, ebx[esi * 4 + 18]}
     */
    // Template#: 802, Serial#: 4249
    public void cmovns(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0x49, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovns  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovns    eax, [ebx + 18]}
     */
    // Template#: 803, Serial#: 4248
    public void cmovns(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0x49, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovns  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovns    eax, ebx[esi * 4]}
     */
    // Template#: 804, Serial#: 4245
    public void cmovns(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0x49, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovns  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovns    eax, eax}
     */
    // Template#: 805, Serial#: 4252
    public void cmovns(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0150((byte) 0x49, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovns  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovns    eax, [ebx]}
     */
    // Template#: 806, Serial#: 4244
    public void cmovns(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0x49, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovns  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovns    eax, [L1: 0x12345678]}
     */
    // Template#: 807, Serial#: 4247
    public void m_cmovns(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0x49, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovns  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovns    eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 808, Serial#: 4251
    public void cmovns(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0x49, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovns  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovns    eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 809, Serial#: 4246
    public void m_cmovns(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0x49, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovns  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovns    eax, [ebx + 305419896]}
     */
    // Template#: 810, Serial#: 4250
    public void cmovns(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0x49, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovo  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovo     ax, ebx[esi * 4 + 18]}
     */
    // Template#: 811, Serial#: 2851
    public void cmovo(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0x40, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovo  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovo     ax, [ebx + 18]}
     */
    // Template#: 812, Serial#: 2850
    public void cmovo(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0x40, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovo  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovo     ax, ebx[esi * 4]}
     */
    // Template#: 813, Serial#: 2847
    public void cmovo(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0x40, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovo  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovo     ax, ax}
     */
    // Template#: 814, Serial#: 2854
    public void cmovo(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0141((byte) 0x40, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovo  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovo     ax, [ebx]}
     */
    // Template#: 815, Serial#: 2846
    public void cmovo(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0x40, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovo  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovo     ax, [L1: 0x12345678]}
     */
    // Template#: 816, Serial#: 2849
    public void m_cmovo(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0x40, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovo  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovo     ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 817, Serial#: 2853
    public void cmovo(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0x40, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovo  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovo     ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 818, Serial#: 2848
    public void m_cmovo(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0x40, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovo  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovo     ax, [ebx + 305419896]}
     */
    // Template#: 819, Serial#: 2852
    public void cmovo(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0x40, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovo  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovo     eax, ebx[esi * 4 + 18]}
     */
    // Template#: 820, Serial#: 2842
    public void cmovo(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0x40, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovo  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovo     eax, [ebx + 18]}
     */
    // Template#: 821, Serial#: 2841
    public void cmovo(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0x40, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovo  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovo     eax, ebx[esi * 4]}
     */
    // Template#: 822, Serial#: 2838
    public void cmovo(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0x40, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovo  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovo     eax, eax}
     */
    // Template#: 823, Serial#: 2845
    public void cmovo(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0150((byte) 0x40, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovo  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovo     eax, [ebx]}
     */
    // Template#: 824, Serial#: 2837
    public void cmovo(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0x40, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovo  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovo     eax, [L1: 0x12345678]}
     */
    // Template#: 825, Serial#: 2840
    public void m_cmovo(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0x40, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovo  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovo     eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 826, Serial#: 2844
    public void cmovo(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0x40, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovo  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovo     eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 827, Serial#: 2839
    public void m_cmovo(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0x40, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovo  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovo     eax, [ebx + 305419896]}
     */
    // Template#: 828, Serial#: 2843
    public void cmovo(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0x40, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovp  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovp     ax, ebx[esi * 4 + 18]}
     */
    // Template#: 829, Serial#: 4276
    public void cmovp(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0x4A, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovp  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovp     ax, [ebx + 18]}
     */
    // Template#: 830, Serial#: 4275
    public void cmovp(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0x4A, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovp  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovp     ax, ebx[esi * 4]}
     */
    // Template#: 831, Serial#: 4272
    public void cmovp(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0x4A, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovp  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovp     ax, ax}
     */
    // Template#: 832, Serial#: 4279
    public void cmovp(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0141((byte) 0x4A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovp  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovp     ax, [ebx]}
     */
    // Template#: 833, Serial#: 4271
    public void cmovp(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0x4A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovp  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovp     ax, [L1: 0x12345678]}
     */
    // Template#: 834, Serial#: 4274
    public void m_cmovp(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0x4A, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovp  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovp     ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 835, Serial#: 4278
    public void cmovp(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0x4A, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovp  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovp     ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 836, Serial#: 4273
    public void m_cmovp(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0x4A, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovp  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovp     ax, [ebx + 305419896]}
     */
    // Template#: 837, Serial#: 4277
    public void cmovp(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0x4A, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovp  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovp     eax, ebx[esi * 4 + 18]}
     */
    // Template#: 838, Serial#: 4267
    public void cmovp(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0x4A, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovp  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovp     eax, [ebx + 18]}
     */
    // Template#: 839, Serial#: 4266
    public void cmovp(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0x4A, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovp  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovp     eax, ebx[esi * 4]}
     */
    // Template#: 840, Serial#: 4263
    public void cmovp(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0x4A, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovp  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovp     eax, eax}
     */
    // Template#: 841, Serial#: 4270
    public void cmovp(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0150((byte) 0x4A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovp  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovp     eax, [ebx]}
     */
    // Template#: 842, Serial#: 4262
    public void cmovp(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0x4A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovp  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovp     eax, [L1: 0x12345678]}
     */
    // Template#: 843, Serial#: 4265
    public void m_cmovp(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0x4A, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovp  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovp     eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 844, Serial#: 4269
    public void cmovp(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0x4A, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovp  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovp     eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 845, Serial#: 4264
    public void m_cmovp(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0x4A, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovp  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovp     eax, [ebx + 305419896]}
     */
    // Template#: 846, Serial#: 4268
    public void cmovp(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0x4A, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovs  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovs     ax, ebx[esi * 4 + 18]}
     */
    // Template#: 847, Serial#: 4240
    public void cmovs(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0x48, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovs  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovs     ax, [ebx + 18]}
     */
    // Template#: 848, Serial#: 4239
    public void cmovs(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0x48, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovs  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovs     ax, ebx[esi * 4]}
     */
    // Template#: 849, Serial#: 4236
    public void cmovs(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0x48, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovs  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovs     ax, ax}
     */
    // Template#: 850, Serial#: 4243
    public void cmovs(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0141((byte) 0x48, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovs  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovs     ax, [ebx]}
     */
    // Template#: 851, Serial#: 4235
    public void cmovs(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0x48, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovs  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovs     ax, [L1: 0x12345678]}
     */
    // Template#: 852, Serial#: 4238
    public void m_cmovs(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0x48, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovs  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovs     ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 853, Serial#: 4242
    public void cmovs(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0x48, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovs  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovs     ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 854, Serial#: 4237
    public void m_cmovs(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0x48, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovs  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovs     ax, [ebx + 305419896]}
     */
    // Template#: 855, Serial#: 4241
    public void cmovs(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0x48, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovs  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovs     eax, ebx[esi * 4 + 18]}
     */
    // Template#: 856, Serial#: 4231
    public void cmovs(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0x48, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovs  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovs     eax, [ebx + 18]}
     */
    // Template#: 857, Serial#: 4230
    public void cmovs(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0x48, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovs  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovs     eax, ebx[esi * 4]}
     */
    // Template#: 858, Serial#: 4227
    public void cmovs(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0x48, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovs  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovs     eax, eax}
     */
    // Template#: 859, Serial#: 4234
    public void cmovs(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0150((byte) 0x48, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovs  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovs     eax, [ebx]}
     */
    // Template#: 860, Serial#: 4226
    public void cmovs(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0x48, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovs  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmovs     eax, [L1: 0x12345678]}
     */
    // Template#: 861, Serial#: 4229
    public void m_cmovs(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0x48, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovs  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovs     eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 862, Serial#: 4233
    public void cmovs(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0x48, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovs  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmovs     eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 863, Serial#: 4228
    public void m_cmovs(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0x48, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmovs  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmovs     eax, [ebx + 305419896]}
     */
    // Template#: 864, Serial#: 4232
    public void cmovs(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0x48, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>imm8</i>
     * Example disassembly syntax: {@code cmp       al, 0x12}
     */
    // Template#: 865, Serial#: 1754
    public void cmp_AL(byte imm8) {
        assemble0002((byte) 0x3C, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpb      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 866, Serial#: 317
    public void cmpb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0x80, (byte) 0x07, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpl      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 867, Serial#: 533
    public void cmpl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0x83, (byte) 0x07, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpw      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 868, Serial#: 605
    public void cmpw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0004((byte) 0x83, (byte) 0x07, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       ebx[esi * 4 + 18], ax}
     */
    // Template#: 869, Serial#: 1726
    public void cmp(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0005((byte) 0x39, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       ebx[esi * 4 + 18], eax}
     */
    // Template#: 870, Serial#: 1717
    public void cmp(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0006((byte) 0x39, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       ebx[esi * 4 + 18], al}
     */
    // Template#: 871, Serial#: 1708
    public void cmp(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0007((byte) 0x38, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code cmpl      ebx[esi * 4 + 18], 0x12345678}
     */
    // Template#: 872, Serial#: 389
    public void cmpl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0008((byte) 0x81, (byte) 0x07, disp8, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code cmpw      ebx[esi * 4 + 18], 0x1234}
     */
    // Template#: 873, Serial#: 461
    public void cmpw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0009((byte) 0x81, (byte) 0x07, disp8, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpb  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpb      [ebx + 18], 0x12}
     */
    // Template#: 874, Serial#: 316
    public void cmpb(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0x80, (byte) 0x07, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpl  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpl      [ebx + 18], 0x12}
     */
    // Template#: 875, Serial#: 532
    public void cmpl(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0x83, (byte) 0x07, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpw  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpw      [ebx + 18], 0x12}
     */
    // Template#: 876, Serial#: 604
    public void cmpw(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0011((byte) 0x83, (byte) 0x07, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       [ebx + 18], ax}
     */
    // Template#: 877, Serial#: 1725
    public void cmp(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0012((byte) 0x39, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       [ebx + 18], eax}
     */
    // Template#: 878, Serial#: 1716
    public void cmp(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0013((byte) 0x39, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       [ebx + 18], al}
     */
    // Template#: 879, Serial#: 1707
    public void cmp(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0014((byte) 0x38, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpl  }<i>disp8</i>, <i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code cmpl      [ebx + 18], 0x12345678}
     */
    // Template#: 880, Serial#: 388
    public void cmpl(byte disp8, IA32IndirectRegister32 destination, int imm32) {
        assemble0015((byte) 0x81, (byte) 0x07, disp8, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpw  }<i>disp8</i>, <i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code cmpw      [ebx + 18], 0x1234}
     */
    // Template#: 881, Serial#: 460
    public void cmpw(byte disp8, IA32IndirectRegister32 destination, short imm16) {
        assemble0016((byte) 0x81, (byte) 0x07, disp8, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpb  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpb      ebx[esi * 4], 0x12}
     */
    // Template#: 882, Serial#: 299
    public void cmpb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0x80, (byte) 0x07, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpl  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpl      ebx[esi * 4], 0x12}
     */
    // Template#: 883, Serial#: 515
    public void cmpl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0x83, (byte) 0x07, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpw      ebx[esi * 4], 0x12}
     */
    // Template#: 884, Serial#: 587
    public void cmpw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0018((byte) 0x83, (byte) 0x07, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       ebx[esi * 4], ax}
     */
    // Template#: 885, Serial#: 1722
    public void cmp(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0019((byte) 0x39, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       ebx[esi * 4], eax}
     */
    // Template#: 886, Serial#: 1713
    public void cmp(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0020((byte) 0x39, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       ebx[esi * 4], al}
     */
    // Template#: 887, Serial#: 1704
    public void cmp(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0021((byte) 0x38, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpl  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code cmpl      ebx[esi * 4], 0x12345678}
     */
    // Template#: 888, Serial#: 371
    public void cmpl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0022((byte) 0x81, (byte) 0x07, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code cmpw      ebx[esi * 4], 0x1234}
     */
    // Template#: 889, Serial#: 443
    public void cmpw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0023((byte) 0x81, (byte) 0x07, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpw      ax, 0x12}
     */
    // Template#: 890, Serial#: 629
    public void cmpw(IA32GeneralRegister16 destination, byte imm8) {
        assemble0024((byte) 0x83, (byte) 0x07, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmp       ax, ebx[esi * 4 + 18]}
     */
    // Template#: 891, Serial#: 1751
    public void cmp(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0025((byte) 0x3B, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       ax, [ebx + 18]}
     */
    // Template#: 892, Serial#: 1750
    public void cmp(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0026((byte) 0x3B, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmp       ax, ebx[esi * 4]}
     */
    // Template#: 893, Serial#: 1747
    public void cmp(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0027((byte) 0x3B, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       ax, ax}
     */
    // Template#: 894, Serial#: 1729
    public void cmp(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0028((byte) 0x39, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       ax, [ebx]}
     */
    // Template#: 895, Serial#: 1746
    public void cmp(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0029((byte) 0x3B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmp       ax, [L1: 0x12345678]}
     */
    // Template#: 896, Serial#: 1749
    public void m_cmp(IA32GeneralRegister16 destination, int m32) {
        assemble0030((byte) 0x3B, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmp       ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 897, Serial#: 1753
    public void cmp(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0031((byte) 0x3B, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmp       ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 898, Serial#: 1748
    public void m_cmp(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0032((byte) 0x3B, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       ax, [ebx + 305419896]}
     */
    // Template#: 899, Serial#: 1752
    public void cmp(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0033((byte) 0x3B, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpw  }<i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code cmpw      ax, 0x1234}
     */
    // Template#: 900, Serial#: 485
    public void cmpw(IA32GeneralRegister16 destination, short imm16) {
        assemble0034((byte) 0x81, (byte) 0x07, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpl  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpl      eax, 0x12}
     */
    // Template#: 901, Serial#: 557
    public void cmpl(IA32GeneralRegister32 destination, byte imm8) {
        assemble0035((byte) 0x83, (byte) 0x07, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmp       eax, ebx[esi * 4 + 18]}
     */
    // Template#: 902, Serial#: 1743
    public void cmp(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0036((byte) 0x3B, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       eax, [ebx + 18]}
     */
    // Template#: 903, Serial#: 1742
    public void cmp(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0037((byte) 0x3B, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmp       eax, ebx[esi * 4]}
     */
    // Template#: 904, Serial#: 1739
    public void cmp(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0038((byte) 0x3B, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       eax, eax}
     */
    // Template#: 905, Serial#: 1720
    public void cmp(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0039((byte) 0x39, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       eax, [ebx]}
     */
    // Template#: 906, Serial#: 1738
    public void cmp(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0040((byte) 0x3B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpl  }<i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code cmpl      eax, 0x12345678}
     */
    // Template#: 907, Serial#: 413
    public void cmpl(IA32GeneralRegister32 destination, int imm32) {
        assemble0042((byte) 0x81, (byte) 0x07, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmp       eax, [L1: 0x12345678]}
     */
    // Template#: 908, Serial#: 1741
    public void m_cmp(IA32GeneralRegister32 destination, int m32) {
        assemble0041((byte) 0x3B, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmp       eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 909, Serial#: 1745
    public void cmp(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0043((byte) 0x3B, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmp       eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 910, Serial#: 1740
    public void m_cmp(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0044((byte) 0x3B, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       eax, [ebx + 305419896]}
     */
    // Template#: 911, Serial#: 1744
    public void cmp(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0045((byte) 0x3B, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpb      al, 0x12}
     */
    // Template#: 912, Serial#: 341
    public void cmpb(IA32GeneralRegister8 destination, byte imm8) {
        assemble0046((byte) 0x80, (byte) 0x07, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmp       al, ebx[esi * 4 + 18]}
     */
    // Template#: 913, Serial#: 1735
    public void cmp(IA32GeneralRegister8 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0047((byte) 0x3A, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       al, [ebx + 18]}
     */
    // Template#: 914, Serial#: 1734
    public void cmp(IA32GeneralRegister8 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0048((byte) 0x3A, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmp       al, ebx[esi * 4]}
     */
    // Template#: 915, Serial#: 1731
    public void cmp(IA32GeneralRegister8 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0049((byte) 0x3A, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       al, al}
     */
    // Template#: 916, Serial#: 1711
    public void cmp(IA32GeneralRegister8 destination, IA32GeneralRegister8 source) {
        assemble0050((byte) 0x38, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       al, [ebx]}
     */
    // Template#: 917, Serial#: 1730
    public void cmp(IA32GeneralRegister8 destination, IA32IndirectRegister32 source) {
        assemble0051((byte) 0x3A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cmp       al, [L1: 0x12345678]}
     */
    // Template#: 918, Serial#: 1733
    public void m_cmp(IA32GeneralRegister8 destination, int m32) {
        assemble0052((byte) 0x3A, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmp       al, ebx[esi * 4 + 305419896]}
     */
    // Template#: 919, Serial#: 1737
    public void cmp(IA32GeneralRegister8 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0053((byte) 0x3A, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmp       al, L1: 0x12345678[esi * 4]}
     */
    // Template#: 920, Serial#: 1732
    public void m_cmp(IA32GeneralRegister8 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0054((byte) 0x3A, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       al, [ebx + 305419896]}
     */
    // Template#: 921, Serial#: 1736
    public void cmp(IA32GeneralRegister8 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0055((byte) 0x3A, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpb      [ebx], 0x12}
     */
    // Template#: 922, Serial#: 298
    public void cmpb(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0x80, (byte) 0x07, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpl  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpl      [ebx], 0x12}
     */
    // Template#: 923, Serial#: 514
    public void cmpl(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0x83, (byte) 0x07, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpw      [ebx], 0x12}
     */
    // Template#: 924, Serial#: 586
    public void cmpw(IA32IndirectRegister32 destination, byte imm8) {
        assemble0057((byte) 0x83, (byte) 0x07, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       [ebx], ax}
     */
    // Template#: 925, Serial#: 1721
    public void cmp(IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0058((byte) 0x39, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       [ebx], eax}
     */
    // Template#: 926, Serial#: 1712
    public void cmp(IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0059((byte) 0x39, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       [ebx], al}
     */
    // Template#: 927, Serial#: 1703
    public void cmp(IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0060((byte) 0x38, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpl  }<i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code cmpl      [ebx], 0x12345678}
     */
    // Template#: 928, Serial#: 370
    public void cmpl(IA32IndirectRegister32 destination, int imm32) {
        assemble0061((byte) 0x81, (byte) 0x07, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpw  }<i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code cmpw      [ebx], 0x1234}
     */
    // Template#: 929, Serial#: 442
    public void cmpw(IA32IndirectRegister32 destination, short imm16) {
        assemble0062((byte) 0x81, (byte) 0x07, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpl  }<i>imm32</i>
     * Example disassembly syntax: {@code cmpl      eax, 0x12345678}
     */
    // Template#: 930, Serial#: 1755
    public void cmp_EAX(int imm32) {
        assemble0063((byte) 0x3D, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpb  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpb      [L1: 0x12345678], 0x12}
     */
    // Template#: 931, Serial#: 301
    public void m_cmpb(int m32, byte imm8) {
        assemble0064((byte) 0x80, (byte) 0x07, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpl  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpl      [L1: 0x12345678], 0x12}
     */
    // Template#: 932, Serial#: 517
    public void m_cmpl(int m32, byte imm8) {
        assemble0064((byte) 0x83, (byte) 0x07, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpw  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpw      [L1: 0x12345678], 0x12}
     */
    // Template#: 933, Serial#: 589
    public void m_cmpw(int m32, byte imm8) {
        assemble0065((byte) 0x83, (byte) 0x07, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpb      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 934, Serial#: 333
    public void cmpb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0x80, (byte) 0x07, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpl      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 935, Serial#: 549
    public void cmpl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0x83, (byte) 0x07, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpw      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 936, Serial#: 621
    public void cmpw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0067((byte) 0x83, (byte) 0x07, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       ebx[esi * 4 + 305419896], ax}
     */
    // Template#: 937, Serial#: 1728
    public void cmp(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0068((byte) 0x39, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       ebx[esi * 4 + 305419896], eax}
     */
    // Template#: 938, Serial#: 1719
    public void cmp(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0069((byte) 0x39, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       ebx[esi * 4 + 305419896], al}
     */
    // Template#: 939, Serial#: 1710
    public void cmp(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0070((byte) 0x38, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code cmpl      ebx[esi * 4 + 305419896], 0x12345678}
     */
    // Template#: 940, Serial#: 405
    public void cmpl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0071((byte) 0x81, (byte) 0x07, disp32, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code cmpw      ebx[esi * 4 + 305419896], 0x1234}
     */
    // Template#: 941, Serial#: 477
    public void cmpw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0072((byte) 0x81, (byte) 0x07, disp32, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       [L1: 0x12345678], ax}
     */
    // Template#: 942, Serial#: 1724
    public void m_cmp(int m32, IA32GeneralRegister16 source) {
        assemble0073((byte) 0x39, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       [L1: 0x12345678], eax}
     */
    // Template#: 943, Serial#: 1715
    public void m_cmp(int m32, IA32GeneralRegister32 source) {
        assemble0074((byte) 0x39, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       [L1: 0x12345678], al}
     */
    // Template#: 944, Serial#: 1706
    public void m_cmp(int m32, IA32GeneralRegister8 source) {
        assemble0075((byte) 0x38, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpb  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpb      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 945, Serial#: 300
    public void m_cmpb(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0x80, (byte) 0x07, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpl  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpl      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 946, Serial#: 516
    public void m_cmpl(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0x83, (byte) 0x07, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpw      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 947, Serial#: 588
    public void m_cmpw(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0077((byte) 0x83, (byte) 0x07, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       L1: 0x12345678[esi * 4], ax}
     */
    // Template#: 948, Serial#: 1723
    public void m_cmp(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0078((byte) 0x39, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       L1: 0x12345678[esi * 4], eax}
     */
    // Template#: 949, Serial#: 1714
    public void m_cmp(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0079((byte) 0x39, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       L1: 0x12345678[esi * 4], al}
     */
    // Template#: 950, Serial#: 1705
    public void m_cmp(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0080((byte) 0x38, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpl  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code cmpl      L1: 0x12345678[esi * 4], 0x12345678}
     */
    // Template#: 951, Serial#: 372
    public void m_cmpl(int m32, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0081((byte) 0x81, (byte) 0x07, m32, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code cmpw      L1: 0x12345678[esi * 4], 0x1234}
     */
    // Template#: 952, Serial#: 444
    public void m_cmpw(int m32, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0082((byte) 0x81, (byte) 0x07, m32, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpb  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpb      [ebx + 305419896], 0x12}
     */
    // Template#: 953, Serial#: 332
    public void cmpb(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0x80, (byte) 0x07, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpl  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpl      [ebx + 305419896], 0x12}
     */
    // Template#: 954, Serial#: 548
    public void cmpl(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0x83, (byte) 0x07, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpw  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code cmpw      [ebx + 305419896], 0x12}
     */
    // Template#: 955, Serial#: 620
    public void cmpw(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0084((byte) 0x83, (byte) 0x07, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       [ebx + 305419896], ax}
     */
    // Template#: 956, Serial#: 1727
    public void cmp(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0085((byte) 0x39, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       [ebx + 305419896], eax}
     */
    // Template#: 957, Serial#: 1718
    public void cmp(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0086((byte) 0x39, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmp  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmp       [ebx + 305419896], al}
     */
    // Template#: 958, Serial#: 1709
    public void cmp(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0087((byte) 0x38, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpl  }<i>disp32</i>, <i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code cmpl      [ebx + 305419896], 0x12345678}
     */
    // Template#: 959, Serial#: 404
    public void cmpl(int disp32, IA32IndirectRegister32 destination, int imm32) {
        assemble0088((byte) 0x81, (byte) 0x07, disp32, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpw  }<i>disp32</i>, <i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code cmpw      [ebx + 305419896], 0x1234}
     */
    // Template#: 960, Serial#: 476
    public void cmpw(int disp32, IA32IndirectRegister32 destination, short imm16) {
        assemble0089((byte) 0x81, (byte) 0x07, disp32, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpl  }<i>m32</i>, <i>imm32</i>
     * Example disassembly syntax: {@code cmpl      [L1: 0x12345678], 0x12345678}
     */
    // Template#: 961, Serial#: 373
    public void m_cmpl(int m32, int imm32) {
        assemble0090((byte) 0x81, (byte) 0x07, m32, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpw  }<i>m32</i>, <i>imm16</i>
     * Example disassembly syntax: {@code cmpw      [L1: 0x12345678], 0x1234}
     */
    // Template#: 962, Serial#: 445
    public void m_cmpw(int m32, short imm16) {
        assemble0091((byte) 0x81, (byte) 0x07, m32, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpw  }<i>imm16</i>
     * Example disassembly syntax: {@code cmpw      ax, 0x1234}
     */
    // Template#: 963, Serial#: 1756
    public void cmp_AX(short imm16) {
        assemble0092((byte) 0x3D, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmppd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmppd     xmm0, ebx[esi * 4 + 18], less_than_or_equal}
     */
    // Template#: 964, Serial#: 3671
    public void cmppd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        assemble0199((byte) 0xC2, destination, disp8, base, index, scale, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmppd  }<i>destination</i>, <i>disp8</i>, <i>source</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmppd     xmm0, [ebx + 18], less_than_or_equal}
     */
    // Template#: 965, Serial#: 3670
    public void cmppd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source, IA32XMMComparison ia32xmmcomparison) {
        assemble0200((byte) 0xC2, destination, disp8, source, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmppd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmppd     xmm0, ebx[esi * 4], less_than_or_equal}
     */
    // Template#: 966, Serial#: 3667
    public void cmppd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        assemble0201((byte) 0xC2, destination, base, index, scale, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmppd  }<i>destination</i>, <i>source</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmppd     xmm0, [ebx], less_than_or_equal}
     */
    // Template#: 967, Serial#: 3666
    public void cmppd(IA32XMMRegister destination, IA32IndirectRegister32 source, IA32XMMComparison ia32xmmcomparison) {
        assemble0202((byte) 0xC2, destination, source, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmppd  }<i>destination</i>, <i>source</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmppd     xmm0, xmm0, less_than_or_equal}
     */
    // Template#: 968, Serial#: 3674
    public void cmppd(IA32XMMRegister destination, IA32XMMRegister source, IA32XMMComparison ia32xmmcomparison) {
        assemble0203((byte) 0xC2, destination, source, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmppd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmppd     xmm0, ebx[esi * 4 + 305419896], less_than_or_equal}
     */
    // Template#: 969, Serial#: 3673
    public void cmppd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        assemble0204((byte) 0xC2, destination, disp32, base, index, scale, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmppd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmppd     xmm0, L1: 0x12345678[esi * 4], less_than_or_equal}
     */
    // Template#: 970, Serial#: 3668
    public void m_cmppd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        assemble0205((byte) 0xC2, destination, m32, index, scale, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmppd  }<i>destination</i>, <i>disp32</i>, <i>source</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmppd     xmm0, [ebx + 305419896], less_than_or_equal}
     */
    // Template#: 971, Serial#: 3672
    public void cmppd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source, IA32XMMComparison ia32xmmcomparison) {
        assemble0206((byte) 0xC2, destination, disp32, source, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmppd  }<i>destination</i>, <i>m32</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmppd     xmm0, [L1: 0x12345678], less_than_or_equal}
     */
    // Template#: 972, Serial#: 3669
    public void m_cmppd(IA32XMMRegister destination, int m32, IA32XMMComparison ia32xmmcomparison) {
        assemble0207((byte) 0xC2, destination, m32, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpps  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpps     xmm0, ebx[esi * 4 + 18], less_than_or_equal}
     */
    // Template#: 973, Serial#: 3627
    public void cmpps(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        assemble0208((byte) 0xC2, destination, disp8, base, index, scale, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpps  }<i>destination</i>, <i>disp8</i>, <i>source</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpps     xmm0, [ebx + 18], less_than_or_equal}
     */
    // Template#: 974, Serial#: 3626
    public void cmpps(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source, IA32XMMComparison ia32xmmcomparison) {
        assemble0209((byte) 0xC2, destination, disp8, source, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpps  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpps     xmm0, ebx[esi * 4], less_than_or_equal}
     */
    // Template#: 975, Serial#: 3623
    public void cmpps(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        assemble0210((byte) 0xC2, destination, base, index, scale, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpps  }<i>destination</i>, <i>source</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpps     xmm0, [ebx], less_than_or_equal}
     */
    // Template#: 976, Serial#: 3622
    public void cmpps(IA32XMMRegister destination, IA32IndirectRegister32 source, IA32XMMComparison ia32xmmcomparison) {
        assemble0211((byte) 0xC2, destination, source, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpps  }<i>destination</i>, <i>source</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpps     xmm0, xmm0, less_than_or_equal}
     */
    // Template#: 977, Serial#: 3630
    public void cmpps(IA32XMMRegister destination, IA32XMMRegister source, IA32XMMComparison ia32xmmcomparison) {
        assemble0212((byte) 0xC2, destination, source, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpps  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpps     xmm0, ebx[esi * 4 + 305419896], less_than_or_equal}
     */
    // Template#: 978, Serial#: 3629
    public void cmpps(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        assemble0213((byte) 0xC2, destination, disp32, base, index, scale, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpps  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpps     xmm0, L1: 0x12345678[esi * 4], less_than_or_equal}
     */
    // Template#: 979, Serial#: 3624
    public void m_cmpps(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        assemble0214((byte) 0xC2, destination, m32, index, scale, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpps  }<i>destination</i>, <i>disp32</i>, <i>source</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpps     xmm0, [ebx + 305419896], less_than_or_equal}
     */
    // Template#: 980, Serial#: 3628
    public void cmpps(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source, IA32XMMComparison ia32xmmcomparison) {
        assemble0215((byte) 0xC2, destination, disp32, source, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpps  }<i>destination</i>, <i>m32</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpps     xmm0, [L1: 0x12345678], less_than_or_equal}
     */
    // Template#: 981, Serial#: 3625
    public void m_cmpps(IA32XMMRegister destination, int m32, IA32XMMComparison ia32xmmcomparison) {
        assemble0216((byte) 0xC2, destination, m32, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpsb  }
     * Example disassembly syntax: {@code cmpsb   }
     */
    // Template#: 982, Serial#: 696
    public void cmpsb() {
        assemble0001((byte) 0xA6);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpsl  }
     * Example disassembly syntax: {@code cmpsl   }
     */
    // Template#: 983, Serial#: 697
    public void cmpsl() {
        assemble0001((byte) 0xA7);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpsw  }
     * Example disassembly syntax: {@code cmpsw   }
     */
    // Template#: 984, Serial#: 698
    public void cmpsw() {
        assemble0197((byte) 0xA7);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpsd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpsd     xmm0, ebx[esi * 4 + 18], less_than_or_equal}
     */
    // Template#: 985, Serial#: 3699
    public void cmpsd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        assemble0217((byte) 0xC2, destination, disp8, base, index, scale, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpsd  }<i>destination</i>, <i>disp8</i>, <i>source</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpsd     xmm0, [ebx + 18], less_than_or_equal}
     */
    // Template#: 986, Serial#: 3698
    public void cmpsd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source, IA32XMMComparison ia32xmmcomparison) {
        assemble0218((byte) 0xC2, destination, disp8, source, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpsd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpsd     xmm0, ebx[esi * 4], less_than_or_equal}
     */
    // Template#: 987, Serial#: 3695
    public void cmpsd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        assemble0219((byte) 0xC2, destination, base, index, scale, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpsd  }<i>destination</i>, <i>source</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpsd     xmm0, [ebx], less_than_or_equal}
     */
    // Template#: 988, Serial#: 3694
    public void cmpsd(IA32XMMRegister destination, IA32IndirectRegister32 source, IA32XMMComparison ia32xmmcomparison) {
        assemble0220((byte) 0xC2, destination, source, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpsd  }<i>destination</i>, <i>source</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpsd     xmm0, xmm0, less_than_or_equal}
     */
    // Template#: 989, Serial#: 3702
    public void cmpsd(IA32XMMRegister destination, IA32XMMRegister source, IA32XMMComparison ia32xmmcomparison) {
        assemble0221((byte) 0xC2, destination, source, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpsd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpsd     xmm0, ebx[esi * 4 + 305419896], less_than_or_equal}
     */
    // Template#: 990, Serial#: 3701
    public void cmpsd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        assemble0222((byte) 0xC2, destination, disp32, base, index, scale, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpsd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpsd     xmm0, L1: 0x12345678[esi * 4], less_than_or_equal}
     */
    // Template#: 991, Serial#: 3696
    public void m_cmpsd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        assemble0223((byte) 0xC2, destination, m32, index, scale, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpsd  }<i>destination</i>, <i>disp32</i>, <i>source</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpsd     xmm0, [ebx + 305419896], less_than_or_equal}
     */
    // Template#: 992, Serial#: 3700
    public void cmpsd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source, IA32XMMComparison ia32xmmcomparison) {
        assemble0224((byte) 0xC2, destination, disp32, source, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpsd  }<i>destination</i>, <i>m32</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpsd     xmm0, [L1: 0x12345678], less_than_or_equal}
     */
    // Template#: 993, Serial#: 3697
    public void m_cmpsd(IA32XMMRegister destination, int m32, IA32XMMComparison ia32xmmcomparison) {
        assemble0225((byte) 0xC2, destination, m32, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpss  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpss     xmm0, ebx[esi * 4 + 18], less_than_or_equal}
     */
    // Template#: 994, Serial#: 3708
    public void cmpss(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        assemble0226((byte) 0xC2, destination, disp8, base, index, scale, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpss  }<i>destination</i>, <i>disp8</i>, <i>source</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpss     xmm0, [ebx + 18], less_than_or_equal}
     */
    // Template#: 995, Serial#: 3707
    public void cmpss(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source, IA32XMMComparison ia32xmmcomparison) {
        assemble0227((byte) 0xC2, destination, disp8, source, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpss  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpss     xmm0, ebx[esi * 4], less_than_or_equal}
     */
    // Template#: 996, Serial#: 3704
    public void cmpss(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        assemble0228((byte) 0xC2, destination, base, index, scale, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpss  }<i>destination</i>, <i>source</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpss     xmm0, [ebx], less_than_or_equal}
     */
    // Template#: 997, Serial#: 3703
    public void cmpss(IA32XMMRegister destination, IA32IndirectRegister32 source, IA32XMMComparison ia32xmmcomparison) {
        assemble0229((byte) 0xC2, destination, source, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpss  }<i>destination</i>, <i>source</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpss     xmm0, xmm0, less_than_or_equal}
     */
    // Template#: 998, Serial#: 3711
    public void cmpss(IA32XMMRegister destination, IA32XMMRegister source, IA32XMMComparison ia32xmmcomparison) {
        assemble0230((byte) 0xC2, destination, source, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpss  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpss     xmm0, ebx[esi * 4 + 305419896], less_than_or_equal}
     */
    // Template#: 999, Serial#: 3710
    public void cmpss(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        assemble0231((byte) 0xC2, destination, disp32, base, index, scale, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpss  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpss     xmm0, L1: 0x12345678[esi * 4], less_than_or_equal}
     */
    // Template#: 1000, Serial#: 3705
    public void m_cmpss(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        assemble0232((byte) 0xC2, destination, m32, index, scale, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpss  }<i>destination</i>, <i>disp32</i>, <i>source</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpss     xmm0, [ebx + 305419896], less_than_or_equal}
     */
    // Template#: 1001, Serial#: 3709
    public void cmpss(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source, IA32XMMComparison ia32xmmcomparison) {
        assemble0233((byte) 0xC2, destination, disp32, source, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpss  }<i>destination</i>, <i>m32</i>, <i>ia32xmmcomparison</i>
     * Example disassembly syntax: {@code cmpss     xmm0, [L1: 0x12345678], less_than_or_equal}
     */
    // Template#: 1002, Serial#: 3706
    public void m_cmpss(IA32XMMRegister destination, int m32, IA32XMMComparison ia32xmmcomparison) {
        assemble0234((byte) 0xC2, destination, m32, ia32xmmcomparison);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   ebx[esi * 4 + 18], ax}
     */
    // Template#: 1003, Serial#: 3489
    public void cmpxchg(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0158((byte) 0xB1, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   ebx[esi * 4 + 18], eax}
     */
    // Template#: 1004, Serial#: 3480
    public void cmpxchg(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0159((byte) 0xB1, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   ebx[esi * 4 + 18], al}
     */
    // Template#: 1005, Serial#: 3471
    public void cmpxchg(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0235((byte) 0xB0, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   [ebx + 18], ax}
     */
    // Template#: 1006, Serial#: 3488
    public void cmpxchg(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0161((byte) 0xB1, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   [ebx + 18], eax}
     */
    // Template#: 1007, Serial#: 3479
    public void cmpxchg(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0162((byte) 0xB1, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   [ebx + 18], al}
     */
    // Template#: 1008, Serial#: 3470
    public void cmpxchg(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0236((byte) 0xB0, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   ebx[esi * 4], ax}
     */
    // Template#: 1009, Serial#: 3485
    public void cmpxchg(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0164((byte) 0xB1, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   ebx[esi * 4], eax}
     */
    // Template#: 1010, Serial#: 3476
    public void cmpxchg(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0165((byte) 0xB1, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   ebx[esi * 4], al}
     */
    // Template#: 1011, Serial#: 3467
    public void cmpxchg(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0237((byte) 0xB0, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   ax, ax}
     */
    // Template#: 1012, Serial#: 3492
    public void cmpxchg(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0167((byte) 0xB1, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   eax, eax}
     */
    // Template#: 1013, Serial#: 3483
    public void cmpxchg(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0169((byte) 0xB1, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   al, al}
     */
    // Template#: 1014, Serial#: 3474
    public void cmpxchg(IA32GeneralRegister8 destination, IA32GeneralRegister8 source) {
        assemble0238((byte) 0xB0, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   [ebx], ax}
     */
    // Template#: 1015, Serial#: 3484
    public void cmpxchg(IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0171((byte) 0xB1, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   [ebx], eax}
     */
    // Template#: 1016, Serial#: 3475
    public void cmpxchg(IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0172((byte) 0xB1, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   [ebx], al}
     */
    // Template#: 1017, Serial#: 3466
    public void cmpxchg(IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0239((byte) 0xB0, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   ebx[esi * 4 + 305419896], ax}
     */
    // Template#: 1018, Serial#: 3491
    public void cmpxchg(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0175((byte) 0xB1, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   ebx[esi * 4 + 305419896], eax}
     */
    // Template#: 1019, Serial#: 3482
    public void cmpxchg(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0176((byte) 0xB1, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   ebx[esi * 4 + 305419896], al}
     */
    // Template#: 1020, Serial#: 3473
    public void cmpxchg(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0240((byte) 0xB0, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   [L1: 0x12345678], ax}
     */
    // Template#: 1021, Serial#: 3487
    public void m_cmpxchg(int m32, IA32GeneralRegister16 source) {
        assemble0177((byte) 0xB1, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   [L1: 0x12345678], eax}
     */
    // Template#: 1022, Serial#: 3478
    public void m_cmpxchg(int m32, IA32GeneralRegister32 source) {
        assemble0178((byte) 0xB1, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   [L1: 0x12345678], al}
     */
    // Template#: 1023, Serial#: 3469
    public void m_cmpxchg(int m32, IA32GeneralRegister8 source) {
        assemble0241((byte) 0xB0, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   L1: 0x12345678[esi * 4], ax}
     */
    // Template#: 1024, Serial#: 3486
    public void m_cmpxchg(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0180((byte) 0xB1, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   L1: 0x12345678[esi * 4], eax}
     */
    // Template#: 1025, Serial#: 3477
    public void m_cmpxchg(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0181((byte) 0xB1, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   L1: 0x12345678[esi * 4], al}
     */
    // Template#: 1026, Serial#: 3468
    public void m_cmpxchg(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0242((byte) 0xB0, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   [ebx + 305419896], ax}
     */
    // Template#: 1027, Serial#: 3490
    public void cmpxchg(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0183((byte) 0xB1, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   [ebx + 305419896], eax}
     */
    // Template#: 1028, Serial#: 3481
    public void cmpxchg(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0184((byte) 0xB1, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cmpxchg   [ebx + 305419896], al}
     */
    // Template#: 1029, Serial#: 3472
    public void cmpxchg(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0243((byte) 0xB0, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg8b  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmpxchg8b  ebx[esi * 4 + 18]}
     */
    // Template#: 1030, Serial#: 3663
    public void cmpxchg8b(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0244((byte) 0xC7, (byte) 0x01, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg8b  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code cmpxchg8b  [ebx + 18]}
     */
    // Template#: 1031, Serial#: 3662
    public void cmpxchg8b(byte disp8, IA32IndirectRegister32 destination) {
        assemble0245((byte) 0xC7, (byte) 0x01, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg8b  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmpxchg8b  ebx[esi * 4]}
     */
    // Template#: 1032, Serial#: 3659
    public void cmpxchg8b(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0246((byte) 0xC7, (byte) 0x01, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg8b  }<i>destination</i>
     * Example disassembly syntax: {@code cmpxchg8b  [ebx]}
     */
    // Template#: 1033, Serial#: 3658
    public void cmpxchg8b(IA32IndirectRegister32 destination) {
        assemble0247((byte) 0xC7, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg8b  }<i>m32</i>
     * Example disassembly syntax: {@code cmpxchg8b  [L1: 0x12345678]}
     */
    // Template#: 1034, Serial#: 3661
    public void m_cmpxchg8b(int m32) {
        assemble0248((byte) 0xC7, (byte) 0x01, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg8b  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmpxchg8b  ebx[esi * 4 + 305419896]}
     */
    // Template#: 1035, Serial#: 3665
    public void cmpxchg8b(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0249((byte) 0xC7, (byte) 0x01, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg8b  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cmpxchg8b  L1: 0x12345678[esi * 4]}
     */
    // Template#: 1036, Serial#: 3660
    public void m_cmpxchg8b(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0250((byte) 0xC7, (byte) 0x01, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cmpxchg8b  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code cmpxchg8b  [ebx + 305419896]}
     */
    // Template#: 1037, Serial#: 3664
    public void cmpxchg8b(int disp32, IA32IndirectRegister32 destination) {
        assemble0251((byte) 0xC7, (byte) 0x01, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code comisd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code comisd    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1038, Serial#: 4168
    public void comisd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x2F, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code comisd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code comisd    xmm0, [ebx + 18]}
     */
    // Template#: 1039, Serial#: 4167
    public void comisd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x2F, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code comisd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code comisd    xmm0, ebx[esi * 4]}
     */
    // Template#: 1040, Serial#: 4164
    public void comisd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x2F, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code comisd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code comisd    xmm0, [ebx]}
     */
    // Template#: 1041, Serial#: 4163
    public void comisd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x2F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code comisd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code comisd    xmm0, xmm0}
     */
    // Template#: 1042, Serial#: 4171
    public void comisd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x2F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code comisd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code comisd    xmm0, [L1: 0x12345678]}
     */
    // Template#: 1043, Serial#: 4166
    public void m_comisd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x2F, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code comisd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code comisd    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1044, Serial#: 4170
    public void comisd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x2F, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code comisd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code comisd    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1045, Serial#: 4165
    public void m_comisd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x2F, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code comisd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code comisd    xmm0, [ebx + 305419896]}
     */
    // Template#: 1046, Serial#: 4169
    public void comisd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x2F, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code comiss  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code comiss    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1047, Serial#: 4098
    public void comiss(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0102((byte) 0x2F, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code comiss  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code comiss    xmm0, [ebx + 18]}
     */
    // Template#: 1048, Serial#: 4097
    public void comiss(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0103((byte) 0x2F, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code comiss  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code comiss    xmm0, ebx[esi * 4]}
     */
    // Template#: 1049, Serial#: 4094
    public void comiss(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0104((byte) 0x2F, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code comiss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code comiss    xmm0, [ebx]}
     */
    // Template#: 1050, Serial#: 4093
    public void comiss(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0105((byte) 0x2F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code comiss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code comiss    xmm0, xmm0}
     */
    // Template#: 1051, Serial#: 4101
    public void comiss(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0106((byte) 0x2F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code comiss  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code comiss    xmm0, [L1: 0x12345678]}
     */
    // Template#: 1052, Serial#: 4096
    public void m_comiss(IA32XMMRegister destination, int m32) {
        assemble0107((byte) 0x2F, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code comiss  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code comiss    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1053, Serial#: 4100
    public void comiss(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0108((byte) 0x2F, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code comiss  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code comiss    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1054, Serial#: 4095
    public void m_comiss(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0109((byte) 0x2F, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code comiss  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code comiss    xmm0, [ebx + 305419896]}
     */
    // Template#: 1055, Serial#: 4099
    public void comiss(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0110((byte) 0x2F, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cpuid  }
     * Example disassembly syntax: {@code cpuid   }
     */
    // Template#: 1056, Serial#: 3411
    public void cpuid() {
        assemble0198((byte) 0xA2);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtdq2pd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtdq2pd  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1057, Serial#: 3926
    public void cvtdq2pd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0120((byte) 0xE6, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtdq2pd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtdq2pd  xmm0, [ebx + 18]}
     */
    // Template#: 1058, Serial#: 3925
    public void cvtdq2pd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0121((byte) 0xE6, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtdq2pd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtdq2pd  xmm0, ebx[esi * 4]}
     */
    // Template#: 1059, Serial#: 3922
    public void cvtdq2pd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0122((byte) 0xE6, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtdq2pd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtdq2pd  xmm0, [ebx]}
     */
    // Template#: 1060, Serial#: 3921
    public void cvtdq2pd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0123((byte) 0xE6, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtdq2pd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtdq2pd  xmm0, xmm0}
     */
    // Template#: 1061, Serial#: 3929
    public void cvtdq2pd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0124((byte) 0xE6, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtdq2pd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cvtdq2pd  xmm0, [L1: 0x12345678]}
     */
    // Template#: 1062, Serial#: 3924
    public void m_cvtdq2pd(IA32XMMRegister destination, int m32) {
        assemble0125((byte) 0xE6, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtdq2pd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtdq2pd  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1063, Serial#: 3928
    public void cvtdq2pd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0126((byte) 0xE6, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtdq2pd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtdq2pd  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1064, Serial#: 3923
    public void m_cvtdq2pd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0127((byte) 0xE6, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtdq2pd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtdq2pd  xmm0, [ebx + 305419896]}
     */
    // Template#: 1065, Serial#: 3927
    public void cvtdq2pd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0128((byte) 0xE6, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtdq2ps  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtdq2ps  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1066, Serial#: 4402
    public void cvtdq2ps(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0102((byte) 0x5B, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtdq2ps  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtdq2ps  xmm0, [ebx + 18]}
     */
    // Template#: 1067, Serial#: 4401
    public void cvtdq2ps(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0103((byte) 0x5B, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtdq2ps  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtdq2ps  xmm0, ebx[esi * 4]}
     */
    // Template#: 1068, Serial#: 4398
    public void cvtdq2ps(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0104((byte) 0x5B, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtdq2ps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtdq2ps  xmm0, [ebx]}
     */
    // Template#: 1069, Serial#: 4397
    public void cvtdq2ps(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0105((byte) 0x5B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtdq2ps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtdq2ps  xmm0, xmm0}
     */
    // Template#: 1070, Serial#: 4405
    public void cvtdq2ps(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0106((byte) 0x5B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtdq2ps  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cvtdq2ps  xmm0, [L1: 0x12345678]}
     */
    // Template#: 1071, Serial#: 4400
    public void m_cvtdq2ps(IA32XMMRegister destination, int m32) {
        assemble0107((byte) 0x5B, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtdq2ps  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtdq2ps  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1072, Serial#: 4404
    public void cvtdq2ps(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0108((byte) 0x5B, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtdq2ps  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtdq2ps  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1073, Serial#: 4399
    public void m_cvtdq2ps(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0109((byte) 0x5B, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtdq2ps  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtdq2ps  xmm0, [ebx + 305419896]}
     */
    // Template#: 1074, Serial#: 4403
    public void cvtdq2ps(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0110((byte) 0x5B, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2dq  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtpd2dq  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1075, Serial#: 3917
    public void cvtpd2dq(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0111((byte) 0xE6, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2dq  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtpd2dq  xmm0, [ebx + 18]}
     */
    // Template#: 1076, Serial#: 3916
    public void cvtpd2dq(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0112((byte) 0xE6, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2dq  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtpd2dq  xmm0, ebx[esi * 4]}
     */
    // Template#: 1077, Serial#: 3913
    public void cvtpd2dq(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0113((byte) 0xE6, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2dq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtpd2dq  xmm0, [ebx]}
     */
    // Template#: 1078, Serial#: 3912
    public void cvtpd2dq(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0114((byte) 0xE6, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2dq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtpd2dq  xmm0, xmm0}
     */
    // Template#: 1079, Serial#: 3920
    public void cvtpd2dq(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0115((byte) 0xE6, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2dq  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cvtpd2dq  xmm0, [L1: 0x12345678]}
     */
    // Template#: 1080, Serial#: 3915
    public void m_cvtpd2dq(IA32XMMRegister destination, int m32) {
        assemble0116((byte) 0xE6, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2dq  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtpd2dq  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1081, Serial#: 3919
    public void cvtpd2dq(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0117((byte) 0xE6, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2dq  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtpd2dq  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1082, Serial#: 3914
    public void m_cvtpd2dq(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0118((byte) 0xE6, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2dq  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtpd2dq  xmm0, [ebx + 305419896]}
     */
    // Template#: 1083, Serial#: 3918
    public void cvtpd2dq(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0119((byte) 0xE6, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2pi  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtpd2pi  mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1084, Serial#: 4150
    public void cvtpd2pi(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0252((byte) 0x2D, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2pi  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtpd2pi  mm0, [ebx + 18]}
     */
    // Template#: 1085, Serial#: 4149
    public void cvtpd2pi(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0253((byte) 0x2D, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2pi  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtpd2pi  mm0, ebx[esi * 4]}
     */
    // Template#: 1086, Serial#: 4146
    public void cvtpd2pi(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0254((byte) 0x2D, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2pi  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtpd2pi  mm0, [ebx]}
     */
    // Template#: 1087, Serial#: 4145
    public void cvtpd2pi(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0255((byte) 0x2D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2pi  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtpd2pi  mm0, xmm0}
     */
    // Template#: 1088, Serial#: 4153
    public void cvtpd2pi(MMXRegister destination, IA32XMMRegister source) {
        assemble0256((byte) 0x2D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2pi  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cvtpd2pi  mm0, [L1: 0x12345678]}
     */
    // Template#: 1089, Serial#: 4148
    public void m_cvtpd2pi(MMXRegister destination, int m32) {
        assemble0257((byte) 0x2D, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2pi  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtpd2pi  mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1090, Serial#: 4152
    public void cvtpd2pi(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0258((byte) 0x2D, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2pi  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtpd2pi  mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1091, Serial#: 4147
    public void m_cvtpd2pi(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0259((byte) 0x2D, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2pi  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtpd2pi  mm0, [ebx + 305419896]}
     */
    // Template#: 1092, Serial#: 4151
    public void cvtpd2pi(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0260((byte) 0x2D, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2ps  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtpd2ps  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1093, Serial#: 4465
    public void cvtpd2ps(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x5A, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2ps  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtpd2ps  xmm0, [ebx + 18]}
     */
    // Template#: 1094, Serial#: 4464
    public void cvtpd2ps(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x5A, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2ps  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtpd2ps  xmm0, ebx[esi * 4]}
     */
    // Template#: 1095, Serial#: 4461
    public void cvtpd2ps(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x5A, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2ps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtpd2ps  xmm0, [ebx]}
     */
    // Template#: 1096, Serial#: 4460
    public void cvtpd2ps(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x5A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2ps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtpd2ps  xmm0, xmm0}
     */
    // Template#: 1097, Serial#: 4468
    public void cvtpd2ps(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x5A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2ps  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cvtpd2ps  xmm0, [L1: 0x12345678]}
     */
    // Template#: 1098, Serial#: 4463
    public void m_cvtpd2ps(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x5A, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2ps  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtpd2ps  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1099, Serial#: 4467
    public void cvtpd2ps(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x5A, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2ps  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtpd2ps  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1100, Serial#: 4462
    public void m_cvtpd2ps(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x5A, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpd2ps  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtpd2ps  xmm0, [ebx + 305419896]}
     */
    // Template#: 1101, Serial#: 4466
    public void cvtpd2ps(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x5A, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpi2pd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtpi2pd  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1102, Serial#: 4124
    public void cvtpi2pd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x2A, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpi2pd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtpi2pd  xmm0, [ebx + 18]}
     */
    // Template#: 1103, Serial#: 4123
    public void cvtpi2pd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x2A, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpi2pd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtpi2pd  xmm0, ebx[esi * 4]}
     */
    // Template#: 1104, Serial#: 4120
    public void cvtpi2pd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x2A, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpi2pd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtpi2pd  xmm0, [ebx]}
     */
    // Template#: 1105, Serial#: 4119
    public void cvtpi2pd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x2A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpi2pd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtpi2pd  xmm0, mm0}
     */
    // Template#: 1106, Serial#: 4127
    public void cvtpi2pd(IA32XMMRegister destination, MMXRegister source) {
        assemble0261((byte) 0x2A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpi2pd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cvtpi2pd  xmm0, [L1: 0x12345678]}
     */
    // Template#: 1107, Serial#: 4122
    public void m_cvtpi2pd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x2A, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpi2pd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtpi2pd  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1108, Serial#: 4126
    public void cvtpi2pd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x2A, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpi2pd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtpi2pd  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1109, Serial#: 4121
    public void m_cvtpi2pd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x2A, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpi2pd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtpi2pd  xmm0, [ebx + 305419896]}
     */
    // Template#: 1110, Serial#: 4125
    public void cvtpi2pd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x2A, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpi2ps  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtpi2ps  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1111, Serial#: 4054
    public void cvtpi2ps(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0102((byte) 0x2A, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpi2ps  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtpi2ps  xmm0, [ebx + 18]}
     */
    // Template#: 1112, Serial#: 4053
    public void cvtpi2ps(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0103((byte) 0x2A, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpi2ps  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtpi2ps  xmm0, ebx[esi * 4]}
     */
    // Template#: 1113, Serial#: 4050
    public void cvtpi2ps(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0104((byte) 0x2A, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpi2ps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtpi2ps  xmm0, [ebx]}
     */
    // Template#: 1114, Serial#: 4049
    public void cvtpi2ps(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0105((byte) 0x2A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpi2ps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtpi2ps  xmm0, mm0}
     */
    // Template#: 1115, Serial#: 4057
    public void cvtpi2ps(IA32XMMRegister destination, MMXRegister source) {
        assemble0262((byte) 0x2A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpi2ps  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cvtpi2ps  xmm0, [L1: 0x12345678]}
     */
    // Template#: 1116, Serial#: 4052
    public void m_cvtpi2ps(IA32XMMRegister destination, int m32) {
        assemble0107((byte) 0x2A, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpi2ps  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtpi2ps  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1117, Serial#: 4056
    public void cvtpi2ps(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0108((byte) 0x2A, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpi2ps  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtpi2ps  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1118, Serial#: 4051
    public void m_cvtpi2ps(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0109((byte) 0x2A, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtpi2ps  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtpi2ps  xmm0, [ebx + 305419896]}
     */
    // Template#: 1119, Serial#: 4055
    public void cvtpi2ps(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0110((byte) 0x2A, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2dq  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtps2dq  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1120, Serial#: 4474
    public void cvtps2dq(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x5B, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2dq  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtps2dq  xmm0, [ebx + 18]}
     */
    // Template#: 1121, Serial#: 4473
    public void cvtps2dq(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x5B, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2dq  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtps2dq  xmm0, ebx[esi * 4]}
     */
    // Template#: 1122, Serial#: 4470
    public void cvtps2dq(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x5B, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2dq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtps2dq  xmm0, [ebx]}
     */
    // Template#: 1123, Serial#: 4469
    public void cvtps2dq(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x5B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2dq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtps2dq  xmm0, xmm0}
     */
    // Template#: 1124, Serial#: 4477
    public void cvtps2dq(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x5B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2dq  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cvtps2dq  xmm0, [L1: 0x12345678]}
     */
    // Template#: 1125, Serial#: 4472
    public void m_cvtps2dq(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x5B, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2dq  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtps2dq  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1126, Serial#: 4476
    public void cvtps2dq(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x5B, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2dq  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtps2dq  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1127, Serial#: 4471
    public void m_cvtps2dq(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x5B, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2dq  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtps2dq  xmm0, [ebx + 305419896]}
     */
    // Template#: 1128, Serial#: 4475
    public void cvtps2dq(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x5B, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2pd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtps2pd  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1129, Serial#: 4393
    public void cvtps2pd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0102((byte) 0x5A, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2pd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtps2pd  xmm0, [ebx + 18]}
     */
    // Template#: 1130, Serial#: 4392
    public void cvtps2pd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0103((byte) 0x5A, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2pd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtps2pd  xmm0, ebx[esi * 4]}
     */
    // Template#: 1131, Serial#: 4389
    public void cvtps2pd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0104((byte) 0x5A, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2pd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtps2pd  xmm0, [ebx]}
     */
    // Template#: 1132, Serial#: 4388
    public void cvtps2pd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0105((byte) 0x5A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2pd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtps2pd  xmm0, xmm0}
     */
    // Template#: 1133, Serial#: 4396
    public void cvtps2pd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0106((byte) 0x5A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2pd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cvtps2pd  xmm0, [L1: 0x12345678]}
     */
    // Template#: 1134, Serial#: 4391
    public void m_cvtps2pd(IA32XMMRegister destination, int m32) {
        assemble0107((byte) 0x5A, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2pd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtps2pd  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1135, Serial#: 4395
    public void cvtps2pd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0108((byte) 0x5A, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2pd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtps2pd  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1136, Serial#: 4390
    public void m_cvtps2pd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0109((byte) 0x5A, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2pd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtps2pd  xmm0, [ebx + 305419896]}
     */
    // Template#: 1137, Serial#: 4394
    public void cvtps2pd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0110((byte) 0x5A, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2pi  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtps2pi  mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1138, Serial#: 4080
    public void cvtps2pi(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0x2D, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2pi  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtps2pi  mm0, [ebx + 18]}
     */
    // Template#: 1139, Serial#: 4079
    public void cvtps2pi(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0x2D, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2pi  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtps2pi  mm0, ebx[esi * 4]}
     */
    // Template#: 1140, Serial#: 4076
    public void cvtps2pi(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0x2D, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2pi  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtps2pi  mm0, [ebx]}
     */
    // Template#: 1141, Serial#: 4075
    public void cvtps2pi(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0x2D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2pi  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtps2pi  mm0, xmm0}
     */
    // Template#: 1142, Serial#: 4083
    public void cvtps2pi(MMXRegister destination, IA32XMMRegister source) {
        assemble0267((byte) 0x2D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2pi  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cvtps2pi  mm0, [L1: 0x12345678]}
     */
    // Template#: 1143, Serial#: 4078
    public void m_cvtps2pi(MMXRegister destination, int m32) {
        assemble0268((byte) 0x2D, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2pi  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtps2pi  mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1144, Serial#: 4082
    public void cvtps2pi(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0x2D, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2pi  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtps2pi  mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1145, Serial#: 4077
    public void m_cvtps2pi(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0x2D, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtps2pi  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtps2pi  mm0, [ebx + 305419896]}
     */
    // Template#: 1146, Serial#: 4081
    public void cvtps2pi(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0x2D, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsd2si  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtsd2si  eax, ebx[esi * 4 + 18]}
     */
    // Template#: 1147, Serial#: 4195
    public void cvtsd2si(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0272((byte) 0x2D, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsd2si  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtsd2si  eax, [ebx + 18]}
     */
    // Template#: 1148, Serial#: 4194
    public void cvtsd2si(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0273((byte) 0x2D, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsd2si  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtsd2si  eax, ebx[esi * 4]}
     */
    // Template#: 1149, Serial#: 4191
    public void cvtsd2si(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0274((byte) 0x2D, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsd2si  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtsd2si  eax, [ebx]}
     */
    // Template#: 1150, Serial#: 4190
    public void cvtsd2si(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0275((byte) 0x2D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsd2si  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtsd2si  eax, xmm0}
     */
    // Template#: 1151, Serial#: 4198
    public void cvtsd2si(IA32GeneralRegister32 destination, IA32XMMRegister source) {
        assemble0276((byte) 0x2D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsd2si  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cvtsd2si  eax, [L1: 0x12345678]}
     */
    // Template#: 1152, Serial#: 4193
    public void m_cvtsd2si(IA32GeneralRegister32 destination, int m32) {
        assemble0277((byte) 0x2D, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsd2si  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtsd2si  eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1153, Serial#: 4197
    public void cvtsd2si(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0278((byte) 0x2D, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsd2si  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtsd2si  eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1154, Serial#: 4192
    public void m_cvtsd2si(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0279((byte) 0x2D, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsd2si  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtsd2si  eax, [ebx + 305419896]}
     */
    // Template#: 1155, Serial#: 4196
    public void cvtsd2si(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0280((byte) 0x2D, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsd2ss  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtsd2ss  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1156, Serial#: 4537
    public void cvtsd2ss(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0111((byte) 0x5A, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsd2ss  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtsd2ss  xmm0, [ebx + 18]}
     */
    // Template#: 1157, Serial#: 4536
    public void cvtsd2ss(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0112((byte) 0x5A, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsd2ss  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtsd2ss  xmm0, ebx[esi * 4]}
     */
    // Template#: 1158, Serial#: 4533
    public void cvtsd2ss(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0113((byte) 0x5A, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsd2ss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtsd2ss  xmm0, [ebx]}
     */
    // Template#: 1159, Serial#: 4532
    public void cvtsd2ss(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0114((byte) 0x5A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsd2ss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtsd2ss  xmm0, xmm0}
     */
    // Template#: 1160, Serial#: 4540
    public void cvtsd2ss(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0115((byte) 0x5A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsd2ss  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cvtsd2ss  xmm0, [L1: 0x12345678]}
     */
    // Template#: 1161, Serial#: 4535
    public void m_cvtsd2ss(IA32XMMRegister destination, int m32) {
        assemble0116((byte) 0x5A, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsd2ss  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtsd2ss  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1162, Serial#: 4539
    public void cvtsd2ss(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0117((byte) 0x5A, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsd2ss  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtsd2ss  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1163, Serial#: 4534
    public void m_cvtsd2ss(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0118((byte) 0x5A, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsd2ss  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtsd2ss  xmm0, [ebx + 305419896]}
     */
    // Template#: 1164, Serial#: 4538
    public void cvtsd2ss(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0119((byte) 0x5A, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsi2sd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtsi2sd  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1165, Serial#: 4177
    public void cvtsi2sd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0111((byte) 0x2A, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsi2sd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtsi2sd  xmm0, [ebx + 18]}
     */
    // Template#: 1166, Serial#: 4176
    public void cvtsi2sd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0112((byte) 0x2A, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsi2sd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtsi2sd  xmm0, ebx[esi * 4]}
     */
    // Template#: 1167, Serial#: 4173
    public void cvtsi2sd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0113((byte) 0x2A, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsi2sd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtsi2sd  xmm0, eax}
     */
    // Template#: 1168, Serial#: 4180
    public void cvtsi2sd(IA32XMMRegister destination, IA32GeneralRegister32 source) {
        assemble0281((byte) 0x2A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsi2sd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtsi2sd  xmm0, [ebx]}
     */
    // Template#: 1169, Serial#: 4172
    public void cvtsi2sd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0114((byte) 0x2A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsi2sd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cvtsi2sd  xmm0, [L1: 0x12345678]}
     */
    // Template#: 1170, Serial#: 4175
    public void m_cvtsi2sd(IA32XMMRegister destination, int m32) {
        assemble0116((byte) 0x2A, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsi2sd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtsi2sd  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1171, Serial#: 4179
    public void cvtsi2sd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0117((byte) 0x2A, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsi2sd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtsi2sd  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1172, Serial#: 4174
    public void m_cvtsi2sd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0118((byte) 0x2A, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsi2sd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtsi2sd  xmm0, [ebx + 305419896]}
     */
    // Template#: 1173, Serial#: 4178
    public void cvtsi2sd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0119((byte) 0x2A, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsi2ss  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtsi2ss  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1174, Serial#: 4204
    public void cvtsi2ss(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0120((byte) 0x2A, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsi2ss  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtsi2ss  xmm0, [ebx + 18]}
     */
    // Template#: 1175, Serial#: 4203
    public void cvtsi2ss(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0121((byte) 0x2A, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsi2ss  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtsi2ss  xmm0, ebx[esi * 4]}
     */
    // Template#: 1176, Serial#: 4200
    public void cvtsi2ss(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0122((byte) 0x2A, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsi2ss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtsi2ss  xmm0, eax}
     */
    // Template#: 1177, Serial#: 4207
    public void cvtsi2ss(IA32XMMRegister destination, IA32GeneralRegister32 source) {
        assemble0282((byte) 0x2A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsi2ss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtsi2ss  xmm0, [ebx]}
     */
    // Template#: 1178, Serial#: 4199
    public void cvtsi2ss(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0123((byte) 0x2A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsi2ss  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cvtsi2ss  xmm0, [L1: 0x12345678]}
     */
    // Template#: 1179, Serial#: 4202
    public void m_cvtsi2ss(IA32XMMRegister destination, int m32) {
        assemble0125((byte) 0x2A, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsi2ss  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtsi2ss  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1180, Serial#: 4206
    public void cvtsi2ss(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0126((byte) 0x2A, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsi2ss  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtsi2ss  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1181, Serial#: 4201
    public void m_cvtsi2ss(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0127((byte) 0x2A, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtsi2ss  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtsi2ss  xmm0, [ebx + 305419896]}
     */
    // Template#: 1182, Serial#: 4205
    public void cvtsi2ss(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0128((byte) 0x2A, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtss2sd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtss2sd  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1183, Serial#: 4600
    public void cvtss2sd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0120((byte) 0x5A, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtss2sd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtss2sd  xmm0, [ebx + 18]}
     */
    // Template#: 1184, Serial#: 4599
    public void cvtss2sd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0121((byte) 0x5A, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtss2sd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtss2sd  xmm0, ebx[esi * 4]}
     */
    // Template#: 1185, Serial#: 4596
    public void cvtss2sd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0122((byte) 0x5A, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtss2sd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtss2sd  xmm0, [ebx]}
     */
    // Template#: 1186, Serial#: 4595
    public void cvtss2sd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0123((byte) 0x5A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtss2sd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtss2sd  xmm0, xmm0}
     */
    // Template#: 1187, Serial#: 4603
    public void cvtss2sd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0124((byte) 0x5A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtss2sd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cvtss2sd  xmm0, [L1: 0x12345678]}
     */
    // Template#: 1188, Serial#: 4598
    public void m_cvtss2sd(IA32XMMRegister destination, int m32) {
        assemble0125((byte) 0x5A, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtss2sd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtss2sd  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1189, Serial#: 4602
    public void cvtss2sd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0126((byte) 0x5A, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtss2sd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtss2sd  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1190, Serial#: 4597
    public void m_cvtss2sd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0127((byte) 0x5A, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtss2sd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtss2sd  xmm0, [ebx + 305419896]}
     */
    // Template#: 1191, Serial#: 4601
    public void cvtss2sd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0128((byte) 0x5A, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtss2si  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtss2si  eax, ebx[esi * 4 + 18]}
     */
    // Template#: 1192, Serial#: 4222
    public void cvtss2si(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0283((byte) 0x2D, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtss2si  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtss2si  eax, [ebx + 18]}
     */
    // Template#: 1193, Serial#: 4221
    public void cvtss2si(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0284((byte) 0x2D, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtss2si  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtss2si  eax, ebx[esi * 4]}
     */
    // Template#: 1194, Serial#: 4218
    public void cvtss2si(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0285((byte) 0x2D, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtss2si  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtss2si  eax, [ebx]}
     */
    // Template#: 1195, Serial#: 4217
    public void cvtss2si(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0286((byte) 0x2D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtss2si  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtss2si  eax, xmm0}
     */
    // Template#: 1196, Serial#: 4225
    public void cvtss2si(IA32GeneralRegister32 destination, IA32XMMRegister source) {
        assemble0287((byte) 0x2D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtss2si  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cvtss2si  eax, [L1: 0x12345678]}
     */
    // Template#: 1197, Serial#: 4220
    public void m_cvtss2si(IA32GeneralRegister32 destination, int m32) {
        assemble0288((byte) 0x2D, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtss2si  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtss2si  eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1198, Serial#: 4224
    public void cvtss2si(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0289((byte) 0x2D, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtss2si  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvtss2si  eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1199, Serial#: 4219
    public void m_cvtss2si(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0290((byte) 0x2D, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvtss2si  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cvtss2si  eax, [ebx + 305419896]}
     */
    // Template#: 1200, Serial#: 4223
    public void cvtss2si(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0291((byte) 0x2D, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttpd2dq  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvttpd2dq  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1201, Serial#: 3900
    public void cvttpd2dq(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xE6, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttpd2dq  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cvttpd2dq  xmm0, [ebx + 18]}
     */
    // Template#: 1202, Serial#: 3899
    public void cvttpd2dq(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xE6, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttpd2dq  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvttpd2dq  xmm0, ebx[esi * 4]}
     */
    // Template#: 1203, Serial#: 3896
    public void cvttpd2dq(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xE6, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttpd2dq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvttpd2dq  xmm0, [ebx]}
     */
    // Template#: 1204, Serial#: 3895
    public void cvttpd2dq(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xE6, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttpd2dq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvttpd2dq  xmm0, xmm0}
     */
    // Template#: 1205, Serial#: 3903
    public void cvttpd2dq(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xE6, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttpd2dq  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cvttpd2dq  xmm0, [L1: 0x12345678]}
     */
    // Template#: 1206, Serial#: 3898
    public void m_cvttpd2dq(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xE6, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttpd2dq  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvttpd2dq  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1207, Serial#: 3902
    public void cvttpd2dq(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xE6, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttpd2dq  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvttpd2dq  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1208, Serial#: 3897
    public void m_cvttpd2dq(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xE6, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttpd2dq  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cvttpd2dq  xmm0, [ebx + 305419896]}
     */
    // Template#: 1209, Serial#: 3901
    public void cvttpd2dq(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xE6, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttpd2pi  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvttpd2pi  mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1210, Serial#: 4141
    public void cvttpd2pi(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0252((byte) 0x2C, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttpd2pi  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cvttpd2pi  mm0, [ebx + 18]}
     */
    // Template#: 1211, Serial#: 4140
    public void cvttpd2pi(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0253((byte) 0x2C, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttpd2pi  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvttpd2pi  mm0, ebx[esi * 4]}
     */
    // Template#: 1212, Serial#: 4137
    public void cvttpd2pi(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0254((byte) 0x2C, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttpd2pi  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvttpd2pi  mm0, [ebx]}
     */
    // Template#: 1213, Serial#: 4136
    public void cvttpd2pi(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0255((byte) 0x2C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttpd2pi  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvttpd2pi  mm0, xmm0}
     */
    // Template#: 1214, Serial#: 4144
    public void cvttpd2pi(MMXRegister destination, IA32XMMRegister source) {
        assemble0256((byte) 0x2C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttpd2pi  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cvttpd2pi  mm0, [L1: 0x12345678]}
     */
    // Template#: 1215, Serial#: 4139
    public void m_cvttpd2pi(MMXRegister destination, int m32) {
        assemble0257((byte) 0x2C, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttpd2pi  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvttpd2pi  mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1216, Serial#: 4143
    public void cvttpd2pi(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0258((byte) 0x2C, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttpd2pi  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvttpd2pi  mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1217, Serial#: 4138
    public void m_cvttpd2pi(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0259((byte) 0x2C, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttpd2pi  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cvttpd2pi  mm0, [ebx + 305419896]}
     */
    // Template#: 1218, Serial#: 4142
    public void cvttpd2pi(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0260((byte) 0x2C, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttps2dq  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvttps2dq  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1219, Serial#: 4609
    public void cvttps2dq(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0120((byte) 0x5B, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttps2dq  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cvttps2dq  xmm0, [ebx + 18]}
     */
    // Template#: 1220, Serial#: 4608
    public void cvttps2dq(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0121((byte) 0x5B, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttps2dq  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvttps2dq  xmm0, ebx[esi * 4]}
     */
    // Template#: 1221, Serial#: 4605
    public void cvttps2dq(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0122((byte) 0x5B, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttps2dq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvttps2dq  xmm0, [ebx]}
     */
    // Template#: 1222, Serial#: 4604
    public void cvttps2dq(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0123((byte) 0x5B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttps2dq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvttps2dq  xmm0, xmm0}
     */
    // Template#: 1223, Serial#: 4612
    public void cvttps2dq(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0124((byte) 0x5B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttps2dq  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cvttps2dq  xmm0, [L1: 0x12345678]}
     */
    // Template#: 1224, Serial#: 4607
    public void m_cvttps2dq(IA32XMMRegister destination, int m32) {
        assemble0125((byte) 0x5B, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttps2dq  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvttps2dq  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1225, Serial#: 4611
    public void cvttps2dq(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0126((byte) 0x5B, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttps2dq  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvttps2dq  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1226, Serial#: 4606
    public void m_cvttps2dq(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0127((byte) 0x5B, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttps2dq  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cvttps2dq  xmm0, [ebx + 305419896]}
     */
    // Template#: 1227, Serial#: 4610
    public void cvttps2dq(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0128((byte) 0x5B, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttps2pi  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvttps2pi  mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1228, Serial#: 4071
    public void cvttps2pi(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0x2C, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttps2pi  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cvttps2pi  mm0, [ebx + 18]}
     */
    // Template#: 1229, Serial#: 4070
    public void cvttps2pi(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0x2C, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttps2pi  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvttps2pi  mm0, ebx[esi * 4]}
     */
    // Template#: 1230, Serial#: 4067
    public void cvttps2pi(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0x2C, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttps2pi  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvttps2pi  mm0, [ebx]}
     */
    // Template#: 1231, Serial#: 4066
    public void cvttps2pi(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0x2C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttps2pi  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvttps2pi  mm0, xmm0}
     */
    // Template#: 1232, Serial#: 4074
    public void cvttps2pi(MMXRegister destination, IA32XMMRegister source) {
        assemble0267((byte) 0x2C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttps2pi  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cvttps2pi  mm0, [L1: 0x12345678]}
     */
    // Template#: 1233, Serial#: 4069
    public void m_cvttps2pi(MMXRegister destination, int m32) {
        assemble0268((byte) 0x2C, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttps2pi  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvttps2pi  mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1234, Serial#: 4073
    public void cvttps2pi(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0x2C, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttps2pi  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvttps2pi  mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1235, Serial#: 4068
    public void m_cvttps2pi(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0x2C, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttps2pi  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cvttps2pi  mm0, [ebx + 305419896]}
     */
    // Template#: 1236, Serial#: 4072
    public void cvttps2pi(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0x2C, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttsd2si  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvttsd2si  eax, ebx[esi * 4 + 18]}
     */
    // Template#: 1237, Serial#: 4186
    public void cvttsd2si(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0272((byte) 0x2C, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttsd2si  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cvttsd2si  eax, [ebx + 18]}
     */
    // Template#: 1238, Serial#: 4185
    public void cvttsd2si(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0273((byte) 0x2C, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttsd2si  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvttsd2si  eax, ebx[esi * 4]}
     */
    // Template#: 1239, Serial#: 4182
    public void cvttsd2si(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0274((byte) 0x2C, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttsd2si  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvttsd2si  eax, [ebx]}
     */
    // Template#: 1240, Serial#: 4181
    public void cvttsd2si(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0275((byte) 0x2C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttsd2si  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvttsd2si  eax, xmm0}
     */
    // Template#: 1241, Serial#: 4189
    public void cvttsd2si(IA32GeneralRegister32 destination, IA32XMMRegister source) {
        assemble0276((byte) 0x2C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttsd2si  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cvttsd2si  eax, [L1: 0x12345678]}
     */
    // Template#: 1242, Serial#: 4184
    public void m_cvttsd2si(IA32GeneralRegister32 destination, int m32) {
        assemble0277((byte) 0x2C, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttsd2si  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvttsd2si  eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1243, Serial#: 4188
    public void cvttsd2si(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0278((byte) 0x2C, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttsd2si  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvttsd2si  eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1244, Serial#: 4183
    public void m_cvttsd2si(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0279((byte) 0x2C, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttsd2si  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cvttsd2si  eax, [ebx + 305419896]}
     */
    // Template#: 1245, Serial#: 4187
    public void cvttsd2si(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0280((byte) 0x2C, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttss2si  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvttss2si  eax, ebx[esi * 4 + 18]}
     */
    // Template#: 1246, Serial#: 4213
    public void cvttss2si(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0283((byte) 0x2C, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttss2si  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code cvttss2si  eax, [ebx + 18]}
     */
    // Template#: 1247, Serial#: 4212
    public void cvttss2si(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0284((byte) 0x2C, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttss2si  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvttss2si  eax, ebx[esi * 4]}
     */
    // Template#: 1248, Serial#: 4209
    public void cvttss2si(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0285((byte) 0x2C, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttss2si  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvttss2si  eax, [ebx]}
     */
    // Template#: 1249, Serial#: 4208
    public void cvttss2si(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0286((byte) 0x2C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttss2si  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code cvttss2si  eax, xmm0}
     */
    // Template#: 1250, Serial#: 4216
    public void cvttss2si(IA32GeneralRegister32 destination, IA32XMMRegister source) {
        assemble0287((byte) 0x2C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttss2si  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code cvttss2si  eax, [L1: 0x12345678]}
     */
    // Template#: 1251, Serial#: 4211
    public void m_cvttss2si(IA32GeneralRegister32 destination, int m32) {
        assemble0288((byte) 0x2C, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttss2si  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvttss2si  eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1252, Serial#: 4215
    public void cvttss2si(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0289((byte) 0x2C, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttss2si  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code cvttss2si  eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1253, Serial#: 4210
    public void m_cvttss2si(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0290((byte) 0x2C, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code cvttss2si  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code cvttss2si  eax, [ebx + 305419896]}
     */
    // Template#: 1254, Serial#: 4214
    public void cvttss2si(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0291((byte) 0x2C, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code cwd  }
     * Example disassembly syntax: {@code cwd     }
     */
    // Template#: 1255, Serial#: 1911
    public void cwd() {
        assemble0197((byte) 0x99);
    }

    /**
     * Pseudo-external assembler syntax: {@code cwde  }
     * Example disassembly syntax: {@code cwde    }
     */
    // Template#: 1256, Serial#: 1910
    public void cwde() {
        assemble0001((byte) 0x98);
    }

    /**
     * Pseudo-external assembler syntax: {@code daa  }
     * Example disassembly syntax: {@code daa     }
     */
    // Template#: 1257, Serial#: 168
    public void daa() {
        assemble0001((byte) 0x27);
    }

    /**
     * Pseudo-external assembler syntax: {@code das  }
     * Example disassembly syntax: {@code das     }
     */
    // Template#: 1258, Serial#: 1702
    public void das() {
        assemble0001((byte) 0x2F);
    }

    /**
     * Pseudo-external assembler syntax: {@code decb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code decb      ebx[esi * 4 + 18]}
     */
    // Template#: 1259, Serial#: 2427
    public void decb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xFE, (byte) 0x01, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code decl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code decl      ebx[esi * 4 + 18]}
     */
    // Template#: 1260, Serial#: 2455
    public void decl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xFF, (byte) 0x01, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code decw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code decw      ebx[esi * 4 + 18]}
     */
    // Template#: 1261, Serial#: 2487
    public void decw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0292((byte) 0xFF, (byte) 0x01, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code decb  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code decb      [ebx + 18]}
     */
    // Template#: 1262, Serial#: 2426
    public void decb(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xFE, (byte) 0x01, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code decl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code decl      [ebx + 18]}
     */
    // Template#: 1263, Serial#: 2454
    public void decl(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xFF, (byte) 0x01, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code decw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code decw      [ebx + 18]}
     */
    // Template#: 1264, Serial#: 2486
    public void decw(byte disp8, IA32IndirectRegister32 destination) {
        assemble0293((byte) 0xFF, (byte) 0x01, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code decb  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code decb      ebx[esi * 4]}
     */
    // Template#: 1265, Serial#: 2421
    public void decb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xFE, (byte) 0x01, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code decl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code decl      ebx[esi * 4]}
     */
    // Template#: 1266, Serial#: 2439
    public void decl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xFF, (byte) 0x01, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code decw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code decw      ebx[esi * 4]}
     */
    // Template#: 1267, Serial#: 2481
    public void decw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0294((byte) 0xFF, (byte) 0x01, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code dec  }<i>register</i>
     * Example disassembly syntax: {@code dec       ax}
     */
    // Template#: 1268, Serial#: 1760
    public void dec(IA32GeneralRegister16 register) {
        assemble0295((byte) 0x48, register);
    }

    /**
     * Pseudo-external assembler syntax: {@code decw  }<i>destination</i>
     * Example disassembly syntax: {@code decw      ax}
     */
    // Template#: 1269, Serial#: 2493
    public void decw(IA32GeneralRegister16 destination) {
        assemble0188((byte) 0xFF, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code dec  }<i>register</i>
     * Example disassembly syntax: {@code dec       eax}
     */
    // Template#: 1270, Serial#: 1759
    public void dec(IA32GeneralRegister32 register) {
        assemble0296((byte) 0x48, register);
    }

    /**
     * Pseudo-external assembler syntax: {@code decl  }<i>destination</i>
     * Example disassembly syntax: {@code decl      eax}
     */
    // Template#: 1271, Serial#: 2473
    public void decl(IA32GeneralRegister32 destination) {
        assemble0189((byte) 0xFF, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code decb  }<i>destination</i>
     * Example disassembly syntax: {@code decb      al}
     */
    // Template#: 1272, Serial#: 2433
    public void decb(IA32GeneralRegister8 destination) {
        assemble0297((byte) 0xFE, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code decb  }<i>destination</i>
     * Example disassembly syntax: {@code decb      [ebx]}
     */
    // Template#: 1273, Serial#: 2420
    public void decb(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xFE, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code decl  }<i>destination</i>
     * Example disassembly syntax: {@code decl      [ebx]}
     */
    // Template#: 1274, Serial#: 2438
    public void decl(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xFF, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code decw  }<i>destination</i>
     * Example disassembly syntax: {@code decw      [ebx]}
     */
    // Template#: 1275, Serial#: 2480
    public void decw(IA32IndirectRegister32 destination) {
        assemble0298((byte) 0xFF, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code decb  }<i>m32</i>
     * Example disassembly syntax: {@code decb      [L1: 0x12345678]}
     */
    // Template#: 1276, Serial#: 2423
    public void m_decb(int m32) {
        assemble0299((byte) 0xFE, (byte) 0x01, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code decl  }<i>m32</i>
     * Example disassembly syntax: {@code decl      [L1: 0x12345678]}
     */
    // Template#: 1277, Serial#: 2441
    public void m_decl(int m32) {
        assemble0299((byte) 0xFF, (byte) 0x01, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code decw  }<i>m32</i>
     * Example disassembly syntax: {@code decw      [L1: 0x12345678]}
     */
    // Template#: 1278, Serial#: 2483
    public void m_decw(int m32) {
        assemble0300((byte) 0xFF, (byte) 0x01, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code decb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code decb      ebx[esi * 4 + 305419896]}
     */
    // Template#: 1279, Serial#: 2431
    public void decb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xFE, (byte) 0x01, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code decl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code decl      ebx[esi * 4 + 305419896]}
     */
    // Template#: 1280, Serial#: 2465
    public void decl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xFF, (byte) 0x01, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code decw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code decw      ebx[esi * 4 + 305419896]}
     */
    // Template#: 1281, Serial#: 2491
    public void decw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0301((byte) 0xFF, (byte) 0x01, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code decb  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code decb      L1: 0x12345678[esi * 4]}
     */
    // Template#: 1282, Serial#: 2422
    public void m_decb(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xFE, (byte) 0x01, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code decl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code decl      L1: 0x12345678[esi * 4]}
     */
    // Template#: 1283, Serial#: 2440
    public void m_decl(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xFF, (byte) 0x01, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code decw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code decw      L1: 0x12345678[esi * 4]}
     */
    // Template#: 1284, Serial#: 2482
    public void m_decw(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0302((byte) 0xFF, (byte) 0x01, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code decb  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code decb      [ebx + 305419896]}
     */
    // Template#: 1285, Serial#: 2430
    public void decb(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xFE, (byte) 0x01, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code decl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code decl      [ebx + 305419896]}
     */
    // Template#: 1286, Serial#: 2464
    public void decl(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xFF, (byte) 0x01, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code decw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code decw      [ebx + 305419896]}
     */
    // Template#: 1287, Serial#: 2490
    public void decw(int disp32, IA32IndirectRegister32 destination) {
        assemble0303((byte) 0xFF, (byte) 0x01, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code divb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divb      ebx[esi * 4 + 18], al}
     */
    // Template#: 1288, Serial#: 1386
    public void divb___AL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xF6, (byte) 0x06, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divl      ebx[esi * 4 + 18], eax}
     */
    // Template#: 1289, Serial#: 1449
    public void divl___EAX(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xF7, (byte) 0x06, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divw      ebx[esi * 4 + 18], ax}
     */
    // Template#: 1290, Serial#: 1512
    public void divw___AX(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0292((byte) 0xF7, (byte) 0x06, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divb  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code divb      [ebx + 18], al}
     */
    // Template#: 1291, Serial#: 1385
    public void divb___AL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xF6, (byte) 0x06, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code divl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code divl      [ebx + 18], eax}
     */
    // Template#: 1292, Serial#: 1448
    public void divl___EAX(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xF7, (byte) 0x06, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code divw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code divw      [ebx + 18], ax}
     */
    // Template#: 1293, Serial#: 1511
    public void divw___AX(byte disp8, IA32IndirectRegister32 destination) {
        assemble0293((byte) 0xF7, (byte) 0x06, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code divb  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divb      ebx[esi * 4], al}
     */
    // Template#: 1294, Serial#: 1368
    public void divb___AL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xF6, (byte) 0x06, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divl      ebx[esi * 4], eax}
     */
    // Template#: 1295, Serial#: 1431
    public void divl___EAX(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xF7, (byte) 0x06, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divw      ebx[esi * 4], ax}
     */
    // Template#: 1296, Serial#: 1494
    public void divw___AX(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0294((byte) 0xF7, (byte) 0x06, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divw  }<i>destination</i>
     * Example disassembly syntax: {@code divw      ax, ax}
     */
    // Template#: 1297, Serial#: 1534
    public void divw___AX(IA32GeneralRegister16 destination) {
        assemble0188((byte) 0xF7, (byte) 0x06, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code divl  }<i>destination</i>
     * Example disassembly syntax: {@code divl      eax, eax}
     */
    // Template#: 1298, Serial#: 1471
    public void divl___EAX(IA32GeneralRegister32 destination) {
        assemble0189((byte) 0xF7, (byte) 0x06, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code divb  }<i>destination</i>
     * Example disassembly syntax: {@code divb      al, al}
     */
    // Template#: 1299, Serial#: 1408
    public void divb___AL(IA32GeneralRegister8 destination) {
        assemble0297((byte) 0xF6, (byte) 0x06, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code divb  }<i>destination</i>
     * Example disassembly syntax: {@code divb      [ebx], al}
     */
    // Template#: 1300, Serial#: 1367
    public void divb___AL(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xF6, (byte) 0x06, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code divl  }<i>destination</i>
     * Example disassembly syntax: {@code divl      [ebx], eax}
     */
    // Template#: 1301, Serial#: 1430
    public void divl___EAX(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xF7, (byte) 0x06, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code divw  }<i>destination</i>
     * Example disassembly syntax: {@code divw      [ebx], ax}
     */
    // Template#: 1302, Serial#: 1493
    public void divw___AX(IA32IndirectRegister32 destination) {
        assemble0298((byte) 0xF7, (byte) 0x06, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code divb  }<i>m32</i>
     * Example disassembly syntax: {@code divb      [L1: 0x12345678], al}
     */
    // Template#: 1303, Serial#: 1370
    public void m_divb___AL(int m32) {
        assemble0299((byte) 0xF6, (byte) 0x06, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code divl  }<i>m32</i>
     * Example disassembly syntax: {@code divl      [L1: 0x12345678], eax}
     */
    // Template#: 1304, Serial#: 1433
    public void m_divl___EAX(int m32) {
        assemble0299((byte) 0xF7, (byte) 0x06, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code divw  }<i>m32</i>
     * Example disassembly syntax: {@code divw      [L1: 0x12345678], ax}
     */
    // Template#: 1305, Serial#: 1496
    public void m_divw___AX(int m32) {
        assemble0300((byte) 0xF7, (byte) 0x06, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code divb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divb      ebx[esi * 4 + 305419896], al}
     */
    // Template#: 1306, Serial#: 1400
    public void divb___AL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xF6, (byte) 0x06, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divl      ebx[esi * 4 + 305419896], eax}
     */
    // Template#: 1307, Serial#: 1463
    public void divl___EAX(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xF7, (byte) 0x06, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divw      ebx[esi * 4 + 305419896], ax}
     */
    // Template#: 1308, Serial#: 1526
    public void divw___AX(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0301((byte) 0xF7, (byte) 0x06, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divb  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divb      L1: 0x12345678[esi * 4], al}
     */
    // Template#: 1309, Serial#: 1369
    public void m_divb___AL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xF6, (byte) 0x06, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divl      L1: 0x12345678[esi * 4], eax}
     */
    // Template#: 1310, Serial#: 1432
    public void m_divl___EAX(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xF7, (byte) 0x06, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divw      L1: 0x12345678[esi * 4], ax}
     */
    // Template#: 1311, Serial#: 1495
    public void m_divw___AX(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0302((byte) 0xF7, (byte) 0x06, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divb  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code divb      [ebx + 305419896], al}
     */
    // Template#: 1312, Serial#: 1399
    public void divb___AL(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xF6, (byte) 0x06, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code divl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code divl      [ebx + 305419896], eax}
     */
    // Template#: 1313, Serial#: 1462
    public void divl___EAX(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xF7, (byte) 0x06, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code divw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code divw      [ebx + 305419896], ax}
     */
    // Template#: 1314, Serial#: 1525
    public void divw___AX(int disp32, IA32IndirectRegister32 destination) {
        assemble0303((byte) 0xF7, (byte) 0x06, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code divpd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divpd     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1315, Serial#: 4501
    public void divpd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x5E, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divpd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code divpd     xmm0, [ebx + 18]}
     */
    // Template#: 1316, Serial#: 4500
    public void divpd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x5E, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code divpd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divpd     xmm0, ebx[esi * 4]}
     */
    // Template#: 1317, Serial#: 4497
    public void divpd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x5E, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code divpd     xmm0, [ebx]}
     */
    // Template#: 1318, Serial#: 4496
    public void divpd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x5E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code divpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code divpd     xmm0, xmm0}
     */
    // Template#: 1319, Serial#: 4504
    public void divpd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x5E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code divpd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code divpd     xmm0, [L1: 0x12345678]}
     */
    // Template#: 1320, Serial#: 4499
    public void m_divpd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x5E, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code divpd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divpd     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1321, Serial#: 4503
    public void divpd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x5E, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divpd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divpd     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1322, Serial#: 4498
    public void m_divpd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x5E, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divpd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code divpd     xmm0, [ebx + 305419896]}
     */
    // Template#: 1323, Serial#: 4502
    public void divpd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x5E, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code divps  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divps     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1324, Serial#: 4429
    public void divps(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0102((byte) 0x5E, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divps  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code divps     xmm0, [ebx + 18]}
     */
    // Template#: 1325, Serial#: 4428
    public void divps(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0103((byte) 0x5E, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code divps  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divps     xmm0, ebx[esi * 4]}
     */
    // Template#: 1326, Serial#: 4425
    public void divps(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0104((byte) 0x5E, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code divps     xmm0, [ebx]}
     */
    // Template#: 1327, Serial#: 4424
    public void divps(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0105((byte) 0x5E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code divps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code divps     xmm0, xmm0}
     */
    // Template#: 1328, Serial#: 4432
    public void divps(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0106((byte) 0x5E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code divps  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code divps     xmm0, [L1: 0x12345678]}
     */
    // Template#: 1329, Serial#: 4427
    public void m_divps(IA32XMMRegister destination, int m32) {
        assemble0107((byte) 0x5E, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code divps  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divps     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1330, Serial#: 4431
    public void divps(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0108((byte) 0x5E, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divps  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divps     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1331, Serial#: 4426
    public void m_divps(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0109((byte) 0x5E, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divps  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code divps     xmm0, [ebx + 305419896]}
     */
    // Template#: 1332, Serial#: 4430
    public void divps(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0110((byte) 0x5E, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code divsd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divsd     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1333, Serial#: 4564
    public void divsd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0111((byte) 0x5E, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divsd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code divsd     xmm0, [ebx + 18]}
     */
    // Template#: 1334, Serial#: 4563
    public void divsd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0112((byte) 0x5E, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code divsd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divsd     xmm0, ebx[esi * 4]}
     */
    // Template#: 1335, Serial#: 4560
    public void divsd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0113((byte) 0x5E, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divsd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code divsd     xmm0, [ebx]}
     */
    // Template#: 1336, Serial#: 4559
    public void divsd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0114((byte) 0x5E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code divsd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code divsd     xmm0, xmm0}
     */
    // Template#: 1337, Serial#: 4567
    public void divsd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0115((byte) 0x5E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code divsd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code divsd     xmm0, [L1: 0x12345678]}
     */
    // Template#: 1338, Serial#: 4562
    public void m_divsd(IA32XMMRegister destination, int m32) {
        assemble0116((byte) 0x5E, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code divsd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divsd     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1339, Serial#: 4566
    public void divsd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0117((byte) 0x5E, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divsd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divsd     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1340, Serial#: 4561
    public void m_divsd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0118((byte) 0x5E, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divsd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code divsd     xmm0, [ebx + 305419896]}
     */
    // Template#: 1341, Serial#: 4565
    public void divsd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0119((byte) 0x5E, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code divss  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divss     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1342, Serial#: 4636
    public void divss(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0120((byte) 0x5E, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divss  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code divss     xmm0, [ebx + 18]}
     */
    // Template#: 1343, Serial#: 4635
    public void divss(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0121((byte) 0x5E, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code divss  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divss     xmm0, ebx[esi * 4]}
     */
    // Template#: 1344, Serial#: 4632
    public void divss(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0122((byte) 0x5E, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code divss     xmm0, [ebx]}
     */
    // Template#: 1345, Serial#: 4631
    public void divss(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0123((byte) 0x5E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code divss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code divss     xmm0, xmm0}
     */
    // Template#: 1346, Serial#: 4639
    public void divss(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0124((byte) 0x5E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code divss  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code divss     xmm0, [L1: 0x12345678]}
     */
    // Template#: 1347, Serial#: 4634
    public void m_divss(IA32XMMRegister destination, int m32) {
        assemble0125((byte) 0x5E, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code divss  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divss     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1348, Serial#: 4638
    public void divss(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0126((byte) 0x5E, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divss  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code divss     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1349, Serial#: 4633
    public void m_divss(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0127((byte) 0x5E, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code divss  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code divss     xmm0, [ebx + 305419896]}
     */
    // Template#: 1350, Serial#: 4637
    public void divss(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0128((byte) 0x5E, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code emms  }
     * Example disassembly syntax: {@code emms    }
     */
    // Template#: 1351, Serial#: 3243
    public void emms() {
        assemble0198((byte) 0x77);
    }

    /**
     * Pseudo-external assembler syntax: {@code enter  }<i>imm16</i>, <i>imm8</i>
     * Example disassembly syntax: {@code enter     0x1234, 0x12}
     */
    // Template#: 1352, Serial#: 1935
    public void enter(short imm16, byte imm8) {
        assemble0304((byte) 0xC8, imm16, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code f2xm1  }
     * Example disassembly syntax: {@code f2xm1   }
     */
    // Template#: 1353, Serial#: 5551
    public void f2xm1() {
        assemble0305((byte) 0xF0);
    }

    /**
     * Pseudo-external assembler syntax: {@code fabs  }
     * Example disassembly syntax: {@code fabs    }
     */
    // Template#: 1354, Serial#: 5541
    public void fabs() {
        assemble0305((byte) 0xE1);
    }

    /**
     * Pseudo-external assembler syntax: {@code fadds  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fadds     ebx[esi * 4 + 18]}
     */
    // Template#: 1355, Serial#: 1976
    public void fadds(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD8, (byte) 0x00, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code faddl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code faddl     ebx[esi * 4 + 18]}
     */
    // Template#: 1356, Serial#: 2200
    public void faddl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDC, (byte) 0x00, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fadds  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fadds     [ebx + 18]}
     */
    // Template#: 1357, Serial#: 1975
    public void fadds(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD8, (byte) 0x00, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code faddl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code faddl     [ebx + 18]}
     */
    // Template#: 1358, Serial#: 2199
    public void faddl(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDC, (byte) 0x00, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fadds  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fadds     ebx[esi * 4]}
     */
    // Template#: 1359, Serial#: 1944
    public void fadds(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD8, (byte) 0x00, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code faddl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code faddl     ebx[esi * 4]}
     */
    // Template#: 1360, Serial#: 2168
    public void faddl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDC, (byte) 0x00, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fadds  }<i>destination</i>
     * Example disassembly syntax: {@code fadds     [ebx]}
     */
    // Template#: 1361, Serial#: 1943
    public void fadds(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD8, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code faddl  }<i>destination</i>
     * Example disassembly syntax: {@code faddl     [ebx]}
     */
    // Template#: 1362, Serial#: 2167
    public void faddl(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDC, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fadd  }<i>st_i</i>
     * Example disassembly syntax: {@code fadd      st, st(0)}
     */
    // Template#: 1363, Serial#: 5528
    public void fadd_ST(FPStackRegister st_i) {
        assemble0306((byte) 0xC0, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fadd  }<i>st_i</i>
     * Example disassembly syntax: {@code fadd      st(0), st}
     */
    // Template#: 1364, Serial#: 5580
    public void fadd___ST(FPStackRegister st_i) {
        assemble0307((byte) 0xC0, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fadds  }<i>m32</i>
     * Example disassembly syntax: {@code fadds     [L1: 0x12345678]}
     */
    // Template#: 1365, Serial#: 1946
    public void m_fadds(int m32) {
        assemble0299((byte) 0xD8, (byte) 0x00, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code faddl  }<i>m32</i>
     * Example disassembly syntax: {@code faddl     [L1: 0x12345678]}
     */
    // Template#: 1366, Serial#: 2170
    public void m_faddl(int m32) {
        assemble0299((byte) 0xDC, (byte) 0x00, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fadds  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fadds     ebx[esi * 4 + 305419896]}
     */
    // Template#: 1367, Serial#: 1992
    public void fadds(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD8, (byte) 0x00, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code faddl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code faddl     ebx[esi * 4 + 305419896]}
     */
    // Template#: 1368, Serial#: 2216
    public void faddl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDC, (byte) 0x00, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fadds  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fadds     L1: 0x12345678[esi * 4]}
     */
    // Template#: 1369, Serial#: 1945
    public void m_fadds(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD8, (byte) 0x00, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code faddl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code faddl     L1: 0x12345678[esi * 4]}
     */
    // Template#: 1370, Serial#: 2169
    public void m_faddl(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDC, (byte) 0x00, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fadds  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fadds     [ebx + 305419896]}
     */
    // Template#: 1371, Serial#: 1991
    public void fadds(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD8, (byte) 0x00, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code faddl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code faddl     [ebx + 305419896]}
     */
    // Template#: 1372, Serial#: 2215
    public void faddl(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDC, (byte) 0x00, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code faddp  }<i>st_i</i>
     * Example disassembly syntax: {@code faddp     st(0), st}
     */
    // Template#: 1373, Serial#: 5594
    public void faddp___ST(FPStackRegister st_i) {
        assemble0308((byte) 0xC0, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fbld  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fbld      ebx[esi * 4 + 18]}
     */
    // Template#: 1374, Serial#: 2378
    public void fbld(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDF, (byte) 0x04, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fbld  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fbld      [ebx + 18]}
     */
    // Template#: 1375, Serial#: 2377
    public void fbld(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDF, (byte) 0x04, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fbld  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fbld      ebx[esi * 4]}
     */
    // Template#: 1376, Serial#: 2356
    public void fbld(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDF, (byte) 0x04, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fbld  }<i>destination</i>
     * Example disassembly syntax: {@code fbld      [ebx]}
     */
    // Template#: 1377, Serial#: 2355
    public void fbld(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDF, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fbld  }<i>m32</i>
     * Example disassembly syntax: {@code fbld      [L1: 0x12345678]}
     */
    // Template#: 1378, Serial#: 2358
    public void m_fbld(int m32) {
        assemble0299((byte) 0xDF, (byte) 0x04, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fbld  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fbld      ebx[esi * 4 + 305419896]}
     */
    // Template#: 1379, Serial#: 2392
    public void fbld(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDF, (byte) 0x04, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fbld  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fbld      L1: 0x12345678[esi * 4]}
     */
    // Template#: 1380, Serial#: 2357
    public void m_fbld(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDF, (byte) 0x04, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fbld  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fbld      [ebx + 305419896]}
     */
    // Template#: 1381, Serial#: 2391
    public void fbld(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDF, (byte) 0x04, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fbstp  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fbstp     ebx[esi * 4 + 18]}
     */
    // Template#: 1382, Serial#: 2382
    public void fbstp(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDF, (byte) 0x06, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fbstp  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fbstp     [ebx + 18]}
     */
    // Template#: 1383, Serial#: 2381
    public void fbstp(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDF, (byte) 0x06, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fbstp  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fbstp     ebx[esi * 4]}
     */
    // Template#: 1384, Serial#: 2364
    public void fbstp(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDF, (byte) 0x06, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fbstp  }<i>destination</i>
     * Example disassembly syntax: {@code fbstp     [ebx]}
     */
    // Template#: 1385, Serial#: 2363
    public void fbstp(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDF, (byte) 0x06, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fbstp  }<i>m32</i>
     * Example disassembly syntax: {@code fbstp     [L1: 0x12345678]}
     */
    // Template#: 1386, Serial#: 2366
    public void m_fbstp(int m32) {
        assemble0299((byte) 0xDF, (byte) 0x06, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fbstp  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fbstp     ebx[esi * 4 + 305419896]}
     */
    // Template#: 1387, Serial#: 2396
    public void fbstp(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDF, (byte) 0x06, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fbstp  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fbstp     L1: 0x12345678[esi * 4]}
     */
    // Template#: 1388, Serial#: 2365
    public void m_fbstp(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDF, (byte) 0x06, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fbstp  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fbstp     [ebx + 305419896]}
     */
    // Template#: 1389, Serial#: 2395
    public void fbstp(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDF, (byte) 0x06, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fchs  }
     * Example disassembly syntax: {@code fchs    }
     */
    // Template#: 1390, Serial#: 5540
    public void fchs() {
        assemble0305((byte) 0xE0);
    }

    /**
     * Pseudo-external assembler syntax: {@code fclex  }
     * Example disassembly syntax: {@code fclex   }
     */
    // Template#: 1391, Serial#: 5576
    public void fclex() {
        assemble0309((byte) 0xE2);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcmovb  }<i>st_i</i>
     * Example disassembly syntax: {@code fcmovb    st, st(0)}
     */
    // Template#: 1392, Serial#: 5567
    public void fcmovb_ST(FPStackRegister st_i) {
        assemble0310((byte) 0xC0, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcmovbe  }<i>st_i</i>
     * Example disassembly syntax: {@code fcmovbe   st, st(0)}
     */
    // Template#: 1393, Serial#: 5569
    public void fcmovbe_ST(FPStackRegister st_i) {
        assemble0310((byte) 0xD0, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcmove  }<i>st_i</i>
     * Example disassembly syntax: {@code fcmove    st, st(0)}
     */
    // Template#: 1394, Serial#: 5568
    public void fcmove_ST(FPStackRegister st_i) {
        assemble0310((byte) 0xC8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcmovnb  }<i>st_i</i>
     * Example disassembly syntax: {@code fcmovnb   st, st(0)}
     */
    // Template#: 1395, Serial#: 5572
    public void fcmovnb_ST(FPStackRegister st_i) {
        assemble0311((byte) 0xC0, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcmovnbe  }<i>st_i</i>
     * Example disassembly syntax: {@code fcmovnbe  st, st(0)}
     */
    // Template#: 1396, Serial#: 5574
    public void fcmovnbe_ST(FPStackRegister st_i) {
        assemble0311((byte) 0xD0, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcmovne  }<i>st_i</i>
     * Example disassembly syntax: {@code fcmovne   st, st(0)}
     */
    // Template#: 1397, Serial#: 5573
    public void fcmovne_ST(FPStackRegister st_i) {
        assemble0311((byte) 0xC8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcmovnu  }<i>st_i</i>
     * Example disassembly syntax: {@code fcmovnu   st, st(0)}
     */
    // Template#: 1398, Serial#: 5575
    public void fcmovnu_ST(FPStackRegister st_i) {
        assemble0311((byte) 0xD8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcmovu  }<i>st_i</i>
     * Example disassembly syntax: {@code fcmovu    st, st(0)}
     */
    // Template#: 1399, Serial#: 5570
    public void fcmovu_ST(FPStackRegister st_i) {
        assemble0310((byte) 0xD8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcoms  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fcoms     ebx[esi * 4 + 18]}
     */
    // Template#: 1400, Serial#: 1980
    public void fcoms(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD8, (byte) 0x02, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcoml  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fcoml     ebx[esi * 4 + 18]}
     */
    // Template#: 1401, Serial#: 2204
    public void fcoml(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDC, (byte) 0x02, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcoms  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fcoms     [ebx + 18]}
     */
    // Template#: 1402, Serial#: 1979
    public void fcoms(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD8, (byte) 0x02, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcoml  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fcoml     [ebx + 18]}
     */
    // Template#: 1403, Serial#: 2203
    public void fcoml(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDC, (byte) 0x02, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcoms  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fcoms     ebx[esi * 4]}
     */
    // Template#: 1404, Serial#: 1952
    public void fcoms(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD8, (byte) 0x02, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcoml  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fcoml     ebx[esi * 4]}
     */
    // Template#: 1405, Serial#: 2176
    public void fcoml(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDC, (byte) 0x02, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcoms  }<i>destination</i>
     * Example disassembly syntax: {@code fcoms     [ebx]}
     */
    // Template#: 1406, Serial#: 1951
    public void fcoms(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD8, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcoml  }<i>destination</i>
     * Example disassembly syntax: {@code fcoml     [ebx]}
     */
    // Template#: 1407, Serial#: 2175
    public void fcoml(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDC, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcom  }<i>st_i</i>
     * Example disassembly syntax: {@code fcom      st(0)}
     */
    // Template#: 1408, Serial#: 5530
    public void fcom(FPStackRegister st_i) {
        assemble0306((byte) 0xD0, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcoms  }<i>m32</i>
     * Example disassembly syntax: {@code fcoms     [L1: 0x12345678]}
     */
    // Template#: 1409, Serial#: 1954
    public void m_fcoms(int m32) {
        assemble0299((byte) 0xD8, (byte) 0x02, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcoml  }<i>m32</i>
     * Example disassembly syntax: {@code fcoml     [L1: 0x12345678]}
     */
    // Template#: 1410, Serial#: 2178
    public void m_fcoml(int m32) {
        assemble0299((byte) 0xDC, (byte) 0x02, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcoms  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fcoms     ebx[esi * 4 + 305419896]}
     */
    // Template#: 1411, Serial#: 1996
    public void fcoms(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD8, (byte) 0x02, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcoml  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fcoml     ebx[esi * 4 + 305419896]}
     */
    // Template#: 1412, Serial#: 2220
    public void fcoml(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDC, (byte) 0x02, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcoms  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fcoms     L1: 0x12345678[esi * 4]}
     */
    // Template#: 1413, Serial#: 1953
    public void m_fcoms(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD8, (byte) 0x02, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcoml  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fcoml     L1: 0x12345678[esi * 4]}
     */
    // Template#: 1414, Serial#: 2177
    public void m_fcoml(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDC, (byte) 0x02, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcoms  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fcoms     [ebx + 305419896]}
     */
    // Template#: 1415, Serial#: 1995
    public void fcoms(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD8, (byte) 0x02, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcoml  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fcoml     [ebx + 305419896]}
     */
    // Template#: 1416, Serial#: 2219
    public void fcoml(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDC, (byte) 0x02, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcom2  }<i>st_i</i>
     * Example disassembly syntax: {@code fcom2     st(0)}
     */
    // Template#: 1417, Serial#: 5582
    public void fcom2(FPStackRegister st_i) {
        assemble0307((byte) 0xD0, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcomi  }<i>st_i</i>
     * Example disassembly syntax: {@code fcomi     st, st(0)}
     */
    // Template#: 1418, Serial#: 5579
    public void fcomi_ST(FPStackRegister st_i) {
        assemble0311((byte) 0xF0, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcomip  }<i>st_i</i>
     * Example disassembly syntax: {@code fcomip    st, st(0)}
     */
    // Template#: 1419, Serial#: 5608
    public void fcomip_ST(FPStackRegister st_i) {
        assemble0312((byte) 0xF0, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcomps  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fcomps    ebx[esi * 4 + 18]}
     */
    // Template#: 1420, Serial#: 1982
    public void fcomps(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD8, (byte) 0x03, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcompl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fcompl    ebx[esi * 4 + 18]}
     */
    // Template#: 1421, Serial#: 2206
    public void fcompl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDC, (byte) 0x03, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcomps  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fcomps    [ebx + 18]}
     */
    // Template#: 1422, Serial#: 1981
    public void fcomps(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD8, (byte) 0x03, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcompl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fcompl    [ebx + 18]}
     */
    // Template#: 1423, Serial#: 2205
    public void fcompl(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDC, (byte) 0x03, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcomps  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fcomps    ebx[esi * 4]}
     */
    // Template#: 1424, Serial#: 1956
    public void fcomps(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD8, (byte) 0x03, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcompl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fcompl    ebx[esi * 4]}
     */
    // Template#: 1425, Serial#: 2180
    public void fcompl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDC, (byte) 0x03, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcomps  }<i>destination</i>
     * Example disassembly syntax: {@code fcomps    [ebx]}
     */
    // Template#: 1426, Serial#: 1955
    public void fcomps(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD8, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcompl  }<i>destination</i>
     * Example disassembly syntax: {@code fcompl    [ebx]}
     */
    // Template#: 1427, Serial#: 2179
    public void fcompl(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDC, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcomp  }<i>st_i</i>
     * Example disassembly syntax: {@code fcomp     st(0)}
     */
    // Template#: 1428, Serial#: 5531
    public void fcomp(FPStackRegister st_i) {
        assemble0306((byte) 0xD8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcomps  }<i>m32</i>
     * Example disassembly syntax: {@code fcomps    [L1: 0x12345678]}
     */
    // Template#: 1429, Serial#: 1958
    public void m_fcomps(int m32) {
        assemble0299((byte) 0xD8, (byte) 0x03, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcompl  }<i>m32</i>
     * Example disassembly syntax: {@code fcompl    [L1: 0x12345678]}
     */
    // Template#: 1430, Serial#: 2182
    public void m_fcompl(int m32) {
        assemble0299((byte) 0xDC, (byte) 0x03, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcomps  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fcomps    ebx[esi * 4 + 305419896]}
     */
    // Template#: 1431, Serial#: 1998
    public void fcomps(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD8, (byte) 0x03, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcompl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fcompl    ebx[esi * 4 + 305419896]}
     */
    // Template#: 1432, Serial#: 2222
    public void fcompl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDC, (byte) 0x03, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcomps  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fcomps    L1: 0x12345678[esi * 4]}
     */
    // Template#: 1433, Serial#: 1957
    public void m_fcomps(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD8, (byte) 0x03, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcompl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fcompl    L1: 0x12345678[esi * 4]}
     */
    // Template#: 1434, Serial#: 2181
    public void m_fcompl(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDC, (byte) 0x03, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcomps  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fcomps    [ebx + 305419896]}
     */
    // Template#: 1435, Serial#: 1997
    public void fcomps(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD8, (byte) 0x03, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcompl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fcompl    [ebx + 305419896]}
     */
    // Template#: 1436, Serial#: 2221
    public void fcompl(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDC, (byte) 0x03, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcomp3  }<i>st_i</i>
     * Example disassembly syntax: {@code fcomp3    st(0)}
     */
    // Template#: 1437, Serial#: 5583
    public void fcomp3(FPStackRegister st_i) {
        assemble0307((byte) 0xD8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcomp5  }<i>st_i</i>
     * Example disassembly syntax: {@code fcomp5    st(0), st}
     */
    // Template#: 1438, Serial#: 5596
    public void fcomp5___ST(FPStackRegister st_i) {
        assemble0308((byte) 0xD0, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcompp  }
     * Example disassembly syntax: {@code fcompp  }
     */
    // Template#: 1439, Serial#: 5597
    public void fcompp() {
        assemble0313((byte) 0xD9);
    }

    /**
     * Pseudo-external assembler syntax: {@code fcos  }
     * Example disassembly syntax: {@code fcos    }
     */
    // Template#: 1440, Serial#: 5566
    public void fcos() {
        assemble0305((byte) 0xFF);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdecstp  }
     * Example disassembly syntax: {@code fdecstp }
     */
    // Template#: 1441, Serial#: 5557
    public void fdecstp() {
        assemble0305((byte) 0xF6);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivs  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fdivs     ebx[esi * 4 + 18]}
     */
    // Template#: 1442, Serial#: 1988
    public void fdivs(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD8, (byte) 0x06, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fdivl     ebx[esi * 4 + 18]}
     */
    // Template#: 1443, Serial#: 2212
    public void fdivl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDC, (byte) 0x06, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivs  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fdivs     [ebx + 18]}
     */
    // Template#: 1444, Serial#: 1987
    public void fdivs(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD8, (byte) 0x06, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fdivl     [ebx + 18]}
     */
    // Template#: 1445, Serial#: 2211
    public void fdivl(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDC, (byte) 0x06, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivs  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fdivs     ebx[esi * 4]}
     */
    // Template#: 1446, Serial#: 1968
    public void fdivs(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD8, (byte) 0x06, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fdivl     ebx[esi * 4]}
     */
    // Template#: 1447, Serial#: 2192
    public void fdivl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDC, (byte) 0x06, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivs  }<i>destination</i>
     * Example disassembly syntax: {@code fdivs     [ebx]}
     */
    // Template#: 1448, Serial#: 1967
    public void fdivs(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD8, (byte) 0x06, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivl  }<i>destination</i>
     * Example disassembly syntax: {@code fdivl     [ebx]}
     */
    // Template#: 1449, Serial#: 2191
    public void fdivl(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDC, (byte) 0x06, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdiv  }<i>st_i</i>
     * Example disassembly syntax: {@code fdiv      st, st(0)}
     */
    // Template#: 1450, Serial#: 5534
    public void fdiv_ST(FPStackRegister st_i) {
        assemble0306((byte) 0xF0, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdiv  }<i>st_i</i>
     * Example disassembly syntax: {@code fdiv      st(0), st}
     */
    // Template#: 1451, Serial#: 5586
    public void fdiv___ST(FPStackRegister st_i) {
        assemble0307((byte) 0xF0, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivs  }<i>m32</i>
     * Example disassembly syntax: {@code fdivs     [L1: 0x12345678]}
     */
    // Template#: 1452, Serial#: 1970
    public void m_fdivs(int m32) {
        assemble0299((byte) 0xD8, (byte) 0x06, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivl  }<i>m32</i>
     * Example disassembly syntax: {@code fdivl     [L1: 0x12345678]}
     */
    // Template#: 1453, Serial#: 2194
    public void m_fdivl(int m32) {
        assemble0299((byte) 0xDC, (byte) 0x06, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivs  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fdivs     ebx[esi * 4 + 305419896]}
     */
    // Template#: 1454, Serial#: 2004
    public void fdivs(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD8, (byte) 0x06, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fdivl     ebx[esi * 4 + 305419896]}
     */
    // Template#: 1455, Serial#: 2228
    public void fdivl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDC, (byte) 0x06, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivs  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fdivs     L1: 0x12345678[esi * 4]}
     */
    // Template#: 1456, Serial#: 1969
    public void m_fdivs(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD8, (byte) 0x06, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fdivl     L1: 0x12345678[esi * 4]}
     */
    // Template#: 1457, Serial#: 2193
    public void m_fdivl(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDC, (byte) 0x06, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivs  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fdivs     [ebx + 305419896]}
     */
    // Template#: 1458, Serial#: 2003
    public void fdivs(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD8, (byte) 0x06, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fdivl     [ebx + 305419896]}
     */
    // Template#: 1459, Serial#: 2227
    public void fdivl(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDC, (byte) 0x06, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivrp  }<i>st_i</i>
     * Example disassembly syntax: {@code fdivrp    st(0), st}
     */
    // Template#: 1460, Serial#: 5601
    public void fdivp___ST(FPStackRegister st_i) {
        assemble0308((byte) 0xF8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivrs  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fdivrs    ebx[esi * 4 + 18]}
     */
    // Template#: 1461, Serial#: 1990
    public void fdivrs(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD8, (byte) 0x07, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivrl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fdivrl    ebx[esi * 4 + 18]}
     */
    // Template#: 1462, Serial#: 2214
    public void fdivrl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDC, (byte) 0x07, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivrs  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fdivrs    [ebx + 18]}
     */
    // Template#: 1463, Serial#: 1989
    public void fdivrs(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD8, (byte) 0x07, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivrl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fdivrl    [ebx + 18]}
     */
    // Template#: 1464, Serial#: 2213
    public void fdivrl(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDC, (byte) 0x07, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivrs  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fdivrs    ebx[esi * 4]}
     */
    // Template#: 1465, Serial#: 1972
    public void fdivrs(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD8, (byte) 0x07, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivrl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fdivrl    ebx[esi * 4]}
     */
    // Template#: 1466, Serial#: 2196
    public void fdivrl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDC, (byte) 0x07, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivrs  }<i>destination</i>
     * Example disassembly syntax: {@code fdivrs    [ebx]}
     */
    // Template#: 1467, Serial#: 1971
    public void fdivrs(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD8, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivrl  }<i>destination</i>
     * Example disassembly syntax: {@code fdivrl    [ebx]}
     */
    // Template#: 1468, Serial#: 2195
    public void fdivrl(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDC, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivr  }<i>st_i</i>
     * Example disassembly syntax: {@code fdivr     st, st(0)}
     */
    // Template#: 1469, Serial#: 5535
    public void fdivr_ST(FPStackRegister st_i) {
        assemble0306((byte) 0xF8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivr  }<i>st_i</i>
     * Example disassembly syntax: {@code fdivr     st(0), st}
     */
    // Template#: 1470, Serial#: 5587
    public void fdivr___ST(FPStackRegister st_i) {
        assemble0307((byte) 0xF8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivrs  }<i>m32</i>
     * Example disassembly syntax: {@code fdivrs    [L1: 0x12345678]}
     */
    // Template#: 1471, Serial#: 1974
    public void m_fdivrs(int m32) {
        assemble0299((byte) 0xD8, (byte) 0x07, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivrl  }<i>m32</i>
     * Example disassembly syntax: {@code fdivrl    [L1: 0x12345678]}
     */
    // Template#: 1472, Serial#: 2198
    public void m_fdivrl(int m32) {
        assemble0299((byte) 0xDC, (byte) 0x07, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivrs  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fdivrs    ebx[esi * 4 + 305419896]}
     */
    // Template#: 1473, Serial#: 2006
    public void fdivrs(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD8, (byte) 0x07, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivrl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fdivrl    ebx[esi * 4 + 305419896]}
     */
    // Template#: 1474, Serial#: 2230
    public void fdivrl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDC, (byte) 0x07, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivrs  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fdivrs    L1: 0x12345678[esi * 4]}
     */
    // Template#: 1475, Serial#: 1973
    public void m_fdivrs(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD8, (byte) 0x07, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivrl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fdivrl    L1: 0x12345678[esi * 4]}
     */
    // Template#: 1476, Serial#: 2197
    public void m_fdivrl(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDC, (byte) 0x07, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivrs  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fdivrs    [ebx + 305419896]}
     */
    // Template#: 1477, Serial#: 2005
    public void fdivrs(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD8, (byte) 0x07, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivrl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fdivrl    [ebx + 305419896]}
     */
    // Template#: 1478, Serial#: 2229
    public void fdivrl(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDC, (byte) 0x07, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fdivp  }<i>st_i</i>
     * Example disassembly syntax: {@code fdivp     st(0), st}
     */
    // Template#: 1479, Serial#: 5600
    public void fdivrp___ST(FPStackRegister st_i) {
        assemble0308((byte) 0xF0, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code ffree  }<i>st_i</i>
     * Example disassembly syntax: {@code ffree     st(0)}
     */
    // Template#: 1480, Serial#: 5588
    public void ffree(FPStackRegister st_i) {
        assemble0314((byte) 0xC0, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code ffreep  }<i>st_i</i>
     * Example disassembly syntax: {@code ffreep    st(0)}
     */
    // Template#: 1481, Serial#: 5602
    public void ffreep(FPStackRegister st_i) {
        assemble0312((byte) 0xC0, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fiaddl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fiaddl    ebx[esi * 4 + 18]}
     */
    // Template#: 1482, Serial#: 2096
    public void fiaddl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDA, (byte) 0x00, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fiadds  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fiadds    ebx[esi * 4 + 18]}
     */
    // Template#: 1483, Serial#: 2312
    public void fiadds(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDE, (byte) 0x00, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fiaddl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fiaddl    [ebx + 18]}
     */
    // Template#: 1484, Serial#: 2095
    public void fiaddl(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDA, (byte) 0x00, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fiadds  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fiadds    [ebx + 18]}
     */
    // Template#: 1485, Serial#: 2311
    public void fiadds(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDE, (byte) 0x00, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fiaddl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fiaddl    ebx[esi * 4]}
     */
    // Template#: 1486, Serial#: 2064
    public void fiaddl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDA, (byte) 0x00, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fiadds  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fiadds    ebx[esi * 4]}
     */
    // Template#: 1487, Serial#: 2280
    public void fiadds(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDE, (byte) 0x00, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fiaddl  }<i>destination</i>
     * Example disassembly syntax: {@code fiaddl    [ebx]}
     */
    // Template#: 1488, Serial#: 2063
    public void fiaddl(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDA, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fiadds  }<i>destination</i>
     * Example disassembly syntax: {@code fiadds    [ebx]}
     */
    // Template#: 1489, Serial#: 2279
    public void fiadds(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDE, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fiaddl  }<i>m32</i>
     * Example disassembly syntax: {@code fiaddl    [L1: 0x12345678]}
     */
    // Template#: 1490, Serial#: 2066
    public void m_fiaddl(int m32) {
        assemble0299((byte) 0xDA, (byte) 0x00, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fiadds  }<i>m32</i>
     * Example disassembly syntax: {@code fiadds    [L1: 0x12345678]}
     */
    // Template#: 1491, Serial#: 2282
    public void m_fiadds(int m32) {
        assemble0299((byte) 0xDE, (byte) 0x00, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fiaddl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fiaddl    ebx[esi * 4 + 305419896]}
     */
    // Template#: 1492, Serial#: 2112
    public void fiaddl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDA, (byte) 0x00, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fiadds  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fiadds    ebx[esi * 4 + 305419896]}
     */
    // Template#: 1493, Serial#: 2328
    public void fiadds(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDE, (byte) 0x00, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fiaddl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fiaddl    L1: 0x12345678[esi * 4]}
     */
    // Template#: 1494, Serial#: 2065
    public void m_fiaddl(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDA, (byte) 0x00, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fiadds  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fiadds    L1: 0x12345678[esi * 4]}
     */
    // Template#: 1495, Serial#: 2281
    public void m_fiadds(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDE, (byte) 0x00, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fiaddl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fiaddl    [ebx + 305419896]}
     */
    // Template#: 1496, Serial#: 2111
    public void fiaddl(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDA, (byte) 0x00, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fiadds  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fiadds    [ebx + 305419896]}
     */
    // Template#: 1497, Serial#: 2327
    public void fiadds(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDE, (byte) 0x00, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficoml  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ficoml    ebx[esi * 4 + 18]}
     */
    // Template#: 1498, Serial#: 2100
    public void ficoml(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDA, (byte) 0x02, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficoms  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ficoms    ebx[esi * 4 + 18]}
     */
    // Template#: 1499, Serial#: 2316
    public void ficoms(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDE, (byte) 0x02, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficoml  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code ficoml    [ebx + 18]}
     */
    // Template#: 1500, Serial#: 2099
    public void ficoml(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDA, (byte) 0x02, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficoms  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code ficoms    [ebx + 18]}
     */
    // Template#: 1501, Serial#: 2315
    public void ficoms(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDE, (byte) 0x02, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficoml  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ficoml    ebx[esi * 4]}
     */
    // Template#: 1502, Serial#: 2072
    public void ficoml(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDA, (byte) 0x02, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficoms  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ficoms    ebx[esi * 4]}
     */
    // Template#: 1503, Serial#: 2288
    public void ficoms(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDE, (byte) 0x02, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficoml  }<i>destination</i>
     * Example disassembly syntax: {@code ficoml    [ebx]}
     */
    // Template#: 1504, Serial#: 2071
    public void ficoml(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDA, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficoms  }<i>destination</i>
     * Example disassembly syntax: {@code ficoms    [ebx]}
     */
    // Template#: 1505, Serial#: 2287
    public void ficoms(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDE, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficoml  }<i>m32</i>
     * Example disassembly syntax: {@code ficoml    [L1: 0x12345678]}
     */
    // Template#: 1506, Serial#: 2074
    public void m_ficoml(int m32) {
        assemble0299((byte) 0xDA, (byte) 0x02, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficoms  }<i>m32</i>
     * Example disassembly syntax: {@code ficoms    [L1: 0x12345678]}
     */
    // Template#: 1507, Serial#: 2290
    public void m_ficoms(int m32) {
        assemble0299((byte) 0xDE, (byte) 0x02, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficoml  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ficoml    ebx[esi * 4 + 305419896]}
     */
    // Template#: 1508, Serial#: 2116
    public void ficoml(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDA, (byte) 0x02, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficoms  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ficoms    ebx[esi * 4 + 305419896]}
     */
    // Template#: 1509, Serial#: 2332
    public void ficoms(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDE, (byte) 0x02, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficoml  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ficoml    L1: 0x12345678[esi * 4]}
     */
    // Template#: 1510, Serial#: 2073
    public void m_ficoml(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDA, (byte) 0x02, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficoms  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ficoms    L1: 0x12345678[esi * 4]}
     */
    // Template#: 1511, Serial#: 2289
    public void m_ficoms(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDE, (byte) 0x02, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficoml  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code ficoml    [ebx + 305419896]}
     */
    // Template#: 1512, Serial#: 2115
    public void ficoml(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDA, (byte) 0x02, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficoms  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code ficoms    [ebx + 305419896]}
     */
    // Template#: 1513, Serial#: 2331
    public void ficoms(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDE, (byte) 0x02, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficompl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ficompl   ebx[esi * 4 + 18]}
     */
    // Template#: 1514, Serial#: 2102
    public void ficompl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDA, (byte) 0x03, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficomps  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ficomps   ebx[esi * 4 + 18]}
     */
    // Template#: 1515, Serial#: 2318
    public void ficomps(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDE, (byte) 0x03, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficompl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code ficompl   [ebx + 18]}
     */
    // Template#: 1516, Serial#: 2101
    public void ficompl(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDA, (byte) 0x03, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficomps  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code ficomps   [ebx + 18]}
     */
    // Template#: 1517, Serial#: 2317
    public void ficomps(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDE, (byte) 0x03, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficompl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ficompl   ebx[esi * 4]}
     */
    // Template#: 1518, Serial#: 2076
    public void ficompl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDA, (byte) 0x03, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficomps  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ficomps   ebx[esi * 4]}
     */
    // Template#: 1519, Serial#: 2292
    public void ficomps(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDE, (byte) 0x03, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficompl  }<i>destination</i>
     * Example disassembly syntax: {@code ficompl   [ebx]}
     */
    // Template#: 1520, Serial#: 2075
    public void ficompl(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDA, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficomps  }<i>destination</i>
     * Example disassembly syntax: {@code ficomps   [ebx]}
     */
    // Template#: 1521, Serial#: 2291
    public void ficomps(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDE, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficompl  }<i>m32</i>
     * Example disassembly syntax: {@code ficompl   [L1: 0x12345678]}
     */
    // Template#: 1522, Serial#: 2078
    public void m_ficompl(int m32) {
        assemble0299((byte) 0xDA, (byte) 0x03, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficomps  }<i>m32</i>
     * Example disassembly syntax: {@code ficomps   [L1: 0x12345678]}
     */
    // Template#: 1523, Serial#: 2294
    public void m_ficomps(int m32) {
        assemble0299((byte) 0xDE, (byte) 0x03, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficompl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ficompl   ebx[esi * 4 + 305419896]}
     */
    // Template#: 1524, Serial#: 2118
    public void ficompl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDA, (byte) 0x03, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficomps  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ficomps   ebx[esi * 4 + 305419896]}
     */
    // Template#: 1525, Serial#: 2334
    public void ficomps(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDE, (byte) 0x03, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficompl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ficompl   L1: 0x12345678[esi * 4]}
     */
    // Template#: 1526, Serial#: 2077
    public void m_ficompl(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDA, (byte) 0x03, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficomps  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ficomps   L1: 0x12345678[esi * 4]}
     */
    // Template#: 1527, Serial#: 2293
    public void m_ficomps(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDE, (byte) 0x03, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficompl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code ficompl   [ebx + 305419896]}
     */
    // Template#: 1528, Serial#: 2117
    public void ficompl(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDA, (byte) 0x03, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code ficomps  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code ficomps   [ebx + 305419896]}
     */
    // Template#: 1529, Serial#: 2333
    public void ficomps(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDE, (byte) 0x03, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fidivl    ebx[esi * 4 + 18]}
     */
    // Template#: 1530, Serial#: 2108
    public void fidivl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDA, (byte) 0x06, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivs  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fidivs    ebx[esi * 4 + 18]}
     */
    // Template#: 1531, Serial#: 2324
    public void fidivs(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDE, (byte) 0x06, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fidivl    [ebx + 18]}
     */
    // Template#: 1532, Serial#: 2107
    public void fidivl(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDA, (byte) 0x06, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivs  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fidivs    [ebx + 18]}
     */
    // Template#: 1533, Serial#: 2323
    public void fidivs(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDE, (byte) 0x06, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fidivl    ebx[esi * 4]}
     */
    // Template#: 1534, Serial#: 2088
    public void fidivl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDA, (byte) 0x06, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivs  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fidivs    ebx[esi * 4]}
     */
    // Template#: 1535, Serial#: 2304
    public void fidivs(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDE, (byte) 0x06, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivl  }<i>destination</i>
     * Example disassembly syntax: {@code fidivl    [ebx]}
     */
    // Template#: 1536, Serial#: 2087
    public void fidivl(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDA, (byte) 0x06, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivs  }<i>destination</i>
     * Example disassembly syntax: {@code fidivs    [ebx]}
     */
    // Template#: 1537, Serial#: 2303
    public void fidivs(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDE, (byte) 0x06, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivl  }<i>m32</i>
     * Example disassembly syntax: {@code fidivl    [L1: 0x12345678]}
     */
    // Template#: 1538, Serial#: 2090
    public void m_fidivl(int m32) {
        assemble0299((byte) 0xDA, (byte) 0x06, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivs  }<i>m32</i>
     * Example disassembly syntax: {@code fidivs    [L1: 0x12345678]}
     */
    // Template#: 1539, Serial#: 2306
    public void m_fidivs(int m32) {
        assemble0299((byte) 0xDE, (byte) 0x06, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fidivl    ebx[esi * 4 + 305419896]}
     */
    // Template#: 1540, Serial#: 2124
    public void fidivl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDA, (byte) 0x06, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivs  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fidivs    ebx[esi * 4 + 305419896]}
     */
    // Template#: 1541, Serial#: 2340
    public void fidivs(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDE, (byte) 0x06, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fidivl    L1: 0x12345678[esi * 4]}
     */
    // Template#: 1542, Serial#: 2089
    public void m_fidivl(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDA, (byte) 0x06, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivs  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fidivs    L1: 0x12345678[esi * 4]}
     */
    // Template#: 1543, Serial#: 2305
    public void m_fidivs(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDE, (byte) 0x06, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fidivl    [ebx + 305419896]}
     */
    // Template#: 1544, Serial#: 2123
    public void fidivl(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDA, (byte) 0x06, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivs  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fidivs    [ebx + 305419896]}
     */
    // Template#: 1545, Serial#: 2339
    public void fidivs(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDE, (byte) 0x06, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivrl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fidivrl   ebx[esi * 4 + 18]}
     */
    // Template#: 1546, Serial#: 2110
    public void fidivrl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDA, (byte) 0x07, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivrs  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fidivrs   ebx[esi * 4 + 18]}
     */
    // Template#: 1547, Serial#: 2326
    public void fidivrs(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDE, (byte) 0x07, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivrl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fidivrl   [ebx + 18]}
     */
    // Template#: 1548, Serial#: 2109
    public void fidivrl(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDA, (byte) 0x07, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivrs  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fidivrs   [ebx + 18]}
     */
    // Template#: 1549, Serial#: 2325
    public void fidivrs(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDE, (byte) 0x07, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivrl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fidivrl   ebx[esi * 4]}
     */
    // Template#: 1550, Serial#: 2092
    public void fidivrl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDA, (byte) 0x07, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivrs  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fidivrs   ebx[esi * 4]}
     */
    // Template#: 1551, Serial#: 2308
    public void fidivrs(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDE, (byte) 0x07, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivrl  }<i>destination</i>
     * Example disassembly syntax: {@code fidivrl   [ebx]}
     */
    // Template#: 1552, Serial#: 2091
    public void fidivrl(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDA, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivrs  }<i>destination</i>
     * Example disassembly syntax: {@code fidivrs   [ebx]}
     */
    // Template#: 1553, Serial#: 2307
    public void fidivrs(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDE, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivrl  }<i>m32</i>
     * Example disassembly syntax: {@code fidivrl   [L1: 0x12345678]}
     */
    // Template#: 1554, Serial#: 2094
    public void m_fidivrl(int m32) {
        assemble0299((byte) 0xDA, (byte) 0x07, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivrs  }<i>m32</i>
     * Example disassembly syntax: {@code fidivrs   [L1: 0x12345678]}
     */
    // Template#: 1555, Serial#: 2310
    public void m_fidivrs(int m32) {
        assemble0299((byte) 0xDE, (byte) 0x07, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivrl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fidivrl   ebx[esi * 4 + 305419896]}
     */
    // Template#: 1556, Serial#: 2126
    public void fidivrl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDA, (byte) 0x07, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivrs  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fidivrs   ebx[esi * 4 + 305419896]}
     */
    // Template#: 1557, Serial#: 2342
    public void fidivrs(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDE, (byte) 0x07, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivrl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fidivrl   L1: 0x12345678[esi * 4]}
     */
    // Template#: 1558, Serial#: 2093
    public void m_fidivrl(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDA, (byte) 0x07, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivrs  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fidivrs   L1: 0x12345678[esi * 4]}
     */
    // Template#: 1559, Serial#: 2309
    public void m_fidivrs(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDE, (byte) 0x07, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivrl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fidivrl   [ebx + 305419896]}
     */
    // Template#: 1560, Serial#: 2125
    public void fidivrl(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDA, (byte) 0x07, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fidivrs  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fidivrs   [ebx + 305419896]}
     */
    // Template#: 1561, Serial#: 2341
    public void fidivrs(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDE, (byte) 0x07, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fildl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fildl     ebx[esi * 4 + 18]}
     */
    // Template#: 1562, Serial#: 2148
    public void fildl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDB, (byte) 0x00, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code filds  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code filds     ebx[esi * 4 + 18]}
     */
    // Template#: 1563, Serial#: 2372
    public void filds(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDF, (byte) 0x00, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fildq  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fildq     ebx[esi * 4 + 18]}
     */
    // Template#: 1564, Serial#: 2380
    public void fildq(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDF, (byte) 0x05, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fildl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fildl     [ebx + 18]}
     */
    // Template#: 1565, Serial#: 2147
    public void fildl(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDB, (byte) 0x00, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code filds  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code filds     [ebx + 18]}
     */
    // Template#: 1566, Serial#: 2371
    public void filds(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDF, (byte) 0x00, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fildq  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fildq     [ebx + 18]}
     */
    // Template#: 1567, Serial#: 2379
    public void fildq(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDF, (byte) 0x05, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fildl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fildl     ebx[esi * 4]}
     */
    // Template#: 1568, Serial#: 2128
    public void fildl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDB, (byte) 0x00, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code filds  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code filds     ebx[esi * 4]}
     */
    // Template#: 1569, Serial#: 2344
    public void filds(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDF, (byte) 0x00, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fildq  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fildq     ebx[esi * 4]}
     */
    // Template#: 1570, Serial#: 2360
    public void fildq(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDF, (byte) 0x05, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fildl  }<i>destination</i>
     * Example disassembly syntax: {@code fildl     [ebx]}
     */
    // Template#: 1571, Serial#: 2127
    public void fildl(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDB, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code filds  }<i>destination</i>
     * Example disassembly syntax: {@code filds     [ebx]}
     */
    // Template#: 1572, Serial#: 2343
    public void filds(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDF, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fildq  }<i>destination</i>
     * Example disassembly syntax: {@code fildq     [ebx]}
     */
    // Template#: 1573, Serial#: 2359
    public void fildq(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDF, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fildl  }<i>m32</i>
     * Example disassembly syntax: {@code fildl     [L1: 0x12345678]}
     */
    // Template#: 1574, Serial#: 2130
    public void m_fildl(int m32) {
        assemble0299((byte) 0xDB, (byte) 0x00, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code filds  }<i>m32</i>
     * Example disassembly syntax: {@code filds     [L1: 0x12345678]}
     */
    // Template#: 1575, Serial#: 2346
    public void m_filds(int m32) {
        assemble0299((byte) 0xDF, (byte) 0x00, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fildq  }<i>m32</i>
     * Example disassembly syntax: {@code fildq     [L1: 0x12345678]}
     */
    // Template#: 1576, Serial#: 2362
    public void m_fildq(int m32) {
        assemble0299((byte) 0xDF, (byte) 0x05, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fildl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fildl     ebx[esi * 4 + 305419896]}
     */
    // Template#: 1577, Serial#: 2158
    public void fildl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDB, (byte) 0x00, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code filds  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code filds     ebx[esi * 4 + 305419896]}
     */
    // Template#: 1578, Serial#: 2386
    public void filds(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDF, (byte) 0x00, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fildq  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fildq     ebx[esi * 4 + 305419896]}
     */
    // Template#: 1579, Serial#: 2394
    public void fildq(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDF, (byte) 0x05, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fildl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fildl     L1: 0x12345678[esi * 4]}
     */
    // Template#: 1580, Serial#: 2129
    public void m_fildl(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDB, (byte) 0x00, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code filds  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code filds     L1: 0x12345678[esi * 4]}
     */
    // Template#: 1581, Serial#: 2345
    public void m_filds(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDF, (byte) 0x00, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fildq  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fildq     L1: 0x12345678[esi * 4]}
     */
    // Template#: 1582, Serial#: 2361
    public void m_fildq(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDF, (byte) 0x05, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fildl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fildl     [ebx + 305419896]}
     */
    // Template#: 1583, Serial#: 2157
    public void fildl(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDB, (byte) 0x00, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code filds  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code filds     [ebx + 305419896]}
     */
    // Template#: 1584, Serial#: 2385
    public void filds(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDF, (byte) 0x00, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fildq  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fildq     [ebx + 305419896]}
     */
    // Template#: 1585, Serial#: 2393
    public void fildq(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDF, (byte) 0x05, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fimull  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fimull    ebx[esi * 4 + 18]}
     */
    // Template#: 1586, Serial#: 2098
    public void fimull(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDA, (byte) 0x01, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fimuls  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fimuls    ebx[esi * 4 + 18]}
     */
    // Template#: 1587, Serial#: 2314
    public void fimuls(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDE, (byte) 0x01, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fimull  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fimull    [ebx + 18]}
     */
    // Template#: 1588, Serial#: 2097
    public void fimull(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDA, (byte) 0x01, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fimuls  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fimuls    [ebx + 18]}
     */
    // Template#: 1589, Serial#: 2313
    public void fimuls(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDE, (byte) 0x01, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fimull  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fimull    ebx[esi * 4]}
     */
    // Template#: 1590, Serial#: 2068
    public void fimull(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDA, (byte) 0x01, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fimuls  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fimuls    ebx[esi * 4]}
     */
    // Template#: 1591, Serial#: 2284
    public void fimuls(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDE, (byte) 0x01, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fimull  }<i>destination</i>
     * Example disassembly syntax: {@code fimull    [ebx]}
     */
    // Template#: 1592, Serial#: 2067
    public void fimull(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDA, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fimuls  }<i>destination</i>
     * Example disassembly syntax: {@code fimuls    [ebx]}
     */
    // Template#: 1593, Serial#: 2283
    public void fimuls(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDE, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fimull  }<i>m32</i>
     * Example disassembly syntax: {@code fimull    [L1: 0x12345678]}
     */
    // Template#: 1594, Serial#: 2070
    public void m_fimull(int m32) {
        assemble0299((byte) 0xDA, (byte) 0x01, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fimuls  }<i>m32</i>
     * Example disassembly syntax: {@code fimuls    [L1: 0x12345678]}
     */
    // Template#: 1595, Serial#: 2286
    public void m_fimuls(int m32) {
        assemble0299((byte) 0xDE, (byte) 0x01, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fimull  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fimull    ebx[esi * 4 + 305419896]}
     */
    // Template#: 1596, Serial#: 2114
    public void fimull(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDA, (byte) 0x01, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fimuls  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fimuls    ebx[esi * 4 + 305419896]}
     */
    // Template#: 1597, Serial#: 2330
    public void fimuls(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDE, (byte) 0x01, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fimull  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fimull    L1: 0x12345678[esi * 4]}
     */
    // Template#: 1598, Serial#: 2069
    public void m_fimull(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDA, (byte) 0x01, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fimuls  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fimuls    L1: 0x12345678[esi * 4]}
     */
    // Template#: 1599, Serial#: 2285
    public void m_fimuls(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDE, (byte) 0x01, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fimull  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fimull    [ebx + 305419896]}
     */
    // Template#: 1600, Serial#: 2113
    public void fimull(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDA, (byte) 0x01, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fimuls  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fimuls    [ebx + 305419896]}
     */
    // Template#: 1601, Serial#: 2329
    public void fimuls(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDE, (byte) 0x01, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fincstp  }
     * Example disassembly syntax: {@code fincstp }
     */
    // Template#: 1602, Serial#: 5558
    public void fincstp() {
        assemble0305((byte) 0xF7);
    }

    /**
     * Pseudo-external assembler syntax: {@code finit  }
     * Example disassembly syntax: {@code finit   }
     */
    // Template#: 1603, Serial#: 5577
    public void finit() {
        assemble0309((byte) 0xE3);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fistl     ebx[esi * 4 + 18]}
     */
    // Template#: 1604, Serial#: 2150
    public void fistl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDB, (byte) 0x02, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fists  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fists     ebx[esi * 4 + 18]}
     */
    // Template#: 1605, Serial#: 2374
    public void fists(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDF, (byte) 0x02, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fistl     [ebx + 18]}
     */
    // Template#: 1606, Serial#: 2149
    public void fistl(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDB, (byte) 0x02, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fists  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fists     [ebx + 18]}
     */
    // Template#: 1607, Serial#: 2373
    public void fists(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDF, (byte) 0x02, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fistl     ebx[esi * 4]}
     */
    // Template#: 1608, Serial#: 2132
    public void fistl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDB, (byte) 0x02, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fists  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fists     ebx[esi * 4]}
     */
    // Template#: 1609, Serial#: 2348
    public void fists(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDF, (byte) 0x02, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistl  }<i>destination</i>
     * Example disassembly syntax: {@code fistl     [ebx]}
     */
    // Template#: 1610, Serial#: 2131
    public void fistl(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDB, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fists  }<i>destination</i>
     * Example disassembly syntax: {@code fists     [ebx]}
     */
    // Template#: 1611, Serial#: 2347
    public void fists(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDF, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistl  }<i>m32</i>
     * Example disassembly syntax: {@code fistl     [L1: 0x12345678]}
     */
    // Template#: 1612, Serial#: 2134
    public void m_fistl(int m32) {
        assemble0299((byte) 0xDB, (byte) 0x02, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fists  }<i>m32</i>
     * Example disassembly syntax: {@code fists     [L1: 0x12345678]}
     */
    // Template#: 1613, Serial#: 2350
    public void m_fists(int m32) {
        assemble0299((byte) 0xDF, (byte) 0x02, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fistl     ebx[esi * 4 + 305419896]}
     */
    // Template#: 1614, Serial#: 2160
    public void fistl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDB, (byte) 0x02, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fists  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fists     ebx[esi * 4 + 305419896]}
     */
    // Template#: 1615, Serial#: 2388
    public void fists(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDF, (byte) 0x02, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fistl     L1: 0x12345678[esi * 4]}
     */
    // Template#: 1616, Serial#: 2133
    public void m_fistl(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDB, (byte) 0x02, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fists  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fists     L1: 0x12345678[esi * 4]}
     */
    // Template#: 1617, Serial#: 2349
    public void m_fists(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDF, (byte) 0x02, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fistl     [ebx + 305419896]}
     */
    // Template#: 1618, Serial#: 2159
    public void fistl(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDB, (byte) 0x02, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fists  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fists     [ebx + 305419896]}
     */
    // Template#: 1619, Serial#: 2387
    public void fists(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDF, (byte) 0x02, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistpl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fistpl    ebx[esi * 4 + 18]}
     */
    // Template#: 1620, Serial#: 2152
    public void fistpl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDB, (byte) 0x03, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistps  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fistps    ebx[esi * 4 + 18]}
     */
    // Template#: 1621, Serial#: 2376
    public void fistps(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDF, (byte) 0x03, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistpq  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fistpq    ebx[esi * 4 + 18]}
     */
    // Template#: 1622, Serial#: 2384
    public void fistpq(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDF, (byte) 0x07, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistpl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fistpl    [ebx + 18]}
     */
    // Template#: 1623, Serial#: 2151
    public void fistpl(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDB, (byte) 0x03, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistps  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fistps    [ebx + 18]}
     */
    // Template#: 1624, Serial#: 2375
    public void fistps(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDF, (byte) 0x03, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistpq  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fistpq    [ebx + 18]}
     */
    // Template#: 1625, Serial#: 2383
    public void fistpq(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDF, (byte) 0x07, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistpl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fistpl    ebx[esi * 4]}
     */
    // Template#: 1626, Serial#: 2136
    public void fistpl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDB, (byte) 0x03, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistps  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fistps    ebx[esi * 4]}
     */
    // Template#: 1627, Serial#: 2352
    public void fistps(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDF, (byte) 0x03, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistpq  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fistpq    ebx[esi * 4]}
     */
    // Template#: 1628, Serial#: 2368
    public void fistpq(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDF, (byte) 0x07, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistpl  }<i>destination</i>
     * Example disassembly syntax: {@code fistpl    [ebx]}
     */
    // Template#: 1629, Serial#: 2135
    public void fistpl(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDB, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistps  }<i>destination</i>
     * Example disassembly syntax: {@code fistps    [ebx]}
     */
    // Template#: 1630, Serial#: 2351
    public void fistps(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDF, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistpq  }<i>destination</i>
     * Example disassembly syntax: {@code fistpq    [ebx]}
     */
    // Template#: 1631, Serial#: 2367
    public void fistpq(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDF, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistpl  }<i>m32</i>
     * Example disassembly syntax: {@code fistpl    [L1: 0x12345678]}
     */
    // Template#: 1632, Serial#: 2138
    public void m_fistpl(int m32) {
        assemble0299((byte) 0xDB, (byte) 0x03, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistps  }<i>m32</i>
     * Example disassembly syntax: {@code fistps    [L1: 0x12345678]}
     */
    // Template#: 1633, Serial#: 2354
    public void m_fistps(int m32) {
        assemble0299((byte) 0xDF, (byte) 0x03, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistpq  }<i>m32</i>
     * Example disassembly syntax: {@code fistpq    [L1: 0x12345678]}
     */
    // Template#: 1634, Serial#: 2370
    public void m_fistpq(int m32) {
        assemble0299((byte) 0xDF, (byte) 0x07, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistpl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fistpl    ebx[esi * 4 + 305419896]}
     */
    // Template#: 1635, Serial#: 2162
    public void fistpl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDB, (byte) 0x03, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistps  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fistps    ebx[esi * 4 + 305419896]}
     */
    // Template#: 1636, Serial#: 2390
    public void fistps(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDF, (byte) 0x03, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistpq  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fistpq    ebx[esi * 4 + 305419896]}
     */
    // Template#: 1637, Serial#: 2398
    public void fistpq(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDF, (byte) 0x07, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistpl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fistpl    L1: 0x12345678[esi * 4]}
     */
    // Template#: 1638, Serial#: 2137
    public void m_fistpl(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDB, (byte) 0x03, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistps  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fistps    L1: 0x12345678[esi * 4]}
     */
    // Template#: 1639, Serial#: 2353
    public void m_fistps(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDF, (byte) 0x03, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistpq  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fistpq    L1: 0x12345678[esi * 4]}
     */
    // Template#: 1640, Serial#: 2369
    public void m_fistpq(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDF, (byte) 0x07, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistpl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fistpl    [ebx + 305419896]}
     */
    // Template#: 1641, Serial#: 2161
    public void fistpl(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDB, (byte) 0x03, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistps  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fistps    [ebx + 305419896]}
     */
    // Template#: 1642, Serial#: 2389
    public void fistps(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDF, (byte) 0x03, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fistpq  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fistpq    [ebx + 305419896]}
     */
    // Template#: 1643, Serial#: 2397
    public void fistpq(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDF, (byte) 0x07, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fisubl    ebx[esi * 4 + 18]}
     */
    // Template#: 1644, Serial#: 2104
    public void fisubl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDA, (byte) 0x04, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubs  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fisubs    ebx[esi * 4 + 18]}
     */
    // Template#: 1645, Serial#: 2320
    public void fisubs(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDE, (byte) 0x04, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fisubl    [ebx + 18]}
     */
    // Template#: 1646, Serial#: 2103
    public void fisubl(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDA, (byte) 0x04, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubs  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fisubs    [ebx + 18]}
     */
    // Template#: 1647, Serial#: 2319
    public void fisubs(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDE, (byte) 0x04, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fisubl    ebx[esi * 4]}
     */
    // Template#: 1648, Serial#: 2080
    public void fisubl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDA, (byte) 0x04, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubs  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fisubs    ebx[esi * 4]}
     */
    // Template#: 1649, Serial#: 2296
    public void fisubs(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDE, (byte) 0x04, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubl  }<i>destination</i>
     * Example disassembly syntax: {@code fisubl    [ebx]}
     */
    // Template#: 1650, Serial#: 2079
    public void fisubl(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDA, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubs  }<i>destination</i>
     * Example disassembly syntax: {@code fisubs    [ebx]}
     */
    // Template#: 1651, Serial#: 2295
    public void fisubs(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDE, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubl  }<i>m32</i>
     * Example disassembly syntax: {@code fisubl    [L1: 0x12345678]}
     */
    // Template#: 1652, Serial#: 2082
    public void m_fisubl(int m32) {
        assemble0299((byte) 0xDA, (byte) 0x04, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubs  }<i>m32</i>
     * Example disassembly syntax: {@code fisubs    [L1: 0x12345678]}
     */
    // Template#: 1653, Serial#: 2298
    public void m_fisubs(int m32) {
        assemble0299((byte) 0xDE, (byte) 0x04, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fisubl    ebx[esi * 4 + 305419896]}
     */
    // Template#: 1654, Serial#: 2120
    public void fisubl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDA, (byte) 0x04, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubs  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fisubs    ebx[esi * 4 + 305419896]}
     */
    // Template#: 1655, Serial#: 2336
    public void fisubs(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDE, (byte) 0x04, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fisubl    L1: 0x12345678[esi * 4]}
     */
    // Template#: 1656, Serial#: 2081
    public void m_fisubl(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDA, (byte) 0x04, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubs  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fisubs    L1: 0x12345678[esi * 4]}
     */
    // Template#: 1657, Serial#: 2297
    public void m_fisubs(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDE, (byte) 0x04, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fisubl    [ebx + 305419896]}
     */
    // Template#: 1658, Serial#: 2119
    public void fisubl(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDA, (byte) 0x04, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubs  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fisubs    [ebx + 305419896]}
     */
    // Template#: 1659, Serial#: 2335
    public void fisubs(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDE, (byte) 0x04, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubrl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fisubrl   ebx[esi * 4 + 18]}
     */
    // Template#: 1660, Serial#: 2106
    public void fisubrl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDA, (byte) 0x05, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubrs  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fisubrs   ebx[esi * 4 + 18]}
     */
    // Template#: 1661, Serial#: 2322
    public void fisubrs(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDE, (byte) 0x05, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubrl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fisubrl   [ebx + 18]}
     */
    // Template#: 1662, Serial#: 2105
    public void fisubrl(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDA, (byte) 0x05, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubrs  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fisubrs   [ebx + 18]}
     */
    // Template#: 1663, Serial#: 2321
    public void fisubrs(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDE, (byte) 0x05, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubrl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fisubrl   ebx[esi * 4]}
     */
    // Template#: 1664, Serial#: 2084
    public void fisubrl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDA, (byte) 0x05, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubrs  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fisubrs   ebx[esi * 4]}
     */
    // Template#: 1665, Serial#: 2300
    public void fisubrs(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDE, (byte) 0x05, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubrl  }<i>destination</i>
     * Example disassembly syntax: {@code fisubrl   [ebx]}
     */
    // Template#: 1666, Serial#: 2083
    public void fisubrl(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDA, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubrs  }<i>destination</i>
     * Example disassembly syntax: {@code fisubrs   [ebx]}
     */
    // Template#: 1667, Serial#: 2299
    public void fisubrs(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDE, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubrl  }<i>m32</i>
     * Example disassembly syntax: {@code fisubrl   [L1: 0x12345678]}
     */
    // Template#: 1668, Serial#: 2086
    public void m_fisubrl(int m32) {
        assemble0299((byte) 0xDA, (byte) 0x05, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubrs  }<i>m32</i>
     * Example disassembly syntax: {@code fisubrs   [L1: 0x12345678]}
     */
    // Template#: 1669, Serial#: 2302
    public void m_fisubrs(int m32) {
        assemble0299((byte) 0xDE, (byte) 0x05, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubrl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fisubrl   ebx[esi * 4 + 305419896]}
     */
    // Template#: 1670, Serial#: 2122
    public void fisubrl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDA, (byte) 0x05, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubrs  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fisubrs   ebx[esi * 4 + 305419896]}
     */
    // Template#: 1671, Serial#: 2338
    public void fisubrs(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDE, (byte) 0x05, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubrl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fisubrl   L1: 0x12345678[esi * 4]}
     */
    // Template#: 1672, Serial#: 2085
    public void m_fisubrl(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDA, (byte) 0x05, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubrs  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fisubrs   L1: 0x12345678[esi * 4]}
     */
    // Template#: 1673, Serial#: 2301
    public void m_fisubrs(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDE, (byte) 0x05, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubrl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fisubrl   [ebx + 305419896]}
     */
    // Template#: 1674, Serial#: 2121
    public void fisubrl(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDA, (byte) 0x05, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fisubrs  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fisubrs   [ebx + 305419896]}
     */
    // Template#: 1675, Serial#: 2337
    public void fisubrs(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDE, (byte) 0x05, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code flds  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code flds      ebx[esi * 4 + 18]}
     */
    // Template#: 1676, Serial#: 2036
    public void flds(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD9, (byte) 0x00, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldt  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fldt      ebx[esi * 4 + 18]}
     */
    // Template#: 1677, Serial#: 2154
    public void fldt(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDB, (byte) 0x05, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fldl      ebx[esi * 4 + 18]}
     */
    // Template#: 1678, Serial#: 2256
    public void fldl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDD, (byte) 0x00, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code flds  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code flds      [ebx + 18]}
     */
    // Template#: 1679, Serial#: 2035
    public void flds(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD9, (byte) 0x00, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldt  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fldt      [ebx + 18]}
     */
    // Template#: 1680, Serial#: 2153
    public void fldt(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDB, (byte) 0x05, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fldl      [ebx + 18]}
     */
    // Template#: 1681, Serial#: 2255
    public void fldl(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDD, (byte) 0x00, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code flds  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code flds      ebx[esi * 4]}
     */
    // Template#: 1682, Serial#: 2008
    public void flds(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD9, (byte) 0x00, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldt  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fldt      ebx[esi * 4]}
     */
    // Template#: 1683, Serial#: 2140
    public void fldt(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDB, (byte) 0x05, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fldl      ebx[esi * 4]}
     */
    // Template#: 1684, Serial#: 2232
    public void fldl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDD, (byte) 0x00, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code flds  }<i>destination</i>
     * Example disassembly syntax: {@code flds      [ebx]}
     */
    // Template#: 1685, Serial#: 2007
    public void flds(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD9, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldt  }<i>destination</i>
     * Example disassembly syntax: {@code fldt      [ebx]}
     */
    // Template#: 1686, Serial#: 2139
    public void fldt(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDB, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldl  }<i>destination</i>
     * Example disassembly syntax: {@code fldl      [ebx]}
     */
    // Template#: 1687, Serial#: 2231
    public void fldl(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDD, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fld  }<i>st_i</i>
     * Example disassembly syntax: {@code fld       st(0)}
     */
    // Template#: 1688, Serial#: 5536
    public void fld(FPStackRegister st_i) {
        assemble0315((byte) 0xC0, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code flds  }<i>m32</i>
     * Example disassembly syntax: {@code flds      [L1: 0x12345678]}
     */
    // Template#: 1689, Serial#: 2010
    public void m_flds(int m32) {
        assemble0299((byte) 0xD9, (byte) 0x00, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldt  }<i>m32</i>
     * Example disassembly syntax: {@code fldt      [L1: 0x12345678]}
     */
    // Template#: 1690, Serial#: 2142
    public void m_fldt(int m32) {
        assemble0299((byte) 0xDB, (byte) 0x05, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldl  }<i>m32</i>
     * Example disassembly syntax: {@code fldl      [L1: 0x12345678]}
     */
    // Template#: 1691, Serial#: 2234
    public void m_fldl(int m32) {
        assemble0299((byte) 0xDD, (byte) 0x00, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code flds  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code flds      ebx[esi * 4 + 305419896]}
     */
    // Template#: 1692, Serial#: 2050
    public void flds(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD9, (byte) 0x00, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldt  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fldt      ebx[esi * 4 + 305419896]}
     */
    // Template#: 1693, Serial#: 2164
    public void fldt(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDB, (byte) 0x05, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fldl      ebx[esi * 4 + 305419896]}
     */
    // Template#: 1694, Serial#: 2268
    public void fldl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDD, (byte) 0x00, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code flds  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code flds      L1: 0x12345678[esi * 4]}
     */
    // Template#: 1695, Serial#: 2009
    public void m_flds(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD9, (byte) 0x00, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldt  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fldt      L1: 0x12345678[esi * 4]}
     */
    // Template#: 1696, Serial#: 2141
    public void m_fldt(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDB, (byte) 0x05, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fldl      L1: 0x12345678[esi * 4]}
     */
    // Template#: 1697, Serial#: 2233
    public void m_fldl(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDD, (byte) 0x00, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code flds  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code flds      [ebx + 305419896]}
     */
    // Template#: 1698, Serial#: 2049
    public void flds(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD9, (byte) 0x00, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldt  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fldt      [ebx + 305419896]}
     */
    // Template#: 1699, Serial#: 2163
    public void fldt(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDB, (byte) 0x05, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fldl      [ebx + 305419896]}
     */
    // Template#: 1700, Serial#: 2267
    public void fldl(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDD, (byte) 0x00, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fld1  }
     * Example disassembly syntax: {@code fld1    }
     */
    // Template#: 1701, Serial#: 5544
    public void fld1() {
        assemble0305((byte) 0xE8);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldcw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fldcw     ebx[esi * 4 + 18]}
     */
    // Template#: 1702, Serial#: 2044
    public void fldcw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD9, (byte) 0x05, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldcw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fldcw     [ebx + 18]}
     */
    // Template#: 1703, Serial#: 2043
    public void fldcw(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD9, (byte) 0x05, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldcw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fldcw     ebx[esi * 4]}
     */
    // Template#: 1704, Serial#: 2024
    public void fldcw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD9, (byte) 0x05, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldcw  }<i>destination</i>
     * Example disassembly syntax: {@code fldcw     [ebx]}
     */
    // Template#: 1705, Serial#: 2023
    public void fldcw(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD9, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldcw  }<i>m32</i>
     * Example disassembly syntax: {@code fldcw     [L1: 0x12345678]}
     */
    // Template#: 1706, Serial#: 2026
    public void m_fldcw(int m32) {
        assemble0299((byte) 0xD9, (byte) 0x05, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldcw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fldcw     ebx[esi * 4 + 305419896]}
     */
    // Template#: 1707, Serial#: 2058
    public void fldcw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD9, (byte) 0x05, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldcw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fldcw     L1: 0x12345678[esi * 4]}
     */
    // Template#: 1708, Serial#: 2025
    public void m_fldcw(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD9, (byte) 0x05, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldcw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fldcw     [ebx + 305419896]}
     */
    // Template#: 1709, Serial#: 2057
    public void fldcw(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD9, (byte) 0x05, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldenv  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fldenv    ebx[esi * 4 + 18]}
     */
    // Template#: 1710, Serial#: 2042
    public void fldenv(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD9, (byte) 0x04, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldenv  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fldenv    [ebx + 18]}
     */
    // Template#: 1711, Serial#: 2041
    public void fldenv(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD9, (byte) 0x04, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldenv  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fldenv    ebx[esi * 4]}
     */
    // Template#: 1712, Serial#: 2020
    public void fldenv(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD9, (byte) 0x04, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldenv  }<i>destination</i>
     * Example disassembly syntax: {@code fldenv    [ebx]}
     */
    // Template#: 1713, Serial#: 2019
    public void fldenv(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD9, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldenv  }<i>m32</i>
     * Example disassembly syntax: {@code fldenv    [L1: 0x12345678]}
     */
    // Template#: 1714, Serial#: 2022
    public void m_fldenv(int m32) {
        assemble0299((byte) 0xD9, (byte) 0x04, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldenv  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fldenv    ebx[esi * 4 + 305419896]}
     */
    // Template#: 1715, Serial#: 2056
    public void fldenv(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD9, (byte) 0x04, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldenv  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fldenv    L1: 0x12345678[esi * 4]}
     */
    // Template#: 1716, Serial#: 2021
    public void m_fldenv(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD9, (byte) 0x04, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldenv  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fldenv    [ebx + 305419896]}
     */
    // Template#: 1717, Serial#: 2055
    public void fldenv(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD9, (byte) 0x04, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldl2e  }
     * Example disassembly syntax: {@code fldl2e  }
     */
    // Template#: 1718, Serial#: 5546
    public void fldl2e() {
        assemble0305((byte) 0xEA);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldl2t  }
     * Example disassembly syntax: {@code fldl2t  }
     */
    // Template#: 1719, Serial#: 5545
    public void fldl2t() {
        assemble0305((byte) 0xE9);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldlg2  }
     * Example disassembly syntax: {@code fldlg2  }
     */
    // Template#: 1720, Serial#: 5548
    public void fldlg2() {
        assemble0305((byte) 0xEC);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldln2  }
     * Example disassembly syntax: {@code fldln2  }
     */
    // Template#: 1721, Serial#: 5549
    public void fldln2() {
        assemble0305((byte) 0xED);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldpi  }
     * Example disassembly syntax: {@code fldpi   }
     */
    // Template#: 1722, Serial#: 5547
    public void fldpi() {
        assemble0305((byte) 0xEB);
    }

    /**
     * Pseudo-external assembler syntax: {@code fldz  }
     * Example disassembly syntax: {@code fldz    }
     */
    // Template#: 1723, Serial#: 5550
    public void fldz() {
        assemble0305((byte) 0xEE);
    }

    /**
     * Pseudo-external assembler syntax: {@code fmuls  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fmuls     ebx[esi * 4 + 18]}
     */
    // Template#: 1724, Serial#: 1978
    public void fmuls(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD8, (byte) 0x01, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fmull  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fmull     ebx[esi * 4 + 18]}
     */
    // Template#: 1725, Serial#: 2202
    public void fmull(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDC, (byte) 0x01, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fmuls  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fmuls     [ebx + 18]}
     */
    // Template#: 1726, Serial#: 1977
    public void fmuls(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD8, (byte) 0x01, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fmull  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fmull     [ebx + 18]}
     */
    // Template#: 1727, Serial#: 2201
    public void fmull(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDC, (byte) 0x01, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fmuls  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fmuls     ebx[esi * 4]}
     */
    // Template#: 1728, Serial#: 1948
    public void fmuls(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD8, (byte) 0x01, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fmull  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fmull     ebx[esi * 4]}
     */
    // Template#: 1729, Serial#: 2172
    public void fmull(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDC, (byte) 0x01, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fmuls  }<i>destination</i>
     * Example disassembly syntax: {@code fmuls     [ebx]}
     */
    // Template#: 1730, Serial#: 1947
    public void fmuls(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD8, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fmull  }<i>destination</i>
     * Example disassembly syntax: {@code fmull     [ebx]}
     */
    // Template#: 1731, Serial#: 2171
    public void fmull(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDC, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fmul  }<i>st_i</i>
     * Example disassembly syntax: {@code fmul      st, st(0)}
     */
    // Template#: 1732, Serial#: 5529
    public void fmul_ST(FPStackRegister st_i) {
        assemble0306((byte) 0xC8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fmul  }<i>st_i</i>
     * Example disassembly syntax: {@code fmul      st(0), st}
     */
    // Template#: 1733, Serial#: 5581
    public void fmul___ST(FPStackRegister st_i) {
        assemble0307((byte) 0xC8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fmuls  }<i>m32</i>
     * Example disassembly syntax: {@code fmuls     [L1: 0x12345678]}
     */
    // Template#: 1734, Serial#: 1950
    public void m_fmuls(int m32) {
        assemble0299((byte) 0xD8, (byte) 0x01, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fmull  }<i>m32</i>
     * Example disassembly syntax: {@code fmull     [L1: 0x12345678]}
     */
    // Template#: 1735, Serial#: 2174
    public void m_fmull(int m32) {
        assemble0299((byte) 0xDC, (byte) 0x01, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fmuls  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fmuls     ebx[esi * 4 + 305419896]}
     */
    // Template#: 1736, Serial#: 1994
    public void fmuls(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD8, (byte) 0x01, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fmull  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fmull     ebx[esi * 4 + 305419896]}
     */
    // Template#: 1737, Serial#: 2218
    public void fmull(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDC, (byte) 0x01, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fmuls  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fmuls     L1: 0x12345678[esi * 4]}
     */
    // Template#: 1738, Serial#: 1949
    public void m_fmuls(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD8, (byte) 0x01, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fmull  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fmull     L1: 0x12345678[esi * 4]}
     */
    // Template#: 1739, Serial#: 2173
    public void m_fmull(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDC, (byte) 0x01, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fmuls  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fmuls     [ebx + 305419896]}
     */
    // Template#: 1740, Serial#: 1993
    public void fmuls(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD8, (byte) 0x01, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fmull  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fmull     [ebx + 305419896]}
     */
    // Template#: 1741, Serial#: 2217
    public void fmull(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDC, (byte) 0x01, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fmulp  }<i>st_i</i>
     * Example disassembly syntax: {@code fmulp     st(0), st}
     */
    // Template#: 1742, Serial#: 5595
    public void fmulp___ST(FPStackRegister st_i) {
        assemble0308((byte) 0xC8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fnop  }
     * Example disassembly syntax: {@code fnop    }
     */
    // Template#: 1743, Serial#: 5538
    public void fnop() {
        assemble0305((byte) 0xD0);
    }

    /**
     * Pseudo-external assembler syntax: {@code fpatan  }
     * Example disassembly syntax: {@code fpatan  }
     */
    // Template#: 1744, Serial#: 5554
    public void fpatan() {
        assemble0305((byte) 0xF3);
    }

    /**
     * Pseudo-external assembler syntax: {@code fprem  }
     * Example disassembly syntax: {@code fprem   }
     */
    // Template#: 1745, Serial#: 5559
    public void fprem() {
        assemble0305((byte) 0xF8);
    }

    /**
     * Pseudo-external assembler syntax: {@code fprem1  }
     * Example disassembly syntax: {@code fprem1  }
     */
    // Template#: 1746, Serial#: 5556
    public void fprem1() {
        assemble0305((byte) 0xF5);
    }

    /**
     * Pseudo-external assembler syntax: {@code fptan  }
     * Example disassembly syntax: {@code fptan   }
     */
    // Template#: 1747, Serial#: 5553
    public void fptan() {
        assemble0305((byte) 0xF2);
    }

    /**
     * Pseudo-external assembler syntax: {@code frndint  }
     * Example disassembly syntax: {@code frndint }
     */
    // Template#: 1748, Serial#: 5563
    public void frndint() {
        assemble0305((byte) 0xFC);
    }

    /**
     * Pseudo-external assembler syntax: {@code frstor  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code frstor    ebx[esi * 4 + 18]}
     */
    // Template#: 1749, Serial#: 2262
    public void frstor(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDD, (byte) 0x04, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code frstor  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code frstor    [ebx + 18]}
     */
    // Template#: 1750, Serial#: 2261
    public void frstor(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDD, (byte) 0x04, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code frstor  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code frstor    ebx[esi * 4]}
     */
    // Template#: 1751, Serial#: 2244
    public void frstor(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDD, (byte) 0x04, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code frstor  }<i>destination</i>
     * Example disassembly syntax: {@code frstor    [ebx]}
     */
    // Template#: 1752, Serial#: 2243
    public void frstor(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDD, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code frstor  }<i>m32</i>
     * Example disassembly syntax: {@code frstor    [L1: 0x12345678]}
     */
    // Template#: 1753, Serial#: 2246
    public void m_frstor(int m32) {
        assemble0299((byte) 0xDD, (byte) 0x04, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code frstor  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code frstor    ebx[esi * 4 + 305419896]}
     */
    // Template#: 1754, Serial#: 2274
    public void frstor(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDD, (byte) 0x04, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code frstor  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code frstor    L1: 0x12345678[esi * 4]}
     */
    // Template#: 1755, Serial#: 2245
    public void m_frstor(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDD, (byte) 0x04, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code frstor  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code frstor    [ebx + 305419896]}
     */
    // Template#: 1756, Serial#: 2273
    public void frstor(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDD, (byte) 0x04, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsave  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fsave     ebx[esi * 4 + 18]}
     */
    // Template#: 1757, Serial#: 2264
    public void fsave(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDD, (byte) 0x06, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsave  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fsave     [ebx + 18]}
     */
    // Template#: 1758, Serial#: 2263
    public void fsave(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDD, (byte) 0x06, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsave  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fsave     ebx[esi * 4]}
     */
    // Template#: 1759, Serial#: 2248
    public void fsave(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDD, (byte) 0x06, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsave  }<i>destination</i>
     * Example disassembly syntax: {@code fsave     [ebx]}
     */
    // Template#: 1760, Serial#: 2247
    public void fsave(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDD, (byte) 0x06, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsave  }<i>m32</i>
     * Example disassembly syntax: {@code fsave     [L1: 0x12345678]}
     */
    // Template#: 1761, Serial#: 2250
    public void m_fsave(int m32) {
        assemble0299((byte) 0xDD, (byte) 0x06, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsave  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fsave     ebx[esi * 4 + 305419896]}
     */
    // Template#: 1762, Serial#: 2276
    public void fsave(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDD, (byte) 0x06, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsave  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fsave     L1: 0x12345678[esi * 4]}
     */
    // Template#: 1763, Serial#: 2249
    public void m_fsave(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDD, (byte) 0x06, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsave  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fsave     [ebx + 305419896]}
     */
    // Template#: 1764, Serial#: 2275
    public void fsave(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDD, (byte) 0x06, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fscale  }
     * Example disassembly syntax: {@code fscale  }
     */
    // Template#: 1765, Serial#: 5564
    public void fscale() {
        assemble0305((byte) 0xFD);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsin  }
     * Example disassembly syntax: {@code fsin    }
     */
    // Template#: 1766, Serial#: 5565
    public void fsin() {
        assemble0305((byte) 0xFE);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsincos  }
     * Example disassembly syntax: {@code fsincos }
     */
    // Template#: 1767, Serial#: 5562
    public void fsincos() {
        assemble0305((byte) 0xFB);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsqrt  }
     * Example disassembly syntax: {@code fsqrt   }
     */
    // Template#: 1768, Serial#: 5561
    public void fsqrt() {
        assemble0305((byte) 0xFA);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsts  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fsts      ebx[esi * 4 + 18]}
     */
    // Template#: 1769, Serial#: 2038
    public void fsts(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD9, (byte) 0x02, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstl      ebx[esi * 4 + 18]}
     */
    // Template#: 1770, Serial#: 2258
    public void fstl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDD, (byte) 0x02, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsts  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fsts      [ebx + 18]}
     */
    // Template#: 1771, Serial#: 2037
    public void fsts(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD9, (byte) 0x02, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fstl      [ebx + 18]}
     */
    // Template#: 1772, Serial#: 2257
    public void fstl(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDD, (byte) 0x02, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsts  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fsts      ebx[esi * 4]}
     */
    // Template#: 1773, Serial#: 2012
    public void fsts(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD9, (byte) 0x02, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstl      ebx[esi * 4]}
     */
    // Template#: 1774, Serial#: 2236
    public void fstl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDD, (byte) 0x02, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsts  }<i>destination</i>
     * Example disassembly syntax: {@code fsts      [ebx]}
     */
    // Template#: 1775, Serial#: 2011
    public void fsts(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD9, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstl  }<i>destination</i>
     * Example disassembly syntax: {@code fstl      [ebx]}
     */
    // Template#: 1776, Serial#: 2235
    public void fstl(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDD, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fst  }<i>st_i</i>
     * Example disassembly syntax: {@code fst       st(0)}
     */
    // Template#: 1777, Serial#: 5590
    public void fst(FPStackRegister st_i) {
        assemble0314((byte) 0xD0, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsts  }<i>m32</i>
     * Example disassembly syntax: {@code fsts      [L1: 0x12345678]}
     */
    // Template#: 1778, Serial#: 2014
    public void m_fsts(int m32) {
        assemble0299((byte) 0xD9, (byte) 0x02, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstl  }<i>m32</i>
     * Example disassembly syntax: {@code fstl      [L1: 0x12345678]}
     */
    // Template#: 1779, Serial#: 2238
    public void m_fstl(int m32) {
        assemble0299((byte) 0xDD, (byte) 0x02, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsts  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fsts      ebx[esi * 4 + 305419896]}
     */
    // Template#: 1780, Serial#: 2052
    public void fsts(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD9, (byte) 0x02, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstl      ebx[esi * 4 + 305419896]}
     */
    // Template#: 1781, Serial#: 2270
    public void fstl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDD, (byte) 0x02, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsts  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fsts      L1: 0x12345678[esi * 4]}
     */
    // Template#: 1782, Serial#: 2013
    public void m_fsts(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD9, (byte) 0x02, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstl      L1: 0x12345678[esi * 4]}
     */
    // Template#: 1783, Serial#: 2237
    public void m_fstl(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDD, (byte) 0x02, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsts  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fsts      [ebx + 305419896]}
     */
    // Template#: 1784, Serial#: 2051
    public void fsts(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD9, (byte) 0x02, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fstl      [ebx + 305419896]}
     */
    // Template#: 1785, Serial#: 2269
    public void fstl(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDD, (byte) 0x02, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstcw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstcw     ebx[esi * 4 + 18]}
     */
    // Template#: 1786, Serial#: 2048
    public void fstcw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD9, (byte) 0x07, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstcw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fstcw     [ebx + 18]}
     */
    // Template#: 1787, Serial#: 2047
    public void fstcw(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD9, (byte) 0x07, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstcw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstcw     ebx[esi * 4]}
     */
    // Template#: 1788, Serial#: 2032
    public void fstcw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD9, (byte) 0x07, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstcw  }<i>destination</i>
     * Example disassembly syntax: {@code fstcw     [ebx]}
     */
    // Template#: 1789, Serial#: 2031
    public void fstcw(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD9, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstcw  }<i>m32</i>
     * Example disassembly syntax: {@code fstcw     [L1: 0x12345678]}
     */
    // Template#: 1790, Serial#: 2034
    public void m_fstcw(int m32) {
        assemble0299((byte) 0xD9, (byte) 0x07, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstcw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstcw     ebx[esi * 4 + 305419896]}
     */
    // Template#: 1791, Serial#: 2062
    public void fstcw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD9, (byte) 0x07, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstcw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstcw     L1: 0x12345678[esi * 4]}
     */
    // Template#: 1792, Serial#: 2033
    public void m_fstcw(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD9, (byte) 0x07, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstcw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fstcw     [ebx + 305419896]}
     */
    // Template#: 1793, Serial#: 2061
    public void fstcw(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD9, (byte) 0x07, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstenv  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstenv    ebx[esi * 4 + 18]}
     */
    // Template#: 1794, Serial#: 2046
    public void fstenv(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD9, (byte) 0x06, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstenv  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fstenv    [ebx + 18]}
     */
    // Template#: 1795, Serial#: 2045
    public void fstenv(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD9, (byte) 0x06, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstenv  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstenv    ebx[esi * 4]}
     */
    // Template#: 1796, Serial#: 2028
    public void fstenv(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD9, (byte) 0x06, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstenv  }<i>destination</i>
     * Example disassembly syntax: {@code fstenv    [ebx]}
     */
    // Template#: 1797, Serial#: 2027
    public void fstenv(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD9, (byte) 0x06, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstenv  }<i>m32</i>
     * Example disassembly syntax: {@code fstenv    [L1: 0x12345678]}
     */
    // Template#: 1798, Serial#: 2030
    public void m_fstenv(int m32) {
        assemble0299((byte) 0xD9, (byte) 0x06, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstenv  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstenv    ebx[esi * 4 + 305419896]}
     */
    // Template#: 1799, Serial#: 2060
    public void fstenv(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD9, (byte) 0x06, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstenv  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstenv    L1: 0x12345678[esi * 4]}
     */
    // Template#: 1800, Serial#: 2029
    public void m_fstenv(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD9, (byte) 0x06, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstenv  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fstenv    [ebx + 305419896]}
     */
    // Template#: 1801, Serial#: 2059
    public void fstenv(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD9, (byte) 0x06, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstps  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstps     ebx[esi * 4 + 18]}
     */
    // Template#: 1802, Serial#: 2040
    public void fstps(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD9, (byte) 0x03, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstpt  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstpt     ebx[esi * 4 + 18]}
     */
    // Template#: 1803, Serial#: 2156
    public void fstpt(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDB, (byte) 0x07, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstpl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstpl     ebx[esi * 4 + 18]}
     */
    // Template#: 1804, Serial#: 2260
    public void fstpl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDD, (byte) 0x03, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstps  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fstps     [ebx + 18]}
     */
    // Template#: 1805, Serial#: 2039
    public void fstps(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD9, (byte) 0x03, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstpt  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fstpt     [ebx + 18]}
     */
    // Template#: 1806, Serial#: 2155
    public void fstpt(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDB, (byte) 0x07, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstpl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fstpl     [ebx + 18]}
     */
    // Template#: 1807, Serial#: 2259
    public void fstpl(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDD, (byte) 0x03, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstps  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstps     ebx[esi * 4]}
     */
    // Template#: 1808, Serial#: 2016
    public void fstps(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD9, (byte) 0x03, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstpt  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstpt     ebx[esi * 4]}
     */
    // Template#: 1809, Serial#: 2144
    public void fstpt(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDB, (byte) 0x07, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstpl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstpl     ebx[esi * 4]}
     */
    // Template#: 1810, Serial#: 2240
    public void fstpl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDD, (byte) 0x03, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstps  }<i>destination</i>
     * Example disassembly syntax: {@code fstps     [ebx]}
     */
    // Template#: 1811, Serial#: 2015
    public void fstps(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD9, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstpt  }<i>destination</i>
     * Example disassembly syntax: {@code fstpt     [ebx]}
     */
    // Template#: 1812, Serial#: 2143
    public void fstpt(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDB, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstpl  }<i>destination</i>
     * Example disassembly syntax: {@code fstpl     [ebx]}
     */
    // Template#: 1813, Serial#: 2239
    public void fstpl(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDD, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstp  }<i>st_i</i>
     * Example disassembly syntax: {@code fstp      st(0)}
     */
    // Template#: 1814, Serial#: 5591
    public void fstp(FPStackRegister st_i) {
        assemble0314((byte) 0xD8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstps  }<i>m32</i>
     * Example disassembly syntax: {@code fstps     [L1: 0x12345678]}
     */
    // Template#: 1815, Serial#: 2018
    public void m_fstps(int m32) {
        assemble0299((byte) 0xD9, (byte) 0x03, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstpt  }<i>m32</i>
     * Example disassembly syntax: {@code fstpt     [L1: 0x12345678]}
     */
    // Template#: 1816, Serial#: 2146
    public void m_fstpt(int m32) {
        assemble0299((byte) 0xDB, (byte) 0x07, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstpl  }<i>m32</i>
     * Example disassembly syntax: {@code fstpl     [L1: 0x12345678]}
     */
    // Template#: 1817, Serial#: 2242
    public void m_fstpl(int m32) {
        assemble0299((byte) 0xDD, (byte) 0x03, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstps  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstps     ebx[esi * 4 + 305419896]}
     */
    // Template#: 1818, Serial#: 2054
    public void fstps(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD9, (byte) 0x03, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstpt  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstpt     ebx[esi * 4 + 305419896]}
     */
    // Template#: 1819, Serial#: 2166
    public void fstpt(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDB, (byte) 0x07, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstpl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstpl     ebx[esi * 4 + 305419896]}
     */
    // Template#: 1820, Serial#: 2272
    public void fstpl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDD, (byte) 0x03, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstps  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstps     L1: 0x12345678[esi * 4]}
     */
    // Template#: 1821, Serial#: 2017
    public void m_fstps(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD9, (byte) 0x03, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstpt  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstpt     L1: 0x12345678[esi * 4]}
     */
    // Template#: 1822, Serial#: 2145
    public void m_fstpt(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDB, (byte) 0x07, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstpl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstpl     L1: 0x12345678[esi * 4]}
     */
    // Template#: 1823, Serial#: 2241
    public void m_fstpl(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDD, (byte) 0x03, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstps  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fstps     [ebx + 305419896]}
     */
    // Template#: 1824, Serial#: 2053
    public void fstps(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD9, (byte) 0x03, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstpt  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fstpt     [ebx + 305419896]}
     */
    // Template#: 1825, Serial#: 2165
    public void fstpt(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDB, (byte) 0x07, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstpl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fstpl     [ebx + 305419896]}
     */
    // Template#: 1826, Serial#: 2271
    public void fstpl(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDD, (byte) 0x03, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstp1  }<i>st_i</i>
     * Example disassembly syntax: {@code fstp1     st(0)}
     */
    // Template#: 1827, Serial#: 5539
    public void fstp1(FPStackRegister st_i) {
        assemble0315((byte) 0xD8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstp8  }<i>st_i</i>
     * Example disassembly syntax: {@code fstp8     st(0)}
     */
    // Template#: 1828, Serial#: 5604
    public void fstp8(FPStackRegister st_i) {
        assemble0312((byte) 0xD0, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstp9  }<i>st_i</i>
     * Example disassembly syntax: {@code fstp9     st(0)}
     */
    // Template#: 1829, Serial#: 5605
    public void fstp9(FPStackRegister st_i) {
        assemble0312((byte) 0xD8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstsw  }
     * Example disassembly syntax: {@code fstsw     ax}
     */
    // Template#: 1830, Serial#: 5606
    public void fstsw_AX() {
        assemble0316((byte) 0xE0);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstsw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstsw     ebx[esi * 4 + 18]}
     */
    // Template#: 1831, Serial#: 2266
    public void fstsw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDD, (byte) 0x07, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstsw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fstsw     [ebx + 18]}
     */
    // Template#: 1832, Serial#: 2265
    public void fstsw(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDD, (byte) 0x07, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstsw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstsw     ebx[esi * 4]}
     */
    // Template#: 1833, Serial#: 2252
    public void fstsw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDD, (byte) 0x07, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstsw  }<i>destination</i>
     * Example disassembly syntax: {@code fstsw     [ebx]}
     */
    // Template#: 1834, Serial#: 2251
    public void fstsw(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDD, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstsw  }<i>m32</i>
     * Example disassembly syntax: {@code fstsw     [L1: 0x12345678]}
     */
    // Template#: 1835, Serial#: 2254
    public void m_fstsw(int m32) {
        assemble0299((byte) 0xDD, (byte) 0x07, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstsw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstsw     ebx[esi * 4 + 305419896]}
     */
    // Template#: 1836, Serial#: 2278
    public void fstsw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDD, (byte) 0x07, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstsw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fstsw     L1: 0x12345678[esi * 4]}
     */
    // Template#: 1837, Serial#: 2253
    public void m_fstsw(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDD, (byte) 0x07, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fstsw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fstsw     [ebx + 305419896]}
     */
    // Template#: 1838, Serial#: 2277
    public void fstsw(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDD, (byte) 0x07, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubs  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fsubs     ebx[esi * 4 + 18]}
     */
    // Template#: 1839, Serial#: 1984
    public void fsubs(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD8, (byte) 0x04, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fsubl     ebx[esi * 4 + 18]}
     */
    // Template#: 1840, Serial#: 2208
    public void fsubl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDC, (byte) 0x04, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubs  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fsubs     [ebx + 18]}
     */
    // Template#: 1841, Serial#: 1983
    public void fsubs(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD8, (byte) 0x04, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fsubl     [ebx + 18]}
     */
    // Template#: 1842, Serial#: 2207
    public void fsubl(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDC, (byte) 0x04, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubs  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fsubs     ebx[esi * 4]}
     */
    // Template#: 1843, Serial#: 1960
    public void fsubs(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD8, (byte) 0x04, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fsubl     ebx[esi * 4]}
     */
    // Template#: 1844, Serial#: 2184
    public void fsubl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDC, (byte) 0x04, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubs  }<i>destination</i>
     * Example disassembly syntax: {@code fsubs     [ebx]}
     */
    // Template#: 1845, Serial#: 1959
    public void fsubs(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD8, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubl  }<i>destination</i>
     * Example disassembly syntax: {@code fsubl     [ebx]}
     */
    // Template#: 1846, Serial#: 2183
    public void fsubl(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDC, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsub  }<i>st_i</i>
     * Example disassembly syntax: {@code fsub      st, st(0)}
     */
    // Template#: 1847, Serial#: 5532
    public void fsub_ST(FPStackRegister st_i) {
        assemble0306((byte) 0xE0, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsub  }<i>st_i</i>
     * Example disassembly syntax: {@code fsub      st(0), st}
     */
    // Template#: 1848, Serial#: 5584
    public void fsub___ST(FPStackRegister st_i) {
        assemble0307((byte) 0xE0, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubs  }<i>m32</i>
     * Example disassembly syntax: {@code fsubs     [L1: 0x12345678]}
     */
    // Template#: 1849, Serial#: 1962
    public void m_fsubs(int m32) {
        assemble0299((byte) 0xD8, (byte) 0x04, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubl  }<i>m32</i>
     * Example disassembly syntax: {@code fsubl     [L1: 0x12345678]}
     */
    // Template#: 1850, Serial#: 2186
    public void m_fsubl(int m32) {
        assemble0299((byte) 0xDC, (byte) 0x04, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubs  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fsubs     ebx[esi * 4 + 305419896]}
     */
    // Template#: 1851, Serial#: 2000
    public void fsubs(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD8, (byte) 0x04, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fsubl     ebx[esi * 4 + 305419896]}
     */
    // Template#: 1852, Serial#: 2224
    public void fsubl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDC, (byte) 0x04, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubs  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fsubs     L1: 0x12345678[esi * 4]}
     */
    // Template#: 1853, Serial#: 1961
    public void m_fsubs(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD8, (byte) 0x04, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fsubl     L1: 0x12345678[esi * 4]}
     */
    // Template#: 1854, Serial#: 2185
    public void m_fsubl(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDC, (byte) 0x04, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubs  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fsubs     [ebx + 305419896]}
     */
    // Template#: 1855, Serial#: 1999
    public void fsubs(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD8, (byte) 0x04, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fsubl     [ebx + 305419896]}
     */
    // Template#: 1856, Serial#: 2223
    public void fsubl(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDC, (byte) 0x04, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubrp  }<i>st_i</i>
     * Example disassembly syntax: {@code fsubrp    st(0), st}
     */
    // Template#: 1857, Serial#: 5599
    public void fsubp___ST(FPStackRegister st_i) {
        assemble0308((byte) 0xE8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubrs  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fsubrs    ebx[esi * 4 + 18]}
     */
    // Template#: 1858, Serial#: 1986
    public void fsubrs(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD8, (byte) 0x05, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubrl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fsubrl    ebx[esi * 4 + 18]}
     */
    // Template#: 1859, Serial#: 2210
    public void fsubrl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xDC, (byte) 0x05, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubrs  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fsubrs    [ebx + 18]}
     */
    // Template#: 1860, Serial#: 1985
    public void fsubrs(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD8, (byte) 0x05, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubrl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code fsubrl    [ebx + 18]}
     */
    // Template#: 1861, Serial#: 2209
    public void fsubrl(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xDC, (byte) 0x05, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubrs  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fsubrs    ebx[esi * 4]}
     */
    // Template#: 1862, Serial#: 1964
    public void fsubrs(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD8, (byte) 0x05, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubrl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fsubrl    ebx[esi * 4]}
     */
    // Template#: 1863, Serial#: 2188
    public void fsubrl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xDC, (byte) 0x05, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubrs  }<i>destination</i>
     * Example disassembly syntax: {@code fsubrs    [ebx]}
     */
    // Template#: 1864, Serial#: 1963
    public void fsubrs(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD8, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubrl  }<i>destination</i>
     * Example disassembly syntax: {@code fsubrl    [ebx]}
     */
    // Template#: 1865, Serial#: 2187
    public void fsubrl(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xDC, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubr  }<i>st_i</i>
     * Example disassembly syntax: {@code fsubr     st, st(0)}
     */
    // Template#: 1866, Serial#: 5533
    public void fsubr_ST(FPStackRegister st_i) {
        assemble0306((byte) 0xE8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubr  }<i>st_i</i>
     * Example disassembly syntax: {@code fsubr     st(0), st}
     */
    // Template#: 1867, Serial#: 5585
    public void fsubr___ST(FPStackRegister st_i) {
        assemble0307((byte) 0xE8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubrs  }<i>m32</i>
     * Example disassembly syntax: {@code fsubrs    [L1: 0x12345678]}
     */
    // Template#: 1868, Serial#: 1966
    public void m_fsubrs(int m32) {
        assemble0299((byte) 0xD8, (byte) 0x05, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubrl  }<i>m32</i>
     * Example disassembly syntax: {@code fsubrl    [L1: 0x12345678]}
     */
    // Template#: 1869, Serial#: 2190
    public void m_fsubrl(int m32) {
        assemble0299((byte) 0xDC, (byte) 0x05, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubrs  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fsubrs    ebx[esi * 4 + 305419896]}
     */
    // Template#: 1870, Serial#: 2002
    public void fsubrs(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD8, (byte) 0x05, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubrl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fsubrl    ebx[esi * 4 + 305419896]}
     */
    // Template#: 1871, Serial#: 2226
    public void fsubrl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xDC, (byte) 0x05, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubrs  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fsubrs    L1: 0x12345678[esi * 4]}
     */
    // Template#: 1872, Serial#: 1965
    public void m_fsubrs(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD8, (byte) 0x05, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubrl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code fsubrl    L1: 0x12345678[esi * 4]}
     */
    // Template#: 1873, Serial#: 2189
    public void m_fsubrl(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xDC, (byte) 0x05, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubrs  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fsubrs    [ebx + 305419896]}
     */
    // Template#: 1874, Serial#: 2001
    public void fsubrs(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD8, (byte) 0x05, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubrl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code fsubrl    [ebx + 305419896]}
     */
    // Template#: 1875, Serial#: 2225
    public void fsubrl(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xDC, (byte) 0x05, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code fsubp  }<i>st_i</i>
     * Example disassembly syntax: {@code fsubp     st(0), st}
     */
    // Template#: 1876, Serial#: 5598
    public void fsubrp___ST(FPStackRegister st_i) {
        assemble0308((byte) 0xE0, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code ftst  }
     * Example disassembly syntax: {@code ftst    }
     */
    // Template#: 1877, Serial#: 5542
    public void ftst() {
        assemble0305((byte) 0xE4);
    }

    /**
     * Pseudo-external assembler syntax: {@code fucom  }<i>st_i</i>
     * Example disassembly syntax: {@code fucom     st(0)}
     */
    // Template#: 1878, Serial#: 5592
    public void fucom(FPStackRegister st_i) {
        assemble0314((byte) 0xE0, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fucomi  }<i>st_i</i>
     * Example disassembly syntax: {@code fucomi    st, st(0)}
     */
    // Template#: 1879, Serial#: 5578
    public void fucomi_ST(FPStackRegister st_i) {
        assemble0311((byte) 0xE8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fucomip  }<i>st_i</i>
     * Example disassembly syntax: {@code fucomip   st, st(0)}
     */
    // Template#: 1880, Serial#: 5607
    public void fucomip_ST(FPStackRegister st_i) {
        assemble0312((byte) 0xE8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fucomp  }<i>st_i</i>
     * Example disassembly syntax: {@code fucomp    st(0)}
     */
    // Template#: 1881, Serial#: 5593
    public void fucomp(FPStackRegister st_i) {
        assemble0314((byte) 0xE8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fucompp  }
     * Example disassembly syntax: {@code fucompp }
     */
    // Template#: 1882, Serial#: 5571
    public void fucompp() {
        assemble0317((byte) 0xE9);
    }

    /**
     * Pseudo-external assembler syntax: {@code fwait  }
     * Example disassembly syntax: {@code fwait   }
     */
    // Template#: 1883, Serial#: 1914
    public void fwait() {
        assemble0001((byte) 0x9B);
    }

    /**
     * Pseudo-external assembler syntax: {@code fxam  }
     * Example disassembly syntax: {@code fxam    }
     */
    // Template#: 1884, Serial#: 5543
    public void fxam() {
        assemble0305((byte) 0xE5);
    }

    /**
     * Pseudo-external assembler syntax: {@code fxch  }<i>st_i</i>
     * Example disassembly syntax: {@code fxch      st(0)}
     */
    // Template#: 1885, Serial#: 5537
    public void fxch(FPStackRegister st_i) {
        assemble0315((byte) 0xC8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fxch4  }<i>st_i</i>
     * Example disassembly syntax: {@code fxch4     st(0)}
     */
    // Template#: 1886, Serial#: 5589
    public void fxch4(FPStackRegister st_i) {
        assemble0314((byte) 0xC8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fxch7  }<i>st_i</i>
     * Example disassembly syntax: {@code fxch7     st(0)}
     */
    // Template#: 1887, Serial#: 5603
    public void fxch7(FPStackRegister st_i) {
        assemble0312((byte) 0xC8, st_i);
    }

    /**
     * Pseudo-external assembler syntax: {@code fxtract  }
     * Example disassembly syntax: {@code fxtract }
     */
    // Template#: 1888, Serial#: 5555
    public void fxtract() {
        assemble0305((byte) 0xF4);
    }

    /**
     * Pseudo-external assembler syntax: {@code fyl2x  }
     * Example disassembly syntax: {@code fyl2x   }
     */
    // Template#: 1889, Serial#: 5552
    public void fyl2x() {
        assemble0305((byte) 0xF1);
    }

    /**
     * Pseudo-external assembler syntax: {@code fyl2xp1  }
     * Example disassembly syntax: {@code fyl2xp1 }
     */
    // Template#: 1890, Serial#: 5560
    public void fyl2xp1() {
        assemble0305((byte) 0xF9);
    }

    /**
     * Pseudo-external assembler syntax: {@code haddpd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code haddpd    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1891, Serial#: 4806
    public void haddpd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x7C, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code haddpd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code haddpd    xmm0, [ebx + 18]}
     */
    // Template#: 1892, Serial#: 4805
    public void haddpd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x7C, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code haddpd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code haddpd    xmm0, ebx[esi * 4]}
     */
    // Template#: 1893, Serial#: 4802
    public void haddpd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x7C, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code haddpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code haddpd    xmm0, [ebx]}
     */
    // Template#: 1894, Serial#: 4801
    public void haddpd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x7C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code haddpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code haddpd    xmm0, xmm0}
     */
    // Template#: 1895, Serial#: 4809
    public void haddpd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x7C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code haddpd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code haddpd    xmm0, [L1: 0x12345678]}
     */
    // Template#: 1896, Serial#: 4804
    public void m_haddpd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x7C, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code haddpd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code haddpd    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1897, Serial#: 4808
    public void haddpd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x7C, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code haddpd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code haddpd    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1898, Serial#: 4803
    public void m_haddpd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x7C, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code haddpd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code haddpd    xmm0, [ebx + 305419896]}
     */
    // Template#: 1899, Serial#: 4807
    public void haddpd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x7C, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code haddps  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code haddps    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1900, Serial#: 4841
    public void haddps(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0111((byte) 0x7C, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code haddps  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code haddps    xmm0, [ebx + 18]}
     */
    // Template#: 1901, Serial#: 4840
    public void haddps(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0112((byte) 0x7C, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code haddps  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code haddps    xmm0, ebx[esi * 4]}
     */
    // Template#: 1902, Serial#: 4837
    public void haddps(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0113((byte) 0x7C, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code haddps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code haddps    xmm0, [ebx]}
     */
    // Template#: 1903, Serial#: 4836
    public void haddps(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0114((byte) 0x7C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code haddps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code haddps    xmm0, xmm0}
     */
    // Template#: 1904, Serial#: 4844
    public void haddps(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0115((byte) 0x7C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code haddps  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code haddps    xmm0, [L1: 0x12345678]}
     */
    // Template#: 1905, Serial#: 4839
    public void m_haddps(IA32XMMRegister destination, int m32) {
        assemble0116((byte) 0x7C, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code haddps  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code haddps    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1906, Serial#: 4843
    public void haddps(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0117((byte) 0x7C, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code haddps  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code haddps    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1907, Serial#: 4838
    public void m_haddps(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0118((byte) 0x7C, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code haddps  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code haddps    xmm0, [ebx + 305419896]}
     */
    // Template#: 1908, Serial#: 4842
    public void haddps(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0119((byte) 0x7C, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code hlt  }
     * Example disassembly syntax: {@code hlt     }
     */
    // Template#: 1909, Serial#: 1345
    public void hlt() {
        assemble0001((byte) 0xF4);
    }

    /**
     * Pseudo-external assembler syntax: {@code hsubpd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code hsubpd    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1910, Serial#: 4815
    public void hsubpd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x7D, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code hsubpd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code hsubpd    xmm0, [ebx + 18]}
     */
    // Template#: 1911, Serial#: 4814
    public void hsubpd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x7D, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code hsubpd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code hsubpd    xmm0, ebx[esi * 4]}
     */
    // Template#: 1912, Serial#: 4811
    public void hsubpd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x7D, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code hsubpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code hsubpd    xmm0, [ebx]}
     */
    // Template#: 1913, Serial#: 4810
    public void hsubpd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x7D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code hsubpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code hsubpd    xmm0, xmm0}
     */
    // Template#: 1914, Serial#: 4818
    public void hsubpd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x7D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code hsubpd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code hsubpd    xmm0, [L1: 0x12345678]}
     */
    // Template#: 1915, Serial#: 4813
    public void m_hsubpd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x7D, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code hsubpd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code hsubpd    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1916, Serial#: 4817
    public void hsubpd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x7D, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code hsubpd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code hsubpd    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1917, Serial#: 4812
    public void m_hsubpd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x7D, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code hsubpd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code hsubpd    xmm0, [ebx + 305419896]}
     */
    // Template#: 1918, Serial#: 4816
    public void hsubpd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x7D, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code hsubps  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code hsubps    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 1919, Serial#: 4850
    public void hsubps(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0111((byte) 0x7D, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code hsubps  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code hsubps    xmm0, [ebx + 18]}
     */
    // Template#: 1920, Serial#: 4849
    public void hsubps(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0112((byte) 0x7D, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code hsubps  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code hsubps    xmm0, ebx[esi * 4]}
     */
    // Template#: 1921, Serial#: 4846
    public void hsubps(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0113((byte) 0x7D, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code hsubps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code hsubps    xmm0, [ebx]}
     */
    // Template#: 1922, Serial#: 4845
    public void hsubps(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0114((byte) 0x7D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code hsubps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code hsubps    xmm0, xmm0}
     */
    // Template#: 1923, Serial#: 4853
    public void hsubps(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0115((byte) 0x7D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code hsubps  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code hsubps    xmm0, [L1: 0x12345678]}
     */
    // Template#: 1924, Serial#: 4848
    public void m_hsubps(IA32XMMRegister destination, int m32) {
        assemble0116((byte) 0x7D, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code hsubps  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code hsubps    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1925, Serial#: 4852
    public void hsubps(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0117((byte) 0x7D, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code hsubps  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code hsubps    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1926, Serial#: 4847
    public void m_hsubps(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0118((byte) 0x7D, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code hsubps  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code hsubps    xmm0, [ebx + 305419896]}
     */
    // Template#: 1927, Serial#: 4851
    public void hsubps(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0119((byte) 0x7D, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code idivb     ebx[esi * 4 + 18], al}
     */
    // Template#: 1928, Serial#: 1388
    public void idivb___AL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xF6, (byte) 0x07, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code idivl     ebx[esi * 4 + 18], eax}
     */
    // Template#: 1929, Serial#: 1451
    public void idivl___EAX(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xF7, (byte) 0x07, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code idivw     ebx[esi * 4 + 18], ax}
     */
    // Template#: 1930, Serial#: 1514
    public void idivw___AX(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0292((byte) 0xF7, (byte) 0x07, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivb  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code idivb     [ebx + 18], al}
     */
    // Template#: 1931, Serial#: 1387
    public void idivb___AL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xF6, (byte) 0x07, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code idivl     [ebx + 18], eax}
     */
    // Template#: 1932, Serial#: 1450
    public void idivl___EAX(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xF7, (byte) 0x07, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code idivw     [ebx + 18], ax}
     */
    // Template#: 1933, Serial#: 1513
    public void idivw___AX(byte disp8, IA32IndirectRegister32 destination) {
        assemble0293((byte) 0xF7, (byte) 0x07, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivb  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code idivb     ebx[esi * 4], al}
     */
    // Template#: 1934, Serial#: 1372
    public void idivb___AL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xF6, (byte) 0x07, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code idivl     ebx[esi * 4], eax}
     */
    // Template#: 1935, Serial#: 1435
    public void idivl___EAX(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xF7, (byte) 0x07, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code idivw     ebx[esi * 4], ax}
     */
    // Template#: 1936, Serial#: 1498
    public void idivw___AX(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0294((byte) 0xF7, (byte) 0x07, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivw  }<i>destination</i>
     * Example disassembly syntax: {@code idivw     ax, ax}
     */
    // Template#: 1937, Serial#: 1535
    public void idivw___AX(IA32GeneralRegister16 destination) {
        assemble0188((byte) 0xF7, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivl  }<i>destination</i>
     * Example disassembly syntax: {@code idivl     eax, eax}
     */
    // Template#: 1938, Serial#: 1472
    public void idivl___EAX(IA32GeneralRegister32 destination) {
        assemble0189((byte) 0xF7, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivb  }<i>destination</i>
     * Example disassembly syntax: {@code idivb     al, al}
     */
    // Template#: 1939, Serial#: 1409
    public void idivb___AL(IA32GeneralRegister8 destination) {
        assemble0297((byte) 0xF6, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivb  }<i>destination</i>
     * Example disassembly syntax: {@code idivb     [ebx], al}
     */
    // Template#: 1940, Serial#: 1371
    public void idivb___AL(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xF6, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivl  }<i>destination</i>
     * Example disassembly syntax: {@code idivl     [ebx], eax}
     */
    // Template#: 1941, Serial#: 1434
    public void idivl___EAX(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xF7, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivw  }<i>destination</i>
     * Example disassembly syntax: {@code idivw     [ebx], ax}
     */
    // Template#: 1942, Serial#: 1497
    public void idivw___AX(IA32IndirectRegister32 destination) {
        assemble0298((byte) 0xF7, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivb  }<i>m32</i>
     * Example disassembly syntax: {@code idivb     [L1: 0x12345678], al}
     */
    // Template#: 1943, Serial#: 1374
    public void m_idivb___AL(int m32) {
        assemble0299((byte) 0xF6, (byte) 0x07, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivl  }<i>m32</i>
     * Example disassembly syntax: {@code idivl     [L1: 0x12345678], eax}
     */
    // Template#: 1944, Serial#: 1437
    public void m_idivl___EAX(int m32) {
        assemble0299((byte) 0xF7, (byte) 0x07, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivw  }<i>m32</i>
     * Example disassembly syntax: {@code idivw     [L1: 0x12345678], ax}
     */
    // Template#: 1945, Serial#: 1500
    public void m_idivw___AX(int m32) {
        assemble0300((byte) 0xF7, (byte) 0x07, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code idivb     ebx[esi * 4 + 305419896], al}
     */
    // Template#: 1946, Serial#: 1402
    public void idivb___AL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xF6, (byte) 0x07, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code idivl     ebx[esi * 4 + 305419896], eax}
     */
    // Template#: 1947, Serial#: 1465
    public void idivl___EAX(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xF7, (byte) 0x07, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code idivw     ebx[esi * 4 + 305419896], ax}
     */
    // Template#: 1948, Serial#: 1528
    public void idivw___AX(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0301((byte) 0xF7, (byte) 0x07, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivb  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code idivb     L1: 0x12345678[esi * 4], al}
     */
    // Template#: 1949, Serial#: 1373
    public void m_idivb___AL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xF6, (byte) 0x07, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code idivl     L1: 0x12345678[esi * 4], eax}
     */
    // Template#: 1950, Serial#: 1436
    public void m_idivl___EAX(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xF7, (byte) 0x07, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code idivw     L1: 0x12345678[esi * 4], ax}
     */
    // Template#: 1951, Serial#: 1499
    public void m_idivw___AX(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0302((byte) 0xF7, (byte) 0x07, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivb  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code idivb     [ebx + 305419896], al}
     */
    // Template#: 1952, Serial#: 1401
    public void idivb___AL(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xF6, (byte) 0x07, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code idivl     [ebx + 305419896], eax}
     */
    // Template#: 1953, Serial#: 1464
    public void idivl___EAX(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xF7, (byte) 0x07, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code idivw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code idivw     [ebx + 305419896], ax}
     */
    // Template#: 1954, Serial#: 1527
    public void idivw___AX(int disp32, IA32IndirectRegister32 destination) {
        assemble0303((byte) 0xF7, (byte) 0x07, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code imulb     ebx[esi * 4 + 18], al}
     */
    // Template#: 1955, Serial#: 1384
    public void imulb___AL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xF6, (byte) 0x05, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code imull  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code imull     ebx[esi * 4 + 18], eax}
     */
    // Template#: 1956, Serial#: 1447
    public void imull___EAX(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xF7, (byte) 0x05, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code imulw     ebx[esi * 4 + 18], ax}
     */
    // Template#: 1957, Serial#: 1510
    public void imulw___AX(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0292((byte) 0xF7, (byte) 0x05, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulb  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code imulb     [ebx + 18], al}
     */
    // Template#: 1958, Serial#: 1383
    public void imulb___AL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xF6, (byte) 0x05, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code imull  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code imull     [ebx + 18], eax}
     */
    // Template#: 1959, Serial#: 1446
    public void imull___EAX(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xF7, (byte) 0x05, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code imulw     [ebx + 18], ax}
     */
    // Template#: 1960, Serial#: 1509
    public void imulw___AX(byte disp8, IA32IndirectRegister32 destination) {
        assemble0293((byte) 0xF7, (byte) 0x05, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulb  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code imulb     ebx[esi * 4], al}
     */
    // Template#: 1961, Serial#: 1364
    public void imulb___AL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xF6, (byte) 0x05, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code imull  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code imull     ebx[esi * 4], eax}
     */
    // Template#: 1962, Serial#: 1427
    public void imull___EAX(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xF7, (byte) 0x05, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code imulw     ebx[esi * 4], ax}
     */
    // Template#: 1963, Serial#: 1490
    public void imulw___AX(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0294((byte) 0xF7, (byte) 0x05, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulw  }<i>destination</i>
     * Example disassembly syntax: {@code imulw     ax, ax}
     */
    // Template#: 1964, Serial#: 1533
    public void imulw___AX(IA32GeneralRegister16 destination) {
        assemble0188((byte) 0xF7, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code imul      ax, ebx[esi * 4 + 18]}
     */
    // Template#: 1965, Serial#: 5024
    public void imul(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0xAF, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code imul      ax, ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 1966, Serial#: 1798
    public void imul(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0318((byte) 0x6B, destination, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code imulw     ax, ebx[esi * 4 + 18], 0x1234}
     */
    // Template#: 1967, Serial#: 1779
    public void imul(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0319((byte) 0x69, destination, disp8, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code imul      ax, [ebx + 18]}
     */
    // Template#: 1968, Serial#: 5023
    public void imul(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0xAF, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>disp8</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code imul      ax, [ebx + 18], 0x12}
     */
    // Template#: 1969, Serial#: 1797
    public void imul(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source, byte imm8) {
        assemble0320((byte) 0x6B, destination, disp8, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulw  }<i>destination</i>, <i>disp8</i>, <i>source</i>, <i>imm16</i>
     * Example disassembly syntax: {@code imulw     ax, [ebx + 18], 0x1234}
     */
    // Template#: 1970, Serial#: 1778
    public void imul(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source, short imm16) {
        assemble0321((byte) 0x69, destination, disp8, source, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code imul      ax, ebx[esi * 4]}
     */
    // Template#: 1971, Serial#: 5020
    public void imul(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0xAF, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code imul      ax, ebx[esi * 4], 0x12}
     */
    // Template#: 1972, Serial#: 1794
    public void imul(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0322((byte) 0x6B, destination, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code imulw     ax, ebx[esi * 4], 0x1234}
     */
    // Template#: 1973, Serial#: 1775
    public void imul(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0323((byte) 0x69, destination, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code imul      ax, ax}
     */
    // Template#: 1974, Serial#: 5027
    public void imul(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0141((byte) 0xAF, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code imul      ax, ax, 0x12}
     */
    // Template#: 1975, Serial#: 1801
    public void imul(IA32GeneralRegister16 destination, IA32GeneralRegister16 source, byte imm8) {
        assemble0324((byte) 0x6B, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulw  }<i>destination</i>, <i>source</i>, <i>imm16</i>
     * Example disassembly syntax: {@code imulw     ax, ax, 0x1234}
     */
    // Template#: 1976, Serial#: 1782
    public void imul(IA32GeneralRegister16 destination, IA32GeneralRegister16 source, short imm16) {
        assemble0325((byte) 0x69, destination, source, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code imul      ax, [ebx]}
     */
    // Template#: 1977, Serial#: 5019
    public void imul(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0xAF, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code imul      ax, [ebx], 0x12}
     */
    // Template#: 1978, Serial#: 1793
    public void imul(IA32GeneralRegister16 destination, IA32IndirectRegister32 source, byte imm8) {
        assemble0326((byte) 0x6B, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulw  }<i>destination</i>, <i>source</i>, <i>imm16</i>
     * Example disassembly syntax: {@code imulw     ax, [ebx], 0x1234}
     */
    // Template#: 1979, Serial#: 1774
    public void imul(IA32GeneralRegister16 destination, IA32IndirectRegister32 source, short imm16) {
        assemble0327((byte) 0x69, destination, source, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code imul      ax, [L1: 0x12345678]}
     */
    // Template#: 1980, Serial#: 5022
    public void m_imul(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0xAF, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code imul      ax, [L1: 0x12345678], 0x12}
     */
    // Template#: 1981, Serial#: 1796
    public void m_imul(IA32GeneralRegister16 destination, int m32, byte imm8) {
        assemble0328((byte) 0x6B, destination, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code imul      ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 1982, Serial#: 5026
    public void imul(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0xAF, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code imul      ax, ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 1983, Serial#: 1800
    public void imul(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0329((byte) 0x6B, destination, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code imulw     ax, ebx[esi * 4 + 305419896], 0x1234}
     */
    // Template#: 1984, Serial#: 1781
    public void imul(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0330((byte) 0x69, destination, disp32, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code imul      ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 1985, Serial#: 5021
    public void m_imul(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0xAF, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code imul      ax, L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 1986, Serial#: 1795
    public void m_imul(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0331((byte) 0x6B, destination, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code imulw     ax, L1: 0x12345678[esi * 4], 0x1234}
     */
    // Template#: 1987, Serial#: 1776
    public void m_imul(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0332((byte) 0x69, destination, m32, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code imul      ax, [ebx + 305419896]}
     */
    // Template#: 1988, Serial#: 5025
    public void imul(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0xAF, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>disp32</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code imul      ax, [ebx + 305419896], 0x12}
     */
    // Template#: 1989, Serial#: 1799
    public void imul(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source, byte imm8) {
        assemble0333((byte) 0x6B, destination, disp32, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulw  }<i>destination</i>, <i>disp32</i>, <i>source</i>, <i>imm16</i>
     * Example disassembly syntax: {@code imulw     ax, [ebx + 305419896], 0x1234}
     */
    // Template#: 1990, Serial#: 1780
    public void imul(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source, short imm16) {
        assemble0334((byte) 0x69, destination, disp32, source, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulw  }<i>destination</i>, <i>m32</i>, <i>imm16</i>
     * Example disassembly syntax: {@code imulw     ax, [L1: 0x12345678], 0x1234}
     */
    // Template#: 1991, Serial#: 1777
    public void m_imul(IA32GeneralRegister16 destination, int m32, short imm16) {
        assemble0335((byte) 0x69, destination, m32, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code imull  }<i>destination</i>
     * Example disassembly syntax: {@code imull     eax, eax}
     */
    // Template#: 1992, Serial#: 1470
    public void imull___EAX(IA32GeneralRegister32 destination) {
        assemble0189((byte) 0xF7, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code imul      eax, ebx[esi * 4 + 18]}
     */
    // Template#: 1993, Serial#: 5015
    public void imul(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0xAF, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code imul      eax, ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 1994, Serial#: 1789
    public void imul(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0336((byte) 0x6B, destination, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code imull  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code imull     eax, ebx[esi * 4 + 18], 0x12345678}
     */
    // Template#: 1995, Serial#: 1770
    public void imul(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0337((byte) 0x69, destination, disp8, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code imul      eax, [ebx + 18]}
     */
    // Template#: 1996, Serial#: 5014
    public void imul(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0xAF, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>disp8</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code imul      eax, [ebx + 18], 0x12}
     */
    // Template#: 1997, Serial#: 1788
    public void imul(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source, byte imm8) {
        assemble0338((byte) 0x6B, destination, disp8, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code imull  }<i>destination</i>, <i>disp8</i>, <i>source</i>, <i>imm32</i>
     * Example disassembly syntax: {@code imull     eax, [ebx + 18], 0x12345678}
     */
    // Template#: 1998, Serial#: 1769
    public void imul(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source, int imm32) {
        assemble0339((byte) 0x69, destination, disp8, source, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code imul      eax, ebx[esi * 4]}
     */
    // Template#: 1999, Serial#: 5011
    public void imul(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0xAF, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code imul      eax, ebx[esi * 4], 0x12}
     */
    // Template#: 2000, Serial#: 1785
    public void imul(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0340((byte) 0x6B, destination, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code imull  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code imull     eax, ebx[esi * 4], 0x12345678}
     */
    // Template#: 2001, Serial#: 1766
    public void imul(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0341((byte) 0x69, destination, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code imul      eax, eax}
     */
    // Template#: 2002, Serial#: 5018
    public void imul(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0150((byte) 0xAF, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code imul      eax, eax, 0x12}
     */
    // Template#: 2003, Serial#: 1792
    public void imul(IA32GeneralRegister32 destination, IA32GeneralRegister32 source, byte imm8) {
        assemble0342((byte) 0x6B, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code imull  }<i>destination</i>, <i>source</i>, <i>imm32</i>
     * Example disassembly syntax: {@code imull     eax, eax, 0x12345678}
     */
    // Template#: 2004, Serial#: 1773
    public void imul(IA32GeneralRegister32 destination, IA32GeneralRegister32 source, int imm32) {
        assemble0343((byte) 0x69, destination, source, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code imul      eax, [ebx]}
     */
    // Template#: 2005, Serial#: 5010
    public void imul(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0xAF, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code imul      eax, [ebx], 0x12}
     */
    // Template#: 2006, Serial#: 1784
    public void imul(IA32GeneralRegister32 destination, IA32IndirectRegister32 source, byte imm8) {
        assemble0344((byte) 0x6B, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code imull  }<i>destination</i>, <i>source</i>, <i>imm32</i>
     * Example disassembly syntax: {@code imull     eax, [ebx], 0x12345678}
     */
    // Template#: 2007, Serial#: 1765
    public void imul(IA32GeneralRegister32 destination, IA32IndirectRegister32 source, int imm32) {
        assemble0345((byte) 0x69, destination, source, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code imul      eax, [L1: 0x12345678]}
     */
    // Template#: 2008, Serial#: 5013
    public void m_imul(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0xAF, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code imul      eax, [L1: 0x12345678], 0x12}
     */
    // Template#: 2009, Serial#: 1787
    public void m_imul(IA32GeneralRegister32 destination, int m32, byte imm8) {
        assemble0346((byte) 0x6B, destination, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code imul      eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2010, Serial#: 5017
    public void imul(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0xAF, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code imul      eax, ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 2011, Serial#: 1791
    public void imul(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0347((byte) 0x6B, destination, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code imull  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code imull     eax, ebx[esi * 4 + 305419896], 0x12345678}
     */
    // Template#: 2012, Serial#: 1772
    public void imul(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0348((byte) 0x69, destination, disp32, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code imul      eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2013, Serial#: 5012
    public void m_imul(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0xAF, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code imul      eax, L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 2014, Serial#: 1786
    public void m_imul(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0349((byte) 0x6B, destination, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code imull  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code imull     eax, L1: 0x12345678[esi * 4], 0x12345678}
     */
    // Template#: 2015, Serial#: 1767
    public void m_imul(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0350((byte) 0x69, destination, m32, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code imul      eax, [ebx + 305419896]}
     */
    // Template#: 2016, Serial#: 5016
    public void imul(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0xAF, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code imul  }<i>destination</i>, <i>disp32</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code imul      eax, [ebx + 305419896], 0x12}
     */
    // Template#: 2017, Serial#: 1790
    public void imul(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source, byte imm8) {
        assemble0351((byte) 0x6B, destination, disp32, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code imull  }<i>destination</i>, <i>disp32</i>, <i>source</i>, <i>imm32</i>
     * Example disassembly syntax: {@code imull     eax, [ebx + 305419896], 0x12345678}
     */
    // Template#: 2018, Serial#: 1771
    public void imul(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source, int imm32) {
        assemble0352((byte) 0x69, destination, disp32, source, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code imull  }<i>destination</i>, <i>m32</i>, <i>imm32</i>
     * Example disassembly syntax: {@code imull     eax, [L1: 0x12345678], 0x12345678}
     */
    // Template#: 2019, Serial#: 1768
    public void m_imul(IA32GeneralRegister32 destination, int m32, int imm32) {
        assemble0353((byte) 0x69, destination, m32, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulb  }<i>destination</i>
     * Example disassembly syntax: {@code imulb     al, al}
     */
    // Template#: 2020, Serial#: 1407
    public void imulb___AL(IA32GeneralRegister8 destination) {
        assemble0297((byte) 0xF6, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulb  }<i>destination</i>
     * Example disassembly syntax: {@code imulb     [ebx], al}
     */
    // Template#: 2021, Serial#: 1363
    public void imulb___AL(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xF6, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code imull  }<i>destination</i>
     * Example disassembly syntax: {@code imull     [ebx], eax}
     */
    // Template#: 2022, Serial#: 1426
    public void imull___EAX(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xF7, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulw  }<i>destination</i>
     * Example disassembly syntax: {@code imulw     [ebx], ax}
     */
    // Template#: 2023, Serial#: 1489
    public void imulw___AX(IA32IndirectRegister32 destination) {
        assemble0298((byte) 0xF7, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulb  }<i>m32</i>
     * Example disassembly syntax: {@code imulb     [L1: 0x12345678], al}
     */
    // Template#: 2024, Serial#: 1366
    public void m_imulb___AL(int m32) {
        assemble0299((byte) 0xF6, (byte) 0x05, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code imull  }<i>m32</i>
     * Example disassembly syntax: {@code imull     [L1: 0x12345678], eax}
     */
    // Template#: 2025, Serial#: 1429
    public void m_imull___EAX(int m32) {
        assemble0299((byte) 0xF7, (byte) 0x05, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulw  }<i>m32</i>
     * Example disassembly syntax: {@code imulw     [L1: 0x12345678], ax}
     */
    // Template#: 2026, Serial#: 1492
    public void m_imulw___AX(int m32) {
        assemble0300((byte) 0xF7, (byte) 0x05, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code imulb     ebx[esi * 4 + 305419896], al}
     */
    // Template#: 2027, Serial#: 1398
    public void imulb___AL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xF6, (byte) 0x05, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code imull  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code imull     ebx[esi * 4 + 305419896], eax}
     */
    // Template#: 2028, Serial#: 1461
    public void imull___EAX(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xF7, (byte) 0x05, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code imulw     ebx[esi * 4 + 305419896], ax}
     */
    // Template#: 2029, Serial#: 1524
    public void imulw___AX(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0301((byte) 0xF7, (byte) 0x05, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulb  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code imulb     L1: 0x12345678[esi * 4], al}
     */
    // Template#: 2030, Serial#: 1365
    public void m_imulb___AL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xF6, (byte) 0x05, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code imull  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code imull     L1: 0x12345678[esi * 4], eax}
     */
    // Template#: 2031, Serial#: 1428
    public void m_imull___EAX(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xF7, (byte) 0x05, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code imulw     L1: 0x12345678[esi * 4], ax}
     */
    // Template#: 2032, Serial#: 1491
    public void m_imulw___AX(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0302((byte) 0xF7, (byte) 0x05, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulb  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code imulb     [ebx + 305419896], al}
     */
    // Template#: 2033, Serial#: 1397
    public void imulb___AL(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xF6, (byte) 0x05, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code imull  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code imull     [ebx + 305419896], eax}
     */
    // Template#: 2034, Serial#: 1460
    public void imull___EAX(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xF7, (byte) 0x05, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code imulw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code imulw     [ebx + 305419896], ax}
     */
    // Template#: 2035, Serial#: 1523
    public void imulw___AX(int disp32, IA32IndirectRegister32 destination) {
        assemble0303((byte) 0xF7, (byte) 0x05, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code in  }
     * Example disassembly syntax: {@code in        al, dx}
     */
    // Template#: 2036, Serial#: 2404
    public void in_AL_DX() {
        assemble0001((byte) 0xEC);
    }

    /**
     * Pseudo-external assembler syntax: {@code in  }
     * Example disassembly syntax: {@code in        eax, dx}
     */
    // Template#: 2037, Serial#: 2405
    public void in_EAX_DX() {
        assemble0001((byte) 0xED);
    }

    /**
     * Pseudo-external assembler syntax: {@code in  }
     * Example disassembly syntax: {@code in        ax, dx}
     */
    // Template#: 2038, Serial#: 2406
    public void in_AX_DX() {
        assemble0197((byte) 0xED);
    }

    /**
     * Pseudo-external assembler syntax: {@code in  }<i>imm8</i>
     * Example disassembly syntax: {@code in        al, 0x12}
     */
    // Template#: 2039, Serial#: 1336
    public void in_AL(byte imm8) {
        assemble0002((byte) 0xE4, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code in  }<i>imm8</i>
     * Example disassembly syntax: {@code in        eax, 0x12}
     */
    // Template#: 2040, Serial#: 1337
    public void in_EAX(byte imm8) {
        assemble0002((byte) 0xE5, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code in  }<i>imm8</i>
     * Example disassembly syntax: {@code in        ax, 0x12}
     */
    // Template#: 2041, Serial#: 1338
    public void in_AX(byte imm8) {
        assemble0354((byte) 0xE5, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code incb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code incb      ebx[esi * 4 + 18]}
     */
    // Template#: 2042, Serial#: 2425
    public void incb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xFE, (byte) 0x00, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code incl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code incl      ebx[esi * 4 + 18]}
     */
    // Template#: 2043, Serial#: 2453
    public void incl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xFF, (byte) 0x00, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code incw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code incw      ebx[esi * 4 + 18]}
     */
    // Template#: 2044, Serial#: 2485
    public void incw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0292((byte) 0xFF, (byte) 0x00, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code incb  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code incb      [ebx + 18]}
     */
    // Template#: 2045, Serial#: 2424
    public void incb(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xFE, (byte) 0x00, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code incl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code incl      [ebx + 18]}
     */
    // Template#: 2046, Serial#: 2452
    public void incl(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xFF, (byte) 0x00, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code incw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code incw      [ebx + 18]}
     */
    // Template#: 2047, Serial#: 2484
    public void incw(byte disp8, IA32IndirectRegister32 destination) {
        assemble0293((byte) 0xFF, (byte) 0x00, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code incb  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code incb      ebx[esi * 4]}
     */
    // Template#: 2048, Serial#: 2417
    public void incb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xFE, (byte) 0x00, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code incl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code incl      ebx[esi * 4]}
     */
    // Template#: 2049, Serial#: 2435
    public void incl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xFF, (byte) 0x00, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code incw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code incw      ebx[esi * 4]}
     */
    // Template#: 2050, Serial#: 2477
    public void incw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0294((byte) 0xFF, (byte) 0x00, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code inc  }<i>register</i>
     * Example disassembly syntax: {@code inc       ax}
     */
    // Template#: 2051, Serial#: 226
    public void inc(IA32GeneralRegister16 register) {
        assemble0295((byte) 0x40, register);
    }

    /**
     * Pseudo-external assembler syntax: {@code incw  }<i>destination</i>
     * Example disassembly syntax: {@code incw      ax}
     */
    // Template#: 2052, Serial#: 2492
    public void incw(IA32GeneralRegister16 destination) {
        assemble0188((byte) 0xFF, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code inc  }<i>register</i>
     * Example disassembly syntax: {@code inc       eax}
     */
    // Template#: 2053, Serial#: 225
    public void inc(IA32GeneralRegister32 register) {
        assemble0296((byte) 0x40, register);
    }

    /**
     * Pseudo-external assembler syntax: {@code incl  }<i>destination</i>
     * Example disassembly syntax: {@code incl      eax}
     */
    // Template#: 2054, Serial#: 2472
    public void incl(IA32GeneralRegister32 destination) {
        assemble0189((byte) 0xFF, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code incb  }<i>destination</i>
     * Example disassembly syntax: {@code incb      al}
     */
    // Template#: 2055, Serial#: 2432
    public void incb(IA32GeneralRegister8 destination) {
        assemble0297((byte) 0xFE, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code incb  }<i>destination</i>
     * Example disassembly syntax: {@code incb      [ebx]}
     */
    // Template#: 2056, Serial#: 2416
    public void incb(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xFE, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code incl  }<i>destination</i>
     * Example disassembly syntax: {@code incl      [ebx]}
     */
    // Template#: 2057, Serial#: 2434
    public void incl(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xFF, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code incw  }<i>destination</i>
     * Example disassembly syntax: {@code incw      [ebx]}
     */
    // Template#: 2058, Serial#: 2476
    public void incw(IA32IndirectRegister32 destination) {
        assemble0298((byte) 0xFF, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code incb  }<i>m32</i>
     * Example disassembly syntax: {@code incb      [L1: 0x12345678]}
     */
    // Template#: 2059, Serial#: 2419
    public void m_incb(int m32) {
        assemble0299((byte) 0xFE, (byte) 0x00, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code incl  }<i>m32</i>
     * Example disassembly syntax: {@code incl      [L1: 0x12345678]}
     */
    // Template#: 2060, Serial#: 2437
    public void m_incl(int m32) {
        assemble0299((byte) 0xFF, (byte) 0x00, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code incw  }<i>m32</i>
     * Example disassembly syntax: {@code incw      [L1: 0x12345678]}
     */
    // Template#: 2061, Serial#: 2479
    public void m_incw(int m32) {
        assemble0300((byte) 0xFF, (byte) 0x00, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code incb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code incb      ebx[esi * 4 + 305419896]}
     */
    // Template#: 2062, Serial#: 2429
    public void incb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xFE, (byte) 0x00, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code incl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code incl      ebx[esi * 4 + 305419896]}
     */
    // Template#: 2063, Serial#: 2463
    public void incl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xFF, (byte) 0x00, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code incw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code incw      ebx[esi * 4 + 305419896]}
     */
    // Template#: 2064, Serial#: 2489
    public void incw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0301((byte) 0xFF, (byte) 0x00, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code incb  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code incb      L1: 0x12345678[esi * 4]}
     */
    // Template#: 2065, Serial#: 2418
    public void m_incb(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xFE, (byte) 0x00, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code incl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code incl      L1: 0x12345678[esi * 4]}
     */
    // Template#: 2066, Serial#: 2436
    public void m_incl(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xFF, (byte) 0x00, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code incw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code incw      L1: 0x12345678[esi * 4]}
     */
    // Template#: 2067, Serial#: 2478
    public void m_incw(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0302((byte) 0xFF, (byte) 0x00, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code incb  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code incb      [ebx + 305419896]}
     */
    // Template#: 2068, Serial#: 2428
    public void incb(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xFE, (byte) 0x00, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code incl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code incl      [ebx + 305419896]}
     */
    // Template#: 2069, Serial#: 2462
    public void incl(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xFF, (byte) 0x00, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code incw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code incw      [ebx + 305419896]}
     */
    // Template#: 2070, Serial#: 2488
    public void incw(int disp32, IA32IndirectRegister32 destination) {
        assemble0303((byte) 0xFF, (byte) 0x00, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code insb  }
     * Example disassembly syntax: {@code insb    }
     */
    // Template#: 2071, Serial#: 1802
    public void insb() {
        assemble0001((byte) 0x6C);
    }

    /**
     * Pseudo-external assembler syntax: {@code insl  }
     * Example disassembly syntax: {@code insl    }
     */
    // Template#: 2072, Serial#: 1803
    public void insl() {
        assemble0001((byte) 0x6D);
    }

    /**
     * Pseudo-external assembler syntax: {@code insw  }
     * Example disassembly syntax: {@code insw    }
     */
    // Template#: 2073, Serial#: 1804
    public void insw() {
        assemble0197((byte) 0x6D);
    }

    /**
     * Pseudo-external assembler syntax: {@code int  }
     * Example disassembly syntax: {@code int       0x3}
     */
    // Template#: 2074, Serial#: 1939
    public void int_3() {
        assemble0001((byte) 0xCC);
    }

    /**
     * Pseudo-external assembler syntax: {@code int  }<i>imm8</i>
     * Example disassembly syntax: {@code int       0x12}
     */
    // Template#: 2075, Serial#: 1940
    public void intb(byte imm8) {
        assemble0002((byte) 0xCD, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code into  }
     * Example disassembly syntax: {@code into    }
     */
    // Template#: 2076, Serial#: 1941
    public void into() {
        assemble0001((byte) 0xCE);
    }

    /**
     * Pseudo-external assembler syntax: {@code invd  }
     * Example disassembly syntax: {@code invd    }
     */
    // Template#: 2077, Serial#: 4029
    public void invd() {
        assemble0198((byte) 0x08);
    }

    /**
     * Pseudo-external assembler syntax: {@code invlpg  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code invlpg    ebx[esi * 4 + 18]}
     */
    // Template#: 2078, Serial#: 2593
    public void invlpg(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0244((byte) 0x01, (byte) 0x07, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code invlpg  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code invlpg    [ebx + 18]}
     */
    // Template#: 2079, Serial#: 2592
    public void invlpg(byte disp8, IA32IndirectRegister32 destination) {
        assemble0245((byte) 0x01, (byte) 0x07, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code invlpg  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code invlpg    ebx[esi * 4]}
     */
    // Template#: 2080, Serial#: 2577
    public void invlpg(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0246((byte) 0x01, (byte) 0x07, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code invlpg  }<i>destination</i>
     * Example disassembly syntax: {@code invlpg    [ebx]}
     */
    // Template#: 2081, Serial#: 2576
    public void invlpg(IA32IndirectRegister32 destination) {
        assemble0247((byte) 0x01, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code invlpg  }<i>m32</i>
     * Example disassembly syntax: {@code invlpg    [L1: 0x12345678]}
     */
    // Template#: 2082, Serial#: 2579
    public void m_invlpg(int m32) {
        assemble0248((byte) 0x01, (byte) 0x07, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code invlpg  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code invlpg    ebx[esi * 4 + 305419896]}
     */
    // Template#: 2083, Serial#: 2607
    public void invlpg(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0249((byte) 0x01, (byte) 0x07, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code invlpg  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code invlpg    L1: 0x12345678[esi * 4]}
     */
    // Template#: 2084, Serial#: 2578
    public void m_invlpg(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0250((byte) 0x01, (byte) 0x07, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code invlpg  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code invlpg    [ebx + 305419896]}
     */
    // Template#: 2085, Serial#: 2606
    public void invlpg(int disp32, IA32IndirectRegister32 destination) {
        assemble0251((byte) 0x01, (byte) 0x07, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code iret  }
     * Example disassembly syntax: {@code iret    }
     */
    // Template#: 2086, Serial#: 1942
    public void iret() {
        assemble0001((byte) 0xCF);
    }

    /**
     * Pseudo-external assembler syntax: {@code jb  }<i>rel8</i>
     * Example disassembly syntax: {@code jb        L1: +18}
     */
    // Template#: 2087, Serial#: 264
    public void jb(byte rel8) {
        assemble0355((byte) 0x72, rel8);
    }

    /**
     * Pseudo-external assembler syntax: {@code jb  }<i>rel32</i>
     * Example disassembly syntax: {@code jb        L1: +305419896}
     */
    // Template#: 2088, Serial#: 3331
    public void jb(int rel32) {
        assemble0356((byte) 0x82, rel32);
    }

    /**
     * Pseudo-external assembler syntax: {@code jbe  }<i>rel8</i>
     * Example disassembly syntax: {@code jbe       L1: +18}
     */
    // Template#: 2089, Serial#: 268
    public void jbe(byte rel8) {
        assemble0355((byte) 0x76, rel8);
    }

    /**
     * Pseudo-external assembler syntax: {@code jbe  }<i>rel32</i>
     * Example disassembly syntax: {@code jbe       L1: +305419896}
     */
    // Template#: 2090, Serial#: 3335
    public void jbe(int rel32) {
        assemble0356((byte) 0x86, rel32);
    }

    /**
     * Pseudo-external assembler syntax: {@code jecxz  }<i>rel8</i>
     * Example disassembly syntax: {@code jecxz     L1: +18}
     */
    // Template#: 2091, Serial#: 1335
    public void jecxz(byte rel8) {
        assemble0355((byte) 0xE3, rel8);
    }

    /**
     * Pseudo-external assembler syntax: {@code jl  }<i>rel8</i>
     * Example disassembly syntax: {@code jl        L1: +18}
     */
    // Template#: 2092, Serial#: 1812
    public void jl(byte rel8) {
        assemble0355((byte) 0x7C, rel8);
    }

    /**
     * Pseudo-external assembler syntax: {@code jl  }<i>rel32</i>
     * Example disassembly syntax: {@code jl        L1: +305419896}
     */
    // Template#: 2093, Serial#: 4874
    public void jl(int rel32) {
        assemble0356((byte) 0x8C, rel32);
    }

    /**
     * Pseudo-external assembler syntax: {@code jle  }<i>rel8</i>
     * Example disassembly syntax: {@code jle       L1: +18}
     */
    // Template#: 2094, Serial#: 1814
    public void jle(byte rel8) {
        assemble0355((byte) 0x7E, rel8);
    }

    /**
     * Pseudo-external assembler syntax: {@code jle  }<i>rel32</i>
     * Example disassembly syntax: {@code jle       L1: +305419896}
     */
    // Template#: 2095, Serial#: 4876
    public void jle(int rel32) {
        assemble0356((byte) 0x8E, rel32);
    }

    /**
     * Pseudo-external assembler syntax: {@code jmp  }<i>rel8</i>
     * Example disassembly syntax: {@code jmp       L1: +18}
     */
    // Template#: 2096, Serial#: 2403
    public void jmp(byte rel8) {
        assemble0355((byte) 0xEB, rel8);
    }

    /**
     * Pseudo-external assembler syntax: {@code jmp  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code jmp       ebx[esi * 4 + 18]}
     */
    // Template#: 2097, Serial#: 2459
    public void jmp(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xFF, (byte) 0x04, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code jmp  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code jmp       [ebx + 18]}
     */
    // Template#: 2098, Serial#: 2458
    public void jmp(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xFF, (byte) 0x04, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code jmp  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code jmp       ebx[esi * 4]}
     */
    // Template#: 2099, Serial#: 2446
    public void jmp(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xFF, (byte) 0x04, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code jmp  }<i>destination</i>
     * Example disassembly syntax: {@code jmp       ax}
     */
    // Template#: 2100, Serial#: 2495
    public void jmp(IA32GeneralRegister16 destination) {
        assemble0188((byte) 0xFF, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code jmp  }<i>destination</i>
     * Example disassembly syntax: {@code jmp       eax}
     */
    // Template#: 2101, Serial#: 2475
    public void jmp(IA32GeneralRegister32 destination) {
        assemble0189((byte) 0xFF, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code jmp  }<i>destination</i>
     * Example disassembly syntax: {@code jmp       [ebx]}
     */
    // Template#: 2102, Serial#: 2445
    public void jmp(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xFF, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code jmp  }<i>rel32</i>
     * Example disassembly syntax: {@code jmp       L1: +305419896}
     */
    // Template#: 2103, Serial#: 2401
    public void jmp(int rel32) {
        assemble0192((byte) 0xE9, rel32);
    }

    /**
     * Pseudo-external assembler syntax: {@code jmp  }<i>m32</i>
     * Example disassembly syntax: {@code jmp       [L1: 0x12345678]}
     */
    // Template#: 2104, Serial#: 2402
    public void m_jmp(int m32) {
        assemble0191((byte) 0xEA, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code jmp  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code jmp       ebx[esi * 4 + 305419896]}
     */
    // Template#: 2105, Serial#: 2469
    public void jmp(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xFF, (byte) 0x04, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code jmp  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code jmp       L1: 0x12345678[esi * 4]}
     */
    // Template#: 2106, Serial#: 2447
    public void m_jmp(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xFF, (byte) 0x04, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code jmp  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code jmp       [ebx + 305419896]}
     */
    // Template#: 2107, Serial#: 2468
    public void jmp(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xFF, (byte) 0x04, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code jnb  }<i>rel8</i>
     * Example disassembly syntax: {@code jnb       L1: +18}
     */
    // Template#: 2108, Serial#: 265
    public void jnb(byte rel8) {
        assemble0355((byte) 0x73, rel8);
    }

    /**
     * Pseudo-external assembler syntax: {@code jnb  }<i>rel32</i>
     * Example disassembly syntax: {@code jnb       L1: +305419896}
     */
    // Template#: 2109, Serial#: 3332
    public void jnb(int rel32) {
        assemble0356((byte) 0x83, rel32);
    }

    /**
     * Pseudo-external assembler syntax: {@code jnbe  }<i>rel8</i>
     * Example disassembly syntax: {@code jnbe      L1: +18}
     */
    // Template#: 2110, Serial#: 269
    public void jnbe(byte rel8) {
        assemble0355((byte) 0x77, rel8);
    }

    /**
     * Pseudo-external assembler syntax: {@code jnbe  }<i>rel32</i>
     * Example disassembly syntax: {@code jnbe      L1: +305419896}
     */
    // Template#: 2111, Serial#: 3336
    public void jnbe(int rel32) {
        assemble0356((byte) 0x87, rel32);
    }

    /**
     * Pseudo-external assembler syntax: {@code jnl  }<i>rel8</i>
     * Example disassembly syntax: {@code jnl       L1: +18}
     */
    // Template#: 2112, Serial#: 1813
    public void jnl(byte rel8) {
        assemble0355((byte) 0x7D, rel8);
    }

    /**
     * Pseudo-external assembler syntax: {@code jnl  }<i>rel32</i>
     * Example disassembly syntax: {@code jnl       L1: +305419896}
     */
    // Template#: 2113, Serial#: 4875
    public void jnl(int rel32) {
        assemble0356((byte) 0x8D, rel32);
    }

    /**
     * Pseudo-external assembler syntax: {@code jnle  }<i>rel8</i>
     * Example disassembly syntax: {@code jnle      L1: +18}
     */
    // Template#: 2114, Serial#: 1815
    public void jnle(byte rel8) {
        assemble0355((byte) 0x7F, rel8);
    }

    /**
     * Pseudo-external assembler syntax: {@code jnle  }<i>rel32</i>
     * Example disassembly syntax: {@code jnle      L1: +305419896}
     */
    // Template#: 2115, Serial#: 4877
    public void jnle(int rel32) {
        assemble0356((byte) 0x8F, rel32);
    }

    /**
     * Pseudo-external assembler syntax: {@code jno  }<i>rel8</i>
     * Example disassembly syntax: {@code jno       L1: +18}
     */
    // Template#: 2116, Serial#: 263
    public void jno(byte rel8) {
        assemble0355((byte) 0x71, rel8);
    }

    /**
     * Pseudo-external assembler syntax: {@code jno  }<i>rel32</i>
     * Example disassembly syntax: {@code jno       L1: +305419896}
     */
    // Template#: 2117, Serial#: 3330
    public void jno(int rel32) {
        assemble0356((byte) 0x81, rel32);
    }

    /**
     * Pseudo-external assembler syntax: {@code jnp  }<i>rel8</i>
     * Example disassembly syntax: {@code jnp       L1: +18}
     */
    // Template#: 2118, Serial#: 1811
    public void jnp(byte rel8) {
        assemble0355((byte) 0x7B, rel8);
    }

    /**
     * Pseudo-external assembler syntax: {@code jnp  }<i>rel32</i>
     * Example disassembly syntax: {@code jnp       L1: +305419896}
     */
    // Template#: 2119, Serial#: 4873
    public void jnp(int rel32) {
        assemble0356((byte) 0x8B, rel32);
    }

    /**
     * Pseudo-external assembler syntax: {@code jns  }<i>rel8</i>
     * Example disassembly syntax: {@code jns       L1: +18}
     */
    // Template#: 2120, Serial#: 1809
    public void jns(byte rel8) {
        assemble0355((byte) 0x79, rel8);
    }

    /**
     * Pseudo-external assembler syntax: {@code jns  }<i>rel32</i>
     * Example disassembly syntax: {@code jns       L1: +305419896}
     */
    // Template#: 2121, Serial#: 4871
    public void jns(int rel32) {
        assemble0356((byte) 0x89, rel32);
    }

    /**
     * Pseudo-external assembler syntax: {@code jnz  }<i>rel8</i>
     * Example disassembly syntax: {@code jnz       L1: +18}
     */
    // Template#: 2122, Serial#: 267
    public void jnz(byte rel8) {
        assemble0355((byte) 0x75, rel8);
    }

    /**
     * Pseudo-external assembler syntax: {@code jnz  }<i>rel32</i>
     * Example disassembly syntax: {@code jnz       L1: +305419896}
     */
    // Template#: 2123, Serial#: 3334
    public void jnz(int rel32) {
        assemble0356((byte) 0x85, rel32);
    }

    /**
     * Pseudo-external assembler syntax: {@code jo  }<i>rel8</i>
     * Example disassembly syntax: {@code jo        L1: +18}
     */
    // Template#: 2124, Serial#: 262
    public void jo(byte rel8) {
        assemble0355((byte) 0x70, rel8);
    }

    /**
     * Pseudo-external assembler syntax: {@code jo  }<i>rel32</i>
     * Example disassembly syntax: {@code jo        L1: +305419896}
     */
    // Template#: 2125, Serial#: 3329
    public void jo(int rel32) {
        assemble0356((byte) 0x80, rel32);
    }

    /**
     * Pseudo-external assembler syntax: {@code jp  }<i>rel8</i>
     * Example disassembly syntax: {@code jp        L1: +18}
     */
    // Template#: 2126, Serial#: 1810
    public void jp(byte rel8) {
        assemble0355((byte) 0x7A, rel8);
    }

    /**
     * Pseudo-external assembler syntax: {@code jp  }<i>rel32</i>
     * Example disassembly syntax: {@code jp        L1: +305419896}
     */
    // Template#: 2127, Serial#: 4872
    public void jp(int rel32) {
        assemble0356((byte) 0x8A, rel32);
    }

    /**
     * Pseudo-external assembler syntax: {@code js  }<i>rel8</i>
     * Example disassembly syntax: {@code js        L1: +18}
     */
    // Template#: 2128, Serial#: 1808
    public void js(byte rel8) {
        assemble0355((byte) 0x78, rel8);
    }

    /**
     * Pseudo-external assembler syntax: {@code js  }<i>rel32</i>
     * Example disassembly syntax: {@code js        L1: +305419896}
     */
    // Template#: 2129, Serial#: 4870
    public void js(int rel32) {
        assemble0356((byte) 0x88, rel32);
    }

    /**
     * Pseudo-external assembler syntax: {@code jz  }<i>rel8</i>
     * Example disassembly syntax: {@code jz        L1: +18}
     */
    // Template#: 2130, Serial#: 266
    public void jz(byte rel8) {
        assemble0355((byte) 0x74, rel8);
    }

    /**
     * Pseudo-external assembler syntax: {@code jz  }<i>rel32</i>
     * Example disassembly syntax: {@code jz        L1: +305419896}
     */
    // Template#: 2131, Serial#: 3333
    public void jz(int rel32) {
        assemble0356((byte) 0x84, rel32);
    }

    /**
     * Pseudo-external assembler syntax: {@code lahf  }
     * Example disassembly syntax: {@code lahf    }
     */
    // Template#: 2132, Serial#: 1920
    public void lahf() {
        assemble0001((byte) 0x9F);
    }

    /**
     * Pseudo-external assembler syntax: {@code lar  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lar       ax, ebx[esi * 4 + 18]}
     */
    // Template#: 2133, Serial#: 2625
    public void lar(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0x02, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lar  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code lar       ax, [ebx + 18]}
     */
    // Template#: 2134, Serial#: 2624
    public void lar(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0x02, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lar  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lar       ax, ebx[esi * 4]}
     */
    // Template#: 2135, Serial#: 2621
    public void lar(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0x02, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lar  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code lar       ax, ax}
     */
    // Template#: 2136, Serial#: 2628
    public void lar(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0141((byte) 0x02, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lar  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code lar       ax, [ebx]}
     */
    // Template#: 2137, Serial#: 2620
    public void lar(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0x02, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lar  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code lar       ax, [L1: 0x12345678]}
     */
    // Template#: 2138, Serial#: 2623
    public void m_lar(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0x02, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code lar  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lar       ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2139, Serial#: 2627
    public void lar(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0x02, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lar  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lar       ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2140, Serial#: 2622
    public void m_lar(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0x02, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lar  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code lar       ax, [ebx + 305419896]}
     */
    // Template#: 2141, Serial#: 2626
    public void lar(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0x02, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lar  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lar       eax, ebx[esi * 4 + 18]}
     */
    // Template#: 2142, Serial#: 2616
    public void lar(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0x02, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lar  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code lar       eax, [ebx + 18]}
     */
    // Template#: 2143, Serial#: 2615
    public void lar(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0x02, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lar  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lar       eax, ebx[esi * 4]}
     */
    // Template#: 2144, Serial#: 2612
    public void lar(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0x02, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lar  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code lar       eax, eax}
     */
    // Template#: 2145, Serial#: 2619
    public void lar(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0150((byte) 0x02, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lar  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code lar       eax, [ebx]}
     */
    // Template#: 2146, Serial#: 2611
    public void lar(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0x02, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lar  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code lar       eax, [L1: 0x12345678]}
     */
    // Template#: 2147, Serial#: 2614
    public void m_lar(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0x02, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code lar  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lar       eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2148, Serial#: 2618
    public void lar(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0x02, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lar  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lar       eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2149, Serial#: 2613
    public void m_lar(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0x02, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lar  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code lar       eax, [ebx + 305419896]}
     */
    // Template#: 2150, Serial#: 2617
    public void lar(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0x02, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lddqu  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lddqu     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2151, Serial#: 4026
    public void lddqu(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0111((byte) 0xF0, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lddqu  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code lddqu     xmm0, [ebx + 18]}
     */
    // Template#: 2152, Serial#: 4025
    public void lddqu(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0112((byte) 0xF0, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lddqu  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lddqu     xmm0, ebx[esi * 4]}
     */
    // Template#: 2153, Serial#: 4022
    public void lddqu(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0113((byte) 0xF0, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lddqu  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code lddqu     xmm0, [ebx]}
     */
    // Template#: 2154, Serial#: 4021
    public void lddqu(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0114((byte) 0xF0, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lddqu  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code lddqu     xmm0, [L1: 0x12345678]}
     */
    // Template#: 2155, Serial#: 4024
    public void m_lddqu(IA32XMMRegister destination, int m32) {
        assemble0116((byte) 0xF0, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code lddqu  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lddqu     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2156, Serial#: 4028
    public void lddqu(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0117((byte) 0xF0, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lddqu  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lddqu     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2157, Serial#: 4023
    public void m_lddqu(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0118((byte) 0xF0, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lddqu  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code lddqu     xmm0, [ebx + 305419896]}
     */
    // Template#: 2158, Serial#: 4027
    public void lddqu(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0119((byte) 0xF0, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lds  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lds       ax, ebx[esi * 4 + 18]}
     */
    // Template#: 2159, Serial#: 920
    public void lds(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0025((byte) 0xC5, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lds  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code lds       ax, [ebx + 18]}
     */
    // Template#: 2160, Serial#: 919
    public void lds(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0026((byte) 0xC5, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lds  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lds       ax, ebx[esi * 4]}
     */
    // Template#: 2161, Serial#: 916
    public void lds(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0027((byte) 0xC5, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lds  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code lds       ax, [ebx]}
     */
    // Template#: 2162, Serial#: 915
    public void lds(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0029((byte) 0xC5, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lds  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code lds       ax, [L1: 0x12345678]}
     */
    // Template#: 2163, Serial#: 918
    public void m_lds(IA32GeneralRegister16 destination, int m32) {
        assemble0030((byte) 0xC5, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code lds  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lds       ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2164, Serial#: 922
    public void lds(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0031((byte) 0xC5, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lds  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lds       ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2165, Serial#: 917
    public void m_lds(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0032((byte) 0xC5, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lds  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code lds       ax, [ebx + 305419896]}
     */
    // Template#: 2166, Serial#: 921
    public void lds(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0033((byte) 0xC5, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lds  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lds       eax, ebx[esi * 4 + 18]}
     */
    // Template#: 2167, Serial#: 912
    public void lds(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0036((byte) 0xC5, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lds  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code lds       eax, [ebx + 18]}
     */
    // Template#: 2168, Serial#: 911
    public void lds(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0037((byte) 0xC5, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lds  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lds       eax, ebx[esi * 4]}
     */
    // Template#: 2169, Serial#: 908
    public void lds(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0038((byte) 0xC5, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lds  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code lds       eax, [ebx]}
     */
    // Template#: 2170, Serial#: 907
    public void lds(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0040((byte) 0xC5, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lds  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code lds       eax, [L1: 0x12345678]}
     */
    // Template#: 2171, Serial#: 910
    public void m_lds(IA32GeneralRegister32 destination, int m32) {
        assemble0041((byte) 0xC5, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code lds  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lds       eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2172, Serial#: 914
    public void lds(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0043((byte) 0xC5, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lds  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lds       eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2173, Serial#: 909
    public void m_lds(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0044((byte) 0xC5, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lds  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code lds       eax, [ebx + 305419896]}
     */
    // Template#: 2174, Serial#: 913
    public void lds(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0045((byte) 0xC5, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lea  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lea       ax, ebx[esi * 4 + 18]}
     */
    // Template#: 2175, Serial#: 1889
    public void lea(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0025((byte) 0x8D, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lea  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code lea       ax, [ebx + 18]}
     */
    // Template#: 2176, Serial#: 1888
    public void lea(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0026((byte) 0x8D, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lea  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lea       ax, ebx[esi * 4]}
     */
    // Template#: 2177, Serial#: 1885
    public void lea(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0027((byte) 0x8D, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lea  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code lea       ax, [ebx]}
     */
    // Template#: 2178, Serial#: 1884
    public void lea(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0029((byte) 0x8D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lea  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code lea       ax, [L1: 0x12345678]}
     */
    // Template#: 2179, Serial#: 1887
    public void m_lea(IA32GeneralRegister16 destination, int m32) {
        assemble0030((byte) 0x8D, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code lea  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lea       ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2180, Serial#: 1891
    public void lea(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0031((byte) 0x8D, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lea  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lea       ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2181, Serial#: 1886
    public void m_lea(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0032((byte) 0x8D, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lea  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code lea       ax, [ebx + 305419896]}
     */
    // Template#: 2182, Serial#: 1890
    public void lea(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0033((byte) 0x8D, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lea  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lea       eax, ebx[esi * 4 + 18]}
     */
    // Template#: 2183, Serial#: 1881
    public void lea(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0036((byte) 0x8D, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lea  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code lea       eax, [ebx + 18]}
     */
    // Template#: 2184, Serial#: 1880
    public void lea(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0037((byte) 0x8D, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lea  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lea       eax, ebx[esi * 4]}
     */
    // Template#: 2185, Serial#: 1877
    public void lea(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0038((byte) 0x8D, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lea  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code lea       eax, [ebx]}
     */
    // Template#: 2186, Serial#: 1876
    public void lea(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0040((byte) 0x8D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lea  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code lea       eax, [L1: 0x12345678]}
     */
    // Template#: 2187, Serial#: 1879
    public void m_lea(IA32GeneralRegister32 destination, int m32) {
        assemble0041((byte) 0x8D, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code lea  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lea       eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2188, Serial#: 1883
    public void lea(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0043((byte) 0x8D, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lea  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lea       eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2189, Serial#: 1878
    public void m_lea(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0044((byte) 0x8D, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lea  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code lea       eax, [ebx + 305419896]}
     */
    // Template#: 2190, Serial#: 1882
    public void lea(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0045((byte) 0x8D, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code leave  }
     * Example disassembly syntax: {@code leave   }
     */
    // Template#: 2191, Serial#: 1936
    public void leave() {
        assemble0001((byte) 0xC9);
    }

    /**
     * Pseudo-external assembler syntax: {@code les  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code les       ax, ebx[esi * 4 + 18]}
     */
    // Template#: 2192, Serial#: 904
    public void les(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0025((byte) 0xC4, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code les  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code les       ax, [ebx + 18]}
     */
    // Template#: 2193, Serial#: 903
    public void les(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0026((byte) 0xC4, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code les  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code les       ax, ebx[esi * 4]}
     */
    // Template#: 2194, Serial#: 900
    public void les(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0027((byte) 0xC4, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code les  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code les       ax, [ebx]}
     */
    // Template#: 2195, Serial#: 899
    public void les(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0029((byte) 0xC4, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code les  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code les       ax, [L1: 0x12345678]}
     */
    // Template#: 2196, Serial#: 902
    public void m_les(IA32GeneralRegister16 destination, int m32) {
        assemble0030((byte) 0xC4, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code les  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code les       ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2197, Serial#: 906
    public void les(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0031((byte) 0xC4, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code les  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code les       ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2198, Serial#: 901
    public void m_les(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0032((byte) 0xC4, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code les  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code les       ax, [ebx + 305419896]}
     */
    // Template#: 2199, Serial#: 905
    public void les(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0033((byte) 0xC4, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code les  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code les       eax, ebx[esi * 4 + 18]}
     */
    // Template#: 2200, Serial#: 896
    public void les(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0036((byte) 0xC4, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code les  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code les       eax, [ebx + 18]}
     */
    // Template#: 2201, Serial#: 895
    public void les(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0037((byte) 0xC4, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code les  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code les       eax, ebx[esi * 4]}
     */
    // Template#: 2202, Serial#: 892
    public void les(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0038((byte) 0xC4, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code les  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code les       eax, [ebx]}
     */
    // Template#: 2203, Serial#: 891
    public void les(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0040((byte) 0xC4, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code les  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code les       eax, [L1: 0x12345678]}
     */
    // Template#: 2204, Serial#: 894
    public void m_les(IA32GeneralRegister32 destination, int m32) {
        assemble0041((byte) 0xC4, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code les  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code les       eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2205, Serial#: 898
    public void les(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0043((byte) 0xC4, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code les  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code les       eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2206, Serial#: 893
    public void m_les(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0044((byte) 0xC4, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code les  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code les       eax, [ebx + 305419896]}
     */
    // Template#: 2207, Serial#: 897
    public void les(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0045((byte) 0xC4, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lfence  }
     * Example disassembly syntax: {@code lfence  }
     */
    // Template#: 2208, Serial#: 5007
    public void lfence() {
        assemble0357((byte) 0xAE, (byte) 0x05);
    }

    /**
     * Pseudo-external assembler syntax: {@code lfs  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lfs       ax, ebx[esi * 4 + 18]}
     */
    // Template#: 2209, Serial#: 3540
    public void lfs(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0xB4, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lfs  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code lfs       ax, [ebx + 18]}
     */
    // Template#: 2210, Serial#: 3539
    public void lfs(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0xB4, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lfs  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lfs       ax, ebx[esi * 4]}
     */
    // Template#: 2211, Serial#: 3536
    public void lfs(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0xB4, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lfs  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code lfs       ax, [ebx]}
     */
    // Template#: 2212, Serial#: 3535
    public void lfs(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0xB4, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lfs  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code lfs       ax, [L1: 0x12345678]}
     */
    // Template#: 2213, Serial#: 3538
    public void m_lfs(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0xB4, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code lfs  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lfs       ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2214, Serial#: 3542
    public void lfs(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0xB4, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lfs  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lfs       ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2215, Serial#: 3537
    public void m_lfs(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0xB4, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lfs  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code lfs       ax, [ebx + 305419896]}
     */
    // Template#: 2216, Serial#: 3541
    public void lfs(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0xB4, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lfs  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lfs       eax, ebx[esi * 4 + 18]}
     */
    // Template#: 2217, Serial#: 3532
    public void lfs(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0xB4, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lfs  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code lfs       eax, [ebx + 18]}
     */
    // Template#: 2218, Serial#: 3531
    public void lfs(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0xB4, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lfs  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lfs       eax, ebx[esi * 4]}
     */
    // Template#: 2219, Serial#: 3528
    public void lfs(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0xB4, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lfs  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code lfs       eax, [ebx]}
     */
    // Template#: 2220, Serial#: 3527
    public void lfs(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0xB4, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lfs  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code lfs       eax, [L1: 0x12345678]}
     */
    // Template#: 2221, Serial#: 3530
    public void m_lfs(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0xB4, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code lfs  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lfs       eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2222, Serial#: 3534
    public void lfs(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0xB4, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lfs  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lfs       eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2223, Serial#: 3529
    public void m_lfs(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0xB4, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lfs  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code lfs       eax, [ebx + 305419896]}
     */
    // Template#: 2224, Serial#: 3533
    public void lfs(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0xB4, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lgdt  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lgdt      ebx[esi * 4 + 18]}
     */
    // Template#: 2225, Serial#: 2585
    public void lgdt(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0244((byte) 0x01, (byte) 0x02, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lgdt  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code lgdt      [ebx + 18]}
     */
    // Template#: 2226, Serial#: 2584
    public void lgdt(byte disp8, IA32IndirectRegister32 destination) {
        assemble0245((byte) 0x01, (byte) 0x02, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code lgdt  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lgdt      ebx[esi * 4]}
     */
    // Template#: 2227, Serial#: 2561
    public void lgdt(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0246((byte) 0x01, (byte) 0x02, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lgdt  }<i>destination</i>
     * Example disassembly syntax: {@code lgdt      [ebx]}
     */
    // Template#: 2228, Serial#: 2560
    public void lgdt(IA32IndirectRegister32 destination) {
        assemble0247((byte) 0x01, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code lgdt  }<i>m32</i>
     * Example disassembly syntax: {@code lgdt      [L1: 0x12345678]}
     */
    // Template#: 2229, Serial#: 2563
    public void m_lgdt(int m32) {
        assemble0248((byte) 0x01, (byte) 0x02, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code lgdt  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lgdt      ebx[esi * 4 + 305419896]}
     */
    // Template#: 2230, Serial#: 2599
    public void lgdt(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0249((byte) 0x01, (byte) 0x02, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lgdt  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lgdt      L1: 0x12345678[esi * 4]}
     */
    // Template#: 2231, Serial#: 2562
    public void m_lgdt(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0250((byte) 0x01, (byte) 0x02, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lgdt  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code lgdt      [ebx + 305419896]}
     */
    // Template#: 2232, Serial#: 2598
    public void lgdt(int disp32, IA32IndirectRegister32 destination) {
        assemble0251((byte) 0x01, (byte) 0x02, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code lgs  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lgs       ax, ebx[esi * 4 + 18]}
     */
    // Template#: 2233, Serial#: 3556
    public void lgs(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0xB5, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lgs  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code lgs       ax, [ebx + 18]}
     */
    // Template#: 2234, Serial#: 3555
    public void lgs(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0xB5, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lgs  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lgs       ax, ebx[esi * 4]}
     */
    // Template#: 2235, Serial#: 3552
    public void lgs(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0xB5, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lgs  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code lgs       ax, [ebx]}
     */
    // Template#: 2236, Serial#: 3551
    public void lgs(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0xB5, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lgs  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code lgs       ax, [L1: 0x12345678]}
     */
    // Template#: 2237, Serial#: 3554
    public void m_lgs(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0xB5, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code lgs  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lgs       ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2238, Serial#: 3558
    public void lgs(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0xB5, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lgs  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lgs       ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2239, Serial#: 3553
    public void m_lgs(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0xB5, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lgs  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code lgs       ax, [ebx + 305419896]}
     */
    // Template#: 2240, Serial#: 3557
    public void lgs(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0xB5, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lgs  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lgs       eax, ebx[esi * 4 + 18]}
     */
    // Template#: 2241, Serial#: 3548
    public void lgs(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0xB5, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lgs  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code lgs       eax, [ebx + 18]}
     */
    // Template#: 2242, Serial#: 3547
    public void lgs(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0xB5, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lgs  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lgs       eax, ebx[esi * 4]}
     */
    // Template#: 2243, Serial#: 3544
    public void lgs(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0xB5, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lgs  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code lgs       eax, [ebx]}
     */
    // Template#: 2244, Serial#: 3543
    public void lgs(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0xB5, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lgs  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code lgs       eax, [L1: 0x12345678]}
     */
    // Template#: 2245, Serial#: 3546
    public void m_lgs(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0xB5, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code lgs  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lgs       eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2246, Serial#: 3550
    public void lgs(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0xB5, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lgs  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lgs       eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2247, Serial#: 3545
    public void m_lgs(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0xB5, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lgs  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code lgs       eax, [ebx + 305419896]}
     */
    // Template#: 2248, Serial#: 3549
    public void lgs(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0xB5, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lidt  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lidt      ebx[esi * 4 + 18]}
     */
    // Template#: 2249, Serial#: 2587
    public void lidt(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0244((byte) 0x01, (byte) 0x03, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lidt  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code lidt      [ebx + 18]}
     */
    // Template#: 2250, Serial#: 2586
    public void lidt(byte disp8, IA32IndirectRegister32 destination) {
        assemble0245((byte) 0x01, (byte) 0x03, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code lidt  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lidt      ebx[esi * 4]}
     */
    // Template#: 2251, Serial#: 2565
    public void lidt(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0246((byte) 0x01, (byte) 0x03, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lidt  }<i>destination</i>
     * Example disassembly syntax: {@code lidt      [ebx]}
     */
    // Template#: 2252, Serial#: 2564
    public void lidt(IA32IndirectRegister32 destination) {
        assemble0247((byte) 0x01, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code lidt  }<i>m32</i>
     * Example disassembly syntax: {@code lidt      [L1: 0x12345678]}
     */
    // Template#: 2253, Serial#: 2567
    public void m_lidt(int m32) {
        assemble0248((byte) 0x01, (byte) 0x03, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code lidt  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lidt      ebx[esi * 4 + 305419896]}
     */
    // Template#: 2254, Serial#: 2601
    public void lidt(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0249((byte) 0x01, (byte) 0x03, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lidt  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lidt      L1: 0x12345678[esi * 4]}
     */
    // Template#: 2255, Serial#: 2566
    public void m_lidt(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0250((byte) 0x01, (byte) 0x03, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lidt  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code lidt      [ebx + 305419896]}
     */
    // Template#: 2256, Serial#: 2600
    public void lidt(int disp32, IA32IndirectRegister32 destination) {
        assemble0251((byte) 0x01, (byte) 0x03, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code lldt  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lldt      ebx[esi * 4 + 18]}
     */
    // Template#: 2257, Serial#: 2525
    public void lldt(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0244((byte) 0x00, (byte) 0x02, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lldt  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code lldt      [ebx + 18]}
     */
    // Template#: 2258, Serial#: 2524
    public void lldt(byte disp8, IA32IndirectRegister32 destination) {
        assemble0245((byte) 0x00, (byte) 0x02, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code lldt  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lldt      ebx[esi * 4]}
     */
    // Template#: 2259, Serial#: 2505
    public void lldt(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0246((byte) 0x00, (byte) 0x02, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lldt  }<i>destination</i>
     * Example disassembly syntax: {@code lldt      ax}
     */
    // Template#: 2260, Serial#: 2544
    public void lldt(IA32GeneralRegister16 destination) {
        assemble0358((byte) 0x00, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code lldt  }<i>destination</i>
     * Example disassembly syntax: {@code lldt      [ebx]}
     */
    // Template#: 2261, Serial#: 2504
    public void lldt(IA32IndirectRegister32 destination) {
        assemble0247((byte) 0x00, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code lldt  }<i>m32</i>
     * Example disassembly syntax: {@code lldt      [L1: 0x12345678]}
     */
    // Template#: 2262, Serial#: 2507
    public void m_lldt(int m32) {
        assemble0248((byte) 0x00, (byte) 0x02, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code lldt  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lldt      ebx[esi * 4 + 305419896]}
     */
    // Template#: 2263, Serial#: 2537
    public void lldt(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0249((byte) 0x00, (byte) 0x02, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lldt  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lldt      L1: 0x12345678[esi * 4]}
     */
    // Template#: 2264, Serial#: 2506
    public void m_lldt(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0250((byte) 0x00, (byte) 0x02, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lldt  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code lldt      [ebx + 305419896]}
     */
    // Template#: 2265, Serial#: 2536
    public void lldt(int disp32, IA32IndirectRegister32 destination) {
        assemble0251((byte) 0x00, (byte) 0x02, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code lmsw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lmsw      ebx[esi * 4 + 18]}
     */
    // Template#: 2266, Serial#: 2591
    public void lmsw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0244((byte) 0x01, (byte) 0x06, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lmsw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code lmsw      [ebx + 18]}
     */
    // Template#: 2267, Serial#: 2590
    public void lmsw(byte disp8, IA32IndirectRegister32 destination) {
        assemble0245((byte) 0x01, (byte) 0x06, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code lmsw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lmsw      ebx[esi * 4]}
     */
    // Template#: 2268, Serial#: 2573
    public void lmsw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0246((byte) 0x01, (byte) 0x06, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lmsw  }<i>destination</i>
     * Example disassembly syntax: {@code lmsw      ax}
     */
    // Template#: 2269, Serial#: 2608
    public void lmsw(IA32GeneralRegister16 destination) {
        assemble0358((byte) 0x01, (byte) 0x06, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code lmsw  }<i>destination</i>
     * Example disassembly syntax: {@code lmsw      [ebx]}
     */
    // Template#: 2270, Serial#: 2572
    public void lmsw(IA32IndirectRegister32 destination) {
        assemble0247((byte) 0x01, (byte) 0x06, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code lmsw  }<i>m32</i>
     * Example disassembly syntax: {@code lmsw      [L1: 0x12345678]}
     */
    // Template#: 2271, Serial#: 2575
    public void m_lmsw(int m32) {
        assemble0248((byte) 0x01, (byte) 0x06, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code lmsw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lmsw      ebx[esi * 4 + 305419896]}
     */
    // Template#: 2272, Serial#: 2605
    public void lmsw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0249((byte) 0x01, (byte) 0x06, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lmsw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lmsw      L1: 0x12345678[esi * 4]}
     */
    // Template#: 2273, Serial#: 2574
    public void m_lmsw(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0250((byte) 0x01, (byte) 0x06, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lmsw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code lmsw      [ebx + 305419896]}
     */
    // Template#: 2274, Serial#: 2604
    public void lmsw(int disp32, IA32IndirectRegister32 destination) {
        assemble0251((byte) 0x01, (byte) 0x06, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code lock  }
     * Example disassembly syntax: {@code lock    }
     */
    // Template#: 2275, Serial#: 1342
    public void lock() {
        assemble0001((byte) 0xF0);
    }

    /**
     * Pseudo-external assembler syntax: {@code lodsb  }
     * Example disassembly syntax: {@code lodsb   }
     */
    // Template#: 2276, Serial#: 1927
    public void lodsb() {
        assemble0001((byte) 0xAC);
    }

    /**
     * Pseudo-external assembler syntax: {@code lodsl  }
     * Example disassembly syntax: {@code lodsl   }
     */
    // Template#: 2277, Serial#: 1928
    public void lodsl() {
        assemble0001((byte) 0xAD);
    }

    /**
     * Pseudo-external assembler syntax: {@code lodsw  }
     * Example disassembly syntax: {@code lodsw   }
     */
    // Template#: 2278, Serial#: 1929
    public void lodsw() {
        assemble0197((byte) 0xAD);
    }

    /**
     * Pseudo-external assembler syntax: {@code loop  }<i>rel8</i>
     * Example disassembly syntax: {@code loop      L1: +18}
     */
    // Template#: 2279, Serial#: 1334
    public void loop(byte rel8) {
        assemble0355((byte) 0xE2, rel8);
    }

    /**
     * Pseudo-external assembler syntax: {@code loope  }<i>rel8</i>
     * Example disassembly syntax: {@code loope     L1: +18}
     */
    // Template#: 2280, Serial#: 1333
    public void loope(byte rel8) {
        assemble0355((byte) 0xE1, rel8);
    }

    /**
     * Pseudo-external assembler syntax: {@code loopne  }<i>rel8</i>
     * Example disassembly syntax: {@code loopne    L1: +18}
     */
    // Template#: 2281, Serial#: 1332
    public void loopne(byte rel8) {
        assemble0355((byte) 0xE0, rel8);
    }

    /**
     * Pseudo-external assembler syntax: {@code lsl  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lsl       ax, ebx[esi * 4 + 18]}
     */
    // Template#: 2282, Serial#: 2643
    public void lsl(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0x03, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lsl  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code lsl       ax, [ebx + 18]}
     */
    // Template#: 2283, Serial#: 2642
    public void lsl(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0x03, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lsl  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lsl       ax, ebx[esi * 4]}
     */
    // Template#: 2284, Serial#: 2639
    public void lsl(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0x03, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lsl  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code lsl       ax, ax}
     */
    // Template#: 2285, Serial#: 2646
    public void lsl(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0141((byte) 0x03, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lsl  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code lsl       ax, [ebx]}
     */
    // Template#: 2286, Serial#: 2638
    public void lsl(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0x03, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lsl  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code lsl       ax, [L1: 0x12345678]}
     */
    // Template#: 2287, Serial#: 2641
    public void m_lsl(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0x03, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code lsl  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lsl       ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2288, Serial#: 2645
    public void lsl(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0x03, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lsl  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lsl       ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2289, Serial#: 2640
    public void m_lsl(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0x03, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lsl  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code lsl       ax, [ebx + 305419896]}
     */
    // Template#: 2290, Serial#: 2644
    public void lsl(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0x03, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lsl  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lsl       eax, ebx[esi * 4 + 18]}
     */
    // Template#: 2291, Serial#: 2634
    public void lsl(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0x03, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lsl  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code lsl       eax, [ebx + 18]}
     */
    // Template#: 2292, Serial#: 2633
    public void lsl(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0x03, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lsl  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lsl       eax, ebx[esi * 4]}
     */
    // Template#: 2293, Serial#: 2630
    public void lsl(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0x03, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lsl  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code lsl       eax, eax}
     */
    // Template#: 2294, Serial#: 2637
    public void lsl(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0150((byte) 0x03, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lsl  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code lsl       eax, [ebx]}
     */
    // Template#: 2295, Serial#: 2629
    public void lsl(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0x03, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lsl  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code lsl       eax, [L1: 0x12345678]}
     */
    // Template#: 2296, Serial#: 2632
    public void m_lsl(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0x03, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code lsl  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lsl       eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2297, Serial#: 2636
    public void lsl(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0x03, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lsl  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lsl       eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2298, Serial#: 2631
    public void m_lsl(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0x03, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lsl  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code lsl       eax, [ebx + 305419896]}
     */
    // Template#: 2299, Serial#: 2635
    public void lsl(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0x03, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lss  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lss       ax, ebx[esi * 4 + 18]}
     */
    // Template#: 2300, Serial#: 3506
    public void lss(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0xB2, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lss  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code lss       ax, [ebx + 18]}
     */
    // Template#: 2301, Serial#: 3505
    public void lss(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0xB2, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lss  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lss       ax, ebx[esi * 4]}
     */
    // Template#: 2302, Serial#: 3502
    public void lss(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0xB2, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code lss       ax, [ebx]}
     */
    // Template#: 2303, Serial#: 3501
    public void lss(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0xB2, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lss  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code lss       ax, [L1: 0x12345678]}
     */
    // Template#: 2304, Serial#: 3504
    public void m_lss(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0xB2, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code lss  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lss       ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2305, Serial#: 3508
    public void lss(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0xB2, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lss  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lss       ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2306, Serial#: 3503
    public void m_lss(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0xB2, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lss  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code lss       ax, [ebx + 305419896]}
     */
    // Template#: 2307, Serial#: 3507
    public void lss(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0xB2, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lss  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lss       eax, ebx[esi * 4 + 18]}
     */
    // Template#: 2308, Serial#: 3498
    public void lss(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0xB2, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lss  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code lss       eax, [ebx + 18]}
     */
    // Template#: 2309, Serial#: 3497
    public void lss(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0xB2, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lss  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lss       eax, ebx[esi * 4]}
     */
    // Template#: 2310, Serial#: 3494
    public void lss(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0xB2, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code lss       eax, [ebx]}
     */
    // Template#: 2311, Serial#: 3493
    public void lss(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0xB2, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code lss  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code lss       eax, [L1: 0x12345678]}
     */
    // Template#: 2312, Serial#: 3496
    public void m_lss(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0xB2, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code lss  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lss       eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2313, Serial#: 3500
    public void lss(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0xB2, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lss  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code lss       eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2314, Serial#: 3495
    public void m_lss(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0xB2, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code lss  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code lss       eax, [ebx + 305419896]}
     */
    // Template#: 2315, Serial#: 3499
    public void lss(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0xB2, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code ltr  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ltr       ebx[esi * 4 + 18]}
     */
    // Template#: 2316, Serial#: 2527
    public void ltr(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0244((byte) 0x00, (byte) 0x03, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ltr  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code ltr       [ebx + 18]}
     */
    // Template#: 2317, Serial#: 2526
    public void ltr(byte disp8, IA32IndirectRegister32 destination) {
        assemble0245((byte) 0x00, (byte) 0x03, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code ltr  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ltr       ebx[esi * 4]}
     */
    // Template#: 2318, Serial#: 2509
    public void ltr(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0246((byte) 0x00, (byte) 0x03, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ltr  }<i>destination</i>
     * Example disassembly syntax: {@code ltr       ax}
     */
    // Template#: 2319, Serial#: 2545
    public void ltr(IA32GeneralRegister16 destination) {
        assemble0358((byte) 0x00, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code ltr  }<i>destination</i>
     * Example disassembly syntax: {@code ltr       [ebx]}
     */
    // Template#: 2320, Serial#: 2508
    public void ltr(IA32IndirectRegister32 destination) {
        assemble0247((byte) 0x00, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code ltr  }<i>m32</i>
     * Example disassembly syntax: {@code ltr       [L1: 0x12345678]}
     */
    // Template#: 2321, Serial#: 2511
    public void m_ltr(int m32) {
        assemble0248((byte) 0x00, (byte) 0x03, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code ltr  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ltr       ebx[esi * 4 + 305419896]}
     */
    // Template#: 2322, Serial#: 2539
    public void ltr(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0249((byte) 0x00, (byte) 0x03, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ltr  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ltr       L1: 0x12345678[esi * 4]}
     */
    // Template#: 2323, Serial#: 2510
    public void m_ltr(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0250((byte) 0x00, (byte) 0x03, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ltr  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code ltr       [ebx + 305419896]}
     */
    // Template#: 2324, Serial#: 2538
    public void ltr(int disp32, IA32IndirectRegister32 destination) {
        assemble0251((byte) 0x00, (byte) 0x03, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code maskmovdqu  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code maskmovdqu  xmm0, xmm0}
     */
    // Template#: 2325, Serial#: 4020
    public void maskmovdqu(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xF7, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxpd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code maxpd     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2326, Serial#: 4510
    public void maxpd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x5F, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxpd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code maxpd     xmm0, [ebx + 18]}
     */
    // Template#: 2327, Serial#: 4509
    public void maxpd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x5F, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxpd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code maxpd     xmm0, ebx[esi * 4]}
     */
    // Template#: 2328, Serial#: 4506
    public void maxpd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x5F, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code maxpd     xmm0, [ebx]}
     */
    // Template#: 2329, Serial#: 4505
    public void maxpd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x5F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code maxpd     xmm0, xmm0}
     */
    // Template#: 2330, Serial#: 4513
    public void maxpd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x5F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxpd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code maxpd     xmm0, [L1: 0x12345678]}
     */
    // Template#: 2331, Serial#: 4508
    public void m_maxpd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x5F, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxpd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code maxpd     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2332, Serial#: 4512
    public void maxpd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x5F, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxpd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code maxpd     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2333, Serial#: 4507
    public void m_maxpd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x5F, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxpd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code maxpd     xmm0, [ebx + 305419896]}
     */
    // Template#: 2334, Serial#: 4511
    public void maxpd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x5F, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxps  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code maxps     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2335, Serial#: 4438
    public void maxps(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0102((byte) 0x5F, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxps  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code maxps     xmm0, [ebx + 18]}
     */
    // Template#: 2336, Serial#: 4437
    public void maxps(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0103((byte) 0x5F, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxps  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code maxps     xmm0, ebx[esi * 4]}
     */
    // Template#: 2337, Serial#: 4434
    public void maxps(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0104((byte) 0x5F, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code maxps     xmm0, [ebx]}
     */
    // Template#: 2338, Serial#: 4433
    public void maxps(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0105((byte) 0x5F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code maxps     xmm0, xmm0}
     */
    // Template#: 2339, Serial#: 4441
    public void maxps(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0106((byte) 0x5F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxps  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code maxps     xmm0, [L1: 0x12345678]}
     */
    // Template#: 2340, Serial#: 4436
    public void m_maxps(IA32XMMRegister destination, int m32) {
        assemble0107((byte) 0x5F, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxps  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code maxps     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2341, Serial#: 4440
    public void maxps(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0108((byte) 0x5F, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxps  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code maxps     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2342, Serial#: 4435
    public void m_maxps(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0109((byte) 0x5F, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxps  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code maxps     xmm0, [ebx + 305419896]}
     */
    // Template#: 2343, Serial#: 4439
    public void maxps(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0110((byte) 0x5F, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxsd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code maxsd     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2344, Serial#: 4573
    public void maxsd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0111((byte) 0x5F, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxsd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code maxsd     xmm0, [ebx + 18]}
     */
    // Template#: 2345, Serial#: 4572
    public void maxsd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0112((byte) 0x5F, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxsd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code maxsd     xmm0, ebx[esi * 4]}
     */
    // Template#: 2346, Serial#: 4569
    public void maxsd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0113((byte) 0x5F, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxsd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code maxsd     xmm0, [ebx]}
     */
    // Template#: 2347, Serial#: 4568
    public void maxsd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0114((byte) 0x5F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxsd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code maxsd     xmm0, xmm0}
     */
    // Template#: 2348, Serial#: 4576
    public void maxsd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0115((byte) 0x5F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxsd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code maxsd     xmm0, [L1: 0x12345678]}
     */
    // Template#: 2349, Serial#: 4571
    public void m_maxsd(IA32XMMRegister destination, int m32) {
        assemble0116((byte) 0x5F, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxsd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code maxsd     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2350, Serial#: 4575
    public void maxsd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0117((byte) 0x5F, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxsd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code maxsd     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2351, Serial#: 4570
    public void m_maxsd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0118((byte) 0x5F, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxsd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code maxsd     xmm0, [ebx + 305419896]}
     */
    // Template#: 2352, Serial#: 4574
    public void maxsd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0119((byte) 0x5F, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxss  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code maxss     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2353, Serial#: 4645
    public void maxss(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0120((byte) 0x5F, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxss  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code maxss     xmm0, [ebx + 18]}
     */
    // Template#: 2354, Serial#: 4644
    public void maxss(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0121((byte) 0x5F, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxss  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code maxss     xmm0, ebx[esi * 4]}
     */
    // Template#: 2355, Serial#: 4641
    public void maxss(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0122((byte) 0x5F, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code maxss     xmm0, [ebx]}
     */
    // Template#: 2356, Serial#: 4640
    public void maxss(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0123((byte) 0x5F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code maxss     xmm0, xmm0}
     */
    // Template#: 2357, Serial#: 4648
    public void maxss(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0124((byte) 0x5F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxss  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code maxss     xmm0, [L1: 0x12345678]}
     */
    // Template#: 2358, Serial#: 4643
    public void m_maxss(IA32XMMRegister destination, int m32) {
        assemble0125((byte) 0x5F, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxss  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code maxss     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2359, Serial#: 4647
    public void maxss(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0126((byte) 0x5F, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxss  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code maxss     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2360, Serial#: 4642
    public void m_maxss(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0127((byte) 0x5F, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code maxss  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code maxss     xmm0, [ebx + 305419896]}
     */
    // Template#: 2361, Serial#: 4646
    public void maxss(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0128((byte) 0x5F, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mfence  }
     * Example disassembly syntax: {@code mfence  }
     */
    // Template#: 2362, Serial#: 5008
    public void mfence() {
        assemble0357((byte) 0xAE, (byte) 0x06);
    }

    /**
     * Pseudo-external assembler syntax: {@code minpd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code minpd     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2363, Serial#: 4492
    public void minpd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x5D, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code minpd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code minpd     xmm0, [ebx + 18]}
     */
    // Template#: 2364, Serial#: 4491
    public void minpd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x5D, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code minpd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code minpd     xmm0, ebx[esi * 4]}
     */
    // Template#: 2365, Serial#: 4488
    public void minpd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x5D, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code minpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code minpd     xmm0, [ebx]}
     */
    // Template#: 2366, Serial#: 4487
    public void minpd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x5D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code minpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code minpd     xmm0, xmm0}
     */
    // Template#: 2367, Serial#: 4495
    public void minpd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x5D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code minpd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code minpd     xmm0, [L1: 0x12345678]}
     */
    // Template#: 2368, Serial#: 4490
    public void m_minpd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x5D, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code minpd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code minpd     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2369, Serial#: 4494
    public void minpd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x5D, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code minpd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code minpd     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2370, Serial#: 4489
    public void m_minpd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x5D, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code minpd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code minpd     xmm0, [ebx + 305419896]}
     */
    // Template#: 2371, Serial#: 4493
    public void minpd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x5D, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code minps  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code minps     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2372, Serial#: 4420
    public void minps(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0102((byte) 0x5D, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code minps  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code minps     xmm0, [ebx + 18]}
     */
    // Template#: 2373, Serial#: 4419
    public void minps(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0103((byte) 0x5D, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code minps  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code minps     xmm0, ebx[esi * 4]}
     */
    // Template#: 2374, Serial#: 4416
    public void minps(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0104((byte) 0x5D, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code minps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code minps     xmm0, [ebx]}
     */
    // Template#: 2375, Serial#: 4415
    public void minps(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0105((byte) 0x5D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code minps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code minps     xmm0, xmm0}
     */
    // Template#: 2376, Serial#: 4423
    public void minps(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0106((byte) 0x5D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code minps  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code minps     xmm0, [L1: 0x12345678]}
     */
    // Template#: 2377, Serial#: 4418
    public void m_minps(IA32XMMRegister destination, int m32) {
        assemble0107((byte) 0x5D, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code minps  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code minps     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2378, Serial#: 4422
    public void minps(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0108((byte) 0x5D, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code minps  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code minps     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2379, Serial#: 4417
    public void m_minps(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0109((byte) 0x5D, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code minps  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code minps     xmm0, [ebx + 305419896]}
     */
    // Template#: 2380, Serial#: 4421
    public void minps(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0110((byte) 0x5D, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code minsd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code minsd     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2381, Serial#: 4555
    public void minsd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0111((byte) 0x5D, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code minsd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code minsd     xmm0, [ebx + 18]}
     */
    // Template#: 2382, Serial#: 4554
    public void minsd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0112((byte) 0x5D, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code minsd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code minsd     xmm0, ebx[esi * 4]}
     */
    // Template#: 2383, Serial#: 4551
    public void minsd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0113((byte) 0x5D, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code minsd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code minsd     xmm0, [ebx]}
     */
    // Template#: 2384, Serial#: 4550
    public void minsd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0114((byte) 0x5D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code minsd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code minsd     xmm0, xmm0}
     */
    // Template#: 2385, Serial#: 4558
    public void minsd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0115((byte) 0x5D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code minsd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code minsd     xmm0, [L1: 0x12345678]}
     */
    // Template#: 2386, Serial#: 4553
    public void m_minsd(IA32XMMRegister destination, int m32) {
        assemble0116((byte) 0x5D, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code minsd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code minsd     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2387, Serial#: 4557
    public void minsd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0117((byte) 0x5D, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code minsd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code minsd     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2388, Serial#: 4552
    public void m_minsd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0118((byte) 0x5D, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code minsd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code minsd     xmm0, [ebx + 305419896]}
     */
    // Template#: 2389, Serial#: 4556
    public void minsd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0119((byte) 0x5D, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code minss  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code minss     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2390, Serial#: 4627
    public void minss(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0120((byte) 0x5D, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code minss  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code minss     xmm0, [ebx + 18]}
     */
    // Template#: 2391, Serial#: 4626
    public void minss(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0121((byte) 0x5D, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code minss  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code minss     xmm0, ebx[esi * 4]}
     */
    // Template#: 2392, Serial#: 4623
    public void minss(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0122((byte) 0x5D, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code minss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code minss     xmm0, [ebx]}
     */
    // Template#: 2393, Serial#: 4622
    public void minss(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0123((byte) 0x5D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code minss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code minss     xmm0, xmm0}
     */
    // Template#: 2394, Serial#: 4630
    public void minss(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0124((byte) 0x5D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code minss  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code minss     xmm0, [L1: 0x12345678]}
     */
    // Template#: 2395, Serial#: 4625
    public void m_minss(IA32XMMRegister destination, int m32) {
        assemble0125((byte) 0x5D, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code minss  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code minss     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2396, Serial#: 4629
    public void minss(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0126((byte) 0x5D, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code minss  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code minss     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2397, Serial#: 4624
    public void m_minss(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0127((byte) 0x5D, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code minss  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code minss     xmm0, [ebx + 305419896]}
     */
    // Template#: 2398, Serial#: 4628
    public void minss(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0128((byte) 0x5D, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code movb      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 2399, Serial#: 928
    public void movb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0359((byte) 0xC6, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       ebx[esi * 4 + 18], ax}
     */
    // Template#: 2400, Serial#: 1839
    public void mov(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0005((byte) 0x89, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       ebx[esi * 4 + 18], eax}
     */
    // Template#: 2401, Serial#: 1830
    public void mov(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0006((byte) 0x89, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       ebx[esi * 4 + 18], al}
     */
    // Template#: 2402, Serial#: 1821
    public void mov(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0007((byte) 0x88, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       ebx[esi * 4 + 18], es}
     */
    // Template#: 2403, Serial#: 1872
    public void mov(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, SegmentRegister source) {
        assemble0360((byte) 0x8C, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code movl      ebx[esi * 4 + 18], 0x12345678}
     */
    // Template#: 2404, Serial#: 937
    public void movl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0361((byte) 0xC7, disp8, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code movw      ebx[esi * 4 + 18], 0x1234}
     */
    // Template#: 2405, Serial#: 946
    public void movw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0362((byte) 0xC7, disp8, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code movb  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code movb      [ebx + 18], 0x12}
     */
    // Template#: 2406, Serial#: 927
    public void movb(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0363((byte) 0xC6, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       [ebx + 18], ax}
     */
    // Template#: 2407, Serial#: 1838
    public void mov(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0012((byte) 0x89, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       [ebx + 18], eax}
     */
    // Template#: 2408, Serial#: 1829
    public void mov(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0013((byte) 0x89, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       [ebx + 18], al}
     */
    // Template#: 2409, Serial#: 1820
    public void mov(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0014((byte) 0x88, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       [ebx + 18], es}
     */
    // Template#: 2410, Serial#: 1871
    public void mov(byte disp8, IA32IndirectRegister32 destination, SegmentRegister source) {
        assemble0364((byte) 0x8C, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movl  }<i>disp8</i>, <i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code movl      [ebx + 18], 0x12345678}
     */
    // Template#: 2411, Serial#: 936
    public void movl(byte disp8, IA32IndirectRegister32 destination, int imm32) {
        assemble0365((byte) 0xC7, disp8, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movw  }<i>disp8</i>, <i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code movw      [ebx + 18], 0x1234}
     */
    // Template#: 2412, Serial#: 945
    public void movw(byte disp8, IA32IndirectRegister32 destination, short imm16) {
        assemble0366((byte) 0xC7, disp8, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code movb  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code movb      ebx[esi * 4], 0x12}
     */
    // Template#: 2413, Serial#: 924
    public void movb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0367((byte) 0xC6, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       ebx[esi * 4], ax}
     */
    // Template#: 2414, Serial#: 1835
    public void mov(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0019((byte) 0x89, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       ebx[esi * 4], eax}
     */
    // Template#: 2415, Serial#: 1826
    public void mov(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0020((byte) 0x89, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       ebx[esi * 4], al}
     */
    // Template#: 2416, Serial#: 1817
    public void mov(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0021((byte) 0x88, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       ebx[esi * 4], es}
     */
    // Template#: 2417, Serial#: 1868
    public void mov(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, SegmentRegister source) {
        assemble0368((byte) 0x8C, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movl  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code movl      ebx[esi * 4], 0x12345678}
     */
    // Template#: 2418, Serial#: 933
    public void movl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0369((byte) 0xC7, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code movw      ebx[esi * 4], 0x1234}
     */
    // Template#: 2419, Serial#: 942
    public void movw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0370((byte) 0xC7, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mov       ax, ebx[esi * 4 + 18]}
     */
    // Template#: 2420, Serial#: 1864
    public void mov(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0025((byte) 0x8B, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       ax, [ebx + 18]}
     */
    // Template#: 2421, Serial#: 1863
    public void mov(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0026((byte) 0x8B, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mov       ax, ebx[esi * 4]}
     */
    // Template#: 2422, Serial#: 1860
    public void mov(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0027((byte) 0x8B, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       ax, ax}
     */
    // Template#: 2423, Serial#: 1842
    public void mov(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0028((byte) 0x89, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       ax, [ebx]}
     */
    // Template#: 2424, Serial#: 1859
    public void mov(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0029((byte) 0x8B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       ax, es}
     */
    // Template#: 2425, Serial#: 1875
    public void mov(IA32GeneralRegister16 destination, SegmentRegister source) {
        assemble0371((byte) 0x8C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code mov       ax, [L1: 0x12345678]}
     */
    // Template#: 2426, Serial#: 1862
    public void m_mov(IA32GeneralRegister16 destination, int m32) {
        assemble0030((byte) 0x8B, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mov       ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2427, Serial#: 1866
    public void mov(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0031((byte) 0x8B, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mov       ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2428, Serial#: 1861
    public void m_mov(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0032((byte) 0x8B, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       ax, [ebx + 305419896]}
     */
    // Template#: 2429, Serial#: 1865
    public void mov(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0033((byte) 0x8B, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movw  }<i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code movw      ax, 0x1234}
     */
    // Template#: 2430, Serial#: 949
    public void movw(IA32GeneralRegister16 destination, short imm16) {
        assemble0372((byte) 0xC7, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code movw  }<i>register</i>, <i>imm16</i>
     * Example disassembly syntax: {@code movw      ax, 0x1234}
     */
    // Template#: 2431, Serial#: 1934
    public void mov(IA32GeneralRegister16 register, short imm16) {
        assemble0373((byte) 0xB8, register, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mov       eax, ebx[esi * 4 + 18]}
     */
    // Template#: 2432, Serial#: 1856
    public void mov(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0036((byte) 0x8B, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       eax, [ebx + 18]}
     */
    // Template#: 2433, Serial#: 1855
    public void mov(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0037((byte) 0x8B, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mov       eax, ebx[esi * 4]}
     */
    // Template#: 2434, Serial#: 1852
    public void mov(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0038((byte) 0x8B, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       eax, eax}
     */
    // Template#: 2435, Serial#: 1833
    public void mov(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0039((byte) 0x89, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       eax, [ebx]}
     */
    // Template#: 2436, Serial#: 1851
    public void mov(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0040((byte) 0x8B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       eax, cr0}
     */
    // Template#: 2437, Serial#: 2829
    public void mov(IA32GeneralRegister32 destination, ControlRegister source) {
        assemble0374((byte) 0x20, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       eax, dr0}
     */
    // Template#: 2438, Serial#: 2830
    public void mov(IA32GeneralRegister32 destination, DebugRegister source) {
        assemble0375((byte) 0x21, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movl  }<i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code movl      eax, 0x12345678}
     */
    // Template#: 2439, Serial#: 940
    public void movl(IA32GeneralRegister32 destination, int imm32) {
        assemble0376((byte) 0xC7, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code mov       eax, [L1: 0x12345678]}
     */
    // Template#: 2440, Serial#: 1854
    public void m_mov(IA32GeneralRegister32 destination, int m32) {
        assemble0041((byte) 0x8B, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movl  }<i>register</i>, <i>imm32</i>
     * Example disassembly syntax: {@code movl      eax, 0x12345678}
     */
    // Template#: 2441, Serial#: 1933
    public void mov(IA32GeneralRegister32 register, int imm32) {
        assemble0377((byte) 0xB8, register, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mov       eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2442, Serial#: 1858
    public void mov(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0043((byte) 0x8B, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mov       eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2443, Serial#: 1853
    public void m_mov(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0044((byte) 0x8B, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       eax, [ebx + 305419896]}
     */
    // Template#: 2444, Serial#: 1857
    public void mov(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0045((byte) 0x8B, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>register</i>, <i>imm8</i>
     * Example disassembly syntax: {@code mov       al, 0x12}
     */
    // Template#: 2445, Serial#: 699
    public void mov(IA32GeneralRegister8 register, byte imm8) {
        assemble0378((byte) 0xB0, register, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code movb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code movb      al, 0x12}
     */
    // Template#: 2446, Serial#: 931
    public void movb(IA32GeneralRegister8 destination, byte imm8) {
        assemble0379((byte) 0xC6, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mov       al, ebx[esi * 4 + 18]}
     */
    // Template#: 2447, Serial#: 1848
    public void mov(IA32GeneralRegister8 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0047((byte) 0x8A, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       al, [ebx + 18]}
     */
    // Template#: 2448, Serial#: 1847
    public void mov(IA32GeneralRegister8 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0048((byte) 0x8A, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mov       al, ebx[esi * 4]}
     */
    // Template#: 2449, Serial#: 1844
    public void mov(IA32GeneralRegister8 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0049((byte) 0x8A, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       al, al}
     */
    // Template#: 2450, Serial#: 1824
    public void mov(IA32GeneralRegister8 destination, IA32GeneralRegister8 source) {
        assemble0050((byte) 0x88, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       al, [ebx]}
     */
    // Template#: 2451, Serial#: 1843
    public void mov(IA32GeneralRegister8 destination, IA32IndirectRegister32 source) {
        assemble0051((byte) 0x8A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code mov       al, [L1: 0x12345678]}
     */
    // Template#: 2452, Serial#: 1846
    public void m_mov(IA32GeneralRegister8 destination, int m32) {
        assemble0052((byte) 0x8A, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mov       al, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2453, Serial#: 1850
    public void mov(IA32GeneralRegister8 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0053((byte) 0x8A, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mov       al, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2454, Serial#: 1845
    public void m_mov(IA32GeneralRegister8 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0054((byte) 0x8A, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       al, [ebx + 305419896]}
     */
    // Template#: 2455, Serial#: 1849
    public void mov(IA32GeneralRegister8 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0055((byte) 0x8A, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code movb      [ebx], 0x12}
     */
    // Template#: 2456, Serial#: 923
    public void movb(IA32IndirectRegister32 destination, byte imm8) {
        assemble0380((byte) 0xC6, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       [ebx], ax}
     */
    // Template#: 2457, Serial#: 1834
    public void mov(IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0058((byte) 0x89, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       [ebx], eax}
     */
    // Template#: 2458, Serial#: 1825
    public void mov(IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0059((byte) 0x89, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       [ebx], al}
     */
    // Template#: 2459, Serial#: 1816
    public void mov(IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0060((byte) 0x88, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       [ebx], es}
     */
    // Template#: 2460, Serial#: 1867
    public void mov(IA32IndirectRegister32 destination, SegmentRegister source) {
        assemble0381((byte) 0x8C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movl  }<i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code movl      [ebx], 0x12345678}
     */
    // Template#: 2461, Serial#: 932
    public void movl(IA32IndirectRegister32 destination, int imm32) {
        assemble0382((byte) 0xC7, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movw  }<i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code movw      [ebx], 0x1234}
     */
    // Template#: 2462, Serial#: 941
    public void movw(IA32IndirectRegister32 destination, short imm16) {
        assemble0383((byte) 0xC7, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       cr0, eax}
     */
    // Template#: 2463, Serial#: 2831
    public void mov(ControlRegister destination, IA32GeneralRegister32 source) {
        assemble0384((byte) 0x22, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       dr0, eax}
     */
    // Template#: 2464, Serial#: 2832
    public void mov(DebugRegister destination, IA32GeneralRegister32 source) {
        assemble0385((byte) 0x23, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mov       es, ebx[esi * 4 + 18]}
     */
    // Template#: 2465, Serial#: 1897
    public void mov(SegmentRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0386((byte) 0x8E, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       es, [ebx + 18]}
     */
    // Template#: 2466, Serial#: 1896
    public void mov(SegmentRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0387((byte) 0x8E, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mov       es, ebx[esi * 4]}
     */
    // Template#: 2467, Serial#: 1893
    public void mov(SegmentRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0388((byte) 0x8E, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       es, ax}
     */
    // Template#: 2468, Serial#: 1900
    public void mov(SegmentRegister destination, IA32GeneralRegister16 source) {
        assemble0389((byte) 0x8E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       es, [ebx]}
     */
    // Template#: 2469, Serial#: 1892
    public void mov(SegmentRegister destination, IA32IndirectRegister32 source) {
        assemble0390((byte) 0x8E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code mov       es, [L1: 0x12345678]}
     */
    // Template#: 2470, Serial#: 1895
    public void m_mov(SegmentRegister destination, int m32) {
        assemble0391((byte) 0x8E, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mov       es, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2471, Serial#: 1899
    public void mov(SegmentRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0392((byte) 0x8E, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mov       es, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2472, Serial#: 1894
    public void m_mov(SegmentRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0393((byte) 0x8E, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       es, [ebx + 305419896]}
     */
    // Template#: 2473, Serial#: 1898
    public void mov(SegmentRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0394((byte) 0x8E, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>m32</i>
     * Example disassembly syntax: {@code mov       al, [L1: 0x12345678]}
     */
    // Template#: 2474, Serial#: 687
    public void m_mov_AL(int m32) {
        assemble0191((byte) 0xA0, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>m32</i>
     * Example disassembly syntax: {@code mov       eax, [L1: 0x12345678]}
     */
    // Template#: 2475, Serial#: 688
    public void m_mov_EAX(int m32) {
        assemble0191((byte) 0xA1, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>m32</i>
     * Example disassembly syntax: {@code mov       ax, [L1: 0x12345678]}
     */
    // Template#: 2476, Serial#: 689
    public void m_mov_AX(int m32) {
        assemble0395((byte) 0xA1, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>m32</i>
     * Example disassembly syntax: {@code mov       [L1: 0x12345678], al}
     */
    // Template#: 2477, Serial#: 690
    public void m_mov___AL(int m32) {
        assemble0191((byte) 0xA2, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>m32</i>
     * Example disassembly syntax: {@code mov       [L1: 0x12345678], eax}
     */
    // Template#: 2478, Serial#: 691
    public void m_mov___EAX(int m32) {
        assemble0191((byte) 0xA3, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>m32</i>
     * Example disassembly syntax: {@code mov       [L1: 0x12345678], ax}
     */
    // Template#: 2479, Serial#: 692
    public void m_mov___AX(int m32) {
        assemble0395((byte) 0xA3, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movb  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code movb      [L1: 0x12345678], 0x12}
     */
    // Template#: 2480, Serial#: 926
    public void m_movb(int m32, byte imm8) {
        assemble0396((byte) 0xC6, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code movb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code movb      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 2481, Serial#: 930
    public void movb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0397((byte) 0xC6, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       ebx[esi * 4 + 305419896], ax}
     */
    // Template#: 2482, Serial#: 1841
    public void mov(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0068((byte) 0x89, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       ebx[esi * 4 + 305419896], eax}
     */
    // Template#: 2483, Serial#: 1832
    public void mov(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0069((byte) 0x89, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       ebx[esi * 4 + 305419896], al}
     */
    // Template#: 2484, Serial#: 1823
    public void mov(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0070((byte) 0x88, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       ebx[esi * 4 + 305419896], es}
     */
    // Template#: 2485, Serial#: 1874
    public void mov(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, SegmentRegister source) {
        assemble0398((byte) 0x8C, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code movl      ebx[esi * 4 + 305419896], 0x12345678}
     */
    // Template#: 2486, Serial#: 939
    public void movl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0399((byte) 0xC7, disp32, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code movw      ebx[esi * 4 + 305419896], 0x1234}
     */
    // Template#: 2487, Serial#: 948
    public void movw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0400((byte) 0xC7, disp32, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       [L1: 0x12345678], ax}
     */
    // Template#: 2488, Serial#: 1837
    public void m_mov(int m32, IA32GeneralRegister16 source) {
        assemble0073((byte) 0x89, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       [L1: 0x12345678], eax}
     */
    // Template#: 2489, Serial#: 1828
    public void m_mov(int m32, IA32GeneralRegister32 source) {
        assemble0074((byte) 0x89, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       [L1: 0x12345678], al}
     */
    // Template#: 2490, Serial#: 1819
    public void m_mov(int m32, IA32GeneralRegister8 source) {
        assemble0075((byte) 0x88, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movb  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code movb      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 2491, Serial#: 925
    public void m_movb(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0401((byte) 0xC6, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       L1: 0x12345678[esi * 4], ax}
     */
    // Template#: 2492, Serial#: 1836
    public void m_mov(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0078((byte) 0x89, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       L1: 0x12345678[esi * 4], eax}
     */
    // Template#: 2493, Serial#: 1827
    public void m_mov(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0079((byte) 0x89, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       L1: 0x12345678[esi * 4], al}
     */
    // Template#: 2494, Serial#: 1818
    public void m_mov(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0080((byte) 0x88, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       L1: 0x12345678[esi * 4], es}
     */
    // Template#: 2495, Serial#: 1869
    public void m_mov(int m32, IA32IndexRegister32 index, Scale scale, SegmentRegister source) {
        assemble0402((byte) 0x8C, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movl  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code movl      L1: 0x12345678[esi * 4], 0x12345678}
     */
    // Template#: 2496, Serial#: 934
    public void m_movl(int m32, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0403((byte) 0xC7, m32, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code movw      L1: 0x12345678[esi * 4], 0x1234}
     */
    // Template#: 2497, Serial#: 943
    public void m_movw(int m32, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0404((byte) 0xC7, m32, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code movb  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code movb      [ebx + 305419896], 0x12}
     */
    // Template#: 2498, Serial#: 929
    public void movb(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0405((byte) 0xC6, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       [ebx + 305419896], ax}
     */
    // Template#: 2499, Serial#: 1840
    public void mov(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0085((byte) 0x89, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       [ebx + 305419896], eax}
     */
    // Template#: 2500, Serial#: 1831
    public void mov(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0086((byte) 0x89, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       [ebx + 305419896], al}
     */
    // Template#: 2501, Serial#: 1822
    public void mov(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0087((byte) 0x88, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       [ebx + 305419896], es}
     */
    // Template#: 2502, Serial#: 1873
    public void mov(int disp32, IA32IndirectRegister32 destination, SegmentRegister source) {
        assemble0406((byte) 0x8C, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movl  }<i>disp32</i>, <i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code movl      [ebx + 305419896], 0x12345678}
     */
    // Template#: 2503, Serial#: 938
    public void movl(int disp32, IA32IndirectRegister32 destination, int imm32) {
        assemble0407((byte) 0xC7, disp32, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movw  }<i>disp32</i>, <i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code movw      [ebx + 305419896], 0x1234}
     */
    // Template#: 2504, Serial#: 947
    public void movw(int disp32, IA32IndirectRegister32 destination, short imm16) {
        assemble0408((byte) 0xC7, disp32, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code mov  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code mov       [L1: 0x12345678], es}
     */
    // Template#: 2505, Serial#: 1870
    public void m_mov(int m32, SegmentRegister source) {
        assemble0409((byte) 0x8C, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movl  }<i>m32</i>, <i>imm32</i>
     * Example disassembly syntax: {@code movl      [L1: 0x12345678], 0x12345678}
     */
    // Template#: 2506, Serial#: 935
    public void m_movl(int m32, int imm32) {
        assemble0410((byte) 0xC7, m32, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movw  }<i>m32</i>, <i>imm16</i>
     * Example disassembly syntax: {@code movw      [L1: 0x12345678], 0x1234}
     */
    // Template#: 2507, Serial#: 944
    public void m_movw(int m32, short imm16) {
        assemble0411((byte) 0xC7, m32, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code movapd  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movapd    ebx[esi * 4 + 18], xmm0}
     */
    // Template#: 2508, Serial#: 4116
    public void movapd(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0412((byte) 0x29, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movapd  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movapd    [ebx + 18], xmm0}
     */
    // Template#: 2509, Serial#: 4115
    public void movapd(byte disp8, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0413((byte) 0x29, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movapd  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movapd    ebx[esi * 4], xmm0}
     */
    // Template#: 2510, Serial#: 4112
    public void movapd(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0414((byte) 0x29, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movapd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movapd    [ebx], xmm0}
     */
    // Template#: 2511, Serial#: 4111
    public void movapd(IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0415((byte) 0x29, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movapd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movapd    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2512, Serial#: 4107
    public void movapd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x28, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movapd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movapd    xmm0, [ebx + 18]}
     */
    // Template#: 2513, Serial#: 4106
    public void movapd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x28, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movapd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movapd    xmm0, ebx[esi * 4]}
     */
    // Template#: 2514, Serial#: 4103
    public void movapd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x28, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movapd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movapd    xmm0, [ebx]}
     */
    // Template#: 2515, Serial#: 4102
    public void movapd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x28, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movapd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movapd    xmm0, xmm0}
     */
    // Template#: 2516, Serial#: 4110
    public void movapd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x28, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movapd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movapd    xmm0, [L1: 0x12345678]}
     */
    // Template#: 2517, Serial#: 4105
    public void m_movapd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x28, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movapd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movapd    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2518, Serial#: 4109
    public void movapd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x28, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movapd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movapd    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2519, Serial#: 4104
    public void m_movapd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x28, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movapd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movapd    xmm0, [ebx + 305419896]}
     */
    // Template#: 2520, Serial#: 4108
    public void movapd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x28, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movapd  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movapd    ebx[esi * 4 + 305419896], xmm0}
     */
    // Template#: 2521, Serial#: 4118
    public void movapd(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0416((byte) 0x29, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movapd  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movapd    L1: 0x12345678[esi * 4], xmm0}
     */
    // Template#: 2522, Serial#: 4113
    public void m_movapd(int m32, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0417((byte) 0x29, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movapd  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movapd    [ebx + 305419896], xmm0}
     */
    // Template#: 2523, Serial#: 4117
    public void movapd(int disp32, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0418((byte) 0x29, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movapd  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code movapd    [L1: 0x12345678], xmm0}
     */
    // Template#: 2524, Serial#: 4114
    public void m_movapd(int m32, IA32XMMRegister source) {
        assemble0419((byte) 0x29, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movaps  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movaps    ebx[esi * 4 + 18], xmm0}
     */
    // Template#: 2525, Serial#: 4046
    public void movaps(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0420((byte) 0x29, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movaps  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movaps    [ebx + 18], xmm0}
     */
    // Template#: 2526, Serial#: 4045
    public void movaps(byte disp8, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0421((byte) 0x29, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movaps  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movaps    ebx[esi * 4], xmm0}
     */
    // Template#: 2527, Serial#: 4042
    public void movaps(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0422((byte) 0x29, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movaps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movaps    [ebx], xmm0}
     */
    // Template#: 2528, Serial#: 4041
    public void movaps(IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0423((byte) 0x29, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movaps  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movaps    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2529, Serial#: 4037
    public void movaps(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0102((byte) 0x28, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movaps  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movaps    xmm0, [ebx + 18]}
     */
    // Template#: 2530, Serial#: 4036
    public void movaps(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0103((byte) 0x28, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movaps  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movaps    xmm0, ebx[esi * 4]}
     */
    // Template#: 2531, Serial#: 4033
    public void movaps(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0104((byte) 0x28, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movaps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movaps    xmm0, [ebx]}
     */
    // Template#: 2532, Serial#: 4032
    public void movaps(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0105((byte) 0x28, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movaps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movaps    xmm0, xmm0}
     */
    // Template#: 2533, Serial#: 4040
    public void movaps(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0106((byte) 0x28, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movaps  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movaps    xmm0, [L1: 0x12345678]}
     */
    // Template#: 2534, Serial#: 4035
    public void m_movaps(IA32XMMRegister destination, int m32) {
        assemble0107((byte) 0x28, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movaps  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movaps    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2535, Serial#: 4039
    public void movaps(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0108((byte) 0x28, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movaps  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movaps    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2536, Serial#: 4034
    public void m_movaps(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0109((byte) 0x28, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movaps  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movaps    xmm0, [ebx + 305419896]}
     */
    // Template#: 2537, Serial#: 4038
    public void movaps(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0110((byte) 0x28, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movaps  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movaps    ebx[esi * 4 + 305419896], xmm0}
     */
    // Template#: 2538, Serial#: 4048
    public void movaps(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0424((byte) 0x29, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movaps  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movaps    L1: 0x12345678[esi * 4], xmm0}
     */
    // Template#: 2539, Serial#: 4043
    public void m_movaps(int m32, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0425((byte) 0x29, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movaps  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movaps    [ebx + 305419896], xmm0}
     */
    // Template#: 2540, Serial#: 4047
    public void movaps(int disp32, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0426((byte) 0x29, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movaps  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code movaps    [L1: 0x12345678], xmm0}
     */
    // Template#: 2541, Serial#: 4044
    public void m_movaps(int m32, IA32XMMRegister source) {
        assemble0427((byte) 0x29, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      ebx[esi * 4 + 18], xmm0}
     */
    // Template#: 2542, Serial#: 4824
    public void movd(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0412((byte) 0x7E, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      ebx[esi * 4 + 18], mm0}
     */
    // Template#: 2543, Serial#: 4789
    public void movd(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, MMXRegister source) {
        assemble0428((byte) 0x7E, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      [ebx + 18], xmm0}
     */
    // Template#: 2544, Serial#: 4823
    public void movd(byte disp8, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0413((byte) 0x7E, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      [ebx + 18], mm0}
     */
    // Template#: 2545, Serial#: 4788
    public void movd(byte disp8, IA32IndirectRegister32 destination, MMXRegister source) {
        assemble0429((byte) 0x7E, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      ebx[esi * 4], xmm0}
     */
    // Template#: 2546, Serial#: 4820
    public void movd(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0414((byte) 0x7E, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      ebx[esi * 4], mm0}
     */
    // Template#: 2547, Serial#: 4785
    public void movd(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, MMXRegister source) {
        assemble0430((byte) 0x7E, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      eax, xmm0}
     */
    // Template#: 2548, Serial#: 4827
    public void movd(IA32GeneralRegister32 destination, IA32XMMRegister source) {
        assemble0431((byte) 0x7E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      eax, mm0}
     */
    // Template#: 2549, Serial#: 4792
    public void movd(IA32GeneralRegister32 destination, MMXRegister source) {
        assemble0432((byte) 0x7E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      [ebx], xmm0}
     */
    // Template#: 2550, Serial#: 4819
    public void movd(IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0415((byte) 0x7E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      [ebx], mm0}
     */
    // Template#: 2551, Serial#: 4784
    public void movd(IA32IndirectRegister32 destination, MMXRegister source) {
        assemble0433((byte) 0x7E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movd      xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2552, Serial#: 4762
    public void movd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x6E, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      xmm0, [ebx + 18]}
     */
    // Template#: 2553, Serial#: 4761
    public void movd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x6E, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movd      xmm0, ebx[esi * 4]}
     */
    // Template#: 2554, Serial#: 4758
    public void movd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x6E, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      xmm0, eax}
     */
    // Template#: 2555, Serial#: 4765
    public void movd(IA32XMMRegister destination, IA32GeneralRegister32 source) {
        assemble0434((byte) 0x6E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      xmm0, [ebx]}
     */
    // Template#: 2556, Serial#: 4757
    public void movd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x6E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movd      xmm0, [L1: 0x12345678]}
     */
    // Template#: 2557, Serial#: 4760
    public void m_movd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x6E, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movd      xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2558, Serial#: 4764
    public void movd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x6E, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movd      xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2559, Serial#: 4759
    public void m_movd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x6E, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      xmm0, [ebx + 305419896]}
     */
    // Template#: 2560, Serial#: 4763
    public void movd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x6E, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movd      mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2561, Serial#: 4690
    public void movd(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0x6E, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      mm0, [ebx + 18]}
     */
    // Template#: 2562, Serial#: 4689
    public void movd(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0x6E, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movd      mm0, ebx[esi * 4]}
     */
    // Template#: 2563, Serial#: 4686
    public void movd(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0x6E, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      mm0, eax}
     */
    // Template#: 2564, Serial#: 4693
    public void movd(MMXRegister destination, IA32GeneralRegister32 source) {
        assemble0435((byte) 0x6E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      mm0, [ebx]}
     */
    // Template#: 2565, Serial#: 4685
    public void movd(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0x6E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movd      mm0, [L1: 0x12345678]}
     */
    // Template#: 2566, Serial#: 4688
    public void m_movd(MMXRegister destination, int m32) {
        assemble0268((byte) 0x6E, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movd      mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2567, Serial#: 4692
    public void movd(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0x6E, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movd      mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2568, Serial#: 4687
    public void m_movd(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0x6E, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      mm0, [ebx + 305419896]}
     */
    // Template#: 2569, Serial#: 4691
    public void movd(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0x6E, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      ebx[esi * 4 + 305419896], xmm0}
     */
    // Template#: 2570, Serial#: 4826
    public void movd(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0416((byte) 0x7E, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      ebx[esi * 4 + 305419896], mm0}
     */
    // Template#: 2571, Serial#: 4791
    public void movd(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, MMXRegister source) {
        assemble0436((byte) 0x7E, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      L1: 0x12345678[esi * 4], xmm0}
     */
    // Template#: 2572, Serial#: 4821
    public void m_movd(int m32, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0417((byte) 0x7E, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      L1: 0x12345678[esi * 4], mm0}
     */
    // Template#: 2573, Serial#: 4786
    public void m_movd(int m32, IA32IndexRegister32 index, Scale scale, MMXRegister source) {
        assemble0437((byte) 0x7E, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      [ebx + 305419896], xmm0}
     */
    // Template#: 2574, Serial#: 4825
    public void movd(int disp32, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0418((byte) 0x7E, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      [ebx + 305419896], mm0}
     */
    // Template#: 2575, Serial#: 4790
    public void movd(int disp32, IA32IndirectRegister32 destination, MMXRegister source) {
        assemble0438((byte) 0x7E, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      [L1: 0x12345678], xmm0}
     */
    // Template#: 2576, Serial#: 4822
    public void m_movd(int m32, IA32XMMRegister source) {
        assemble0419((byte) 0x7E, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movd  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code movd      [L1: 0x12345678], mm0}
     */
    // Template#: 2577, Serial#: 4787
    public void m_movd(int m32, MMXRegister source) {
        assemble0439((byte) 0x7E, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movddup  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movddup   xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2578, Serial#: 2790
    public void movddup(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0111((byte) 0x12, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movddup  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movddup   xmm0, [ebx + 18]}
     */
    // Template#: 2579, Serial#: 2789
    public void movddup(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0112((byte) 0x12, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movddup  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movddup   xmm0, ebx[esi * 4]}
     */
    // Template#: 2580, Serial#: 2786
    public void movddup(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0113((byte) 0x12, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movddup  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movddup   xmm0, [ebx]}
     */
    // Template#: 2581, Serial#: 2785
    public void movddup(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0114((byte) 0x12, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movddup  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movddup   xmm0, xmm0}
     */
    // Template#: 2582, Serial#: 2793
    public void movddup(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0115((byte) 0x12, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movddup  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movddup   xmm0, [L1: 0x12345678]}
     */
    // Template#: 2583, Serial#: 2788
    public void m_movddup(IA32XMMRegister destination, int m32) {
        assemble0116((byte) 0x12, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movddup  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movddup   xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2584, Serial#: 2792
    public void movddup(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0117((byte) 0x12, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movddup  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movddup   xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2585, Serial#: 2787
    public void m_movddup(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0118((byte) 0x12, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movddup  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movddup   xmm0, [ebx + 305419896]}
     */
    // Template#: 2586, Serial#: 2791
    public void movddup(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0119((byte) 0x12, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdq2q  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movdq2q   mm0, xmm0}
     */
    // Template#: 2587, Serial#: 3812
    public void movdq2q(MMXRegister destination, IA32XMMRegister source) {
        assemble0440((byte) 0xD6, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqa  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movdqa    ebx[esi * 4 + 18], xmm0}
     */
    // Template#: 2588, Serial#: 4833
    public void movdqa(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0412((byte) 0x7F, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqa  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movdqa    [ebx + 18], xmm0}
     */
    // Template#: 2589, Serial#: 4832
    public void movdqa(byte disp8, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0413((byte) 0x7F, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqa  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movdqa    ebx[esi * 4], xmm0}
     */
    // Template#: 2590, Serial#: 4829
    public void movdqa(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0414((byte) 0x7F, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqa  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movdqa    [ebx], xmm0}
     */
    // Template#: 2591, Serial#: 4828
    public void movdqa(IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0415((byte) 0x7F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqa  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movdqa    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2592, Serial#: 4771
    public void movdqa(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x6F, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqa  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movdqa    xmm0, [ebx + 18]}
     */
    // Template#: 2593, Serial#: 4770
    public void movdqa(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x6F, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqa  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movdqa    xmm0, ebx[esi * 4]}
     */
    // Template#: 2594, Serial#: 4767
    public void movdqa(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x6F, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqa  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movdqa    xmm0, [ebx]}
     */
    // Template#: 2595, Serial#: 4766
    public void movdqa(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x6F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqa  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movdqa    xmm0, xmm0}
     */
    // Template#: 2596, Serial#: 4774
    public void movdqa(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x6F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqa  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movdqa    xmm0, [L1: 0x12345678]}
     */
    // Template#: 2597, Serial#: 4769
    public void m_movdqa(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x6F, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqa  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movdqa    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2598, Serial#: 4773
    public void movdqa(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x6F, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqa  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movdqa    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2599, Serial#: 4768
    public void m_movdqa(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x6F, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqa  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movdqa    xmm0, [ebx + 305419896]}
     */
    // Template#: 2600, Serial#: 4772
    public void movdqa(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x6F, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqa  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movdqa    ebx[esi * 4 + 305419896], xmm0}
     */
    // Template#: 2601, Serial#: 4835
    public void movdqa(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0416((byte) 0x7F, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqa  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movdqa    L1: 0x12345678[esi * 4], xmm0}
     */
    // Template#: 2602, Serial#: 4830
    public void m_movdqa(int m32, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0417((byte) 0x7F, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqa  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movdqa    [ebx + 305419896], xmm0}
     */
    // Template#: 2603, Serial#: 4834
    public void movdqa(int disp32, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0418((byte) 0x7F, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqa  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code movdqa    [L1: 0x12345678], xmm0}
     */
    // Template#: 2604, Serial#: 4831
    public void m_movdqa(int m32, IA32XMMRegister source) {
        assemble0419((byte) 0x7F, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqu  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movdqu    ebx[esi * 4 + 18], xmm0}
     */
    // Template#: 2605, Serial#: 4867
    public void movdqu(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0441((byte) 0x7F, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqu  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movdqu    [ebx + 18], xmm0}
     */
    // Template#: 2606, Serial#: 4866
    public void movdqu(byte disp8, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0442((byte) 0x7F, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqu  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movdqu    ebx[esi * 4], xmm0}
     */
    // Template#: 2607, Serial#: 4863
    public void movdqu(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0443((byte) 0x7F, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqu  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movdqu    [ebx], xmm0}
     */
    // Template#: 2608, Serial#: 4862
    public void movdqu(IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0444((byte) 0x7F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqu  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movdqu    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2609, Serial#: 4780
    public void movdqu(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0120((byte) 0x6F, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqu  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movdqu    xmm0, [ebx + 18]}
     */
    // Template#: 2610, Serial#: 4779
    public void movdqu(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0121((byte) 0x6F, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqu  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movdqu    xmm0, ebx[esi * 4]}
     */
    // Template#: 2611, Serial#: 4776
    public void movdqu(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0122((byte) 0x6F, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqu  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movdqu    xmm0, [ebx]}
     */
    // Template#: 2612, Serial#: 4775
    public void movdqu(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0123((byte) 0x6F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqu  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movdqu    xmm0, xmm0}
     */
    // Template#: 2613, Serial#: 4783
    public void movdqu(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0124((byte) 0x6F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqu  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movdqu    xmm0, [L1: 0x12345678]}
     */
    // Template#: 2614, Serial#: 4778
    public void m_movdqu(IA32XMMRegister destination, int m32) {
        assemble0125((byte) 0x6F, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqu  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movdqu    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2615, Serial#: 4782
    public void movdqu(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0126((byte) 0x6F, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqu  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movdqu    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2616, Serial#: 4777
    public void m_movdqu(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0127((byte) 0x6F, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqu  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movdqu    xmm0, [ebx + 305419896]}
     */
    // Template#: 2617, Serial#: 4781
    public void movdqu(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0128((byte) 0x6F, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqu  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movdqu    ebx[esi * 4 + 305419896], xmm0}
     */
    // Template#: 2618, Serial#: 4869
    public void movdqu(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0445((byte) 0x7F, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqu  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movdqu    L1: 0x12345678[esi * 4], xmm0}
     */
    // Template#: 2619, Serial#: 4864
    public void m_movdqu(int m32, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0446((byte) 0x7F, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqu  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movdqu    [ebx + 305419896], xmm0}
     */
    // Template#: 2620, Serial#: 4868
    public void movdqu(int disp32, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0447((byte) 0x7F, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movdqu  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code movdqu    [L1: 0x12345678], xmm0}
     */
    // Template#: 2621, Serial#: 4865
    public void m_movdqu(int m32, IA32XMMRegister source) {
        assemble0448((byte) 0x7F, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhlps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movhlps   xmm0, xmm0}
     */
    // Template#: 2622, Serial#: 2665
    public void movhlps(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0106((byte) 0x12, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhpd  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movhpd    ebx[esi * 4 + 18], xmm0}
     */
    // Template#: 2623, Serial#: 2765
    public void movhpd(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0412((byte) 0x17, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhpd  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movhpd    [ebx + 18], xmm0}
     */
    // Template#: 2624, Serial#: 2764
    public void movhpd(byte disp8, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0413((byte) 0x17, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhpd  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movhpd    ebx[esi * 4], xmm0}
     */
    // Template#: 2625, Serial#: 2761
    public void movhpd(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0414((byte) 0x17, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movhpd    [ebx], xmm0}
     */
    // Template#: 2626, Serial#: 2760
    public void movhpd(IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0415((byte) 0x17, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhpd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movhpd    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2627, Serial#: 2757
    public void movhpd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x16, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhpd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movhpd    xmm0, [ebx + 18]}
     */
    // Template#: 2628, Serial#: 2756
    public void movhpd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x16, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhpd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movhpd    xmm0, ebx[esi * 4]}
     */
    // Template#: 2629, Serial#: 2753
    public void movhpd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x16, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movhpd    xmm0, [ebx]}
     */
    // Template#: 2630, Serial#: 2752
    public void movhpd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x16, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhpd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movhpd    xmm0, [L1: 0x12345678]}
     */
    // Template#: 2631, Serial#: 2755
    public void m_movhpd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x16, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhpd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movhpd    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2632, Serial#: 2759
    public void movhpd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x16, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhpd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movhpd    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2633, Serial#: 2754
    public void m_movhpd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x16, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhpd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movhpd    xmm0, [ebx + 305419896]}
     */
    // Template#: 2634, Serial#: 2758
    public void movhpd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x16, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhpd  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movhpd    ebx[esi * 4 + 305419896], xmm0}
     */
    // Template#: 2635, Serial#: 2767
    public void movhpd(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0416((byte) 0x17, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhpd  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movhpd    L1: 0x12345678[esi * 4], xmm0}
     */
    // Template#: 2636, Serial#: 2762
    public void m_movhpd(int m32, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0417((byte) 0x17, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhpd  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movhpd    [ebx + 305419896], xmm0}
     */
    // Template#: 2637, Serial#: 2766
    public void movhpd(int disp32, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0418((byte) 0x17, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhpd  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code movhpd    [L1: 0x12345678], xmm0}
     */
    // Template#: 2638, Serial#: 2763
    public void m_movhpd(int m32, IA32XMMRegister source) {
        assemble0419((byte) 0x17, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhps  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movhps    ebx[esi * 4 + 18], xmm0}
     */
    // Template#: 2639, Serial#: 2698
    public void movhps(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0420((byte) 0x17, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhps  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movhps    [ebx + 18], xmm0}
     */
    // Template#: 2640, Serial#: 2697
    public void movhps(byte disp8, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0421((byte) 0x17, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhps  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movhps    ebx[esi * 4], xmm0}
     */
    // Template#: 2641, Serial#: 2694
    public void movhps(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0422((byte) 0x17, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movhps    [ebx], xmm0}
     */
    // Template#: 2642, Serial#: 2693
    public void movhps(IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0423((byte) 0x17, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhps  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movhps    ebx[esi * 4 + 305419896], xmm0}
     */
    // Template#: 2643, Serial#: 2700
    public void movhps(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0424((byte) 0x17, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhps  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movhps    L1: 0x12345678[esi * 4], xmm0}
     */
    // Template#: 2644, Serial#: 2695
    public void m_movhps(int m32, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0425((byte) 0x17, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhps  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movhps    [ebx + 305419896], xmm0}
     */
    // Template#: 2645, Serial#: 2699
    public void movhps(int disp32, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0426((byte) 0x17, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movhps  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code movhps    [L1: 0x12345678], xmm0}
     */
    // Template#: 2646, Serial#: 2696
    public void m_movhps(int m32, IA32XMMRegister source) {
        assemble0427((byte) 0x17, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlhps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movlhps   xmm0, xmm0}
     */
    // Template#: 2647, Serial#: 2692
    public void movlhps(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0106((byte) 0x16, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlpd  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movlpd    ebx[esi * 4 + 18], xmm0}
     */
    // Template#: 2648, Serial#: 2731
    public void movlpd(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0412((byte) 0x13, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlpd  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movlpd    [ebx + 18], xmm0}
     */
    // Template#: 2649, Serial#: 2730
    public void movlpd(byte disp8, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0413((byte) 0x13, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlpd  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movlpd    ebx[esi * 4], xmm0}
     */
    // Template#: 2650, Serial#: 2727
    public void movlpd(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0414((byte) 0x13, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movlpd    [ebx], xmm0}
     */
    // Template#: 2651, Serial#: 2726
    public void movlpd(IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0415((byte) 0x13, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlpd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movlpd    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2652, Serial#: 2723
    public void movlpd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x12, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlpd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movlpd    xmm0, [ebx + 18]}
     */
    // Template#: 2653, Serial#: 2722
    public void movlpd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x12, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlpd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movlpd    xmm0, ebx[esi * 4]}
     */
    // Template#: 2654, Serial#: 2719
    public void movlpd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x12, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movlpd    xmm0, [ebx]}
     */
    // Template#: 2655, Serial#: 2718
    public void movlpd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x12, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlpd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movlpd    xmm0, [L1: 0x12345678]}
     */
    // Template#: 2656, Serial#: 2721
    public void m_movlpd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x12, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlpd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movlpd    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2657, Serial#: 2725
    public void movlpd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x12, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlpd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movlpd    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2658, Serial#: 2720
    public void m_movlpd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x12, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlpd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movlpd    xmm0, [ebx + 305419896]}
     */
    // Template#: 2659, Serial#: 2724
    public void movlpd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x12, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlpd  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movlpd    ebx[esi * 4 + 305419896], xmm0}
     */
    // Template#: 2660, Serial#: 2733
    public void movlpd(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0416((byte) 0x13, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlpd  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movlpd    L1: 0x12345678[esi * 4], xmm0}
     */
    // Template#: 2661, Serial#: 2728
    public void m_movlpd(int m32, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0417((byte) 0x13, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlpd  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movlpd    [ebx + 305419896], xmm0}
     */
    // Template#: 2662, Serial#: 2732
    public void movlpd(int disp32, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0418((byte) 0x13, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlpd  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code movlpd    [L1: 0x12345678], xmm0}
     */
    // Template#: 2663, Serial#: 2729
    public void m_movlpd(int m32, IA32XMMRegister source) {
        assemble0419((byte) 0x13, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlps  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movlps    ebx[esi * 4 + 18], xmm0}
     */
    // Template#: 2664, Serial#: 2671
    public void movlps(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0420((byte) 0x13, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlps  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movlps    [ebx + 18], xmm0}
     */
    // Template#: 2665, Serial#: 2670
    public void movlps(byte disp8, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0421((byte) 0x13, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlps  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movlps    ebx[esi * 4], xmm0}
     */
    // Template#: 2666, Serial#: 2667
    public void movlps(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0422((byte) 0x13, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movlps    [ebx], xmm0}
     */
    // Template#: 2667, Serial#: 2666
    public void movlps(IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0423((byte) 0x13, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlps  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movlps    ebx[esi * 4 + 305419896], xmm0}
     */
    // Template#: 2668, Serial#: 2673
    public void movlps(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0424((byte) 0x13, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlps  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movlps    L1: 0x12345678[esi * 4], xmm0}
     */
    // Template#: 2669, Serial#: 2668
    public void m_movlps(int m32, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0425((byte) 0x13, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlps  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movlps    [ebx + 305419896], xmm0}
     */
    // Template#: 2670, Serial#: 2672
    public void movlps(int disp32, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0426((byte) 0x13, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movlps  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code movlps    [L1: 0x12345678], xmm0}
     */
    // Template#: 2671, Serial#: 2669
    public void m_movlps(int m32, IA32XMMRegister source) {
        assemble0427((byte) 0x13, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movmskpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movmskpd  eax, xmm0}
     */
    // Template#: 2672, Serial#: 3045
    public void movmskpd(IA32GeneralRegister32 destination, IA32XMMRegister source) {
        assemble0449((byte) 0x50, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movmskps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movmskps  eax, xmm0}
     */
    // Template#: 2673, Serial#: 2981
    public void movmskps(IA32GeneralRegister32 destination, IA32XMMRegister source) {
        assemble0450((byte) 0x50, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movnti  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movnti    ebx[esi * 4 + 18], eax}
     */
    // Template#: 2674, Serial#: 3636
    public void movnti(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0159((byte) 0xC3, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movnti  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movnti    [ebx + 18], eax}
     */
    // Template#: 2675, Serial#: 3635
    public void movnti(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0162((byte) 0xC3, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movnti  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movnti    ebx[esi * 4], eax}
     */
    // Template#: 2676, Serial#: 3632
    public void movnti(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0165((byte) 0xC3, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movnti  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movnti    [ebx], eax}
     */
    // Template#: 2677, Serial#: 3631
    public void movnti(IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0172((byte) 0xC3, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movnti  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movnti    ebx[esi * 4 + 305419896], eax}
     */
    // Template#: 2678, Serial#: 3638
    public void movnti(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0176((byte) 0xC3, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movnti  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code movnti    [L1: 0x12345678], eax}
     */
    // Template#: 2679, Serial#: 3634
    public void m_movnti(int m32, IA32GeneralRegister32 source) {
        assemble0178((byte) 0xC3, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movnti  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movnti    L1: 0x12345678[esi * 4], eax}
     */
    // Template#: 2680, Serial#: 3633
    public void m_movnti(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0181((byte) 0xC3, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movnti  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movnti    [ebx + 305419896], eax}
     */
    // Template#: 2681, Serial#: 3637
    public void movnti(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0184((byte) 0xC3, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movntpd  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movntpd   ebx[esi * 4 + 18], xmm0}
     */
    // Template#: 2682, Serial#: 4133
    public void movntpd(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0412((byte) 0x2B, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movntpd  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movntpd   [ebx + 18], xmm0}
     */
    // Template#: 2683, Serial#: 4132
    public void movntpd(byte disp8, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0413((byte) 0x2B, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movntpd  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movntpd   ebx[esi * 4], xmm0}
     */
    // Template#: 2684, Serial#: 4129
    public void movntpd(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0414((byte) 0x2B, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movntpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movntpd   [ebx], xmm0}
     */
    // Template#: 2685, Serial#: 4128
    public void movntpd(IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0415((byte) 0x2B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movntpd  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movntpd   ebx[esi * 4 + 305419896], xmm0}
     */
    // Template#: 2686, Serial#: 4135
    public void movntpd(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0416((byte) 0x2B, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movntpd  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movntpd   L1: 0x12345678[esi * 4], xmm0}
     */
    // Template#: 2687, Serial#: 4130
    public void m_movntpd(int m32, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0417((byte) 0x2B, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movntpd  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movntpd   [ebx + 305419896], xmm0}
     */
    // Template#: 2688, Serial#: 4134
    public void movntpd(int disp32, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0418((byte) 0x2B, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movntpd  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code movntpd   [L1: 0x12345678], xmm0}
     */
    // Template#: 2689, Serial#: 4131
    public void m_movntpd(int m32, IA32XMMRegister source) {
        assemble0419((byte) 0x2B, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movntps  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movntps   ebx[esi * 4 + 18], xmm0}
     */
    // Template#: 2690, Serial#: 4063
    public void movntps(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0420((byte) 0x2B, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movntps  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movntps   [ebx + 18], xmm0}
     */
    // Template#: 2691, Serial#: 4062
    public void movntps(byte disp8, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0421((byte) 0x2B, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movntps  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movntps   ebx[esi * 4], xmm0}
     */
    // Template#: 2692, Serial#: 4059
    public void movntps(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0422((byte) 0x2B, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movntps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movntps   [ebx], xmm0}
     */
    // Template#: 2693, Serial#: 4058
    public void movntps(IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0423((byte) 0x2B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movntps  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movntps   ebx[esi * 4 + 305419896], xmm0}
     */
    // Template#: 2694, Serial#: 4065
    public void movntps(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0424((byte) 0x2B, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movntps  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movntps   L1: 0x12345678[esi * 4], xmm0}
     */
    // Template#: 2695, Serial#: 4060
    public void m_movntps(int m32, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0425((byte) 0x2B, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movntps  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movntps   [ebx + 305419896], xmm0}
     */
    // Template#: 2696, Serial#: 4064
    public void movntps(int disp32, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0426((byte) 0x2B, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movntps  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code movntps   [L1: 0x12345678], xmm0}
     */
    // Template#: 2697, Serial#: 4061
    public void m_movntps(int m32, IA32XMMRegister source) {
        assemble0427((byte) 0x2B, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movq      ebx[esi * 4 + 18], xmm0}
     */
    // Template#: 2698, Serial#: 3807
    public void movq(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0412((byte) 0xD6, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movq      ebx[esi * 4 + 18], mm0}
     */
    // Template#: 2699, Serial#: 4798
    public void movq(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, MMXRegister source) {
        assemble0428((byte) 0x7F, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movq      [ebx + 18], xmm0}
     */
    // Template#: 2700, Serial#: 3806
    public void movq(byte disp8, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0413((byte) 0xD6, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movq      [ebx + 18], mm0}
     */
    // Template#: 2701, Serial#: 4797
    public void movq(byte disp8, IA32IndirectRegister32 destination, MMXRegister source) {
        assemble0429((byte) 0x7F, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movq      ebx[esi * 4], xmm0}
     */
    // Template#: 2702, Serial#: 3803
    public void movq(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0414((byte) 0xD6, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movq      ebx[esi * 4], mm0}
     */
    // Template#: 2703, Serial#: 4794
    public void movq(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, MMXRegister source) {
        assemble0430((byte) 0x7F, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movq      [ebx], xmm0}
     */
    // Template#: 2704, Serial#: 3802
    public void movq(IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0415((byte) 0xD6, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movq      [ebx], mm0}
     */
    // Template#: 2705, Serial#: 4793
    public void movq(IA32IndirectRegister32 destination, MMXRegister source) {
        assemble0433((byte) 0x7F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movq      xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2706, Serial#: 4859
    public void movq(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0120((byte) 0x7E, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movq      xmm0, [ebx + 18]}
     */
    // Template#: 2707, Serial#: 4858
    public void movq(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0121((byte) 0x7E, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movq      xmm0, ebx[esi * 4]}
     */
    // Template#: 2708, Serial#: 4855
    public void movq(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0122((byte) 0x7E, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movq      xmm0, [ebx]}
     */
    // Template#: 2709, Serial#: 4854
    public void movq(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0123((byte) 0x7E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movq      xmm0, xmm0}
     */
    // Template#: 2710, Serial#: 3810
    public void movq(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0451((byte) 0xD6, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movq      xmm0, [L1: 0x12345678]}
     */
    // Template#: 2711, Serial#: 4857
    public void m_movq(IA32XMMRegister destination, int m32) {
        assemble0125((byte) 0x7E, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movq      xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2712, Serial#: 4861
    public void movq(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0126((byte) 0x7E, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movq      xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2713, Serial#: 4856
    public void m_movq(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0127((byte) 0x7E, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movq      xmm0, [ebx + 305419896]}
     */
    // Template#: 2714, Serial#: 4860
    public void movq(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0128((byte) 0x7E, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movq      mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2715, Serial#: 4699
    public void movq(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0x6F, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movq      mm0, [ebx + 18]}
     */
    // Template#: 2716, Serial#: 4698
    public void movq(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0x6F, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movq      mm0, ebx[esi * 4]}
     */
    // Template#: 2717, Serial#: 4695
    public void movq(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0x6F, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movq      mm0, [ebx]}
     */
    // Template#: 2718, Serial#: 4694
    public void movq(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0x6F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movq      mm0, mm0}
     */
    // Template#: 2719, Serial#: 4702
    public void movq(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0x6F, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movq      mm0, [L1: 0x12345678]}
     */
    // Template#: 2720, Serial#: 4697
    public void m_movq(MMXRegister destination, int m32) {
        assemble0268((byte) 0x6F, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movq      mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2721, Serial#: 4701
    public void movq(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0x6F, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movq      mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2722, Serial#: 4696
    public void m_movq(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0x6F, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movq      mm0, [ebx + 305419896]}
     */
    // Template#: 2723, Serial#: 4700
    public void movq(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0x6F, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movq      ebx[esi * 4 + 305419896], xmm0}
     */
    // Template#: 2724, Serial#: 3809
    public void movq(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0416((byte) 0xD6, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movq      ebx[esi * 4 + 305419896], mm0}
     */
    // Template#: 2725, Serial#: 4800
    public void movq(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, MMXRegister source) {
        assemble0436((byte) 0x7F, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movq      L1: 0x12345678[esi * 4], xmm0}
     */
    // Template#: 2726, Serial#: 3804
    public void m_movq(int m32, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0417((byte) 0xD6, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movq      L1: 0x12345678[esi * 4], mm0}
     */
    // Template#: 2727, Serial#: 4795
    public void m_movq(int m32, IA32IndexRegister32 index, Scale scale, MMXRegister source) {
        assemble0437((byte) 0x7F, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movq      [ebx + 305419896], xmm0}
     */
    // Template#: 2728, Serial#: 3808
    public void movq(int disp32, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0418((byte) 0xD6, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movq      [ebx + 305419896], mm0}
     */
    // Template#: 2729, Serial#: 4799
    public void movq(int disp32, IA32IndirectRegister32 destination, MMXRegister source) {
        assemble0438((byte) 0x7F, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code movq      [L1: 0x12345678], xmm0}
     */
    // Template#: 2730, Serial#: 3805
    public void m_movq(int m32, IA32XMMRegister source) {
        assemble0419((byte) 0xD6, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code movq      [L1: 0x12345678], mm0}
     */
    // Template#: 2731, Serial#: 4796
    public void m_movq(int m32, MMXRegister source) {
        assemble0439((byte) 0x7F, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movq2dq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movq2dq   xmm0, mm0}
     */
    // Template#: 2732, Serial#: 3813
    public void movq2dq(IA32XMMRegister destination, MMXRegister source) {
        assemble0453((byte) 0xD6, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsb  }
     * Example disassembly syntax: {@code movsb   }
     */
    // Template#: 2733, Serial#: 693
    public void movsb() {
        assemble0001((byte) 0xA4);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsl  }
     * Example disassembly syntax: {@code movsl   }
     */
    // Template#: 2734, Serial#: 694
    public void movsl() {
        assemble0001((byte) 0xA5);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsw  }
     * Example disassembly syntax: {@code movsw   }
     */
    // Template#: 2735, Serial#: 695
    public void movsw() {
        assemble0197((byte) 0xA5);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsd  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movsd     ebx[esi * 4 + 18], xmm0}
     */
    // Template#: 2736, Serial#: 2782
    public void movsd(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0454((byte) 0x11, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsd  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movsd     [ebx + 18], xmm0}
     */
    // Template#: 2737, Serial#: 2781
    public void movsd(byte disp8, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0455((byte) 0x11, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsd  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movsd     ebx[esi * 4], xmm0}
     */
    // Template#: 2738, Serial#: 2778
    public void movsd(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0456((byte) 0x11, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movsd     [ebx], xmm0}
     */
    // Template#: 2739, Serial#: 2777
    public void movsd(IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0457((byte) 0x11, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movsd     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2740, Serial#: 2773
    public void movsd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0111((byte) 0x10, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movsd     xmm0, [ebx + 18]}
     */
    // Template#: 2741, Serial#: 2772
    public void movsd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0112((byte) 0x10, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movsd     xmm0, ebx[esi * 4]}
     */
    // Template#: 2742, Serial#: 2769
    public void movsd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0113((byte) 0x10, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movsd     xmm0, [ebx]}
     */
    // Template#: 2743, Serial#: 2768
    public void movsd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0114((byte) 0x10, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movsd     xmm0, xmm0}
     */
    // Template#: 2744, Serial#: 2776
    public void movsd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0115((byte) 0x10, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movsd     xmm0, [L1: 0x12345678]}
     */
    // Template#: 2745, Serial#: 2771
    public void m_movsd(IA32XMMRegister destination, int m32) {
        assemble0116((byte) 0x10, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movsd     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2746, Serial#: 2775
    public void movsd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0117((byte) 0x10, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movsd     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2747, Serial#: 2770
    public void m_movsd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0118((byte) 0x10, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movsd     xmm0, [ebx + 305419896]}
     */
    // Template#: 2748, Serial#: 2774
    public void movsd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0119((byte) 0x10, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsd  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movsd     ebx[esi * 4 + 305419896], xmm0}
     */
    // Template#: 2749, Serial#: 2784
    public void movsd(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0458((byte) 0x11, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsd  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movsd     L1: 0x12345678[esi * 4], xmm0}
     */
    // Template#: 2750, Serial#: 2779
    public void m_movsd(int m32, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0459((byte) 0x11, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsd  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movsd     [ebx + 305419896], xmm0}
     */
    // Template#: 2751, Serial#: 2783
    public void movsd(int disp32, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0460((byte) 0x11, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsd  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code movsd     [L1: 0x12345678], xmm0}
     */
    // Template#: 2752, Serial#: 2780
    public void m_movsd(int m32, IA32XMMRegister source) {
        assemble0461((byte) 0x11, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movshdup  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movshdup  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2753, Serial#: 2825
    public void movshdup(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0120((byte) 0x16, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movshdup  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movshdup  xmm0, [ebx + 18]}
     */
    // Template#: 2754, Serial#: 2824
    public void movshdup(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0121((byte) 0x16, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movshdup  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movshdup  xmm0, ebx[esi * 4]}
     */
    // Template#: 2755, Serial#: 2821
    public void movshdup(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0122((byte) 0x16, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movshdup  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movshdup  xmm0, [ebx]}
     */
    // Template#: 2756, Serial#: 2820
    public void movshdup(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0123((byte) 0x16, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movshdup  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movshdup  xmm0, xmm0}
     */
    // Template#: 2757, Serial#: 2828
    public void movshdup(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0124((byte) 0x16, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movshdup  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movshdup  xmm0, [L1: 0x12345678]}
     */
    // Template#: 2758, Serial#: 2823
    public void m_movshdup(IA32XMMRegister destination, int m32) {
        assemble0125((byte) 0x16, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movshdup  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movshdup  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2759, Serial#: 2827
    public void movshdup(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0126((byte) 0x16, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movshdup  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movshdup  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2760, Serial#: 2822
    public void m_movshdup(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0127((byte) 0x16, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movshdup  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movshdup  xmm0, [ebx + 305419896]}
     */
    // Template#: 2761, Serial#: 2826
    public void movshdup(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0128((byte) 0x16, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsldup  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movsldup  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2762, Serial#: 2816
    public void movsldup(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0120((byte) 0x12, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsldup  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movsldup  xmm0, [ebx + 18]}
     */
    // Template#: 2763, Serial#: 2815
    public void movsldup(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0121((byte) 0x12, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsldup  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movsldup  xmm0, ebx[esi * 4]}
     */
    // Template#: 2764, Serial#: 2812
    public void movsldup(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0122((byte) 0x12, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsldup  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movsldup  xmm0, [ebx]}
     */
    // Template#: 2765, Serial#: 2811
    public void movsldup(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0123((byte) 0x12, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsldup  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movsldup  xmm0, xmm0}
     */
    // Template#: 2766, Serial#: 2819
    public void movsldup(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0124((byte) 0x12, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsldup  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movsldup  xmm0, [L1: 0x12345678]}
     */
    // Template#: 2767, Serial#: 2814
    public void m_movsldup(IA32XMMRegister destination, int m32) {
        assemble0125((byte) 0x12, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsldup  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movsldup  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2768, Serial#: 2818
    public void movsldup(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0126((byte) 0x12, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsldup  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movsldup  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2769, Serial#: 2813
    public void m_movsldup(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0127((byte) 0x12, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsldup  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movsldup  xmm0, [ebx + 305419896]}
     */
    // Template#: 2770, Serial#: 2817
    public void movsldup(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0128((byte) 0x12, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movss  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movss     ebx[esi * 4 + 18], xmm0}
     */
    // Template#: 2771, Serial#: 2808
    public void movss(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0441((byte) 0x11, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movss  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movss     [ebx + 18], xmm0}
     */
    // Template#: 2772, Serial#: 2807
    public void movss(byte disp8, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0442((byte) 0x11, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movss  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movss     ebx[esi * 4], xmm0}
     */
    // Template#: 2773, Serial#: 2804
    public void movss(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0443((byte) 0x11, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movss     [ebx], xmm0}
     */
    // Template#: 2774, Serial#: 2803
    public void movss(IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0444((byte) 0x11, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movss  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movss     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2775, Serial#: 2799
    public void movss(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0120((byte) 0x10, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movss  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movss     xmm0, [ebx + 18]}
     */
    // Template#: 2776, Serial#: 2798
    public void movss(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0121((byte) 0x10, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movss  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movss     xmm0, ebx[esi * 4]}
     */
    // Template#: 2777, Serial#: 2795
    public void movss(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0122((byte) 0x10, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movss     xmm0, [ebx]}
     */
    // Template#: 2778, Serial#: 2794
    public void movss(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0123((byte) 0x10, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movss     xmm0, xmm0}
     */
    // Template#: 2779, Serial#: 2802
    public void movss(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0124((byte) 0x10, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movss  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movss     xmm0, [L1: 0x12345678]}
     */
    // Template#: 2780, Serial#: 2797
    public void m_movss(IA32XMMRegister destination, int m32) {
        assemble0125((byte) 0x10, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movss  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movss     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2781, Serial#: 2801
    public void movss(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0126((byte) 0x10, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movss  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movss     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2782, Serial#: 2796
    public void m_movss(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0127((byte) 0x10, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movss  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movss     xmm0, [ebx + 305419896]}
     */
    // Template#: 2783, Serial#: 2800
    public void movss(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0128((byte) 0x10, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movss  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movss     ebx[esi * 4 + 305419896], xmm0}
     */
    // Template#: 2784, Serial#: 2810
    public void movss(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0445((byte) 0x11, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movss  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movss     L1: 0x12345678[esi * 4], xmm0}
     */
    // Template#: 2785, Serial#: 2805
    public void m_movss(int m32, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0446((byte) 0x11, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movss  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movss     [ebx + 305419896], xmm0}
     */
    // Template#: 2786, Serial#: 2809
    public void movss(int disp32, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0447((byte) 0x11, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movss  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code movss     [L1: 0x12345678], xmm0}
     */
    // Template#: 2787, Serial#: 2806
    public void m_movss(int m32, IA32XMMRegister source) {
        assemble0448((byte) 0x11, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsx  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movsx     ax, ebx[esi * 4 + 18]}
     */
    // Template#: 2788, Serial#: 5136
    public void movsxb(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0xBE, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsx  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movsx     ax, [ebx + 18]}
     */
    // Template#: 2789, Serial#: 5135
    public void movsxb(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0xBE, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsx  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movsx     ax, ebx[esi * 4]}
     */
    // Template#: 2790, Serial#: 5132
    public void movsxb(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0xBE, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsx  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movsx     ax, al}
     */
    // Template#: 2791, Serial#: 5139
    public void movsxb(IA32GeneralRegister16 destination, IA32GeneralRegister8 source) {
        assemble0462((byte) 0xBE, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsx  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movsx     ax, [ebx]}
     */
    // Template#: 2792, Serial#: 5131
    public void movsxb(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0xBE, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsx  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movsx     ax, [L1: 0x12345678]}
     */
    // Template#: 2793, Serial#: 5134
    public void m_movsxb(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0xBE, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsx  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movsx     ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2794, Serial#: 5138
    public void movsxb(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0xBE, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsx  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movsx     ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2795, Serial#: 5133
    public void m_movsxb(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0xBE, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsx  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movsx     ax, [ebx + 305419896]}
     */
    // Template#: 2796, Serial#: 5137
    public void movsxb(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0xBE, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsx  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movsx     eax, ebx[esi * 4 + 18]}
     */
    // Template#: 2797, Serial#: 5127
    public void movsxb(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0xBE, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsx  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movsx     eax, [ebx + 18]}
     */
    // Template#: 2798, Serial#: 5126
    public void movsxb(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0xBE, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsx  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movsx     eax, ebx[esi * 4]}
     */
    // Template#: 2799, Serial#: 5123
    public void movsxb(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0xBE, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsx  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movsx     eax, al}
     */
    // Template#: 2800, Serial#: 5130
    public void movsxb(IA32GeneralRegister32 destination, IA32GeneralRegister8 source) {
        assemble0463((byte) 0xBE, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsx  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movsx     eax, [ebx]}
     */
    // Template#: 2801, Serial#: 5122
    public void movsxb(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0xBE, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsx  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movsx     eax, [L1: 0x12345678]}
     */
    // Template#: 2802, Serial#: 5125
    public void m_movsxb(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0xBE, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsx  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movsx     eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2803, Serial#: 5129
    public void movsxb(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0xBE, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsx  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movsx     eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2804, Serial#: 5124
    public void m_movsxb(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0xBE, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsx  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movsx     eax, [ebx + 305419896]}
     */
    // Template#: 2805, Serial#: 5128
    public void movsxb(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0xBE, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsxw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movsxw    ax, ebx[esi * 4 + 18]}
     */
    // Template#: 2806, Serial#: 5154
    public void movsxw(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0xBF, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsxw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movsxw    ax, [ebx + 18]}
     */
    // Template#: 2807, Serial#: 5153
    public void movsxw(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0xBF, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsxw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movsxw    ax, ebx[esi * 4]}
     */
    // Template#: 2808, Serial#: 5150
    public void movsxw(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0xBF, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsxw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movsxw    ax, ax}
     */
    // Template#: 2809, Serial#: 5157
    public void movsxw(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0141((byte) 0xBF, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsxw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movsxw    ax, [ebx]}
     */
    // Template#: 2810, Serial#: 5149
    public void movsxw(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0xBF, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsxw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movsxw    ax, [L1: 0x12345678]}
     */
    // Template#: 2811, Serial#: 5152
    public void m_movsxw(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0xBF, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsxw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movsxw    ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2812, Serial#: 5156
    public void movsxw(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0xBF, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsxw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movsxw    ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2813, Serial#: 5151
    public void m_movsxw(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0xBF, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsxw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movsxw    ax, [ebx + 305419896]}
     */
    // Template#: 2814, Serial#: 5155
    public void movsxw(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0xBF, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsxw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movsxw    eax, ebx[esi * 4 + 18]}
     */
    // Template#: 2815, Serial#: 5145
    public void movsxw(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0xBF, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsxw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movsxw    eax, [ebx + 18]}
     */
    // Template#: 2816, Serial#: 5144
    public void movsxw(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0xBF, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsxw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movsxw    eax, ebx[esi * 4]}
     */
    // Template#: 2817, Serial#: 5141
    public void movsxw(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0xBF, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsxw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movsxw    eax, ax}
     */
    // Template#: 2818, Serial#: 5148
    public void movsxw(IA32GeneralRegister32 destination, IA32GeneralRegister16 source) {
        assemble0464((byte) 0xBF, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsxw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movsxw    eax, [ebx]}
     */
    // Template#: 2819, Serial#: 5140
    public void movsxw(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0xBF, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsxw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movsxw    eax, [L1: 0x12345678]}
     */
    // Template#: 2820, Serial#: 5143
    public void m_movsxw(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0xBF, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsxw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movsxw    eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2821, Serial#: 5147
    public void movsxw(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0xBF, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsxw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movsxw    eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2822, Serial#: 5142
    public void m_movsxw(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0xBF, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movsxw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movsxw    eax, [ebx + 305419896]}
     */
    // Template#: 2823, Serial#: 5146
    public void movsxw(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0xBF, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movupd  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movupd    ebx[esi * 4 + 18], xmm0}
     */
    // Template#: 2824, Serial#: 2715
    public void movupd(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0412((byte) 0x11, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movupd  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movupd    [ebx + 18], xmm0}
     */
    // Template#: 2825, Serial#: 2714
    public void movupd(byte disp8, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0413((byte) 0x11, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movupd  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movupd    ebx[esi * 4], xmm0}
     */
    // Template#: 2826, Serial#: 2711
    public void movupd(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0414((byte) 0x11, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movupd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movupd    [ebx], xmm0}
     */
    // Template#: 2827, Serial#: 2710
    public void movupd(IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0415((byte) 0x11, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movupd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movupd    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2828, Serial#: 2706
    public void movupd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x10, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movupd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movupd    xmm0, [ebx + 18]}
     */
    // Template#: 2829, Serial#: 2705
    public void movupd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x10, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movupd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movupd    xmm0, ebx[esi * 4]}
     */
    // Template#: 2830, Serial#: 2702
    public void movupd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x10, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movupd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movupd    xmm0, [ebx]}
     */
    // Template#: 2831, Serial#: 2701
    public void movupd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x10, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movupd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movupd    xmm0, xmm0}
     */
    // Template#: 2832, Serial#: 2709
    public void movupd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x10, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movupd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movupd    xmm0, [L1: 0x12345678]}
     */
    // Template#: 2833, Serial#: 2704
    public void m_movupd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x10, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movupd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movupd    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2834, Serial#: 2708
    public void movupd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x10, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movupd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movupd    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2835, Serial#: 2703
    public void m_movupd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x10, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movupd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movupd    xmm0, [ebx + 305419896]}
     */
    // Template#: 2836, Serial#: 2707
    public void movupd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x10, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movupd  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movupd    ebx[esi * 4 + 305419896], xmm0}
     */
    // Template#: 2837, Serial#: 2717
    public void movupd(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0416((byte) 0x11, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movupd  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movupd    L1: 0x12345678[esi * 4], xmm0}
     */
    // Template#: 2838, Serial#: 2712
    public void m_movupd(int m32, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0417((byte) 0x11, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movupd  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movupd    [ebx + 305419896], xmm0}
     */
    // Template#: 2839, Serial#: 2716
    public void movupd(int disp32, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0418((byte) 0x11, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movupd  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code movupd    [L1: 0x12345678], xmm0}
     */
    // Template#: 2840, Serial#: 2713
    public void m_movupd(int m32, IA32XMMRegister source) {
        assemble0419((byte) 0x11, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movups  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movups    ebx[esi * 4 + 18], xmm0}
     */
    // Template#: 2841, Serial#: 2662
    public void movups(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0420((byte) 0x11, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movups  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movups    [ebx + 18], xmm0}
     */
    // Template#: 2842, Serial#: 2661
    public void movups(byte disp8, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0421((byte) 0x11, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movups  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movups    ebx[esi * 4], xmm0}
     */
    // Template#: 2843, Serial#: 2658
    public void movups(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0422((byte) 0x11, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movups  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movups    [ebx], xmm0}
     */
    // Template#: 2844, Serial#: 2657
    public void movups(IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0423((byte) 0x11, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movups  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movups    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2845, Serial#: 2653
    public void movups(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0102((byte) 0x10, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movups  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movups    xmm0, [ebx + 18]}
     */
    // Template#: 2846, Serial#: 2652
    public void movups(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0103((byte) 0x10, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movups  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movups    xmm0, ebx[esi * 4]}
     */
    // Template#: 2847, Serial#: 2649
    public void movups(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0104((byte) 0x10, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movups  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movups    xmm0, [ebx]}
     */
    // Template#: 2848, Serial#: 2648
    public void movups(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0105((byte) 0x10, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movups  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movups    xmm0, xmm0}
     */
    // Template#: 2849, Serial#: 2656
    public void movups(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0106((byte) 0x10, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movups  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movups    xmm0, [L1: 0x12345678]}
     */
    // Template#: 2850, Serial#: 2651
    public void m_movups(IA32XMMRegister destination, int m32) {
        assemble0107((byte) 0x10, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movups  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movups    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2851, Serial#: 2655
    public void movups(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0108((byte) 0x10, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movups  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movups    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2852, Serial#: 2650
    public void m_movups(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0109((byte) 0x10, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movups  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movups    xmm0, [ebx + 305419896]}
     */
    // Template#: 2853, Serial#: 2654
    public void movups(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0110((byte) 0x10, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movups  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movups    ebx[esi * 4 + 305419896], xmm0}
     */
    // Template#: 2854, Serial#: 2664
    public void movups(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0424((byte) 0x11, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movups  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code movups    L1: 0x12345678[esi * 4], xmm0}
     */
    // Template#: 2855, Serial#: 2659
    public void m_movups(int m32, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0425((byte) 0x11, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movups  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movups    [ebx + 305419896], xmm0}
     */
    // Template#: 2856, Serial#: 2663
    public void movups(int disp32, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0426((byte) 0x11, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movups  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code movups    [L1: 0x12345678], xmm0}
     */
    // Template#: 2857, Serial#: 2660
    public void m_movups(int m32, IA32XMMRegister source) {
        assemble0427((byte) 0x11, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzx  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movzx     ax, ebx[esi * 4 + 18]}
     */
    // Template#: 2858, Serial#: 3573
    public void movzxb(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0xB6, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzx  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movzx     ax, [ebx + 18]}
     */
    // Template#: 2859, Serial#: 3572
    public void movzxb(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0xB6, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzx  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movzx     ax, ebx[esi * 4]}
     */
    // Template#: 2860, Serial#: 3569
    public void movzxb(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0xB6, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzx  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movzx     ax, al}
     */
    // Template#: 2861, Serial#: 3576
    public void movzxb(IA32GeneralRegister16 destination, IA32GeneralRegister8 source) {
        assemble0462((byte) 0xB6, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzx  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movzx     ax, [ebx]}
     */
    // Template#: 2862, Serial#: 3568
    public void movzxb(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0xB6, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzx  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movzx     ax, [L1: 0x12345678]}
     */
    // Template#: 2863, Serial#: 3571
    public void m_movzxb(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0xB6, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzx  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movzx     ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2864, Serial#: 3575
    public void movzxb(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0xB6, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzx  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movzx     ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2865, Serial#: 3570
    public void m_movzxb(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0xB6, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzx  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movzx     ax, [ebx + 305419896]}
     */
    // Template#: 2866, Serial#: 3574
    public void movzxb(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0xB6, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzx  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movzx     eax, ebx[esi * 4 + 18]}
     */
    // Template#: 2867, Serial#: 3564
    public void movzxb(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0xB6, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzx  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movzx     eax, [ebx + 18]}
     */
    // Template#: 2868, Serial#: 3563
    public void movzxb(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0xB6, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzx  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movzx     eax, ebx[esi * 4]}
     */
    // Template#: 2869, Serial#: 3560
    public void movzxb(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0xB6, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzx  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movzx     eax, al}
     */
    // Template#: 2870, Serial#: 3567
    public void movzxb(IA32GeneralRegister32 destination, IA32GeneralRegister8 source) {
        assemble0463((byte) 0xB6, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzx  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movzx     eax, [ebx]}
     */
    // Template#: 2871, Serial#: 3559
    public void movzxb(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0xB6, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzx  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movzx     eax, [L1: 0x12345678]}
     */
    // Template#: 2872, Serial#: 3562
    public void m_movzxb(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0xB6, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzx  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movzx     eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2873, Serial#: 3566
    public void movzxb(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0xB6, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzx  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movzx     eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2874, Serial#: 3561
    public void m_movzxb(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0xB6, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzx  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movzx     eax, [ebx + 305419896]}
     */
    // Template#: 2875, Serial#: 3565
    public void movzxb(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0xB6, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzxw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movzxw    ax, ebx[esi * 4 + 18]}
     */
    // Template#: 2876, Serial#: 3591
    public void movzxw(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0138((byte) 0xB7, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzxw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movzxw    ax, [ebx + 18]}
     */
    // Template#: 2877, Serial#: 3590
    public void movzxw(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0139((byte) 0xB7, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzxw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movzxw    ax, ebx[esi * 4]}
     */
    // Template#: 2878, Serial#: 3587
    public void movzxw(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0140((byte) 0xB7, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzxw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movzxw    ax, ax}
     */
    // Template#: 2879, Serial#: 3594
    public void movzxw(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0141((byte) 0xB7, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzxw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movzxw    ax, [ebx]}
     */
    // Template#: 2880, Serial#: 3586
    public void movzxw(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0142((byte) 0xB7, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzxw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movzxw    ax, [L1: 0x12345678]}
     */
    // Template#: 2881, Serial#: 3589
    public void m_movzxw(IA32GeneralRegister16 destination, int m32) {
        assemble0143((byte) 0xB7, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzxw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movzxw    ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2882, Serial#: 3593
    public void movzxw(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0144((byte) 0xB7, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzxw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movzxw    ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2883, Serial#: 3588
    public void m_movzxw(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0145((byte) 0xB7, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzxw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movzxw    ax, [ebx + 305419896]}
     */
    // Template#: 2884, Serial#: 3592
    public void movzxw(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0146((byte) 0xB7, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzxw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movzxw    eax, ebx[esi * 4 + 18]}
     */
    // Template#: 2885, Serial#: 3582
    public void movzxw(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0147((byte) 0xB7, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzxw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code movzxw    eax, [ebx + 18]}
     */
    // Template#: 2886, Serial#: 3581
    public void movzxw(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0148((byte) 0xB7, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzxw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movzxw    eax, ebx[esi * 4]}
     */
    // Template#: 2887, Serial#: 3578
    public void movzxw(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0149((byte) 0xB7, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzxw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movzxw    eax, ax}
     */
    // Template#: 2888, Serial#: 3585
    public void movzxw(IA32GeneralRegister32 destination, IA32GeneralRegister16 source) {
        assemble0464((byte) 0xB7, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzxw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code movzxw    eax, [ebx]}
     */
    // Template#: 2889, Serial#: 3577
    public void movzxw(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0151((byte) 0xB7, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzxw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code movzxw    eax, [L1: 0x12345678]}
     */
    // Template#: 2890, Serial#: 3580
    public void m_movzxw(IA32GeneralRegister32 destination, int m32) {
        assemble0152((byte) 0xB7, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzxw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movzxw    eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2891, Serial#: 3584
    public void movzxw(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0153((byte) 0xB7, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzxw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code movzxw    eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2892, Serial#: 3579
    public void m_movzxw(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0154((byte) 0xB7, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code movzxw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code movzxw    eax, [ebx + 305419896]}
     */
    // Template#: 2893, Serial#: 3583
    public void movzxw(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0155((byte) 0xB7, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mulb      ebx[esi * 4 + 18], al}
     */
    // Template#: 2894, Serial#: 1382
    public void mulb___AL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xF6, (byte) 0x04, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mull  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mull      ebx[esi * 4 + 18], eax}
     */
    // Template#: 2895, Serial#: 1445
    public void mull___EAX(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xF7, (byte) 0x04, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mulw      ebx[esi * 4 + 18], ax}
     */
    // Template#: 2896, Serial#: 1508
    public void mulw___AX(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0292((byte) 0xF7, (byte) 0x04, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulb  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code mulb      [ebx + 18], al}
     */
    // Template#: 2897, Serial#: 1381
    public void mulb___AL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xF6, (byte) 0x04, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code mull  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code mull      [ebx + 18], eax}
     */
    // Template#: 2898, Serial#: 1444
    public void mull___EAX(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xF7, (byte) 0x04, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code mulw      [ebx + 18], ax}
     */
    // Template#: 2899, Serial#: 1507
    public void mulw___AX(byte disp8, IA32IndirectRegister32 destination) {
        assemble0293((byte) 0xF7, (byte) 0x04, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulb  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mulb      ebx[esi * 4], al}
     */
    // Template#: 2900, Serial#: 1360
    public void mulb___AL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xF6, (byte) 0x04, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mull  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mull      ebx[esi * 4], eax}
     */
    // Template#: 2901, Serial#: 1423
    public void mull___EAX(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xF7, (byte) 0x04, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mulw      ebx[esi * 4], ax}
     */
    // Template#: 2902, Serial#: 1486
    public void mulw___AX(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0294((byte) 0xF7, (byte) 0x04, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulw  }<i>destination</i>
     * Example disassembly syntax: {@code mulw      ax, ax}
     */
    // Template#: 2903, Serial#: 1532
    public void mulw___AX(IA32GeneralRegister16 destination) {
        assemble0188((byte) 0xF7, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code mull  }<i>destination</i>
     * Example disassembly syntax: {@code mull      eax, eax}
     */
    // Template#: 2904, Serial#: 1469
    public void mull___EAX(IA32GeneralRegister32 destination) {
        assemble0189((byte) 0xF7, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulb  }<i>destination</i>
     * Example disassembly syntax: {@code mulb      al, al}
     */
    // Template#: 2905, Serial#: 1406
    public void mulb___AL(IA32GeneralRegister8 destination) {
        assemble0297((byte) 0xF6, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulb  }<i>destination</i>
     * Example disassembly syntax: {@code mulb      [ebx], al}
     */
    // Template#: 2906, Serial#: 1359
    public void mulb___AL(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xF6, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code mull  }<i>destination</i>
     * Example disassembly syntax: {@code mull      [ebx], eax}
     */
    // Template#: 2907, Serial#: 1422
    public void mull___EAX(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xF7, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulw  }<i>destination</i>
     * Example disassembly syntax: {@code mulw      [ebx], ax}
     */
    // Template#: 2908, Serial#: 1485
    public void mulw___AX(IA32IndirectRegister32 destination) {
        assemble0298((byte) 0xF7, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulb  }<i>m32</i>
     * Example disassembly syntax: {@code mulb      [L1: 0x12345678], al}
     */
    // Template#: 2909, Serial#: 1362
    public void m_mulb___AL(int m32) {
        assemble0299((byte) 0xF6, (byte) 0x04, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code mull  }<i>m32</i>
     * Example disassembly syntax: {@code mull      [L1: 0x12345678], eax}
     */
    // Template#: 2910, Serial#: 1425
    public void m_mull___EAX(int m32) {
        assemble0299((byte) 0xF7, (byte) 0x04, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulw  }<i>m32</i>
     * Example disassembly syntax: {@code mulw      [L1: 0x12345678], ax}
     */
    // Template#: 2911, Serial#: 1488
    public void m_mulw___AX(int m32) {
        assemble0300((byte) 0xF7, (byte) 0x04, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mulb      ebx[esi * 4 + 305419896], al}
     */
    // Template#: 2912, Serial#: 1396
    public void mulb___AL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xF6, (byte) 0x04, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mull  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mull      ebx[esi * 4 + 305419896], eax}
     */
    // Template#: 2913, Serial#: 1459
    public void mull___EAX(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xF7, (byte) 0x04, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mulw      ebx[esi * 4 + 305419896], ax}
     */
    // Template#: 2914, Serial#: 1522
    public void mulw___AX(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0301((byte) 0xF7, (byte) 0x04, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulb  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mulb      L1: 0x12345678[esi * 4], al}
     */
    // Template#: 2915, Serial#: 1361
    public void m_mulb___AL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xF6, (byte) 0x04, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mull  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mull      L1: 0x12345678[esi * 4], eax}
     */
    // Template#: 2916, Serial#: 1424
    public void m_mull___EAX(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xF7, (byte) 0x04, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mulw      L1: 0x12345678[esi * 4], ax}
     */
    // Template#: 2917, Serial#: 1487
    public void m_mulw___AX(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0302((byte) 0xF7, (byte) 0x04, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulb  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code mulb      [ebx + 305419896], al}
     */
    // Template#: 2918, Serial#: 1395
    public void mulb___AL(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xF6, (byte) 0x04, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code mull  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code mull      [ebx + 305419896], eax}
     */
    // Template#: 2919, Serial#: 1458
    public void mull___EAX(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xF7, (byte) 0x04, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code mulw      [ebx + 305419896], ax}
     */
    // Template#: 2920, Serial#: 1521
    public void mulw___AX(int disp32, IA32IndirectRegister32 destination) {
        assemble0303((byte) 0xF7, (byte) 0x04, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulpd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mulpd     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2921, Serial#: 4456
    public void mulpd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x59, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulpd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code mulpd     xmm0, [ebx + 18]}
     */
    // Template#: 2922, Serial#: 4455
    public void mulpd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x59, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulpd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mulpd     xmm0, ebx[esi * 4]}
     */
    // Template#: 2923, Serial#: 4452
    public void mulpd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x59, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mulpd     xmm0, [ebx]}
     */
    // Template#: 2924, Serial#: 4451
    public void mulpd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x59, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mulpd     xmm0, xmm0}
     */
    // Template#: 2925, Serial#: 4459
    public void mulpd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x59, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulpd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code mulpd     xmm0, [L1: 0x12345678]}
     */
    // Template#: 2926, Serial#: 4454
    public void m_mulpd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x59, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulpd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mulpd     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2927, Serial#: 4458
    public void mulpd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x59, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulpd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mulpd     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2928, Serial#: 4453
    public void m_mulpd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x59, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulpd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code mulpd     xmm0, [ebx + 305419896]}
     */
    // Template#: 2929, Serial#: 4457
    public void mulpd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x59, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulps  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mulps     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2930, Serial#: 4384
    public void mulps(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0102((byte) 0x59, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulps  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code mulps     xmm0, [ebx + 18]}
     */
    // Template#: 2931, Serial#: 4383
    public void mulps(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0103((byte) 0x59, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulps  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mulps     xmm0, ebx[esi * 4]}
     */
    // Template#: 2932, Serial#: 4380
    public void mulps(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0104((byte) 0x59, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mulps     xmm0, [ebx]}
     */
    // Template#: 2933, Serial#: 4379
    public void mulps(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0105((byte) 0x59, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mulps     xmm0, xmm0}
     */
    // Template#: 2934, Serial#: 4387
    public void mulps(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0106((byte) 0x59, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulps  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code mulps     xmm0, [L1: 0x12345678]}
     */
    // Template#: 2935, Serial#: 4382
    public void m_mulps(IA32XMMRegister destination, int m32) {
        assemble0107((byte) 0x59, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulps  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mulps     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2936, Serial#: 4386
    public void mulps(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0108((byte) 0x59, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulps  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mulps     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2937, Serial#: 4381
    public void m_mulps(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0109((byte) 0x59, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulps  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code mulps     xmm0, [ebx + 305419896]}
     */
    // Template#: 2938, Serial#: 4385
    public void mulps(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0110((byte) 0x59, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulsd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mulsd     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2939, Serial#: 4528
    public void mulsd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0111((byte) 0x59, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulsd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code mulsd     xmm0, [ebx + 18]}
     */
    // Template#: 2940, Serial#: 4527
    public void mulsd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0112((byte) 0x59, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulsd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mulsd     xmm0, ebx[esi * 4]}
     */
    // Template#: 2941, Serial#: 4524
    public void mulsd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0113((byte) 0x59, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulsd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mulsd     xmm0, [ebx]}
     */
    // Template#: 2942, Serial#: 4523
    public void mulsd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0114((byte) 0x59, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulsd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mulsd     xmm0, xmm0}
     */
    // Template#: 2943, Serial#: 4531
    public void mulsd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0115((byte) 0x59, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulsd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code mulsd     xmm0, [L1: 0x12345678]}
     */
    // Template#: 2944, Serial#: 4526
    public void m_mulsd(IA32XMMRegister destination, int m32) {
        assemble0116((byte) 0x59, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulsd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mulsd     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2945, Serial#: 4530
    public void mulsd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0117((byte) 0x59, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulsd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mulsd     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2946, Serial#: 4525
    public void m_mulsd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0118((byte) 0x59, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulsd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code mulsd     xmm0, [ebx + 305419896]}
     */
    // Template#: 2947, Serial#: 4529
    public void mulsd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0119((byte) 0x59, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulss  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mulss     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 2948, Serial#: 4591
    public void mulss(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0120((byte) 0x59, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulss  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code mulss     xmm0, [ebx + 18]}
     */
    // Template#: 2949, Serial#: 4590
    public void mulss(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0121((byte) 0x59, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulss  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mulss     xmm0, ebx[esi * 4]}
     */
    // Template#: 2950, Serial#: 4587
    public void mulss(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0122((byte) 0x59, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mulss     xmm0, [ebx]}
     */
    // Template#: 2951, Serial#: 4586
    public void mulss(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0123((byte) 0x59, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mulss     xmm0, xmm0}
     */
    // Template#: 2952, Serial#: 4594
    public void mulss(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0124((byte) 0x59, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulss  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code mulss     xmm0, [L1: 0x12345678]}
     */
    // Template#: 2953, Serial#: 4589
    public void m_mulss(IA32XMMRegister destination, int m32) {
        assemble0125((byte) 0x59, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulss  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mulss     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 2954, Serial#: 4593
    public void mulss(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0126((byte) 0x59, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulss  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code mulss     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 2955, Serial#: 4588
    public void m_mulss(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0127((byte) 0x59, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code mulss  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code mulss     xmm0, [ebx + 305419896]}
     */
    // Template#: 2956, Serial#: 4592
    public void mulss(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0128((byte) 0x59, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mvntdq  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code mvntdq    ebx[esi * 4 + 18], xmm0}
     */
    // Template#: 2957, Serial#: 3909
    public void mvntdq(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0412((byte) 0xE7, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mvntdq  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mvntdq    [ebx + 18], xmm0}
     */
    // Template#: 2958, Serial#: 3908
    public void mvntdq(byte disp8, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0413((byte) 0xE7, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mvntdq  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code mvntdq    ebx[esi * 4], xmm0}
     */
    // Template#: 2959, Serial#: 3905
    public void mvntdq(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0414((byte) 0xE7, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mvntdq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mvntdq    [ebx], xmm0}
     */
    // Template#: 2960, Serial#: 3904
    public void mvntdq(IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0415((byte) 0xE7, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mvntdq  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code mvntdq    ebx[esi * 4 + 305419896], xmm0}
     */
    // Template#: 2961, Serial#: 3911
    public void mvntdq(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0416((byte) 0xE7, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mvntdq  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code mvntdq    L1: 0x12345678[esi * 4], xmm0}
     */
    // Template#: 2962, Serial#: 3906
    public void m_mvntdq(int m32, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        assemble0417((byte) 0xE7, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mvntdq  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code mvntdq    [ebx + 305419896], xmm0}
     */
    // Template#: 2963, Serial#: 3910
    public void mvntdq(int disp32, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        assemble0418((byte) 0xE7, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code mvntdq  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code mvntdq    [L1: 0x12345678], xmm0}
     */
    // Template#: 2964, Serial#: 3907
    public void m_mvntdq(int m32, IA32XMMRegister source) {
        assemble0419((byte) 0xE7, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code negb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code negb      ebx[esi * 4 + 18]}
     */
    // Template#: 2965, Serial#: 1380
    public void negb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xF6, (byte) 0x03, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code negl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code negl      ebx[esi * 4 + 18]}
     */
    // Template#: 2966, Serial#: 1443
    public void negl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xF7, (byte) 0x03, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code negw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code negw      ebx[esi * 4 + 18]}
     */
    // Template#: 2967, Serial#: 1506
    public void negw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0292((byte) 0xF7, (byte) 0x03, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code negb  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code negb      [ebx + 18]}
     */
    // Template#: 2968, Serial#: 1379
    public void negb(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xF6, (byte) 0x03, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code negl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code negl      [ebx + 18]}
     */
    // Template#: 2969, Serial#: 1442
    public void negl(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xF7, (byte) 0x03, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code negw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code negw      [ebx + 18]}
     */
    // Template#: 2970, Serial#: 1505
    public void negw(byte disp8, IA32IndirectRegister32 destination) {
        assemble0293((byte) 0xF7, (byte) 0x03, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code negb  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code negb      ebx[esi * 4]}
     */
    // Template#: 2971, Serial#: 1356
    public void negb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xF6, (byte) 0x03, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code negl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code negl      ebx[esi * 4]}
     */
    // Template#: 2972, Serial#: 1419
    public void negl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xF7, (byte) 0x03, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code negw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code negw      ebx[esi * 4]}
     */
    // Template#: 2973, Serial#: 1482
    public void negw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0294((byte) 0xF7, (byte) 0x03, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code negw  }<i>destination</i>
     * Example disassembly syntax: {@code negw      ax}
     */
    // Template#: 2974, Serial#: 1531
    public void negw(IA32GeneralRegister16 destination) {
        assemble0188((byte) 0xF7, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code negl  }<i>destination</i>
     * Example disassembly syntax: {@code negl      eax}
     */
    // Template#: 2975, Serial#: 1468
    public void negl(IA32GeneralRegister32 destination) {
        assemble0189((byte) 0xF7, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code negb  }<i>destination</i>
     * Example disassembly syntax: {@code negb      al}
     */
    // Template#: 2976, Serial#: 1405
    public void negb(IA32GeneralRegister8 destination) {
        assemble0297((byte) 0xF6, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code negb  }<i>destination</i>
     * Example disassembly syntax: {@code negb      [ebx]}
     */
    // Template#: 2977, Serial#: 1355
    public void negb(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xF6, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code negl  }<i>destination</i>
     * Example disassembly syntax: {@code negl      [ebx]}
     */
    // Template#: 2978, Serial#: 1418
    public void negl(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xF7, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code negw  }<i>destination</i>
     * Example disassembly syntax: {@code negw      [ebx]}
     */
    // Template#: 2979, Serial#: 1481
    public void negw(IA32IndirectRegister32 destination) {
        assemble0298((byte) 0xF7, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code negb  }<i>m32</i>
     * Example disassembly syntax: {@code negb      [L1: 0x12345678]}
     */
    // Template#: 2980, Serial#: 1358
    public void m_negb(int m32) {
        assemble0299((byte) 0xF6, (byte) 0x03, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code negl  }<i>m32</i>
     * Example disassembly syntax: {@code negl      [L1: 0x12345678]}
     */
    // Template#: 2981, Serial#: 1421
    public void m_negl(int m32) {
        assemble0299((byte) 0xF7, (byte) 0x03, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code negw  }<i>m32</i>
     * Example disassembly syntax: {@code negw      [L1: 0x12345678]}
     */
    // Template#: 2982, Serial#: 1484
    public void m_negw(int m32) {
        assemble0300((byte) 0xF7, (byte) 0x03, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code negb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code negb      ebx[esi * 4 + 305419896]}
     */
    // Template#: 2983, Serial#: 1394
    public void negb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xF6, (byte) 0x03, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code negl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code negl      ebx[esi * 4 + 305419896]}
     */
    // Template#: 2984, Serial#: 1457
    public void negl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xF7, (byte) 0x03, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code negw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code negw      ebx[esi * 4 + 305419896]}
     */
    // Template#: 2985, Serial#: 1520
    public void negw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0301((byte) 0xF7, (byte) 0x03, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code negb  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code negb      L1: 0x12345678[esi * 4]}
     */
    // Template#: 2986, Serial#: 1357
    public void m_negb(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xF6, (byte) 0x03, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code negl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code negl      L1: 0x12345678[esi * 4]}
     */
    // Template#: 2987, Serial#: 1420
    public void m_negl(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xF7, (byte) 0x03, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code negw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code negw      L1: 0x12345678[esi * 4]}
     */
    // Template#: 2988, Serial#: 1483
    public void m_negw(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0302((byte) 0xF7, (byte) 0x03, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code negb  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code negb      [ebx + 305419896]}
     */
    // Template#: 2989, Serial#: 1393
    public void negb(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xF6, (byte) 0x03, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code negl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code negl      [ebx + 305419896]}
     */
    // Template#: 2990, Serial#: 1456
    public void negl(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xF7, (byte) 0x03, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code negw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code negw      [ebx + 305419896]}
     */
    // Template#: 2991, Serial#: 1519
    public void negw(int disp32, IA32IndirectRegister32 destination) {
        assemble0303((byte) 0xF7, (byte) 0x03, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code nop  }
     * Example disassembly syntax: {@code nop     }
     */
    // Template#: 2992, Serial#: 684
    public void nop() {
        assemble0001((byte) 0x90);
    }

    /**
     * Pseudo-external assembler syntax: {@code notb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code notb      ebx[esi * 4 + 18]}
     */
    // Template#: 2993, Serial#: 1378
    public void notb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xF6, (byte) 0x02, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code notl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code notl      ebx[esi * 4 + 18]}
     */
    // Template#: 2994, Serial#: 1441
    public void notl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xF7, (byte) 0x02, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code notw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code notw      ebx[esi * 4 + 18]}
     */
    // Template#: 2995, Serial#: 1504
    public void notw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0292((byte) 0xF7, (byte) 0x02, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code notb  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code notb      [ebx + 18]}
     */
    // Template#: 2996, Serial#: 1377
    public void notb(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xF6, (byte) 0x02, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code notl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code notl      [ebx + 18]}
     */
    // Template#: 2997, Serial#: 1440
    public void notl(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xF7, (byte) 0x02, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code notw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code notw      [ebx + 18]}
     */
    // Template#: 2998, Serial#: 1503
    public void notw(byte disp8, IA32IndirectRegister32 destination) {
        assemble0293((byte) 0xF7, (byte) 0x02, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code notb  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code notb      ebx[esi * 4]}
     */
    // Template#: 2999, Serial#: 1352
    public void notb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xF6, (byte) 0x02, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code notl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code notl      ebx[esi * 4]}
     */
    // Template#: 3000, Serial#: 1415
    public void notl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xF7, (byte) 0x02, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code notw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code notw      ebx[esi * 4]}
     */
    // Template#: 3001, Serial#: 1478
    public void notw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0294((byte) 0xF7, (byte) 0x02, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code notw  }<i>destination</i>
     * Example disassembly syntax: {@code notw      ax}
     */
    // Template#: 3002, Serial#: 1530
    public void notw(IA32GeneralRegister16 destination) {
        assemble0188((byte) 0xF7, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code notl  }<i>destination</i>
     * Example disassembly syntax: {@code notl      eax}
     */
    // Template#: 3003, Serial#: 1467
    public void notl(IA32GeneralRegister32 destination) {
        assemble0189((byte) 0xF7, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code notb  }<i>destination</i>
     * Example disassembly syntax: {@code notb      al}
     */
    // Template#: 3004, Serial#: 1404
    public void notb(IA32GeneralRegister8 destination) {
        assemble0297((byte) 0xF6, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code notb  }<i>destination</i>
     * Example disassembly syntax: {@code notb      [ebx]}
     */
    // Template#: 3005, Serial#: 1351
    public void notb(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xF6, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code notl  }<i>destination</i>
     * Example disassembly syntax: {@code notl      [ebx]}
     */
    // Template#: 3006, Serial#: 1414
    public void notl(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xF7, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code notw  }<i>destination</i>
     * Example disassembly syntax: {@code notw      [ebx]}
     */
    // Template#: 3007, Serial#: 1477
    public void notw(IA32IndirectRegister32 destination) {
        assemble0298((byte) 0xF7, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code notb  }<i>m32</i>
     * Example disassembly syntax: {@code notb      [L1: 0x12345678]}
     */
    // Template#: 3008, Serial#: 1354
    public void m_notb(int m32) {
        assemble0299((byte) 0xF6, (byte) 0x02, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code notl  }<i>m32</i>
     * Example disassembly syntax: {@code notl      [L1: 0x12345678]}
     */
    // Template#: 3009, Serial#: 1417
    public void m_notl(int m32) {
        assemble0299((byte) 0xF7, (byte) 0x02, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code notw  }<i>m32</i>
     * Example disassembly syntax: {@code notw      [L1: 0x12345678]}
     */
    // Template#: 3010, Serial#: 1480
    public void m_notw(int m32) {
        assemble0300((byte) 0xF7, (byte) 0x02, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code notb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code notb      ebx[esi * 4 + 305419896]}
     */
    // Template#: 3011, Serial#: 1392
    public void notb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xF6, (byte) 0x02, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code notl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code notl      ebx[esi * 4 + 305419896]}
     */
    // Template#: 3012, Serial#: 1455
    public void notl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xF7, (byte) 0x02, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code notw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code notw      ebx[esi * 4 + 305419896]}
     */
    // Template#: 3013, Serial#: 1518
    public void notw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0301((byte) 0xF7, (byte) 0x02, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code notb  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code notb      L1: 0x12345678[esi * 4]}
     */
    // Template#: 3014, Serial#: 1353
    public void m_notb(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xF6, (byte) 0x02, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code notl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code notl      L1: 0x12345678[esi * 4]}
     */
    // Template#: 3015, Serial#: 1416
    public void m_notl(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xF7, (byte) 0x02, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code notw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code notw      L1: 0x12345678[esi * 4]}
     */
    // Template#: 3016, Serial#: 1479
    public void m_notw(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0302((byte) 0xF7, (byte) 0x02, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code notb  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code notb      [ebx + 305419896]}
     */
    // Template#: 3017, Serial#: 1391
    public void notb(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xF6, (byte) 0x02, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code notl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code notl      [ebx + 305419896]}
     */
    // Template#: 3018, Serial#: 1454
    public void notl(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xF7, (byte) 0x02, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code notw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code notw      [ebx + 305419896]}
     */
    // Template#: 3019, Serial#: 1517
    public void notw(int disp32, IA32IndirectRegister32 destination) {
        assemble0303((byte) 0xF7, (byte) 0x02, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code operand_size  }
     * Example disassembly syntax: {@code operand_size}
     */
    // Template#: 3020, Serial#: 260
    public void operand_size() {
        assemble0001((byte) 0x66);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>imm8</i>
     * Example disassembly syntax: {@code or        al, 0x12}
     */
    // Template#: 3021, Serial#: 1587
    public void or_AL(byte imm8) {
        assemble0002((byte) 0x0C, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code orb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orb       ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 3022, Serial#: 305
    public void orb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0x80, (byte) 0x01, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code orl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orl       ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 3023, Serial#: 521
    public void orl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0x83, (byte) 0x01, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code orw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orw       ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 3024, Serial#: 593
    public void orw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0004((byte) 0x83, (byte) 0x01, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code or        ebx[esi * 4 + 18], ax}
     */
    // Template#: 3025, Serial#: 1559
    public void or(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0005((byte) 0x09, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code or        ebx[esi * 4 + 18], eax}
     */
    // Template#: 3026, Serial#: 1550
    public void or(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0006((byte) 0x09, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code or        ebx[esi * 4 + 18], al}
     */
    // Template#: 3027, Serial#: 1541
    public void or(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0007((byte) 0x08, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code orl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code orl       ebx[esi * 4 + 18], 0x12345678}
     */
    // Template#: 3028, Serial#: 377
    public void orl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0008((byte) 0x81, (byte) 0x01, disp8, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code orw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code orw       ebx[esi * 4 + 18], 0x1234}
     */
    // Template#: 3029, Serial#: 449
    public void orw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0009((byte) 0x81, (byte) 0x01, disp8, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code orb  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orb       [ebx + 18], 0x12}
     */
    // Template#: 3030, Serial#: 304
    public void orb(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0x80, (byte) 0x01, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code orl  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orl       [ebx + 18], 0x12}
     */
    // Template#: 3031, Serial#: 520
    public void orl(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0x83, (byte) 0x01, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code orw  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orw       [ebx + 18], 0x12}
     */
    // Template#: 3032, Serial#: 592
    public void orw(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0011((byte) 0x83, (byte) 0x01, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code or        [ebx + 18], ax}
     */
    // Template#: 3033, Serial#: 1558
    public void or(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0012((byte) 0x09, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code or        [ebx + 18], eax}
     */
    // Template#: 3034, Serial#: 1549
    public void or(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0013((byte) 0x09, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code or        [ebx + 18], al}
     */
    // Template#: 3035, Serial#: 1540
    public void or(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0014((byte) 0x08, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code orl  }<i>disp8</i>, <i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code orl       [ebx + 18], 0x12345678}
     */
    // Template#: 3036, Serial#: 376
    public void orl(byte disp8, IA32IndirectRegister32 destination, int imm32) {
        assemble0015((byte) 0x81, (byte) 0x01, disp8, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code orw  }<i>disp8</i>, <i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code orw       [ebx + 18], 0x1234}
     */
    // Template#: 3037, Serial#: 448
    public void orw(byte disp8, IA32IndirectRegister32 destination, short imm16) {
        assemble0016((byte) 0x81, (byte) 0x01, disp8, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code orb  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orb       ebx[esi * 4], 0x12}
     */
    // Template#: 3038, Serial#: 275
    public void orb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0x80, (byte) 0x01, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code orl  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orl       ebx[esi * 4], 0x12}
     */
    // Template#: 3039, Serial#: 491
    public void orl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0x83, (byte) 0x01, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code orw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orw       ebx[esi * 4], 0x12}
     */
    // Template#: 3040, Serial#: 563
    public void orw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0018((byte) 0x83, (byte) 0x01, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code or        ebx[esi * 4], ax}
     */
    // Template#: 3041, Serial#: 1555
    public void or(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0019((byte) 0x09, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code or        ebx[esi * 4], eax}
     */
    // Template#: 3042, Serial#: 1546
    public void or(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0020((byte) 0x09, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code or        ebx[esi * 4], al}
     */
    // Template#: 3043, Serial#: 1537
    public void or(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0021((byte) 0x08, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code orl  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code orl       ebx[esi * 4], 0x12345678}
     */
    // Template#: 3044, Serial#: 347
    public void orl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0022((byte) 0x81, (byte) 0x01, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code orw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code orw       ebx[esi * 4], 0x1234}
     */
    // Template#: 3045, Serial#: 419
    public void orw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0023((byte) 0x81, (byte) 0x01, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code orw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orw       ax, 0x12}
     */
    // Template#: 3046, Serial#: 623
    public void orw(IA32GeneralRegister16 destination, byte imm8) {
        assemble0024((byte) 0x83, (byte) 0x01, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code or        ax, ebx[esi * 4 + 18]}
     */
    // Template#: 3047, Serial#: 1584
    public void or(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0025((byte) 0x0B, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code or        ax, [ebx + 18]}
     */
    // Template#: 3048, Serial#: 1583
    public void or(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0026((byte) 0x0B, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code or        ax, ebx[esi * 4]}
     */
    // Template#: 3049, Serial#: 1580
    public void or(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0027((byte) 0x0B, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code or        ax, ax}
     */
    // Template#: 3050, Serial#: 1562
    public void or(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0028((byte) 0x09, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code or        ax, [ebx]}
     */
    // Template#: 3051, Serial#: 1579
    public void or(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0029((byte) 0x0B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code or        ax, [L1: 0x12345678]}
     */
    // Template#: 3052, Serial#: 1582
    public void m_or(IA32GeneralRegister16 destination, int m32) {
        assemble0030((byte) 0x0B, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code or        ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3053, Serial#: 1586
    public void or(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0031((byte) 0x0B, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code or        ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3054, Serial#: 1581
    public void m_or(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0032((byte) 0x0B, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code or        ax, [ebx + 305419896]}
     */
    // Template#: 3055, Serial#: 1585
    public void or(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0033((byte) 0x0B, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code orw  }<i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code orw       ax, 0x1234}
     */
    // Template#: 3056, Serial#: 479
    public void orw(IA32GeneralRegister16 destination, short imm16) {
        assemble0034((byte) 0x81, (byte) 0x01, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code orl  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orl       eax, 0x12}
     */
    // Template#: 3057, Serial#: 551
    public void orl(IA32GeneralRegister32 destination, byte imm8) {
        assemble0035((byte) 0x83, (byte) 0x01, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code or        eax, ebx[esi * 4 + 18]}
     */
    // Template#: 3058, Serial#: 1576
    public void or(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0036((byte) 0x0B, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code or        eax, [ebx + 18]}
     */
    // Template#: 3059, Serial#: 1575
    public void or(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0037((byte) 0x0B, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code or        eax, ebx[esi * 4]}
     */
    // Template#: 3060, Serial#: 1572
    public void or(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0038((byte) 0x0B, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code or        eax, eax}
     */
    // Template#: 3061, Serial#: 1553
    public void or(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0039((byte) 0x09, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code or        eax, [ebx]}
     */
    // Template#: 3062, Serial#: 1571
    public void or(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0040((byte) 0x0B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code orl  }<i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code orl       eax, 0x12345678}
     */
    // Template#: 3063, Serial#: 407
    public void orl(IA32GeneralRegister32 destination, int imm32) {
        assemble0042((byte) 0x81, (byte) 0x01, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code or        eax, [L1: 0x12345678]}
     */
    // Template#: 3064, Serial#: 1574
    public void m_or(IA32GeneralRegister32 destination, int m32) {
        assemble0041((byte) 0x0B, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code or        eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3065, Serial#: 1578
    public void or(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0043((byte) 0x0B, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code or        eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3066, Serial#: 1573
    public void m_or(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0044((byte) 0x0B, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code or        eax, [ebx + 305419896]}
     */
    // Template#: 3067, Serial#: 1577
    public void or(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0045((byte) 0x0B, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code orb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orb       al, 0x12}
     */
    // Template#: 3068, Serial#: 335
    public void orb(IA32GeneralRegister8 destination, byte imm8) {
        assemble0046((byte) 0x80, (byte) 0x01, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code or        al, ebx[esi * 4 + 18]}
     */
    // Template#: 3069, Serial#: 1568
    public void or(IA32GeneralRegister8 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0047((byte) 0x0A, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code or        al, [ebx + 18]}
     */
    // Template#: 3070, Serial#: 1567
    public void or(IA32GeneralRegister8 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0048((byte) 0x0A, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code or        al, ebx[esi * 4]}
     */
    // Template#: 3071, Serial#: 1564
    public void or(IA32GeneralRegister8 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0049((byte) 0x0A, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code or        al, al}
     */
    // Template#: 3072, Serial#: 1544
    public void or(IA32GeneralRegister8 destination, IA32GeneralRegister8 source) {
        assemble0050((byte) 0x08, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code or        al, [ebx]}
     */
    // Template#: 3073, Serial#: 1563
    public void or(IA32GeneralRegister8 destination, IA32IndirectRegister32 source) {
        assemble0051((byte) 0x0A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code or        al, [L1: 0x12345678]}
     */
    // Template#: 3074, Serial#: 1566
    public void m_or(IA32GeneralRegister8 destination, int m32) {
        assemble0052((byte) 0x0A, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code or        al, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3075, Serial#: 1570
    public void or(IA32GeneralRegister8 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0053((byte) 0x0A, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code or        al, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3076, Serial#: 1565
    public void m_or(IA32GeneralRegister8 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0054((byte) 0x0A, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code or        al, [ebx + 305419896]}
     */
    // Template#: 3077, Serial#: 1569
    public void or(IA32GeneralRegister8 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0055((byte) 0x0A, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code orb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orb       [ebx], 0x12}
     */
    // Template#: 3078, Serial#: 274
    public void orb(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0x80, (byte) 0x01, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code orl  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orl       [ebx], 0x12}
     */
    // Template#: 3079, Serial#: 490
    public void orl(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0x83, (byte) 0x01, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code orw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orw       [ebx], 0x12}
     */
    // Template#: 3080, Serial#: 562
    public void orw(IA32IndirectRegister32 destination, byte imm8) {
        assemble0057((byte) 0x83, (byte) 0x01, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code or        [ebx], ax}
     */
    // Template#: 3081, Serial#: 1554
    public void or(IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0058((byte) 0x09, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code or        [ebx], eax}
     */
    // Template#: 3082, Serial#: 1545
    public void or(IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0059((byte) 0x09, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code or        [ebx], al}
     */
    // Template#: 3083, Serial#: 1536
    public void or(IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0060((byte) 0x08, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code orl  }<i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code orl       [ebx], 0x12345678}
     */
    // Template#: 3084, Serial#: 346
    public void orl(IA32IndirectRegister32 destination, int imm32) {
        assemble0061((byte) 0x81, (byte) 0x01, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code orw  }<i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code orw       [ebx], 0x1234}
     */
    // Template#: 3085, Serial#: 418
    public void orw(IA32IndirectRegister32 destination, short imm16) {
        assemble0062((byte) 0x81, (byte) 0x01, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code orl  }<i>imm32</i>
     * Example disassembly syntax: {@code orl       eax, 0x12345678}
     */
    // Template#: 3086, Serial#: 1588
    public void or_EAX(int imm32) {
        assemble0063((byte) 0x0D, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code orb  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orb       [L1: 0x12345678], 0x12}
     */
    // Template#: 3087, Serial#: 277
    public void m_orb(int m32, byte imm8) {
        assemble0064((byte) 0x80, (byte) 0x01, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code orl  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orl       [L1: 0x12345678], 0x12}
     */
    // Template#: 3088, Serial#: 493
    public void m_orl(int m32, byte imm8) {
        assemble0064((byte) 0x83, (byte) 0x01, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code orw  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orw       [L1: 0x12345678], 0x12}
     */
    // Template#: 3089, Serial#: 565
    public void m_orw(int m32, byte imm8) {
        assemble0065((byte) 0x83, (byte) 0x01, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code orb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orb       ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 3090, Serial#: 321
    public void orb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0x80, (byte) 0x01, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code orl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orl       ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 3091, Serial#: 537
    public void orl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0x83, (byte) 0x01, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code orw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orw       ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 3092, Serial#: 609
    public void orw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0067((byte) 0x83, (byte) 0x01, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code or        ebx[esi * 4 + 305419896], ax}
     */
    // Template#: 3093, Serial#: 1561
    public void or(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0068((byte) 0x09, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code or        ebx[esi * 4 + 305419896], eax}
     */
    // Template#: 3094, Serial#: 1552
    public void or(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0069((byte) 0x09, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code or        ebx[esi * 4 + 305419896], al}
     */
    // Template#: 3095, Serial#: 1543
    public void or(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0070((byte) 0x08, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code orl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code orl       ebx[esi * 4 + 305419896], 0x12345678}
     */
    // Template#: 3096, Serial#: 393
    public void orl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0071((byte) 0x81, (byte) 0x01, disp32, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code orw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code orw       ebx[esi * 4 + 305419896], 0x1234}
     */
    // Template#: 3097, Serial#: 465
    public void orw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0072((byte) 0x81, (byte) 0x01, disp32, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code or        [L1: 0x12345678], ax}
     */
    // Template#: 3098, Serial#: 1557
    public void m_or(int m32, IA32GeneralRegister16 source) {
        assemble0073((byte) 0x09, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code or        [L1: 0x12345678], eax}
     */
    // Template#: 3099, Serial#: 1548
    public void m_or(int m32, IA32GeneralRegister32 source) {
        assemble0074((byte) 0x09, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code or        [L1: 0x12345678], al}
     */
    // Template#: 3100, Serial#: 1539
    public void m_or(int m32, IA32GeneralRegister8 source) {
        assemble0075((byte) 0x08, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code orb  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orb       L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 3101, Serial#: 276
    public void m_orb(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0x80, (byte) 0x01, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code orl  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orl       L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 3102, Serial#: 492
    public void m_orl(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0x83, (byte) 0x01, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code orw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orw       L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 3103, Serial#: 564
    public void m_orw(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0077((byte) 0x83, (byte) 0x01, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code or        L1: 0x12345678[esi * 4], ax}
     */
    // Template#: 3104, Serial#: 1556
    public void m_or(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0078((byte) 0x09, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code or        L1: 0x12345678[esi * 4], eax}
     */
    // Template#: 3105, Serial#: 1547
    public void m_or(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0079((byte) 0x09, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code or        L1: 0x12345678[esi * 4], al}
     */
    // Template#: 3106, Serial#: 1538
    public void m_or(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0080((byte) 0x08, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code orl  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code orl       L1: 0x12345678[esi * 4], 0x12345678}
     */
    // Template#: 3107, Serial#: 348
    public void m_orl(int m32, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0081((byte) 0x81, (byte) 0x01, m32, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code orw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code orw       L1: 0x12345678[esi * 4], 0x1234}
     */
    // Template#: 3108, Serial#: 420
    public void m_orw(int m32, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0082((byte) 0x81, (byte) 0x01, m32, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code orb  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orb       [ebx + 305419896], 0x12}
     */
    // Template#: 3109, Serial#: 320
    public void orb(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0x80, (byte) 0x01, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code orl  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orl       [ebx + 305419896], 0x12}
     */
    // Template#: 3110, Serial#: 536
    public void orl(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0x83, (byte) 0x01, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code orw  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code orw       [ebx + 305419896], 0x12}
     */
    // Template#: 3111, Serial#: 608
    public void orw(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0084((byte) 0x83, (byte) 0x01, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code or        [ebx + 305419896], ax}
     */
    // Template#: 3112, Serial#: 1560
    public void or(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0085((byte) 0x09, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code or        [ebx + 305419896], eax}
     */
    // Template#: 3113, Serial#: 1551
    public void or(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0086((byte) 0x09, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code or  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code or        [ebx + 305419896], al}
     */
    // Template#: 3114, Serial#: 1542
    public void or(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0087((byte) 0x08, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code orl  }<i>disp32</i>, <i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code orl       [ebx + 305419896], 0x12345678}
     */
    // Template#: 3115, Serial#: 392
    public void orl(int disp32, IA32IndirectRegister32 destination, int imm32) {
        assemble0088((byte) 0x81, (byte) 0x01, disp32, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code orw  }<i>disp32</i>, <i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code orw       [ebx + 305419896], 0x1234}
     */
    // Template#: 3116, Serial#: 464
    public void orw(int disp32, IA32IndirectRegister32 destination, short imm16) {
        assemble0089((byte) 0x81, (byte) 0x01, disp32, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code orl  }<i>m32</i>, <i>imm32</i>
     * Example disassembly syntax: {@code orl       [L1: 0x12345678], 0x12345678}
     */
    // Template#: 3117, Serial#: 349
    public void m_orl(int m32, int imm32) {
        assemble0090((byte) 0x81, (byte) 0x01, m32, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code orw  }<i>m32</i>, <i>imm16</i>
     * Example disassembly syntax: {@code orw       [L1: 0x12345678], 0x1234}
     */
    // Template#: 3118, Serial#: 421
    public void m_orw(int m32, short imm16) {
        assemble0091((byte) 0x81, (byte) 0x01, m32, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code orw  }<i>imm16</i>
     * Example disassembly syntax: {@code orw       ax, 0x1234}
     */
    // Template#: 3119, Serial#: 1589
    public void or_AX(short imm16) {
        assemble0092((byte) 0x0D, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code orpd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code orpd      xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3120, Serial#: 3078
    public void orpd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x56, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code orpd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code orpd      xmm0, [ebx + 18]}
     */
    // Template#: 3121, Serial#: 3077
    public void orpd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x56, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code orpd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code orpd      xmm0, ebx[esi * 4]}
     */
    // Template#: 3122, Serial#: 3074
    public void orpd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x56, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code orpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code orpd      xmm0, [ebx]}
     */
    // Template#: 3123, Serial#: 3073
    public void orpd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x56, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code orpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code orpd      xmm0, xmm0}
     */
    // Template#: 3124, Serial#: 3081
    public void orpd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x56, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code orpd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code orpd      xmm0, [L1: 0x12345678]}
     */
    // Template#: 3125, Serial#: 3076
    public void m_orpd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x56, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code orpd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code orpd      xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3126, Serial#: 3080
    public void orpd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x56, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code orpd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code orpd      xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3127, Serial#: 3075
    public void m_orpd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x56, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code orpd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code orpd      xmm0, [ebx + 305419896]}
     */
    // Template#: 3128, Serial#: 3079
    public void orpd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x56, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code orps  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code orps      xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3129, Serial#: 3032
    public void orps(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0102((byte) 0x56, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code orps  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code orps      xmm0, [ebx + 18]}
     */
    // Template#: 3130, Serial#: 3031
    public void orps(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0103((byte) 0x56, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code orps  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code orps      xmm0, ebx[esi * 4]}
     */
    // Template#: 3131, Serial#: 3028
    public void orps(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0104((byte) 0x56, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code orps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code orps      xmm0, [ebx]}
     */
    // Template#: 3132, Serial#: 3027
    public void orps(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0105((byte) 0x56, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code orps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code orps      xmm0, xmm0}
     */
    // Template#: 3133, Serial#: 3035
    public void orps(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0106((byte) 0x56, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code orps  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code orps      xmm0, [L1: 0x12345678]}
     */
    // Template#: 3134, Serial#: 3030
    public void m_orps(IA32XMMRegister destination, int m32) {
        assemble0107((byte) 0x56, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code orps  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code orps      xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3135, Serial#: 3034
    public void orps(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0108((byte) 0x56, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code orps  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code orps      xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3136, Serial#: 3029
    public void m_orps(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0109((byte) 0x56, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code orps  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code orps      xmm0, [ebx + 305419896]}
     */
    // Template#: 3137, Serial#: 3033
    public void orps(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0110((byte) 0x56, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code out  }
     * Example disassembly syntax: {@code out       dx, al}
     */
    // Template#: 3138, Serial#: 2407
    public void out_DX_AL() {
        assemble0001((byte) 0xEE);
    }

    /**
     * Pseudo-external assembler syntax: {@code out  }
     * Example disassembly syntax: {@code out       dx, eax}
     */
    // Template#: 3139, Serial#: 2408
    public void out_DX_EAX() {
        assemble0001((byte) 0xEF);
    }

    /**
     * Pseudo-external assembler syntax: {@code out  }
     * Example disassembly syntax: {@code out       dx, ax}
     */
    // Template#: 3140, Serial#: 2409
    public void out_DX_AX() {
        assemble0197((byte) 0xEF);
    }

    /**
     * Pseudo-external assembler syntax: {@code out  }<i>imm8</i>
     * Example disassembly syntax: {@code out       0x12, al}
     */
    // Template#: 3141, Serial#: 1339
    public void out___AL(byte imm8) {
        assemble0002((byte) 0xE6, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code out  }<i>imm8</i>
     * Example disassembly syntax: {@code out       0x12, eax}
     */
    // Template#: 3142, Serial#: 1340
    public void out___EAX(byte imm8) {
        assemble0002((byte) 0xE7, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code out  }<i>imm8</i>
     * Example disassembly syntax: {@code out       0x12, ax}
     */
    // Template#: 3143, Serial#: 1341
    public void out___AX(byte imm8) {
        assemble0354((byte) 0xE7, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code outsb  }
     * Example disassembly syntax: {@code outsb   }
     */
    // Template#: 3144, Serial#: 1805
    public void outsb() {
        assemble0001((byte) 0x6E);
    }

    /**
     * Pseudo-external assembler syntax: {@code outsl  }
     * Example disassembly syntax: {@code outsl   }
     */
    // Template#: 3145, Serial#: 1806
    public void outsl() {
        assemble0001((byte) 0x6F);
    }

    /**
     * Pseudo-external assembler syntax: {@code outsw  }
     * Example disassembly syntax: {@code outsw   }
     */
    // Template#: 3146, Serial#: 1807
    public void outsw() {
        assemble0197((byte) 0x6F);
    }

    /**
     * Pseudo-external assembler syntax: {@code packssdw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code packssdw  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3147, Serial#: 4735
    public void packssdw(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x6B, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code packssdw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code packssdw  xmm0, [ebx + 18]}
     */
    // Template#: 3148, Serial#: 4734
    public void packssdw(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x6B, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code packssdw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code packssdw  xmm0, ebx[esi * 4]}
     */
    // Template#: 3149, Serial#: 4731
    public void packssdw(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x6B, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code packssdw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code packssdw  xmm0, [ebx]}
     */
    // Template#: 3150, Serial#: 4730
    public void packssdw(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x6B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code packssdw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code packssdw  xmm0, xmm0}
     */
    // Template#: 3151, Serial#: 4738
    public void packssdw(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x6B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code packssdw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code packssdw  xmm0, [L1: 0x12345678]}
     */
    // Template#: 3152, Serial#: 4733
    public void m_packssdw(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x6B, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code packssdw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code packssdw  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3153, Serial#: 4737
    public void packssdw(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x6B, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code packssdw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code packssdw  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3154, Serial#: 4732
    public void m_packssdw(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x6B, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code packssdw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code packssdw  xmm0, [ebx + 305419896]}
     */
    // Template#: 3155, Serial#: 4736
    public void packssdw(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x6B, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code packssdw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code packssdw  mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3156, Serial#: 4681
    public void packssdw(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0x6B, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code packssdw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code packssdw  mm0, [ebx + 18]}
     */
    // Template#: 3157, Serial#: 4680
    public void packssdw(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0x6B, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code packssdw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code packssdw  mm0, ebx[esi * 4]}
     */
    // Template#: 3158, Serial#: 4677
    public void packssdw(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0x6B, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code packssdw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code packssdw  mm0, [ebx]}
     */
    // Template#: 3159, Serial#: 4676
    public void packssdw(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0x6B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code packssdw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code packssdw  mm0, mm0}
     */
    // Template#: 3160, Serial#: 4684
    public void packssdw(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0x6B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code packssdw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code packssdw  mm0, [L1: 0x12345678]}
     */
    // Template#: 3161, Serial#: 4679
    public void m_packssdw(MMXRegister destination, int m32) {
        assemble0268((byte) 0x6B, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code packssdw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code packssdw  mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3162, Serial#: 4683
    public void packssdw(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0x6B, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code packssdw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code packssdw  mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3163, Serial#: 4678
    public void m_packssdw(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0x6B, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code packssdw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code packssdw  mm0, [ebx + 305419896]}
     */
    // Template#: 3164, Serial#: 4682
    public void packssdw(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0x6B, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code packsswb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code packsswb  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3165, Serial#: 3276
    public void packsswb(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x63, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code packsswb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code packsswb  xmm0, [ebx + 18]}
     */
    // Template#: 3166, Serial#: 3275
    public void packsswb(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x63, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code packsswb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code packsswb  xmm0, ebx[esi * 4]}
     */
    // Template#: 3167, Serial#: 3272
    public void packsswb(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x63, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code packsswb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code packsswb  xmm0, [ebx]}
     */
    // Template#: 3168, Serial#: 3271
    public void packsswb(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x63, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code packsswb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code packsswb  xmm0, xmm0}
     */
    // Template#: 3169, Serial#: 3279
    public void packsswb(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x63, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code packsswb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code packsswb  xmm0, [L1: 0x12345678]}
     */
    // Template#: 3170, Serial#: 3274
    public void m_packsswb(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x63, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code packsswb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code packsswb  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3171, Serial#: 3278
    public void packsswb(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x63, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code packsswb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code packsswb  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3172, Serial#: 3273
    public void m_packsswb(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x63, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code packsswb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code packsswb  xmm0, [ebx + 305419896]}
     */
    // Template#: 3173, Serial#: 3277
    public void packsswb(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x63, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code packsswb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code packsswb  mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3174, Serial#: 3159
    public void packsswb(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0x63, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code packsswb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code packsswb  mm0, [ebx + 18]}
     */
    // Template#: 3175, Serial#: 3158
    public void packsswb(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0x63, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code packsswb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code packsswb  mm0, ebx[esi * 4]}
     */
    // Template#: 3176, Serial#: 3155
    public void packsswb(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0x63, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code packsswb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code packsswb  mm0, [ebx]}
     */
    // Template#: 3177, Serial#: 3154
    public void packsswb(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0x63, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code packsswb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code packsswb  mm0, mm0}
     */
    // Template#: 3178, Serial#: 3162
    public void packsswb(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0x63, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code packsswb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code packsswb  mm0, [L1: 0x12345678]}
     */
    // Template#: 3179, Serial#: 3157
    public void m_packsswb(MMXRegister destination, int m32) {
        assemble0268((byte) 0x63, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code packsswb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code packsswb  mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3180, Serial#: 3161
    public void packsswb(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0x63, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code packsswb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code packsswb  mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3181, Serial#: 3156
    public void m_packsswb(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0x63, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code packsswb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code packsswb  mm0, [ebx + 305419896]}
     */
    // Template#: 3182, Serial#: 3160
    public void packsswb(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0x63, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code packuswb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code packuswb  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3183, Serial#: 3312
    public void packuswb(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x67, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code packuswb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code packuswb  xmm0, [ebx + 18]}
     */
    // Template#: 3184, Serial#: 3311
    public void packuswb(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x67, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code packuswb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code packuswb  xmm0, ebx[esi * 4]}
     */
    // Template#: 3185, Serial#: 3308
    public void packuswb(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x67, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code packuswb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code packuswb  xmm0, [ebx]}
     */
    // Template#: 3186, Serial#: 3307
    public void packuswb(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x67, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code packuswb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code packuswb  xmm0, xmm0}
     */
    // Template#: 3187, Serial#: 3315
    public void packuswb(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x67, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code packuswb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code packuswb  xmm0, [L1: 0x12345678]}
     */
    // Template#: 3188, Serial#: 3310
    public void m_packuswb(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x67, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code packuswb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code packuswb  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3189, Serial#: 3314
    public void packuswb(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x67, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code packuswb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code packuswb  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3190, Serial#: 3309
    public void m_packuswb(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x67, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code packuswb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code packuswb  xmm0, [ebx + 305419896]}
     */
    // Template#: 3191, Serial#: 3313
    public void packuswb(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x67, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code packuswb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code packuswb  mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3192, Serial#: 3195
    public void packuswb(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0x67, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code packuswb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code packuswb  mm0, [ebx + 18]}
     */
    // Template#: 3193, Serial#: 3194
    public void packuswb(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0x67, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code packuswb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code packuswb  mm0, ebx[esi * 4]}
     */
    // Template#: 3194, Serial#: 3191
    public void packuswb(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0x67, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code packuswb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code packuswb  mm0, [ebx]}
     */
    // Template#: 3195, Serial#: 3190
    public void packuswb(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0x67, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code packuswb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code packuswb  mm0, mm0}
     */
    // Template#: 3196, Serial#: 3198
    public void packuswb(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0x67, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code packuswb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code packuswb  mm0, [L1: 0x12345678]}
     */
    // Template#: 3197, Serial#: 3193
    public void m_packuswb(MMXRegister destination, int m32) {
        assemble0268((byte) 0x67, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code packuswb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code packuswb  mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3198, Serial#: 3197
    public void packuswb(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0x67, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code packuswb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code packuswb  mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3199, Serial#: 3192
    public void m_packuswb(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0x67, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code packuswb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code packuswb  mm0, [ebx + 305419896]}
     */
    // Template#: 3200, Serial#: 3196
    public void packuswb(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0x67, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddb     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3201, Serial#: 5506
    public void paddb(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xFC, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code paddb     xmm0, [ebx + 18]}
     */
    // Template#: 3202, Serial#: 5505
    public void paddb(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xFC, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddb     xmm0, ebx[esi * 4]}
     */
    // Template#: 3203, Serial#: 5502
    public void paddb(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xFC, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddb     xmm0, [ebx]}
     */
    // Template#: 3204, Serial#: 5501
    public void paddb(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xFC, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddb     xmm0, xmm0}
     */
    // Template#: 3205, Serial#: 5509
    public void paddb(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xFC, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code paddb     xmm0, [L1: 0x12345678]}
     */
    // Template#: 3206, Serial#: 5504
    public void m_paddb(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xFC, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddb     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3207, Serial#: 5508
    public void paddb(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xFC, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddb     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3208, Serial#: 5503
    public void m_paddb(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xFC, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code paddb     xmm0, [ebx + 305419896]}
     */
    // Template#: 3209, Serial#: 5507
    public void paddb(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xFC, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddb     mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3210, Serial#: 5443
    public void paddb(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xFC, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code paddb     mm0, [ebx + 18]}
     */
    // Template#: 3211, Serial#: 5442
    public void paddb(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xFC, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddb     mm0, ebx[esi * 4]}
     */
    // Template#: 3212, Serial#: 5439
    public void paddb(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xFC, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddb     mm0, [ebx]}
     */
    // Template#: 3213, Serial#: 5438
    public void paddb(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xFC, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddb     mm0, mm0}
     */
    // Template#: 3214, Serial#: 5446
    public void paddb(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xFC, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code paddb     mm0, [L1: 0x12345678]}
     */
    // Template#: 3215, Serial#: 5441
    public void m_paddb(MMXRegister destination, int m32) {
        assemble0268((byte) 0xFC, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddb     mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3216, Serial#: 5445
    public void paddb(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xFC, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddb     mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3217, Serial#: 5440
    public void m_paddb(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xFC, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code paddb     mm0, [ebx + 305419896]}
     */
    // Template#: 3218, Serial#: 5444
    public void paddb(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xFC, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddd     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3219, Serial#: 5524
    public void paddd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xFE, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code paddd     xmm0, [ebx + 18]}
     */
    // Template#: 3220, Serial#: 5523
    public void paddd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xFE, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddd     xmm0, ebx[esi * 4]}
     */
    // Template#: 3221, Serial#: 5520
    public void paddd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xFE, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddd     xmm0, [ebx]}
     */
    // Template#: 3222, Serial#: 5519
    public void paddd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xFE, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddd     xmm0, xmm0}
     */
    // Template#: 3223, Serial#: 5527
    public void paddd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xFE, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code paddd     xmm0, [L1: 0x12345678]}
     */
    // Template#: 3224, Serial#: 5522
    public void m_paddd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xFE, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddd     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3225, Serial#: 5526
    public void paddd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xFE, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddd     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3226, Serial#: 5521
    public void m_paddd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xFE, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code paddd     xmm0, [ebx + 305419896]}
     */
    // Template#: 3227, Serial#: 5525
    public void paddd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xFE, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddd     mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3228, Serial#: 5461
    public void paddd(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xFE, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code paddd     mm0, [ebx + 18]}
     */
    // Template#: 3229, Serial#: 5460
    public void paddd(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xFE, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddd     mm0, ebx[esi * 4]}
     */
    // Template#: 3230, Serial#: 5457
    public void paddd(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xFE, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddd     mm0, [ebx]}
     */
    // Template#: 3231, Serial#: 5456
    public void paddd(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xFE, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddd     mm0, mm0}
     */
    // Template#: 3232, Serial#: 5464
    public void paddd(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xFE, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code paddd     mm0, [L1: 0x12345678]}
     */
    // Template#: 3233, Serial#: 5459
    public void m_paddd(MMXRegister destination, int m32) {
        assemble0268((byte) 0xFE, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddd     mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3234, Serial#: 5463
    public void paddd(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xFE, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddd     mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3235, Serial#: 5458
    public void m_paddd(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xFE, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code paddd     mm0, [ebx + 305419896]}
     */
    // Template#: 3236, Serial#: 5462
    public void paddd(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xFE, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddq  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddq     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3237, Serial#: 3789
    public void paddq(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xD4, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddq  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code paddq     xmm0, [ebx + 18]}
     */
    // Template#: 3238, Serial#: 3788
    public void paddq(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xD4, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddq  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddq     xmm0, ebx[esi * 4]}
     */
    // Template#: 3239, Serial#: 3785
    public void paddq(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xD4, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddq     xmm0, [ebx]}
     */
    // Template#: 3240, Serial#: 3784
    public void paddq(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xD4, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddq     xmm0, xmm0}
     */
    // Template#: 3241, Serial#: 3792
    public void paddq(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xD4, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddq  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code paddq     xmm0, [L1: 0x12345678]}
     */
    // Template#: 3242, Serial#: 3787
    public void m_paddq(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xD4, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddq  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddq     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3243, Serial#: 3791
    public void paddq(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xD4, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddq  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddq     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3244, Serial#: 3786
    public void m_paddq(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xD4, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddq  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code paddq     xmm0, [ebx + 305419896]}
     */
    // Template#: 3245, Serial#: 3790
    public void paddq(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xD4, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddsb    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3246, Serial#: 5380
    public void paddsb(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xEC, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code paddsb    xmm0, [ebx + 18]}
     */
    // Template#: 3247, Serial#: 5379
    public void paddsb(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xEC, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddsb    xmm0, ebx[esi * 4]}
     */
    // Template#: 3248, Serial#: 5376
    public void paddsb(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xEC, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddsb    xmm0, [ebx]}
     */
    // Template#: 3249, Serial#: 5375
    public void paddsb(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xEC, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddsb    xmm0, xmm0}
     */
    // Template#: 3250, Serial#: 5383
    public void paddsb(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xEC, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code paddsb    xmm0, [L1: 0x12345678]}
     */
    // Template#: 3251, Serial#: 5378
    public void m_paddsb(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xEC, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddsb    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3252, Serial#: 5382
    public void paddsb(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xEC, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddsb    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3253, Serial#: 5377
    public void m_paddsb(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xEC, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code paddsb    xmm0, [ebx + 305419896]}
     */
    // Template#: 3254, Serial#: 5381
    public void paddsb(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xEC, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddsb    mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3255, Serial#: 5317
    public void paddsb(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xEC, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code paddsb    mm0, [ebx + 18]}
     */
    // Template#: 3256, Serial#: 5316
    public void paddsb(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xEC, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddsb    mm0, ebx[esi * 4]}
     */
    // Template#: 3257, Serial#: 5313
    public void paddsb(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xEC, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddsb    mm0, [ebx]}
     */
    // Template#: 3258, Serial#: 5312
    public void paddsb(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xEC, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddsb    mm0, mm0}
     */
    // Template#: 3259, Serial#: 5320
    public void paddsb(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xEC, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code paddsb    mm0, [L1: 0x12345678]}
     */
    // Template#: 3260, Serial#: 5315
    public void m_paddsb(MMXRegister destination, int m32) {
        assemble0268((byte) 0xEC, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddsb    mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3261, Serial#: 5319
    public void paddsb(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xEC, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddsb    mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3262, Serial#: 5314
    public void m_paddsb(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xEC, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code paddsb    mm0, [ebx + 305419896]}
     */
    // Template#: 3263, Serial#: 5318
    public void paddsb(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xEC, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddsw    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3264, Serial#: 5389
    public void paddsw(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xED, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code paddsw    xmm0, [ebx + 18]}
     */
    // Template#: 3265, Serial#: 5388
    public void paddsw(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xED, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddsw    xmm0, ebx[esi * 4]}
     */
    // Template#: 3266, Serial#: 5385
    public void paddsw(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xED, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddsw    xmm0, [ebx]}
     */
    // Template#: 3267, Serial#: 5384
    public void paddsw(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xED, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddsw    xmm0, xmm0}
     */
    // Template#: 3268, Serial#: 5392
    public void paddsw(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xED, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code paddsw    xmm0, [L1: 0x12345678]}
     */
    // Template#: 3269, Serial#: 5387
    public void m_paddsw(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xED, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddsw    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3270, Serial#: 5391
    public void paddsw(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xED, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddsw    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3271, Serial#: 5386
    public void m_paddsw(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xED, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code paddsw    xmm0, [ebx + 305419896]}
     */
    // Template#: 3272, Serial#: 5390
    public void paddsw(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xED, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddsw    mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3273, Serial#: 5326
    public void paddsw(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xED, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code paddsw    mm0, [ebx + 18]}
     */
    // Template#: 3274, Serial#: 5325
    public void paddsw(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xED, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddsw    mm0, ebx[esi * 4]}
     */
    // Template#: 3275, Serial#: 5322
    public void paddsw(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xED, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddsw    mm0, [ebx]}
     */
    // Template#: 3276, Serial#: 5321
    public void paddsw(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xED, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddsw    mm0, mm0}
     */
    // Template#: 3277, Serial#: 5329
    public void paddsw(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xED, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code paddsw    mm0, [L1: 0x12345678]}
     */
    // Template#: 3278, Serial#: 5324
    public void m_paddsw(MMXRegister destination, int m32) {
        assemble0268((byte) 0xED, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddsw    mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3279, Serial#: 5328
    public void paddsw(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xED, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddsw    mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3280, Serial#: 5323
    public void m_paddsw(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xED, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddsw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code paddsw    mm0, [ebx + 305419896]}
     */
    // Template#: 3281, Serial#: 5327
    public void paddsw(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xED, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddusb   xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3282, Serial#: 5254
    public void paddusb(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xDC, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code paddusb   xmm0, [ebx + 18]}
     */
    // Template#: 3283, Serial#: 5253
    public void paddusb(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xDC, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddusb   xmm0, ebx[esi * 4]}
     */
    // Template#: 3284, Serial#: 5250
    public void paddusb(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xDC, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddusb   xmm0, [ebx]}
     */
    // Template#: 3285, Serial#: 5249
    public void paddusb(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xDC, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddusb   xmm0, xmm0}
     */
    // Template#: 3286, Serial#: 5257
    public void paddusb(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xDC, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code paddusb   xmm0, [L1: 0x12345678]}
     */
    // Template#: 3287, Serial#: 5252
    public void m_paddusb(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xDC, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddusb   xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3288, Serial#: 5256
    public void paddusb(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xDC, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddusb   xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3289, Serial#: 5251
    public void m_paddusb(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xDC, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code paddusb   xmm0, [ebx + 305419896]}
     */
    // Template#: 3290, Serial#: 5255
    public void paddusb(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xDC, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddusb   mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3291, Serial#: 5191
    public void paddusb(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xDC, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code paddusb   mm0, [ebx + 18]}
     */
    // Template#: 3292, Serial#: 5190
    public void paddusb(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xDC, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddusb   mm0, ebx[esi * 4]}
     */
    // Template#: 3293, Serial#: 5187
    public void paddusb(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xDC, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddusb   mm0, [ebx]}
     */
    // Template#: 3294, Serial#: 5186
    public void paddusb(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xDC, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddusb   mm0, mm0}
     */
    // Template#: 3295, Serial#: 5194
    public void paddusb(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xDC, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code paddusb   mm0, [L1: 0x12345678]}
     */
    // Template#: 3296, Serial#: 5189
    public void m_paddusb(MMXRegister destination, int m32) {
        assemble0268((byte) 0xDC, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddusb   mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3297, Serial#: 5193
    public void paddusb(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xDC, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddusb   mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3298, Serial#: 5188
    public void m_paddusb(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xDC, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code paddusb   mm0, [ebx + 305419896]}
     */
    // Template#: 3299, Serial#: 5192
    public void paddusb(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xDC, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddusw   xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3300, Serial#: 5263
    public void paddusw(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xDD, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code paddusw   xmm0, [ebx + 18]}
     */
    // Template#: 3301, Serial#: 5262
    public void paddusw(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xDD, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddusw   xmm0, ebx[esi * 4]}
     */
    // Template#: 3302, Serial#: 5259
    public void paddusw(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xDD, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddusw   xmm0, [ebx]}
     */
    // Template#: 3303, Serial#: 5258
    public void paddusw(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xDD, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddusw   xmm0, xmm0}
     */
    // Template#: 3304, Serial#: 5266
    public void paddusw(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xDD, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code paddusw   xmm0, [L1: 0x12345678]}
     */
    // Template#: 3305, Serial#: 5261
    public void m_paddusw(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xDD, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddusw   xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3306, Serial#: 5265
    public void paddusw(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xDD, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddusw   xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3307, Serial#: 5260
    public void m_paddusw(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xDD, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code paddusw   xmm0, [ebx + 305419896]}
     */
    // Template#: 3308, Serial#: 5264
    public void paddusw(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xDD, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddusw   mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3309, Serial#: 5200
    public void paddusw(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xDD, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code paddusw   mm0, [ebx + 18]}
     */
    // Template#: 3310, Serial#: 5199
    public void paddusw(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xDD, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddusw   mm0, ebx[esi * 4]}
     */
    // Template#: 3311, Serial#: 5196
    public void paddusw(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xDD, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddusw   mm0, [ebx]}
     */
    // Template#: 3312, Serial#: 5195
    public void paddusw(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xDD, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddusw   mm0, mm0}
     */
    // Template#: 3313, Serial#: 5203
    public void paddusw(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xDD, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code paddusw   mm0, [L1: 0x12345678]}
     */
    // Template#: 3314, Serial#: 5198
    public void m_paddusw(MMXRegister destination, int m32) {
        assemble0268((byte) 0xDD, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddusw   mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3315, Serial#: 5202
    public void paddusw(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xDD, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddusw   mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3316, Serial#: 5197
    public void m_paddusw(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xDD, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddusw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code paddusw   mm0, [ebx + 305419896]}
     */
    // Template#: 3317, Serial#: 5201
    public void paddusw(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xDD, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddw     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3318, Serial#: 5515
    public void paddw(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xFD, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code paddw     xmm0, [ebx + 18]}
     */
    // Template#: 3319, Serial#: 5514
    public void paddw(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xFD, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddw     xmm0, ebx[esi * 4]}
     */
    // Template#: 3320, Serial#: 5511
    public void paddw(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xFD, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddw     xmm0, [ebx]}
     */
    // Template#: 3321, Serial#: 5510
    public void paddw(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xFD, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddw     xmm0, xmm0}
     */
    // Template#: 3322, Serial#: 5518
    public void paddw(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xFD, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code paddw     xmm0, [L1: 0x12345678]}
     */
    // Template#: 3323, Serial#: 5513
    public void m_paddw(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xFD, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddw     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3324, Serial#: 5517
    public void paddw(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xFD, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddw     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3325, Serial#: 5512
    public void m_paddw(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xFD, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code paddw     xmm0, [ebx + 305419896]}
     */
    // Template#: 3326, Serial#: 5516
    public void paddw(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xFD, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddw     mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3327, Serial#: 5452
    public void paddw(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xFD, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code paddw     mm0, [ebx + 18]}
     */
    // Template#: 3328, Serial#: 5451
    public void paddw(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xFD, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddw     mm0, ebx[esi * 4]}
     */
    // Template#: 3329, Serial#: 5448
    public void paddw(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xFD, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddw     mm0, [ebx]}
     */
    // Template#: 3330, Serial#: 5447
    public void paddw(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xFD, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code paddw     mm0, mm0}
     */
    // Template#: 3331, Serial#: 5455
    public void paddw(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xFD, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code paddw     mm0, [L1: 0x12345678]}
     */
    // Template#: 3332, Serial#: 5450
    public void m_paddw(MMXRegister destination, int m32) {
        assemble0268((byte) 0xFD, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddw     mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3333, Serial#: 5454
    public void paddw(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xFD, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code paddw     mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3334, Serial#: 5449
    public void m_paddw(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xFD, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code paddw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code paddw     mm0, [ebx + 305419896]}
     */
    // Template#: 3335, Serial#: 5453
    public void paddw(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xFD, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pand  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pand      xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3336, Serial#: 5245
    public void pand(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xDB, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pand  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pand      xmm0, [ebx + 18]}
     */
    // Template#: 3337, Serial#: 5244
    public void pand(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xDB, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pand  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pand      xmm0, ebx[esi * 4]}
     */
    // Template#: 3338, Serial#: 5241
    public void pand(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xDB, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pand  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pand      xmm0, [ebx]}
     */
    // Template#: 3339, Serial#: 5240
    public void pand(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xDB, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pand  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pand      xmm0, xmm0}
     */
    // Template#: 3340, Serial#: 5248
    public void pand(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xDB, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pand  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pand      xmm0, [L1: 0x12345678]}
     */
    // Template#: 3341, Serial#: 5243
    public void m_pand(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xDB, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pand  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pand      xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3342, Serial#: 5247
    public void pand(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xDB, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pand  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pand      xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3343, Serial#: 5242
    public void m_pand(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xDB, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pand  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pand      xmm0, [ebx + 305419896]}
     */
    // Template#: 3344, Serial#: 5246
    public void pand(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xDB, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pand  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pand      mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3345, Serial#: 5182
    public void pand(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xDB, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pand  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pand      mm0, [ebx + 18]}
     */
    // Template#: 3346, Serial#: 5181
    public void pand(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xDB, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pand  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pand      mm0, ebx[esi * 4]}
     */
    // Template#: 3347, Serial#: 5178
    public void pand(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xDB, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pand  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pand      mm0, [ebx]}
     */
    // Template#: 3348, Serial#: 5177
    public void pand(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xDB, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pand  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pand      mm0, mm0}
     */
    // Template#: 3349, Serial#: 5185
    public void pand(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xDB, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pand  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pand      mm0, [L1: 0x12345678]}
     */
    // Template#: 3350, Serial#: 5180
    public void m_pand(MMXRegister destination, int m32) {
        assemble0268((byte) 0xDB, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pand  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pand      mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3351, Serial#: 5184
    public void pand(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xDB, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pand  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pand      mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3352, Serial#: 5179
    public void m_pand(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xDB, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pand  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pand      mm0, [ebx + 305419896]}
     */
    // Template#: 3353, Serial#: 5183
    public void pand(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xDB, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pandn  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pandn     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3354, Serial#: 5281
    public void pandn(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xDF, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pandn  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pandn     xmm0, [ebx + 18]}
     */
    // Template#: 3355, Serial#: 5280
    public void pandn(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xDF, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pandn  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pandn     xmm0, ebx[esi * 4]}
     */
    // Template#: 3356, Serial#: 5277
    public void pandn(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xDF, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pandn  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pandn     xmm0, [ebx]}
     */
    // Template#: 3357, Serial#: 5276
    public void pandn(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xDF, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pandn  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pandn     xmm0, xmm0}
     */
    // Template#: 3358, Serial#: 5284
    public void pandn(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xDF, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pandn  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pandn     xmm0, [L1: 0x12345678]}
     */
    // Template#: 3359, Serial#: 5279
    public void m_pandn(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xDF, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pandn  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pandn     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3360, Serial#: 5283
    public void pandn(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xDF, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pandn  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pandn     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3361, Serial#: 5278
    public void m_pandn(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xDF, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pandn  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pandn     xmm0, [ebx + 305419896]}
     */
    // Template#: 3362, Serial#: 5282
    public void pandn(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xDF, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pandn  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pandn     mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3363, Serial#: 5209
    public void pandn(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xDF, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pandn  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pandn     mm0, [ebx + 18]}
     */
    // Template#: 3364, Serial#: 5208
    public void pandn(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xDF, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pandn  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pandn     mm0, ebx[esi * 4]}
     */
    // Template#: 3365, Serial#: 5205
    public void pandn(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xDF, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pandn  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pandn     mm0, [ebx]}
     */
    // Template#: 3366, Serial#: 5204
    public void pandn(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xDF, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pandn  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pandn     mm0, mm0}
     */
    // Template#: 3367, Serial#: 5212
    public void pandn(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xDF, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pandn  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pandn     mm0, [L1: 0x12345678]}
     */
    // Template#: 3368, Serial#: 5207
    public void m_pandn(MMXRegister destination, int m32) {
        assemble0268((byte) 0xDF, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pandn  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pandn     mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3369, Serial#: 5211
    public void pandn(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xDF, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pandn  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pandn     mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3370, Serial#: 5206
    public void m_pandn(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xDF, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pandn  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pandn     mm0, [ebx + 305419896]}
     */
    // Template#: 3371, Serial#: 5210
    public void pandn(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xDF, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pavgb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pavgb     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3372, Serial#: 3846
    public void pavgb(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xE0, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pavgb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pavgb     xmm0, [ebx + 18]}
     */
    // Template#: 3373, Serial#: 3845
    public void pavgb(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xE0, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pavgb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pavgb     xmm0, ebx[esi * 4]}
     */
    // Template#: 3374, Serial#: 3842
    public void pavgb(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xE0, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pavgb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pavgb     xmm0, [ebx]}
     */
    // Template#: 3375, Serial#: 3841
    public void pavgb(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xE0, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pavgb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pavgb     xmm0, xmm0}
     */
    // Template#: 3376, Serial#: 3849
    public void pavgb(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xE0, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pavgb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pavgb     xmm0, [L1: 0x12345678]}
     */
    // Template#: 3377, Serial#: 3844
    public void m_pavgb(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xE0, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pavgb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pavgb     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3378, Serial#: 3848
    public void pavgb(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xE0, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pavgb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pavgb     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3379, Serial#: 3843
    public void m_pavgb(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xE0, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pavgb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pavgb     xmm0, [ebx + 305419896]}
     */
    // Template#: 3380, Serial#: 3847
    public void pavgb(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xE0, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pavgw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pavgw     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3381, Serial#: 3873
    public void pavgw(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xE3, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pavgw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pavgw     xmm0, [ebx + 18]}
     */
    // Template#: 3382, Serial#: 3872
    public void pavgw(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xE3, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pavgw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pavgw     xmm0, ebx[esi * 4]}
     */
    // Template#: 3383, Serial#: 3869
    public void pavgw(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xE3, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pavgw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pavgw     xmm0, [ebx]}
     */
    // Template#: 3384, Serial#: 3868
    public void pavgw(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xE3, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pavgw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pavgw     xmm0, xmm0}
     */
    // Template#: 3385, Serial#: 3876
    public void pavgw(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xE3, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pavgw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pavgw     xmm0, [L1: 0x12345678]}
     */
    // Template#: 3386, Serial#: 3871
    public void m_pavgw(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xE3, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pavgw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pavgw     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3387, Serial#: 3875
    public void pavgw(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xE3, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pavgw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pavgw     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3388, Serial#: 3870
    public void m_pavgw(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xE3, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pavgw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pavgw     xmm0, [ebx + 305419896]}
     */
    // Template#: 3389, Serial#: 3874
    public void pavgw(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xE3, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpeqb   mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3390, Serial#: 3221
    public void pcmpeqb(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0x74, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpeqb   mm0, [ebx + 18]}
     */
    // Template#: 3391, Serial#: 3220
    public void pcmpeqb(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0x74, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpeqb   mm0, ebx[esi * 4]}
     */
    // Template#: 3392, Serial#: 3217
    public void pcmpeqb(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0x74, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpeqb   mm0, [ebx]}
     */
    // Template#: 3393, Serial#: 3216
    public void pcmpeqb(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0x74, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpeqb   mm0, mm0}
     */
    // Template#: 3394, Serial#: 3224
    public void pcmpeqb(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0x74, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pcmpeqb   mm0, [L1: 0x12345678]}
     */
    // Template#: 3395, Serial#: 3219
    public void m_pcmpeqb(MMXRegister destination, int m32) {
        assemble0268((byte) 0x74, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpeqb   mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3396, Serial#: 3223
    public void pcmpeqb(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0x74, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpeqb   mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3397, Serial#: 3218
    public void m_pcmpeqb(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0x74, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpeqb   mm0, [ebx + 305419896]}
     */
    // Template#: 3398, Serial#: 3222
    public void pcmpeqb(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0x74, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpeqd   xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3399, Serial#: 3325
    public void pcmpeqd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x76, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpeqd   xmm0, [ebx + 18]}
     */
    // Template#: 3400, Serial#: 3324
    public void pcmpeqd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x76, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpeqd   xmm0, ebx[esi * 4]}
     */
    // Template#: 3401, Serial#: 3321
    public void pcmpeqd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x76, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpeqd   xmm0, [ebx]}
     */
    // Template#: 3402, Serial#: 3320
    public void pcmpeqd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x76, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpeqd   xmm0, xmm0}
     */
    // Template#: 3403, Serial#: 3328
    public void pcmpeqd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x76, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pcmpeqd   xmm0, [L1: 0x12345678]}
     */
    // Template#: 3404, Serial#: 3323
    public void m_pcmpeqd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x76, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpeqd   xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3405, Serial#: 3327
    public void pcmpeqd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x76, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpeqd   xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3406, Serial#: 3322
    public void m_pcmpeqd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x76, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpeqd   xmm0, [ebx + 305419896]}
     */
    // Template#: 3407, Serial#: 3326
    public void pcmpeqd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x76, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpeqd   mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3408, Serial#: 3239
    public void pcmpeqd(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0x76, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpeqd   mm0, [ebx + 18]}
     */
    // Template#: 3409, Serial#: 3238
    public void pcmpeqd(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0x76, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpeqd   mm0, ebx[esi * 4]}
     */
    // Template#: 3410, Serial#: 3235
    public void pcmpeqd(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0x76, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpeqd   mm0, [ebx]}
     */
    // Template#: 3411, Serial#: 3234
    public void pcmpeqd(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0x76, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpeqd   mm0, mm0}
     */
    // Template#: 3412, Serial#: 3242
    public void pcmpeqd(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0x76, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pcmpeqd   mm0, [L1: 0x12345678]}
     */
    // Template#: 3413, Serial#: 3237
    public void m_pcmpeqd(MMXRegister destination, int m32) {
        assemble0268((byte) 0x76, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpeqd   mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3414, Serial#: 3241
    public void pcmpeqd(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0x76, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpeqd   mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3415, Serial#: 3236
    public void m_pcmpeqd(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0x76, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpeqd   mm0, [ebx + 305419896]}
     */
    // Template#: 3416, Serial#: 3240
    public void pcmpeqd(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0x76, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpeqw   mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3417, Serial#: 3230
    public void pcmpeqw(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0x75, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpeqw   mm0, [ebx + 18]}
     */
    // Template#: 3418, Serial#: 3229
    public void pcmpeqw(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0x75, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpeqw   mm0, ebx[esi * 4]}
     */
    // Template#: 3419, Serial#: 3226
    public void pcmpeqw(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0x75, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpeqw   mm0, [ebx]}
     */
    // Template#: 3420, Serial#: 3225
    public void pcmpeqw(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0x75, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpeqw   mm0, mm0}
     */
    // Template#: 3421, Serial#: 3233
    public void pcmpeqw(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0x75, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pcmpeqw   mm0, [L1: 0x12345678]}
     */
    // Template#: 3422, Serial#: 3228
    public void m_pcmpeqw(MMXRegister destination, int m32) {
        assemble0268((byte) 0x75, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpeqw   mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3423, Serial#: 3232
    public void pcmpeqw(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0x75, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpeqw   mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3424, Serial#: 3227
    public void m_pcmpeqw(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0x75, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpeqw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpeqw   mm0, [ebx + 305419896]}
     */
    // Template#: 3425, Serial#: 3231
    public void pcmpeqw(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0x75, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpgtb   xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3426, Serial#: 3285
    public void pcmpgtb(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x64, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpgtb   xmm0, [ebx + 18]}
     */
    // Template#: 3427, Serial#: 3284
    public void pcmpgtb(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x64, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpgtb   xmm0, ebx[esi * 4]}
     */
    // Template#: 3428, Serial#: 3281
    public void pcmpgtb(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x64, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpgtb   xmm0, [ebx]}
     */
    // Template#: 3429, Serial#: 3280
    public void pcmpgtb(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x64, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpgtb   xmm0, xmm0}
     */
    // Template#: 3430, Serial#: 3288
    public void pcmpgtb(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x64, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pcmpgtb   xmm0, [L1: 0x12345678]}
     */
    // Template#: 3431, Serial#: 3283
    public void m_pcmpgtb(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x64, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpgtb   xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3432, Serial#: 3287
    public void pcmpgtb(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x64, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpgtb   xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3433, Serial#: 3282
    public void m_pcmpgtb(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x64, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpgtb   xmm0, [ebx + 305419896]}
     */
    // Template#: 3434, Serial#: 3286
    public void pcmpgtb(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x64, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpgtb   mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3435, Serial#: 3168
    public void pcmpgtb(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0x64, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpgtb   mm0, [ebx + 18]}
     */
    // Template#: 3436, Serial#: 3167
    public void pcmpgtb(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0x64, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpgtb   mm0, ebx[esi * 4]}
     */
    // Template#: 3437, Serial#: 3164
    public void pcmpgtb(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0x64, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpgtb   mm0, [ebx]}
     */
    // Template#: 3438, Serial#: 3163
    public void pcmpgtb(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0x64, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpgtb   mm0, mm0}
     */
    // Template#: 3439, Serial#: 3171
    public void pcmpgtb(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0x64, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pcmpgtb   mm0, [L1: 0x12345678]}
     */
    // Template#: 3440, Serial#: 3166
    public void m_pcmpgtb(MMXRegister destination, int m32) {
        assemble0268((byte) 0x64, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpgtb   mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3441, Serial#: 3170
    public void pcmpgtb(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0x64, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpgtb   mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3442, Serial#: 3165
    public void m_pcmpgtb(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0x64, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpgtb   mm0, [ebx + 305419896]}
     */
    // Template#: 3443, Serial#: 3169
    public void pcmpgtb(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0x64, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpgtd   xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3444, Serial#: 3303
    public void pcmpgtd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x66, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpgtd   xmm0, [ebx + 18]}
     */
    // Template#: 3445, Serial#: 3302
    public void pcmpgtd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x66, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpgtd   xmm0, ebx[esi * 4]}
     */
    // Template#: 3446, Serial#: 3299
    public void pcmpgtd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x66, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpgtd   xmm0, [ebx]}
     */
    // Template#: 3447, Serial#: 3298
    public void pcmpgtd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x66, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpgtd   xmm0, xmm0}
     */
    // Template#: 3448, Serial#: 3306
    public void pcmpgtd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x66, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pcmpgtd   xmm0, [L1: 0x12345678]}
     */
    // Template#: 3449, Serial#: 3301
    public void m_pcmpgtd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x66, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpgtd   xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3450, Serial#: 3305
    public void pcmpgtd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x66, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpgtd   xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3451, Serial#: 3300
    public void m_pcmpgtd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x66, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpgtd   xmm0, [ebx + 305419896]}
     */
    // Template#: 3452, Serial#: 3304
    public void pcmpgtd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x66, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpgtd   mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3453, Serial#: 3186
    public void pcmpgtd(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0x66, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpgtd   mm0, [ebx + 18]}
     */
    // Template#: 3454, Serial#: 3185
    public void pcmpgtd(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0x66, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpgtd   mm0, ebx[esi * 4]}
     */
    // Template#: 3455, Serial#: 3182
    public void pcmpgtd(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0x66, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpgtd   mm0, [ebx]}
     */
    // Template#: 3456, Serial#: 3181
    public void pcmpgtd(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0x66, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpgtd   mm0, mm0}
     */
    // Template#: 3457, Serial#: 3189
    public void pcmpgtd(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0x66, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pcmpgtd   mm0, [L1: 0x12345678]}
     */
    // Template#: 3458, Serial#: 3184
    public void m_pcmpgtd(MMXRegister destination, int m32) {
        assemble0268((byte) 0x66, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpgtd   mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3459, Serial#: 3188
    public void pcmpgtd(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0x66, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpgtd   mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3460, Serial#: 3183
    public void m_pcmpgtd(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0x66, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpgtd   mm0, [ebx + 305419896]}
     */
    // Template#: 3461, Serial#: 3187
    public void pcmpgtd(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0x66, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpgtw   xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3462, Serial#: 3294
    public void pcmpgtw(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x65, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpgtw   xmm0, [ebx + 18]}
     */
    // Template#: 3463, Serial#: 3293
    public void pcmpgtw(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x65, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpgtw   xmm0, ebx[esi * 4]}
     */
    // Template#: 3464, Serial#: 3290
    public void pcmpgtw(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x65, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpgtw   xmm0, [ebx]}
     */
    // Template#: 3465, Serial#: 3289
    public void pcmpgtw(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x65, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpgtw   xmm0, xmm0}
     */
    // Template#: 3466, Serial#: 3297
    public void pcmpgtw(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x65, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pcmpgtw   xmm0, [L1: 0x12345678]}
     */
    // Template#: 3467, Serial#: 3292
    public void m_pcmpgtw(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x65, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpgtw   xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3468, Serial#: 3296
    public void pcmpgtw(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x65, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpgtw   xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3469, Serial#: 3291
    public void m_pcmpgtw(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x65, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpgtw   xmm0, [ebx + 305419896]}
     */
    // Template#: 3470, Serial#: 3295
    public void pcmpgtw(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x65, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpgtw   mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3471, Serial#: 3177
    public void pcmpgtw(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0x65, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpgtw   mm0, [ebx + 18]}
     */
    // Template#: 3472, Serial#: 3176
    public void pcmpgtw(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0x65, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpgtw   mm0, ebx[esi * 4]}
     */
    // Template#: 3473, Serial#: 3173
    public void pcmpgtw(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0x65, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpgtw   mm0, [ebx]}
     */
    // Template#: 3474, Serial#: 3172
    public void pcmpgtw(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0x65, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpgtw   mm0, mm0}
     */
    // Template#: 3475, Serial#: 3180
    public void pcmpgtw(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0x65, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pcmpgtw   mm0, [L1: 0x12345678]}
     */
    // Template#: 3476, Serial#: 3175
    public void m_pcmpgtw(MMXRegister destination, int m32) {
        assemble0268((byte) 0x65, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpgtw   mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3477, Serial#: 3179
    public void pcmpgtw(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0x65, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pcmpgtw   mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3478, Serial#: 3174
    public void m_pcmpgtw(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0x65, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pcmpgtw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pcmpgtw   mm0, [ebx + 305419896]}
     */
    // Template#: 3479, Serial#: 3178
    public void pcmpgtw(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0x65, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pextrw  }<i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pextrw    eax, xmm0, 0x12}
     */
    // Template#: 3480, Serial#: 3684
    public void pextrw(IA32GeneralRegister32 destination, IA32XMMRegister source, byte imm8) {
        assemble0465((byte) 0xC5, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pextrw  }<i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pextrw    eax, mm0, 0x12}
     */
    // Template#: 3481, Serial#: 3648
    public void pextrw(IA32GeneralRegister32 destination, MMXRegister source, byte imm8) {
        assemble0466((byte) 0xC5, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pinsrw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pinsrw    xmm0, ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 3482, Serial#: 3680
    public void pinsrw(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0467((byte) 0xC4, destination, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pinsrw  }<i>destination</i>, <i>disp8</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pinsrw    xmm0, [ebx + 18], 0x12}
     */
    // Template#: 3483, Serial#: 3679
    public void pinsrw(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source, byte imm8) {
        assemble0468((byte) 0xC4, destination, disp8, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pinsrw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pinsrw    xmm0, ebx[esi * 4], 0x12}
     */
    // Template#: 3484, Serial#: 3676
    public void pinsrw(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0469((byte) 0xC4, destination, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pinsrw  }<i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pinsrw    xmm0, eax, 0x12}
     */
    // Template#: 3485, Serial#: 3683
    public void pinsrw(IA32XMMRegister destination, IA32GeneralRegister32 source, byte imm8) {
        assemble0470((byte) 0xC4, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pinsrw  }<i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pinsrw    xmm0, [ebx], 0x12}
     */
    // Template#: 3486, Serial#: 3675
    public void pinsrw(IA32XMMRegister destination, IA32IndirectRegister32 source, byte imm8) {
        assemble0471((byte) 0xC4, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pinsrw  }<i>destination</i>, <i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pinsrw    xmm0, [L1: 0x12345678], 0x12}
     */
    // Template#: 3487, Serial#: 3678
    public void m_pinsrw(IA32XMMRegister destination, int m32, byte imm8) {
        assemble0472((byte) 0xC4, destination, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pinsrw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pinsrw    xmm0, ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 3488, Serial#: 3682
    public void pinsrw(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0473((byte) 0xC4, destination, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pinsrw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pinsrw    xmm0, L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 3489, Serial#: 3677
    public void m_pinsrw(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0474((byte) 0xC4, destination, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pinsrw  }<i>destination</i>, <i>disp32</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pinsrw    xmm0, [ebx + 305419896], 0x12}
     */
    // Template#: 3490, Serial#: 3681
    public void pinsrw(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source, byte imm8) {
        assemble0475((byte) 0xC4, destination, disp32, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pinsrw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pinsrw    mm0, ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 3491, Serial#: 3644
    public void pinsrw(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0476((byte) 0xC4, destination, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pinsrw  }<i>destination</i>, <i>disp8</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pinsrw    mm0, [ebx + 18], 0x12}
     */
    // Template#: 3492, Serial#: 3643
    public void pinsrw(MMXRegister destination, byte disp8, IA32IndirectRegister32 source, byte imm8) {
        assemble0477((byte) 0xC4, destination, disp8, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pinsrw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pinsrw    mm0, ebx[esi * 4], 0x12}
     */
    // Template#: 3493, Serial#: 3640
    public void pinsrw(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0478((byte) 0xC4, destination, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pinsrw  }<i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pinsrw    mm0, eax, 0x12}
     */
    // Template#: 3494, Serial#: 3647
    public void pinsrw(MMXRegister destination, IA32GeneralRegister32 source, byte imm8) {
        assemble0479((byte) 0xC4, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pinsrw  }<i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pinsrw    mm0, [ebx], 0x12}
     */
    // Template#: 3495, Serial#: 3639
    public void pinsrw(MMXRegister destination, IA32IndirectRegister32 source, byte imm8) {
        assemble0480((byte) 0xC4, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pinsrw  }<i>destination</i>, <i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pinsrw    mm0, [L1: 0x12345678], 0x12}
     */
    // Template#: 3496, Serial#: 3642
    public void m_pinsrw(MMXRegister destination, int m32, byte imm8) {
        assemble0481((byte) 0xC4, destination, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pinsrw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pinsrw    mm0, ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 3497, Serial#: 3646
    public void pinsrw(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0482((byte) 0xC4, destination, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pinsrw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pinsrw    mm0, L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 3498, Serial#: 3641
    public void m_pinsrw(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0483((byte) 0xC4, destination, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pinsrw  }<i>destination</i>, <i>disp32</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pinsrw    mm0, [ebx + 305419896], 0x12}
     */
    // Template#: 3499, Serial#: 3645
    public void pinsrw(MMXRegister destination, int disp32, IA32IndirectRegister32 source, byte imm8) {
        assemble0484((byte) 0xC4, destination, disp32, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaddwd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmaddwd   xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3500, Serial#: 4007
    public void pmaddwd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xF5, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaddwd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pmaddwd   xmm0, [ebx + 18]}
     */
    // Template#: 3501, Serial#: 4006
    public void pmaddwd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xF5, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaddwd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmaddwd   xmm0, ebx[esi * 4]}
     */
    // Template#: 3502, Serial#: 4003
    public void pmaddwd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xF5, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaddwd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pmaddwd   xmm0, [ebx]}
     */
    // Template#: 3503, Serial#: 4002
    public void pmaddwd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xF5, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaddwd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pmaddwd   xmm0, xmm0}
     */
    // Template#: 3504, Serial#: 4010
    public void pmaddwd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xF5, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaddwd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pmaddwd   xmm0, [L1: 0x12345678]}
     */
    // Template#: 3505, Serial#: 4005
    public void m_pmaddwd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xF5, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaddwd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmaddwd   xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3506, Serial#: 4009
    public void pmaddwd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xF5, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaddwd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmaddwd   xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3507, Serial#: 4004
    public void m_pmaddwd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xF5, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaddwd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pmaddwd   xmm0, [ebx + 305419896]}
     */
    // Template#: 3508, Serial#: 4008
    public void pmaddwd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xF5, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaddwd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmaddwd   mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3509, Serial#: 3962
    public void pmaddwd(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xF5, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaddwd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pmaddwd   mm0, [ebx + 18]}
     */
    // Template#: 3510, Serial#: 3961
    public void pmaddwd(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xF5, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaddwd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmaddwd   mm0, ebx[esi * 4]}
     */
    // Template#: 3511, Serial#: 3958
    public void pmaddwd(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xF5, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaddwd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pmaddwd   mm0, [ebx]}
     */
    // Template#: 3512, Serial#: 3957
    public void pmaddwd(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xF5, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaddwd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pmaddwd   mm0, mm0}
     */
    // Template#: 3513, Serial#: 3965
    public void pmaddwd(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xF5, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaddwd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pmaddwd   mm0, [L1: 0x12345678]}
     */
    // Template#: 3514, Serial#: 3960
    public void m_pmaddwd(MMXRegister destination, int m32) {
        assemble0268((byte) 0xF5, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaddwd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmaddwd   mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3515, Serial#: 3964
    public void pmaddwd(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xF5, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaddwd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmaddwd   mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3516, Serial#: 3959
    public void m_pmaddwd(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xF5, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaddwd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pmaddwd   mm0, [ebx + 305419896]}
     */
    // Template#: 3517, Serial#: 3963
    public void pmaddwd(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xF5, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaxsw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmaxsw    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3518, Serial#: 5398
    public void pmaxsw(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xEE, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaxsw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pmaxsw    xmm0, [ebx + 18]}
     */
    // Template#: 3519, Serial#: 5397
    public void pmaxsw(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xEE, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaxsw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmaxsw    xmm0, ebx[esi * 4]}
     */
    // Template#: 3520, Serial#: 5394
    public void pmaxsw(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xEE, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaxsw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pmaxsw    xmm0, [ebx]}
     */
    // Template#: 3521, Serial#: 5393
    public void pmaxsw(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xEE, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaxsw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pmaxsw    xmm0, xmm0}
     */
    // Template#: 3522, Serial#: 5401
    public void pmaxsw(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xEE, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaxsw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pmaxsw    xmm0, [L1: 0x12345678]}
     */
    // Template#: 3523, Serial#: 5396
    public void m_pmaxsw(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xEE, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaxsw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmaxsw    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3524, Serial#: 5400
    public void pmaxsw(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xEE, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaxsw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmaxsw    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3525, Serial#: 5395
    public void m_pmaxsw(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xEE, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaxsw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pmaxsw    xmm0, [ebx + 305419896]}
     */
    // Template#: 3526, Serial#: 5399
    public void pmaxsw(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xEE, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaxub  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmaxub    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3527, Serial#: 5272
    public void pmaxub(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xDE, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaxub  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pmaxub    xmm0, [ebx + 18]}
     */
    // Template#: 3528, Serial#: 5271
    public void pmaxub(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xDE, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaxub  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmaxub    xmm0, ebx[esi * 4]}
     */
    // Template#: 3529, Serial#: 5268
    public void pmaxub(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xDE, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaxub  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pmaxub    xmm0, [ebx]}
     */
    // Template#: 3530, Serial#: 5267
    public void pmaxub(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xDE, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaxub  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pmaxub    xmm0, xmm0}
     */
    // Template#: 3531, Serial#: 5275
    public void pmaxub(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xDE, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaxub  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pmaxub    xmm0, [L1: 0x12345678]}
     */
    // Template#: 3532, Serial#: 5270
    public void m_pmaxub(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xDE, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaxub  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmaxub    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3533, Serial#: 5274
    public void pmaxub(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xDE, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaxub  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmaxub    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3534, Serial#: 5269
    public void m_pmaxub(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xDE, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmaxub  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pmaxub    xmm0, [ebx + 305419896]}
     */
    // Template#: 3535, Serial#: 5273
    public void pmaxub(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xDE, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pminsw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pminsw    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3536, Serial#: 5362
    public void pminsw(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xEA, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pminsw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pminsw    xmm0, [ebx + 18]}
     */
    // Template#: 3537, Serial#: 5361
    public void pminsw(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xEA, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pminsw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pminsw    xmm0, ebx[esi * 4]}
     */
    // Template#: 3538, Serial#: 5358
    public void pminsw(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xEA, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pminsw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pminsw    xmm0, [ebx]}
     */
    // Template#: 3539, Serial#: 5357
    public void pminsw(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xEA, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pminsw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pminsw    xmm0, xmm0}
     */
    // Template#: 3540, Serial#: 5365
    public void pminsw(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xEA, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pminsw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pminsw    xmm0, [L1: 0x12345678]}
     */
    // Template#: 3541, Serial#: 5360
    public void m_pminsw(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xEA, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pminsw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pminsw    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3542, Serial#: 5364
    public void pminsw(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xEA, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pminsw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pminsw    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3543, Serial#: 5359
    public void m_pminsw(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xEA, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pminsw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pminsw    xmm0, [ebx + 305419896]}
     */
    // Template#: 3544, Serial#: 5363
    public void pminsw(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xEA, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pminub  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pminub    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3545, Serial#: 5236
    public void pminub(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xDA, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pminub  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pminub    xmm0, [ebx + 18]}
     */
    // Template#: 3546, Serial#: 5235
    public void pminub(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xDA, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pminub  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pminub    xmm0, ebx[esi * 4]}
     */
    // Template#: 3547, Serial#: 5232
    public void pminub(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xDA, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pminub  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pminub    xmm0, [ebx]}
     */
    // Template#: 3548, Serial#: 5231
    public void pminub(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xDA, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pminub  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pminub    xmm0, xmm0}
     */
    // Template#: 3549, Serial#: 5239
    public void pminub(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xDA, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pminub  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pminub    xmm0, [L1: 0x12345678]}
     */
    // Template#: 3550, Serial#: 5234
    public void m_pminub(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xDA, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pminub  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pminub    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3551, Serial#: 5238
    public void pminub(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xDA, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pminub  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pminub    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3552, Serial#: 5233
    public void m_pminub(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xDA, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pminub  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pminub    xmm0, [ebx + 305419896]}
     */
    // Template#: 3553, Serial#: 5237
    public void pminub(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xDA, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmovmskb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pmovmskb  eax, xmm0}
     */
    // Template#: 3554, Serial#: 3811
    public void pmovmskb(IA32GeneralRegister32 destination, IA32XMMRegister source) {
        assemble0449((byte) 0xD7, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhuw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmulhuw   xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3555, Serial#: 3882
    public void pmulhuw(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xE4, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhuw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pmulhuw   xmm0, [ebx + 18]}
     */
    // Template#: 3556, Serial#: 3881
    public void pmulhuw(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xE4, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhuw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmulhuw   xmm0, ebx[esi * 4]}
     */
    // Template#: 3557, Serial#: 3878
    public void pmulhuw(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xE4, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhuw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pmulhuw   xmm0, [ebx]}
     */
    // Template#: 3558, Serial#: 3877
    public void pmulhuw(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xE4, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhuw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pmulhuw   xmm0, xmm0}
     */
    // Template#: 3559, Serial#: 3885
    public void pmulhuw(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xE4, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhuw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pmulhuw   xmm0, [L1: 0x12345678]}
     */
    // Template#: 3560, Serial#: 3880
    public void m_pmulhuw(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xE4, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhuw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmulhuw   xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3561, Serial#: 3884
    public void pmulhuw(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xE4, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhuw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmulhuw   xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3562, Serial#: 3879
    public void m_pmulhuw(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xE4, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhuw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pmulhuw   xmm0, [ebx + 305419896]}
     */
    // Template#: 3563, Serial#: 3883
    public void pmulhuw(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xE4, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmulhw    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3564, Serial#: 3891
    public void pmulhw(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xE5, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pmulhw    xmm0, [ebx + 18]}
     */
    // Template#: 3565, Serial#: 3890
    public void pmulhw(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xE5, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmulhw    xmm0, ebx[esi * 4]}
     */
    // Template#: 3566, Serial#: 3887
    public void pmulhw(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xE5, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pmulhw    xmm0, [ebx]}
     */
    // Template#: 3567, Serial#: 3886
    public void pmulhw(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xE5, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pmulhw    xmm0, xmm0}
     */
    // Template#: 3568, Serial#: 3894
    public void pmulhw(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xE5, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pmulhw    xmm0, [L1: 0x12345678]}
     */
    // Template#: 3569, Serial#: 3889
    public void m_pmulhw(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xE5, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmulhw    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3570, Serial#: 3893
    public void pmulhw(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xE5, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmulhw    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3571, Serial#: 3888
    public void m_pmulhw(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xE5, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pmulhw    xmm0, [ebx + 305419896]}
     */
    // Template#: 3572, Serial#: 3892
    public void pmulhw(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xE5, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmulhw    mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3573, Serial#: 3837
    public void pmulhw(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xE5, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pmulhw    mm0, [ebx + 18]}
     */
    // Template#: 3574, Serial#: 3836
    public void pmulhw(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xE5, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmulhw    mm0, ebx[esi * 4]}
     */
    // Template#: 3575, Serial#: 3833
    public void pmulhw(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xE5, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pmulhw    mm0, [ebx]}
     */
    // Template#: 3576, Serial#: 3832
    public void pmulhw(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xE5, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pmulhw    mm0, mm0}
     */
    // Template#: 3577, Serial#: 3840
    public void pmulhw(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xE5, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pmulhw    mm0, [L1: 0x12345678]}
     */
    // Template#: 3578, Serial#: 3835
    public void m_pmulhw(MMXRegister destination, int m32) {
        assemble0268((byte) 0xE5, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmulhw    mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3579, Serial#: 3839
    public void pmulhw(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xE5, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmulhw    mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3580, Serial#: 3834
    public void m_pmulhw(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xE5, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmulhw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pmulhw    mm0, [ebx + 305419896]}
     */
    // Template#: 3581, Serial#: 3838
    public void pmulhw(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xE5, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmullw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmullw    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3582, Serial#: 3798
    public void pmullw(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xD5, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmullw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pmullw    xmm0, [ebx + 18]}
     */
    // Template#: 3583, Serial#: 3797
    public void pmullw(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xD5, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmullw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmullw    xmm0, ebx[esi * 4]}
     */
    // Template#: 3584, Serial#: 3794
    public void pmullw(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xD5, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmullw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pmullw    xmm0, [ebx]}
     */
    // Template#: 3585, Serial#: 3793
    public void pmullw(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xD5, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmullw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pmullw    xmm0, xmm0}
     */
    // Template#: 3586, Serial#: 3801
    public void pmullw(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xD5, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmullw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pmullw    xmm0, [L1: 0x12345678]}
     */
    // Template#: 3587, Serial#: 3796
    public void m_pmullw(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xD5, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmullw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmullw    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3588, Serial#: 3800
    public void pmullw(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xD5, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmullw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmullw    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3589, Serial#: 3795
    public void m_pmullw(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xD5, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmullw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pmullw    xmm0, [ebx + 305419896]}
     */
    // Template#: 3590, Serial#: 3799
    public void pmullw(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xD5, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmullw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmullw    mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3591, Serial#: 3744
    public void pmullw(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xD5, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmullw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pmullw    mm0, [ebx + 18]}
     */
    // Template#: 3592, Serial#: 3743
    public void pmullw(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xD5, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmullw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmullw    mm0, ebx[esi * 4]}
     */
    // Template#: 3593, Serial#: 3740
    public void pmullw(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xD5, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmullw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pmullw    mm0, [ebx]}
     */
    // Template#: 3594, Serial#: 3739
    public void pmullw(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xD5, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmullw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pmullw    mm0, mm0}
     */
    // Template#: 3595, Serial#: 3747
    public void pmullw(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xD5, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmullw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pmullw    mm0, [L1: 0x12345678]}
     */
    // Template#: 3596, Serial#: 3742
    public void m_pmullw(MMXRegister destination, int m32) {
        assemble0268((byte) 0xD5, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmullw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmullw    mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3597, Serial#: 3746
    public void pmullw(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xD5, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmullw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmullw    mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3598, Serial#: 3741
    public void m_pmullw(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xD5, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmullw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pmullw    mm0, [ebx + 305419896]}
     */
    // Template#: 3599, Serial#: 3745
    public void pmullw(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xD5, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmuludq  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmuludq   xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3600, Serial#: 3998
    public void pmuludq(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xF4, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmuludq  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pmuludq   xmm0, [ebx + 18]}
     */
    // Template#: 3601, Serial#: 3997
    public void pmuludq(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xF4, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmuludq  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmuludq   xmm0, ebx[esi * 4]}
     */
    // Template#: 3602, Serial#: 3994
    public void pmuludq(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xF4, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmuludq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pmuludq   xmm0, [ebx]}
     */
    // Template#: 3603, Serial#: 3993
    public void pmuludq(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xF4, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmuludq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pmuludq   xmm0, xmm0}
     */
    // Template#: 3604, Serial#: 4001
    public void pmuludq(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xF4, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmuludq  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pmuludq   xmm0, [L1: 0x12345678]}
     */
    // Template#: 3605, Serial#: 3996
    public void m_pmuludq(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xF4, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmuludq  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmuludq   xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3606, Serial#: 4000
    public void pmuludq(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xF4, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmuludq  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pmuludq   xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3607, Serial#: 3995
    public void m_pmuludq(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xF4, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pmuludq  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pmuludq   xmm0, [ebx + 305419896]}
     */
    // Template#: 3608, Serial#: 3999
    public void pmuludq(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xF4, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pop  }
     * Example disassembly syntax: {@code pop       es}
     */
    // Template#: 3609, Serial#: 56
    public void pop_ES() {
        assemble0001((byte) 0x07);
    }

    /**
     * Pseudo-external assembler syntax: {@code pop  }
     * Example disassembly syntax: {@code pop       ss}
     */
    // Template#: 3610, Serial#: 112
    public void pop_SS() {
        assemble0001((byte) 0x17);
    }

    /**
     * Pseudo-external assembler syntax: {@code pop  }
     * Example disassembly syntax: {@code pop       ds}
     */
    // Template#: 3611, Serial#: 1646
    public void pop_DS() {
        assemble0001((byte) 0x1F);
    }

    /**
     * Pseudo-external assembler syntax: {@code pop  }
     * Example disassembly syntax: {@code pop       fs}
     */
    // Template#: 3612, Serial#: 3410
    public void pop_FS() {
        assemble0198((byte) 0xA1);
    }

    /**
     * Pseudo-external assembler syntax: {@code pop  }
     * Example disassembly syntax: {@code pop       gs}
     */
    // Template#: 3613, Serial#: 4951
    public void pop_GS() {
        assemble0198((byte) 0xA9);
    }

    /**
     * Pseudo-external assembler syntax: {@code pop  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pop       ebx[esi * 4 + 18]}
     */
    // Template#: 3614, Serial#: 1906
    public void pop(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0485((byte) 0x8F, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pop  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code pop       [ebx + 18]}
     */
    // Template#: 3615, Serial#: 1905
    public void pop(byte disp8, IA32IndirectRegister32 destination) {
        assemble0486((byte) 0x8F, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code pop  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pop       ebx[esi * 4]}
     */
    // Template#: 3616, Serial#: 1902
    public void pop(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0487((byte) 0x8F, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pop  }<i>register</i>
     * Example disassembly syntax: {@code pop       ax}
     */
    // Template#: 3617, Serial#: 1762
    public void pop(IA32GeneralRegister16 register) {
        assemble0295((byte) 0x58, register);
    }

    /**
     * Pseudo-external assembler syntax: {@code pop  }<i>register</i>
     * Example disassembly syntax: {@code pop       eax}
     */
    // Template#: 3618, Serial#: 1761
    public void pop(IA32GeneralRegister32 register) {
        assemble0296((byte) 0x58, register);
    }

    /**
     * Pseudo-external assembler syntax: {@code pop  }<i>destination</i>
     * Example disassembly syntax: {@code pop       [ebx]}
     */
    // Template#: 3619, Serial#: 1901
    public void pop(IA32IndirectRegister32 destination) {
        assemble0488((byte) 0x8F, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code pop  }<i>m32</i>
     * Example disassembly syntax: {@code pop       [L1: 0x12345678]}
     */
    // Template#: 3620, Serial#: 1904
    public void m_pop(int m32) {
        assemble0489((byte) 0x8F, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pop  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pop       ebx[esi * 4 + 305419896]}
     */
    // Template#: 3621, Serial#: 1908
    public void pop(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0490((byte) 0x8F, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pop  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pop       L1: 0x12345678[esi * 4]}
     */
    // Template#: 3622, Serial#: 1903
    public void m_pop(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0491((byte) 0x8F, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pop  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code pop       [ebx + 305419896]}
     */
    // Template#: 3623, Serial#: 1907
    public void pop(int disp32, IA32IndirectRegister32 destination) {
        assemble0492((byte) 0x8F, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code popa  }
     * Example disassembly syntax: {@code popa    }
     */
    // Template#: 3624, Serial#: 231
    public void popa() {
        assemble0197((byte) 0x61);
    }

    /**
     * Pseudo-external assembler syntax: {@code popa  }
     * Example disassembly syntax: {@code popa    }
     */
    // Template#: 3625, Serial#: 232
    public void popad() {
        assemble0001((byte) 0x61);
    }

    /**
     * Pseudo-external assembler syntax: {@code popfl  }
     * Example disassembly syntax: {@code popfl   }
     */
    // Template#: 3626, Serial#: 1917
    public void popfl() {
        assemble0001((byte) 0x9D);
    }

    /**
     * Pseudo-external assembler syntax: {@code popfw  }
     * Example disassembly syntax: {@code popfw   }
     */
    // Template#: 3627, Serial#: 1918
    public void popfw() {
        assemble0197((byte) 0x9D);
    }

    /**
     * Pseudo-external assembler syntax: {@code por  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code por       xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3628, Serial#: 5371
    public void por(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xEB, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code por  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code por       xmm0, [ebx + 18]}
     */
    // Template#: 3629, Serial#: 5370
    public void por(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xEB, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code por  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code por       xmm0, ebx[esi * 4]}
     */
    // Template#: 3630, Serial#: 5367
    public void por(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xEB, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code por  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code por       xmm0, [ebx]}
     */
    // Template#: 3631, Serial#: 5366
    public void por(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xEB, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code por  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code por       xmm0, xmm0}
     */
    // Template#: 3632, Serial#: 5374
    public void por(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xEB, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code por  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code por       xmm0, [L1: 0x12345678]}
     */
    // Template#: 3633, Serial#: 5369
    public void m_por(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xEB, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code por  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code por       xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3634, Serial#: 5373
    public void por(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xEB, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code por  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code por       xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3635, Serial#: 5368
    public void m_por(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xEB, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code por  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code por       xmm0, [ebx + 305419896]}
     */
    // Template#: 3636, Serial#: 5372
    public void por(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xEB, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code por  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code por       mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3637, Serial#: 5308
    public void por(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xEB, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code por  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code por       mm0, [ebx + 18]}
     */
    // Template#: 3638, Serial#: 5307
    public void por(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xEB, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code por  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code por       mm0, ebx[esi * 4]}
     */
    // Template#: 3639, Serial#: 5304
    public void por(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xEB, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code por  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code por       mm0, [ebx]}
     */
    // Template#: 3640, Serial#: 5303
    public void por(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xEB, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code por  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code por       mm0, mm0}
     */
    // Template#: 3641, Serial#: 5311
    public void por(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xEB, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code por  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code por       mm0, [L1: 0x12345678]}
     */
    // Template#: 3642, Serial#: 5306
    public void m_por(MMXRegister destination, int m32) {
        assemble0268((byte) 0xEB, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code por  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code por       mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3643, Serial#: 5310
    public void por(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xEB, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code por  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code por       mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3644, Serial#: 5305
    public void m_por(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xEB, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code por  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code por       mm0, [ebx + 305419896]}
     */
    // Template#: 3645, Serial#: 5309
    public void por(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xEB, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psadbw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psadbw    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3646, Serial#: 4016
    public void psadbw(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xF6, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psadbw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psadbw    xmm0, [ebx + 18]}
     */
    // Template#: 3647, Serial#: 4015
    public void psadbw(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xF6, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psadbw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psadbw    xmm0, ebx[esi * 4]}
     */
    // Template#: 3648, Serial#: 4012
    public void psadbw(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xF6, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psadbw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psadbw    xmm0, [ebx]}
     */
    // Template#: 3649, Serial#: 4011
    public void psadbw(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xF6, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psadbw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psadbw    xmm0, xmm0}
     */
    // Template#: 3650, Serial#: 4019
    public void psadbw(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xF6, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psadbw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psadbw    xmm0, [L1: 0x12345678]}
     */
    // Template#: 3651, Serial#: 4014
    public void m_psadbw(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xF6, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psadbw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psadbw    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3652, Serial#: 4018
    public void psadbw(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xF6, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psadbw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psadbw    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3653, Serial#: 4013
    public void m_psadbw(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xF6, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psadbw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psadbw    xmm0, [ebx + 305419896]}
     */
    // Template#: 3654, Serial#: 4017
    public void psadbw(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xF6, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pshufw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pshufw    mm0, ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 3655, Serial#: 3204
    public void pshufw(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0476((byte) 0x70, destination, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pshufw  }<i>destination</i>, <i>disp8</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pshufw    mm0, [ebx + 18], 0x12}
     */
    // Template#: 3656, Serial#: 3203
    public void pshufw(MMXRegister destination, byte disp8, IA32IndirectRegister32 source, byte imm8) {
        assemble0477((byte) 0x70, destination, disp8, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pshufw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pshufw    mm0, ebx[esi * 4], 0x12}
     */
    // Template#: 3657, Serial#: 3200
    public void pshufw(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0478((byte) 0x70, destination, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pshufw  }<i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pshufw    mm0, [ebx], 0x12}
     */
    // Template#: 3658, Serial#: 3199
    public void pshufw(MMXRegister destination, IA32IndirectRegister32 source, byte imm8) {
        assemble0480((byte) 0x70, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pshufw  }<i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pshufw    mm0, mm0, 0x12}
     */
    // Template#: 3659, Serial#: 3207
    public void pshufw(MMXRegister destination, MMXRegister source, byte imm8) {
        assemble0493((byte) 0x70, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pshufw  }<i>destination</i>, <i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pshufw    mm0, [L1: 0x12345678], 0x12}
     */
    // Template#: 3660, Serial#: 3202
    public void m_pshufw(MMXRegister destination, int m32, byte imm8) {
        assemble0481((byte) 0x70, destination, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pshufw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pshufw    mm0, ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 3661, Serial#: 3206
    public void pshufw(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0482((byte) 0x70, destination, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pshufw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pshufw    mm0, L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 3662, Serial#: 3201
    public void m_pshufw(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0483((byte) 0x70, destination, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pshufw  }<i>destination</i>, <i>disp32</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pshufw    mm0, [ebx + 305419896], 0x12}
     */
    // Template#: 3663, Serial#: 3205
    public void pshufw(MMXRegister destination, int disp32, IA32IndirectRegister32 source, byte imm8) {
        assemble0484((byte) 0x70, destination, disp32, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pslld  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pslld     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3664, Serial#: 3980
    public void pslld(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xF2, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pslld  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pslld     xmm0, [ebx + 18]}
     */
    // Template#: 3665, Serial#: 3979
    public void pslld(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xF2, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pslld  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pslld     xmm0, ebx[esi * 4]}
     */
    // Template#: 3666, Serial#: 3976
    public void pslld(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xF2, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pslld  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pslld     xmm0, [ebx]}
     */
    // Template#: 3667, Serial#: 3975
    public void pslld(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xF2, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pslld  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pslld     xmm0, xmm0}
     */
    // Template#: 3668, Serial#: 3983
    public void pslld(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xF2, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pslld  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pslld     xmm0, [L1: 0x12345678]}
     */
    // Template#: 3669, Serial#: 3978
    public void m_pslld(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xF2, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pslld  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pslld     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3670, Serial#: 3982
    public void pslld(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xF2, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pslld  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pslld     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3671, Serial#: 3977
    public void m_pslld(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xF2, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pslld  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pslld     xmm0, [ebx + 305419896]}
     */
    // Template#: 3672, Serial#: 3981
    public void pslld(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xF2, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pslld  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pslld     mm0, 0x12}
     */
    // Template#: 3673, Serial#: 3213
    public void pslld(MMXRegister destination, byte imm8) {
        assemble0494((byte) 0x72, (byte) 0x06, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code pslld  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pslld     mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3674, Serial#: 3944
    public void pslld(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xF2, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pslld  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pslld     mm0, [ebx + 18]}
     */
    // Template#: 3675, Serial#: 3943
    public void pslld(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xF2, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pslld  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pslld     mm0, ebx[esi * 4]}
     */
    // Template#: 3676, Serial#: 3940
    public void pslld(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xF2, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pslld  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pslld     mm0, [ebx]}
     */
    // Template#: 3677, Serial#: 3939
    public void pslld(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xF2, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pslld  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pslld     mm0, mm0}
     */
    // Template#: 3678, Serial#: 3947
    public void pslld(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xF2, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pslld  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pslld     mm0, [L1: 0x12345678]}
     */
    // Template#: 3679, Serial#: 3942
    public void m_pslld(MMXRegister destination, int m32) {
        assemble0268((byte) 0xF2, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pslld  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pslld     mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3680, Serial#: 3946
    public void pslld(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xF2, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pslld  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pslld     mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3681, Serial#: 3941
    public void m_pslld(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xF2, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pslld  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pslld     mm0, [ebx + 305419896]}
     */
    // Template#: 3682, Serial#: 3945
    public void pslld(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xF2, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pslldq  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code pslldq    xmm0, 0x12}
     */
    // Template#: 3683, Serial#: 3319
    public void pslldq(IA32XMMRegister destination, byte imm8) {
        assemble0495((byte) 0x73, (byte) 0x07, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllq  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code psllq     xmm0, 0x12}
     */
    // Template#: 3684, Serial#: 3318
    public void psllq(IA32XMMRegister destination, byte imm8) {
        assemble0495((byte) 0x73, (byte) 0x06, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllq  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psllq     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3685, Serial#: 3989
    public void psllq(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xF3, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllq  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psllq     xmm0, [ebx + 18]}
     */
    // Template#: 3686, Serial#: 3988
    public void psllq(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xF3, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllq  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psllq     xmm0, ebx[esi * 4]}
     */
    // Template#: 3687, Serial#: 3985
    public void psllq(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xF3, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psllq     xmm0, [ebx]}
     */
    // Template#: 3688, Serial#: 3984
    public void psllq(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xF3, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psllq     xmm0, xmm0}
     */
    // Template#: 3689, Serial#: 3992
    public void psllq(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xF3, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllq  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psllq     xmm0, [L1: 0x12345678]}
     */
    // Template#: 3690, Serial#: 3987
    public void m_psllq(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xF3, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllq  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psllq     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3691, Serial#: 3991
    public void psllq(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xF3, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllq  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psllq     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3692, Serial#: 3986
    public void m_psllq(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xF3, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllq  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psllq     xmm0, [ebx + 305419896]}
     */
    // Template#: 3693, Serial#: 3990
    public void psllq(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xF3, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllq  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code psllq     mm0, 0x12}
     */
    // Template#: 3694, Serial#: 3215
    public void psllq(MMXRegister destination, byte imm8) {
        assemble0494((byte) 0x73, (byte) 0x06, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllq  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psllq     mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3695, Serial#: 3953
    public void psllq(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xF3, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllq  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psllq     mm0, [ebx + 18]}
     */
    // Template#: 3696, Serial#: 3952
    public void psllq(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xF3, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllq  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psllq     mm0, ebx[esi * 4]}
     */
    // Template#: 3697, Serial#: 3949
    public void psllq(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xF3, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psllq     mm0, [ebx]}
     */
    // Template#: 3698, Serial#: 3948
    public void psllq(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xF3, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psllq     mm0, mm0}
     */
    // Template#: 3699, Serial#: 3956
    public void psllq(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xF3, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllq  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psllq     mm0, [L1: 0x12345678]}
     */
    // Template#: 3700, Serial#: 3951
    public void m_psllq(MMXRegister destination, int m32) {
        assemble0268((byte) 0xF3, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllq  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psllq     mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3701, Serial#: 3955
    public void psllq(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xF3, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllq  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psllq     mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3702, Serial#: 3950
    public void m_psllq(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xF3, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllq  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psllq     mm0, [ebx + 305419896]}
     */
    // Template#: 3703, Serial#: 3954
    public void psllq(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xF3, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psllw     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3704, Serial#: 3971
    public void psllw(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xF1, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psllw     xmm0, [ebx + 18]}
     */
    // Template#: 3705, Serial#: 3970
    public void psllw(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xF1, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psllw     xmm0, ebx[esi * 4]}
     */
    // Template#: 3706, Serial#: 3967
    public void psllw(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xF1, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psllw     xmm0, [ebx]}
     */
    // Template#: 3707, Serial#: 3966
    public void psllw(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xF1, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psllw     xmm0, xmm0}
     */
    // Template#: 3708, Serial#: 3974
    public void psllw(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xF1, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psllw     xmm0, [L1: 0x12345678]}
     */
    // Template#: 3709, Serial#: 3969
    public void m_psllw(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xF1, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psllw     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3710, Serial#: 3973
    public void psllw(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xF1, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psllw     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3711, Serial#: 3968
    public void m_psllw(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xF1, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psllw     xmm0, [ebx + 305419896]}
     */
    // Template#: 3712, Serial#: 3972
    public void psllw(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xF1, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code psllw     mm0, 0x12}
     */
    // Template#: 3713, Serial#: 3210
    public void psllw(MMXRegister destination, byte imm8) {
        assemble0494((byte) 0x71, (byte) 0x06, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psllw     mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3714, Serial#: 3935
    public void psllw(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xF1, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psllw     mm0, [ebx + 18]}
     */
    // Template#: 3715, Serial#: 3934
    public void psllw(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xF1, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psllw     mm0, ebx[esi * 4]}
     */
    // Template#: 3716, Serial#: 3931
    public void psllw(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xF1, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psllw     mm0, [ebx]}
     */
    // Template#: 3717, Serial#: 3930
    public void psllw(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xF1, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psllw     mm0, mm0}
     */
    // Template#: 3718, Serial#: 3938
    public void psllw(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xF1, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psllw     mm0, [L1: 0x12345678]}
     */
    // Template#: 3719, Serial#: 3933
    public void m_psllw(MMXRegister destination, int m32) {
        assemble0268((byte) 0xF1, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psllw     mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3720, Serial#: 3937
    public void psllw(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xF1, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psllw     mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3721, Serial#: 3932
    public void m_psllw(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xF1, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psllw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psllw     mm0, [ebx + 305419896]}
     */
    // Template#: 3722, Serial#: 3936
    public void psllw(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xF1, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrad  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrad     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3723, Serial#: 3864
    public void psrad(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xE2, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrad  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psrad     xmm0, [ebx + 18]}
     */
    // Template#: 3724, Serial#: 3863
    public void psrad(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xE2, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrad  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrad     xmm0, ebx[esi * 4]}
     */
    // Template#: 3725, Serial#: 3860
    public void psrad(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xE2, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrad  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psrad     xmm0, [ebx]}
     */
    // Template#: 3726, Serial#: 3859
    public void psrad(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xE2, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrad  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psrad     xmm0, xmm0}
     */
    // Template#: 3727, Serial#: 3867
    public void psrad(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xE2, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrad  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psrad     xmm0, [L1: 0x12345678]}
     */
    // Template#: 3728, Serial#: 3862
    public void m_psrad(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xE2, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrad  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrad     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3729, Serial#: 3866
    public void psrad(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xE2, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrad  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrad     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3730, Serial#: 3861
    public void m_psrad(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xE2, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrad  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psrad     xmm0, [ebx + 305419896]}
     */
    // Template#: 3731, Serial#: 3865
    public void psrad(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xE2, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrad  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code psrad     mm0, 0x12}
     */
    // Template#: 3732, Serial#: 3212
    public void psrad(MMXRegister destination, byte imm8) {
        assemble0494((byte) 0x72, (byte) 0x04, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrad  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrad     mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3733, Serial#: 3828
    public void psrad(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xE2, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrad  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psrad     mm0, [ebx + 18]}
     */
    // Template#: 3734, Serial#: 3827
    public void psrad(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xE2, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrad  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrad     mm0, ebx[esi * 4]}
     */
    // Template#: 3735, Serial#: 3824
    public void psrad(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xE2, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrad  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psrad     mm0, [ebx]}
     */
    // Template#: 3736, Serial#: 3823
    public void psrad(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xE2, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrad  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psrad     mm0, mm0}
     */
    // Template#: 3737, Serial#: 3831
    public void psrad(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xE2, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrad  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psrad     mm0, [L1: 0x12345678]}
     */
    // Template#: 3738, Serial#: 3826
    public void m_psrad(MMXRegister destination, int m32) {
        assemble0268((byte) 0xE2, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrad  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrad     mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3739, Serial#: 3830
    public void psrad(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xE2, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrad  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrad     mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3740, Serial#: 3825
    public void m_psrad(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xE2, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrad  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psrad     mm0, [ebx + 305419896]}
     */
    // Template#: 3741, Serial#: 3829
    public void psrad(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xE2, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psraw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psraw     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3742, Serial#: 3855
    public void psraw(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xE1, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psraw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psraw     xmm0, [ebx + 18]}
     */
    // Template#: 3743, Serial#: 3854
    public void psraw(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xE1, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psraw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psraw     xmm0, ebx[esi * 4]}
     */
    // Template#: 3744, Serial#: 3851
    public void psraw(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xE1, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psraw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psraw     xmm0, [ebx]}
     */
    // Template#: 3745, Serial#: 3850
    public void psraw(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xE1, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psraw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psraw     xmm0, xmm0}
     */
    // Template#: 3746, Serial#: 3858
    public void psraw(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xE1, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psraw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psraw     xmm0, [L1: 0x12345678]}
     */
    // Template#: 3747, Serial#: 3853
    public void m_psraw(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xE1, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psraw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psraw     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3748, Serial#: 3857
    public void psraw(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xE1, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psraw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psraw     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3749, Serial#: 3852
    public void m_psraw(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xE1, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psraw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psraw     xmm0, [ebx + 305419896]}
     */
    // Template#: 3750, Serial#: 3856
    public void psraw(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xE1, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psraw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code psraw     mm0, 0x12}
     */
    // Template#: 3751, Serial#: 3209
    public void psraw(MMXRegister destination, byte imm8) {
        assemble0494((byte) 0x71, (byte) 0x04, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code psraw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psraw     mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3752, Serial#: 3819
    public void psraw(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xE1, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psraw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psraw     mm0, [ebx + 18]}
     */
    // Template#: 3753, Serial#: 3818
    public void psraw(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xE1, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psraw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psraw     mm0, ebx[esi * 4]}
     */
    // Template#: 3754, Serial#: 3815
    public void psraw(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xE1, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psraw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psraw     mm0, [ebx]}
     */
    // Template#: 3755, Serial#: 3814
    public void psraw(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xE1, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psraw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psraw     mm0, mm0}
     */
    // Template#: 3756, Serial#: 3822
    public void psraw(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xE1, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psraw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psraw     mm0, [L1: 0x12345678]}
     */
    // Template#: 3757, Serial#: 3817
    public void m_psraw(MMXRegister destination, int m32) {
        assemble0268((byte) 0xE1, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psraw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psraw     mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3758, Serial#: 3821
    public void psraw(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xE1, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psraw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psraw     mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3759, Serial#: 3816
    public void m_psraw(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xE1, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psraw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psraw     mm0, [ebx + 305419896]}
     */
    // Template#: 3760, Serial#: 3820
    public void psraw(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xE1, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrld  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrld     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3761, Serial#: 3771
    public void psrld(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xD2, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrld  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psrld     xmm0, [ebx + 18]}
     */
    // Template#: 3762, Serial#: 3770
    public void psrld(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xD2, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrld  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrld     xmm0, ebx[esi * 4]}
     */
    // Template#: 3763, Serial#: 3767
    public void psrld(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xD2, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrld  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psrld     xmm0, [ebx]}
     */
    // Template#: 3764, Serial#: 3766
    public void psrld(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xD2, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrld  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psrld     xmm0, xmm0}
     */
    // Template#: 3765, Serial#: 3774
    public void psrld(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xD2, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrld  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psrld     xmm0, [L1: 0x12345678]}
     */
    // Template#: 3766, Serial#: 3769
    public void m_psrld(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xD2, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrld  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrld     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3767, Serial#: 3773
    public void psrld(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xD2, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrld  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrld     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3768, Serial#: 3768
    public void m_psrld(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xD2, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrld  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psrld     xmm0, [ebx + 305419896]}
     */
    // Template#: 3769, Serial#: 3772
    public void psrld(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xD2, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrld  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code psrld     mm0, 0x12}
     */
    // Template#: 3770, Serial#: 3211
    public void psrld(MMXRegister destination, byte imm8) {
        assemble0494((byte) 0x72, (byte) 0x02, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrld  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrld     mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3771, Serial#: 3726
    public void psrld(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xD2, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrld  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psrld     mm0, [ebx + 18]}
     */
    // Template#: 3772, Serial#: 3725
    public void psrld(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xD2, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrld  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrld     mm0, ebx[esi * 4]}
     */
    // Template#: 3773, Serial#: 3722
    public void psrld(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xD2, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrld  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psrld     mm0, [ebx]}
     */
    // Template#: 3774, Serial#: 3721
    public void psrld(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xD2, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrld  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psrld     mm0, mm0}
     */
    // Template#: 3775, Serial#: 3729
    public void psrld(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xD2, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrld  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psrld     mm0, [L1: 0x12345678]}
     */
    // Template#: 3776, Serial#: 3724
    public void m_psrld(MMXRegister destination, int m32) {
        assemble0268((byte) 0xD2, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrld  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrld     mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3777, Serial#: 3728
    public void psrld(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xD2, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrld  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrld     mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3778, Serial#: 3723
    public void m_psrld(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xD2, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrld  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psrld     mm0, [ebx + 305419896]}
     */
    // Template#: 3779, Serial#: 3727
    public void psrld(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xD2, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrldq  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code psrldq    xmm0, 0x12}
     */
    // Template#: 3780, Serial#: 3317
    public void psrldq(IA32XMMRegister destination, byte imm8) {
        assemble0495((byte) 0x73, (byte) 0x03, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlq  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code psrlq     xmm0, 0x12}
     */
    // Template#: 3781, Serial#: 3316
    public void psrlq(IA32XMMRegister destination, byte imm8) {
        assemble0495((byte) 0x73, (byte) 0x02, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlq  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrlq     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3782, Serial#: 3780
    public void psrlq(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xD3, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlq  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psrlq     xmm0, [ebx + 18]}
     */
    // Template#: 3783, Serial#: 3779
    public void psrlq(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xD3, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlq  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrlq     xmm0, ebx[esi * 4]}
     */
    // Template#: 3784, Serial#: 3776
    public void psrlq(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xD3, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psrlq     xmm0, [ebx]}
     */
    // Template#: 3785, Serial#: 3775
    public void psrlq(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xD3, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psrlq     xmm0, xmm0}
     */
    // Template#: 3786, Serial#: 3783
    public void psrlq(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xD3, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlq  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psrlq     xmm0, [L1: 0x12345678]}
     */
    // Template#: 3787, Serial#: 3778
    public void m_psrlq(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xD3, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlq  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrlq     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3788, Serial#: 3782
    public void psrlq(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xD3, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlq  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrlq     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3789, Serial#: 3777
    public void m_psrlq(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xD3, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlq  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psrlq     xmm0, [ebx + 305419896]}
     */
    // Template#: 3790, Serial#: 3781
    public void psrlq(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xD3, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlq  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code psrlq     mm0, 0x12}
     */
    // Template#: 3791, Serial#: 3214
    public void psrlq(MMXRegister destination, byte imm8) {
        assemble0494((byte) 0x73, (byte) 0x02, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlq  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrlq     mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3792, Serial#: 3735
    public void psrlq(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xD3, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlq  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psrlq     mm0, [ebx + 18]}
     */
    // Template#: 3793, Serial#: 3734
    public void psrlq(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xD3, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlq  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrlq     mm0, ebx[esi * 4]}
     */
    // Template#: 3794, Serial#: 3731
    public void psrlq(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xD3, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psrlq     mm0, [ebx]}
     */
    // Template#: 3795, Serial#: 3730
    public void psrlq(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xD3, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psrlq     mm0, mm0}
     */
    // Template#: 3796, Serial#: 3738
    public void psrlq(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xD3, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlq  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psrlq     mm0, [L1: 0x12345678]}
     */
    // Template#: 3797, Serial#: 3733
    public void m_psrlq(MMXRegister destination, int m32) {
        assemble0268((byte) 0xD3, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlq  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrlq     mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3798, Serial#: 3737
    public void psrlq(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xD3, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlq  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrlq     mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3799, Serial#: 3732
    public void m_psrlq(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xD3, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlq  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psrlq     mm0, [ebx + 305419896]}
     */
    // Template#: 3800, Serial#: 3736
    public void psrlq(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xD3, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrlw     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3801, Serial#: 3762
    public void psrlw(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xD1, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psrlw     xmm0, [ebx + 18]}
     */
    // Template#: 3802, Serial#: 3761
    public void psrlw(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xD1, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrlw     xmm0, ebx[esi * 4]}
     */
    // Template#: 3803, Serial#: 3758
    public void psrlw(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xD1, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psrlw     xmm0, [ebx]}
     */
    // Template#: 3804, Serial#: 3757
    public void psrlw(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xD1, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psrlw     xmm0, xmm0}
     */
    // Template#: 3805, Serial#: 3765
    public void psrlw(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xD1, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psrlw     xmm0, [L1: 0x12345678]}
     */
    // Template#: 3806, Serial#: 3760
    public void m_psrlw(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xD1, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrlw     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3807, Serial#: 3764
    public void psrlw(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xD1, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrlw     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3808, Serial#: 3759
    public void m_psrlw(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xD1, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psrlw     xmm0, [ebx + 305419896]}
     */
    // Template#: 3809, Serial#: 3763
    public void psrlw(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xD1, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code psrlw     mm0, 0x12}
     */
    // Template#: 3810, Serial#: 3208
    public void psrlw(MMXRegister destination, byte imm8) {
        assemble0494((byte) 0x71, (byte) 0x02, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrlw     mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3811, Serial#: 3717
    public void psrlw(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xD1, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psrlw     mm0, [ebx + 18]}
     */
    // Template#: 3812, Serial#: 3716
    public void psrlw(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xD1, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrlw     mm0, ebx[esi * 4]}
     */
    // Template#: 3813, Serial#: 3713
    public void psrlw(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xD1, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psrlw     mm0, [ebx]}
     */
    // Template#: 3814, Serial#: 3712
    public void psrlw(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xD1, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psrlw     mm0, mm0}
     */
    // Template#: 3815, Serial#: 3720
    public void psrlw(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xD1, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psrlw     mm0, [L1: 0x12345678]}
     */
    // Template#: 3816, Serial#: 3715
    public void m_psrlw(MMXRegister destination, int m32) {
        assemble0268((byte) 0xD1, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrlw     mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3817, Serial#: 3719
    public void psrlw(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xD1, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psrlw     mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3818, Serial#: 3714
    public void m_psrlw(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xD1, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psrlw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psrlw     mm0, [ebx + 305419896]}
     */
    // Template#: 3819, Serial#: 3718
    public void psrlw(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xD1, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubb     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3820, Serial#: 5470
    public void psubb(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xF8, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psubb     xmm0, [ebx + 18]}
     */
    // Template#: 3821, Serial#: 5469
    public void psubb(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xF8, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubb     xmm0, ebx[esi * 4]}
     */
    // Template#: 3822, Serial#: 5466
    public void psubb(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xF8, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubb     xmm0, [ebx]}
     */
    // Template#: 3823, Serial#: 5465
    public void psubb(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xF8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubb     xmm0, xmm0}
     */
    // Template#: 3824, Serial#: 5473
    public void psubb(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xF8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psubb     xmm0, [L1: 0x12345678]}
     */
    // Template#: 3825, Serial#: 5468
    public void m_psubb(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xF8, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubb     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3826, Serial#: 5472
    public void psubb(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xF8, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubb     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3827, Serial#: 5467
    public void m_psubb(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xF8, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psubb     xmm0, [ebx + 305419896]}
     */
    // Template#: 3828, Serial#: 5471
    public void psubb(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xF8, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubb     mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3829, Serial#: 5416
    public void psubb(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xF8, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psubb     mm0, [ebx + 18]}
     */
    // Template#: 3830, Serial#: 5415
    public void psubb(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xF8, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubb     mm0, ebx[esi * 4]}
     */
    // Template#: 3831, Serial#: 5412
    public void psubb(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xF8, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubb     mm0, [ebx]}
     */
    // Template#: 3832, Serial#: 5411
    public void psubb(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xF8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubb     mm0, mm0}
     */
    // Template#: 3833, Serial#: 5419
    public void psubb(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xF8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psubb     mm0, [L1: 0x12345678]}
     */
    // Template#: 3834, Serial#: 5414
    public void m_psubb(MMXRegister destination, int m32) {
        assemble0268((byte) 0xF8, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubb     mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3835, Serial#: 5418
    public void psubb(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xF8, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubb     mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3836, Serial#: 5413
    public void m_psubb(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xF8, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psubb     mm0, [ebx + 305419896]}
     */
    // Template#: 3837, Serial#: 5417
    public void psubb(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xF8, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubd     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3838, Serial#: 5488
    public void psubd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xFA, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psubd     xmm0, [ebx + 18]}
     */
    // Template#: 3839, Serial#: 5487
    public void psubd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xFA, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubd     xmm0, ebx[esi * 4]}
     */
    // Template#: 3840, Serial#: 5484
    public void psubd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xFA, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubd     xmm0, [ebx]}
     */
    // Template#: 3841, Serial#: 5483
    public void psubd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xFA, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubd     xmm0, xmm0}
     */
    // Template#: 3842, Serial#: 5491
    public void psubd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xFA, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psubd     xmm0, [L1: 0x12345678]}
     */
    // Template#: 3843, Serial#: 5486
    public void m_psubd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xFA, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubd     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3844, Serial#: 5490
    public void psubd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xFA, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubd     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3845, Serial#: 5485
    public void m_psubd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xFA, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psubd     xmm0, [ebx + 305419896]}
     */
    // Template#: 3846, Serial#: 5489
    public void psubd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xFA, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubd     mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3847, Serial#: 5434
    public void psubd(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xFA, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psubd     mm0, [ebx + 18]}
     */
    // Template#: 3848, Serial#: 5433
    public void psubd(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xFA, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubd     mm0, ebx[esi * 4]}
     */
    // Template#: 3849, Serial#: 5430
    public void psubd(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xFA, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubd     mm0, [ebx]}
     */
    // Template#: 3850, Serial#: 5429
    public void psubd(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xFA, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubd     mm0, mm0}
     */
    // Template#: 3851, Serial#: 5437
    public void psubd(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xFA, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psubd     mm0, [L1: 0x12345678]}
     */
    // Template#: 3852, Serial#: 5432
    public void m_psubd(MMXRegister destination, int m32) {
        assemble0268((byte) 0xFA, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubd     mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3853, Serial#: 5436
    public void psubd(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xFA, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubd     mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3854, Serial#: 5431
    public void m_psubd(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xFA, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psubd     mm0, [ebx + 305419896]}
     */
    // Template#: 3855, Serial#: 5435
    public void psubd(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xFA, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubq  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubq     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3856, Serial#: 5497
    public void psubq(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xFB, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubq  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psubq     xmm0, [ebx + 18]}
     */
    // Template#: 3857, Serial#: 5496
    public void psubq(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xFB, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubq  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubq     xmm0, ebx[esi * 4]}
     */
    // Template#: 3858, Serial#: 5493
    public void psubq(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xFB, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubq     xmm0, [ebx]}
     */
    // Template#: 3859, Serial#: 5492
    public void psubq(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xFB, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubq     xmm0, xmm0}
     */
    // Template#: 3860, Serial#: 5500
    public void psubq(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xFB, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubq  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psubq     xmm0, [L1: 0x12345678]}
     */
    // Template#: 3861, Serial#: 5495
    public void m_psubq(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xFB, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubq  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubq     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3862, Serial#: 5499
    public void psubq(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xFB, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubq  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubq     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3863, Serial#: 5494
    public void m_psubq(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xFB, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubq  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psubq     xmm0, [ebx + 305419896]}
     */
    // Template#: 3864, Serial#: 5498
    public void psubq(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xFB, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubsb    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3865, Serial#: 5344
    public void psubsb(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xE8, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psubsb    xmm0, [ebx + 18]}
     */
    // Template#: 3866, Serial#: 5343
    public void psubsb(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xE8, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubsb    xmm0, ebx[esi * 4]}
     */
    // Template#: 3867, Serial#: 5340
    public void psubsb(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xE8, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubsb    xmm0, [ebx]}
     */
    // Template#: 3868, Serial#: 5339
    public void psubsb(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xE8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubsb    xmm0, xmm0}
     */
    // Template#: 3869, Serial#: 5347
    public void psubsb(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xE8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psubsb    xmm0, [L1: 0x12345678]}
     */
    // Template#: 3870, Serial#: 5342
    public void m_psubsb(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xE8, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubsb    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3871, Serial#: 5346
    public void psubsb(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xE8, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubsb    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3872, Serial#: 5341
    public void m_psubsb(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xE8, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psubsb    xmm0, [ebx + 305419896]}
     */
    // Template#: 3873, Serial#: 5345
    public void psubsb(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xE8, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubsb    mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3874, Serial#: 5290
    public void psubsb(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xE8, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psubsb    mm0, [ebx + 18]}
     */
    // Template#: 3875, Serial#: 5289
    public void psubsb(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xE8, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubsb    mm0, ebx[esi * 4]}
     */
    // Template#: 3876, Serial#: 5286
    public void psubsb(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xE8, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubsb    mm0, [ebx]}
     */
    // Template#: 3877, Serial#: 5285
    public void psubsb(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xE8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubsb    mm0, mm0}
     */
    // Template#: 3878, Serial#: 5293
    public void psubsb(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xE8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psubsb    mm0, [L1: 0x12345678]}
     */
    // Template#: 3879, Serial#: 5288
    public void m_psubsb(MMXRegister destination, int m32) {
        assemble0268((byte) 0xE8, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubsb    mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3880, Serial#: 5292
    public void psubsb(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xE8, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubsb    mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3881, Serial#: 5287
    public void m_psubsb(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xE8, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psubsb    mm0, [ebx + 305419896]}
     */
    // Template#: 3882, Serial#: 5291
    public void psubsb(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xE8, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubsw    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3883, Serial#: 5353
    public void psubsw(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xE9, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psubsw    xmm0, [ebx + 18]}
     */
    // Template#: 3884, Serial#: 5352
    public void psubsw(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xE9, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubsw    xmm0, ebx[esi * 4]}
     */
    // Template#: 3885, Serial#: 5349
    public void psubsw(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xE9, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubsw    xmm0, [ebx]}
     */
    // Template#: 3886, Serial#: 5348
    public void psubsw(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xE9, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubsw    xmm0, xmm0}
     */
    // Template#: 3887, Serial#: 5356
    public void psubsw(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xE9, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psubsw    xmm0, [L1: 0x12345678]}
     */
    // Template#: 3888, Serial#: 5351
    public void m_psubsw(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xE9, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubsw    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3889, Serial#: 5355
    public void psubsw(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xE9, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubsw    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3890, Serial#: 5350
    public void m_psubsw(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xE9, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psubsw    xmm0, [ebx + 305419896]}
     */
    // Template#: 3891, Serial#: 5354
    public void psubsw(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xE9, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubsw    mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3892, Serial#: 5299
    public void psubsw(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xE9, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psubsw    mm0, [ebx + 18]}
     */
    // Template#: 3893, Serial#: 5298
    public void psubsw(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xE9, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubsw    mm0, ebx[esi * 4]}
     */
    // Template#: 3894, Serial#: 5295
    public void psubsw(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xE9, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubsw    mm0, [ebx]}
     */
    // Template#: 3895, Serial#: 5294
    public void psubsw(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xE9, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubsw    mm0, mm0}
     */
    // Template#: 3896, Serial#: 5302
    public void psubsw(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xE9, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psubsw    mm0, [L1: 0x12345678]}
     */
    // Template#: 3897, Serial#: 5297
    public void m_psubsw(MMXRegister destination, int m32) {
        assemble0268((byte) 0xE9, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubsw    mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3898, Serial#: 5301
    public void psubsw(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xE9, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubsw    mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3899, Serial#: 5296
    public void m_psubsw(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xE9, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubsw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psubsw    mm0, [ebx + 305419896]}
     */
    // Template#: 3900, Serial#: 5300
    public void psubsw(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xE9, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubusb   xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3901, Serial#: 5218
    public void psubusb(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xD8, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psubusb   xmm0, [ebx + 18]}
     */
    // Template#: 3902, Serial#: 5217
    public void psubusb(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xD8, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubusb   xmm0, ebx[esi * 4]}
     */
    // Template#: 3903, Serial#: 5214
    public void psubusb(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xD8, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubusb   xmm0, [ebx]}
     */
    // Template#: 3904, Serial#: 5213
    public void psubusb(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xD8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubusb   xmm0, xmm0}
     */
    // Template#: 3905, Serial#: 5221
    public void psubusb(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xD8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psubusb   xmm0, [L1: 0x12345678]}
     */
    // Template#: 3906, Serial#: 5216
    public void m_psubusb(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xD8, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubusb   xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3907, Serial#: 5220
    public void psubusb(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xD8, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubusb   xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3908, Serial#: 5215
    public void m_psubusb(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xD8, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psubusb   xmm0, [ebx + 305419896]}
     */
    // Template#: 3909, Serial#: 5219
    public void psubusb(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xD8, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubusb   mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3910, Serial#: 5164
    public void psubusb(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xD8, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psubusb   mm0, [ebx + 18]}
     */
    // Template#: 3911, Serial#: 5163
    public void psubusb(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xD8, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubusb   mm0, ebx[esi * 4]}
     */
    // Template#: 3912, Serial#: 5160
    public void psubusb(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xD8, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubusb   mm0, [ebx]}
     */
    // Template#: 3913, Serial#: 5159
    public void psubusb(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xD8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubusb   mm0, mm0}
     */
    // Template#: 3914, Serial#: 5167
    public void psubusb(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xD8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psubusb   mm0, [L1: 0x12345678]}
     */
    // Template#: 3915, Serial#: 5162
    public void m_psubusb(MMXRegister destination, int m32) {
        assemble0268((byte) 0xD8, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubusb   mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3916, Serial#: 5166
    public void psubusb(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xD8, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubusb   mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3917, Serial#: 5161
    public void m_psubusb(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xD8, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psubusb   mm0, [ebx + 305419896]}
     */
    // Template#: 3918, Serial#: 5165
    public void psubusb(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xD8, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubusw   xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3919, Serial#: 5227
    public void psubusw(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xD9, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psubusw   xmm0, [ebx + 18]}
     */
    // Template#: 3920, Serial#: 5226
    public void psubusw(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xD9, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubusw   xmm0, ebx[esi * 4]}
     */
    // Template#: 3921, Serial#: 5223
    public void psubusw(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xD9, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubusw   xmm0, [ebx]}
     */
    // Template#: 3922, Serial#: 5222
    public void psubusw(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xD9, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubusw   xmm0, xmm0}
     */
    // Template#: 3923, Serial#: 5230
    public void psubusw(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xD9, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psubusw   xmm0, [L1: 0x12345678]}
     */
    // Template#: 3924, Serial#: 5225
    public void m_psubusw(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xD9, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubusw   xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3925, Serial#: 5229
    public void psubusw(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xD9, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubusw   xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3926, Serial#: 5224
    public void m_psubusw(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xD9, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psubusw   xmm0, [ebx + 305419896]}
     */
    // Template#: 3927, Serial#: 5228
    public void psubusw(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xD9, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubusw   mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3928, Serial#: 5173
    public void psubusw(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xD9, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psubusw   mm0, [ebx + 18]}
     */
    // Template#: 3929, Serial#: 5172
    public void psubusw(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xD9, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubusw   mm0, ebx[esi * 4]}
     */
    // Template#: 3930, Serial#: 5169
    public void psubusw(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xD9, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubusw   mm0, [ebx]}
     */
    // Template#: 3931, Serial#: 5168
    public void psubusw(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xD9, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubusw   mm0, mm0}
     */
    // Template#: 3932, Serial#: 5176
    public void psubusw(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xD9, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psubusw   mm0, [L1: 0x12345678]}
     */
    // Template#: 3933, Serial#: 5171
    public void m_psubusw(MMXRegister destination, int m32) {
        assemble0268((byte) 0xD9, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubusw   mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3934, Serial#: 5175
    public void psubusw(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xD9, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubusw   mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3935, Serial#: 5170
    public void m_psubusw(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xD9, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubusw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psubusw   mm0, [ebx + 305419896]}
     */
    // Template#: 3936, Serial#: 5174
    public void psubusw(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xD9, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubw     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3937, Serial#: 5479
    public void psubw(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xF9, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psubw     xmm0, [ebx + 18]}
     */
    // Template#: 3938, Serial#: 5478
    public void psubw(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xF9, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubw     xmm0, ebx[esi * 4]}
     */
    // Template#: 3939, Serial#: 5475
    public void psubw(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xF9, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubw     xmm0, [ebx]}
     */
    // Template#: 3940, Serial#: 5474
    public void psubw(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xF9, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubw     xmm0, xmm0}
     */
    // Template#: 3941, Serial#: 5482
    public void psubw(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xF9, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psubw     xmm0, [L1: 0x12345678]}
     */
    // Template#: 3942, Serial#: 5477
    public void m_psubw(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xF9, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubw     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3943, Serial#: 5481
    public void psubw(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xF9, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubw     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3944, Serial#: 5476
    public void m_psubw(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xF9, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psubw     xmm0, [ebx + 305419896]}
     */
    // Template#: 3945, Serial#: 5480
    public void psubw(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xF9, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubw     mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3946, Serial#: 5425
    public void psubw(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xF9, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code psubw     mm0, [ebx + 18]}
     */
    // Template#: 3947, Serial#: 5424
    public void psubw(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xF9, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubw     mm0, ebx[esi * 4]}
     */
    // Template#: 3948, Serial#: 5421
    public void psubw(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xF9, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubw     mm0, [ebx]}
     */
    // Template#: 3949, Serial#: 5420
    public void psubw(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xF9, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code psubw     mm0, mm0}
     */
    // Template#: 3950, Serial#: 5428
    public void psubw(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xF9, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code psubw     mm0, [L1: 0x12345678]}
     */
    // Template#: 3951, Serial#: 5423
    public void m_psubw(MMXRegister destination, int m32) {
        assemble0268((byte) 0xF9, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubw     mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3952, Serial#: 5427
    public void psubw(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xF9, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code psubw     mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3953, Serial#: 5422
    public void m_psubw(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xF9, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code psubw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code psubw     mm0, [ebx + 305419896]}
     */
    // Template#: 3954, Serial#: 5426
    public void psubw(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xF9, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhbw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhbw  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3955, Serial#: 4708
    public void punpckhbw(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x68, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhbw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhbw  xmm0, [ebx + 18]}
     */
    // Template#: 3956, Serial#: 4707
    public void punpckhbw(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x68, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhbw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhbw  xmm0, ebx[esi * 4]}
     */
    // Template#: 3957, Serial#: 4704
    public void punpckhbw(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x68, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhbw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhbw  xmm0, [ebx]}
     */
    // Template#: 3958, Serial#: 4703
    public void punpckhbw(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x68, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhbw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhbw  xmm0, xmm0}
     */
    // Template#: 3959, Serial#: 4711
    public void punpckhbw(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x68, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhbw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code punpckhbw  xmm0, [L1: 0x12345678]}
     */
    // Template#: 3960, Serial#: 4706
    public void m_punpckhbw(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x68, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhbw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhbw  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3961, Serial#: 4710
    public void punpckhbw(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x68, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhbw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhbw  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3962, Serial#: 4705
    public void m_punpckhbw(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x68, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhbw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhbw  xmm0, [ebx + 305419896]}
     */
    // Template#: 3963, Serial#: 4709
    public void punpckhbw(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x68, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhbw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhbw  mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3964, Serial#: 4654
    public void punpckhbw(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0x68, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhbw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhbw  mm0, [ebx + 18]}
     */
    // Template#: 3965, Serial#: 4653
    public void punpckhbw(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0x68, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhbw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhbw  mm0, ebx[esi * 4]}
     */
    // Template#: 3966, Serial#: 4650
    public void punpckhbw(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0x68, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhbw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhbw  mm0, [ebx]}
     */
    // Template#: 3967, Serial#: 4649
    public void punpckhbw(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0x68, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhbw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhbw  mm0, mm0}
     */
    // Template#: 3968, Serial#: 4657
    public void punpckhbw(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0x68, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhbw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code punpckhbw  mm0, [L1: 0x12345678]}
     */
    // Template#: 3969, Serial#: 4652
    public void m_punpckhbw(MMXRegister destination, int m32) {
        assemble0268((byte) 0x68, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhbw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhbw  mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3970, Serial#: 4656
    public void punpckhbw(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0x68, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhbw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhbw  mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3971, Serial#: 4651
    public void m_punpckhbw(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0x68, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhbw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhbw  mm0, [ebx + 305419896]}
     */
    // Template#: 3972, Serial#: 4655
    public void punpckhbw(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0x68, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhdq  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhdq  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3973, Serial#: 4726
    public void punpckhdq(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x6A, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhdq  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhdq  xmm0, [ebx + 18]}
     */
    // Template#: 3974, Serial#: 4725
    public void punpckhdq(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x6A, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhdq  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhdq  xmm0, ebx[esi * 4]}
     */
    // Template#: 3975, Serial#: 4722
    public void punpckhdq(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x6A, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhdq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhdq  xmm0, [ebx]}
     */
    // Template#: 3976, Serial#: 4721
    public void punpckhdq(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x6A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhdq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhdq  xmm0, xmm0}
     */
    // Template#: 3977, Serial#: 4729
    public void punpckhdq(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x6A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhdq  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code punpckhdq  xmm0, [L1: 0x12345678]}
     */
    // Template#: 3978, Serial#: 4724
    public void m_punpckhdq(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x6A, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhdq  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhdq  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3979, Serial#: 4728
    public void punpckhdq(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x6A, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhdq  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhdq  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3980, Serial#: 4723
    public void m_punpckhdq(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x6A, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhdq  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhdq  xmm0, [ebx + 305419896]}
     */
    // Template#: 3981, Serial#: 4727
    public void punpckhdq(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x6A, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhdq  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhdq  mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3982, Serial#: 4672
    public void punpckhdq(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0x6A, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhdq  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhdq  mm0, [ebx + 18]}
     */
    // Template#: 3983, Serial#: 4671
    public void punpckhdq(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0x6A, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhdq  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhdq  mm0, ebx[esi * 4]}
     */
    // Template#: 3984, Serial#: 4668
    public void punpckhdq(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0x6A, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhdq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhdq  mm0, [ebx]}
     */
    // Template#: 3985, Serial#: 4667
    public void punpckhdq(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0x6A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhdq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhdq  mm0, mm0}
     */
    // Template#: 3986, Serial#: 4675
    public void punpckhdq(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0x6A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhdq  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code punpckhdq  mm0, [L1: 0x12345678]}
     */
    // Template#: 3987, Serial#: 4670
    public void m_punpckhdq(MMXRegister destination, int m32) {
        assemble0268((byte) 0x6A, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhdq  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhdq  mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3988, Serial#: 4674
    public void punpckhdq(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0x6A, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhdq  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhdq  mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3989, Serial#: 4669
    public void m_punpckhdq(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0x6A, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhdq  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhdq  mm0, [ebx + 305419896]}
     */
    // Template#: 3990, Serial#: 4673
    public void punpckhdq(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0x6A, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhqdq  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhqdq  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 3991, Serial#: 4753
    public void punpckhqdq(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x6D, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhqdq  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhqdq  xmm0, [ebx + 18]}
     */
    // Template#: 3992, Serial#: 4752
    public void punpckhqdq(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x6D, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhqdq  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhqdq  xmm0, ebx[esi * 4]}
     */
    // Template#: 3993, Serial#: 4749
    public void punpckhqdq(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x6D, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhqdq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhqdq  xmm0, [ebx]}
     */
    // Template#: 3994, Serial#: 4748
    public void punpckhqdq(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x6D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhqdq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhqdq  xmm0, xmm0}
     */
    // Template#: 3995, Serial#: 4756
    public void punpckhqdq(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x6D, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhqdq  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code punpckhqdq  xmm0, [L1: 0x12345678]}
     */
    // Template#: 3996, Serial#: 4751
    public void m_punpckhqdq(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x6D, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhqdq  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhqdq  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 3997, Serial#: 4755
    public void punpckhqdq(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x6D, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhqdq  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhqdq  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 3998, Serial#: 4750
    public void m_punpckhqdq(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x6D, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhqdq  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhqdq  xmm0, [ebx + 305419896]}
     */
    // Template#: 3999, Serial#: 4754
    public void punpckhqdq(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x6D, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhwd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhwd  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 4000, Serial#: 4717
    public void punpckhwd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x69, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhwd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhwd  xmm0, [ebx + 18]}
     */
    // Template#: 4001, Serial#: 4716
    public void punpckhwd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x69, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhwd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhwd  xmm0, ebx[esi * 4]}
     */
    // Template#: 4002, Serial#: 4713
    public void punpckhwd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x69, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhwd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhwd  xmm0, [ebx]}
     */
    // Template#: 4003, Serial#: 4712
    public void punpckhwd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x69, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhwd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhwd  xmm0, xmm0}
     */
    // Template#: 4004, Serial#: 4720
    public void punpckhwd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x69, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhwd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code punpckhwd  xmm0, [L1: 0x12345678]}
     */
    // Template#: 4005, Serial#: 4715
    public void m_punpckhwd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x69, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhwd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhwd  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 4006, Serial#: 4719
    public void punpckhwd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x69, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhwd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhwd  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 4007, Serial#: 4714
    public void m_punpckhwd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x69, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhwd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhwd  xmm0, [ebx + 305419896]}
     */
    // Template#: 4008, Serial#: 4718
    public void punpckhwd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x69, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhwd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhwd  mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 4009, Serial#: 4663
    public void punpckhwd(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0x69, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhwd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhwd  mm0, [ebx + 18]}
     */
    // Template#: 4010, Serial#: 4662
    public void punpckhwd(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0x69, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhwd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhwd  mm0, ebx[esi * 4]}
     */
    // Template#: 4011, Serial#: 4659
    public void punpckhwd(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0x69, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhwd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhwd  mm0, [ebx]}
     */
    // Template#: 4012, Serial#: 4658
    public void punpckhwd(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0x69, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhwd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhwd  mm0, mm0}
     */
    // Template#: 4013, Serial#: 4666
    public void punpckhwd(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0x69, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhwd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code punpckhwd  mm0, [L1: 0x12345678]}
     */
    // Template#: 4014, Serial#: 4661
    public void m_punpckhwd(MMXRegister destination, int m32) {
        assemble0268((byte) 0x69, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhwd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhwd  mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 4015, Serial#: 4665
    public void punpckhwd(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0x69, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhwd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckhwd  mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 4016, Serial#: 4660
    public void m_punpckhwd(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0x69, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckhwd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckhwd  mm0, [ebx + 305419896]}
     */
    // Template#: 4017, Serial#: 4664
    public void punpckhwd(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0x69, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklbw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpcklbw  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 4018, Serial#: 3249
    public void punpcklbw(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x60, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklbw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code punpcklbw  xmm0, [ebx + 18]}
     */
    // Template#: 4019, Serial#: 3248
    public void punpcklbw(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x60, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklbw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpcklbw  xmm0, ebx[esi * 4]}
     */
    // Template#: 4020, Serial#: 3245
    public void punpcklbw(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x60, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklbw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpcklbw  xmm0, [ebx]}
     */
    // Template#: 4021, Serial#: 3244
    public void punpcklbw(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x60, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklbw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpcklbw  xmm0, xmm0}
     */
    // Template#: 4022, Serial#: 3252
    public void punpcklbw(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x60, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklbw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code punpcklbw  xmm0, [L1: 0x12345678]}
     */
    // Template#: 4023, Serial#: 3247
    public void m_punpcklbw(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x60, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklbw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpcklbw  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 4024, Serial#: 3251
    public void punpcklbw(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x60, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklbw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpcklbw  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 4025, Serial#: 3246
    public void m_punpcklbw(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x60, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklbw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code punpcklbw  xmm0, [ebx + 305419896]}
     */
    // Template#: 4026, Serial#: 3250
    public void punpcklbw(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x60, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklbw  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpcklbw  mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 4027, Serial#: 3132
    public void punpcklbw(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0x60, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklbw  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code punpcklbw  mm0, [ebx + 18]}
     */
    // Template#: 4028, Serial#: 3131
    public void punpcklbw(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0x60, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklbw  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpcklbw  mm0, ebx[esi * 4]}
     */
    // Template#: 4029, Serial#: 3128
    public void punpcklbw(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0x60, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklbw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpcklbw  mm0, [ebx]}
     */
    // Template#: 4030, Serial#: 3127
    public void punpcklbw(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0x60, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklbw  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpcklbw  mm0, mm0}
     */
    // Template#: 4031, Serial#: 3135
    public void punpcklbw(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0x60, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklbw  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code punpcklbw  mm0, [L1: 0x12345678]}
     */
    // Template#: 4032, Serial#: 3130
    public void m_punpcklbw(MMXRegister destination, int m32) {
        assemble0268((byte) 0x60, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklbw  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpcklbw  mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 4033, Serial#: 3134
    public void punpcklbw(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0x60, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklbw  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpcklbw  mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 4034, Serial#: 3129
    public void m_punpcklbw(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0x60, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklbw  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code punpcklbw  mm0, [ebx + 305419896]}
     */
    // Template#: 4035, Serial#: 3133
    public void punpcklbw(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0x60, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckldq  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckldq  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 4036, Serial#: 3267
    public void punpckldq(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x62, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckldq  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckldq  xmm0, [ebx + 18]}
     */
    // Template#: 4037, Serial#: 3266
    public void punpckldq(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x62, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckldq  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckldq  xmm0, ebx[esi * 4]}
     */
    // Template#: 4038, Serial#: 3263
    public void punpckldq(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x62, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckldq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckldq  xmm0, [ebx]}
     */
    // Template#: 4039, Serial#: 3262
    public void punpckldq(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x62, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckldq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckldq  xmm0, xmm0}
     */
    // Template#: 4040, Serial#: 3270
    public void punpckldq(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x62, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckldq  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code punpckldq  xmm0, [L1: 0x12345678]}
     */
    // Template#: 4041, Serial#: 3265
    public void m_punpckldq(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x62, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckldq  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckldq  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 4042, Serial#: 3269
    public void punpckldq(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x62, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckldq  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckldq  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 4043, Serial#: 3264
    public void m_punpckldq(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x62, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckldq  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckldq  xmm0, [ebx + 305419896]}
     */
    // Template#: 4044, Serial#: 3268
    public void punpckldq(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x62, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckldq  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckldq  mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 4045, Serial#: 3150
    public void punpckldq(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0x62, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckldq  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckldq  mm0, [ebx + 18]}
     */
    // Template#: 4046, Serial#: 3149
    public void punpckldq(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0x62, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckldq  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckldq  mm0, ebx[esi * 4]}
     */
    // Template#: 4047, Serial#: 3146
    public void punpckldq(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0x62, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckldq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckldq  mm0, [ebx]}
     */
    // Template#: 4048, Serial#: 3145
    public void punpckldq(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0x62, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckldq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckldq  mm0, mm0}
     */
    // Template#: 4049, Serial#: 3153
    public void punpckldq(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0x62, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckldq  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code punpckldq  mm0, [L1: 0x12345678]}
     */
    // Template#: 4050, Serial#: 3148
    public void m_punpckldq(MMXRegister destination, int m32) {
        assemble0268((byte) 0x62, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckldq  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckldq  mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 4051, Serial#: 3152
    public void punpckldq(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0x62, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckldq  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpckldq  mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 4052, Serial#: 3147
    public void m_punpckldq(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0x62, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpckldq  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code punpckldq  mm0, [ebx + 305419896]}
     */
    // Template#: 4053, Serial#: 3151
    public void punpckldq(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0x62, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklqdq  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpcklqdq  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 4054, Serial#: 4744
    public void punpcklqdq(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x6C, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklqdq  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code punpcklqdq  xmm0, [ebx + 18]}
     */
    // Template#: 4055, Serial#: 4743
    public void punpcklqdq(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x6C, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklqdq  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpcklqdq  xmm0, ebx[esi * 4]}
     */
    // Template#: 4056, Serial#: 4740
    public void punpcklqdq(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x6C, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklqdq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpcklqdq  xmm0, [ebx]}
     */
    // Template#: 4057, Serial#: 4739
    public void punpcklqdq(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x6C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklqdq  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpcklqdq  xmm0, xmm0}
     */
    // Template#: 4058, Serial#: 4747
    public void punpcklqdq(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x6C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklqdq  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code punpcklqdq  xmm0, [L1: 0x12345678]}
     */
    // Template#: 4059, Serial#: 4742
    public void m_punpcklqdq(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x6C, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklqdq  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpcklqdq  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 4060, Serial#: 4746
    public void punpcklqdq(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x6C, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklqdq  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpcklqdq  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 4061, Serial#: 4741
    public void m_punpcklqdq(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x6C, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklqdq  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code punpcklqdq  xmm0, [ebx + 305419896]}
     */
    // Template#: 4062, Serial#: 4745
    public void punpcklqdq(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x6C, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklwd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpcklwd  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 4063, Serial#: 3258
    public void punpcklwd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x61, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklwd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code punpcklwd  xmm0, [ebx + 18]}
     */
    // Template#: 4064, Serial#: 3257
    public void punpcklwd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x61, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklwd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpcklwd  xmm0, ebx[esi * 4]}
     */
    // Template#: 4065, Serial#: 3254
    public void punpcklwd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x61, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklwd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpcklwd  xmm0, [ebx]}
     */
    // Template#: 4066, Serial#: 3253
    public void punpcklwd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x61, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklwd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpcklwd  xmm0, xmm0}
     */
    // Template#: 4067, Serial#: 3261
    public void punpcklwd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x61, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklwd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code punpcklwd  xmm0, [L1: 0x12345678]}
     */
    // Template#: 4068, Serial#: 3256
    public void m_punpcklwd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x61, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklwd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpcklwd  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 4069, Serial#: 3260
    public void punpcklwd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x61, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklwd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpcklwd  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 4070, Serial#: 3255
    public void m_punpcklwd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x61, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklwd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code punpcklwd  xmm0, [ebx + 305419896]}
     */
    // Template#: 4071, Serial#: 3259
    public void punpcklwd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x61, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklwd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpcklwd  mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 4072, Serial#: 3141
    public void punpcklwd(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0x61, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklwd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code punpcklwd  mm0, [ebx + 18]}
     */
    // Template#: 4073, Serial#: 3140
    public void punpcklwd(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0x61, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklwd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpcklwd  mm0, ebx[esi * 4]}
     */
    // Template#: 4074, Serial#: 3137
    public void punpcklwd(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0x61, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklwd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpcklwd  mm0, [ebx]}
     */
    // Template#: 4075, Serial#: 3136
    public void punpcklwd(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0x61, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklwd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code punpcklwd  mm0, mm0}
     */
    // Template#: 4076, Serial#: 3144
    public void punpcklwd(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0x61, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklwd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code punpcklwd  mm0, [L1: 0x12345678]}
     */
    // Template#: 4077, Serial#: 3139
    public void m_punpcklwd(MMXRegister destination, int m32) {
        assemble0268((byte) 0x61, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklwd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpcklwd  mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 4078, Serial#: 3143
    public void punpcklwd(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0x61, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklwd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code punpcklwd  mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 4079, Serial#: 3138
    public void m_punpcklwd(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0x61, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code punpcklwd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code punpcklwd  mm0, [ebx + 305419896]}
     */
    // Template#: 4080, Serial#: 3142
    public void punpcklwd(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0x61, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code push  }
     * Example disassembly syntax: {@code push      es}
     */
    // Template#: 4081, Serial#: 55
    public void push_ES() {
        assemble0001((byte) 0x06);
    }

    /**
     * Pseudo-external assembler syntax: {@code push  }
     * Example disassembly syntax: {@code push      ss}
     */
    // Template#: 4082, Serial#: 111
    public void push_SS() {
        assemble0001((byte) 0x16);
    }

    /**
     * Pseudo-external assembler syntax: {@code push  }
     * Example disassembly syntax: {@code push      cs}
     */
    // Template#: 4083, Serial#: 1590
    public void push_CS() {
        assemble0001((byte) 0x0E);
    }

    /**
     * Pseudo-external assembler syntax: {@code push  }
     * Example disassembly syntax: {@code push      ds}
     */
    // Template#: 4084, Serial#: 1645
    public void push_DS() {
        assemble0001((byte) 0x1E);
    }

    /**
     * Pseudo-external assembler syntax: {@code push  }
     * Example disassembly syntax: {@code push      fs}
     */
    // Template#: 4085, Serial#: 3409
    public void push_FS() {
        assemble0198((byte) 0xA0);
    }

    /**
     * Pseudo-external assembler syntax: {@code push  }
     * Example disassembly syntax: {@code push      gs}
     */
    // Template#: 4086, Serial#: 4950
    public void push_GS() {
        assemble0198((byte) 0xA8);
    }

    /**
     * Pseudo-external assembler syntax: {@code push  }<i>imm8</i>
     * Example disassembly syntax: {@code push      0x12}
     */
    // Template#: 4087, Serial#: 1783
    public void push(byte imm8) {
        assemble0002((byte) 0x6A, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code push  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code push      ebx[esi * 4 + 18]}
     */
    // Template#: 4088, Serial#: 2461
    public void push(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xFF, (byte) 0x06, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code push  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code push      [ebx + 18]}
     */
    // Template#: 4089, Serial#: 2460
    public void push(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xFF, (byte) 0x06, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code push  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code push      ebx[esi * 4]}
     */
    // Template#: 4090, Serial#: 2449
    public void push(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xFF, (byte) 0x06, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code push  }<i>register</i>
     * Example disassembly syntax: {@code push      ax}
     */
    // Template#: 4091, Serial#: 228
    public void push(IA32GeneralRegister16 register) {
        assemble0295((byte) 0x50, register);
    }

    /**
     * Pseudo-external assembler syntax: {@code push  }<i>register</i>
     * Example disassembly syntax: {@code push      eax}
     */
    // Template#: 4092, Serial#: 227
    public void push(IA32GeneralRegister32 register) {
        assemble0296((byte) 0x50, register);
    }

    /**
     * Pseudo-external assembler syntax: {@code push  }<i>destination</i>
     * Example disassembly syntax: {@code push      [ebx]}
     */
    // Template#: 4093, Serial#: 2448
    public void push(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xFF, (byte) 0x06, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code pushl  }<i>imm32</i>
     * Example disassembly syntax: {@code pushl     0x12345678}
     */
    // Template#: 4094, Serial#: 1763
    public void push(int imm32) {
        assemble0063((byte) 0x68, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code push  }<i>m32</i>
     * Example disassembly syntax: {@code push      [L1: 0x12345678]}
     */
    // Template#: 4095, Serial#: 2451
    public void m_push(int m32) {
        assemble0299((byte) 0xFF, (byte) 0x06, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code push  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code push      ebx[esi * 4 + 305419896]}
     */
    // Template#: 4096, Serial#: 2471
    public void push(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xFF, (byte) 0x06, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code push  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code push      L1: 0x12345678[esi * 4]}
     */
    // Template#: 4097, Serial#: 2450
    public void m_push(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xFF, (byte) 0x06, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code push  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code push      [ebx + 305419896]}
     */
    // Template#: 4098, Serial#: 2470
    public void push(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xFF, (byte) 0x06, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code pushw  }<i>imm16</i>
     * Example disassembly syntax: {@code pushw     0x1234}
     */
    // Template#: 4099, Serial#: 1764
    public void push(short imm16) {
        assemble0092((byte) 0x68, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code pusha  }
     * Example disassembly syntax: {@code pusha   }
     */
    // Template#: 4100, Serial#: 229
    public void pusha() {
        assemble0197((byte) 0x60);
    }

    /**
     * Pseudo-external assembler syntax: {@code pusha  }
     * Example disassembly syntax: {@code pusha   }
     */
    // Template#: 4101, Serial#: 230
    public void pushad() {
        assemble0001((byte) 0x60);
    }

    /**
     * Pseudo-external assembler syntax: {@code pushfl  }
     * Example disassembly syntax: {@code pushfl  }
     */
    // Template#: 4102, Serial#: 1915
    public void pushfl() {
        assemble0001((byte) 0x9C);
    }

    /**
     * Pseudo-external assembler syntax: {@code pushfw  }
     * Example disassembly syntax: {@code pushfw  }
     */
    // Template#: 4103, Serial#: 1916
    public void pushfw() {
        assemble0197((byte) 0x9C);
    }

    /**
     * Pseudo-external assembler syntax: {@code pxor  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pxor      xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 4104, Serial#: 5407
    public void pxor(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0xEF, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pxor  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pxor      xmm0, [ebx + 18]}
     */
    // Template#: 4105, Serial#: 5406
    public void pxor(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0xEF, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pxor  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pxor      xmm0, ebx[esi * 4]}
     */
    // Template#: 4106, Serial#: 5403
    public void pxor(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0xEF, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pxor  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pxor      xmm0, [ebx]}
     */
    // Template#: 4107, Serial#: 5402
    public void pxor(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0xEF, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pxor  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pxor      xmm0, xmm0}
     */
    // Template#: 4108, Serial#: 5410
    public void pxor(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0xEF, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pxor  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pxor      xmm0, [L1: 0x12345678]}
     */
    // Template#: 4109, Serial#: 5405
    public void m_pxor(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0xEF, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pxor  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pxor      xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 4110, Serial#: 5409
    public void pxor(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0xEF, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pxor  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pxor      xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 4111, Serial#: 5404
    public void m_pxor(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0xEF, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pxor  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pxor      xmm0, [ebx + 305419896]}
     */
    // Template#: 4112, Serial#: 5408
    public void pxor(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0xEF, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pxor  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pxor      mm0, ebx[esi * 4 + 18]}
     */
    // Template#: 4113, Serial#: 5335
    public void pxor(MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0263((byte) 0xEF, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pxor  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code pxor      mm0, [ebx + 18]}
     */
    // Template#: 4114, Serial#: 5334
    public void pxor(MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0264((byte) 0xEF, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pxor  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pxor      mm0, ebx[esi * 4]}
     */
    // Template#: 4115, Serial#: 5331
    public void pxor(MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0265((byte) 0xEF, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pxor  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pxor      mm0, [ebx]}
     */
    // Template#: 4116, Serial#: 5330
    public void pxor(MMXRegister destination, IA32IndirectRegister32 source) {
        assemble0266((byte) 0xEF, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pxor  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code pxor      mm0, mm0}
     */
    // Template#: 4117, Serial#: 5338
    public void pxor(MMXRegister destination, MMXRegister source) {
        assemble0452((byte) 0xEF, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code pxor  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code pxor      mm0, [L1: 0x12345678]}
     */
    // Template#: 4118, Serial#: 5333
    public void m_pxor(MMXRegister destination, int m32) {
        assemble0268((byte) 0xEF, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code pxor  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pxor      mm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 4119, Serial#: 5337
    public void pxor(MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0269((byte) 0xEF, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pxor  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code pxor      mm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 4120, Serial#: 5332
    public void m_pxor(MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0270((byte) 0xEF, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code pxor  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code pxor      mm0, [ebx + 305419896]}
     */
    // Template#: 4121, Serial#: 5336
    public void pxor(MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0271((byte) 0xEF, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rclb      ebx[esi * 4 + 18], 0x1}
     */
    // Template#: 4122, Serial#: 983
    public void rclb___1(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD0, (byte) 0x02, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcll      ebx[esi * 4 + 18], 0x1}
     */
    // Template#: 4123, Serial#: 1046
    public void rcll___1(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD1, (byte) 0x02, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rclw      ebx[esi * 4 + 18], 0x1}
     */
    // Template#: 4124, Serial#: 1109
    public void rclw___1(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0292((byte) 0xD1, (byte) 0x02, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rclb      ebx[esi * 4 + 18], cl}
     */
    // Template#: 4125, Serial#: 1172
    public void rclb___CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD2, (byte) 0x02, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcll      ebx[esi * 4 + 18], cl}
     */
    // Template#: 4126, Serial#: 1235
    public void rcll___CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD3, (byte) 0x02, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rclw      ebx[esi * 4 + 18], cl}
     */
    // Template#: 4127, Serial#: 1298
    public void rclw___CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0292((byte) 0xD3, (byte) 0x02, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rclb      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 4128, Serial#: 733
    public void rclb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0xC0, (byte) 0x02, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcll      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 4129, Serial#: 796
    public void rcll(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0xC1, (byte) 0x02, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rclw      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 4130, Serial#: 859
    public void rclw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0004((byte) 0xC1, (byte) 0x02, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code rclb      [ebx + 18], 0x1}
     */
    // Template#: 4131, Serial#: 982
    public void rclb___1(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD0, (byte) 0x02, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code rcll      [ebx + 18], 0x1}
     */
    // Template#: 4132, Serial#: 1045
    public void rcll___1(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD1, (byte) 0x02, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code rclw      [ebx + 18], 0x1}
     */
    // Template#: 4133, Serial#: 1108
    public void rclw___1(byte disp8, IA32IndirectRegister32 destination) {
        assemble0293((byte) 0xD1, (byte) 0x02, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code rclb      [ebx + 18], cl}
     */
    // Template#: 4134, Serial#: 1171
    public void rclb___CL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD2, (byte) 0x02, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code rcll      [ebx + 18], cl}
     */
    // Template#: 4135, Serial#: 1234
    public void rcll___CL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD3, (byte) 0x02, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code rclw      [ebx + 18], cl}
     */
    // Template#: 4136, Serial#: 1297
    public void rclw___CL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0293((byte) 0xD3, (byte) 0x02, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rclb      [ebx + 18], 0x12}
     */
    // Template#: 4137, Serial#: 732
    public void rclb(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0xC0, (byte) 0x02, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcll      [ebx + 18], 0x12}
     */
    // Template#: 4138, Serial#: 795
    public void rcll(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0xC1, (byte) 0x02, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rclw      [ebx + 18], 0x12}
     */
    // Template#: 4139, Serial#: 858
    public void rclw(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0011((byte) 0xC1, (byte) 0x02, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rclb      ebx[esi * 4], 0x1}
     */
    // Template#: 4140, Serial#: 959
    public void rclb___1(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD0, (byte) 0x02, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcll      ebx[esi * 4], 0x1}
     */
    // Template#: 4141, Serial#: 1022
    public void rcll___1(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD1, (byte) 0x02, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rclw      ebx[esi * 4], 0x1}
     */
    // Template#: 4142, Serial#: 1085
    public void rclw___1(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0294((byte) 0xD1, (byte) 0x02, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rclb      ebx[esi * 4], cl}
     */
    // Template#: 4143, Serial#: 1148
    public void rclb___CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD2, (byte) 0x02, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcll      ebx[esi * 4], cl}
     */
    // Template#: 4144, Serial#: 1211
    public void rcll___CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD3, (byte) 0x02, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rclw      ebx[esi * 4], cl}
     */
    // Template#: 4145, Serial#: 1274
    public void rclw___CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0294((byte) 0xD3, (byte) 0x02, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rclb      ebx[esi * 4], 0x12}
     */
    // Template#: 4146, Serial#: 709
    public void rclb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0xC0, (byte) 0x02, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcll      ebx[esi * 4], 0x12}
     */
    // Template#: 4147, Serial#: 772
    public void rcll(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0xC1, (byte) 0x02, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rclw      ebx[esi * 4], 0x12}
     */
    // Template#: 4148, Serial#: 835
    public void rclw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0018((byte) 0xC1, (byte) 0x02, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>destination</i>
     * Example disassembly syntax: {@code rclw      ax, 0x1}
     */
    // Template#: 4149, Serial#: 1134
    public void rclw___1(IA32GeneralRegister16 destination) {
        assemble0188((byte) 0xD1, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>destination</i>
     * Example disassembly syntax: {@code rclw      ax, cl}
     */
    // Template#: 4150, Serial#: 1323
    public void rclw___CL(IA32GeneralRegister16 destination) {
        assemble0188((byte) 0xD3, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rclw      ax, 0x12}
     */
    // Template#: 4151, Serial#: 884
    public void rclw(IA32GeneralRegister16 destination, byte imm8) {
        assemble0024((byte) 0xC1, (byte) 0x02, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>destination</i>
     * Example disassembly syntax: {@code rcll      eax, 0x1}
     */
    // Template#: 4152, Serial#: 1071
    public void rcll___1(IA32GeneralRegister32 destination) {
        assemble0189((byte) 0xD1, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>destination</i>
     * Example disassembly syntax: {@code rcll      eax, cl}
     */
    // Template#: 4153, Serial#: 1260
    public void rcll___CL(IA32GeneralRegister32 destination) {
        assemble0189((byte) 0xD3, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcll      eax, 0x12}
     */
    // Template#: 4154, Serial#: 821
    public void rcll(IA32GeneralRegister32 destination, byte imm8) {
        assemble0035((byte) 0xC1, (byte) 0x02, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>destination</i>
     * Example disassembly syntax: {@code rclb      al, 0x1}
     */
    // Template#: 4155, Serial#: 1008
    public void rclb___1(IA32GeneralRegister8 destination) {
        assemble0297((byte) 0xD0, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>destination</i>
     * Example disassembly syntax: {@code rclb      al, cl}
     */
    // Template#: 4156, Serial#: 1197
    public void rclb___CL(IA32GeneralRegister8 destination) {
        assemble0297((byte) 0xD2, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rclb      al, 0x12}
     */
    // Template#: 4157, Serial#: 758
    public void rclb(IA32GeneralRegister8 destination, byte imm8) {
        assemble0046((byte) 0xC0, (byte) 0x02, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>destination</i>
     * Example disassembly syntax: {@code rclb      [ebx], 0x1}
     */
    // Template#: 4158, Serial#: 958
    public void rclb___1(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD0, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>destination</i>
     * Example disassembly syntax: {@code rcll      [ebx], 0x1}
     */
    // Template#: 4159, Serial#: 1021
    public void rcll___1(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD1, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>destination</i>
     * Example disassembly syntax: {@code rclw      [ebx], 0x1}
     */
    // Template#: 4160, Serial#: 1084
    public void rclw___1(IA32IndirectRegister32 destination) {
        assemble0298((byte) 0xD1, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>destination</i>
     * Example disassembly syntax: {@code rclb      [ebx], cl}
     */
    // Template#: 4161, Serial#: 1147
    public void rclb___CL(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD2, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>destination</i>
     * Example disassembly syntax: {@code rcll      [ebx], cl}
     */
    // Template#: 4162, Serial#: 1210
    public void rcll___CL(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD3, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>destination</i>
     * Example disassembly syntax: {@code rclw      [ebx], cl}
     */
    // Template#: 4163, Serial#: 1273
    public void rclw___CL(IA32IndirectRegister32 destination) {
        assemble0298((byte) 0xD3, (byte) 0x02, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rclb      [ebx], 0x12}
     */
    // Template#: 4164, Serial#: 708
    public void rclb(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0xC0, (byte) 0x02, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcll      [ebx], 0x12}
     */
    // Template#: 4165, Serial#: 771
    public void rcll(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0xC1, (byte) 0x02, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rclw      [ebx], 0x12}
     */
    // Template#: 4166, Serial#: 834
    public void rclw(IA32IndirectRegister32 destination, byte imm8) {
        assemble0057((byte) 0xC1, (byte) 0x02, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>m32</i>
     * Example disassembly syntax: {@code rclb      [L1: 0x12345678], 0x1}
     */
    // Template#: 4167, Serial#: 961
    public void m_rclb___1(int m32) {
        assemble0299((byte) 0xD0, (byte) 0x02, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>m32</i>
     * Example disassembly syntax: {@code rcll      [L1: 0x12345678], 0x1}
     */
    // Template#: 4168, Serial#: 1024
    public void m_rcll___1(int m32) {
        assemble0299((byte) 0xD1, (byte) 0x02, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>m32</i>
     * Example disassembly syntax: {@code rclw      [L1: 0x12345678], 0x1}
     */
    // Template#: 4169, Serial#: 1087
    public void m_rclw___1(int m32) {
        assemble0300((byte) 0xD1, (byte) 0x02, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>m32</i>
     * Example disassembly syntax: {@code rclb      [L1: 0x12345678], cl}
     */
    // Template#: 4170, Serial#: 1150
    public void m_rclb___CL(int m32) {
        assemble0299((byte) 0xD2, (byte) 0x02, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>m32</i>
     * Example disassembly syntax: {@code rcll      [L1: 0x12345678], cl}
     */
    // Template#: 4171, Serial#: 1213
    public void m_rcll___CL(int m32) {
        assemble0299((byte) 0xD3, (byte) 0x02, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>m32</i>
     * Example disassembly syntax: {@code rclw      [L1: 0x12345678], cl}
     */
    // Template#: 4172, Serial#: 1276
    public void m_rclw___CL(int m32) {
        assemble0300((byte) 0xD3, (byte) 0x02, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rclb      [L1: 0x12345678], 0x12}
     */
    // Template#: 4173, Serial#: 711
    public void m_rclb(int m32, byte imm8) {
        assemble0064((byte) 0xC0, (byte) 0x02, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcll      [L1: 0x12345678], 0x12}
     */
    // Template#: 4174, Serial#: 774
    public void m_rcll(int m32, byte imm8) {
        assemble0064((byte) 0xC1, (byte) 0x02, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rclw      [L1: 0x12345678], 0x12}
     */
    // Template#: 4175, Serial#: 837
    public void m_rclw(int m32, byte imm8) {
        assemble0065((byte) 0xC1, (byte) 0x02, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rclb      ebx[esi * 4 + 305419896], 0x1}
     */
    // Template#: 4176, Serial#: 997
    public void rclb___1(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD0, (byte) 0x02, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcll      ebx[esi * 4 + 305419896], 0x1}
     */
    // Template#: 4177, Serial#: 1060
    public void rcll___1(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD1, (byte) 0x02, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rclw      ebx[esi * 4 + 305419896], 0x1}
     */
    // Template#: 4178, Serial#: 1123
    public void rclw___1(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0301((byte) 0xD1, (byte) 0x02, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rclb      ebx[esi * 4 + 305419896], cl}
     */
    // Template#: 4179, Serial#: 1186
    public void rclb___CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD2, (byte) 0x02, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcll      ebx[esi * 4 + 305419896], cl}
     */
    // Template#: 4180, Serial#: 1249
    public void rcll___CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD3, (byte) 0x02, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rclw      ebx[esi * 4 + 305419896], cl}
     */
    // Template#: 4181, Serial#: 1312
    public void rclw___CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0301((byte) 0xD3, (byte) 0x02, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rclb      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 4182, Serial#: 747
    public void rclb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0xC0, (byte) 0x02, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcll      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 4183, Serial#: 810
    public void rcll(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0xC1, (byte) 0x02, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rclw      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 4184, Serial#: 873
    public void rclw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0067((byte) 0xC1, (byte) 0x02, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rclb      L1: 0x12345678[esi * 4], 0x1}
     */
    // Template#: 4185, Serial#: 960
    public void m_rclb___1(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD0, (byte) 0x02, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcll      L1: 0x12345678[esi * 4], 0x1}
     */
    // Template#: 4186, Serial#: 1023
    public void m_rcll___1(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD1, (byte) 0x02, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rclw      L1: 0x12345678[esi * 4], 0x1}
     */
    // Template#: 4187, Serial#: 1086
    public void m_rclw___1(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0302((byte) 0xD1, (byte) 0x02, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rclb      L1: 0x12345678[esi * 4], cl}
     */
    // Template#: 4188, Serial#: 1149
    public void m_rclb___CL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD2, (byte) 0x02, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcll      L1: 0x12345678[esi * 4], cl}
     */
    // Template#: 4189, Serial#: 1212
    public void m_rcll___CL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD3, (byte) 0x02, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rclw      L1: 0x12345678[esi * 4], cl}
     */
    // Template#: 4190, Serial#: 1275
    public void m_rclw___CL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0302((byte) 0xD3, (byte) 0x02, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rclb      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 4191, Serial#: 710
    public void m_rclb(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0xC0, (byte) 0x02, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcll      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 4192, Serial#: 773
    public void m_rcll(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0xC1, (byte) 0x02, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rclw      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 4193, Serial#: 836
    public void m_rclw(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0077((byte) 0xC1, (byte) 0x02, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code rclb      [ebx + 305419896], 0x1}
     */
    // Template#: 4194, Serial#: 996
    public void rclb___1(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD0, (byte) 0x02, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code rcll      [ebx + 305419896], 0x1}
     */
    // Template#: 4195, Serial#: 1059
    public void rcll___1(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD1, (byte) 0x02, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code rclw      [ebx + 305419896], 0x1}
     */
    // Template#: 4196, Serial#: 1122
    public void rclw___1(int disp32, IA32IndirectRegister32 destination) {
        assemble0303((byte) 0xD1, (byte) 0x02, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code rclb      [ebx + 305419896], cl}
     */
    // Template#: 4197, Serial#: 1185
    public void rclb___CL(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD2, (byte) 0x02, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code rcll      [ebx + 305419896], cl}
     */
    // Template#: 4198, Serial#: 1248
    public void rcll___CL(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD3, (byte) 0x02, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code rclw      [ebx + 305419896], cl}
     */
    // Template#: 4199, Serial#: 1311
    public void rclw___CL(int disp32, IA32IndirectRegister32 destination) {
        assemble0303((byte) 0xD3, (byte) 0x02, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclb  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rclb      [ebx + 305419896], 0x12}
     */
    // Template#: 4200, Serial#: 746
    public void rclb(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0xC0, (byte) 0x02, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcll  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcll      [ebx + 305419896], 0x12}
     */
    // Template#: 4201, Serial#: 809
    public void rcll(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0xC1, (byte) 0x02, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rclw  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rclw      [ebx + 305419896], 0x12}
     */
    // Template#: 4202, Serial#: 872
    public void rclw(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0084((byte) 0xC1, (byte) 0x02, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcpps  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcpps     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 4203, Serial#: 3005
    public void rcpps(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0102((byte) 0x53, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcpps  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code rcpps     xmm0, [ebx + 18]}
     */
    // Template#: 4204, Serial#: 3004
    public void rcpps(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0103((byte) 0x53, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcpps  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcpps     xmm0, ebx[esi * 4]}
     */
    // Template#: 4205, Serial#: 3001
    public void rcpps(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0104((byte) 0x53, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcpps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code rcpps     xmm0, [ebx]}
     */
    // Template#: 4206, Serial#: 3000
    public void rcpps(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0105((byte) 0x53, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcpps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code rcpps     xmm0, xmm0}
     */
    // Template#: 4207, Serial#: 3008
    public void rcpps(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0106((byte) 0x53, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcpps  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code rcpps     xmm0, [L1: 0x12345678]}
     */
    // Template#: 4208, Serial#: 3003
    public void m_rcpps(IA32XMMRegister destination, int m32) {
        assemble0107((byte) 0x53, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcpps  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcpps     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 4209, Serial#: 3007
    public void rcpps(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0108((byte) 0x53, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcpps  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcpps     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 4210, Serial#: 3002
    public void m_rcpps(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0109((byte) 0x53, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcpps  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code rcpps     xmm0, [ebx + 305419896]}
     */
    // Template#: 4211, Serial#: 3006
    public void rcpps(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0110((byte) 0x53, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcpss  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcpss     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 4212, Serial#: 3123
    public void rcpss(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0120((byte) 0x53, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcpss  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code rcpss     xmm0, [ebx + 18]}
     */
    // Template#: 4213, Serial#: 3122
    public void rcpss(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0121((byte) 0x53, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcpss  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcpss     xmm0, ebx[esi * 4]}
     */
    // Template#: 4214, Serial#: 3119
    public void rcpss(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0122((byte) 0x53, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcpss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code rcpss     xmm0, [ebx]}
     */
    // Template#: 4215, Serial#: 3118
    public void rcpss(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0123((byte) 0x53, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcpss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code rcpss     xmm0, xmm0}
     */
    // Template#: 4216, Serial#: 3126
    public void rcpss(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0124((byte) 0x53, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcpss  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code rcpss     xmm0, [L1: 0x12345678]}
     */
    // Template#: 4217, Serial#: 3121
    public void m_rcpss(IA32XMMRegister destination, int m32) {
        assemble0125((byte) 0x53, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcpss  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcpss     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 4218, Serial#: 3125
    public void rcpss(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0126((byte) 0x53, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcpss  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcpss     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 4219, Serial#: 3120
    public void m_rcpss(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0127((byte) 0x53, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcpss  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code rcpss     xmm0, [ebx + 305419896]}
     */
    // Template#: 4220, Serial#: 3124
    public void rcpss(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0128((byte) 0x53, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcrb      ebx[esi * 4 + 18], 0x1}
     */
    // Template#: 4221, Serial#: 985
    public void rcrb___1(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD0, (byte) 0x03, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcrl      ebx[esi * 4 + 18], 0x1}
     */
    // Template#: 4222, Serial#: 1048
    public void rcrl___1(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD1, (byte) 0x03, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcrw      ebx[esi * 4 + 18], 0x1}
     */
    // Template#: 4223, Serial#: 1111
    public void rcrw___1(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0292((byte) 0xD1, (byte) 0x03, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcrb      ebx[esi * 4 + 18], cl}
     */
    // Template#: 4224, Serial#: 1174
    public void rcrb___CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD2, (byte) 0x03, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcrl      ebx[esi * 4 + 18], cl}
     */
    // Template#: 4225, Serial#: 1237
    public void rcrl___CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD3, (byte) 0x03, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcrw      ebx[esi * 4 + 18], cl}
     */
    // Template#: 4226, Serial#: 1300
    public void rcrw___CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0292((byte) 0xD3, (byte) 0x03, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrb      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 4227, Serial#: 735
    public void rcrb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0xC0, (byte) 0x03, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrl      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 4228, Serial#: 798
    public void rcrl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0xC1, (byte) 0x03, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrw      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 4229, Serial#: 861
    public void rcrw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0004((byte) 0xC1, (byte) 0x03, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code rcrb      [ebx + 18], 0x1}
     */
    // Template#: 4230, Serial#: 984
    public void rcrb___1(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD0, (byte) 0x03, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code rcrl      [ebx + 18], 0x1}
     */
    // Template#: 4231, Serial#: 1047
    public void rcrl___1(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD1, (byte) 0x03, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code rcrw      [ebx + 18], 0x1}
     */
    // Template#: 4232, Serial#: 1110
    public void rcrw___1(byte disp8, IA32IndirectRegister32 destination) {
        assemble0293((byte) 0xD1, (byte) 0x03, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code rcrb      [ebx + 18], cl}
     */
    // Template#: 4233, Serial#: 1173
    public void rcrb___CL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD2, (byte) 0x03, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code rcrl      [ebx + 18], cl}
     */
    // Template#: 4234, Serial#: 1236
    public void rcrl___CL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD3, (byte) 0x03, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code rcrw      [ebx + 18], cl}
     */
    // Template#: 4235, Serial#: 1299
    public void rcrw___CL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0293((byte) 0xD3, (byte) 0x03, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrb      [ebx + 18], 0x12}
     */
    // Template#: 4236, Serial#: 734
    public void rcrb(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0xC0, (byte) 0x03, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrl      [ebx + 18], 0x12}
     */
    // Template#: 4237, Serial#: 797
    public void rcrl(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0xC1, (byte) 0x03, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrw      [ebx + 18], 0x12}
     */
    // Template#: 4238, Serial#: 860
    public void rcrw(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0011((byte) 0xC1, (byte) 0x03, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcrb      ebx[esi * 4], 0x1}
     */
    // Template#: 4239, Serial#: 963
    public void rcrb___1(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD0, (byte) 0x03, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcrl      ebx[esi * 4], 0x1}
     */
    // Template#: 4240, Serial#: 1026
    public void rcrl___1(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD1, (byte) 0x03, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcrw      ebx[esi * 4], 0x1}
     */
    // Template#: 4241, Serial#: 1089
    public void rcrw___1(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0294((byte) 0xD1, (byte) 0x03, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcrb      ebx[esi * 4], cl}
     */
    // Template#: 4242, Serial#: 1152
    public void rcrb___CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD2, (byte) 0x03, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcrl      ebx[esi * 4], cl}
     */
    // Template#: 4243, Serial#: 1215
    public void rcrl___CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD3, (byte) 0x03, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcrw      ebx[esi * 4], cl}
     */
    // Template#: 4244, Serial#: 1278
    public void rcrw___CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0294((byte) 0xD3, (byte) 0x03, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrb      ebx[esi * 4], 0x12}
     */
    // Template#: 4245, Serial#: 713
    public void rcrb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0xC0, (byte) 0x03, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrl      ebx[esi * 4], 0x12}
     */
    // Template#: 4246, Serial#: 776
    public void rcrl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0xC1, (byte) 0x03, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrw      ebx[esi * 4], 0x12}
     */
    // Template#: 4247, Serial#: 839
    public void rcrw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0018((byte) 0xC1, (byte) 0x03, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>destination</i>
     * Example disassembly syntax: {@code rcrw      ax, 0x1}
     */
    // Template#: 4248, Serial#: 1135
    public void rcrw___1(IA32GeneralRegister16 destination) {
        assemble0188((byte) 0xD1, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>destination</i>
     * Example disassembly syntax: {@code rcrw      ax, cl}
     */
    // Template#: 4249, Serial#: 1324
    public void rcrw___CL(IA32GeneralRegister16 destination) {
        assemble0188((byte) 0xD3, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrw      ax, 0x12}
     */
    // Template#: 4250, Serial#: 885
    public void rcrw(IA32GeneralRegister16 destination, byte imm8) {
        assemble0024((byte) 0xC1, (byte) 0x03, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>destination</i>
     * Example disassembly syntax: {@code rcrl      eax, 0x1}
     */
    // Template#: 4251, Serial#: 1072
    public void rcrl___1(IA32GeneralRegister32 destination) {
        assemble0189((byte) 0xD1, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>destination</i>
     * Example disassembly syntax: {@code rcrl      eax, cl}
     */
    // Template#: 4252, Serial#: 1261
    public void rcrl___CL(IA32GeneralRegister32 destination) {
        assemble0189((byte) 0xD3, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrl      eax, 0x12}
     */
    // Template#: 4253, Serial#: 822
    public void rcrl(IA32GeneralRegister32 destination, byte imm8) {
        assemble0035((byte) 0xC1, (byte) 0x03, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>destination</i>
     * Example disassembly syntax: {@code rcrb      al, 0x1}
     */
    // Template#: 4254, Serial#: 1009
    public void rcrb___1(IA32GeneralRegister8 destination) {
        assemble0297((byte) 0xD0, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>destination</i>
     * Example disassembly syntax: {@code rcrb      al, cl}
     */
    // Template#: 4255, Serial#: 1198
    public void rcrb___CL(IA32GeneralRegister8 destination) {
        assemble0297((byte) 0xD2, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrb      al, 0x12}
     */
    // Template#: 4256, Serial#: 759
    public void rcrb(IA32GeneralRegister8 destination, byte imm8) {
        assemble0046((byte) 0xC0, (byte) 0x03, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>destination</i>
     * Example disassembly syntax: {@code rcrb      [ebx], 0x1}
     */
    // Template#: 4257, Serial#: 962
    public void rcrb___1(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD0, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>destination</i>
     * Example disassembly syntax: {@code rcrl      [ebx], 0x1}
     */
    // Template#: 4258, Serial#: 1025
    public void rcrl___1(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD1, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>destination</i>
     * Example disassembly syntax: {@code rcrw      [ebx], 0x1}
     */
    // Template#: 4259, Serial#: 1088
    public void rcrw___1(IA32IndirectRegister32 destination) {
        assemble0298((byte) 0xD1, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>destination</i>
     * Example disassembly syntax: {@code rcrb      [ebx], cl}
     */
    // Template#: 4260, Serial#: 1151
    public void rcrb___CL(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD2, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>destination</i>
     * Example disassembly syntax: {@code rcrl      [ebx], cl}
     */
    // Template#: 4261, Serial#: 1214
    public void rcrl___CL(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD3, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>destination</i>
     * Example disassembly syntax: {@code rcrw      [ebx], cl}
     */
    // Template#: 4262, Serial#: 1277
    public void rcrw___CL(IA32IndirectRegister32 destination) {
        assemble0298((byte) 0xD3, (byte) 0x03, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrb      [ebx], 0x12}
     */
    // Template#: 4263, Serial#: 712
    public void rcrb(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0xC0, (byte) 0x03, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrl      [ebx], 0x12}
     */
    // Template#: 4264, Serial#: 775
    public void rcrl(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0xC1, (byte) 0x03, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrw      [ebx], 0x12}
     */
    // Template#: 4265, Serial#: 838
    public void rcrw(IA32IndirectRegister32 destination, byte imm8) {
        assemble0057((byte) 0xC1, (byte) 0x03, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>m32</i>
     * Example disassembly syntax: {@code rcrb      [L1: 0x12345678], 0x1}
     */
    // Template#: 4266, Serial#: 965
    public void m_rcrb___1(int m32) {
        assemble0299((byte) 0xD0, (byte) 0x03, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>m32</i>
     * Example disassembly syntax: {@code rcrl      [L1: 0x12345678], 0x1}
     */
    // Template#: 4267, Serial#: 1028
    public void m_rcrl___1(int m32) {
        assemble0299((byte) 0xD1, (byte) 0x03, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>m32</i>
     * Example disassembly syntax: {@code rcrw      [L1: 0x12345678], 0x1}
     */
    // Template#: 4268, Serial#: 1091
    public void m_rcrw___1(int m32) {
        assemble0300((byte) 0xD1, (byte) 0x03, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>m32</i>
     * Example disassembly syntax: {@code rcrb      [L1: 0x12345678], cl}
     */
    // Template#: 4269, Serial#: 1154
    public void m_rcrb___CL(int m32) {
        assemble0299((byte) 0xD2, (byte) 0x03, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>m32</i>
     * Example disassembly syntax: {@code rcrl      [L1: 0x12345678], cl}
     */
    // Template#: 4270, Serial#: 1217
    public void m_rcrl___CL(int m32) {
        assemble0299((byte) 0xD3, (byte) 0x03, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>m32</i>
     * Example disassembly syntax: {@code rcrw      [L1: 0x12345678], cl}
     */
    // Template#: 4271, Serial#: 1280
    public void m_rcrw___CL(int m32) {
        assemble0300((byte) 0xD3, (byte) 0x03, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrb      [L1: 0x12345678], 0x12}
     */
    // Template#: 4272, Serial#: 715
    public void m_rcrb(int m32, byte imm8) {
        assemble0064((byte) 0xC0, (byte) 0x03, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrl      [L1: 0x12345678], 0x12}
     */
    // Template#: 4273, Serial#: 778
    public void m_rcrl(int m32, byte imm8) {
        assemble0064((byte) 0xC1, (byte) 0x03, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrw      [L1: 0x12345678], 0x12}
     */
    // Template#: 4274, Serial#: 841
    public void m_rcrw(int m32, byte imm8) {
        assemble0065((byte) 0xC1, (byte) 0x03, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcrb      ebx[esi * 4 + 305419896], 0x1}
     */
    // Template#: 4275, Serial#: 999
    public void rcrb___1(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD0, (byte) 0x03, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcrl      ebx[esi * 4 + 305419896], 0x1}
     */
    // Template#: 4276, Serial#: 1062
    public void rcrl___1(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD1, (byte) 0x03, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcrw      ebx[esi * 4 + 305419896], 0x1}
     */
    // Template#: 4277, Serial#: 1125
    public void rcrw___1(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0301((byte) 0xD1, (byte) 0x03, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcrb      ebx[esi * 4 + 305419896], cl}
     */
    // Template#: 4278, Serial#: 1188
    public void rcrb___CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD2, (byte) 0x03, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcrl      ebx[esi * 4 + 305419896], cl}
     */
    // Template#: 4279, Serial#: 1251
    public void rcrl___CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD3, (byte) 0x03, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcrw      ebx[esi * 4 + 305419896], cl}
     */
    // Template#: 4280, Serial#: 1314
    public void rcrw___CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0301((byte) 0xD3, (byte) 0x03, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrb      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 4281, Serial#: 749
    public void rcrb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0xC0, (byte) 0x03, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrl      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 4282, Serial#: 812
    public void rcrl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0xC1, (byte) 0x03, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrw      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 4283, Serial#: 875
    public void rcrw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0067((byte) 0xC1, (byte) 0x03, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcrb      L1: 0x12345678[esi * 4], 0x1}
     */
    // Template#: 4284, Serial#: 964
    public void m_rcrb___1(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD0, (byte) 0x03, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcrl      L1: 0x12345678[esi * 4], 0x1}
     */
    // Template#: 4285, Serial#: 1027
    public void m_rcrl___1(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD1, (byte) 0x03, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcrw      L1: 0x12345678[esi * 4], 0x1}
     */
    // Template#: 4286, Serial#: 1090
    public void m_rcrw___1(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0302((byte) 0xD1, (byte) 0x03, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcrb      L1: 0x12345678[esi * 4], cl}
     */
    // Template#: 4287, Serial#: 1153
    public void m_rcrb___CL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD2, (byte) 0x03, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcrl      L1: 0x12345678[esi * 4], cl}
     */
    // Template#: 4288, Serial#: 1216
    public void m_rcrl___CL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD3, (byte) 0x03, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rcrw      L1: 0x12345678[esi * 4], cl}
     */
    // Template#: 4289, Serial#: 1279
    public void m_rcrw___CL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0302((byte) 0xD3, (byte) 0x03, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrb      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 4290, Serial#: 714
    public void m_rcrb(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0xC0, (byte) 0x03, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrl      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 4291, Serial#: 777
    public void m_rcrl(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0xC1, (byte) 0x03, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrw      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 4292, Serial#: 840
    public void m_rcrw(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0077((byte) 0xC1, (byte) 0x03, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code rcrb      [ebx + 305419896], 0x1}
     */
    // Template#: 4293, Serial#: 998
    public void rcrb___1(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD0, (byte) 0x03, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code rcrl      [ebx + 305419896], 0x1}
     */
    // Template#: 4294, Serial#: 1061
    public void rcrl___1(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD1, (byte) 0x03, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code rcrw      [ebx + 305419896], 0x1}
     */
    // Template#: 4295, Serial#: 1124
    public void rcrw___1(int disp32, IA32IndirectRegister32 destination) {
        assemble0303((byte) 0xD1, (byte) 0x03, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code rcrb      [ebx + 305419896], cl}
     */
    // Template#: 4296, Serial#: 1187
    public void rcrb___CL(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD2, (byte) 0x03, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code rcrl      [ebx + 305419896], cl}
     */
    // Template#: 4297, Serial#: 1250
    public void rcrl___CL(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD3, (byte) 0x03, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code rcrw      [ebx + 305419896], cl}
     */
    // Template#: 4298, Serial#: 1313
    public void rcrw___CL(int disp32, IA32IndirectRegister32 destination) {
        assemble0303((byte) 0xD3, (byte) 0x03, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrb  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrb      [ebx + 305419896], 0x12}
     */
    // Template#: 4299, Serial#: 748
    public void rcrb(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0xC0, (byte) 0x03, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrl  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrl      [ebx + 305419896], 0x12}
     */
    // Template#: 4300, Serial#: 811
    public void rcrl(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0xC1, (byte) 0x03, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rcrw  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rcrw      [ebx + 305419896], 0x12}
     */
    // Template#: 4301, Serial#: 874
    public void rcrw(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0084((byte) 0xC1, (byte) 0x03, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rdmsr  }
     * Example disassembly syntax: {@code rdmsr   }
     */
    // Template#: 4302, Serial#: 2835
    public void rdmsr() {
        assemble0198((byte) 0x32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rdpmc  }
     * Example disassembly syntax: {@code rdpmc   }
     */
    // Template#: 4303, Serial#: 2836
    public void rdpmc() {
        assemble0198((byte) 0x33);
    }

    /**
     * Pseudo-external assembler syntax: {@code rdtsc  }
     * Example disassembly syntax: {@code rdtsc   }
     */
    // Template#: 4304, Serial#: 2834
    public void rdtsc() {
        assemble0198((byte) 0x31);
    }

    /**
     * Pseudo-external assembler syntax: {@code repe  }
     * Example disassembly syntax: {@code repe    }
     */
    // Template#: 4305, Serial#: 1344
    public void repe() {
        assemble0001((byte) 0xF3);
    }

    /**
     * Pseudo-external assembler syntax: {@code repne  }
     * Example disassembly syntax: {@code repne   }
     */
    // Template#: 4306, Serial#: 1343
    public void repne() {
        assemble0001((byte) 0xF2);
    }

    /**
     * Pseudo-external assembler syntax: {@code ret  }
     * Example disassembly syntax: {@code ret     }
     */
    // Template#: 4307, Serial#: 890
    public void ret() {
        assemble0001((byte) 0xC3);
    }

    /**
     * Pseudo-external assembler syntax: {@code ret  }<i>imm16</i>
     * Example disassembly syntax: {@code ret       0x1234}
     */
    // Template#: 4308, Serial#: 889
    public void ret(short imm16) {
        assemble0496((byte) 0xC2, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code retf  }
     * Example disassembly syntax: {@code retf    }
     */
    // Template#: 4309, Serial#: 1938
    public void retf() {
        assemble0001((byte) 0xCB);
    }

    /**
     * Pseudo-external assembler syntax: {@code retf  }<i>imm16</i>
     * Example disassembly syntax: {@code retf      0x1234}
     */
    // Template#: 4310, Serial#: 1937
    public void retf(short imm16) {
        assemble0496((byte) 0xCA, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rolb      ebx[esi * 4 + 18], 0x1}
     */
    // Template#: 4311, Serial#: 979
    public void rolb___1(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD0, (byte) 0x00, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code roll      ebx[esi * 4 + 18], 0x1}
     */
    // Template#: 4312, Serial#: 1042
    public void roll___1(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD1, (byte) 0x00, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rolw      ebx[esi * 4 + 18], 0x1}
     */
    // Template#: 4313, Serial#: 1105
    public void rolw___1(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0292((byte) 0xD1, (byte) 0x00, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rolb      ebx[esi * 4 + 18], cl}
     */
    // Template#: 4314, Serial#: 1168
    public void rolb___CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD2, (byte) 0x00, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code roll      ebx[esi * 4 + 18], cl}
     */
    // Template#: 4315, Serial#: 1231
    public void roll___CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD3, (byte) 0x00, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rolw      ebx[esi * 4 + 18], cl}
     */
    // Template#: 4316, Serial#: 1294
    public void rolw___CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0292((byte) 0xD3, (byte) 0x00, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rolb      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 4317, Serial#: 729
    public void rolb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0xC0, (byte) 0x00, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code roll      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 4318, Serial#: 792
    public void roll(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0xC1, (byte) 0x00, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rolw      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 4319, Serial#: 855
    public void rolw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0004((byte) 0xC1, (byte) 0x00, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code rolb      [ebx + 18], 0x1}
     */
    // Template#: 4320, Serial#: 978
    public void rolb___1(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD0, (byte) 0x00, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code roll      [ebx + 18], 0x1}
     */
    // Template#: 4321, Serial#: 1041
    public void roll___1(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD1, (byte) 0x00, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code rolw      [ebx + 18], 0x1}
     */
    // Template#: 4322, Serial#: 1104
    public void rolw___1(byte disp8, IA32IndirectRegister32 destination) {
        assemble0293((byte) 0xD1, (byte) 0x00, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code rolb      [ebx + 18], cl}
     */
    // Template#: 4323, Serial#: 1167
    public void rolb___CL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD2, (byte) 0x00, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code roll      [ebx + 18], cl}
     */
    // Template#: 4324, Serial#: 1230
    public void roll___CL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD3, (byte) 0x00, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code rolw      [ebx + 18], cl}
     */
    // Template#: 4325, Serial#: 1293
    public void rolw___CL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0293((byte) 0xD3, (byte) 0x00, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rolb      [ebx + 18], 0x12}
     */
    // Template#: 4326, Serial#: 728
    public void rolb(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0xC0, (byte) 0x00, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code roll      [ebx + 18], 0x12}
     */
    // Template#: 4327, Serial#: 791
    public void roll(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0xC1, (byte) 0x00, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rolw      [ebx + 18], 0x12}
     */
    // Template#: 4328, Serial#: 854
    public void rolw(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0011((byte) 0xC1, (byte) 0x00, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rolb      ebx[esi * 4], 0x1}
     */
    // Template#: 4329, Serial#: 951
    public void rolb___1(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD0, (byte) 0x00, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code roll      ebx[esi * 4], 0x1}
     */
    // Template#: 4330, Serial#: 1014
    public void roll___1(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD1, (byte) 0x00, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rolw      ebx[esi * 4], 0x1}
     */
    // Template#: 4331, Serial#: 1077
    public void rolw___1(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0294((byte) 0xD1, (byte) 0x00, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rolb      ebx[esi * 4], cl}
     */
    // Template#: 4332, Serial#: 1140
    public void rolb___CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD2, (byte) 0x00, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code roll      ebx[esi * 4], cl}
     */
    // Template#: 4333, Serial#: 1203
    public void roll___CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD3, (byte) 0x00, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rolw      ebx[esi * 4], cl}
     */
    // Template#: 4334, Serial#: 1266
    public void rolw___CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0294((byte) 0xD3, (byte) 0x00, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rolb      ebx[esi * 4], 0x12}
     */
    // Template#: 4335, Serial#: 701
    public void rolb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0xC0, (byte) 0x00, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code roll      ebx[esi * 4], 0x12}
     */
    // Template#: 4336, Serial#: 764
    public void roll(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0xC1, (byte) 0x00, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rolw      ebx[esi * 4], 0x12}
     */
    // Template#: 4337, Serial#: 827
    public void rolw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0018((byte) 0xC1, (byte) 0x00, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>destination</i>
     * Example disassembly syntax: {@code rolw      ax, 0x1}
     */
    // Template#: 4338, Serial#: 1132
    public void rolw___1(IA32GeneralRegister16 destination) {
        assemble0188((byte) 0xD1, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>destination</i>
     * Example disassembly syntax: {@code rolw      ax, cl}
     */
    // Template#: 4339, Serial#: 1321
    public void rolw___CL(IA32GeneralRegister16 destination) {
        assemble0188((byte) 0xD3, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rolw      ax, 0x12}
     */
    // Template#: 4340, Serial#: 882
    public void rolw(IA32GeneralRegister16 destination, byte imm8) {
        assemble0024((byte) 0xC1, (byte) 0x00, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>destination</i>
     * Example disassembly syntax: {@code roll      eax, 0x1}
     */
    // Template#: 4341, Serial#: 1069
    public void roll___1(IA32GeneralRegister32 destination) {
        assemble0189((byte) 0xD1, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>destination</i>
     * Example disassembly syntax: {@code roll      eax, cl}
     */
    // Template#: 4342, Serial#: 1258
    public void roll___CL(IA32GeneralRegister32 destination) {
        assemble0189((byte) 0xD3, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code roll      eax, 0x12}
     */
    // Template#: 4343, Serial#: 819
    public void roll(IA32GeneralRegister32 destination, byte imm8) {
        assemble0035((byte) 0xC1, (byte) 0x00, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>destination</i>
     * Example disassembly syntax: {@code rolb      al, 0x1}
     */
    // Template#: 4344, Serial#: 1006
    public void rolb___1(IA32GeneralRegister8 destination) {
        assemble0297((byte) 0xD0, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>destination</i>
     * Example disassembly syntax: {@code rolb      al, cl}
     */
    // Template#: 4345, Serial#: 1195
    public void rolb___CL(IA32GeneralRegister8 destination) {
        assemble0297((byte) 0xD2, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rolb      al, 0x12}
     */
    // Template#: 4346, Serial#: 756
    public void rolb(IA32GeneralRegister8 destination, byte imm8) {
        assemble0046((byte) 0xC0, (byte) 0x00, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>destination</i>
     * Example disassembly syntax: {@code rolb      [ebx], 0x1}
     */
    // Template#: 4347, Serial#: 950
    public void rolb___1(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD0, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>destination</i>
     * Example disassembly syntax: {@code roll      [ebx], 0x1}
     */
    // Template#: 4348, Serial#: 1013
    public void roll___1(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD1, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>destination</i>
     * Example disassembly syntax: {@code rolw      [ebx], 0x1}
     */
    // Template#: 4349, Serial#: 1076
    public void rolw___1(IA32IndirectRegister32 destination) {
        assemble0298((byte) 0xD1, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>destination</i>
     * Example disassembly syntax: {@code rolb      [ebx], cl}
     */
    // Template#: 4350, Serial#: 1139
    public void rolb___CL(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD2, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>destination</i>
     * Example disassembly syntax: {@code roll      [ebx], cl}
     */
    // Template#: 4351, Serial#: 1202
    public void roll___CL(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD3, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>destination</i>
     * Example disassembly syntax: {@code rolw      [ebx], cl}
     */
    // Template#: 4352, Serial#: 1265
    public void rolw___CL(IA32IndirectRegister32 destination) {
        assemble0298((byte) 0xD3, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rolb      [ebx], 0x12}
     */
    // Template#: 4353, Serial#: 700
    public void rolb(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0xC0, (byte) 0x00, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code roll      [ebx], 0x12}
     */
    // Template#: 4354, Serial#: 763
    public void roll(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0xC1, (byte) 0x00, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rolw      [ebx], 0x12}
     */
    // Template#: 4355, Serial#: 826
    public void rolw(IA32IndirectRegister32 destination, byte imm8) {
        assemble0057((byte) 0xC1, (byte) 0x00, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>m32</i>
     * Example disassembly syntax: {@code rolb      [L1: 0x12345678], 0x1}
     */
    // Template#: 4356, Serial#: 953
    public void m_rolb___1(int m32) {
        assemble0299((byte) 0xD0, (byte) 0x00, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>m32</i>
     * Example disassembly syntax: {@code roll      [L1: 0x12345678], 0x1}
     */
    // Template#: 4357, Serial#: 1016
    public void m_roll___1(int m32) {
        assemble0299((byte) 0xD1, (byte) 0x00, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>m32</i>
     * Example disassembly syntax: {@code rolw      [L1: 0x12345678], 0x1}
     */
    // Template#: 4358, Serial#: 1079
    public void m_rolw___1(int m32) {
        assemble0300((byte) 0xD1, (byte) 0x00, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>m32</i>
     * Example disassembly syntax: {@code rolb      [L1: 0x12345678], cl}
     */
    // Template#: 4359, Serial#: 1142
    public void m_rolb___CL(int m32) {
        assemble0299((byte) 0xD2, (byte) 0x00, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>m32</i>
     * Example disassembly syntax: {@code roll      [L1: 0x12345678], cl}
     */
    // Template#: 4360, Serial#: 1205
    public void m_roll___CL(int m32) {
        assemble0299((byte) 0xD3, (byte) 0x00, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>m32</i>
     * Example disassembly syntax: {@code rolw      [L1: 0x12345678], cl}
     */
    // Template#: 4361, Serial#: 1268
    public void m_rolw___CL(int m32) {
        assemble0300((byte) 0xD3, (byte) 0x00, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rolb      [L1: 0x12345678], 0x12}
     */
    // Template#: 4362, Serial#: 703
    public void m_rolb(int m32, byte imm8) {
        assemble0064((byte) 0xC0, (byte) 0x00, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code roll      [L1: 0x12345678], 0x12}
     */
    // Template#: 4363, Serial#: 766
    public void m_roll(int m32, byte imm8) {
        assemble0064((byte) 0xC1, (byte) 0x00, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rolw      [L1: 0x12345678], 0x12}
     */
    // Template#: 4364, Serial#: 829
    public void m_rolw(int m32, byte imm8) {
        assemble0065((byte) 0xC1, (byte) 0x00, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rolb      ebx[esi * 4 + 305419896], 0x1}
     */
    // Template#: 4365, Serial#: 993
    public void rolb___1(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD0, (byte) 0x00, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code roll      ebx[esi * 4 + 305419896], 0x1}
     */
    // Template#: 4366, Serial#: 1056
    public void roll___1(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD1, (byte) 0x00, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rolw      ebx[esi * 4 + 305419896], 0x1}
     */
    // Template#: 4367, Serial#: 1119
    public void rolw___1(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0301((byte) 0xD1, (byte) 0x00, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rolb      ebx[esi * 4 + 305419896], cl}
     */
    // Template#: 4368, Serial#: 1182
    public void rolb___CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD2, (byte) 0x00, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code roll      ebx[esi * 4 + 305419896], cl}
     */
    // Template#: 4369, Serial#: 1245
    public void roll___CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD3, (byte) 0x00, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rolw      ebx[esi * 4 + 305419896], cl}
     */
    // Template#: 4370, Serial#: 1308
    public void rolw___CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0301((byte) 0xD3, (byte) 0x00, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rolb      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 4371, Serial#: 743
    public void rolb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0xC0, (byte) 0x00, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code roll      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 4372, Serial#: 806
    public void roll(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0xC1, (byte) 0x00, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rolw      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 4373, Serial#: 869
    public void rolw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0067((byte) 0xC1, (byte) 0x00, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rolb      L1: 0x12345678[esi * 4], 0x1}
     */
    // Template#: 4374, Serial#: 952
    public void m_rolb___1(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD0, (byte) 0x00, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code roll      L1: 0x12345678[esi * 4], 0x1}
     */
    // Template#: 4375, Serial#: 1015
    public void m_roll___1(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD1, (byte) 0x00, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rolw      L1: 0x12345678[esi * 4], 0x1}
     */
    // Template#: 4376, Serial#: 1078
    public void m_rolw___1(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0302((byte) 0xD1, (byte) 0x00, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rolb      L1: 0x12345678[esi * 4], cl}
     */
    // Template#: 4377, Serial#: 1141
    public void m_rolb___CL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD2, (byte) 0x00, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code roll      L1: 0x12345678[esi * 4], cl}
     */
    // Template#: 4378, Serial#: 1204
    public void m_roll___CL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD3, (byte) 0x00, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rolw      L1: 0x12345678[esi * 4], cl}
     */
    // Template#: 4379, Serial#: 1267
    public void m_rolw___CL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0302((byte) 0xD3, (byte) 0x00, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rolb      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 4380, Serial#: 702
    public void m_rolb(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0xC0, (byte) 0x00, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code roll      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 4381, Serial#: 765
    public void m_roll(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0xC1, (byte) 0x00, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rolw      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 4382, Serial#: 828
    public void m_rolw(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0077((byte) 0xC1, (byte) 0x00, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code rolb      [ebx + 305419896], 0x1}
     */
    // Template#: 4383, Serial#: 992
    public void rolb___1(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD0, (byte) 0x00, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code roll      [ebx + 305419896], 0x1}
     */
    // Template#: 4384, Serial#: 1055
    public void roll___1(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD1, (byte) 0x00, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code rolw      [ebx + 305419896], 0x1}
     */
    // Template#: 4385, Serial#: 1118
    public void rolw___1(int disp32, IA32IndirectRegister32 destination) {
        assemble0303((byte) 0xD1, (byte) 0x00, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code rolb      [ebx + 305419896], cl}
     */
    // Template#: 4386, Serial#: 1181
    public void rolb___CL(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD2, (byte) 0x00, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code roll      [ebx + 305419896], cl}
     */
    // Template#: 4387, Serial#: 1244
    public void roll___CL(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD3, (byte) 0x00, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code rolw      [ebx + 305419896], cl}
     */
    // Template#: 4388, Serial#: 1307
    public void rolw___CL(int disp32, IA32IndirectRegister32 destination) {
        assemble0303((byte) 0xD3, (byte) 0x00, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolb  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rolb      [ebx + 305419896], 0x12}
     */
    // Template#: 4389, Serial#: 742
    public void rolb(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0xC0, (byte) 0x00, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code roll  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code roll      [ebx + 305419896], 0x12}
     */
    // Template#: 4390, Serial#: 805
    public void roll(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0xC1, (byte) 0x00, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rolw  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rolw      [ebx + 305419896], 0x12}
     */
    // Template#: 4391, Serial#: 868
    public void rolw(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0084((byte) 0xC1, (byte) 0x00, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rorb      ebx[esi * 4 + 18], 0x1}
     */
    // Template#: 4392, Serial#: 981
    public void rorb___1(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD0, (byte) 0x01, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rorl      ebx[esi * 4 + 18], 0x1}
     */
    // Template#: 4393, Serial#: 1044
    public void rorl___1(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD1, (byte) 0x01, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rorw      ebx[esi * 4 + 18], 0x1}
     */
    // Template#: 4394, Serial#: 1107
    public void rorw___1(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0292((byte) 0xD1, (byte) 0x01, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rorb      ebx[esi * 4 + 18], cl}
     */
    // Template#: 4395, Serial#: 1170
    public void rorb___CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD2, (byte) 0x01, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rorl      ebx[esi * 4 + 18], cl}
     */
    // Template#: 4396, Serial#: 1233
    public void rorl___CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD3, (byte) 0x01, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rorw      ebx[esi * 4 + 18], cl}
     */
    // Template#: 4397, Serial#: 1296
    public void rorw___CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0292((byte) 0xD3, (byte) 0x01, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorb      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 4398, Serial#: 731
    public void rorb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0xC0, (byte) 0x01, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorl      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 4399, Serial#: 794
    public void rorl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0xC1, (byte) 0x01, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorw      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 4400, Serial#: 857
    public void rorw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0004((byte) 0xC1, (byte) 0x01, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code rorb      [ebx + 18], 0x1}
     */
    // Template#: 4401, Serial#: 980
    public void rorb___1(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD0, (byte) 0x01, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code rorl      [ebx + 18], 0x1}
     */
    // Template#: 4402, Serial#: 1043
    public void rorl___1(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD1, (byte) 0x01, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code rorw      [ebx + 18], 0x1}
     */
    // Template#: 4403, Serial#: 1106
    public void rorw___1(byte disp8, IA32IndirectRegister32 destination) {
        assemble0293((byte) 0xD1, (byte) 0x01, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code rorb      [ebx + 18], cl}
     */
    // Template#: 4404, Serial#: 1169
    public void rorb___CL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD2, (byte) 0x01, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code rorl      [ebx + 18], cl}
     */
    // Template#: 4405, Serial#: 1232
    public void rorl___CL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD3, (byte) 0x01, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code rorw      [ebx + 18], cl}
     */
    // Template#: 4406, Serial#: 1295
    public void rorw___CL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0293((byte) 0xD3, (byte) 0x01, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorb      [ebx + 18], 0x12}
     */
    // Template#: 4407, Serial#: 730
    public void rorb(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0xC0, (byte) 0x01, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorl      [ebx + 18], 0x12}
     */
    // Template#: 4408, Serial#: 793
    public void rorl(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0xC1, (byte) 0x01, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorw      [ebx + 18], 0x12}
     */
    // Template#: 4409, Serial#: 856
    public void rorw(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0011((byte) 0xC1, (byte) 0x01, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rorb      ebx[esi * 4], 0x1}
     */
    // Template#: 4410, Serial#: 955
    public void rorb___1(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD0, (byte) 0x01, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rorl      ebx[esi * 4], 0x1}
     */
    // Template#: 4411, Serial#: 1018
    public void rorl___1(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD1, (byte) 0x01, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rorw      ebx[esi * 4], 0x1}
     */
    // Template#: 4412, Serial#: 1081
    public void rorw___1(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0294((byte) 0xD1, (byte) 0x01, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rorb      ebx[esi * 4], cl}
     */
    // Template#: 4413, Serial#: 1144
    public void rorb___CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD2, (byte) 0x01, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rorl      ebx[esi * 4], cl}
     */
    // Template#: 4414, Serial#: 1207
    public void rorl___CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD3, (byte) 0x01, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rorw      ebx[esi * 4], cl}
     */
    // Template#: 4415, Serial#: 1270
    public void rorw___CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0294((byte) 0xD3, (byte) 0x01, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorb      ebx[esi * 4], 0x12}
     */
    // Template#: 4416, Serial#: 705
    public void rorb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0xC0, (byte) 0x01, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorl      ebx[esi * 4], 0x12}
     */
    // Template#: 4417, Serial#: 768
    public void rorl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0xC1, (byte) 0x01, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorw      ebx[esi * 4], 0x12}
     */
    // Template#: 4418, Serial#: 831
    public void rorw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0018((byte) 0xC1, (byte) 0x01, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>destination</i>
     * Example disassembly syntax: {@code rorw      ax, 0x1}
     */
    // Template#: 4419, Serial#: 1133
    public void rorw___1(IA32GeneralRegister16 destination) {
        assemble0188((byte) 0xD1, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>destination</i>
     * Example disassembly syntax: {@code rorw      ax, cl}
     */
    // Template#: 4420, Serial#: 1322
    public void rorw___CL(IA32GeneralRegister16 destination) {
        assemble0188((byte) 0xD3, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorw      ax, 0x12}
     */
    // Template#: 4421, Serial#: 883
    public void rorw(IA32GeneralRegister16 destination, byte imm8) {
        assemble0024((byte) 0xC1, (byte) 0x01, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>destination</i>
     * Example disassembly syntax: {@code rorl      eax, 0x1}
     */
    // Template#: 4422, Serial#: 1070
    public void rorl___1(IA32GeneralRegister32 destination) {
        assemble0189((byte) 0xD1, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>destination</i>
     * Example disassembly syntax: {@code rorl      eax, cl}
     */
    // Template#: 4423, Serial#: 1259
    public void rorl___CL(IA32GeneralRegister32 destination) {
        assemble0189((byte) 0xD3, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorl      eax, 0x12}
     */
    // Template#: 4424, Serial#: 820
    public void rorl(IA32GeneralRegister32 destination, byte imm8) {
        assemble0035((byte) 0xC1, (byte) 0x01, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>destination</i>
     * Example disassembly syntax: {@code rorb      al, 0x1}
     */
    // Template#: 4425, Serial#: 1007
    public void rorb___1(IA32GeneralRegister8 destination) {
        assemble0297((byte) 0xD0, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>destination</i>
     * Example disassembly syntax: {@code rorb      al, cl}
     */
    // Template#: 4426, Serial#: 1196
    public void rorb___CL(IA32GeneralRegister8 destination) {
        assemble0297((byte) 0xD2, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorb      al, 0x12}
     */
    // Template#: 4427, Serial#: 757
    public void rorb(IA32GeneralRegister8 destination, byte imm8) {
        assemble0046((byte) 0xC0, (byte) 0x01, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>destination</i>
     * Example disassembly syntax: {@code rorb      [ebx], 0x1}
     */
    // Template#: 4428, Serial#: 954
    public void rorb___1(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD0, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>destination</i>
     * Example disassembly syntax: {@code rorl      [ebx], 0x1}
     */
    // Template#: 4429, Serial#: 1017
    public void rorl___1(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD1, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>destination</i>
     * Example disassembly syntax: {@code rorw      [ebx], 0x1}
     */
    // Template#: 4430, Serial#: 1080
    public void rorw___1(IA32IndirectRegister32 destination) {
        assemble0298((byte) 0xD1, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>destination</i>
     * Example disassembly syntax: {@code rorb      [ebx], cl}
     */
    // Template#: 4431, Serial#: 1143
    public void rorb___CL(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD2, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>destination</i>
     * Example disassembly syntax: {@code rorl      [ebx], cl}
     */
    // Template#: 4432, Serial#: 1206
    public void rorl___CL(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD3, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>destination</i>
     * Example disassembly syntax: {@code rorw      [ebx], cl}
     */
    // Template#: 4433, Serial#: 1269
    public void rorw___CL(IA32IndirectRegister32 destination) {
        assemble0298((byte) 0xD3, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorb      [ebx], 0x12}
     */
    // Template#: 4434, Serial#: 704
    public void rorb(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0xC0, (byte) 0x01, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorl      [ebx], 0x12}
     */
    // Template#: 4435, Serial#: 767
    public void rorl(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0xC1, (byte) 0x01, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorw      [ebx], 0x12}
     */
    // Template#: 4436, Serial#: 830
    public void rorw(IA32IndirectRegister32 destination, byte imm8) {
        assemble0057((byte) 0xC1, (byte) 0x01, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>m32</i>
     * Example disassembly syntax: {@code rorb      [L1: 0x12345678], 0x1}
     */
    // Template#: 4437, Serial#: 957
    public void m_rorb___1(int m32) {
        assemble0299((byte) 0xD0, (byte) 0x01, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>m32</i>
     * Example disassembly syntax: {@code rorl      [L1: 0x12345678], 0x1}
     */
    // Template#: 4438, Serial#: 1020
    public void m_rorl___1(int m32) {
        assemble0299((byte) 0xD1, (byte) 0x01, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>m32</i>
     * Example disassembly syntax: {@code rorw      [L1: 0x12345678], 0x1}
     */
    // Template#: 4439, Serial#: 1083
    public void m_rorw___1(int m32) {
        assemble0300((byte) 0xD1, (byte) 0x01, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>m32</i>
     * Example disassembly syntax: {@code rorb      [L1: 0x12345678], cl}
     */
    // Template#: 4440, Serial#: 1146
    public void m_rorb___CL(int m32) {
        assemble0299((byte) 0xD2, (byte) 0x01, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>m32</i>
     * Example disassembly syntax: {@code rorl      [L1: 0x12345678], cl}
     */
    // Template#: 4441, Serial#: 1209
    public void m_rorl___CL(int m32) {
        assemble0299((byte) 0xD3, (byte) 0x01, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>m32</i>
     * Example disassembly syntax: {@code rorw      [L1: 0x12345678], cl}
     */
    // Template#: 4442, Serial#: 1272
    public void m_rorw___CL(int m32) {
        assemble0300((byte) 0xD3, (byte) 0x01, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorb      [L1: 0x12345678], 0x12}
     */
    // Template#: 4443, Serial#: 707
    public void m_rorb(int m32, byte imm8) {
        assemble0064((byte) 0xC0, (byte) 0x01, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorl      [L1: 0x12345678], 0x12}
     */
    // Template#: 4444, Serial#: 770
    public void m_rorl(int m32, byte imm8) {
        assemble0064((byte) 0xC1, (byte) 0x01, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorw      [L1: 0x12345678], 0x12}
     */
    // Template#: 4445, Serial#: 833
    public void m_rorw(int m32, byte imm8) {
        assemble0065((byte) 0xC1, (byte) 0x01, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rorb      ebx[esi * 4 + 305419896], 0x1}
     */
    // Template#: 4446, Serial#: 995
    public void rorb___1(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD0, (byte) 0x01, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rorl      ebx[esi * 4 + 305419896], 0x1}
     */
    // Template#: 4447, Serial#: 1058
    public void rorl___1(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD1, (byte) 0x01, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rorw      ebx[esi * 4 + 305419896], 0x1}
     */
    // Template#: 4448, Serial#: 1121
    public void rorw___1(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0301((byte) 0xD1, (byte) 0x01, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rorb      ebx[esi * 4 + 305419896], cl}
     */
    // Template#: 4449, Serial#: 1184
    public void rorb___CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD2, (byte) 0x01, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rorl      ebx[esi * 4 + 305419896], cl}
     */
    // Template#: 4450, Serial#: 1247
    public void rorl___CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD3, (byte) 0x01, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rorw      ebx[esi * 4 + 305419896], cl}
     */
    // Template#: 4451, Serial#: 1310
    public void rorw___CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0301((byte) 0xD3, (byte) 0x01, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorb      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 4452, Serial#: 745
    public void rorb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0xC0, (byte) 0x01, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorl      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 4453, Serial#: 808
    public void rorl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0xC1, (byte) 0x01, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorw      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 4454, Serial#: 871
    public void rorw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0067((byte) 0xC1, (byte) 0x01, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rorb      L1: 0x12345678[esi * 4], 0x1}
     */
    // Template#: 4455, Serial#: 956
    public void m_rorb___1(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD0, (byte) 0x01, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rorl      L1: 0x12345678[esi * 4], 0x1}
     */
    // Template#: 4456, Serial#: 1019
    public void m_rorl___1(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD1, (byte) 0x01, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rorw      L1: 0x12345678[esi * 4], 0x1}
     */
    // Template#: 4457, Serial#: 1082
    public void m_rorw___1(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0302((byte) 0xD1, (byte) 0x01, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rorb      L1: 0x12345678[esi * 4], cl}
     */
    // Template#: 4458, Serial#: 1145
    public void m_rorb___CL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD2, (byte) 0x01, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rorl      L1: 0x12345678[esi * 4], cl}
     */
    // Template#: 4459, Serial#: 1208
    public void m_rorl___CL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD3, (byte) 0x01, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rorw      L1: 0x12345678[esi * 4], cl}
     */
    // Template#: 4460, Serial#: 1271
    public void m_rorw___CL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0302((byte) 0xD3, (byte) 0x01, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorb      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 4461, Serial#: 706
    public void m_rorb(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0xC0, (byte) 0x01, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorl      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 4462, Serial#: 769
    public void m_rorl(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0xC1, (byte) 0x01, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorw      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 4463, Serial#: 832
    public void m_rorw(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0077((byte) 0xC1, (byte) 0x01, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code rorb      [ebx + 305419896], 0x1}
     */
    // Template#: 4464, Serial#: 994
    public void rorb___1(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD0, (byte) 0x01, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code rorl      [ebx + 305419896], 0x1}
     */
    // Template#: 4465, Serial#: 1057
    public void rorl___1(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD1, (byte) 0x01, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code rorw      [ebx + 305419896], 0x1}
     */
    // Template#: 4466, Serial#: 1120
    public void rorw___1(int disp32, IA32IndirectRegister32 destination) {
        assemble0303((byte) 0xD1, (byte) 0x01, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code rorb      [ebx + 305419896], cl}
     */
    // Template#: 4467, Serial#: 1183
    public void rorb___CL(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD2, (byte) 0x01, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code rorl      [ebx + 305419896], cl}
     */
    // Template#: 4468, Serial#: 1246
    public void rorl___CL(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD3, (byte) 0x01, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code rorw      [ebx + 305419896], cl}
     */
    // Template#: 4469, Serial#: 1309
    public void rorw___CL(int disp32, IA32IndirectRegister32 destination) {
        assemble0303((byte) 0xD3, (byte) 0x01, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorb  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorb      [ebx + 305419896], 0x12}
     */
    // Template#: 4470, Serial#: 744
    public void rorb(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0xC0, (byte) 0x01, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorl  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorl      [ebx + 305419896], 0x12}
     */
    // Template#: 4471, Serial#: 807
    public void rorl(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0xC1, (byte) 0x01, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rorw  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code rorw      [ebx + 305419896], 0x12}
     */
    // Template#: 4472, Serial#: 870
    public void rorw(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0084((byte) 0xC1, (byte) 0x01, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code rsm  }
     * Example disassembly syntax: {@code rsm     }
     */
    // Template#: 4473, Serial#: 4952
    public void rsm() {
        assemble0198((byte) 0xAA);
    }

    /**
     * Pseudo-external assembler syntax: {@code rsqrtps  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rsqrtps   xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 4474, Serial#: 2996
    public void rsqrtps(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0102((byte) 0x52, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rsqrtps  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code rsqrtps   xmm0, [ebx + 18]}
     */
    // Template#: 4475, Serial#: 2995
    public void rsqrtps(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0103((byte) 0x52, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code rsqrtps  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rsqrtps   xmm0, ebx[esi * 4]}
     */
    // Template#: 4476, Serial#: 2992
    public void rsqrtps(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0104((byte) 0x52, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rsqrtps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code rsqrtps   xmm0, [ebx]}
     */
    // Template#: 4477, Serial#: 2991
    public void rsqrtps(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0105((byte) 0x52, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code rsqrtps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code rsqrtps   xmm0, xmm0}
     */
    // Template#: 4478, Serial#: 2999
    public void rsqrtps(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0106((byte) 0x52, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code rsqrtps  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code rsqrtps   xmm0, [L1: 0x12345678]}
     */
    // Template#: 4479, Serial#: 2994
    public void m_rsqrtps(IA32XMMRegister destination, int m32) {
        assemble0107((byte) 0x52, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rsqrtps  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rsqrtps   xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 4480, Serial#: 2998
    public void rsqrtps(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0108((byte) 0x52, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rsqrtps  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rsqrtps   xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 4481, Serial#: 2993
    public void m_rsqrtps(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0109((byte) 0x52, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rsqrtps  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code rsqrtps   xmm0, [ebx + 305419896]}
     */
    // Template#: 4482, Serial#: 2997
    public void rsqrtps(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0110((byte) 0x52, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code rsqrtss  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rsqrtss   xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 4483, Serial#: 3114
    public void rsqrtss(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0120((byte) 0x52, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rsqrtss  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code rsqrtss   xmm0, [ebx + 18]}
     */
    // Template#: 4484, Serial#: 3113
    public void rsqrtss(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0121((byte) 0x52, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code rsqrtss  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rsqrtss   xmm0, ebx[esi * 4]}
     */
    // Template#: 4485, Serial#: 3110
    public void rsqrtss(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0122((byte) 0x52, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rsqrtss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code rsqrtss   xmm0, [ebx]}
     */
    // Template#: 4486, Serial#: 3109
    public void rsqrtss(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0123((byte) 0x52, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code rsqrtss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code rsqrtss   xmm0, xmm0}
     */
    // Template#: 4487, Serial#: 3117
    public void rsqrtss(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0124((byte) 0x52, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code rsqrtss  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code rsqrtss   xmm0, [L1: 0x12345678]}
     */
    // Template#: 4488, Serial#: 3112
    public void m_rsqrtss(IA32XMMRegister destination, int m32) {
        assemble0125((byte) 0x52, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code rsqrtss  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rsqrtss   xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 4489, Serial#: 3116
    public void rsqrtss(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0126((byte) 0x52, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rsqrtss  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code rsqrtss   xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 4490, Serial#: 3111
    public void m_rsqrtss(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0127((byte) 0x52, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code rsqrtss  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code rsqrtss   xmm0, [ebx + 305419896]}
     */
    // Template#: 4491, Serial#: 3115
    public void rsqrtss(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0128((byte) 0x52, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sahf  }
     * Example disassembly syntax: {@code sahf    }
     */
    // Template#: 4492, Serial#: 1919
    public void sahf() {
        assemble0001((byte) 0x9E);
    }

    /**
     * Pseudo-external assembler syntax: {@code salc  }
     * Example disassembly syntax: {@code salc    }
     */
    // Template#: 4493, Serial#: 1330
    public void salc() {
        assemble0001((byte) 0xD6);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sarb      ebx[esi * 4 + 18], 0x1}
     */
    // Template#: 4494, Serial#: 991
    public void sarb___1(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD0, (byte) 0x07, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sarl      ebx[esi * 4 + 18], 0x1}
     */
    // Template#: 4495, Serial#: 1054
    public void sarl___1(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD1, (byte) 0x07, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sarw      ebx[esi * 4 + 18], 0x1}
     */
    // Template#: 4496, Serial#: 1117
    public void sarw___1(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0292((byte) 0xD1, (byte) 0x07, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sarb      ebx[esi * 4 + 18], cl}
     */
    // Template#: 4497, Serial#: 1180
    public void sarb___CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD2, (byte) 0x07, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sarl      ebx[esi * 4 + 18], cl}
     */
    // Template#: 4498, Serial#: 1243
    public void sarl___CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD3, (byte) 0x07, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sarw      ebx[esi * 4 + 18], cl}
     */
    // Template#: 4499, Serial#: 1306
    public void sarw___CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0292((byte) 0xD3, (byte) 0x07, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarb      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 4500, Serial#: 741
    public void sarb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0xC0, (byte) 0x07, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarl      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 4501, Serial#: 804
    public void sarl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0xC1, (byte) 0x07, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarw      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 4502, Serial#: 867
    public void sarw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0004((byte) 0xC1, (byte) 0x07, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code sarb      [ebx + 18], 0x1}
     */
    // Template#: 4503, Serial#: 990
    public void sarb___1(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD0, (byte) 0x07, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code sarl      [ebx + 18], 0x1}
     */
    // Template#: 4504, Serial#: 1053
    public void sarl___1(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD1, (byte) 0x07, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code sarw      [ebx + 18], 0x1}
     */
    // Template#: 4505, Serial#: 1116
    public void sarw___1(byte disp8, IA32IndirectRegister32 destination) {
        assemble0293((byte) 0xD1, (byte) 0x07, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code sarb      [ebx + 18], cl}
     */
    // Template#: 4506, Serial#: 1179
    public void sarb___CL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD2, (byte) 0x07, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code sarl      [ebx + 18], cl}
     */
    // Template#: 4507, Serial#: 1242
    public void sarl___CL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD3, (byte) 0x07, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code sarw      [ebx + 18], cl}
     */
    // Template#: 4508, Serial#: 1305
    public void sarw___CL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0293((byte) 0xD3, (byte) 0x07, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarb      [ebx + 18], 0x12}
     */
    // Template#: 4509, Serial#: 740
    public void sarb(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0xC0, (byte) 0x07, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarl      [ebx + 18], 0x12}
     */
    // Template#: 4510, Serial#: 803
    public void sarl(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0xC1, (byte) 0x07, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarw      [ebx + 18], 0x12}
     */
    // Template#: 4511, Serial#: 866
    public void sarw(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0011((byte) 0xC1, (byte) 0x07, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sarb      ebx[esi * 4], 0x1}
     */
    // Template#: 4512, Serial#: 975
    public void sarb___1(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD0, (byte) 0x07, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sarl      ebx[esi * 4], 0x1}
     */
    // Template#: 4513, Serial#: 1038
    public void sarl___1(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD1, (byte) 0x07, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sarw      ebx[esi * 4], 0x1}
     */
    // Template#: 4514, Serial#: 1101
    public void sarw___1(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0294((byte) 0xD1, (byte) 0x07, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sarb      ebx[esi * 4], cl}
     */
    // Template#: 4515, Serial#: 1164
    public void sarb___CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD2, (byte) 0x07, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sarl      ebx[esi * 4], cl}
     */
    // Template#: 4516, Serial#: 1227
    public void sarl___CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD3, (byte) 0x07, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sarw      ebx[esi * 4], cl}
     */
    // Template#: 4517, Serial#: 1290
    public void sarw___CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0294((byte) 0xD3, (byte) 0x07, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarb      ebx[esi * 4], 0x12}
     */
    // Template#: 4518, Serial#: 725
    public void sarb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0xC0, (byte) 0x07, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarl      ebx[esi * 4], 0x12}
     */
    // Template#: 4519, Serial#: 788
    public void sarl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0xC1, (byte) 0x07, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarw      ebx[esi * 4], 0x12}
     */
    // Template#: 4520, Serial#: 851
    public void sarw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0018((byte) 0xC1, (byte) 0x07, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>destination</i>
     * Example disassembly syntax: {@code sarw      ax, 0x1}
     */
    // Template#: 4521, Serial#: 1138
    public void sarw___1(IA32GeneralRegister16 destination) {
        assemble0188((byte) 0xD1, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>destination</i>
     * Example disassembly syntax: {@code sarw      ax, cl}
     */
    // Template#: 4522, Serial#: 1327
    public void sarw___CL(IA32GeneralRegister16 destination) {
        assemble0188((byte) 0xD3, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarw      ax, 0x12}
     */
    // Template#: 4523, Serial#: 888
    public void sarw(IA32GeneralRegister16 destination, byte imm8) {
        assemble0024((byte) 0xC1, (byte) 0x07, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>destination</i>
     * Example disassembly syntax: {@code sarl      eax, 0x1}
     */
    // Template#: 4524, Serial#: 1075
    public void sarl___1(IA32GeneralRegister32 destination) {
        assemble0189((byte) 0xD1, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>destination</i>
     * Example disassembly syntax: {@code sarl      eax, cl}
     */
    // Template#: 4525, Serial#: 1264
    public void sarl___CL(IA32GeneralRegister32 destination) {
        assemble0189((byte) 0xD3, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarl      eax, 0x12}
     */
    // Template#: 4526, Serial#: 825
    public void sarl(IA32GeneralRegister32 destination, byte imm8) {
        assemble0035((byte) 0xC1, (byte) 0x07, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>destination</i>
     * Example disassembly syntax: {@code sarb      al, 0x1}
     */
    // Template#: 4527, Serial#: 1012
    public void sarb___1(IA32GeneralRegister8 destination) {
        assemble0297((byte) 0xD0, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>destination</i>
     * Example disassembly syntax: {@code sarb      al, cl}
     */
    // Template#: 4528, Serial#: 1201
    public void sarb___CL(IA32GeneralRegister8 destination) {
        assemble0297((byte) 0xD2, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarb      al, 0x12}
     */
    // Template#: 4529, Serial#: 762
    public void sarb(IA32GeneralRegister8 destination, byte imm8) {
        assemble0046((byte) 0xC0, (byte) 0x07, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>destination</i>
     * Example disassembly syntax: {@code sarb      [ebx], 0x1}
     */
    // Template#: 4530, Serial#: 974
    public void sarb___1(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD0, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>destination</i>
     * Example disassembly syntax: {@code sarl      [ebx], 0x1}
     */
    // Template#: 4531, Serial#: 1037
    public void sarl___1(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD1, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>destination</i>
     * Example disassembly syntax: {@code sarw      [ebx], 0x1}
     */
    // Template#: 4532, Serial#: 1100
    public void sarw___1(IA32IndirectRegister32 destination) {
        assemble0298((byte) 0xD1, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>destination</i>
     * Example disassembly syntax: {@code sarb      [ebx], cl}
     */
    // Template#: 4533, Serial#: 1163
    public void sarb___CL(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD2, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>destination</i>
     * Example disassembly syntax: {@code sarl      [ebx], cl}
     */
    // Template#: 4534, Serial#: 1226
    public void sarl___CL(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD3, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>destination</i>
     * Example disassembly syntax: {@code sarw      [ebx], cl}
     */
    // Template#: 4535, Serial#: 1289
    public void sarw___CL(IA32IndirectRegister32 destination) {
        assemble0298((byte) 0xD3, (byte) 0x07, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarb      [ebx], 0x12}
     */
    // Template#: 4536, Serial#: 724
    public void sarb(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0xC0, (byte) 0x07, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarl      [ebx], 0x12}
     */
    // Template#: 4537, Serial#: 787
    public void sarl(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0xC1, (byte) 0x07, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarw      [ebx], 0x12}
     */
    // Template#: 4538, Serial#: 850
    public void sarw(IA32IndirectRegister32 destination, byte imm8) {
        assemble0057((byte) 0xC1, (byte) 0x07, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>m32</i>
     * Example disassembly syntax: {@code sarb      [L1: 0x12345678], 0x1}
     */
    // Template#: 4539, Serial#: 977
    public void m_sarb___1(int m32) {
        assemble0299((byte) 0xD0, (byte) 0x07, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>m32</i>
     * Example disassembly syntax: {@code sarl      [L1: 0x12345678], 0x1}
     */
    // Template#: 4540, Serial#: 1040
    public void m_sarl___1(int m32) {
        assemble0299((byte) 0xD1, (byte) 0x07, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>m32</i>
     * Example disassembly syntax: {@code sarw      [L1: 0x12345678], 0x1}
     */
    // Template#: 4541, Serial#: 1103
    public void m_sarw___1(int m32) {
        assemble0300((byte) 0xD1, (byte) 0x07, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>m32</i>
     * Example disassembly syntax: {@code sarb      [L1: 0x12345678], cl}
     */
    // Template#: 4542, Serial#: 1166
    public void m_sarb___CL(int m32) {
        assemble0299((byte) 0xD2, (byte) 0x07, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>m32</i>
     * Example disassembly syntax: {@code sarl      [L1: 0x12345678], cl}
     */
    // Template#: 4543, Serial#: 1229
    public void m_sarl___CL(int m32) {
        assemble0299((byte) 0xD3, (byte) 0x07, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>m32</i>
     * Example disassembly syntax: {@code sarw      [L1: 0x12345678], cl}
     */
    // Template#: 4544, Serial#: 1292
    public void m_sarw___CL(int m32) {
        assemble0300((byte) 0xD3, (byte) 0x07, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarb      [L1: 0x12345678], 0x12}
     */
    // Template#: 4545, Serial#: 727
    public void m_sarb(int m32, byte imm8) {
        assemble0064((byte) 0xC0, (byte) 0x07, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarl      [L1: 0x12345678], 0x12}
     */
    // Template#: 4546, Serial#: 790
    public void m_sarl(int m32, byte imm8) {
        assemble0064((byte) 0xC1, (byte) 0x07, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarw      [L1: 0x12345678], 0x12}
     */
    // Template#: 4547, Serial#: 853
    public void m_sarw(int m32, byte imm8) {
        assemble0065((byte) 0xC1, (byte) 0x07, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sarb      ebx[esi * 4 + 305419896], 0x1}
     */
    // Template#: 4548, Serial#: 1005
    public void sarb___1(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD0, (byte) 0x07, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sarl      ebx[esi * 4 + 305419896], 0x1}
     */
    // Template#: 4549, Serial#: 1068
    public void sarl___1(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD1, (byte) 0x07, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sarw      ebx[esi * 4 + 305419896], 0x1}
     */
    // Template#: 4550, Serial#: 1131
    public void sarw___1(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0301((byte) 0xD1, (byte) 0x07, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sarb      ebx[esi * 4 + 305419896], cl}
     */
    // Template#: 4551, Serial#: 1194
    public void sarb___CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD2, (byte) 0x07, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sarl      ebx[esi * 4 + 305419896], cl}
     */
    // Template#: 4552, Serial#: 1257
    public void sarl___CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD3, (byte) 0x07, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sarw      ebx[esi * 4 + 305419896], cl}
     */
    // Template#: 4553, Serial#: 1320
    public void sarw___CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0301((byte) 0xD3, (byte) 0x07, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarb      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 4554, Serial#: 755
    public void sarb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0xC0, (byte) 0x07, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarl      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 4555, Serial#: 818
    public void sarl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0xC1, (byte) 0x07, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarw      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 4556, Serial#: 881
    public void sarw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0067((byte) 0xC1, (byte) 0x07, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sarb      L1: 0x12345678[esi * 4], 0x1}
     */
    // Template#: 4557, Serial#: 976
    public void m_sarb___1(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD0, (byte) 0x07, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sarl      L1: 0x12345678[esi * 4], 0x1}
     */
    // Template#: 4558, Serial#: 1039
    public void m_sarl___1(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD1, (byte) 0x07, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sarw      L1: 0x12345678[esi * 4], 0x1}
     */
    // Template#: 4559, Serial#: 1102
    public void m_sarw___1(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0302((byte) 0xD1, (byte) 0x07, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sarb      L1: 0x12345678[esi * 4], cl}
     */
    // Template#: 4560, Serial#: 1165
    public void m_sarb___CL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD2, (byte) 0x07, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sarl      L1: 0x12345678[esi * 4], cl}
     */
    // Template#: 4561, Serial#: 1228
    public void m_sarl___CL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD3, (byte) 0x07, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sarw      L1: 0x12345678[esi * 4], cl}
     */
    // Template#: 4562, Serial#: 1291
    public void m_sarw___CL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0302((byte) 0xD3, (byte) 0x07, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarb      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 4563, Serial#: 726
    public void m_sarb(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0xC0, (byte) 0x07, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarl      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 4564, Serial#: 789
    public void m_sarl(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0xC1, (byte) 0x07, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarw      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 4565, Serial#: 852
    public void m_sarw(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0077((byte) 0xC1, (byte) 0x07, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code sarb      [ebx + 305419896], 0x1}
     */
    // Template#: 4566, Serial#: 1004
    public void sarb___1(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD0, (byte) 0x07, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code sarl      [ebx + 305419896], 0x1}
     */
    // Template#: 4567, Serial#: 1067
    public void sarl___1(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD1, (byte) 0x07, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code sarw      [ebx + 305419896], 0x1}
     */
    // Template#: 4568, Serial#: 1130
    public void sarw___1(int disp32, IA32IndirectRegister32 destination) {
        assemble0303((byte) 0xD1, (byte) 0x07, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code sarb      [ebx + 305419896], cl}
     */
    // Template#: 4569, Serial#: 1193
    public void sarb___CL(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD2, (byte) 0x07, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code sarl      [ebx + 305419896], cl}
     */
    // Template#: 4570, Serial#: 1256
    public void sarl___CL(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD3, (byte) 0x07, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code sarw      [ebx + 305419896], cl}
     */
    // Template#: 4571, Serial#: 1319
    public void sarw___CL(int disp32, IA32IndirectRegister32 destination) {
        assemble0303((byte) 0xD3, (byte) 0x07, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarb  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarb      [ebx + 305419896], 0x12}
     */
    // Template#: 4572, Serial#: 754
    public void sarb(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0xC0, (byte) 0x07, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarl  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarl      [ebx + 305419896], 0x12}
     */
    // Template#: 4573, Serial#: 817
    public void sarl(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0xC1, (byte) 0x07, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sarw  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sarw      [ebx + 305419896], 0x12}
     */
    // Template#: 4574, Serial#: 880
    public void sarw(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0084((byte) 0xC1, (byte) 0x07, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>imm8</i>
     * Example disassembly syntax: {@code sbb       al, 0x12}
     */
    // Template#: 4575, Serial#: 1642
    public void sbb_AL(byte imm8) {
        assemble0002((byte) 0x1C, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbb      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 4576, Serial#: 309
    public void sbbb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0x80, (byte) 0x03, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbl      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 4577, Serial#: 525
    public void sbbl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0x83, (byte) 0x03, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbw      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 4578, Serial#: 597
    public void sbbw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0004((byte) 0x83, (byte) 0x03, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       ebx[esi * 4 + 18], ax}
     */
    // Template#: 4579, Serial#: 1614
    public void sbb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0005((byte) 0x19, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       ebx[esi * 4 + 18], eax}
     */
    // Template#: 4580, Serial#: 1605
    public void sbb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0006((byte) 0x19, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       ebx[esi * 4 + 18], al}
     */
    // Template#: 4581, Serial#: 1596
    public void sbb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0007((byte) 0x18, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code sbbl      ebx[esi * 4 + 18], 0x12345678}
     */
    // Template#: 4582, Serial#: 381
    public void sbbl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0008((byte) 0x81, (byte) 0x03, disp8, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code sbbw      ebx[esi * 4 + 18], 0x1234}
     */
    // Template#: 4583, Serial#: 453
    public void sbbw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0009((byte) 0x81, (byte) 0x03, disp8, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbb  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbb      [ebx + 18], 0x12}
     */
    // Template#: 4584, Serial#: 308
    public void sbbb(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0x80, (byte) 0x03, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbl  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbl      [ebx + 18], 0x12}
     */
    // Template#: 4585, Serial#: 524
    public void sbbl(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0x83, (byte) 0x03, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbw  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbw      [ebx + 18], 0x12}
     */
    // Template#: 4586, Serial#: 596
    public void sbbw(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0011((byte) 0x83, (byte) 0x03, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       [ebx + 18], ax}
     */
    // Template#: 4587, Serial#: 1613
    public void sbb(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0012((byte) 0x19, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       [ebx + 18], eax}
     */
    // Template#: 4588, Serial#: 1604
    public void sbb(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0013((byte) 0x19, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       [ebx + 18], al}
     */
    // Template#: 4589, Serial#: 1595
    public void sbb(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0014((byte) 0x18, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbl  }<i>disp8</i>, <i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code sbbl      [ebx + 18], 0x12345678}
     */
    // Template#: 4590, Serial#: 380
    public void sbbl(byte disp8, IA32IndirectRegister32 destination, int imm32) {
        assemble0015((byte) 0x81, (byte) 0x03, disp8, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbw  }<i>disp8</i>, <i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code sbbw      [ebx + 18], 0x1234}
     */
    // Template#: 4591, Serial#: 452
    public void sbbw(byte disp8, IA32IndirectRegister32 destination, short imm16) {
        assemble0016((byte) 0x81, (byte) 0x03, disp8, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbb  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbb      ebx[esi * 4], 0x12}
     */
    // Template#: 4592, Serial#: 283
    public void sbbb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0x80, (byte) 0x03, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbl  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbl      ebx[esi * 4], 0x12}
     */
    // Template#: 4593, Serial#: 499
    public void sbbl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0x83, (byte) 0x03, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbw      ebx[esi * 4], 0x12}
     */
    // Template#: 4594, Serial#: 571
    public void sbbw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0018((byte) 0x83, (byte) 0x03, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       ebx[esi * 4], ax}
     */
    // Template#: 4595, Serial#: 1610
    public void sbb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0019((byte) 0x19, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       ebx[esi * 4], eax}
     */
    // Template#: 4596, Serial#: 1601
    public void sbb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0020((byte) 0x19, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       ebx[esi * 4], al}
     */
    // Template#: 4597, Serial#: 1592
    public void sbb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0021((byte) 0x18, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbl  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code sbbl      ebx[esi * 4], 0x12345678}
     */
    // Template#: 4598, Serial#: 355
    public void sbbl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0022((byte) 0x81, (byte) 0x03, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code sbbw      ebx[esi * 4], 0x1234}
     */
    // Template#: 4599, Serial#: 427
    public void sbbw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0023((byte) 0x81, (byte) 0x03, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbw      ax, 0x12}
     */
    // Template#: 4600, Serial#: 625
    public void sbbw(IA32GeneralRegister16 destination, byte imm8) {
        assemble0024((byte) 0x83, (byte) 0x03, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sbb       ax, ebx[esi * 4 + 18]}
     */
    // Template#: 4601, Serial#: 1639
    public void sbb(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0025((byte) 0x1B, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       ax, [ebx + 18]}
     */
    // Template#: 4602, Serial#: 1638
    public void sbb(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0026((byte) 0x1B, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sbb       ax, ebx[esi * 4]}
     */
    // Template#: 4603, Serial#: 1635
    public void sbb(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0027((byte) 0x1B, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       ax, ax}
     */
    // Template#: 4604, Serial#: 1617
    public void sbb(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0028((byte) 0x19, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       ax, [ebx]}
     */
    // Template#: 4605, Serial#: 1634
    public void sbb(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0029((byte) 0x1B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code sbb       ax, [L1: 0x12345678]}
     */
    // Template#: 4606, Serial#: 1637
    public void m_sbb(IA32GeneralRegister16 destination, int m32) {
        assemble0030((byte) 0x1B, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sbb       ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 4607, Serial#: 1641
    public void sbb(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0031((byte) 0x1B, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sbb       ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 4608, Serial#: 1636
    public void m_sbb(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0032((byte) 0x1B, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       ax, [ebx + 305419896]}
     */
    // Template#: 4609, Serial#: 1640
    public void sbb(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0033((byte) 0x1B, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbw  }<i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code sbbw      ax, 0x1234}
     */
    // Template#: 4610, Serial#: 481
    public void sbbw(IA32GeneralRegister16 destination, short imm16) {
        assemble0034((byte) 0x81, (byte) 0x03, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbl  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbl      eax, 0x12}
     */
    // Template#: 4611, Serial#: 553
    public void sbbl(IA32GeneralRegister32 destination, byte imm8) {
        assemble0035((byte) 0x83, (byte) 0x03, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sbb       eax, ebx[esi * 4 + 18]}
     */
    // Template#: 4612, Serial#: 1631
    public void sbb(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0036((byte) 0x1B, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       eax, [ebx + 18]}
     */
    // Template#: 4613, Serial#: 1630
    public void sbb(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0037((byte) 0x1B, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sbb       eax, ebx[esi * 4]}
     */
    // Template#: 4614, Serial#: 1627
    public void sbb(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0038((byte) 0x1B, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       eax, eax}
     */
    // Template#: 4615, Serial#: 1608
    public void sbb(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0039((byte) 0x19, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       eax, [ebx]}
     */
    // Template#: 4616, Serial#: 1626
    public void sbb(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0040((byte) 0x1B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbl  }<i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code sbbl      eax, 0x12345678}
     */
    // Template#: 4617, Serial#: 409
    public void sbbl(IA32GeneralRegister32 destination, int imm32) {
        assemble0042((byte) 0x81, (byte) 0x03, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code sbb       eax, [L1: 0x12345678]}
     */
    // Template#: 4618, Serial#: 1629
    public void m_sbb(IA32GeneralRegister32 destination, int m32) {
        assemble0041((byte) 0x1B, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sbb       eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 4619, Serial#: 1633
    public void sbb(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0043((byte) 0x1B, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sbb       eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 4620, Serial#: 1628
    public void m_sbb(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0044((byte) 0x1B, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       eax, [ebx + 305419896]}
     */
    // Template#: 4621, Serial#: 1632
    public void sbb(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0045((byte) 0x1B, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbb      al, 0x12}
     */
    // Template#: 4622, Serial#: 337
    public void sbbb(IA32GeneralRegister8 destination, byte imm8) {
        assemble0046((byte) 0x80, (byte) 0x03, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sbb       al, ebx[esi * 4 + 18]}
     */
    // Template#: 4623, Serial#: 1623
    public void sbb(IA32GeneralRegister8 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0047((byte) 0x1A, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       al, [ebx + 18]}
     */
    // Template#: 4624, Serial#: 1622
    public void sbb(IA32GeneralRegister8 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0048((byte) 0x1A, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sbb       al, ebx[esi * 4]}
     */
    // Template#: 4625, Serial#: 1619
    public void sbb(IA32GeneralRegister8 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0049((byte) 0x1A, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       al, al}
     */
    // Template#: 4626, Serial#: 1599
    public void sbb(IA32GeneralRegister8 destination, IA32GeneralRegister8 source) {
        assemble0050((byte) 0x18, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       al, [ebx]}
     */
    // Template#: 4627, Serial#: 1618
    public void sbb(IA32GeneralRegister8 destination, IA32IndirectRegister32 source) {
        assemble0051((byte) 0x1A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code sbb       al, [L1: 0x12345678]}
     */
    // Template#: 4628, Serial#: 1621
    public void m_sbb(IA32GeneralRegister8 destination, int m32) {
        assemble0052((byte) 0x1A, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sbb       al, ebx[esi * 4 + 305419896]}
     */
    // Template#: 4629, Serial#: 1625
    public void sbb(IA32GeneralRegister8 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0053((byte) 0x1A, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sbb       al, L1: 0x12345678[esi * 4]}
     */
    // Template#: 4630, Serial#: 1620
    public void m_sbb(IA32GeneralRegister8 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0054((byte) 0x1A, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       al, [ebx + 305419896]}
     */
    // Template#: 4631, Serial#: 1624
    public void sbb(IA32GeneralRegister8 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0055((byte) 0x1A, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbb      [ebx], 0x12}
     */
    // Template#: 4632, Serial#: 282
    public void sbbb(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0x80, (byte) 0x03, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbl  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbl      [ebx], 0x12}
     */
    // Template#: 4633, Serial#: 498
    public void sbbl(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0x83, (byte) 0x03, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbw      [ebx], 0x12}
     */
    // Template#: 4634, Serial#: 570
    public void sbbw(IA32IndirectRegister32 destination, byte imm8) {
        assemble0057((byte) 0x83, (byte) 0x03, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       [ebx], ax}
     */
    // Template#: 4635, Serial#: 1609
    public void sbb(IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0058((byte) 0x19, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       [ebx], eax}
     */
    // Template#: 4636, Serial#: 1600
    public void sbb(IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0059((byte) 0x19, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       [ebx], al}
     */
    // Template#: 4637, Serial#: 1591
    public void sbb(IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0060((byte) 0x18, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbl  }<i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code sbbl      [ebx], 0x12345678}
     */
    // Template#: 4638, Serial#: 354
    public void sbbl(IA32IndirectRegister32 destination, int imm32) {
        assemble0061((byte) 0x81, (byte) 0x03, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbw  }<i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code sbbw      [ebx], 0x1234}
     */
    // Template#: 4639, Serial#: 426
    public void sbbw(IA32IndirectRegister32 destination, short imm16) {
        assemble0062((byte) 0x81, (byte) 0x03, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbl  }<i>imm32</i>
     * Example disassembly syntax: {@code sbbl      eax, 0x12345678}
     */
    // Template#: 4640, Serial#: 1643
    public void sbb_EAX(int imm32) {
        assemble0063((byte) 0x1D, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbb  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbb      [L1: 0x12345678], 0x12}
     */
    // Template#: 4641, Serial#: 285
    public void m_sbbb(int m32, byte imm8) {
        assemble0064((byte) 0x80, (byte) 0x03, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbl  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbl      [L1: 0x12345678], 0x12}
     */
    // Template#: 4642, Serial#: 501
    public void m_sbbl(int m32, byte imm8) {
        assemble0064((byte) 0x83, (byte) 0x03, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbw  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbw      [L1: 0x12345678], 0x12}
     */
    // Template#: 4643, Serial#: 573
    public void m_sbbw(int m32, byte imm8) {
        assemble0065((byte) 0x83, (byte) 0x03, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbb      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 4644, Serial#: 325
    public void sbbb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0x80, (byte) 0x03, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbl      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 4645, Serial#: 541
    public void sbbl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0x83, (byte) 0x03, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbw      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 4646, Serial#: 613
    public void sbbw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0067((byte) 0x83, (byte) 0x03, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       ebx[esi * 4 + 305419896], ax}
     */
    // Template#: 4647, Serial#: 1616
    public void sbb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0068((byte) 0x19, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       ebx[esi * 4 + 305419896], eax}
     */
    // Template#: 4648, Serial#: 1607
    public void sbb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0069((byte) 0x19, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       ebx[esi * 4 + 305419896], al}
     */
    // Template#: 4649, Serial#: 1598
    public void sbb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0070((byte) 0x18, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code sbbl      ebx[esi * 4 + 305419896], 0x12345678}
     */
    // Template#: 4650, Serial#: 397
    public void sbbl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0071((byte) 0x81, (byte) 0x03, disp32, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code sbbw      ebx[esi * 4 + 305419896], 0x1234}
     */
    // Template#: 4651, Serial#: 469
    public void sbbw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0072((byte) 0x81, (byte) 0x03, disp32, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       [L1: 0x12345678], ax}
     */
    // Template#: 4652, Serial#: 1612
    public void m_sbb(int m32, IA32GeneralRegister16 source) {
        assemble0073((byte) 0x19, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       [L1: 0x12345678], eax}
     */
    // Template#: 4653, Serial#: 1603
    public void m_sbb(int m32, IA32GeneralRegister32 source) {
        assemble0074((byte) 0x19, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       [L1: 0x12345678], al}
     */
    // Template#: 4654, Serial#: 1594
    public void m_sbb(int m32, IA32GeneralRegister8 source) {
        assemble0075((byte) 0x18, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbb  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbb      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 4655, Serial#: 284
    public void m_sbbb(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0x80, (byte) 0x03, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbl  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbl      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 4656, Serial#: 500
    public void m_sbbl(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0x83, (byte) 0x03, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbw      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 4657, Serial#: 572
    public void m_sbbw(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0077((byte) 0x83, (byte) 0x03, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       L1: 0x12345678[esi * 4], ax}
     */
    // Template#: 4658, Serial#: 1611
    public void m_sbb(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0078((byte) 0x19, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       L1: 0x12345678[esi * 4], eax}
     */
    // Template#: 4659, Serial#: 1602
    public void m_sbb(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0079((byte) 0x19, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       L1: 0x12345678[esi * 4], al}
     */
    // Template#: 4660, Serial#: 1593
    public void m_sbb(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0080((byte) 0x18, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbl  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code sbbl      L1: 0x12345678[esi * 4], 0x12345678}
     */
    // Template#: 4661, Serial#: 356
    public void m_sbbl(int m32, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0081((byte) 0x81, (byte) 0x03, m32, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code sbbw      L1: 0x12345678[esi * 4], 0x1234}
     */
    // Template#: 4662, Serial#: 428
    public void m_sbbw(int m32, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0082((byte) 0x81, (byte) 0x03, m32, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbb  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbb      [ebx + 305419896], 0x12}
     */
    // Template#: 4663, Serial#: 324
    public void sbbb(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0x80, (byte) 0x03, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbl  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbl      [ebx + 305419896], 0x12}
     */
    // Template#: 4664, Serial#: 540
    public void sbbl(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0x83, (byte) 0x03, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbw  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code sbbw      [ebx + 305419896], 0x12}
     */
    // Template#: 4665, Serial#: 612
    public void sbbw(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0084((byte) 0x83, (byte) 0x03, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       [ebx + 305419896], ax}
     */
    // Template#: 4666, Serial#: 1615
    public void sbb(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0085((byte) 0x19, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       [ebx + 305419896], eax}
     */
    // Template#: 4667, Serial#: 1606
    public void sbb(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0086((byte) 0x19, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbb  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sbb       [ebx + 305419896], al}
     */
    // Template#: 4668, Serial#: 1597
    public void sbb(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0087((byte) 0x18, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbl  }<i>disp32</i>, <i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code sbbl      [ebx + 305419896], 0x12345678}
     */
    // Template#: 4669, Serial#: 396
    public void sbbl(int disp32, IA32IndirectRegister32 destination, int imm32) {
        assemble0088((byte) 0x81, (byte) 0x03, disp32, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbw  }<i>disp32</i>, <i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code sbbw      [ebx + 305419896], 0x1234}
     */
    // Template#: 4670, Serial#: 468
    public void sbbw(int disp32, IA32IndirectRegister32 destination, short imm16) {
        assemble0089((byte) 0x81, (byte) 0x03, disp32, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbl  }<i>m32</i>, <i>imm32</i>
     * Example disassembly syntax: {@code sbbl      [L1: 0x12345678], 0x12345678}
     */
    // Template#: 4671, Serial#: 357
    public void m_sbbl(int m32, int imm32) {
        assemble0090((byte) 0x81, (byte) 0x03, m32, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbw  }<i>m32</i>, <i>imm16</i>
     * Example disassembly syntax: {@code sbbw      [L1: 0x12345678], 0x1234}
     */
    // Template#: 4672, Serial#: 429
    public void m_sbbw(int m32, short imm16) {
        assemble0091((byte) 0x81, (byte) 0x03, m32, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code sbbw  }<i>imm16</i>
     * Example disassembly syntax: {@code sbbw      ax, 0x1234}
     */
    // Template#: 4673, Serial#: 1644
    public void sbb_AX(short imm16) {
        assemble0092((byte) 0x1D, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code scasb  }
     * Example disassembly syntax: {@code scasb   }
     */
    // Template#: 4674, Serial#: 1930
    public void scasb() {
        assemble0001((byte) 0xAE);
    }

    /**
     * Pseudo-external assembler syntax: {@code scasl  }
     * Example disassembly syntax: {@code scasl   }
     */
    // Template#: 4675, Serial#: 1931
    public void scasl() {
        assemble0001((byte) 0xAF);
    }

    /**
     * Pseudo-external assembler syntax: {@code scasw  }
     * Example disassembly syntax: {@code scasw   }
     */
    // Template#: 4676, Serial#: 1932
    public void scasw() {
        assemble0197((byte) 0xAF);
    }

    /**
     * Pseudo-external assembler syntax: {@code seg_cs  }
     * Example disassembly syntax: {@code seg_cs  }
     */
    // Template#: 4677, Serial#: 1701
    public void seg_cs() {
        assemble0001((byte) 0x2E);
    }

    /**
     * Pseudo-external assembler syntax: {@code seg_ds  }
     * Example disassembly syntax: {@code seg_ds  }
     */
    // Template#: 4678, Serial#: 1757
    public void seg_ds() {
        assemble0001((byte) 0x3E);
    }

    /**
     * Pseudo-external assembler syntax: {@code seg_es  }
     * Example disassembly syntax: {@code seg_es  }
     */
    // Template#: 4679, Serial#: 167
    public void seg_es() {
        assemble0001((byte) 0x26);
    }

    /**
     * Pseudo-external assembler syntax: {@code seg_fs  }
     * Example disassembly syntax: {@code seg_fs  }
     */
    // Template#: 4680, Serial#: 258
    public void seg_fs() {
        assemble0001((byte) 0x64);
    }

    /**
     * Pseudo-external assembler syntax: {@code seg_gs  }
     * Example disassembly syntax: {@code seg_gs  }
     */
    // Template#: 4681, Serial#: 259
    public void seg_gs() {
        assemble0001((byte) 0x65);
    }

    /**
     * Pseudo-external assembler syntax: {@code seg_ss  }
     * Example disassembly syntax: {@code seg_ss  }
     */
    // Template#: 4682, Serial#: 223
    public void seg_ss() {
        assemble0001((byte) 0x36);
    }

    /**
     * Pseudo-external assembler syntax: {@code setb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setb      ebx[esi * 4 + 18]}
     */
    // Template#: 4683, Serial#: 3360
    public void setb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0497((byte) 0x92, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setb  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code setb      [ebx + 18]}
     */
    // Template#: 4684, Serial#: 3359
    public void setb(byte disp8, IA32IndirectRegister32 destination) {
        assemble0498((byte) 0x92, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setb  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setb      ebx[esi * 4]}
     */
    // Template#: 4685, Serial#: 3356
    public void setb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0499((byte) 0x92, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setb  }<i>destination</i>
     * Example disassembly syntax: {@code setb      al}
     */
    // Template#: 4686, Serial#: 3363
    public void setb(IA32GeneralRegister8 destination) {
        assemble0500((byte) 0x92, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setb  }<i>destination</i>
     * Example disassembly syntax: {@code setb      [ebx]}
     */
    // Template#: 4687, Serial#: 3355
    public void setb(IA32IndirectRegister32 destination) {
        assemble0501((byte) 0x92, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setb  }<i>m32</i>
     * Example disassembly syntax: {@code setb      [L1: 0x12345678]}
     */
    // Template#: 4688, Serial#: 3358
    public void m_setb(int m32) {
        assemble0502((byte) 0x92, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code setb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setb      ebx[esi * 4 + 305419896]}
     */
    // Template#: 4689, Serial#: 3362
    public void setb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0503((byte) 0x92, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setb  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setb      L1: 0x12345678[esi * 4]}
     */
    // Template#: 4690, Serial#: 3357
    public void m_setb(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0504((byte) 0x92, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setb  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code setb      [ebx + 305419896]}
     */
    // Template#: 4691, Serial#: 3361
    public void setb(int disp32, IA32IndirectRegister32 destination) {
        assemble0505((byte) 0x92, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setbe  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setbe     ebx[esi * 4 + 18]}
     */
    // Template#: 4692, Serial#: 3396
    public void setbe(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0497((byte) 0x96, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setbe  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code setbe     [ebx + 18]}
     */
    // Template#: 4693, Serial#: 3395
    public void setbe(byte disp8, IA32IndirectRegister32 destination) {
        assemble0498((byte) 0x96, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setbe  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setbe     ebx[esi * 4]}
     */
    // Template#: 4694, Serial#: 3392
    public void setbe(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0499((byte) 0x96, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setbe  }<i>destination</i>
     * Example disassembly syntax: {@code setbe     al}
     */
    // Template#: 4695, Serial#: 3399
    public void setbe(IA32GeneralRegister8 destination) {
        assemble0500((byte) 0x96, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setbe  }<i>destination</i>
     * Example disassembly syntax: {@code setbe     [ebx]}
     */
    // Template#: 4696, Serial#: 3391
    public void setbe(IA32IndirectRegister32 destination) {
        assemble0501((byte) 0x96, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setbe  }<i>m32</i>
     * Example disassembly syntax: {@code setbe     [L1: 0x12345678]}
     */
    // Template#: 4697, Serial#: 3394
    public void m_setbe(int m32) {
        assemble0502((byte) 0x96, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code setbe  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setbe     ebx[esi * 4 + 305419896]}
     */
    // Template#: 4698, Serial#: 3398
    public void setbe(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0503((byte) 0x96, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setbe  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setbe     L1: 0x12345678[esi * 4]}
     */
    // Template#: 4699, Serial#: 3393
    public void m_setbe(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0504((byte) 0x96, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setbe  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code setbe     [ebx + 305419896]}
     */
    // Template#: 4700, Serial#: 3397
    public void setbe(int disp32, IA32IndirectRegister32 destination) {
        assemble0505((byte) 0x96, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setl      ebx[esi * 4 + 18]}
     */
    // Template#: 4701, Serial#: 4919
    public void setl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0497((byte) 0x9C, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code setl      [ebx + 18]}
     */
    // Template#: 4702, Serial#: 4918
    public void setl(byte disp8, IA32IndirectRegister32 destination) {
        assemble0498((byte) 0x9C, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setl      ebx[esi * 4]}
     */
    // Template#: 4703, Serial#: 4915
    public void setl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0499((byte) 0x9C, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setl  }<i>destination</i>
     * Example disassembly syntax: {@code setl      al}
     */
    // Template#: 4704, Serial#: 4922
    public void setl(IA32GeneralRegister8 destination) {
        assemble0500((byte) 0x9C, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setl  }<i>destination</i>
     * Example disassembly syntax: {@code setl      [ebx]}
     */
    // Template#: 4705, Serial#: 4914
    public void setl(IA32IndirectRegister32 destination) {
        assemble0501((byte) 0x9C, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setl  }<i>m32</i>
     * Example disassembly syntax: {@code setl      [L1: 0x12345678]}
     */
    // Template#: 4706, Serial#: 4917
    public void m_setl(int m32) {
        assemble0502((byte) 0x9C, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code setl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setl      ebx[esi * 4 + 305419896]}
     */
    // Template#: 4707, Serial#: 4921
    public void setl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0503((byte) 0x9C, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setl      L1: 0x12345678[esi * 4]}
     */
    // Template#: 4708, Serial#: 4916
    public void m_setl(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0504((byte) 0x9C, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code setl      [ebx + 305419896]}
     */
    // Template#: 4709, Serial#: 4920
    public void setl(int disp32, IA32IndirectRegister32 destination) {
        assemble0505((byte) 0x9C, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setle  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setle     ebx[esi * 4 + 18]}
     */
    // Template#: 4710, Serial#: 4937
    public void setle(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0497((byte) 0x9E, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setle  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code setle     [ebx + 18]}
     */
    // Template#: 4711, Serial#: 4936
    public void setle(byte disp8, IA32IndirectRegister32 destination) {
        assemble0498((byte) 0x9E, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setle  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setle     ebx[esi * 4]}
     */
    // Template#: 4712, Serial#: 4933
    public void setle(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0499((byte) 0x9E, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setle  }<i>destination</i>
     * Example disassembly syntax: {@code setle     al}
     */
    // Template#: 4713, Serial#: 4940
    public void setle(IA32GeneralRegister8 destination) {
        assemble0500((byte) 0x9E, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setle  }<i>destination</i>
     * Example disassembly syntax: {@code setle     [ebx]}
     */
    // Template#: 4714, Serial#: 4932
    public void setle(IA32IndirectRegister32 destination) {
        assemble0501((byte) 0x9E, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setle  }<i>m32</i>
     * Example disassembly syntax: {@code setle     [L1: 0x12345678]}
     */
    // Template#: 4715, Serial#: 4935
    public void m_setle(int m32) {
        assemble0502((byte) 0x9E, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code setle  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setle     ebx[esi * 4 + 305419896]}
     */
    // Template#: 4716, Serial#: 4939
    public void setle(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0503((byte) 0x9E, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setle  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setle     L1: 0x12345678[esi * 4]}
     */
    // Template#: 4717, Serial#: 4934
    public void m_setle(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0504((byte) 0x9E, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setle  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code setle     [ebx + 305419896]}
     */
    // Template#: 4718, Serial#: 4938
    public void setle(int disp32, IA32IndirectRegister32 destination) {
        assemble0505((byte) 0x9E, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setnb     ebx[esi * 4 + 18]}
     */
    // Template#: 4719, Serial#: 3369
    public void setnb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0497((byte) 0x93, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnb  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code setnb     [ebx + 18]}
     */
    // Template#: 4720, Serial#: 3368
    public void setnb(byte disp8, IA32IndirectRegister32 destination) {
        assemble0498((byte) 0x93, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnb  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setnb     ebx[esi * 4]}
     */
    // Template#: 4721, Serial#: 3365
    public void setnb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0499((byte) 0x93, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnb  }<i>destination</i>
     * Example disassembly syntax: {@code setnb     al}
     */
    // Template#: 4722, Serial#: 3372
    public void setnb(IA32GeneralRegister8 destination) {
        assemble0500((byte) 0x93, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnb  }<i>destination</i>
     * Example disassembly syntax: {@code setnb     [ebx]}
     */
    // Template#: 4723, Serial#: 3364
    public void setnb(IA32IndirectRegister32 destination) {
        assemble0501((byte) 0x93, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnb  }<i>m32</i>
     * Example disassembly syntax: {@code setnb     [L1: 0x12345678]}
     */
    // Template#: 4724, Serial#: 3367
    public void m_setnb(int m32) {
        assemble0502((byte) 0x93, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setnb     ebx[esi * 4 + 305419896]}
     */
    // Template#: 4725, Serial#: 3371
    public void setnb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0503((byte) 0x93, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnb  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setnb     L1: 0x12345678[esi * 4]}
     */
    // Template#: 4726, Serial#: 3366
    public void m_setnb(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0504((byte) 0x93, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnb  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code setnb     [ebx + 305419896]}
     */
    // Template#: 4727, Serial#: 3370
    public void setnb(int disp32, IA32IndirectRegister32 destination) {
        assemble0505((byte) 0x93, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnbe  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setnbe    ebx[esi * 4 + 18]}
     */
    // Template#: 4728, Serial#: 3405
    public void setnbe(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0497((byte) 0x97, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnbe  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code setnbe    [ebx + 18]}
     */
    // Template#: 4729, Serial#: 3404
    public void setnbe(byte disp8, IA32IndirectRegister32 destination) {
        assemble0498((byte) 0x97, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnbe  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setnbe    ebx[esi * 4]}
     */
    // Template#: 4730, Serial#: 3401
    public void setnbe(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0499((byte) 0x97, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnbe  }<i>destination</i>
     * Example disassembly syntax: {@code setnbe    al}
     */
    // Template#: 4731, Serial#: 3408
    public void setnbe(IA32GeneralRegister8 destination) {
        assemble0500((byte) 0x97, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnbe  }<i>destination</i>
     * Example disassembly syntax: {@code setnbe    [ebx]}
     */
    // Template#: 4732, Serial#: 3400
    public void setnbe(IA32IndirectRegister32 destination) {
        assemble0501((byte) 0x97, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnbe  }<i>m32</i>
     * Example disassembly syntax: {@code setnbe    [L1: 0x12345678]}
     */
    // Template#: 4733, Serial#: 3403
    public void m_setnbe(int m32) {
        assemble0502((byte) 0x97, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnbe  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setnbe    ebx[esi * 4 + 305419896]}
     */
    // Template#: 4734, Serial#: 3407
    public void setnbe(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0503((byte) 0x97, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnbe  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setnbe    L1: 0x12345678[esi * 4]}
     */
    // Template#: 4735, Serial#: 3402
    public void m_setnbe(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0504((byte) 0x97, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnbe  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code setnbe    [ebx + 305419896]}
     */
    // Template#: 4736, Serial#: 3406
    public void setnbe(int disp32, IA32IndirectRegister32 destination) {
        assemble0505((byte) 0x97, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setnl     ebx[esi * 4 + 18]}
     */
    // Template#: 4737, Serial#: 4928
    public void setnl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0497((byte) 0x9D, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code setnl     [ebx + 18]}
     */
    // Template#: 4738, Serial#: 4927
    public void setnl(byte disp8, IA32IndirectRegister32 destination) {
        assemble0498((byte) 0x9D, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setnl     ebx[esi * 4]}
     */
    // Template#: 4739, Serial#: 4924
    public void setnl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0499((byte) 0x9D, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnl  }<i>destination</i>
     * Example disassembly syntax: {@code setnl     al}
     */
    // Template#: 4740, Serial#: 4931
    public void setnl(IA32GeneralRegister8 destination) {
        assemble0500((byte) 0x9D, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnl  }<i>destination</i>
     * Example disassembly syntax: {@code setnl     [ebx]}
     */
    // Template#: 4741, Serial#: 4923
    public void setnl(IA32IndirectRegister32 destination) {
        assemble0501((byte) 0x9D, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnl  }<i>m32</i>
     * Example disassembly syntax: {@code setnl     [L1: 0x12345678]}
     */
    // Template#: 4742, Serial#: 4926
    public void m_setnl(int m32) {
        assemble0502((byte) 0x9D, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setnl     ebx[esi * 4 + 305419896]}
     */
    // Template#: 4743, Serial#: 4930
    public void setnl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0503((byte) 0x9D, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setnl     L1: 0x12345678[esi * 4]}
     */
    // Template#: 4744, Serial#: 4925
    public void m_setnl(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0504((byte) 0x9D, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code setnl     [ebx + 305419896]}
     */
    // Template#: 4745, Serial#: 4929
    public void setnl(int disp32, IA32IndirectRegister32 destination) {
        assemble0505((byte) 0x9D, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnle  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setnle    ebx[esi * 4 + 18]}
     */
    // Template#: 4746, Serial#: 4946
    public void setnle(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0497((byte) 0x9F, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnle  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code setnle    [ebx + 18]}
     */
    // Template#: 4747, Serial#: 4945
    public void setnle(byte disp8, IA32IndirectRegister32 destination) {
        assemble0498((byte) 0x9F, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnle  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setnle    ebx[esi * 4]}
     */
    // Template#: 4748, Serial#: 4942
    public void setnle(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0499((byte) 0x9F, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnle  }<i>destination</i>
     * Example disassembly syntax: {@code setnle    al}
     */
    // Template#: 4749, Serial#: 4949
    public void setnle(IA32GeneralRegister8 destination) {
        assemble0500((byte) 0x9F, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnle  }<i>destination</i>
     * Example disassembly syntax: {@code setnle    [ebx]}
     */
    // Template#: 4750, Serial#: 4941
    public void setnle(IA32IndirectRegister32 destination) {
        assemble0501((byte) 0x9F, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnle  }<i>m32</i>
     * Example disassembly syntax: {@code setnle    [L1: 0x12345678]}
     */
    // Template#: 4751, Serial#: 4944
    public void m_setnle(int m32) {
        assemble0502((byte) 0x9F, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnle  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setnle    ebx[esi * 4 + 305419896]}
     */
    // Template#: 4752, Serial#: 4948
    public void setnle(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0503((byte) 0x9F, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnle  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setnle    L1: 0x12345678[esi * 4]}
     */
    // Template#: 4753, Serial#: 4943
    public void m_setnle(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0504((byte) 0x9F, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnle  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code setnle    [ebx + 305419896]}
     */
    // Template#: 4754, Serial#: 4947
    public void setnle(int disp32, IA32IndirectRegister32 destination) {
        assemble0505((byte) 0x9F, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setno  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setno     ebx[esi * 4 + 18]}
     */
    // Template#: 4755, Serial#: 3351
    public void setno(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0497((byte) 0x91, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setno  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code setno     [ebx + 18]}
     */
    // Template#: 4756, Serial#: 3350
    public void setno(byte disp8, IA32IndirectRegister32 destination) {
        assemble0498((byte) 0x91, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setno  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setno     ebx[esi * 4]}
     */
    // Template#: 4757, Serial#: 3347
    public void setno(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0499((byte) 0x91, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setno  }<i>destination</i>
     * Example disassembly syntax: {@code setno     al}
     */
    // Template#: 4758, Serial#: 3354
    public void setno(IA32GeneralRegister8 destination) {
        assemble0500((byte) 0x91, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setno  }<i>destination</i>
     * Example disassembly syntax: {@code setno     [ebx]}
     */
    // Template#: 4759, Serial#: 3346
    public void setno(IA32IndirectRegister32 destination) {
        assemble0501((byte) 0x91, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setno  }<i>m32</i>
     * Example disassembly syntax: {@code setno     [L1: 0x12345678]}
     */
    // Template#: 4760, Serial#: 3349
    public void m_setno(int m32) {
        assemble0502((byte) 0x91, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code setno  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setno     ebx[esi * 4 + 305419896]}
     */
    // Template#: 4761, Serial#: 3353
    public void setno(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0503((byte) 0x91, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setno  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setno     L1: 0x12345678[esi * 4]}
     */
    // Template#: 4762, Serial#: 3348
    public void m_setno(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0504((byte) 0x91, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setno  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code setno     [ebx + 305419896]}
     */
    // Template#: 4763, Serial#: 3352
    public void setno(int disp32, IA32IndirectRegister32 destination) {
        assemble0505((byte) 0x91, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnp  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setnp     ebx[esi * 4 + 18]}
     */
    // Template#: 4764, Serial#: 4910
    public void setnp(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0497((byte) 0x9B, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnp  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code setnp     [ebx + 18]}
     */
    // Template#: 4765, Serial#: 4909
    public void setnp(byte disp8, IA32IndirectRegister32 destination) {
        assemble0498((byte) 0x9B, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnp  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setnp     ebx[esi * 4]}
     */
    // Template#: 4766, Serial#: 4906
    public void setnp(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0499((byte) 0x9B, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnp  }<i>destination</i>
     * Example disassembly syntax: {@code setnp     al}
     */
    // Template#: 4767, Serial#: 4913
    public void setnp(IA32GeneralRegister8 destination) {
        assemble0500((byte) 0x9B, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnp  }<i>destination</i>
     * Example disassembly syntax: {@code setnp     [ebx]}
     */
    // Template#: 4768, Serial#: 4905
    public void setnp(IA32IndirectRegister32 destination) {
        assemble0501((byte) 0x9B, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnp  }<i>m32</i>
     * Example disassembly syntax: {@code setnp     [L1: 0x12345678]}
     */
    // Template#: 4769, Serial#: 4908
    public void m_setnp(int m32) {
        assemble0502((byte) 0x9B, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnp  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setnp     ebx[esi * 4 + 305419896]}
     */
    // Template#: 4770, Serial#: 4912
    public void setnp(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0503((byte) 0x9B, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnp  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setnp     L1: 0x12345678[esi * 4]}
     */
    // Template#: 4771, Serial#: 4907
    public void m_setnp(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0504((byte) 0x9B, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnp  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code setnp     [ebx + 305419896]}
     */
    // Template#: 4772, Serial#: 4911
    public void setnp(int disp32, IA32IndirectRegister32 destination) {
        assemble0505((byte) 0x9B, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setns  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setns     ebx[esi * 4 + 18]}
     */
    // Template#: 4773, Serial#: 4892
    public void setns(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0497((byte) 0x99, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setns  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code setns     [ebx + 18]}
     */
    // Template#: 4774, Serial#: 4891
    public void setns(byte disp8, IA32IndirectRegister32 destination) {
        assemble0498((byte) 0x99, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setns  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setns     ebx[esi * 4]}
     */
    // Template#: 4775, Serial#: 4888
    public void setns(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0499((byte) 0x99, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setns  }<i>destination</i>
     * Example disassembly syntax: {@code setns     al}
     */
    // Template#: 4776, Serial#: 4895
    public void setns(IA32GeneralRegister8 destination) {
        assemble0500((byte) 0x99, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setns  }<i>destination</i>
     * Example disassembly syntax: {@code setns     [ebx]}
     */
    // Template#: 4777, Serial#: 4887
    public void setns(IA32IndirectRegister32 destination) {
        assemble0501((byte) 0x99, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setns  }<i>m32</i>
     * Example disassembly syntax: {@code setns     [L1: 0x12345678]}
     */
    // Template#: 4778, Serial#: 4890
    public void m_setns(int m32) {
        assemble0502((byte) 0x99, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code setns  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setns     ebx[esi * 4 + 305419896]}
     */
    // Template#: 4779, Serial#: 4894
    public void setns(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0503((byte) 0x99, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setns  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setns     L1: 0x12345678[esi * 4]}
     */
    // Template#: 4780, Serial#: 4889
    public void m_setns(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0504((byte) 0x99, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setns  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code setns     [ebx + 305419896]}
     */
    // Template#: 4781, Serial#: 4893
    public void setns(int disp32, IA32IndirectRegister32 destination) {
        assemble0505((byte) 0x99, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnz  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setnz     ebx[esi * 4 + 18]}
     */
    // Template#: 4782, Serial#: 3387
    public void setnz(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0497((byte) 0x95, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnz  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code setnz     [ebx + 18]}
     */
    // Template#: 4783, Serial#: 3386
    public void setnz(byte disp8, IA32IndirectRegister32 destination) {
        assemble0498((byte) 0x95, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnz  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setnz     ebx[esi * 4]}
     */
    // Template#: 4784, Serial#: 3383
    public void setnz(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0499((byte) 0x95, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnz  }<i>destination</i>
     * Example disassembly syntax: {@code setnz     al}
     */
    // Template#: 4785, Serial#: 3390
    public void setnz(IA32GeneralRegister8 destination) {
        assemble0500((byte) 0x95, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnz  }<i>destination</i>
     * Example disassembly syntax: {@code setnz     [ebx]}
     */
    // Template#: 4786, Serial#: 3382
    public void setnz(IA32IndirectRegister32 destination) {
        assemble0501((byte) 0x95, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnz  }<i>m32</i>
     * Example disassembly syntax: {@code setnz     [L1: 0x12345678]}
     */
    // Template#: 4787, Serial#: 3385
    public void m_setnz(int m32) {
        assemble0502((byte) 0x95, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnz  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setnz     ebx[esi * 4 + 305419896]}
     */
    // Template#: 4788, Serial#: 3389
    public void setnz(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0503((byte) 0x95, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnz  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setnz     L1: 0x12345678[esi * 4]}
     */
    // Template#: 4789, Serial#: 3384
    public void m_setnz(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0504((byte) 0x95, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setnz  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code setnz     [ebx + 305419896]}
     */
    // Template#: 4790, Serial#: 3388
    public void setnz(int disp32, IA32IndirectRegister32 destination) {
        assemble0505((byte) 0x95, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code seto  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code seto      ebx[esi * 4 + 18]}
     */
    // Template#: 4791, Serial#: 3342
    public void seto(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0497((byte) 0x90, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code seto  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code seto      [ebx + 18]}
     */
    // Template#: 4792, Serial#: 3341
    public void seto(byte disp8, IA32IndirectRegister32 destination) {
        assemble0498((byte) 0x90, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code seto  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code seto      ebx[esi * 4]}
     */
    // Template#: 4793, Serial#: 3338
    public void seto(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0499((byte) 0x90, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code seto  }<i>destination</i>
     * Example disassembly syntax: {@code seto      al}
     */
    // Template#: 4794, Serial#: 3345
    public void seto(IA32GeneralRegister8 destination) {
        assemble0500((byte) 0x90, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code seto  }<i>destination</i>
     * Example disassembly syntax: {@code seto      [ebx]}
     */
    // Template#: 4795, Serial#: 3337
    public void seto(IA32IndirectRegister32 destination) {
        assemble0501((byte) 0x90, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code seto  }<i>m32</i>
     * Example disassembly syntax: {@code seto      [L1: 0x12345678]}
     */
    // Template#: 4796, Serial#: 3340
    public void m_seto(int m32) {
        assemble0502((byte) 0x90, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code seto  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code seto      ebx[esi * 4 + 305419896]}
     */
    // Template#: 4797, Serial#: 3344
    public void seto(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0503((byte) 0x90, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code seto  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code seto      L1: 0x12345678[esi * 4]}
     */
    // Template#: 4798, Serial#: 3339
    public void m_seto(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0504((byte) 0x90, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code seto  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code seto      [ebx + 305419896]}
     */
    // Template#: 4799, Serial#: 3343
    public void seto(int disp32, IA32IndirectRegister32 destination) {
        assemble0505((byte) 0x90, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setp  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setp      ebx[esi * 4 + 18]}
     */
    // Template#: 4800, Serial#: 4901
    public void setp(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0497((byte) 0x9A, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setp  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code setp      [ebx + 18]}
     */
    // Template#: 4801, Serial#: 4900
    public void setp(byte disp8, IA32IndirectRegister32 destination) {
        assemble0498((byte) 0x9A, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setp  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setp      ebx[esi * 4]}
     */
    // Template#: 4802, Serial#: 4897
    public void setp(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0499((byte) 0x9A, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setp  }<i>destination</i>
     * Example disassembly syntax: {@code setp      al}
     */
    // Template#: 4803, Serial#: 4904
    public void setp(IA32GeneralRegister8 destination) {
        assemble0500((byte) 0x9A, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setp  }<i>destination</i>
     * Example disassembly syntax: {@code setp      [ebx]}
     */
    // Template#: 4804, Serial#: 4896
    public void setp(IA32IndirectRegister32 destination) {
        assemble0501((byte) 0x9A, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setp  }<i>m32</i>
     * Example disassembly syntax: {@code setp      [L1: 0x12345678]}
     */
    // Template#: 4805, Serial#: 4899
    public void m_setp(int m32) {
        assemble0502((byte) 0x9A, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code setp  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setp      ebx[esi * 4 + 305419896]}
     */
    // Template#: 4806, Serial#: 4903
    public void setp(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0503((byte) 0x9A, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setp  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setp      L1: 0x12345678[esi * 4]}
     */
    // Template#: 4807, Serial#: 4898
    public void m_setp(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0504((byte) 0x9A, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setp  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code setp      [ebx + 305419896]}
     */
    // Template#: 4808, Serial#: 4902
    public void setp(int disp32, IA32IndirectRegister32 destination) {
        assemble0505((byte) 0x9A, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sets  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sets      ebx[esi * 4 + 18]}
     */
    // Template#: 4809, Serial#: 4883
    public void sets(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0497((byte) 0x98, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sets  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code sets      [ebx + 18]}
     */
    // Template#: 4810, Serial#: 4882
    public void sets(byte disp8, IA32IndirectRegister32 destination) {
        assemble0498((byte) 0x98, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sets  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sets      ebx[esi * 4]}
     */
    // Template#: 4811, Serial#: 4879
    public void sets(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0499((byte) 0x98, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sets  }<i>destination</i>
     * Example disassembly syntax: {@code sets      al}
     */
    // Template#: 4812, Serial#: 4886
    public void sets(IA32GeneralRegister8 destination) {
        assemble0500((byte) 0x98, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sets  }<i>destination</i>
     * Example disassembly syntax: {@code sets      [ebx]}
     */
    // Template#: 4813, Serial#: 4878
    public void sets(IA32IndirectRegister32 destination) {
        assemble0501((byte) 0x98, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sets  }<i>m32</i>
     * Example disassembly syntax: {@code sets      [L1: 0x12345678]}
     */
    // Template#: 4814, Serial#: 4881
    public void m_sets(int m32) {
        assemble0502((byte) 0x98, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sets  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sets      ebx[esi * 4 + 305419896]}
     */
    // Template#: 4815, Serial#: 4885
    public void sets(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0503((byte) 0x98, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sets  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sets      L1: 0x12345678[esi * 4]}
     */
    // Template#: 4816, Serial#: 4880
    public void m_sets(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0504((byte) 0x98, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sets  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code sets      [ebx + 305419896]}
     */
    // Template#: 4817, Serial#: 4884
    public void sets(int disp32, IA32IndirectRegister32 destination) {
        assemble0505((byte) 0x98, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setz  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setz      ebx[esi * 4 + 18]}
     */
    // Template#: 4818, Serial#: 3378
    public void setz(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0497((byte) 0x94, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setz  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code setz      [ebx + 18]}
     */
    // Template#: 4819, Serial#: 3377
    public void setz(byte disp8, IA32IndirectRegister32 destination) {
        assemble0498((byte) 0x94, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setz  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setz      ebx[esi * 4]}
     */
    // Template#: 4820, Serial#: 3374
    public void setz(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0499((byte) 0x94, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setz  }<i>destination</i>
     * Example disassembly syntax: {@code setz      al}
     */
    // Template#: 4821, Serial#: 3381
    public void setz(IA32GeneralRegister8 destination) {
        assemble0500((byte) 0x94, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setz  }<i>destination</i>
     * Example disassembly syntax: {@code setz      [ebx]}
     */
    // Template#: 4822, Serial#: 3373
    public void setz(IA32IndirectRegister32 destination) {
        assemble0501((byte) 0x94, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code setz  }<i>m32</i>
     * Example disassembly syntax: {@code setz      [L1: 0x12345678]}
     */
    // Template#: 4823, Serial#: 3376
    public void m_setz(int m32) {
        assemble0502((byte) 0x94, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code setz  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setz      ebx[esi * 4 + 305419896]}
     */
    // Template#: 4824, Serial#: 3380
    public void setz(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0503((byte) 0x94, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setz  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code setz      L1: 0x12345678[esi * 4]}
     */
    // Template#: 4825, Serial#: 3375
    public void m_setz(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0504((byte) 0x94, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code setz  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code setz      [ebx + 305419896]}
     */
    // Template#: 4826, Serial#: 3379
    public void setz(int disp32, IA32IndirectRegister32 destination) {
        assemble0505((byte) 0x94, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sfence  }
     * Example disassembly syntax: {@code sfence  }
     */
    // Template#: 4827, Serial#: 5009
    public void sfence() {
        assemble0357((byte) 0xAE, (byte) 0x07);
    }

    /**
     * Pseudo-external assembler syntax: {@code sgdt  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sgdt      ebx[esi * 4 + 18]}
     */
    // Template#: 4828, Serial#: 2581
    public void sgdt(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0244((byte) 0x01, (byte) 0x00, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sgdt  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code sgdt      [ebx + 18]}
     */
    // Template#: 4829, Serial#: 2580
    public void sgdt(byte disp8, IA32IndirectRegister32 destination) {
        assemble0245((byte) 0x01, (byte) 0x00, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sgdt  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sgdt      ebx[esi * 4]}
     */
    // Template#: 4830, Serial#: 2553
    public void sgdt(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0246((byte) 0x01, (byte) 0x00, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sgdt  }<i>destination</i>
     * Example disassembly syntax: {@code sgdt      [ebx]}
     */
    // Template#: 4831, Serial#: 2552
    public void sgdt(IA32IndirectRegister32 destination) {
        assemble0247((byte) 0x01, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sgdt  }<i>m32</i>
     * Example disassembly syntax: {@code sgdt      [L1: 0x12345678]}
     */
    // Template#: 4832, Serial#: 2555
    public void m_sgdt(int m32) {
        assemble0248((byte) 0x01, (byte) 0x00, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sgdt  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sgdt      ebx[esi * 4 + 305419896]}
     */
    // Template#: 4833, Serial#: 2595
    public void sgdt(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0249((byte) 0x01, (byte) 0x00, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sgdt  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sgdt      L1: 0x12345678[esi * 4]}
     */
    // Template#: 4834, Serial#: 2554
    public void m_sgdt(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0250((byte) 0x01, (byte) 0x00, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sgdt  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code sgdt      [ebx + 305419896]}
     */
    // Template#: 4835, Serial#: 2594
    public void sgdt(int disp32, IA32IndirectRegister32 destination) {
        assemble0251((byte) 0x01, (byte) 0x00, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shlb      ebx[esi * 4 + 18], 0x1}
     */
    // Template#: 4836, Serial#: 987
    public void shlb___1(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD0, (byte) 0x04, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shll      ebx[esi * 4 + 18], 0x1}
     */
    // Template#: 4837, Serial#: 1050
    public void shll___1(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD1, (byte) 0x04, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shlw      ebx[esi * 4 + 18], 0x1}
     */
    // Template#: 4838, Serial#: 1113
    public void shlw___1(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0292((byte) 0xD1, (byte) 0x04, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shlb      ebx[esi * 4 + 18], cl}
     */
    // Template#: 4839, Serial#: 1176
    public void shlb___CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD2, (byte) 0x04, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shll      ebx[esi * 4 + 18], cl}
     */
    // Template#: 4840, Serial#: 1239
    public void shll___CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD3, (byte) 0x04, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shlw      ebx[esi * 4 + 18], cl}
     */
    // Template#: 4841, Serial#: 1302
    public void shlw___CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0292((byte) 0xD3, (byte) 0x04, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shlb      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 4842, Serial#: 737
    public void shlb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0xC0, (byte) 0x04, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shll      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 4843, Serial#: 800
    public void shll(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0xC1, (byte) 0x04, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shlw      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 4844, Serial#: 863
    public void shlw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0004((byte) 0xC1, (byte) 0x04, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code shlb      [ebx + 18], 0x1}
     */
    // Template#: 4845, Serial#: 986
    public void shlb___1(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD0, (byte) 0x04, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code shll      [ebx + 18], 0x1}
     */
    // Template#: 4846, Serial#: 1049
    public void shll___1(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD1, (byte) 0x04, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code shlw      [ebx + 18], 0x1}
     */
    // Template#: 4847, Serial#: 1112
    public void shlw___1(byte disp8, IA32IndirectRegister32 destination) {
        assemble0293((byte) 0xD1, (byte) 0x04, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code shlb      [ebx + 18], cl}
     */
    // Template#: 4848, Serial#: 1175
    public void shlb___CL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD2, (byte) 0x04, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code shll      [ebx + 18], cl}
     */
    // Template#: 4849, Serial#: 1238
    public void shll___CL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD3, (byte) 0x04, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code shlw      [ebx + 18], cl}
     */
    // Template#: 4850, Serial#: 1301
    public void shlw___CL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0293((byte) 0xD3, (byte) 0x04, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shlb      [ebx + 18], 0x12}
     */
    // Template#: 4851, Serial#: 736
    public void shlb(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0xC0, (byte) 0x04, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shll      [ebx + 18], 0x12}
     */
    // Template#: 4852, Serial#: 799
    public void shll(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0xC1, (byte) 0x04, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shlw      [ebx + 18], 0x12}
     */
    // Template#: 4853, Serial#: 862
    public void shlw(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0011((byte) 0xC1, (byte) 0x04, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shlb      ebx[esi * 4], 0x1}
     */
    // Template#: 4854, Serial#: 967
    public void shlb___1(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD0, (byte) 0x04, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shll      ebx[esi * 4], 0x1}
     */
    // Template#: 4855, Serial#: 1030
    public void shll___1(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD1, (byte) 0x04, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shlw      ebx[esi * 4], 0x1}
     */
    // Template#: 4856, Serial#: 1093
    public void shlw___1(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0294((byte) 0xD1, (byte) 0x04, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shlb      ebx[esi * 4], cl}
     */
    // Template#: 4857, Serial#: 1156
    public void shlb___CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD2, (byte) 0x04, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shll      ebx[esi * 4], cl}
     */
    // Template#: 4858, Serial#: 1219
    public void shll___CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD3, (byte) 0x04, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shlw      ebx[esi * 4], cl}
     */
    // Template#: 4859, Serial#: 1282
    public void shlw___CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0294((byte) 0xD3, (byte) 0x04, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shlb      ebx[esi * 4], 0x12}
     */
    // Template#: 4860, Serial#: 717
    public void shlb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0xC0, (byte) 0x04, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shll      ebx[esi * 4], 0x12}
     */
    // Template#: 4861, Serial#: 780
    public void shll(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0xC1, (byte) 0x04, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shlw      ebx[esi * 4], 0x12}
     */
    // Template#: 4862, Serial#: 843
    public void shlw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0018((byte) 0xC1, (byte) 0x04, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>destination</i>
     * Example disassembly syntax: {@code shlw      ax, 0x1}
     */
    // Template#: 4863, Serial#: 1136
    public void shlw___1(IA32GeneralRegister16 destination) {
        assemble0188((byte) 0xD1, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>destination</i>
     * Example disassembly syntax: {@code shlw      ax, cl}
     */
    // Template#: 4864, Serial#: 1325
    public void shlw___CL(IA32GeneralRegister16 destination) {
        assemble0188((byte) 0xD3, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shlw      ax, 0x12}
     */
    // Template#: 4865, Serial#: 886
    public void shlw(IA32GeneralRegister16 destination, byte imm8) {
        assemble0024((byte) 0xC1, (byte) 0x04, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>destination</i>
     * Example disassembly syntax: {@code shll      eax, 0x1}
     */
    // Template#: 4866, Serial#: 1073
    public void shll___1(IA32GeneralRegister32 destination) {
        assemble0189((byte) 0xD1, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>destination</i>
     * Example disassembly syntax: {@code shll      eax, cl}
     */
    // Template#: 4867, Serial#: 1262
    public void shll___CL(IA32GeneralRegister32 destination) {
        assemble0189((byte) 0xD3, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shll      eax, 0x12}
     */
    // Template#: 4868, Serial#: 823
    public void shll(IA32GeneralRegister32 destination, byte imm8) {
        assemble0035((byte) 0xC1, (byte) 0x04, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>destination</i>
     * Example disassembly syntax: {@code shlb      al, 0x1}
     */
    // Template#: 4869, Serial#: 1010
    public void shlb___1(IA32GeneralRegister8 destination) {
        assemble0297((byte) 0xD0, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>destination</i>
     * Example disassembly syntax: {@code shlb      al, cl}
     */
    // Template#: 4870, Serial#: 1199
    public void shlb___CL(IA32GeneralRegister8 destination) {
        assemble0297((byte) 0xD2, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shlb      al, 0x12}
     */
    // Template#: 4871, Serial#: 760
    public void shlb(IA32GeneralRegister8 destination, byte imm8) {
        assemble0046((byte) 0xC0, (byte) 0x04, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>destination</i>
     * Example disassembly syntax: {@code shlb      [ebx], 0x1}
     */
    // Template#: 4872, Serial#: 966
    public void shlb___1(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD0, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>destination</i>
     * Example disassembly syntax: {@code shll      [ebx], 0x1}
     */
    // Template#: 4873, Serial#: 1029
    public void shll___1(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD1, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>destination</i>
     * Example disassembly syntax: {@code shlw      [ebx], 0x1}
     */
    // Template#: 4874, Serial#: 1092
    public void shlw___1(IA32IndirectRegister32 destination) {
        assemble0298((byte) 0xD1, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>destination</i>
     * Example disassembly syntax: {@code shlb      [ebx], cl}
     */
    // Template#: 4875, Serial#: 1155
    public void shlb___CL(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD2, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>destination</i>
     * Example disassembly syntax: {@code shll      [ebx], cl}
     */
    // Template#: 4876, Serial#: 1218
    public void shll___CL(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD3, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>destination</i>
     * Example disassembly syntax: {@code shlw      [ebx], cl}
     */
    // Template#: 4877, Serial#: 1281
    public void shlw___CL(IA32IndirectRegister32 destination) {
        assemble0298((byte) 0xD3, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shlb      [ebx], 0x12}
     */
    // Template#: 4878, Serial#: 716
    public void shlb(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0xC0, (byte) 0x04, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shll      [ebx], 0x12}
     */
    // Template#: 4879, Serial#: 779
    public void shll(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0xC1, (byte) 0x04, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shlw      [ebx], 0x12}
     */
    // Template#: 4880, Serial#: 842
    public void shlw(IA32IndirectRegister32 destination, byte imm8) {
        assemble0057((byte) 0xC1, (byte) 0x04, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>m32</i>
     * Example disassembly syntax: {@code shlb      [L1: 0x12345678], 0x1}
     */
    // Template#: 4881, Serial#: 969
    public void m_shlb___1(int m32) {
        assemble0299((byte) 0xD0, (byte) 0x04, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>m32</i>
     * Example disassembly syntax: {@code shll      [L1: 0x12345678], 0x1}
     */
    // Template#: 4882, Serial#: 1032
    public void m_shll___1(int m32) {
        assemble0299((byte) 0xD1, (byte) 0x04, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>m32</i>
     * Example disassembly syntax: {@code shlw      [L1: 0x12345678], 0x1}
     */
    // Template#: 4883, Serial#: 1095
    public void m_shlw___1(int m32) {
        assemble0300((byte) 0xD1, (byte) 0x04, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>m32</i>
     * Example disassembly syntax: {@code shlb      [L1: 0x12345678], cl}
     */
    // Template#: 4884, Serial#: 1158
    public void m_shlb___CL(int m32) {
        assemble0299((byte) 0xD2, (byte) 0x04, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>m32</i>
     * Example disassembly syntax: {@code shll      [L1: 0x12345678], cl}
     */
    // Template#: 4885, Serial#: 1221
    public void m_shll___CL(int m32) {
        assemble0299((byte) 0xD3, (byte) 0x04, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>m32</i>
     * Example disassembly syntax: {@code shlw      [L1: 0x12345678], cl}
     */
    // Template#: 4886, Serial#: 1284
    public void m_shlw___CL(int m32) {
        assemble0300((byte) 0xD3, (byte) 0x04, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shlb      [L1: 0x12345678], 0x12}
     */
    // Template#: 4887, Serial#: 719
    public void m_shlb(int m32, byte imm8) {
        assemble0064((byte) 0xC0, (byte) 0x04, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shll      [L1: 0x12345678], 0x12}
     */
    // Template#: 4888, Serial#: 782
    public void m_shll(int m32, byte imm8) {
        assemble0064((byte) 0xC1, (byte) 0x04, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shlw      [L1: 0x12345678], 0x12}
     */
    // Template#: 4889, Serial#: 845
    public void m_shlw(int m32, byte imm8) {
        assemble0065((byte) 0xC1, (byte) 0x04, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shlb      ebx[esi * 4 + 305419896], 0x1}
     */
    // Template#: 4890, Serial#: 1001
    public void shlb___1(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD0, (byte) 0x04, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shll      ebx[esi * 4 + 305419896], 0x1}
     */
    // Template#: 4891, Serial#: 1064
    public void shll___1(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD1, (byte) 0x04, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shlw      ebx[esi * 4 + 305419896], 0x1}
     */
    // Template#: 4892, Serial#: 1127
    public void shlw___1(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0301((byte) 0xD1, (byte) 0x04, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shlb      ebx[esi * 4 + 305419896], cl}
     */
    // Template#: 4893, Serial#: 1190
    public void shlb___CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD2, (byte) 0x04, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shll      ebx[esi * 4 + 305419896], cl}
     */
    // Template#: 4894, Serial#: 1253
    public void shll___CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD3, (byte) 0x04, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shlw      ebx[esi * 4 + 305419896], cl}
     */
    // Template#: 4895, Serial#: 1316
    public void shlw___CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0301((byte) 0xD3, (byte) 0x04, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shlb      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 4896, Serial#: 751
    public void shlb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0xC0, (byte) 0x04, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shll      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 4897, Serial#: 814
    public void shll(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0xC1, (byte) 0x04, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shlw      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 4898, Serial#: 877
    public void shlw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0067((byte) 0xC1, (byte) 0x04, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shlb      L1: 0x12345678[esi * 4], 0x1}
     */
    // Template#: 4899, Serial#: 968
    public void m_shlb___1(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD0, (byte) 0x04, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shll      L1: 0x12345678[esi * 4], 0x1}
     */
    // Template#: 4900, Serial#: 1031
    public void m_shll___1(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD1, (byte) 0x04, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shlw      L1: 0x12345678[esi * 4], 0x1}
     */
    // Template#: 4901, Serial#: 1094
    public void m_shlw___1(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0302((byte) 0xD1, (byte) 0x04, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shlb      L1: 0x12345678[esi * 4], cl}
     */
    // Template#: 4902, Serial#: 1157
    public void m_shlb___CL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD2, (byte) 0x04, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shll      L1: 0x12345678[esi * 4], cl}
     */
    // Template#: 4903, Serial#: 1220
    public void m_shll___CL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD3, (byte) 0x04, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shlw      L1: 0x12345678[esi * 4], cl}
     */
    // Template#: 4904, Serial#: 1283
    public void m_shlw___CL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0302((byte) 0xD3, (byte) 0x04, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shlb      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 4905, Serial#: 718
    public void m_shlb(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0xC0, (byte) 0x04, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shll      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 4906, Serial#: 781
    public void m_shll(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0xC1, (byte) 0x04, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shlw      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 4907, Serial#: 844
    public void m_shlw(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0077((byte) 0xC1, (byte) 0x04, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code shlb      [ebx + 305419896], 0x1}
     */
    // Template#: 4908, Serial#: 1000
    public void shlb___1(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD0, (byte) 0x04, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code shll      [ebx + 305419896], 0x1}
     */
    // Template#: 4909, Serial#: 1063
    public void shll___1(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD1, (byte) 0x04, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code shlw      [ebx + 305419896], 0x1}
     */
    // Template#: 4910, Serial#: 1126
    public void shlw___1(int disp32, IA32IndirectRegister32 destination) {
        assemble0303((byte) 0xD1, (byte) 0x04, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code shlb      [ebx + 305419896], cl}
     */
    // Template#: 4911, Serial#: 1189
    public void shlb___CL(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD2, (byte) 0x04, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code shll      [ebx + 305419896], cl}
     */
    // Template#: 4912, Serial#: 1252
    public void shll___CL(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD3, (byte) 0x04, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code shlw      [ebx + 305419896], cl}
     */
    // Template#: 4913, Serial#: 1315
    public void shlw___CL(int disp32, IA32IndirectRegister32 destination) {
        assemble0303((byte) 0xD3, (byte) 0x04, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlb  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shlb      [ebx + 305419896], 0x12}
     */
    // Template#: 4914, Serial#: 750
    public void shlb(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0xC0, (byte) 0x04, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shll  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shll      [ebx + 305419896], 0x12}
     */
    // Template#: 4915, Serial#: 813
    public void shll(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0xC1, (byte) 0x04, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shlw  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shlw      [ebx + 305419896], 0x12}
     */
    // Template#: 4916, Serial#: 876
    public void shlw(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0084((byte) 0xC1, (byte) 0x04, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code shld      ebx[esi * 4 + 18], ax, cl}
     */
    // Template#: 4917, Serial#: 3462
    public void shld_CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0158((byte) 0xA5, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shld      ebx[esi * 4 + 18], ax, 0x12}
     */
    // Template#: 4918, Serial#: 3444
    public void shld(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source, byte imm8) {
        assemble0506((byte) 0xA4, disp8, base, index, scale, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code shld      ebx[esi * 4 + 18], eax, cl}
     */
    // Template#: 4919, Serial#: 3453
    public void shld_CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0159((byte) 0xA5, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shld      ebx[esi * 4 + 18], eax, 0x12}
     */
    // Template#: 4920, Serial#: 3435
    public void shld(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source, byte imm8) {
        assemble0507((byte) 0xA4, disp8, base, index, scale, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code shld      [ebx + 18], ax, cl}
     */
    // Template#: 4921, Serial#: 3461
    public void shld_CL(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0161((byte) 0xA5, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>disp8</i>, <i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shld      [ebx + 18], ax, 0x12}
     */
    // Template#: 4922, Serial#: 3443
    public void shld(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source, byte imm8) {
        assemble0508((byte) 0xA4, disp8, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code shld      [ebx + 18], eax, cl}
     */
    // Template#: 4923, Serial#: 3452
    public void shld_CL(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0162((byte) 0xA5, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>disp8</i>, <i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shld      [ebx + 18], eax, 0x12}
     */
    // Template#: 4924, Serial#: 3434
    public void shld(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source, byte imm8) {
        assemble0509((byte) 0xA4, disp8, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code shld      ebx[esi * 4], ax, cl}
     */
    // Template#: 4925, Serial#: 3458
    public void shld_CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0164((byte) 0xA5, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shld      ebx[esi * 4], ax, 0x12}
     */
    // Template#: 4926, Serial#: 3440
    public void shld(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source, byte imm8) {
        assemble0510((byte) 0xA4, base, index, scale, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code shld      ebx[esi * 4], eax, cl}
     */
    // Template#: 4927, Serial#: 3449
    public void shld_CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0165((byte) 0xA5, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shld      ebx[esi * 4], eax, 0x12}
     */
    // Template#: 4928, Serial#: 3431
    public void shld(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source, byte imm8) {
        assemble0511((byte) 0xA4, base, index, scale, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code shld      ax, ax, cl}
     */
    // Template#: 4929, Serial#: 3465
    public void shld_CL(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0167((byte) 0xA5, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shld      ax, ax, 0x12}
     */
    // Template#: 4930, Serial#: 3447
    public void shld(IA32GeneralRegister16 destination, IA32GeneralRegister16 source, byte imm8) {
        assemble0512((byte) 0xA4, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code shld      eax, eax, cl}
     */
    // Template#: 4931, Serial#: 3456
    public void shld_CL(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0169((byte) 0xA5, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shld      eax, eax, 0x12}
     */
    // Template#: 4932, Serial#: 3438
    public void shld(IA32GeneralRegister32 destination, IA32GeneralRegister32 source, byte imm8) {
        assemble0513((byte) 0xA4, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code shld      [ebx], ax, cl}
     */
    // Template#: 4933, Serial#: 3457
    public void shld_CL(IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0171((byte) 0xA5, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shld      [ebx], ax, 0x12}
     */
    // Template#: 4934, Serial#: 3439
    public void shld(IA32IndirectRegister32 destination, IA32GeneralRegister16 source, byte imm8) {
        assemble0514((byte) 0xA4, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code shld      [ebx], eax, cl}
     */
    // Template#: 4935, Serial#: 3448
    public void shld_CL(IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0172((byte) 0xA5, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shld      [ebx], eax, 0x12}
     */
    // Template#: 4936, Serial#: 3430
    public void shld(IA32IndirectRegister32 destination, IA32GeneralRegister32 source, byte imm8) {
        assemble0515((byte) 0xA4, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code shld      ebx[esi * 4 + 305419896], ax, cl}
     */
    // Template#: 4937, Serial#: 3464
    public void shld_CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0175((byte) 0xA5, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shld      ebx[esi * 4 + 305419896], ax, 0x12}
     */
    // Template#: 4938, Serial#: 3446
    public void shld(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source, byte imm8) {
        assemble0516((byte) 0xA4, disp32, base, index, scale, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code shld      ebx[esi * 4 + 305419896], eax, cl}
     */
    // Template#: 4939, Serial#: 3455
    public void shld_CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0176((byte) 0xA5, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shld      ebx[esi * 4 + 305419896], eax, 0x12}
     */
    // Template#: 4940, Serial#: 3437
    public void shld(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source, byte imm8) {
        assemble0517((byte) 0xA4, disp32, base, index, scale, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code shld      [L1: 0x12345678], ax, cl}
     */
    // Template#: 4941, Serial#: 3460
    public void m_shld_CL(int m32, IA32GeneralRegister16 source) {
        assemble0177((byte) 0xA5, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>m32</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shld      [L1: 0x12345678], ax, 0x12}
     */
    // Template#: 4942, Serial#: 3442
    public void m_shld(int m32, IA32GeneralRegister16 source, byte imm8) {
        assemble0518((byte) 0xA4, m32, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code shld      [L1: 0x12345678], eax, cl}
     */
    // Template#: 4943, Serial#: 3451
    public void m_shld_CL(int m32, IA32GeneralRegister32 source) {
        assemble0178((byte) 0xA5, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>m32</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shld      [L1: 0x12345678], eax, 0x12}
     */
    // Template#: 4944, Serial#: 3433
    public void m_shld(int m32, IA32GeneralRegister32 source, byte imm8) {
        assemble0519((byte) 0xA4, m32, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code shld      L1: 0x12345678[esi * 4], ax, cl}
     */
    // Template#: 4945, Serial#: 3459
    public void m_shld_CL(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0180((byte) 0xA5, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shld      L1: 0x12345678[esi * 4], ax, 0x12}
     */
    // Template#: 4946, Serial#: 3441
    public void m_shld(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source, byte imm8) {
        assemble0520((byte) 0xA4, m32, index, scale, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code shld      L1: 0x12345678[esi * 4], eax, cl}
     */
    // Template#: 4947, Serial#: 3450
    public void m_shld_CL(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0181((byte) 0xA5, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shld      L1: 0x12345678[esi * 4], eax, 0x12}
     */
    // Template#: 4948, Serial#: 3432
    public void m_shld(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source, byte imm8) {
        assemble0521((byte) 0xA4, m32, index, scale, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code shld      [ebx + 305419896], ax, cl}
     */
    // Template#: 4949, Serial#: 3463
    public void shld_CL(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0183((byte) 0xA5, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>disp32</i>, <i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shld      [ebx + 305419896], ax, 0x12}
     */
    // Template#: 4950, Serial#: 3445
    public void shld(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source, byte imm8) {
        assemble0522((byte) 0xA4, disp32, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code shld      [ebx + 305419896], eax, cl}
     */
    // Template#: 4951, Serial#: 3454
    public void shld_CL(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0184((byte) 0xA5, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shld  }<i>disp32</i>, <i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shld      [ebx + 305419896], eax, 0x12}
     */
    // Template#: 4952, Serial#: 3436
    public void shld(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source, byte imm8) {
        assemble0523((byte) 0xA4, disp32, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shrb      ebx[esi * 4 + 18], 0x1}
     */
    // Template#: 4953, Serial#: 989
    public void shrb___1(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD0, (byte) 0x05, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shrl      ebx[esi * 4 + 18], 0x1}
     */
    // Template#: 4954, Serial#: 1052
    public void shrl___1(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD1, (byte) 0x05, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shrw      ebx[esi * 4 + 18], 0x1}
     */
    // Template#: 4955, Serial#: 1115
    public void shrw___1(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0292((byte) 0xD1, (byte) 0x05, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shrb      ebx[esi * 4 + 18], cl}
     */
    // Template#: 4956, Serial#: 1178
    public void shrb___CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD2, (byte) 0x05, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shrl      ebx[esi * 4 + 18], cl}
     */
    // Template#: 4957, Serial#: 1241
    public void shrl___CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0185((byte) 0xD3, (byte) 0x05, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shrw      ebx[esi * 4 + 18], cl}
     */
    // Template#: 4958, Serial#: 1304
    public void shrw___CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0292((byte) 0xD3, (byte) 0x05, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrb      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 4959, Serial#: 739
    public void shrb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0xC0, (byte) 0x05, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrl      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 4960, Serial#: 802
    public void shrl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0xC1, (byte) 0x05, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrw      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 4961, Serial#: 865
    public void shrw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0004((byte) 0xC1, (byte) 0x05, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code shrb      [ebx + 18], 0x1}
     */
    // Template#: 4962, Serial#: 988
    public void shrb___1(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD0, (byte) 0x05, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code shrl      [ebx + 18], 0x1}
     */
    // Template#: 4963, Serial#: 1051
    public void shrl___1(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD1, (byte) 0x05, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code shrw      [ebx + 18], 0x1}
     */
    // Template#: 4964, Serial#: 1114
    public void shrw___1(byte disp8, IA32IndirectRegister32 destination) {
        assemble0293((byte) 0xD1, (byte) 0x05, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code shrb      [ebx + 18], cl}
     */
    // Template#: 4965, Serial#: 1177
    public void shrb___CL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD2, (byte) 0x05, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code shrl      [ebx + 18], cl}
     */
    // Template#: 4966, Serial#: 1240
    public void shrl___CL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0186((byte) 0xD3, (byte) 0x05, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code shrw      [ebx + 18], cl}
     */
    // Template#: 4967, Serial#: 1303
    public void shrw___CL(byte disp8, IA32IndirectRegister32 destination) {
        assemble0293((byte) 0xD3, (byte) 0x05, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrb      [ebx + 18], 0x12}
     */
    // Template#: 4968, Serial#: 738
    public void shrb(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0xC0, (byte) 0x05, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrl      [ebx + 18], 0x12}
     */
    // Template#: 4969, Serial#: 801
    public void shrl(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0xC1, (byte) 0x05, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrw      [ebx + 18], 0x12}
     */
    // Template#: 4970, Serial#: 864
    public void shrw(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0011((byte) 0xC1, (byte) 0x05, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shrb      ebx[esi * 4], 0x1}
     */
    // Template#: 4971, Serial#: 971
    public void shrb___1(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD0, (byte) 0x05, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shrl      ebx[esi * 4], 0x1}
     */
    // Template#: 4972, Serial#: 1034
    public void shrl___1(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD1, (byte) 0x05, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shrw      ebx[esi * 4], 0x1}
     */
    // Template#: 4973, Serial#: 1097
    public void shrw___1(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0294((byte) 0xD1, (byte) 0x05, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shrb      ebx[esi * 4], cl}
     */
    // Template#: 4974, Serial#: 1160
    public void shrb___CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD2, (byte) 0x05, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shrl      ebx[esi * 4], cl}
     */
    // Template#: 4975, Serial#: 1223
    public void shrl___CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0187((byte) 0xD3, (byte) 0x05, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shrw      ebx[esi * 4], cl}
     */
    // Template#: 4976, Serial#: 1286
    public void shrw___CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0294((byte) 0xD3, (byte) 0x05, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrb      ebx[esi * 4], 0x12}
     */
    // Template#: 4977, Serial#: 721
    public void shrb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0xC0, (byte) 0x05, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrl      ebx[esi * 4], 0x12}
     */
    // Template#: 4978, Serial#: 784
    public void shrl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0xC1, (byte) 0x05, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrw      ebx[esi * 4], 0x12}
     */
    // Template#: 4979, Serial#: 847
    public void shrw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0018((byte) 0xC1, (byte) 0x05, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>destination</i>
     * Example disassembly syntax: {@code shrw      ax, 0x1}
     */
    // Template#: 4980, Serial#: 1137
    public void shrw___1(IA32GeneralRegister16 destination) {
        assemble0188((byte) 0xD1, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>destination</i>
     * Example disassembly syntax: {@code shrw      ax, cl}
     */
    // Template#: 4981, Serial#: 1326
    public void shrw___CL(IA32GeneralRegister16 destination) {
        assemble0188((byte) 0xD3, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrw      ax, 0x12}
     */
    // Template#: 4982, Serial#: 887
    public void shrw(IA32GeneralRegister16 destination, byte imm8) {
        assemble0024((byte) 0xC1, (byte) 0x05, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>destination</i>
     * Example disassembly syntax: {@code shrl      eax, 0x1}
     */
    // Template#: 4983, Serial#: 1074
    public void shrl___1(IA32GeneralRegister32 destination) {
        assemble0189((byte) 0xD1, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>destination</i>
     * Example disassembly syntax: {@code shrl      eax, cl}
     */
    // Template#: 4984, Serial#: 1263
    public void shrl___CL(IA32GeneralRegister32 destination) {
        assemble0189((byte) 0xD3, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrl      eax, 0x12}
     */
    // Template#: 4985, Serial#: 824
    public void shrl(IA32GeneralRegister32 destination, byte imm8) {
        assemble0035((byte) 0xC1, (byte) 0x05, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>destination</i>
     * Example disassembly syntax: {@code shrb      al, 0x1}
     */
    // Template#: 4986, Serial#: 1011
    public void shrb___1(IA32GeneralRegister8 destination) {
        assemble0297((byte) 0xD0, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>destination</i>
     * Example disassembly syntax: {@code shrb      al, cl}
     */
    // Template#: 4987, Serial#: 1200
    public void shrb___CL(IA32GeneralRegister8 destination) {
        assemble0297((byte) 0xD2, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrb      al, 0x12}
     */
    // Template#: 4988, Serial#: 761
    public void shrb(IA32GeneralRegister8 destination, byte imm8) {
        assemble0046((byte) 0xC0, (byte) 0x05, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>destination</i>
     * Example disassembly syntax: {@code shrb      [ebx], 0x1}
     */
    // Template#: 4989, Serial#: 970
    public void shrb___1(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD0, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>destination</i>
     * Example disassembly syntax: {@code shrl      [ebx], 0x1}
     */
    // Template#: 4990, Serial#: 1033
    public void shrl___1(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD1, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>destination</i>
     * Example disassembly syntax: {@code shrw      [ebx], 0x1}
     */
    // Template#: 4991, Serial#: 1096
    public void shrw___1(IA32IndirectRegister32 destination) {
        assemble0298((byte) 0xD1, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>destination</i>
     * Example disassembly syntax: {@code shrb      [ebx], cl}
     */
    // Template#: 4992, Serial#: 1159
    public void shrb___CL(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD2, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>destination</i>
     * Example disassembly syntax: {@code shrl      [ebx], cl}
     */
    // Template#: 4993, Serial#: 1222
    public void shrl___CL(IA32IndirectRegister32 destination) {
        assemble0190((byte) 0xD3, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>destination</i>
     * Example disassembly syntax: {@code shrw      [ebx], cl}
     */
    // Template#: 4994, Serial#: 1285
    public void shrw___CL(IA32IndirectRegister32 destination) {
        assemble0298((byte) 0xD3, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrb      [ebx], 0x12}
     */
    // Template#: 4995, Serial#: 720
    public void shrb(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0xC0, (byte) 0x05, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrl      [ebx], 0x12}
     */
    // Template#: 4996, Serial#: 783
    public void shrl(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0xC1, (byte) 0x05, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrw      [ebx], 0x12}
     */
    // Template#: 4997, Serial#: 846
    public void shrw(IA32IndirectRegister32 destination, byte imm8) {
        assemble0057((byte) 0xC1, (byte) 0x05, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>m32</i>
     * Example disassembly syntax: {@code shrb      [L1: 0x12345678], 0x1}
     */
    // Template#: 4998, Serial#: 973
    public void m_shrb___1(int m32) {
        assemble0299((byte) 0xD0, (byte) 0x05, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>m32</i>
     * Example disassembly syntax: {@code shrl      [L1: 0x12345678], 0x1}
     */
    // Template#: 4999, Serial#: 1036
    public void m_shrl___1(int m32) {
        assemble0299((byte) 0xD1, (byte) 0x05, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>m32</i>
     * Example disassembly syntax: {@code shrw      [L1: 0x12345678], 0x1}
     */
    // Template#: 5000, Serial#: 1099
    public void m_shrw___1(int m32) {
        assemble0300((byte) 0xD1, (byte) 0x05, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>m32</i>
     * Example disassembly syntax: {@code shrb      [L1: 0x12345678], cl}
     */
    // Template#: 5001, Serial#: 1162
    public void m_shrb___CL(int m32) {
        assemble0299((byte) 0xD2, (byte) 0x05, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>m32</i>
     * Example disassembly syntax: {@code shrl      [L1: 0x12345678], cl}
     */
    // Template#: 5002, Serial#: 1225
    public void m_shrl___CL(int m32) {
        assemble0299((byte) 0xD3, (byte) 0x05, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>m32</i>
     * Example disassembly syntax: {@code shrw      [L1: 0x12345678], cl}
     */
    // Template#: 5003, Serial#: 1288
    public void m_shrw___CL(int m32) {
        assemble0300((byte) 0xD3, (byte) 0x05, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrb      [L1: 0x12345678], 0x12}
     */
    // Template#: 5004, Serial#: 723
    public void m_shrb(int m32, byte imm8) {
        assemble0064((byte) 0xC0, (byte) 0x05, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrl      [L1: 0x12345678], 0x12}
     */
    // Template#: 5005, Serial#: 786
    public void m_shrl(int m32, byte imm8) {
        assemble0064((byte) 0xC1, (byte) 0x05, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrw      [L1: 0x12345678], 0x12}
     */
    // Template#: 5006, Serial#: 849
    public void m_shrw(int m32, byte imm8) {
        assemble0065((byte) 0xC1, (byte) 0x05, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shrb      ebx[esi * 4 + 305419896], 0x1}
     */
    // Template#: 5007, Serial#: 1003
    public void shrb___1(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD0, (byte) 0x05, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shrl      ebx[esi * 4 + 305419896], 0x1}
     */
    // Template#: 5008, Serial#: 1066
    public void shrl___1(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD1, (byte) 0x05, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shrw      ebx[esi * 4 + 305419896], 0x1}
     */
    // Template#: 5009, Serial#: 1129
    public void shrw___1(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0301((byte) 0xD1, (byte) 0x05, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shrb      ebx[esi * 4 + 305419896], cl}
     */
    // Template#: 5010, Serial#: 1192
    public void shrb___CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD2, (byte) 0x05, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shrl      ebx[esi * 4 + 305419896], cl}
     */
    // Template#: 5011, Serial#: 1255
    public void shrl___CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0193((byte) 0xD3, (byte) 0x05, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shrw      ebx[esi * 4 + 305419896], cl}
     */
    // Template#: 5012, Serial#: 1318
    public void shrw___CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0301((byte) 0xD3, (byte) 0x05, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrb      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 5013, Serial#: 753
    public void shrb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0xC0, (byte) 0x05, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrl      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 5014, Serial#: 816
    public void shrl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0xC1, (byte) 0x05, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrw      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 5015, Serial#: 879
    public void shrw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0067((byte) 0xC1, (byte) 0x05, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shrb      L1: 0x12345678[esi * 4], 0x1}
     */
    // Template#: 5016, Serial#: 972
    public void m_shrb___1(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD0, (byte) 0x05, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shrl      L1: 0x12345678[esi * 4], 0x1}
     */
    // Template#: 5017, Serial#: 1035
    public void m_shrl___1(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD1, (byte) 0x05, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shrw      L1: 0x12345678[esi * 4], 0x1}
     */
    // Template#: 5018, Serial#: 1098
    public void m_shrw___1(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0302((byte) 0xD1, (byte) 0x05, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shrb      L1: 0x12345678[esi * 4], cl}
     */
    // Template#: 5019, Serial#: 1161
    public void m_shrb___CL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD2, (byte) 0x05, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shrl      L1: 0x12345678[esi * 4], cl}
     */
    // Template#: 5020, Serial#: 1224
    public void m_shrl___CL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0194((byte) 0xD3, (byte) 0x05, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code shrw      L1: 0x12345678[esi * 4], cl}
     */
    // Template#: 5021, Serial#: 1287
    public void m_shrw___CL(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0302((byte) 0xD3, (byte) 0x05, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrb      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 5022, Serial#: 722
    public void m_shrb(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0xC0, (byte) 0x05, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrl      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 5023, Serial#: 785
    public void m_shrl(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0xC1, (byte) 0x05, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrw      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 5024, Serial#: 848
    public void m_shrw(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0077((byte) 0xC1, (byte) 0x05, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code shrb      [ebx + 305419896], 0x1}
     */
    // Template#: 5025, Serial#: 1002
    public void shrb___1(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD0, (byte) 0x05, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code shrl      [ebx + 305419896], 0x1}
     */
    // Template#: 5026, Serial#: 1065
    public void shrl___1(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD1, (byte) 0x05, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code shrw      [ebx + 305419896], 0x1}
     */
    // Template#: 5027, Serial#: 1128
    public void shrw___1(int disp32, IA32IndirectRegister32 destination) {
        assemble0303((byte) 0xD1, (byte) 0x05, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code shrb      [ebx + 305419896], cl}
     */
    // Template#: 5028, Serial#: 1191
    public void shrb___CL(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD2, (byte) 0x05, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code shrl      [ebx + 305419896], cl}
     */
    // Template#: 5029, Serial#: 1254
    public void shrl___CL(int disp32, IA32IndirectRegister32 destination) {
        assemble0195((byte) 0xD3, (byte) 0x05, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code shrw      [ebx + 305419896], cl}
     */
    // Template#: 5030, Serial#: 1317
    public void shrw___CL(int disp32, IA32IndirectRegister32 destination) {
        assemble0303((byte) 0xD3, (byte) 0x05, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrb  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrb      [ebx + 305419896], 0x12}
     */
    // Template#: 5031, Serial#: 752
    public void shrb(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0xC0, (byte) 0x05, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrl  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrl      [ebx + 305419896], 0x12}
     */
    // Template#: 5032, Serial#: 815
    public void shrl(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0xC1, (byte) 0x05, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrw  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrw      [ebx + 305419896], 0x12}
     */
    // Template#: 5033, Serial#: 878
    public void shrw(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0084((byte) 0xC1, (byte) 0x05, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code shrd      ebx[esi * 4 + 18], ax, cl}
     */
    // Template#: 5034, Serial#: 5003
    public void shrd_CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0158((byte) 0xAD, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrd      ebx[esi * 4 + 18], ax, 0x12}
     */
    // Template#: 5035, Serial#: 4985
    public void shrd(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source, byte imm8) {
        assemble0506((byte) 0xAC, disp8, base, index, scale, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code shrd      ebx[esi * 4 + 18], eax, cl}
     */
    // Template#: 5036, Serial#: 4994
    public void shrd_CL(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0159((byte) 0xAD, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrd      ebx[esi * 4 + 18], eax, 0x12}
     */
    // Template#: 5037, Serial#: 4976
    public void shrd(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source, byte imm8) {
        assemble0507((byte) 0xAC, disp8, base, index, scale, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code shrd      [ebx + 18], ax, cl}
     */
    // Template#: 5038, Serial#: 5002
    public void shrd_CL(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0161((byte) 0xAD, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>disp8</i>, <i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrd      [ebx + 18], ax, 0x12}
     */
    // Template#: 5039, Serial#: 4984
    public void shrd(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source, byte imm8) {
        assemble0508((byte) 0xAC, disp8, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code shrd      [ebx + 18], eax, cl}
     */
    // Template#: 5040, Serial#: 4993
    public void shrd_CL(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0162((byte) 0xAD, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>disp8</i>, <i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrd      [ebx + 18], eax, 0x12}
     */
    // Template#: 5041, Serial#: 4975
    public void shrd(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source, byte imm8) {
        assemble0509((byte) 0xAC, disp8, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code shrd      ebx[esi * 4], ax, cl}
     */
    // Template#: 5042, Serial#: 4999
    public void shrd_CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0164((byte) 0xAD, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrd      ebx[esi * 4], ax, 0x12}
     */
    // Template#: 5043, Serial#: 4981
    public void shrd(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source, byte imm8) {
        assemble0510((byte) 0xAC, base, index, scale, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code shrd      ebx[esi * 4], eax, cl}
     */
    // Template#: 5044, Serial#: 4990
    public void shrd_CL(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0165((byte) 0xAD, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrd      ebx[esi * 4], eax, 0x12}
     */
    // Template#: 5045, Serial#: 4972
    public void shrd(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source, byte imm8) {
        assemble0511((byte) 0xAC, base, index, scale, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code shrd      ax, ax, cl}
     */
    // Template#: 5046, Serial#: 5006
    public void shrd_CL(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0167((byte) 0xAD, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrd      ax, ax, 0x12}
     */
    // Template#: 5047, Serial#: 4988
    public void shrd(IA32GeneralRegister16 destination, IA32GeneralRegister16 source, byte imm8) {
        assemble0512((byte) 0xAC, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code shrd      eax, eax, cl}
     */
    // Template#: 5048, Serial#: 4997
    public void shrd_CL(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0169((byte) 0xAD, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrd      eax, eax, 0x12}
     */
    // Template#: 5049, Serial#: 4979
    public void shrd(IA32GeneralRegister32 destination, IA32GeneralRegister32 source, byte imm8) {
        assemble0513((byte) 0xAC, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code shrd      [ebx], ax, cl}
     */
    // Template#: 5050, Serial#: 4998
    public void shrd_CL(IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0171((byte) 0xAD, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrd      [ebx], ax, 0x12}
     */
    // Template#: 5051, Serial#: 4980
    public void shrd(IA32IndirectRegister32 destination, IA32GeneralRegister16 source, byte imm8) {
        assemble0514((byte) 0xAC, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code shrd      [ebx], eax, cl}
     */
    // Template#: 5052, Serial#: 4989
    public void shrd_CL(IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0172((byte) 0xAD, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrd      [ebx], eax, 0x12}
     */
    // Template#: 5053, Serial#: 4971
    public void shrd(IA32IndirectRegister32 destination, IA32GeneralRegister32 source, byte imm8) {
        assemble0515((byte) 0xAC, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code shrd      ebx[esi * 4 + 305419896], ax, cl}
     */
    // Template#: 5054, Serial#: 5005
    public void shrd_CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0175((byte) 0xAD, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrd      ebx[esi * 4 + 305419896], ax, 0x12}
     */
    // Template#: 5055, Serial#: 4987
    public void shrd(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source, byte imm8) {
        assemble0516((byte) 0xAC, disp32, base, index, scale, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code shrd      ebx[esi * 4 + 305419896], eax, cl}
     */
    // Template#: 5056, Serial#: 4996
    public void shrd_CL(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0176((byte) 0xAD, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrd      ebx[esi * 4 + 305419896], eax, 0x12}
     */
    // Template#: 5057, Serial#: 4978
    public void shrd(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source, byte imm8) {
        assemble0517((byte) 0xAC, disp32, base, index, scale, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code shrd      [L1: 0x12345678], ax, cl}
     */
    // Template#: 5058, Serial#: 5001
    public void m_shrd_CL(int m32, IA32GeneralRegister16 source) {
        assemble0177((byte) 0xAD, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>m32</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrd      [L1: 0x12345678], ax, 0x12}
     */
    // Template#: 5059, Serial#: 4983
    public void m_shrd(int m32, IA32GeneralRegister16 source, byte imm8) {
        assemble0518((byte) 0xAC, m32, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code shrd      [L1: 0x12345678], eax, cl}
     */
    // Template#: 5060, Serial#: 4992
    public void m_shrd_CL(int m32, IA32GeneralRegister32 source) {
        assemble0178((byte) 0xAD, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>m32</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrd      [L1: 0x12345678], eax, 0x12}
     */
    // Template#: 5061, Serial#: 4974
    public void m_shrd(int m32, IA32GeneralRegister32 source, byte imm8) {
        assemble0519((byte) 0xAC, m32, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code shrd      L1: 0x12345678[esi * 4], ax, cl}
     */
    // Template#: 5062, Serial#: 5000
    public void m_shrd_CL(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0180((byte) 0xAD, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrd      L1: 0x12345678[esi * 4], ax, 0x12}
     */
    // Template#: 5063, Serial#: 4982
    public void m_shrd(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source, byte imm8) {
        assemble0520((byte) 0xAC, m32, index, scale, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code shrd      L1: 0x12345678[esi * 4], eax, cl}
     */
    // Template#: 5064, Serial#: 4991
    public void m_shrd_CL(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0181((byte) 0xAD, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrd      L1: 0x12345678[esi * 4], eax, 0x12}
     */
    // Template#: 5065, Serial#: 4973
    public void m_shrd(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source, byte imm8) {
        assemble0521((byte) 0xAC, m32, index, scale, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code shrd      [ebx + 305419896], ax, cl}
     */
    // Template#: 5066, Serial#: 5004
    public void shrd_CL(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0183((byte) 0xAD, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>disp32</i>, <i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrd      [ebx + 305419896], ax, 0x12}
     */
    // Template#: 5067, Serial#: 4986
    public void shrd(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source, byte imm8) {
        assemble0522((byte) 0xAC, disp32, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code shrd      [ebx + 305419896], eax, cl}
     */
    // Template#: 5068, Serial#: 4995
    public void shrd_CL(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0184((byte) 0xAD, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code shrd  }<i>disp32</i>, <i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shrd      [ebx + 305419896], eax, 0x12}
     */
    // Template#: 5069, Serial#: 4977
    public void shrd(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source, byte imm8) {
        assemble0523((byte) 0xAC, disp32, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shufpd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shufpd    xmm0, ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 5070, Serial#: 3690
    public void shufpd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0467((byte) 0xC6, destination, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shufpd  }<i>destination</i>, <i>disp8</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shufpd    xmm0, [ebx + 18], 0x12}
     */
    // Template#: 5071, Serial#: 3689
    public void shufpd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source, byte imm8) {
        assemble0468((byte) 0xC6, destination, disp8, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shufpd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shufpd    xmm0, ebx[esi * 4], 0x12}
     */
    // Template#: 5072, Serial#: 3686
    public void shufpd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0469((byte) 0xC6, destination, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shufpd  }<i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shufpd    xmm0, [ebx], 0x12}
     */
    // Template#: 5073, Serial#: 3685
    public void shufpd(IA32XMMRegister destination, IA32IndirectRegister32 source, byte imm8) {
        assemble0471((byte) 0xC6, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shufpd  }<i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shufpd    xmm0, xmm0, 0x12}
     */
    // Template#: 5074, Serial#: 3693
    public void shufpd(IA32XMMRegister destination, IA32XMMRegister source, byte imm8) {
        assemble0524((byte) 0xC6, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shufpd  }<i>destination</i>, <i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shufpd    xmm0, [L1: 0x12345678], 0x12}
     */
    // Template#: 5075, Serial#: 3688
    public void m_shufpd(IA32XMMRegister destination, int m32, byte imm8) {
        assemble0472((byte) 0xC6, destination, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shufpd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shufpd    xmm0, ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 5076, Serial#: 3692
    public void shufpd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0473((byte) 0xC6, destination, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shufpd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shufpd    xmm0, L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 5077, Serial#: 3687
    public void m_shufpd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0474((byte) 0xC6, destination, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shufpd  }<i>destination</i>, <i>disp32</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shufpd    xmm0, [ebx + 305419896], 0x12}
     */
    // Template#: 5078, Serial#: 3691
    public void shufpd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source, byte imm8) {
        assemble0475((byte) 0xC6, destination, disp32, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shufps  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shufps    xmm0, ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 5079, Serial#: 3654
    public void shufps(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0525((byte) 0xC6, destination, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shufps  }<i>destination</i>, <i>disp8</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shufps    xmm0, [ebx + 18], 0x12}
     */
    // Template#: 5080, Serial#: 3653
    public void shufps(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source, byte imm8) {
        assemble0526((byte) 0xC6, destination, disp8, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shufps  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shufps    xmm0, ebx[esi * 4], 0x12}
     */
    // Template#: 5081, Serial#: 3650
    public void shufps(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0527((byte) 0xC6, destination, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shufps  }<i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shufps    xmm0, [ebx], 0x12}
     */
    // Template#: 5082, Serial#: 3649
    public void shufps(IA32XMMRegister destination, IA32IndirectRegister32 source, byte imm8) {
        assemble0528((byte) 0xC6, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shufps  }<i>destination</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shufps    xmm0, xmm0, 0x12}
     */
    // Template#: 5083, Serial#: 3657
    public void shufps(IA32XMMRegister destination, IA32XMMRegister source, byte imm8) {
        assemble0529((byte) 0xC6, destination, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shufps  }<i>destination</i>, <i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shufps    xmm0, [L1: 0x12345678], 0x12}
     */
    // Template#: 5084, Serial#: 3652
    public void m_shufps(IA32XMMRegister destination, int m32, byte imm8) {
        assemble0530((byte) 0xC6, destination, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shufps  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shufps    xmm0, ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 5085, Serial#: 3656
    public void shufps(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0531((byte) 0xC6, destination, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shufps  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shufps    xmm0, L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 5086, Serial#: 3651
    public void m_shufps(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0532((byte) 0xC6, destination, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code shufps  }<i>destination</i>, <i>disp32</i>, <i>source</i>, <i>imm8</i>
     * Example disassembly syntax: {@code shufps    xmm0, [ebx + 305419896], 0x12}
     */
    // Template#: 5087, Serial#: 3655
    public void shufps(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source, byte imm8) {
        assemble0533((byte) 0xC6, destination, disp32, source, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sidt  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sidt      ebx[esi * 4 + 18]}
     */
    // Template#: 5088, Serial#: 2583
    public void sidt(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0244((byte) 0x01, (byte) 0x01, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sidt  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code sidt      [ebx + 18]}
     */
    // Template#: 5089, Serial#: 2582
    public void sidt(byte disp8, IA32IndirectRegister32 destination) {
        assemble0245((byte) 0x01, (byte) 0x01, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sidt  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sidt      ebx[esi * 4]}
     */
    // Template#: 5090, Serial#: 2557
    public void sidt(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0246((byte) 0x01, (byte) 0x01, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sidt  }<i>destination</i>
     * Example disassembly syntax: {@code sidt      [ebx]}
     */
    // Template#: 5091, Serial#: 2556
    public void sidt(IA32IndirectRegister32 destination) {
        assemble0247((byte) 0x01, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sidt  }<i>m32</i>
     * Example disassembly syntax: {@code sidt      [L1: 0x12345678]}
     */
    // Template#: 5092, Serial#: 2559
    public void m_sidt(int m32) {
        assemble0248((byte) 0x01, (byte) 0x01, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sidt  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sidt      ebx[esi * 4 + 305419896]}
     */
    // Template#: 5093, Serial#: 2597
    public void sidt(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0249((byte) 0x01, (byte) 0x01, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sidt  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sidt      L1: 0x12345678[esi * 4]}
     */
    // Template#: 5094, Serial#: 2558
    public void m_sidt(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0250((byte) 0x01, (byte) 0x01, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sidt  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code sidt      [ebx + 305419896]}
     */
    // Template#: 5095, Serial#: 2596
    public void sidt(int disp32, IA32IndirectRegister32 destination) {
        assemble0251((byte) 0x01, (byte) 0x01, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sldt  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sldt      ebx[esi * 4 + 18]}
     */
    // Template#: 5096, Serial#: 2521
    public void sldt(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0244((byte) 0x00, (byte) 0x00, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sldt  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code sldt      [ebx + 18]}
     */
    // Template#: 5097, Serial#: 2520
    public void sldt(byte disp8, IA32IndirectRegister32 destination) {
        assemble0245((byte) 0x00, (byte) 0x00, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sldt  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sldt      ebx[esi * 4]}
     */
    // Template#: 5098, Serial#: 2497
    public void sldt(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0246((byte) 0x00, (byte) 0x00, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sldt  }<i>destination</i>
     * Example disassembly syntax: {@code sldt      ax}
     */
    // Template#: 5099, Serial#: 2550
    public void sldt(IA32GeneralRegister16 destination) {
        assemble0534((byte) 0x00, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sldt  }<i>destination</i>
     * Example disassembly syntax: {@code sldt      eax}
     */
    // Template#: 5100, Serial#: 2548
    public void sldt(IA32GeneralRegister32 destination) {
        assemble0535((byte) 0x00, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sldt  }<i>destination</i>
     * Example disassembly syntax: {@code sldt      [ebx]}
     */
    // Template#: 5101, Serial#: 2496
    public void sldt(IA32IndirectRegister32 destination) {
        assemble0247((byte) 0x00, (byte) 0x00, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sldt  }<i>m32</i>
     * Example disassembly syntax: {@code sldt      [L1: 0x12345678]}
     */
    // Template#: 5102, Serial#: 2499
    public void m_sldt(int m32) {
        assemble0248((byte) 0x00, (byte) 0x00, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sldt  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sldt      ebx[esi * 4 + 305419896]}
     */
    // Template#: 5103, Serial#: 2533
    public void sldt(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0249((byte) 0x00, (byte) 0x00, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sldt  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sldt      L1: 0x12345678[esi * 4]}
     */
    // Template#: 5104, Serial#: 2498
    public void m_sldt(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0250((byte) 0x00, (byte) 0x00, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sldt  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code sldt      [ebx + 305419896]}
     */
    // Template#: 5105, Serial#: 2532
    public void sldt(int disp32, IA32IndirectRegister32 destination) {
        assemble0251((byte) 0x00, (byte) 0x00, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code smsw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code smsw      ebx[esi * 4 + 18]}
     */
    // Template#: 5106, Serial#: 2589
    public void smsw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0244((byte) 0x01, (byte) 0x04, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code smsw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code smsw      [ebx + 18]}
     */
    // Template#: 5107, Serial#: 2588
    public void smsw(byte disp8, IA32IndirectRegister32 destination) {
        assemble0245((byte) 0x01, (byte) 0x04, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code smsw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code smsw      ebx[esi * 4]}
     */
    // Template#: 5108, Serial#: 2569
    public void smsw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0246((byte) 0x01, (byte) 0x04, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code smsw  }<i>destination</i>
     * Example disassembly syntax: {@code smsw      ax}
     */
    // Template#: 5109, Serial#: 2610
    public void smsw(IA32GeneralRegister16 destination) {
        assemble0534((byte) 0x01, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code smsw  }<i>destination</i>
     * Example disassembly syntax: {@code smsw      eax}
     */
    // Template#: 5110, Serial#: 2609
    public void smsw(IA32GeneralRegister32 destination) {
        assemble0535((byte) 0x01, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code smsw  }<i>destination</i>
     * Example disassembly syntax: {@code smsw      [ebx]}
     */
    // Template#: 5111, Serial#: 2568
    public void smsw(IA32IndirectRegister32 destination) {
        assemble0247((byte) 0x01, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code smsw  }<i>m32</i>
     * Example disassembly syntax: {@code smsw      [L1: 0x12345678]}
     */
    // Template#: 5112, Serial#: 2571
    public void m_smsw(int m32) {
        assemble0248((byte) 0x01, (byte) 0x04, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code smsw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code smsw      ebx[esi * 4 + 305419896]}
     */
    // Template#: 5113, Serial#: 2603
    public void smsw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0249((byte) 0x01, (byte) 0x04, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code smsw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code smsw      L1: 0x12345678[esi * 4]}
     */
    // Template#: 5114, Serial#: 2570
    public void m_smsw(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0250((byte) 0x01, (byte) 0x04, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code smsw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code smsw      [ebx + 305419896]}
     */
    // Template#: 5115, Serial#: 2602
    public void smsw(int disp32, IA32IndirectRegister32 destination) {
        assemble0251((byte) 0x01, (byte) 0x04, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtpd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sqrtpd    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 5116, Serial#: 3051
    public void sqrtpd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x51, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtpd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code sqrtpd    xmm0, [ebx + 18]}
     */
    // Template#: 5117, Serial#: 3050
    public void sqrtpd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x51, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtpd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sqrtpd    xmm0, ebx[esi * 4]}
     */
    // Template#: 5118, Serial#: 3047
    public void sqrtpd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x51, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sqrtpd    xmm0, [ebx]}
     */
    // Template#: 5119, Serial#: 3046
    public void sqrtpd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x51, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sqrtpd    xmm0, xmm0}
     */
    // Template#: 5120, Serial#: 3054
    public void sqrtpd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x51, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtpd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code sqrtpd    xmm0, [L1: 0x12345678]}
     */
    // Template#: 5121, Serial#: 3049
    public void m_sqrtpd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x51, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtpd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sqrtpd    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 5122, Serial#: 3053
    public void sqrtpd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x51, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtpd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sqrtpd    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 5123, Serial#: 3048
    public void m_sqrtpd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x51, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtpd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code sqrtpd    xmm0, [ebx + 305419896]}
     */
    // Template#: 5124, Serial#: 3052
    public void sqrtpd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x51, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtps  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sqrtps    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 5125, Serial#: 2987
    public void sqrtps(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0102((byte) 0x51, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtps  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code sqrtps    xmm0, [ebx + 18]}
     */
    // Template#: 5126, Serial#: 2986
    public void sqrtps(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0103((byte) 0x51, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtps  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sqrtps    xmm0, ebx[esi * 4]}
     */
    // Template#: 5127, Serial#: 2983
    public void sqrtps(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0104((byte) 0x51, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sqrtps    xmm0, [ebx]}
     */
    // Template#: 5128, Serial#: 2982
    public void sqrtps(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0105((byte) 0x51, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sqrtps    xmm0, xmm0}
     */
    // Template#: 5129, Serial#: 2990
    public void sqrtps(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0106((byte) 0x51, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtps  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code sqrtps    xmm0, [L1: 0x12345678]}
     */
    // Template#: 5130, Serial#: 2985
    public void m_sqrtps(IA32XMMRegister destination, int m32) {
        assemble0107((byte) 0x51, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtps  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sqrtps    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 5131, Serial#: 2989
    public void sqrtps(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0108((byte) 0x51, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtps  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sqrtps    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 5132, Serial#: 2984
    public void m_sqrtps(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0109((byte) 0x51, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtps  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code sqrtps    xmm0, [ebx + 305419896]}
     */
    // Template#: 5133, Serial#: 2988
    public void sqrtps(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0110((byte) 0x51, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtsd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sqrtsd    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 5134, Serial#: 3096
    public void sqrtsd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0111((byte) 0x51, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtsd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code sqrtsd    xmm0, [ebx + 18]}
     */
    // Template#: 5135, Serial#: 3095
    public void sqrtsd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0112((byte) 0x51, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtsd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sqrtsd    xmm0, ebx[esi * 4]}
     */
    // Template#: 5136, Serial#: 3092
    public void sqrtsd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0113((byte) 0x51, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtsd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sqrtsd    xmm0, [ebx]}
     */
    // Template#: 5137, Serial#: 3091
    public void sqrtsd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0114((byte) 0x51, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtsd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sqrtsd    xmm0, xmm0}
     */
    // Template#: 5138, Serial#: 3099
    public void sqrtsd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0115((byte) 0x51, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtsd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code sqrtsd    xmm0, [L1: 0x12345678]}
     */
    // Template#: 5139, Serial#: 3094
    public void m_sqrtsd(IA32XMMRegister destination, int m32) {
        assemble0116((byte) 0x51, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtsd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sqrtsd    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 5140, Serial#: 3098
    public void sqrtsd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0117((byte) 0x51, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtsd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sqrtsd    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 5141, Serial#: 3093
    public void m_sqrtsd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0118((byte) 0x51, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtsd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code sqrtsd    xmm0, [ebx + 305419896]}
     */
    // Template#: 5142, Serial#: 3097
    public void sqrtsd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0119((byte) 0x51, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtss  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sqrtss    xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 5143, Serial#: 3105
    public void sqrtss(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0120((byte) 0x51, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtss  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code sqrtss    xmm0, [ebx + 18]}
     */
    // Template#: 5144, Serial#: 3104
    public void sqrtss(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0121((byte) 0x51, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtss  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sqrtss    xmm0, ebx[esi * 4]}
     */
    // Template#: 5145, Serial#: 3101
    public void sqrtss(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0122((byte) 0x51, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sqrtss    xmm0, [ebx]}
     */
    // Template#: 5146, Serial#: 3100
    public void sqrtss(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0123((byte) 0x51, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sqrtss    xmm0, xmm0}
     */
    // Template#: 5147, Serial#: 3108
    public void sqrtss(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0124((byte) 0x51, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtss  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code sqrtss    xmm0, [L1: 0x12345678]}
     */
    // Template#: 5148, Serial#: 3103
    public void m_sqrtss(IA32XMMRegister destination, int m32) {
        assemble0125((byte) 0x51, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtss  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sqrtss    xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 5149, Serial#: 3107
    public void sqrtss(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0126((byte) 0x51, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtss  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sqrtss    xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 5150, Serial#: 3102
    public void m_sqrtss(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0127((byte) 0x51, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sqrtss  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code sqrtss    xmm0, [ebx + 305419896]}
     */
    // Template#: 5151, Serial#: 3106
    public void sqrtss(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0128((byte) 0x51, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code stc  }
     * Example disassembly syntax: {@code stc     }
     */
    // Template#: 5152, Serial#: 2411
    public void stc() {
        assemble0001((byte) 0xF9);
    }

    /**
     * Pseudo-external assembler syntax: {@code std  }
     * Example disassembly syntax: {@code std     }
     */
    // Template#: 5153, Serial#: 2415
    public void std() {
        assemble0001((byte) 0xFD);
    }

    /**
     * Pseudo-external assembler syntax: {@code sti  }
     * Example disassembly syntax: {@code sti     }
     */
    // Template#: 5154, Serial#: 2413
    public void sti() {
        assemble0001((byte) 0xFB);
    }

    /**
     * Pseudo-external assembler syntax: {@code stosb  }
     * Example disassembly syntax: {@code stosb   }
     */
    // Template#: 5155, Serial#: 1924
    public void stosb() {
        assemble0001((byte) 0xAA);
    }

    /**
     * Pseudo-external assembler syntax: {@code stosl  }
     * Example disassembly syntax: {@code stosl   }
     */
    // Template#: 5156, Serial#: 1925
    public void stosl() {
        assemble0001((byte) 0xAB);
    }

    /**
     * Pseudo-external assembler syntax: {@code stosw  }
     * Example disassembly syntax: {@code stosw   }
     */
    // Template#: 5157, Serial#: 1926
    public void stosw() {
        assemble0197((byte) 0xAB);
    }

    /**
     * Pseudo-external assembler syntax: {@code str  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code str       ebx[esi * 4 + 18]}
     */
    // Template#: 5158, Serial#: 2523
    public void str(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0244((byte) 0x00, (byte) 0x01, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code str  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code str       [ebx + 18]}
     */
    // Template#: 5159, Serial#: 2522
    public void str(byte disp8, IA32IndirectRegister32 destination) {
        assemble0245((byte) 0x00, (byte) 0x01, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code str  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code str       ebx[esi * 4]}
     */
    // Template#: 5160, Serial#: 2501
    public void str(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0246((byte) 0x00, (byte) 0x01, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code str  }<i>destination</i>
     * Example disassembly syntax: {@code str       ax}
     */
    // Template#: 5161, Serial#: 2551
    public void str(IA32GeneralRegister16 destination) {
        assemble0534((byte) 0x00, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code str  }<i>destination</i>
     * Example disassembly syntax: {@code str       eax}
     */
    // Template#: 5162, Serial#: 2549
    public void str(IA32GeneralRegister32 destination) {
        assemble0535((byte) 0x00, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code str  }<i>destination</i>
     * Example disassembly syntax: {@code str       [ebx]}
     */
    // Template#: 5163, Serial#: 2500
    public void str(IA32IndirectRegister32 destination) {
        assemble0247((byte) 0x00, (byte) 0x01, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code str  }<i>m32</i>
     * Example disassembly syntax: {@code str       [L1: 0x12345678]}
     */
    // Template#: 5164, Serial#: 2503
    public void m_str(int m32) {
        assemble0248((byte) 0x00, (byte) 0x01, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code str  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code str       ebx[esi * 4 + 305419896]}
     */
    // Template#: 5165, Serial#: 2535
    public void str(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0249((byte) 0x00, (byte) 0x01, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code str  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code str       L1: 0x12345678[esi * 4]}
     */
    // Template#: 5166, Serial#: 2502
    public void m_str(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0250((byte) 0x00, (byte) 0x01, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code str  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code str       [ebx + 305419896]}
     */
    // Template#: 5167, Serial#: 2534
    public void str(int disp32, IA32IndirectRegister32 destination) {
        assemble0251((byte) 0x00, (byte) 0x01, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>imm8</i>
     * Example disassembly syntax: {@code sub       al, 0x12}
     */
    // Template#: 5168, Serial#: 1698
    public void sub_AL(byte imm8) {
        assemble0002((byte) 0x2C, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code subb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subb      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 5169, Serial#: 313
    public void subb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0x80, (byte) 0x05, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code subl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subl      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 5170, Serial#: 529
    public void subl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0x83, (byte) 0x05, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code subw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subw      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 5171, Serial#: 601
    public void subw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0004((byte) 0x83, (byte) 0x05, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       ebx[esi * 4 + 18], ax}
     */
    // Template#: 5172, Serial#: 1670
    public void sub(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0005((byte) 0x29, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       ebx[esi * 4 + 18], eax}
     */
    // Template#: 5173, Serial#: 1661
    public void sub(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0006((byte) 0x29, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       ebx[esi * 4 + 18], al}
     */
    // Template#: 5174, Serial#: 1652
    public void sub(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0007((byte) 0x28, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code subl      ebx[esi * 4 + 18], 0x12345678}
     */
    // Template#: 5175, Serial#: 385
    public void subl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0008((byte) 0x81, (byte) 0x05, disp8, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code subw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code subw      ebx[esi * 4 + 18], 0x1234}
     */
    // Template#: 5176, Serial#: 457
    public void subw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0009((byte) 0x81, (byte) 0x05, disp8, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code subb  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subb      [ebx + 18], 0x12}
     */
    // Template#: 5177, Serial#: 312
    public void subb(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0x80, (byte) 0x05, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code subl  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subl      [ebx + 18], 0x12}
     */
    // Template#: 5178, Serial#: 528
    public void subl(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0x83, (byte) 0x05, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code subw  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subw      [ebx + 18], 0x12}
     */
    // Template#: 5179, Serial#: 600
    public void subw(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0011((byte) 0x83, (byte) 0x05, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       [ebx + 18], ax}
     */
    // Template#: 5180, Serial#: 1669
    public void sub(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0012((byte) 0x29, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       [ebx + 18], eax}
     */
    // Template#: 5181, Serial#: 1660
    public void sub(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0013((byte) 0x29, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       [ebx + 18], al}
     */
    // Template#: 5182, Serial#: 1651
    public void sub(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0014((byte) 0x28, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subl  }<i>disp8</i>, <i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code subl      [ebx + 18], 0x12345678}
     */
    // Template#: 5183, Serial#: 384
    public void subl(byte disp8, IA32IndirectRegister32 destination, int imm32) {
        assemble0015((byte) 0x81, (byte) 0x05, disp8, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code subw  }<i>disp8</i>, <i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code subw      [ebx + 18], 0x1234}
     */
    // Template#: 5184, Serial#: 456
    public void subw(byte disp8, IA32IndirectRegister32 destination, short imm16) {
        assemble0016((byte) 0x81, (byte) 0x05, disp8, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code subb  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subb      ebx[esi * 4], 0x12}
     */
    // Template#: 5185, Serial#: 291
    public void subb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0x80, (byte) 0x05, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code subl  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subl      ebx[esi * 4], 0x12}
     */
    // Template#: 5186, Serial#: 507
    public void subl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0x83, (byte) 0x05, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code subw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subw      ebx[esi * 4], 0x12}
     */
    // Template#: 5187, Serial#: 579
    public void subw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0018((byte) 0x83, (byte) 0x05, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       ebx[esi * 4], ax}
     */
    // Template#: 5188, Serial#: 1666
    public void sub(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0019((byte) 0x29, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       ebx[esi * 4], eax}
     */
    // Template#: 5189, Serial#: 1657
    public void sub(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0020((byte) 0x29, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       ebx[esi * 4], al}
     */
    // Template#: 5190, Serial#: 1648
    public void sub(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0021((byte) 0x28, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subl  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code subl      ebx[esi * 4], 0x12345678}
     */
    // Template#: 5191, Serial#: 363
    public void subl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0022((byte) 0x81, (byte) 0x05, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code subw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code subw      ebx[esi * 4], 0x1234}
     */
    // Template#: 5192, Serial#: 435
    public void subw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0023((byte) 0x81, (byte) 0x05, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code subw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subw      ax, 0x12}
     */
    // Template#: 5193, Serial#: 627
    public void subw(IA32GeneralRegister16 destination, byte imm8) {
        assemble0024((byte) 0x83, (byte) 0x05, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sub       ax, ebx[esi * 4 + 18]}
     */
    // Template#: 5194, Serial#: 1695
    public void sub(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0025((byte) 0x2B, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       ax, [ebx + 18]}
     */
    // Template#: 5195, Serial#: 1694
    public void sub(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0026((byte) 0x2B, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sub       ax, ebx[esi * 4]}
     */
    // Template#: 5196, Serial#: 1691
    public void sub(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0027((byte) 0x2B, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       ax, ax}
     */
    // Template#: 5197, Serial#: 1673
    public void sub(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0028((byte) 0x29, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       ax, [ebx]}
     */
    // Template#: 5198, Serial#: 1690
    public void sub(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0029((byte) 0x2B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code sub       ax, [L1: 0x12345678]}
     */
    // Template#: 5199, Serial#: 1693
    public void m_sub(IA32GeneralRegister16 destination, int m32) {
        assemble0030((byte) 0x2B, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sub       ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 5200, Serial#: 1697
    public void sub(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0031((byte) 0x2B, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sub       ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 5201, Serial#: 1692
    public void m_sub(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0032((byte) 0x2B, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       ax, [ebx + 305419896]}
     */
    // Template#: 5202, Serial#: 1696
    public void sub(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0033((byte) 0x2B, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subw  }<i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code subw      ax, 0x1234}
     */
    // Template#: 5203, Serial#: 483
    public void subw(IA32GeneralRegister16 destination, short imm16) {
        assemble0034((byte) 0x81, (byte) 0x05, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code subl  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subl      eax, 0x12}
     */
    // Template#: 5204, Serial#: 555
    public void subl(IA32GeneralRegister32 destination, byte imm8) {
        assemble0035((byte) 0x83, (byte) 0x05, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sub       eax, ebx[esi * 4 + 18]}
     */
    // Template#: 5205, Serial#: 1687
    public void sub(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0036((byte) 0x2B, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       eax, [ebx + 18]}
     */
    // Template#: 5206, Serial#: 1686
    public void sub(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0037((byte) 0x2B, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sub       eax, ebx[esi * 4]}
     */
    // Template#: 5207, Serial#: 1683
    public void sub(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0038((byte) 0x2B, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       eax, eax}
     */
    // Template#: 5208, Serial#: 1664
    public void sub(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0039((byte) 0x29, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       eax, [ebx]}
     */
    // Template#: 5209, Serial#: 1682
    public void sub(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0040((byte) 0x2B, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subl  }<i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code subl      eax, 0x12345678}
     */
    // Template#: 5210, Serial#: 411
    public void subl(IA32GeneralRegister32 destination, int imm32) {
        assemble0042((byte) 0x81, (byte) 0x05, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code sub       eax, [L1: 0x12345678]}
     */
    // Template#: 5211, Serial#: 1685
    public void m_sub(IA32GeneralRegister32 destination, int m32) {
        assemble0041((byte) 0x2B, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sub       eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 5212, Serial#: 1689
    public void sub(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0043((byte) 0x2B, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sub       eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 5213, Serial#: 1684
    public void m_sub(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0044((byte) 0x2B, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       eax, [ebx + 305419896]}
     */
    // Template#: 5214, Serial#: 1688
    public void sub(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0045((byte) 0x2B, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subb      al, 0x12}
     */
    // Template#: 5215, Serial#: 339
    public void subb(IA32GeneralRegister8 destination, byte imm8) {
        assemble0046((byte) 0x80, (byte) 0x05, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sub       al, ebx[esi * 4 + 18]}
     */
    // Template#: 5216, Serial#: 1679
    public void sub(IA32GeneralRegister8 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0047((byte) 0x2A, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       al, [ebx + 18]}
     */
    // Template#: 5217, Serial#: 1678
    public void sub(IA32GeneralRegister8 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0048((byte) 0x2A, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sub       al, ebx[esi * 4]}
     */
    // Template#: 5218, Serial#: 1675
    public void sub(IA32GeneralRegister8 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0049((byte) 0x2A, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       al, al}
     */
    // Template#: 5219, Serial#: 1655
    public void sub(IA32GeneralRegister8 destination, IA32GeneralRegister8 source) {
        assemble0050((byte) 0x28, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       al, [ebx]}
     */
    // Template#: 5220, Serial#: 1674
    public void sub(IA32GeneralRegister8 destination, IA32IndirectRegister32 source) {
        assemble0051((byte) 0x2A, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code sub       al, [L1: 0x12345678]}
     */
    // Template#: 5221, Serial#: 1677
    public void m_sub(IA32GeneralRegister8 destination, int m32) {
        assemble0052((byte) 0x2A, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sub       al, ebx[esi * 4 + 305419896]}
     */
    // Template#: 5222, Serial#: 1681
    public void sub(IA32GeneralRegister8 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0053((byte) 0x2A, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code sub       al, L1: 0x12345678[esi * 4]}
     */
    // Template#: 5223, Serial#: 1676
    public void m_sub(IA32GeneralRegister8 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0054((byte) 0x2A, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       al, [ebx + 305419896]}
     */
    // Template#: 5224, Serial#: 1680
    public void sub(IA32GeneralRegister8 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0055((byte) 0x2A, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subb      [ebx], 0x12}
     */
    // Template#: 5225, Serial#: 290
    public void subb(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0x80, (byte) 0x05, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code subl  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subl      [ebx], 0x12}
     */
    // Template#: 5226, Serial#: 506
    public void subl(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0x83, (byte) 0x05, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code subw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subw      [ebx], 0x12}
     */
    // Template#: 5227, Serial#: 578
    public void subw(IA32IndirectRegister32 destination, byte imm8) {
        assemble0057((byte) 0x83, (byte) 0x05, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       [ebx], ax}
     */
    // Template#: 5228, Serial#: 1665
    public void sub(IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0058((byte) 0x29, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       [ebx], eax}
     */
    // Template#: 5229, Serial#: 1656
    public void sub(IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0059((byte) 0x29, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       [ebx], al}
     */
    // Template#: 5230, Serial#: 1647
    public void sub(IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0060((byte) 0x28, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subl  }<i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code subl      [ebx], 0x12345678}
     */
    // Template#: 5231, Serial#: 362
    public void subl(IA32IndirectRegister32 destination, int imm32) {
        assemble0061((byte) 0x81, (byte) 0x05, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code subw  }<i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code subw      [ebx], 0x1234}
     */
    // Template#: 5232, Serial#: 434
    public void subw(IA32IndirectRegister32 destination, short imm16) {
        assemble0062((byte) 0x81, (byte) 0x05, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code subl  }<i>imm32</i>
     * Example disassembly syntax: {@code subl      eax, 0x12345678}
     */
    // Template#: 5233, Serial#: 1699
    public void sub_EAX(int imm32) {
        assemble0063((byte) 0x2D, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code subb  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subb      [L1: 0x12345678], 0x12}
     */
    // Template#: 5234, Serial#: 293
    public void m_subb(int m32, byte imm8) {
        assemble0064((byte) 0x80, (byte) 0x05, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code subl  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subl      [L1: 0x12345678], 0x12}
     */
    // Template#: 5235, Serial#: 509
    public void m_subl(int m32, byte imm8) {
        assemble0064((byte) 0x83, (byte) 0x05, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code subw  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subw      [L1: 0x12345678], 0x12}
     */
    // Template#: 5236, Serial#: 581
    public void m_subw(int m32, byte imm8) {
        assemble0065((byte) 0x83, (byte) 0x05, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code subb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subb      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 5237, Serial#: 329
    public void subb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0x80, (byte) 0x05, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code subl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subl      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 5238, Serial#: 545
    public void subl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0x83, (byte) 0x05, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code subw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subw      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 5239, Serial#: 617
    public void subw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0067((byte) 0x83, (byte) 0x05, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       ebx[esi * 4 + 305419896], ax}
     */
    // Template#: 5240, Serial#: 1672
    public void sub(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0068((byte) 0x29, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       ebx[esi * 4 + 305419896], eax}
     */
    // Template#: 5241, Serial#: 1663
    public void sub(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0069((byte) 0x29, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       ebx[esi * 4 + 305419896], al}
     */
    // Template#: 5242, Serial#: 1654
    public void sub(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0070((byte) 0x28, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code subl      ebx[esi * 4 + 305419896], 0x12345678}
     */
    // Template#: 5243, Serial#: 401
    public void subl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0071((byte) 0x81, (byte) 0x05, disp32, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code subw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code subw      ebx[esi * 4 + 305419896], 0x1234}
     */
    // Template#: 5244, Serial#: 473
    public void subw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0072((byte) 0x81, (byte) 0x05, disp32, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       [L1: 0x12345678], ax}
     */
    // Template#: 5245, Serial#: 1668
    public void m_sub(int m32, IA32GeneralRegister16 source) {
        assemble0073((byte) 0x29, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       [L1: 0x12345678], eax}
     */
    // Template#: 5246, Serial#: 1659
    public void m_sub(int m32, IA32GeneralRegister32 source) {
        assemble0074((byte) 0x29, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       [L1: 0x12345678], al}
     */
    // Template#: 5247, Serial#: 1650
    public void m_sub(int m32, IA32GeneralRegister8 source) {
        assemble0075((byte) 0x28, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subb  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subb      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 5248, Serial#: 292
    public void m_subb(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0x80, (byte) 0x05, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code subl  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subl      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 5249, Serial#: 508
    public void m_subl(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0x83, (byte) 0x05, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code subw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subw      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 5250, Serial#: 580
    public void m_subw(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0077((byte) 0x83, (byte) 0x05, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       L1: 0x12345678[esi * 4], ax}
     */
    // Template#: 5251, Serial#: 1667
    public void m_sub(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0078((byte) 0x29, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       L1: 0x12345678[esi * 4], eax}
     */
    // Template#: 5252, Serial#: 1658
    public void m_sub(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0079((byte) 0x29, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       L1: 0x12345678[esi * 4], al}
     */
    // Template#: 5253, Serial#: 1649
    public void m_sub(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0080((byte) 0x28, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subl  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code subl      L1: 0x12345678[esi * 4], 0x12345678}
     */
    // Template#: 5254, Serial#: 364
    public void m_subl(int m32, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0081((byte) 0x81, (byte) 0x05, m32, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code subw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code subw      L1: 0x12345678[esi * 4], 0x1234}
     */
    // Template#: 5255, Serial#: 436
    public void m_subw(int m32, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0082((byte) 0x81, (byte) 0x05, m32, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code subb  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subb      [ebx + 305419896], 0x12}
     */
    // Template#: 5256, Serial#: 328
    public void subb(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0x80, (byte) 0x05, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code subl  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subl      [ebx + 305419896], 0x12}
     */
    // Template#: 5257, Serial#: 544
    public void subl(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0x83, (byte) 0x05, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code subw  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code subw      [ebx + 305419896], 0x12}
     */
    // Template#: 5258, Serial#: 616
    public void subw(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0084((byte) 0x83, (byte) 0x05, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       [ebx + 305419896], ax}
     */
    // Template#: 5259, Serial#: 1671
    public void sub(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0085((byte) 0x29, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       [ebx + 305419896], eax}
     */
    // Template#: 5260, Serial#: 1662
    public void sub(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0086((byte) 0x29, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code sub  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code sub       [ebx + 305419896], al}
     */
    // Template#: 5261, Serial#: 1653
    public void sub(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0087((byte) 0x28, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subl  }<i>disp32</i>, <i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code subl      [ebx + 305419896], 0x12345678}
     */
    // Template#: 5262, Serial#: 400
    public void subl(int disp32, IA32IndirectRegister32 destination, int imm32) {
        assemble0088((byte) 0x81, (byte) 0x05, disp32, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code subw  }<i>disp32</i>, <i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code subw      [ebx + 305419896], 0x1234}
     */
    // Template#: 5263, Serial#: 472
    public void subw(int disp32, IA32IndirectRegister32 destination, short imm16) {
        assemble0089((byte) 0x81, (byte) 0x05, disp32, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code subl  }<i>m32</i>, <i>imm32</i>
     * Example disassembly syntax: {@code subl      [L1: 0x12345678], 0x12345678}
     */
    // Template#: 5264, Serial#: 365
    public void m_subl(int m32, int imm32) {
        assemble0090((byte) 0x81, (byte) 0x05, m32, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code subw  }<i>m32</i>, <i>imm16</i>
     * Example disassembly syntax: {@code subw      [L1: 0x12345678], 0x1234}
     */
    // Template#: 5265, Serial#: 437
    public void m_subw(int m32, short imm16) {
        assemble0091((byte) 0x81, (byte) 0x05, m32, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code subw  }<i>imm16</i>
     * Example disassembly syntax: {@code subw      ax, 0x1234}
     */
    // Template#: 5266, Serial#: 1700
    public void sub_AX(short imm16) {
        assemble0092((byte) 0x2D, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code subpd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code subpd     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 5267, Serial#: 4483
    public void subpd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x5C, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code subpd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code subpd     xmm0, [ebx + 18]}
     */
    // Template#: 5268, Serial#: 4482
    public void subpd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x5C, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subpd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code subpd     xmm0, ebx[esi * 4]}
     */
    // Template#: 5269, Serial#: 4479
    public void subpd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x5C, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code subpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code subpd     xmm0, [ebx]}
     */
    // Template#: 5270, Serial#: 4478
    public void subpd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x5C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code subpd     xmm0, xmm0}
     */
    // Template#: 5271, Serial#: 4486
    public void subpd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x5C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subpd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code subpd     xmm0, [L1: 0x12345678]}
     */
    // Template#: 5272, Serial#: 4481
    public void m_subpd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x5C, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code subpd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code subpd     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 5273, Serial#: 4485
    public void subpd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x5C, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code subpd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code subpd     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 5274, Serial#: 4480
    public void m_subpd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x5C, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code subpd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code subpd     xmm0, [ebx + 305419896]}
     */
    // Template#: 5275, Serial#: 4484
    public void subpd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x5C, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subps  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code subps     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 5276, Serial#: 4411
    public void subps(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0102((byte) 0x5C, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code subps  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code subps     xmm0, [ebx + 18]}
     */
    // Template#: 5277, Serial#: 4410
    public void subps(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0103((byte) 0x5C, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subps  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code subps     xmm0, ebx[esi * 4]}
     */
    // Template#: 5278, Serial#: 4407
    public void subps(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0104((byte) 0x5C, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code subps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code subps     xmm0, [ebx]}
     */
    // Template#: 5279, Serial#: 4406
    public void subps(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0105((byte) 0x5C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code subps     xmm0, xmm0}
     */
    // Template#: 5280, Serial#: 4414
    public void subps(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0106((byte) 0x5C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subps  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code subps     xmm0, [L1: 0x12345678]}
     */
    // Template#: 5281, Serial#: 4409
    public void m_subps(IA32XMMRegister destination, int m32) {
        assemble0107((byte) 0x5C, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code subps  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code subps     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 5282, Serial#: 4413
    public void subps(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0108((byte) 0x5C, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code subps  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code subps     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 5283, Serial#: 4408
    public void m_subps(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0109((byte) 0x5C, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code subps  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code subps     xmm0, [ebx + 305419896]}
     */
    // Template#: 5284, Serial#: 4412
    public void subps(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0110((byte) 0x5C, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subsd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code subsd     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 5285, Serial#: 4546
    public void subsd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0111((byte) 0x5C, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code subsd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code subsd     xmm0, [ebx + 18]}
     */
    // Template#: 5286, Serial#: 4545
    public void subsd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0112((byte) 0x5C, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subsd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code subsd     xmm0, ebx[esi * 4]}
     */
    // Template#: 5287, Serial#: 4542
    public void subsd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0113((byte) 0x5C, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code subsd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code subsd     xmm0, [ebx]}
     */
    // Template#: 5288, Serial#: 4541
    public void subsd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0114((byte) 0x5C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subsd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code subsd     xmm0, xmm0}
     */
    // Template#: 5289, Serial#: 4549
    public void subsd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0115((byte) 0x5C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subsd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code subsd     xmm0, [L1: 0x12345678]}
     */
    // Template#: 5290, Serial#: 4544
    public void m_subsd(IA32XMMRegister destination, int m32) {
        assemble0116((byte) 0x5C, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code subsd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code subsd     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 5291, Serial#: 4548
    public void subsd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0117((byte) 0x5C, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code subsd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code subsd     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 5292, Serial#: 4543
    public void m_subsd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0118((byte) 0x5C, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code subsd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code subsd     xmm0, [ebx + 305419896]}
     */
    // Template#: 5293, Serial#: 4547
    public void subsd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0119((byte) 0x5C, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subss  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code subss     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 5294, Serial#: 4618
    public void subss(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0120((byte) 0x5C, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code subss  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code subss     xmm0, [ebx + 18]}
     */
    // Template#: 5295, Serial#: 4617
    public void subss(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0121((byte) 0x5C, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subss  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code subss     xmm0, ebx[esi * 4]}
     */
    // Template#: 5296, Serial#: 4614
    public void subss(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0122((byte) 0x5C, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code subss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code subss     xmm0, [ebx]}
     */
    // Template#: 5297, Serial#: 4613
    public void subss(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0123((byte) 0x5C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code subss     xmm0, xmm0}
     */
    // Template#: 5298, Serial#: 4621
    public void subss(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0124((byte) 0x5C, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code subss  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code subss     xmm0, [L1: 0x12345678]}
     */
    // Template#: 5299, Serial#: 4616
    public void m_subss(IA32XMMRegister destination, int m32) {
        assemble0125((byte) 0x5C, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code subss  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code subss     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 5300, Serial#: 4620
    public void subss(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0126((byte) 0x5C, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code subss  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code subss     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 5301, Serial#: 4615
    public void m_subss(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0127((byte) 0x5C, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code subss  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code subss     xmm0, [ebx + 305419896]}
     */
    // Template#: 5302, Serial#: 4619
    public void subss(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0128((byte) 0x5C, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>imm8</i>
     * Example disassembly syntax: {@code test      al, 0x12}
     */
    // Template#: 5303, Serial#: 1921
    public void test_AL(byte imm8) {
        assemble0002((byte) 0xA8, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code testb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code testb     ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 5304, Serial#: 1376
    public void testb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0xF6, (byte) 0x00, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code test      ebx[esi * 4 + 18], ax}
     */
    // Template#: 5305, Serial#: 653
    public void test(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0005((byte) 0x85, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code test      ebx[esi * 4 + 18], eax}
     */
    // Template#: 5306, Serial#: 644
    public void test(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0006((byte) 0x85, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code test      ebx[esi * 4 + 18], al}
     */
    // Template#: 5307, Serial#: 635
    public void test(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0007((byte) 0x84, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code testl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code testl     ebx[esi * 4 + 18], 0x12345678}
     */
    // Template#: 5308, Serial#: 1439
    public void testl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0008((byte) 0xF7, (byte) 0x00, disp8, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code testw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code testw     ebx[esi * 4 + 18], 0x1234}
     */
    // Template#: 5309, Serial#: 1502
    public void testw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0009((byte) 0xF7, (byte) 0x00, disp8, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code testb  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code testb     [ebx + 18], 0x12}
     */
    // Template#: 5310, Serial#: 1375
    public void testb(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0xF6, (byte) 0x00, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code test      [ebx + 18], ax}
     */
    // Template#: 5311, Serial#: 652
    public void test(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0012((byte) 0x85, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code test      [ebx + 18], eax}
     */
    // Template#: 5312, Serial#: 643
    public void test(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0013((byte) 0x85, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code test      [ebx + 18], al}
     */
    // Template#: 5313, Serial#: 634
    public void test(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0014((byte) 0x84, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code testl  }<i>disp8</i>, <i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code testl     [ebx + 18], 0x12345678}
     */
    // Template#: 5314, Serial#: 1438
    public void testl(byte disp8, IA32IndirectRegister32 destination, int imm32) {
        assemble0015((byte) 0xF7, (byte) 0x00, disp8, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code testw  }<i>disp8</i>, <i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code testw     [ebx + 18], 0x1234}
     */
    // Template#: 5315, Serial#: 1501
    public void testw(byte disp8, IA32IndirectRegister32 destination, short imm16) {
        assemble0016((byte) 0xF7, (byte) 0x00, disp8, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code testb  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code testb     ebx[esi * 4], 0x12}
     */
    // Template#: 5316, Serial#: 1348
    public void testb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0xF6, (byte) 0x00, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code test      ebx[esi * 4], ax}
     */
    // Template#: 5317, Serial#: 649
    public void test(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0019((byte) 0x85, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code test      ebx[esi * 4], eax}
     */
    // Template#: 5318, Serial#: 640
    public void test(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0020((byte) 0x85, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code test      ebx[esi * 4], al}
     */
    // Template#: 5319, Serial#: 631
    public void test(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0021((byte) 0x84, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code testl  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code testl     ebx[esi * 4], 0x12345678}
     */
    // Template#: 5320, Serial#: 1411
    public void testl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0022((byte) 0xF7, (byte) 0x00, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code testw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code testw     ebx[esi * 4], 0x1234}
     */
    // Template#: 5321, Serial#: 1474
    public void testw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0023((byte) 0xF7, (byte) 0x00, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code test      ax, ax}
     */
    // Template#: 5322, Serial#: 656
    public void test(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0028((byte) 0x85, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code testw  }<i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code testw     ax, 0x1234}
     */
    // Template#: 5323, Serial#: 1529
    public void testw(IA32GeneralRegister16 destination, short imm16) {
        assemble0034((byte) 0xF7, (byte) 0x00, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code test      eax, eax}
     */
    // Template#: 5324, Serial#: 647
    public void test(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0039((byte) 0x85, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code testl  }<i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code testl     eax, 0x12345678}
     */
    // Template#: 5325, Serial#: 1466
    public void testl(IA32GeneralRegister32 destination, int imm32) {
        assemble0042((byte) 0xF7, (byte) 0x00, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code testb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code testb     al, 0x12}
     */
    // Template#: 5326, Serial#: 1403
    public void testb(IA32GeneralRegister8 destination, byte imm8) {
        assemble0046((byte) 0xF6, (byte) 0x00, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code test      al, al}
     */
    // Template#: 5327, Serial#: 638
    public void test(IA32GeneralRegister8 destination, IA32GeneralRegister8 source) {
        assemble0050((byte) 0x84, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code testb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code testb     [ebx], 0x12}
     */
    // Template#: 5328, Serial#: 1347
    public void testb(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0xF6, (byte) 0x00, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code test      [ebx], ax}
     */
    // Template#: 5329, Serial#: 648
    public void test(IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0058((byte) 0x85, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code test      [ebx], eax}
     */
    // Template#: 5330, Serial#: 639
    public void test(IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0059((byte) 0x85, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code test      [ebx], al}
     */
    // Template#: 5331, Serial#: 630
    public void test(IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0060((byte) 0x84, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code testl  }<i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code testl     [ebx], 0x12345678}
     */
    // Template#: 5332, Serial#: 1410
    public void testl(IA32IndirectRegister32 destination, int imm32) {
        assemble0061((byte) 0xF7, (byte) 0x00, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code testw  }<i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code testw     [ebx], 0x1234}
     */
    // Template#: 5333, Serial#: 1473
    public void testw(IA32IndirectRegister32 destination, short imm16) {
        assemble0062((byte) 0xF7, (byte) 0x00, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code testl  }<i>imm32</i>
     * Example disassembly syntax: {@code testl     eax, 0x12345678}
     */
    // Template#: 5334, Serial#: 1922
    public void test_EAX(int imm32) {
        assemble0063((byte) 0xA9, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code testb  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code testb     [L1: 0x12345678], 0x12}
     */
    // Template#: 5335, Serial#: 1350
    public void m_testb(int m32, byte imm8) {
        assemble0064((byte) 0xF6, (byte) 0x00, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code testb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code testb     ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 5336, Serial#: 1390
    public void testb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0xF6, (byte) 0x00, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code test      ebx[esi * 4 + 305419896], ax}
     */
    // Template#: 5337, Serial#: 655
    public void test(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0068((byte) 0x85, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code test      ebx[esi * 4 + 305419896], eax}
     */
    // Template#: 5338, Serial#: 646
    public void test(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0069((byte) 0x85, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code test      ebx[esi * 4 + 305419896], al}
     */
    // Template#: 5339, Serial#: 637
    public void test(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0070((byte) 0x84, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code testl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code testl     ebx[esi * 4 + 305419896], 0x12345678}
     */
    // Template#: 5340, Serial#: 1453
    public void testl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0071((byte) 0xF7, (byte) 0x00, disp32, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code testw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code testw     ebx[esi * 4 + 305419896], 0x1234}
     */
    // Template#: 5341, Serial#: 1516
    public void testw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0072((byte) 0xF7, (byte) 0x00, disp32, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code test      [L1: 0x12345678], ax}
     */
    // Template#: 5342, Serial#: 651
    public void m_test(int m32, IA32GeneralRegister16 source) {
        assemble0073((byte) 0x85, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code test      [L1: 0x12345678], eax}
     */
    // Template#: 5343, Serial#: 642
    public void m_test(int m32, IA32GeneralRegister32 source) {
        assemble0074((byte) 0x85, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code test      [L1: 0x12345678], al}
     */
    // Template#: 5344, Serial#: 633
    public void m_test(int m32, IA32GeneralRegister8 source) {
        assemble0075((byte) 0x84, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code testb  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code testb     L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 5345, Serial#: 1349
    public void m_testb(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0xF6, (byte) 0x00, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code test      L1: 0x12345678[esi * 4], ax}
     */
    // Template#: 5346, Serial#: 650
    public void m_test(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0078((byte) 0x85, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code test      L1: 0x12345678[esi * 4], eax}
     */
    // Template#: 5347, Serial#: 641
    public void m_test(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0079((byte) 0x85, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code test      L1: 0x12345678[esi * 4], al}
     */
    // Template#: 5348, Serial#: 632
    public void m_test(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0080((byte) 0x84, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code testl  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code testl     L1: 0x12345678[esi * 4], 0x12345678}
     */
    // Template#: 5349, Serial#: 1412
    public void m_testl(int m32, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0081((byte) 0xF7, (byte) 0x00, m32, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code testw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code testw     L1: 0x12345678[esi * 4], 0x1234}
     */
    // Template#: 5350, Serial#: 1475
    public void m_testw(int m32, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0082((byte) 0xF7, (byte) 0x00, m32, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code testb  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code testb     [ebx + 305419896], 0x12}
     */
    // Template#: 5351, Serial#: 1389
    public void testb(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0xF6, (byte) 0x00, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code test      [ebx + 305419896], ax}
     */
    // Template#: 5352, Serial#: 654
    public void test(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0085((byte) 0x85, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code test      [ebx + 305419896], eax}
     */
    // Template#: 5353, Serial#: 645
    public void test(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0086((byte) 0x85, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code test  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code test      [ebx + 305419896], al}
     */
    // Template#: 5354, Serial#: 636
    public void test(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0087((byte) 0x84, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code testl  }<i>disp32</i>, <i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code testl     [ebx + 305419896], 0x12345678}
     */
    // Template#: 5355, Serial#: 1452
    public void testl(int disp32, IA32IndirectRegister32 destination, int imm32) {
        assemble0088((byte) 0xF7, (byte) 0x00, disp32, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code testw  }<i>disp32</i>, <i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code testw     [ebx + 305419896], 0x1234}
     */
    // Template#: 5356, Serial#: 1515
    public void testw(int disp32, IA32IndirectRegister32 destination, short imm16) {
        assemble0089((byte) 0xF7, (byte) 0x00, disp32, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code testl  }<i>m32</i>, <i>imm32</i>
     * Example disassembly syntax: {@code testl     [L1: 0x12345678], 0x12345678}
     */
    // Template#: 5357, Serial#: 1413
    public void m_testl(int m32, int imm32) {
        assemble0090((byte) 0xF7, (byte) 0x00, m32, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code testw  }<i>m32</i>, <i>imm16</i>
     * Example disassembly syntax: {@code testw     [L1: 0x12345678], 0x1234}
     */
    // Template#: 5358, Serial#: 1476
    public void m_testw(int m32, short imm16) {
        assemble0091((byte) 0xF7, (byte) 0x00, m32, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code testw  }<i>imm16</i>
     * Example disassembly syntax: {@code testw     ax, 0x1234}
     */
    // Template#: 5359, Serial#: 1923
    public void test_AX(short imm16) {
        assemble0092((byte) 0xA9, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code ucomisd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ucomisd   xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 5360, Serial#: 4159
    public void ucomisd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x2E, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ucomisd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code ucomisd   xmm0, [ebx + 18]}
     */
    // Template#: 5361, Serial#: 4158
    public void ucomisd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x2E, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code ucomisd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ucomisd   xmm0, ebx[esi * 4]}
     */
    // Template#: 5362, Serial#: 4155
    public void ucomisd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x2E, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ucomisd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code ucomisd   xmm0, [ebx]}
     */
    // Template#: 5363, Serial#: 4154
    public void ucomisd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x2E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code ucomisd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code ucomisd   xmm0, xmm0}
     */
    // Template#: 5364, Serial#: 4162
    public void ucomisd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x2E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code ucomisd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code ucomisd   xmm0, [L1: 0x12345678]}
     */
    // Template#: 5365, Serial#: 4157
    public void m_ucomisd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x2E, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code ucomisd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ucomisd   xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 5366, Serial#: 4161
    public void ucomisd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x2E, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ucomisd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ucomisd   xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 5367, Serial#: 4156
    public void m_ucomisd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x2E, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ucomisd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code ucomisd   xmm0, [ebx + 305419896]}
     */
    // Template#: 5368, Serial#: 4160
    public void ucomisd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x2E, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code ucomiss  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ucomiss   xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 5369, Serial#: 4089
    public void ucomiss(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0102((byte) 0x2E, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ucomiss  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code ucomiss   xmm0, [ebx + 18]}
     */
    // Template#: 5370, Serial#: 4088
    public void ucomiss(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0103((byte) 0x2E, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code ucomiss  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ucomiss   xmm0, ebx[esi * 4]}
     */
    // Template#: 5371, Serial#: 4085
    public void ucomiss(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0104((byte) 0x2E, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ucomiss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code ucomiss   xmm0, [ebx]}
     */
    // Template#: 5372, Serial#: 4084
    public void ucomiss(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0105((byte) 0x2E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code ucomiss  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code ucomiss   xmm0, xmm0}
     */
    // Template#: 5373, Serial#: 4092
    public void ucomiss(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0106((byte) 0x2E, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code ucomiss  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code ucomiss   xmm0, [L1: 0x12345678]}
     */
    // Template#: 5374, Serial#: 4087
    public void m_ucomiss(IA32XMMRegister destination, int m32) {
        assemble0107((byte) 0x2E, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code ucomiss  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ucomiss   xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 5375, Serial#: 4091
    public void ucomiss(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0108((byte) 0x2E, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ucomiss  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code ucomiss   xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 5376, Serial#: 4086
    public void m_ucomiss(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0109((byte) 0x2E, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code ucomiss  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code ucomiss   xmm0, [ebx + 305419896]}
     */
    // Template#: 5377, Serial#: 4090
    public void ucomiss(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0110((byte) 0x2E, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code ud2  }
     * Example disassembly syntax: {@code ud2     }
     */
    // Template#: 5378, Serial#: 4031
    public void ud2() {
        assemble0198((byte) 0x0B);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpckhpd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code unpckhpd  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 5379, Serial#: 2748
    public void unpckhpd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x15, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpckhpd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code unpckhpd  xmm0, [ebx + 18]}
     */
    // Template#: 5380, Serial#: 2747
    public void unpckhpd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x15, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpckhpd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code unpckhpd  xmm0, ebx[esi * 4]}
     */
    // Template#: 5381, Serial#: 2744
    public void unpckhpd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x15, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpckhpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code unpckhpd  xmm0, [ebx]}
     */
    // Template#: 5382, Serial#: 2743
    public void unpckhpd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x15, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpckhpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code unpckhpd  xmm0, xmm0}
     */
    // Template#: 5383, Serial#: 2751
    public void unpckhpd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x15, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpckhpd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code unpckhpd  xmm0, [L1: 0x12345678]}
     */
    // Template#: 5384, Serial#: 2746
    public void m_unpckhpd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x15, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpckhpd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code unpckhpd  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 5385, Serial#: 2750
    public void unpckhpd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x15, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpckhpd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code unpckhpd  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 5386, Serial#: 2745
    public void m_unpckhpd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x15, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpckhpd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code unpckhpd  xmm0, [ebx + 305419896]}
     */
    // Template#: 5387, Serial#: 2749
    public void unpckhpd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x15, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpckhps  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code unpckhps  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 5388, Serial#: 2688
    public void unpckhps(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0102((byte) 0x15, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpckhps  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code unpckhps  xmm0, [ebx + 18]}
     */
    // Template#: 5389, Serial#: 2687
    public void unpckhps(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0103((byte) 0x15, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpckhps  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code unpckhps  xmm0, ebx[esi * 4]}
     */
    // Template#: 5390, Serial#: 2684
    public void unpckhps(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0104((byte) 0x15, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpckhps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code unpckhps  xmm0, [ebx]}
     */
    // Template#: 5391, Serial#: 2683
    public void unpckhps(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0105((byte) 0x15, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpckhps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code unpckhps  xmm0, xmm0}
     */
    // Template#: 5392, Serial#: 2691
    public void unpckhps(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0106((byte) 0x15, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpckhps  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code unpckhps  xmm0, [L1: 0x12345678]}
     */
    // Template#: 5393, Serial#: 2686
    public void m_unpckhps(IA32XMMRegister destination, int m32) {
        assemble0107((byte) 0x15, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpckhps  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code unpckhps  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 5394, Serial#: 2690
    public void unpckhps(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0108((byte) 0x15, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpckhps  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code unpckhps  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 5395, Serial#: 2685
    public void m_unpckhps(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0109((byte) 0x15, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpckhps  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code unpckhps  xmm0, [ebx + 305419896]}
     */
    // Template#: 5396, Serial#: 2689
    public void unpckhps(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0110((byte) 0x15, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpcklpd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code unpcklpd  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 5397, Serial#: 2739
    public void unpcklpd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x14, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpcklpd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code unpcklpd  xmm0, [ebx + 18]}
     */
    // Template#: 5398, Serial#: 2738
    public void unpcklpd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x14, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpcklpd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code unpcklpd  xmm0, ebx[esi * 4]}
     */
    // Template#: 5399, Serial#: 2735
    public void unpcklpd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x14, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpcklpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code unpcklpd  xmm0, [ebx]}
     */
    // Template#: 5400, Serial#: 2734
    public void unpcklpd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x14, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpcklpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code unpcklpd  xmm0, xmm0}
     */
    // Template#: 5401, Serial#: 2742
    public void unpcklpd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x14, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpcklpd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code unpcklpd  xmm0, [L1: 0x12345678]}
     */
    // Template#: 5402, Serial#: 2737
    public void m_unpcklpd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x14, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpcklpd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code unpcklpd  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 5403, Serial#: 2741
    public void unpcklpd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x14, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpcklpd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code unpcklpd  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 5404, Serial#: 2736
    public void m_unpcklpd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x14, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpcklpd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code unpcklpd  xmm0, [ebx + 305419896]}
     */
    // Template#: 5405, Serial#: 2740
    public void unpcklpd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x14, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpcklps  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code unpcklps  xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 5406, Serial#: 2679
    public void unpcklps(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0102((byte) 0x14, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpcklps  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code unpcklps  xmm0, [ebx + 18]}
     */
    // Template#: 5407, Serial#: 2678
    public void unpcklps(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0103((byte) 0x14, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpcklps  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code unpcklps  xmm0, ebx[esi * 4]}
     */
    // Template#: 5408, Serial#: 2675
    public void unpcklps(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0104((byte) 0x14, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpcklps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code unpcklps  xmm0, [ebx]}
     */
    // Template#: 5409, Serial#: 2674
    public void unpcklps(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0105((byte) 0x14, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpcklps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code unpcklps  xmm0, xmm0}
     */
    // Template#: 5410, Serial#: 2682
    public void unpcklps(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0106((byte) 0x14, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpcklps  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code unpcklps  xmm0, [L1: 0x12345678]}
     */
    // Template#: 5411, Serial#: 2677
    public void m_unpcklps(IA32XMMRegister destination, int m32) {
        assemble0107((byte) 0x14, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpcklps  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code unpcklps  xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 5412, Serial#: 2681
    public void unpcklps(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0108((byte) 0x14, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpcklps  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code unpcklps  xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 5413, Serial#: 2676
    public void m_unpcklps(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0109((byte) 0x14, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code unpcklps  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code unpcklps  xmm0, [ebx + 305419896]}
     */
    // Template#: 5414, Serial#: 2680
    public void unpcklps(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0110((byte) 0x14, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code verr  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code verr      ebx[esi * 4 + 18]}
     */
    // Template#: 5415, Serial#: 2529
    public void verr(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0244((byte) 0x00, (byte) 0x04, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code verr  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code verr      [ebx + 18]}
     */
    // Template#: 5416, Serial#: 2528
    public void verr(byte disp8, IA32IndirectRegister32 destination) {
        assemble0245((byte) 0x00, (byte) 0x04, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code verr  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code verr      ebx[esi * 4]}
     */
    // Template#: 5417, Serial#: 2513
    public void verr(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0246((byte) 0x00, (byte) 0x04, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code verr  }<i>destination</i>
     * Example disassembly syntax: {@code verr      ax}
     */
    // Template#: 5418, Serial#: 2546
    public void verr(IA32GeneralRegister16 destination) {
        assemble0358((byte) 0x00, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code verr  }<i>destination</i>
     * Example disassembly syntax: {@code verr      [ebx]}
     */
    // Template#: 5419, Serial#: 2512
    public void verr(IA32IndirectRegister32 destination) {
        assemble0247((byte) 0x00, (byte) 0x04, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code verr  }<i>m32</i>
     * Example disassembly syntax: {@code verr      [L1: 0x12345678]}
     */
    // Template#: 5420, Serial#: 2515
    public void m_verr(int m32) {
        assemble0248((byte) 0x00, (byte) 0x04, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code verr  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code verr      ebx[esi * 4 + 305419896]}
     */
    // Template#: 5421, Serial#: 2541
    public void verr(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0249((byte) 0x00, (byte) 0x04, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code verr  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code verr      L1: 0x12345678[esi * 4]}
     */
    // Template#: 5422, Serial#: 2514
    public void m_verr(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0250((byte) 0x00, (byte) 0x04, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code verr  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code verr      [ebx + 305419896]}
     */
    // Template#: 5423, Serial#: 2540
    public void verr(int disp32, IA32IndirectRegister32 destination) {
        assemble0251((byte) 0x00, (byte) 0x04, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code verw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code verw      ebx[esi * 4 + 18]}
     */
    // Template#: 5424, Serial#: 2531
    public void verw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0244((byte) 0x00, (byte) 0x05, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code verw  }<i>disp8</i>, <i>destination</i>
     * Example disassembly syntax: {@code verw      [ebx + 18]}
     */
    // Template#: 5425, Serial#: 2530
    public void verw(byte disp8, IA32IndirectRegister32 destination) {
        assemble0245((byte) 0x00, (byte) 0x05, disp8, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code verw  }<i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code verw      ebx[esi * 4]}
     */
    // Template#: 5426, Serial#: 2517
    public void verw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0246((byte) 0x00, (byte) 0x05, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code verw  }<i>destination</i>
     * Example disassembly syntax: {@code verw      ax}
     */
    // Template#: 5427, Serial#: 2547
    public void verw(IA32GeneralRegister16 destination) {
        assemble0358((byte) 0x00, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code verw  }<i>destination</i>
     * Example disassembly syntax: {@code verw      [ebx]}
     */
    // Template#: 5428, Serial#: 2516
    public void verw(IA32IndirectRegister32 destination) {
        assemble0247((byte) 0x00, (byte) 0x05, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code verw  }<i>m32</i>
     * Example disassembly syntax: {@code verw      [L1: 0x12345678]}
     */
    // Template#: 5429, Serial#: 2519
    public void m_verw(int m32) {
        assemble0248((byte) 0x00, (byte) 0x05, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code verw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code verw      ebx[esi * 4 + 305419896]}
     */
    // Template#: 5430, Serial#: 2543
    public void verw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0249((byte) 0x00, (byte) 0x05, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code verw  }<i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code verw      L1: 0x12345678[esi * 4]}
     */
    // Template#: 5431, Serial#: 2518
    public void m_verw(int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0250((byte) 0x00, (byte) 0x05, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code verw  }<i>disp32</i>, <i>destination</i>
     * Example disassembly syntax: {@code verw      [ebx + 305419896]}
     */
    // Template#: 5432, Serial#: 2542
    public void verw(int disp32, IA32IndirectRegister32 destination) {
        assemble0251((byte) 0x00, (byte) 0x05, disp32, destination);
    }

    /**
     * Pseudo-external assembler syntax: {@code wbinvd  }
     * Example disassembly syntax: {@code wbinvd  }
     */
    // Template#: 5433, Serial#: 4030
    public void wbinvd() {
        assemble0198((byte) 0x09);
    }

    /**
     * Pseudo-external assembler syntax: {@code wrmsr  }
     * Example disassembly syntax: {@code wrmsr   }
     */
    // Template#: 5434, Serial#: 2833
    public void wrmsr() {
        assemble0198((byte) 0x30);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      ebx[esi * 4 + 18], ax}
     */
    // Template#: 5435, Serial#: 3618
    public void xadd(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0158((byte) 0xC1, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      ebx[esi * 4 + 18], eax}
     */
    // Template#: 5436, Serial#: 3609
    public void xadd(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0159((byte) 0xC1, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      ebx[esi * 4 + 18], al}
     */
    // Template#: 5437, Serial#: 3600
    public void xadd(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0235((byte) 0xC0, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      [ebx + 18], ax}
     */
    // Template#: 5438, Serial#: 3617
    public void xadd(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0161((byte) 0xC1, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      [ebx + 18], eax}
     */
    // Template#: 5439, Serial#: 3608
    public void xadd(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0162((byte) 0xC1, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      [ebx + 18], al}
     */
    // Template#: 5440, Serial#: 3599
    public void xadd(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0236((byte) 0xC0, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      ebx[esi * 4], ax}
     */
    // Template#: 5441, Serial#: 3614
    public void xadd(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0164((byte) 0xC1, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      ebx[esi * 4], eax}
     */
    // Template#: 5442, Serial#: 3605
    public void xadd(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0165((byte) 0xC1, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      ebx[esi * 4], al}
     */
    // Template#: 5443, Serial#: 3596
    public void xadd(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0237((byte) 0xC0, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      ax, ax}
     */
    // Template#: 5444, Serial#: 3621
    public void xadd(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0167((byte) 0xC1, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      eax, eax}
     */
    // Template#: 5445, Serial#: 3612
    public void xadd(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0169((byte) 0xC1, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      al, al}
     */
    // Template#: 5446, Serial#: 3603
    public void xadd(IA32GeneralRegister8 destination, IA32GeneralRegister8 source) {
        assemble0238((byte) 0xC0, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      [ebx], ax}
     */
    // Template#: 5447, Serial#: 3613
    public void xadd(IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0171((byte) 0xC1, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      [ebx], eax}
     */
    // Template#: 5448, Serial#: 3604
    public void xadd(IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0172((byte) 0xC1, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      [ebx], al}
     */
    // Template#: 5449, Serial#: 3595
    public void xadd(IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0239((byte) 0xC0, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      ebx[esi * 4 + 305419896], ax}
     */
    // Template#: 5450, Serial#: 3620
    public void xadd(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0175((byte) 0xC1, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      ebx[esi * 4 + 305419896], eax}
     */
    // Template#: 5451, Serial#: 3611
    public void xadd(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0176((byte) 0xC1, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      ebx[esi * 4 + 305419896], al}
     */
    // Template#: 5452, Serial#: 3602
    public void xadd(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0240((byte) 0xC0, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      [L1: 0x12345678], ax}
     */
    // Template#: 5453, Serial#: 3616
    public void m_xadd(int m32, IA32GeneralRegister16 source) {
        assemble0177((byte) 0xC1, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      [L1: 0x12345678], eax}
     */
    // Template#: 5454, Serial#: 3607
    public void m_xadd(int m32, IA32GeneralRegister32 source) {
        assemble0178((byte) 0xC1, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      [L1: 0x12345678], al}
     */
    // Template#: 5455, Serial#: 3598
    public void m_xadd(int m32, IA32GeneralRegister8 source) {
        assemble0241((byte) 0xC0, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      L1: 0x12345678[esi * 4], ax}
     */
    // Template#: 5456, Serial#: 3615
    public void m_xadd(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0180((byte) 0xC1, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      L1: 0x12345678[esi * 4], eax}
     */
    // Template#: 5457, Serial#: 3606
    public void m_xadd(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0181((byte) 0xC1, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      L1: 0x12345678[esi * 4], al}
     */
    // Template#: 5458, Serial#: 3597
    public void m_xadd(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0242((byte) 0xC0, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      [ebx + 305419896], ax}
     */
    // Template#: 5459, Serial#: 3619
    public void xadd(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0183((byte) 0xC1, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      [ebx + 305419896], eax}
     */
    // Template#: 5460, Serial#: 3610
    public void xadd(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0184((byte) 0xC1, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xadd  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xadd      [ebx + 305419896], al}
     */
    // Template#: 5461, Serial#: 3601
    public void xadd(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0243((byte) 0xC0, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      ebx[esi * 4 + 18], ax}
     */
    // Template#: 5462, Serial#: 680
    public void xchg(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0005((byte) 0x87, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      ebx[esi * 4 + 18], eax}
     */
    // Template#: 5463, Serial#: 671
    public void xchg(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0006((byte) 0x87, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      ebx[esi * 4 + 18], al}
     */
    // Template#: 5464, Serial#: 662
    public void xchg(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0007((byte) 0x86, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      [ebx + 18], ax}
     */
    // Template#: 5465, Serial#: 679
    public void xchg(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0012((byte) 0x87, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      [ebx + 18], eax}
     */
    // Template#: 5466, Serial#: 670
    public void xchg(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0013((byte) 0x87, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      [ebx + 18], al}
     */
    // Template#: 5467, Serial#: 661
    public void xchg(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0014((byte) 0x86, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      ebx[esi * 4], ax}
     */
    // Template#: 5468, Serial#: 676
    public void xchg(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0019((byte) 0x87, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      ebx[esi * 4], eax}
     */
    // Template#: 5469, Serial#: 667
    public void xchg(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0020((byte) 0x87, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      ebx[esi * 4], al}
     */
    // Template#: 5470, Serial#: 658
    public void xchg(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0021((byte) 0x86, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>register</i>
     * Example disassembly syntax: {@code xchg      ax, ax}
     */
    // Template#: 5471, Serial#: 686
    public void xchg___AX(IA32GeneralRegister16 register) {
        assemble0295((byte) 0x90, register);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      ax, ax}
     */
    // Template#: 5472, Serial#: 683
    public void xchg(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0028((byte) 0x87, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>register</i>
     * Example disassembly syntax: {@code xchg      eax, eax}
     */
    // Template#: 5473, Serial#: 685
    public void xchg___EAX(IA32GeneralRegister32 register) {
        assemble0296((byte) 0x90, register);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      eax, eax}
     */
    // Template#: 5474, Serial#: 674
    public void xchg(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0039((byte) 0x87, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      al, al}
     */
    // Template#: 5475, Serial#: 665
    public void xchg(IA32GeneralRegister8 destination, IA32GeneralRegister8 source) {
        assemble0050((byte) 0x86, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      [ebx], ax}
     */
    // Template#: 5476, Serial#: 675
    public void xchg(IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0058((byte) 0x87, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      [ebx], eax}
     */
    // Template#: 5477, Serial#: 666
    public void xchg(IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0059((byte) 0x87, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      [ebx], al}
     */
    // Template#: 5478, Serial#: 657
    public void xchg(IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0060((byte) 0x86, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      ebx[esi * 4 + 305419896], ax}
     */
    // Template#: 5479, Serial#: 682
    public void xchg(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0068((byte) 0x87, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      ebx[esi * 4 + 305419896], eax}
     */
    // Template#: 5480, Serial#: 673
    public void xchg(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0069((byte) 0x87, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      ebx[esi * 4 + 305419896], al}
     */
    // Template#: 5481, Serial#: 664
    public void xchg(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0070((byte) 0x86, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      [L1: 0x12345678], ax}
     */
    // Template#: 5482, Serial#: 678
    public void m_xchg(int m32, IA32GeneralRegister16 source) {
        assemble0073((byte) 0x87, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      [L1: 0x12345678], eax}
     */
    // Template#: 5483, Serial#: 669
    public void m_xchg(int m32, IA32GeneralRegister32 source) {
        assemble0074((byte) 0x87, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      [L1: 0x12345678], al}
     */
    // Template#: 5484, Serial#: 660
    public void m_xchg(int m32, IA32GeneralRegister8 source) {
        assemble0075((byte) 0x86, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      L1: 0x12345678[esi * 4], ax}
     */
    // Template#: 5485, Serial#: 677
    public void m_xchg(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0078((byte) 0x87, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      L1: 0x12345678[esi * 4], eax}
     */
    // Template#: 5486, Serial#: 668
    public void m_xchg(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0079((byte) 0x87, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      L1: 0x12345678[esi * 4], al}
     */
    // Template#: 5487, Serial#: 659
    public void m_xchg(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0080((byte) 0x86, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      [ebx + 305419896], ax}
     */
    // Template#: 5488, Serial#: 681
    public void xchg(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0085((byte) 0x87, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      [ebx + 305419896], eax}
     */
    // Template#: 5489, Serial#: 672
    public void xchg(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0086((byte) 0x87, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xchg  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xchg      [ebx + 305419896], al}
     */
    // Template#: 5490, Serial#: 663
    public void xchg(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0087((byte) 0x86, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xlat  }
     * Example disassembly syntax: {@code xlat    }
     */
    // Template#: 5491, Serial#: 1331
    public void xlat() {
        assemble0001((byte) 0xD7);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>imm8</i>
     * Example disassembly syntax: {@code xor       al, 0x12}
     */
    // Template#: 5492, Serial#: 220
    public void xor_AL(byte imm8) {
        assemble0002((byte) 0x34, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorb  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorb      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 5493, Serial#: 315
    public void xorb(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0x80, (byte) 0x06, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorl      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 5494, Serial#: 531
    public void xorl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0003((byte) 0x83, (byte) 0x06, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorw      ebx[esi * 4 + 18], 0x12}
     */
    // Template#: 5495, Serial#: 603
    public void xorw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0004((byte) 0x83, (byte) 0x06, disp8, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       ebx[esi * 4 + 18], ax}
     */
    // Template#: 5496, Serial#: 192
    public void xor(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0005((byte) 0x31, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       ebx[esi * 4 + 18], eax}
     */
    // Template#: 5497, Serial#: 183
    public void xor(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0006((byte) 0x31, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       ebx[esi * 4 + 18], al}
     */
    // Template#: 5498, Serial#: 174
    public void xor(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0007((byte) 0x30, disp8, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorl  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code xorl      ebx[esi * 4 + 18], 0x12345678}
     */
    // Template#: 5499, Serial#: 387
    public void xorl(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0008((byte) 0x81, (byte) 0x06, disp8, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorw  }<i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code xorw      ebx[esi * 4 + 18], 0x1234}
     */
    // Template#: 5500, Serial#: 459
    public void xorw(byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0009((byte) 0x81, (byte) 0x06, disp8, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorb  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorb      [ebx + 18], 0x12}
     */
    // Template#: 5501, Serial#: 314
    public void xorb(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0x80, (byte) 0x06, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorl  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorl      [ebx + 18], 0x12}
     */
    // Template#: 5502, Serial#: 530
    public void xorl(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0010((byte) 0x83, (byte) 0x06, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorw  }<i>disp8</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorw      [ebx + 18], 0x12}
     */
    // Template#: 5503, Serial#: 602
    public void xorw(byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        assemble0011((byte) 0x83, (byte) 0x06, disp8, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       [ebx + 18], ax}
     */
    // Template#: 5504, Serial#: 191
    public void xor(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0012((byte) 0x31, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       [ebx + 18], eax}
     */
    // Template#: 5505, Serial#: 182
    public void xor(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0013((byte) 0x31, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>disp8</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       [ebx + 18], al}
     */
    // Template#: 5506, Serial#: 173
    public void xor(byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0014((byte) 0x30, disp8, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorl  }<i>disp8</i>, <i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code xorl      [ebx + 18], 0x12345678}
     */
    // Template#: 5507, Serial#: 386
    public void xorl(byte disp8, IA32IndirectRegister32 destination, int imm32) {
        assemble0015((byte) 0x81, (byte) 0x06, disp8, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorw  }<i>disp8</i>, <i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code xorw      [ebx + 18], 0x1234}
     */
    // Template#: 5508, Serial#: 458
    public void xorw(byte disp8, IA32IndirectRegister32 destination, short imm16) {
        assemble0016((byte) 0x81, (byte) 0x06, disp8, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorb  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorb      ebx[esi * 4], 0x12}
     */
    // Template#: 5509, Serial#: 295
    public void xorb(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0x80, (byte) 0x06, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorl  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorl      ebx[esi * 4], 0x12}
     */
    // Template#: 5510, Serial#: 511
    public void xorl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0017((byte) 0x83, (byte) 0x06, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorw      ebx[esi * 4], 0x12}
     */
    // Template#: 5511, Serial#: 583
    public void xorw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0018((byte) 0x83, (byte) 0x06, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       ebx[esi * 4], ax}
     */
    // Template#: 5512, Serial#: 188
    public void xor(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0019((byte) 0x31, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       ebx[esi * 4], eax}
     */
    // Template#: 5513, Serial#: 179
    public void xor(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0020((byte) 0x31, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       ebx[esi * 4], al}
     */
    // Template#: 5514, Serial#: 170
    public void xor(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0021((byte) 0x30, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorl  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code xorl      ebx[esi * 4], 0x12345678}
     */
    // Template#: 5515, Serial#: 367
    public void xorl(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0022((byte) 0x81, (byte) 0x06, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorw  }<i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code xorw      ebx[esi * 4], 0x1234}
     */
    // Template#: 5516, Serial#: 439
    public void xorw(IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0023((byte) 0x81, (byte) 0x06, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorw      ax, 0x12}
     */
    // Template#: 5517, Serial#: 628
    public void xorw(IA32GeneralRegister16 destination, byte imm8) {
        assemble0024((byte) 0x83, (byte) 0x06, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code xor       ax, ebx[esi * 4 + 18]}
     */
    // Template#: 5518, Serial#: 217
    public void xor(IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0025((byte) 0x33, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       ax, [ebx + 18]}
     */
    // Template#: 5519, Serial#: 216
    public void xor(IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0026((byte) 0x33, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code xor       ax, ebx[esi * 4]}
     */
    // Template#: 5520, Serial#: 213
    public void xor(IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0027((byte) 0x33, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       ax, ax}
     */
    // Template#: 5521, Serial#: 195
    public void xor(IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        assemble0028((byte) 0x31, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       ax, [ebx]}
     */
    // Template#: 5522, Serial#: 212
    public void xor(IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        assemble0029((byte) 0x33, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code xor       ax, [L1: 0x12345678]}
     */
    // Template#: 5523, Serial#: 215
    public void m_xor(IA32GeneralRegister16 destination, int m32) {
        assemble0030((byte) 0x33, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code xor       ax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 5524, Serial#: 219
    public void xor(IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0031((byte) 0x33, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code xor       ax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 5525, Serial#: 214
    public void m_xor(IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0032((byte) 0x33, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       ax, [ebx + 305419896]}
     */
    // Template#: 5526, Serial#: 218
    public void xor(IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0033((byte) 0x33, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorw  }<i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code xorw      ax, 0x1234}
     */
    // Template#: 5527, Serial#: 484
    public void xorw(IA32GeneralRegister16 destination, short imm16) {
        assemble0034((byte) 0x81, (byte) 0x06, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorl  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorl      eax, 0x12}
     */
    // Template#: 5528, Serial#: 556
    public void xorl(IA32GeneralRegister32 destination, byte imm8) {
        assemble0035((byte) 0x83, (byte) 0x06, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code xor       eax, ebx[esi * 4 + 18]}
     */
    // Template#: 5529, Serial#: 209
    public void xor(IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0036((byte) 0x33, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       eax, [ebx + 18]}
     */
    // Template#: 5530, Serial#: 208
    public void xor(IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0037((byte) 0x33, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code xor       eax, ebx[esi * 4]}
     */
    // Template#: 5531, Serial#: 205
    public void xor(IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0038((byte) 0x33, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       eax, eax}
     */
    // Template#: 5532, Serial#: 186
    public void xor(IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        assemble0039((byte) 0x31, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       eax, [ebx]}
     */
    // Template#: 5533, Serial#: 204
    public void xor(IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        assemble0040((byte) 0x33, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code xor       eax, [L1: 0x12345678]}
     */
    // Template#: 5534, Serial#: 207
    public void m_xor(IA32GeneralRegister32 destination, int m32) {
        assemble0041((byte) 0x33, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorl  }<i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code xorl      eax, 0x12345678}
     */
    // Template#: 5535, Serial#: 412
    public void xorl(IA32GeneralRegister32 destination, int imm32) {
        assemble0042((byte) 0x81, (byte) 0x06, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code xor       eax, ebx[esi * 4 + 305419896]}
     */
    // Template#: 5536, Serial#: 211
    public void xor(IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0043((byte) 0x33, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code xor       eax, L1: 0x12345678[esi * 4]}
     */
    // Template#: 5537, Serial#: 206
    public void m_xor(IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0044((byte) 0x33, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       eax, [ebx + 305419896]}
     */
    // Template#: 5538, Serial#: 210
    public void xor(IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0045((byte) 0x33, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorb      al, 0x12}
     */
    // Template#: 5539, Serial#: 340
    public void xorb(IA32GeneralRegister8 destination, byte imm8) {
        assemble0046((byte) 0x80, (byte) 0x06, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code xor       al, ebx[esi * 4 + 18]}
     */
    // Template#: 5540, Serial#: 201
    public void xor(IA32GeneralRegister8 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0047((byte) 0x32, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       al, [ebx + 18]}
     */
    // Template#: 5541, Serial#: 200
    public void xor(IA32GeneralRegister8 destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0048((byte) 0x32, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code xor       al, ebx[esi * 4]}
     */
    // Template#: 5542, Serial#: 197
    public void xor(IA32GeneralRegister8 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0049((byte) 0x32, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       al, al}
     */
    // Template#: 5543, Serial#: 177
    public void xor(IA32GeneralRegister8 destination, IA32GeneralRegister8 source) {
        assemble0050((byte) 0x30, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       al, [ebx]}
     */
    // Template#: 5544, Serial#: 196
    public void xor(IA32GeneralRegister8 destination, IA32IndirectRegister32 source) {
        assemble0051((byte) 0x32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code xor       al, [L1: 0x12345678]}
     */
    // Template#: 5545, Serial#: 199
    public void m_xor(IA32GeneralRegister8 destination, int m32) {
        assemble0052((byte) 0x32, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code xor       al, ebx[esi * 4 + 305419896]}
     */
    // Template#: 5546, Serial#: 203
    public void xor(IA32GeneralRegister8 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0053((byte) 0x32, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code xor       al, L1: 0x12345678[esi * 4]}
     */
    // Template#: 5547, Serial#: 198
    public void m_xor(IA32GeneralRegister8 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0054((byte) 0x32, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       al, [ebx + 305419896]}
     */
    // Template#: 5548, Serial#: 202
    public void xor(IA32GeneralRegister8 destination, int disp32, IA32IndirectRegister32 source) {
        assemble0055((byte) 0x32, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorb  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorb      [ebx], 0x12}
     */
    // Template#: 5549, Serial#: 294
    public void xorb(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0x80, (byte) 0x06, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorl  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorl      [ebx], 0x12}
     */
    // Template#: 5550, Serial#: 510
    public void xorl(IA32IndirectRegister32 destination, byte imm8) {
        assemble0056((byte) 0x83, (byte) 0x06, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorw  }<i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorw      [ebx], 0x12}
     */
    // Template#: 5551, Serial#: 582
    public void xorw(IA32IndirectRegister32 destination, byte imm8) {
        assemble0057((byte) 0x83, (byte) 0x06, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       [ebx], ax}
     */
    // Template#: 5552, Serial#: 187
    public void xor(IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0058((byte) 0x31, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       [ebx], eax}
     */
    // Template#: 5553, Serial#: 178
    public void xor(IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0059((byte) 0x31, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       [ebx], al}
     */
    // Template#: 5554, Serial#: 169
    public void xor(IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0060((byte) 0x30, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorl  }<i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code xorl      [ebx], 0x12345678}
     */
    // Template#: 5555, Serial#: 366
    public void xorl(IA32IndirectRegister32 destination, int imm32) {
        assemble0061((byte) 0x81, (byte) 0x06, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorw  }<i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code xorw      [ebx], 0x1234}
     */
    // Template#: 5556, Serial#: 438
    public void xorw(IA32IndirectRegister32 destination, short imm16) {
        assemble0062((byte) 0x81, (byte) 0x06, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorl  }<i>imm32</i>
     * Example disassembly syntax: {@code xorl      eax, 0x12345678}
     */
    // Template#: 5557, Serial#: 221
    public void xor_EAX(int imm32) {
        assemble0063((byte) 0x35, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorb  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorb      [L1: 0x12345678], 0x12}
     */
    // Template#: 5558, Serial#: 297
    public void m_xorb(int m32, byte imm8) {
        assemble0064((byte) 0x80, (byte) 0x06, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorl  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorl      [L1: 0x12345678], 0x12}
     */
    // Template#: 5559, Serial#: 513
    public void m_xorl(int m32, byte imm8) {
        assemble0064((byte) 0x83, (byte) 0x06, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorw  }<i>m32</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorw      [L1: 0x12345678], 0x12}
     */
    // Template#: 5560, Serial#: 585
    public void m_xorw(int m32, byte imm8) {
        assemble0065((byte) 0x83, (byte) 0x06, m32, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorb  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorb      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 5561, Serial#: 331
    public void xorb(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0x80, (byte) 0x06, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorl      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 5562, Serial#: 547
    public void xorl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0066((byte) 0x83, (byte) 0x06, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorw      ebx[esi * 4 + 305419896], 0x12}
     */
    // Template#: 5563, Serial#: 619
    public void xorw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0067((byte) 0x83, (byte) 0x06, disp32, base, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       ebx[esi * 4 + 305419896], ax}
     */
    // Template#: 5564, Serial#: 194
    public void xor(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0068((byte) 0x31, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       ebx[esi * 4 + 305419896], eax}
     */
    // Template#: 5565, Serial#: 185
    public void xor(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0069((byte) 0x31, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       ebx[esi * 4 + 305419896], al}
     */
    // Template#: 5566, Serial#: 176
    public void xor(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0070((byte) 0x30, disp32, base, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorl  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code xorl      ebx[esi * 4 + 305419896], 0x12345678}
     */
    // Template#: 5567, Serial#: 403
    public void xorl(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0071((byte) 0x81, (byte) 0x06, disp32, base, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorw  }<i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code xorw      ebx[esi * 4 + 305419896], 0x1234}
     */
    // Template#: 5568, Serial#: 475
    public void xorw(int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0072((byte) 0x81, (byte) 0x06, disp32, base, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       [L1: 0x12345678], ax}
     */
    // Template#: 5569, Serial#: 190
    public void m_xor(int m32, IA32GeneralRegister16 source) {
        assemble0073((byte) 0x31, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       [L1: 0x12345678], eax}
     */
    // Template#: 5570, Serial#: 181
    public void m_xor(int m32, IA32GeneralRegister32 source) {
        assemble0074((byte) 0x31, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>m32</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       [L1: 0x12345678], al}
     */
    // Template#: 5571, Serial#: 172
    public void m_xor(int m32, IA32GeneralRegister8 source) {
        assemble0075((byte) 0x30, m32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorb  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorb      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 5572, Serial#: 296
    public void m_xorb(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0x80, (byte) 0x06, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorl  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorl      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 5573, Serial#: 512
    public void m_xorl(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0076((byte) 0x83, (byte) 0x06, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorw      L1: 0x12345678[esi * 4], 0x12}
     */
    // Template#: 5574, Serial#: 584
    public void m_xorw(int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        assemble0077((byte) 0x83, (byte) 0x06, m32, index, scale, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       L1: 0x12345678[esi * 4], ax}
     */
    // Template#: 5575, Serial#: 189
    public void m_xor(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        assemble0078((byte) 0x31, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       L1: 0x12345678[esi * 4], eax}
     */
    // Template#: 5576, Serial#: 180
    public void m_xor(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        assemble0079((byte) 0x31, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       L1: 0x12345678[esi * 4], al}
     */
    // Template#: 5577, Serial#: 171
    public void m_xor(int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        assemble0080((byte) 0x30, m32, index, scale, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorl  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm32</i>
     * Example disassembly syntax: {@code xorl      L1: 0x12345678[esi * 4], 0x12345678}
     */
    // Template#: 5578, Serial#: 368
    public void m_xorl(int m32, IA32IndexRegister32 index, Scale scale, int imm32) {
        assemble0081((byte) 0x81, (byte) 0x06, m32, index, scale, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorw  }<i>m32</i>, <i>index</i>, <i>scale</i>, <i>imm16</i>
     * Example disassembly syntax: {@code xorw      L1: 0x12345678[esi * 4], 0x1234}
     */
    // Template#: 5579, Serial#: 440
    public void m_xorw(int m32, IA32IndexRegister32 index, Scale scale, short imm16) {
        assemble0082((byte) 0x81, (byte) 0x06, m32, index, scale, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorb  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorb      [ebx + 305419896], 0x12}
     */
    // Template#: 5580, Serial#: 330
    public void xorb(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0x80, (byte) 0x06, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorl  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorl      [ebx + 305419896], 0x12}
     */
    // Template#: 5581, Serial#: 546
    public void xorl(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0083((byte) 0x83, (byte) 0x06, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorw  }<i>disp32</i>, <i>destination</i>, <i>imm8</i>
     * Example disassembly syntax: {@code xorw      [ebx + 305419896], 0x12}
     */
    // Template#: 5582, Serial#: 618
    public void xorw(int disp32, IA32IndirectRegister32 destination, byte imm8) {
        assemble0084((byte) 0x83, (byte) 0x06, disp32, destination, imm8);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       [ebx + 305419896], ax}
     */
    // Template#: 5583, Serial#: 193
    public void xor(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        assemble0085((byte) 0x31, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       [ebx + 305419896], eax}
     */
    // Template#: 5584, Serial#: 184
    public void xor(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        assemble0086((byte) 0x31, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xor  }<i>disp32</i>, <i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xor       [ebx + 305419896], al}
     */
    // Template#: 5585, Serial#: 175
    public void xor(int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        assemble0087((byte) 0x30, disp32, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorl  }<i>disp32</i>, <i>destination</i>, <i>imm32</i>
     * Example disassembly syntax: {@code xorl      [ebx + 305419896], 0x12345678}
     */
    // Template#: 5586, Serial#: 402
    public void xorl(int disp32, IA32IndirectRegister32 destination, int imm32) {
        assemble0088((byte) 0x81, (byte) 0x06, disp32, destination, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorw  }<i>disp32</i>, <i>destination</i>, <i>imm16</i>
     * Example disassembly syntax: {@code xorw      [ebx + 305419896], 0x1234}
     */
    // Template#: 5587, Serial#: 474
    public void xorw(int disp32, IA32IndirectRegister32 destination, short imm16) {
        assemble0089((byte) 0x81, (byte) 0x06, disp32, destination, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorl  }<i>m32</i>, <i>imm32</i>
     * Example disassembly syntax: {@code xorl      [L1: 0x12345678], 0x12345678}
     */
    // Template#: 5588, Serial#: 369
    public void m_xorl(int m32, int imm32) {
        assemble0090((byte) 0x81, (byte) 0x06, m32, imm32);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorw  }<i>m32</i>, <i>imm16</i>
     * Example disassembly syntax: {@code xorw      [L1: 0x12345678], 0x1234}
     */
    // Template#: 5589, Serial#: 441
    public void m_xorw(int m32, short imm16) {
        assemble0091((byte) 0x81, (byte) 0x06, m32, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorw  }<i>imm16</i>
     * Example disassembly syntax: {@code xorw      ax, 0x1234}
     */
    // Template#: 5590, Serial#: 222
    public void xor_AX(short imm16) {
        assemble0092((byte) 0x35, imm16);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorpd  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code xorpd     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 5591, Serial#: 3087
    public void xorpd(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0093((byte) 0x57, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorpd  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code xorpd     xmm0, [ebx + 18]}
     */
    // Template#: 5592, Serial#: 3086
    public void xorpd(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0094((byte) 0x57, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorpd  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code xorpd     xmm0, ebx[esi * 4]}
     */
    // Template#: 5593, Serial#: 3083
    public void xorpd(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0095((byte) 0x57, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xorpd     xmm0, [ebx]}
     */
    // Template#: 5594, Serial#: 3082
    public void xorpd(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0096((byte) 0x57, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorpd  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xorpd     xmm0, xmm0}
     */
    // Template#: 5595, Serial#: 3090
    public void xorpd(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0097((byte) 0x57, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorpd  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code xorpd     xmm0, [L1: 0x12345678]}
     */
    // Template#: 5596, Serial#: 3085
    public void m_xorpd(IA32XMMRegister destination, int m32) {
        assemble0098((byte) 0x57, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorpd  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code xorpd     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 5597, Serial#: 3089
    public void xorpd(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0099((byte) 0x57, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorpd  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code xorpd     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 5598, Serial#: 3084
    public void m_xorpd(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0100((byte) 0x57, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorpd  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code xorpd     xmm0, [ebx + 305419896]}
     */
    // Template#: 5599, Serial#: 3088
    public void xorpd(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0101((byte) 0x57, destination, disp32, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorps  }<i>destination</i>, <i>disp8</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code xorps     xmm0, ebx[esi * 4 + 18]}
     */
    // Template#: 5600, Serial#: 3041
    public void xorps(IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0102((byte) 0x57, destination, disp8, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorps  }<i>destination</i>, <i>disp8</i>, <i>source</i>
     * Example disassembly syntax: {@code xorps     xmm0, [ebx + 18]}
     */
    // Template#: 5601, Serial#: 3040
    public void xorps(IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        assemble0103((byte) 0x57, destination, disp8, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorps  }<i>destination</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code xorps     xmm0, ebx[esi * 4]}
     */
    // Template#: 5602, Serial#: 3037
    public void xorps(IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0104((byte) 0x57, destination, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xorps     xmm0, [ebx]}
     */
    // Template#: 5603, Serial#: 3036
    public void xorps(IA32XMMRegister destination, IA32IndirectRegister32 source) {
        assemble0105((byte) 0x57, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorps  }<i>destination</i>, <i>source</i>
     * Example disassembly syntax: {@code xorps     xmm0, xmm0}
     */
    // Template#: 5604, Serial#: 3044
    public void xorps(IA32XMMRegister destination, IA32XMMRegister source) {
        assemble0106((byte) 0x57, destination, source);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorps  }<i>destination</i>, <i>m32</i>
     * Example disassembly syntax: {@code xorps     xmm0, [L1: 0x12345678]}
     */
    // Template#: 5605, Serial#: 3039
    public void m_xorps(IA32XMMRegister destination, int m32) {
        assemble0107((byte) 0x57, destination, m32);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorps  }<i>destination</i>, <i>disp32</i>, <i>base</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code xorps     xmm0, ebx[esi * 4 + 305419896]}
     */
    // Template#: 5606, Serial#: 3043
    public void xorps(IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        assemble0108((byte) 0x57, destination, disp32, base, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorps  }<i>destination</i>, <i>m32</i>, <i>index</i>, <i>scale</i>
     * Example disassembly syntax: {@code xorps     xmm0, L1: 0x12345678[esi * 4]}
     */
    // Template#: 5607, Serial#: 3038
    public void m_xorps(IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        assemble0109((byte) 0x57, destination, m32, index, scale);
    }

    /**
     * Pseudo-external assembler syntax: {@code xorps  }<i>destination</i>, <i>disp32</i>, <i>source</i>
     * Example disassembly syntax: {@code xorps     xmm0, [ebx + 305419896]}
     */
    // Template#: 5608, Serial#: 3042
    public void xorps(IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        assemble0110((byte) 0x57, destination, disp32, source);
    }

    private void assemble0001(byte opcode1) {
        emitByte(opcode1);
    }

    private void assemble0002(byte opcode1, byte imm8) {
        emitByte(opcode1);
        emitByte(imm8); // appended
    }

    private void assemble0003(byte opcode1, byte modRmOpcode, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
        emitByte(imm8); // appended
    }

    private void assemble0004(byte opcode1, byte modRmOpcode, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
        emitByte(imm8); // appended
    }

    private void assemble0005(byte opcode1, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0006(byte opcode1, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0007(byte opcode1, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0008(byte opcode1, byte modRmOpcode, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0009(byte opcode1, byte modRmOpcode, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0010(byte opcode1, byte modRmOpcode, byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
        emitByte(imm8); // appended
    }

    private void assemble0011(byte opcode1, byte modRmOpcode, byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
        emitByte(imm8); // appended
    }

    private void assemble0012(byte opcode1, byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0013(byte opcode1, byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0014(byte opcode1, byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0015(byte opcode1, byte modRmOpcode, byte disp8, IA32IndirectRegister32 destination, int imm32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0016(byte opcode1, byte modRmOpcode, byte disp8, IA32IndirectRegister32 destination, short imm16) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0017(byte opcode1, byte modRmOpcode, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0003(opcode1, modRmOpcode, (byte) 0, base, index, scale, imm8);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(imm8); // appended
    }

    private void assemble0018(byte opcode1, byte modRmOpcode, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0004(opcode1, modRmOpcode, (byte) 0, base, index, scale, imm8);
            return;
        }
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(imm8); // appended
    }

    private void assemble0019(byte opcode1, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0005(opcode1, (byte) 0, base, index, scale, source);
            return;
        }
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0020(byte opcode1, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0006(opcode1, (byte) 0, base, index, scale, source);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0021(byte opcode1, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0007(opcode1, (byte) 0, base, index, scale, source);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0022(byte opcode1, byte modRmOpcode, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0008(opcode1, modRmOpcode, (byte) 0, base, index, scale, imm32);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0023(byte opcode1, byte modRmOpcode, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0009(opcode1, modRmOpcode, (byte) 0, base, index, scale, imm16);
            return;
        }
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0024(byte opcode1, byte modRmOpcode, IA32GeneralRegister16 destination, byte imm8) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((3 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        emitByte(imm8); // appended
    }

    private void assemble0025(byte opcode1, IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0026(byte opcode1, IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0027(byte opcode1, IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0025(opcode1, destination, (byte) 0, base, index, scale);
            return;
        }
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0028(byte opcode1, IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
    }

    private void assemble0029(byte opcode1, IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0026(opcode1, destination, (byte) 0, source);
            return;
        }
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0030(byte opcode1, IA32GeneralRegister16 destination, int m32) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0031(byte opcode1, IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0032(byte opcode1, IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0033(byte opcode1, IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0034(byte opcode1, byte modRmOpcode, IA32GeneralRegister16 destination, short imm16) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((3 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0035(byte opcode1, byte modRmOpcode, IA32GeneralRegister32 destination, byte imm8) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((3 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        emitByte(imm8); // appended
    }

    private void assemble0036(byte opcode1, IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0037(byte opcode1, IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0038(byte opcode1, IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0036(opcode1, destination, (byte) 0, base, index, scale);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0039(byte opcode1, IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
    }

    private void assemble0040(byte opcode1, IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0037(opcode1, destination, (byte) 0, source);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0041(byte opcode1, IA32GeneralRegister32 destination, int m32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0042(byte opcode1, byte modRmOpcode, IA32GeneralRegister32 destination, int imm32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((3 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0043(byte opcode1, IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0044(byte opcode1, IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0045(byte opcode1, IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0046(byte opcode1, byte modRmOpcode, IA32GeneralRegister8 destination, byte imm8) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((3 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        emitByte(imm8); // appended
    }

    private void assemble0047(byte opcode1, IA32GeneralRegister8 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0048(byte opcode1, IA32GeneralRegister8 destination, byte disp8, IA32IndirectRegister32 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0049(byte opcode1, IA32GeneralRegister8 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0047(opcode1, destination, (byte) 0, base, index, scale);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0050(byte opcode1, IA32GeneralRegister8 destination, IA32GeneralRegister8 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
    }

    private void assemble0051(byte opcode1, IA32GeneralRegister8 destination, IA32IndirectRegister32 source) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0048(opcode1, destination, (byte) 0, source);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0052(byte opcode1, IA32GeneralRegister8 destination, int m32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0053(byte opcode1, IA32GeneralRegister8 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0054(byte opcode1, IA32GeneralRegister8 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0055(byte opcode1, IA32GeneralRegister8 destination, int disp32, IA32IndirectRegister32 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0056(byte opcode1, byte modRmOpcode, IA32IndirectRegister32 destination, byte imm8) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0010(opcode1, modRmOpcode, (byte) 0, destination, imm8);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(imm8); // appended
    }

    private void assemble0057(byte opcode1, byte modRmOpcode, IA32IndirectRegister32 destination, byte imm8) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0011(opcode1, modRmOpcode, (byte) 0, destination, imm8);
            return;
        }
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(imm8); // appended
    }

    private void assemble0058(byte opcode1, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0012(opcode1, (byte) 0, destination, source);
            return;
        }
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0059(byte opcode1, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0013(opcode1, (byte) 0, destination, source);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0060(byte opcode1, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0014(opcode1, (byte) 0, destination, source);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0061(byte opcode1, byte modRmOpcode, IA32IndirectRegister32 destination, int imm32) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0015(opcode1, modRmOpcode, (byte) 0, destination, imm32);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0062(byte opcode1, byte modRmOpcode, IA32IndirectRegister32 destination, short imm16) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0016(opcode1, modRmOpcode, (byte) 0, destination, imm16);
            return;
        }
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0063(byte opcode1, int imm32) {
        emitByte(opcode1);
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0064(byte opcode1, byte modRmOpcode, int m32, byte imm8) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 5 << 0; // rm field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0065(byte opcode1, byte modRmOpcode, int m32, byte imm8) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 5 << 0; // rm field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0066(byte opcode1, byte modRmOpcode, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0067(byte opcode1, byte modRmOpcode, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0068(byte opcode1, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0069(byte opcode1, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0070(byte opcode1, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0071(byte opcode1, byte modRmOpcode, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0072(byte opcode1, byte modRmOpcode, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0073(byte opcode1, int m32, IA32GeneralRegister16 source) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0074(byte opcode1, int m32, IA32GeneralRegister32 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0075(byte opcode1, int m32, IA32GeneralRegister8 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0076(byte opcode1, byte modRmOpcode, int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0077(byte opcode1, byte modRmOpcode, int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0078(byte opcode1, int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0079(byte opcode1, int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0080(byte opcode1, int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0081(byte opcode1, byte modRmOpcode, int m32, IA32IndexRegister32 index, Scale scale, int imm32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0082(byte opcode1, byte modRmOpcode, int m32, IA32IndexRegister32 index, Scale scale, short imm16) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0083(byte opcode1, byte modRmOpcode, int disp32, IA32IndirectRegister32 destination, byte imm8) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0084(byte opcode1, byte modRmOpcode, int disp32, IA32IndirectRegister32 destination, byte imm8) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0085(byte opcode1, int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0086(byte opcode1, int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0087(byte opcode1, int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0088(byte opcode1, byte modRmOpcode, int disp32, IA32IndirectRegister32 destination, int imm32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0089(byte opcode1, byte modRmOpcode, int disp32, IA32IndirectRegister32 destination, short imm16) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0090(byte opcode1, byte modRmOpcode, int m32, int imm32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 5 << 0; // rm field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0091(byte opcode1, byte modRmOpcode, int m32, short imm16) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 5 << 0; // rm field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0092(byte opcode1, short imm16) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0093(byte opcode2, IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0094(byte opcode2, IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0095(byte opcode2, IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0093(opcode2, destination, (byte) 0, base, index, scale);
            return;
        }
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0096(byte opcode2, IA32XMMRegister destination, IA32IndirectRegister32 source) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0094(opcode2, destination, (byte) 0, source);
            return;
        }
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0097(byte opcode2, IA32XMMRegister destination, IA32XMMRegister source) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0098(byte opcode2, IA32XMMRegister destination, int m32) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0099(byte opcode2, IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0100(byte opcode2, IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0101(byte opcode2, IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0102(byte opcode2, IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0103(byte opcode2, IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0104(byte opcode2, IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0102(opcode2, destination, (byte) 0, base, index, scale);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0105(byte opcode2, IA32XMMRegister destination, IA32IndirectRegister32 source) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0103(opcode2, destination, (byte) 0, source);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0106(byte opcode2, IA32XMMRegister destination, IA32XMMRegister source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0107(byte opcode2, IA32XMMRegister destination, int m32) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0108(byte opcode2, IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0109(byte opcode2, IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0110(byte opcode2, IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0111(byte opcode2, IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0112(byte opcode2, IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0113(byte opcode2, IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0111(opcode2, destination, (byte) 0, base, index, scale);
            return;
        }
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0114(byte opcode2, IA32XMMRegister destination, IA32IndirectRegister32 source) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0112(opcode2, destination, (byte) 0, source);
            return;
        }
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0115(byte opcode2, IA32XMMRegister destination, IA32XMMRegister source) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0116(byte opcode2, IA32XMMRegister destination, int m32) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0117(byte opcode2, IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0118(byte opcode2, IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0119(byte opcode2, IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0120(byte opcode2, IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0121(byte opcode2, IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0122(byte opcode2, IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0120(opcode2, destination, (byte) 0, base, index, scale);
            return;
        }
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0123(byte opcode2, IA32XMMRegister destination, IA32IndirectRegister32 source) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0121(opcode2, destination, (byte) 0, source);
            return;
        }
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0124(byte opcode2, IA32XMMRegister destination, IA32XMMRegister source) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0125(byte opcode2, IA32XMMRegister destination, int m32) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0126(byte opcode2, IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0127(byte opcode2, IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0128(byte opcode2, IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0129(byte opcode1, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0130(byte opcode1, byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0131(byte opcode1, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0129(opcode1, (byte) 0, base, index, scale, source);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0132(byte opcode1, IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
    }

    private void assemble0133(byte opcode1, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0130(opcode1, (byte) 0, destination, source);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0134(byte opcode1, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0135(byte opcode1, int m32, IA32GeneralRegister16 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0136(byte opcode1, int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0137(byte opcode1, int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0138(byte opcode2, IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0139(byte opcode2, IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0140(byte opcode2, IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0138(opcode2, destination, (byte) 0, base, index, scale);
            return;
        }
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0141(byte opcode2, IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0142(byte opcode2, IA32GeneralRegister16 destination, IA32IndirectRegister32 source) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0139(opcode2, destination, (byte) 0, source);
            return;
        }
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0143(byte opcode2, IA32GeneralRegister16 destination, int m32) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0144(byte opcode2, IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0145(byte opcode2, IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0146(byte opcode2, IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0147(byte opcode2, IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0148(byte opcode2, IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0149(byte opcode2, IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0147(opcode2, destination, (byte) 0, base, index, scale);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0150(byte opcode2, IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0151(byte opcode2, IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0148(opcode2, destination, (byte) 0, source);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0152(byte opcode2, IA32GeneralRegister32 destination, int m32) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0153(byte opcode2, IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0154(byte opcode2, IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0155(byte opcode2, IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0156(byte opcode2, IA32GeneralRegister32 st_i) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte((byte) (opcode2 + st_i.value())); // opcode2
    }

    private void assemble0157(byte opcode2, byte modRmOpcode, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
        emitByte(imm8); // appended
    }

    private void assemble0158(byte opcode2, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0159(byte opcode2, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0160(byte opcode2, byte modRmOpcode, byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
        emitByte(imm8); // appended
    }

    private void assemble0161(byte opcode2, byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0162(byte opcode2, byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0163(byte opcode2, byte modRmOpcode, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0157(opcode2, modRmOpcode, (byte) 0, base, index, scale, imm8);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(imm8); // appended
    }

    private void assemble0164(byte opcode2, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0158(opcode2, (byte) 0, base, index, scale, source);
            return;
        }
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0165(byte opcode2, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0159(opcode2, (byte) 0, base, index, scale, source);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0166(byte opcode2, byte modRmOpcode, IA32GeneralRegister16 destination, byte imm8) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        emitByte(imm8); // appended
    }

    private void assemble0167(byte opcode2, IA32GeneralRegister16 destination, IA32GeneralRegister16 source) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
    }

    private void assemble0168(byte opcode2, byte modRmOpcode, IA32GeneralRegister32 destination, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        emitByte(imm8); // appended
    }

    private void assemble0169(byte opcode2, IA32GeneralRegister32 destination, IA32GeneralRegister32 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
    }

    private void assemble0170(byte opcode2, byte modRmOpcode, IA32IndirectRegister32 destination, byte imm8) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0160(opcode2, modRmOpcode, (byte) 0, destination, imm8);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(imm8); // appended
    }

    private void assemble0171(byte opcode2, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0161(opcode2, (byte) 0, destination, source);
            return;
        }
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0172(byte opcode2, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0162(opcode2, (byte) 0, destination, source);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0173(byte opcode2, byte modRmOpcode, int m32, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 5 << 0; // rm field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0174(byte opcode2, byte modRmOpcode, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0175(byte opcode2, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0176(byte opcode2, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0177(byte opcode2, int m32, IA32GeneralRegister16 source) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0178(byte opcode2, int m32, IA32GeneralRegister32 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0179(byte opcode2, byte modRmOpcode, int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0180(byte opcode2, int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0181(byte opcode2, int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0182(byte opcode2, byte modRmOpcode, int disp32, IA32IndirectRegister32 destination, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0183(byte opcode2, int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0184(byte opcode2, int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0185(byte opcode1, byte modRmOpcode, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0186(byte opcode1, byte modRmOpcode, byte disp8, IA32IndirectRegister32 destination) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0187(byte opcode1, byte modRmOpcode, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0185(opcode1, modRmOpcode, (byte) 0, base, index, scale);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0188(byte opcode1, byte modRmOpcode, IA32GeneralRegister16 destination) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((3 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0189(byte opcode1, byte modRmOpcode, IA32GeneralRegister32 destination) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((3 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0190(byte opcode1, byte modRmOpcode, IA32IndirectRegister32 destination) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0186(opcode1, modRmOpcode, (byte) 0, destination);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0191(byte opcode1, int m32) {
        emitByte(opcode1);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0192(byte opcode1, int rel32) {
        emitByte(opcode1);
        // appended:
        emitByte((byte) (rel32 & 0xff));
        rel32 >>= 8;
        emitByte((byte) (rel32 & 0xff));
        rel32 >>= 8;
        emitByte((byte) (rel32 & 0xff));
        rel32 >>= 8;
        emitByte((byte) (rel32 & 0xff));
    }

    private void assemble0193(byte opcode1, byte modRmOpcode, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0194(byte opcode1, byte modRmOpcode, int m32, IA32IndexRegister32 index, Scale scale) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0195(byte opcode1, byte modRmOpcode, int disp32, IA32IndirectRegister32 destination) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0196(byte opcode1, short rel16) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        // appended:
        emitByte((byte) (rel16 & 0xff));
        rel16 >>= 8;
        emitByte((byte) (rel16 & 0xff));
    }

    private void assemble0197(byte opcode1) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
    }

    private void assemble0198(byte opcode2) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
    }

    private void assemble0199(byte opcode2, IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0200(byte opcode2, IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source, IA32XMMComparison ia32xmmcomparison) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0201(byte opcode2, IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0199(opcode2, destination, (byte) 0, base, index, scale, ia32xmmcomparison);
            return;
        }
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0202(byte opcode2, IA32XMMRegister destination, IA32IndirectRegister32 source, IA32XMMComparison ia32xmmcomparison) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0200(opcode2, destination, (byte) 0, source, ia32xmmcomparison);
            return;
        }
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0203(byte opcode2, IA32XMMRegister destination, IA32XMMRegister source, IA32XMMComparison ia32xmmcomparison) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0204(byte opcode2, IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0205(byte opcode2, IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0206(byte opcode2, IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source, IA32XMMComparison ia32xmmcomparison) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0207(byte opcode2, IA32XMMRegister destination, int m32, IA32XMMComparison ia32xmmcomparison) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0208(byte opcode2, IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0209(byte opcode2, IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source, IA32XMMComparison ia32xmmcomparison) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0210(byte opcode2, IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0208(opcode2, destination, (byte) 0, base, index, scale, ia32xmmcomparison);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0211(byte opcode2, IA32XMMRegister destination, IA32IndirectRegister32 source, IA32XMMComparison ia32xmmcomparison) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0209(opcode2, destination, (byte) 0, source, ia32xmmcomparison);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0212(byte opcode2, IA32XMMRegister destination, IA32XMMRegister source, IA32XMMComparison ia32xmmcomparison) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0213(byte opcode2, IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0214(byte opcode2, IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0215(byte opcode2, IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source, IA32XMMComparison ia32xmmcomparison) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0216(byte opcode2, IA32XMMRegister destination, int m32, IA32XMMComparison ia32xmmcomparison) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0217(byte opcode2, IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0218(byte opcode2, IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source, IA32XMMComparison ia32xmmcomparison) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0219(byte opcode2, IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0217(opcode2, destination, (byte) 0, base, index, scale, ia32xmmcomparison);
            return;
        }
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0220(byte opcode2, IA32XMMRegister destination, IA32IndirectRegister32 source, IA32XMMComparison ia32xmmcomparison) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0218(opcode2, destination, (byte) 0, source, ia32xmmcomparison);
            return;
        }
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0221(byte opcode2, IA32XMMRegister destination, IA32XMMRegister source, IA32XMMComparison ia32xmmcomparison) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0222(byte opcode2, IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0223(byte opcode2, IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0224(byte opcode2, IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source, IA32XMMComparison ia32xmmcomparison) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0225(byte opcode2, IA32XMMRegister destination, int m32, IA32XMMComparison ia32xmmcomparison) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0226(byte opcode2, IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0227(byte opcode2, IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source, IA32XMMComparison ia32xmmcomparison) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0228(byte opcode2, IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0226(opcode2, destination, (byte) 0, base, index, scale, ia32xmmcomparison);
            return;
        }
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0229(byte opcode2, IA32XMMRegister destination, IA32IndirectRegister32 source, IA32XMMComparison ia32xmmcomparison) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0227(opcode2, destination, (byte) 0, source, ia32xmmcomparison);
            return;
        }
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0230(byte opcode2, IA32XMMRegister destination, IA32XMMRegister source, IA32XMMComparison ia32xmmcomparison) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0231(byte opcode2, IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0232(byte opcode2, IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale, IA32XMMComparison ia32xmmcomparison) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0233(byte opcode2, IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source, IA32XMMComparison ia32xmmcomparison) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0234(byte opcode2, IA32XMMRegister destination, int m32, IA32XMMComparison ia32xmmcomparison) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte((byte) ia32xmmcomparison.value()); // appended
    }

    private void assemble0235(byte opcode2, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0236(byte opcode2, byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0237(byte opcode2, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0235(opcode2, (byte) 0, base, index, scale, source);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0238(byte opcode2, IA32GeneralRegister8 destination, IA32GeneralRegister8 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
    }

    private void assemble0239(byte opcode2, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0236(opcode2, (byte) 0, destination, source);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0240(byte opcode2, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0241(byte opcode2, int m32, IA32GeneralRegister8 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0242(byte opcode2, int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister8 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0243(byte opcode2, int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister8 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0244(byte opcode2, byte modRmOpcode, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0245(byte opcode2, byte modRmOpcode, byte disp8, IA32IndirectRegister32 destination) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0246(byte opcode2, byte modRmOpcode, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0244(opcode2, modRmOpcode, (byte) 0, base, index, scale);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0247(byte opcode2, byte modRmOpcode, IA32IndirectRegister32 destination) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0245(opcode2, modRmOpcode, (byte) 0, destination);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0248(byte opcode2, byte modRmOpcode, int m32) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 5 << 0; // rm field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0249(byte opcode2, byte modRmOpcode, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0250(byte opcode2, byte modRmOpcode, int m32, IA32IndexRegister32 index, Scale scale) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0251(byte opcode2, byte modRmOpcode, int disp32, IA32IndirectRegister32 destination) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0252(byte opcode2, MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0253(byte opcode2, MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0254(byte opcode2, MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0252(opcode2, destination, (byte) 0, base, index, scale);
            return;
        }
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0255(byte opcode2, MMXRegister destination, IA32IndirectRegister32 source) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0253(opcode2, destination, (byte) 0, source);
            return;
        }
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0256(byte opcode2, MMXRegister destination, IA32XMMRegister source) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0257(byte opcode2, MMXRegister destination, int m32) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0258(byte opcode2, MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0259(byte opcode2, MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0260(byte opcode2, MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0261(byte opcode2, IA32XMMRegister destination, MMXRegister source) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0262(byte opcode2, IA32XMMRegister destination, MMXRegister source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0263(byte opcode2, MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0264(byte opcode2, MMXRegister destination, byte disp8, IA32IndirectRegister32 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0265(byte opcode2, MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0263(opcode2, destination, (byte) 0, base, index, scale);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0266(byte opcode2, MMXRegister destination, IA32IndirectRegister32 source) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0264(opcode2, destination, (byte) 0, source);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0267(byte opcode2, MMXRegister destination, IA32XMMRegister source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0268(byte opcode2, MMXRegister destination, int m32) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0269(byte opcode2, MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0270(byte opcode2, MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0271(byte opcode2, MMXRegister destination, int disp32, IA32IndirectRegister32 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0272(byte opcode2, IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0273(byte opcode2, IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0274(byte opcode2, IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0272(opcode2, destination, (byte) 0, base, index, scale);
            return;
        }
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0275(byte opcode2, IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0273(opcode2, destination, (byte) 0, source);
            return;
        }
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0276(byte opcode2, IA32GeneralRegister32 destination, IA32XMMRegister source) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0277(byte opcode2, IA32GeneralRegister32 destination, int m32) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0278(byte opcode2, IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0279(byte opcode2, IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0280(byte opcode2, IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0281(byte opcode2, IA32XMMRegister destination, IA32GeneralRegister32 source) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0282(byte opcode2, IA32XMMRegister destination, IA32GeneralRegister32 source) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0283(byte opcode2, IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0284(byte opcode2, IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0285(byte opcode2, IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0283(opcode2, destination, (byte) 0, base, index, scale);
            return;
        }
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0286(byte opcode2, IA32GeneralRegister32 destination, IA32IndirectRegister32 source) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0284(opcode2, destination, (byte) 0, source);
            return;
        }
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0287(byte opcode2, IA32GeneralRegister32 destination, IA32XMMRegister source) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0288(byte opcode2, IA32GeneralRegister32 destination, int m32) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0289(byte opcode2, IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0290(byte opcode2, IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0291(byte opcode2, IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0292(byte opcode1, byte modRmOpcode, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0293(byte opcode1, byte modRmOpcode, byte disp8, IA32IndirectRegister32 destination) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0294(byte opcode1, byte modRmOpcode, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0292(opcode1, modRmOpcode, (byte) 0, base, index, scale);
            return;
        }
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0295(byte opcode1, IA32GeneralRegister16 register) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (opcode1 + register.value())); // opcode1
    }

    private void assemble0296(byte opcode1, IA32GeneralRegister32 register) {
        emitByte((byte) (opcode1 + register.value())); // opcode1
    }

    private void assemble0297(byte opcode1, byte modRmOpcode, IA32GeneralRegister8 destination) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((3 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0298(byte opcode1, byte modRmOpcode, IA32IndirectRegister32 destination) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0293(opcode1, modRmOpcode, (byte) 0, destination);
            return;
        }
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0299(byte opcode1, byte modRmOpcode, int m32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 5 << 0; // rm field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0300(byte opcode1, byte modRmOpcode, int m32) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 5 << 0; // rm field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0301(byte opcode1, byte modRmOpcode, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0302(byte opcode1, byte modRmOpcode, int m32, IA32IndexRegister32 index, Scale scale) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0303(byte opcode1, byte modRmOpcode, int disp32, IA32IndirectRegister32 destination) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0304(byte opcode1, short imm16, byte imm8) {
        emitByte(opcode1);
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0305(byte opcode2) {
        emitByte((byte) (0xD9)); // opcode1
        emitByte(opcode2);
    }

    private void assemble0306(byte opcode2, FPStackRegister st_i) {
        emitByte((byte) (0xD8)); // opcode1
        emitByte((byte) (opcode2 + st_i.value())); // opcode2
    }

    private void assemble0307(byte opcode2, FPStackRegister st_i) {
        emitByte((byte) (0xDC)); // opcode1
        emitByte((byte) (opcode2 + st_i.value())); // opcode2
    }

    private void assemble0308(byte opcode2, FPStackRegister st_i) {
        emitByte((byte) (0xDE)); // opcode1
        emitByte((byte) (opcode2 + st_i.value())); // opcode2
    }

    private void assemble0309(byte opcode2) {
        emitByte((byte) (0xDB)); // opcode1
        emitByte(opcode2);
    }

    private void assemble0310(byte opcode2, FPStackRegister st_i) {
        emitByte((byte) (0xDA)); // opcode1
        emitByte((byte) (opcode2 + st_i.value())); // opcode2
    }

    private void assemble0311(byte opcode2, FPStackRegister st_i) {
        emitByte((byte) (0xDB)); // opcode1
        emitByte((byte) (opcode2 + st_i.value())); // opcode2
    }

    private void assemble0312(byte opcode2, FPStackRegister st_i) {
        emitByte((byte) (0xDF)); // opcode1
        emitByte((byte) (opcode2 + st_i.value())); // opcode2
    }

    private void assemble0313(byte opcode2) {
        emitByte((byte) (0xDE)); // opcode1
        emitByte(opcode2);
    }

    private void assemble0314(byte opcode2, FPStackRegister st_i) {
        emitByte((byte) (0xDD)); // opcode1
        emitByte((byte) (opcode2 + st_i.value())); // opcode2
    }

    private void assemble0315(byte opcode2, FPStackRegister st_i) {
        emitByte((byte) (0xD9)); // opcode1
        emitByte((byte) (opcode2 + st_i.value())); // opcode2
    }

    private void assemble0316(byte opcode2) {
        emitByte((byte) (0xDF)); // opcode1
        emitByte(opcode2);
    }

    private void assemble0317(byte opcode2) {
        emitByte((byte) (0xDA)); // opcode1
        emitByte(opcode2);
    }

    private void assemble0318(byte opcode1, IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
        emitByte(imm8); // appended
    }

    private void assemble0319(byte opcode1, IA32GeneralRegister16 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0320(byte opcode1, IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source, byte imm8) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
        emitByte(imm8); // appended
    }

    private void assemble0321(byte opcode1, IA32GeneralRegister16 destination, byte disp8, IA32IndirectRegister32 source, short imm16) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0322(byte opcode1, IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0318(opcode1, destination, (byte) 0, base, index, scale, imm8);
            return;
        }
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(imm8); // appended
    }

    private void assemble0323(byte opcode1, IA32GeneralRegister16 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0319(opcode1, destination, (byte) 0, base, index, scale, imm16);
            return;
        }
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0324(byte opcode1, IA32GeneralRegister16 destination, IA32GeneralRegister16 source, byte imm8) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        emitByte(imm8); // appended
    }

    private void assemble0325(byte opcode1, IA32GeneralRegister16 destination, IA32GeneralRegister16 source, short imm16) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0326(byte opcode1, IA32GeneralRegister16 destination, IA32IndirectRegister32 source, byte imm8) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0320(opcode1, destination, (byte) 0, source, imm8);
            return;
        }
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(imm8); // appended
    }

    private void assemble0327(byte opcode1, IA32GeneralRegister16 destination, IA32IndirectRegister32 source, short imm16) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0321(opcode1, destination, (byte) 0, source, imm16);
            return;
        }
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0328(byte opcode1, IA32GeneralRegister16 destination, int m32, byte imm8) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0329(byte opcode1, IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0330(byte opcode1, IA32GeneralRegister16 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0331(byte opcode1, IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0332(byte opcode1, IA32GeneralRegister16 destination, int m32, IA32IndexRegister32 index, Scale scale, short imm16) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0333(byte opcode1, IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source, byte imm8) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0334(byte opcode1, IA32GeneralRegister16 destination, int disp32, IA32IndirectRegister32 source, short imm16) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0335(byte opcode1, IA32GeneralRegister16 destination, int m32, short imm16) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0336(byte opcode1, IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
        emitByte(imm8); // appended
    }

    private void assemble0337(byte opcode1, IA32GeneralRegister32 destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0338(byte opcode1, IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source, byte imm8) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
        emitByte(imm8); // appended
    }

    private void assemble0339(byte opcode1, IA32GeneralRegister32 destination, byte disp8, IA32IndirectRegister32 source, int imm32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0340(byte opcode1, IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0336(opcode1, destination, (byte) 0, base, index, scale, imm8);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(imm8); // appended
    }

    private void assemble0341(byte opcode1, IA32GeneralRegister32 destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0337(opcode1, destination, (byte) 0, base, index, scale, imm32);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0342(byte opcode1, IA32GeneralRegister32 destination, IA32GeneralRegister32 source, byte imm8) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        emitByte(imm8); // appended
    }

    private void assemble0343(byte opcode1, IA32GeneralRegister32 destination, IA32GeneralRegister32 source, int imm32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0344(byte opcode1, IA32GeneralRegister32 destination, IA32IndirectRegister32 source, byte imm8) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0338(opcode1, destination, (byte) 0, source, imm8);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(imm8); // appended
    }

    private void assemble0345(byte opcode1, IA32GeneralRegister32 destination, IA32IndirectRegister32 source, int imm32) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0339(opcode1, destination, (byte) 0, source, imm32);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0346(byte opcode1, IA32GeneralRegister32 destination, int m32, byte imm8) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0347(byte opcode1, IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0348(byte opcode1, IA32GeneralRegister32 destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0349(byte opcode1, IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0350(byte opcode1, IA32GeneralRegister32 destination, int m32, IA32IndexRegister32 index, Scale scale, int imm32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0351(byte opcode1, IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source, byte imm8) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0352(byte opcode1, IA32GeneralRegister32 destination, int disp32, IA32IndirectRegister32 source, int imm32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0353(byte opcode1, IA32GeneralRegister32 destination, int m32, int imm32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0354(byte opcode1, byte imm8) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        emitByte(imm8); // appended
    }

    private void assemble0355(byte opcode1, byte rel8) {
        emitByte(opcode1);
        emitByte(rel8); // appended
    }

    private void assemble0356(byte opcode2, int rel32) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        // appended:
        emitByte((byte) (rel32 & 0xff));
        rel32 >>= 8;
        emitByte((byte) (rel32 & 0xff));
        rel32 >>= 8;
        emitByte((byte) (rel32 & 0xff));
        rel32 >>= 8;
        emitByte((byte) (rel32 & 0xff));
    }

    private void assemble0357(byte opcode2, byte modRmOpcode) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        emitByte(modRMByte);
    }

    private void assemble0358(byte opcode2, byte modRmOpcode, IA32GeneralRegister16 destination) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0359(byte opcode1, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
        emitByte(imm8); // appended
    }

    private void assemble0360(byte opcode1, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, SegmentRegister source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0361(byte opcode1, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0362(byte opcode1, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0363(byte opcode1, byte disp8, IA32IndirectRegister32 destination, byte imm8) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
        emitByte(imm8); // appended
    }

    private void assemble0364(byte opcode1, byte disp8, IA32IndirectRegister32 destination, SegmentRegister source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0365(byte opcode1, byte disp8, IA32IndirectRegister32 destination, int imm32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0366(byte opcode1, byte disp8, IA32IndirectRegister32 destination, short imm16) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0367(byte opcode1, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0359(opcode1, (byte) 0, base, index, scale, imm8);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(imm8); // appended
    }

    private void assemble0368(byte opcode1, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, SegmentRegister source) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0360(opcode1, (byte) 0, base, index, scale, source);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0369(byte opcode1, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0361(opcode1, (byte) 0, base, index, scale, imm32);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0370(byte opcode1, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0362(opcode1, (byte) 0, base, index, scale, imm16);
            return;
        }
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0371(byte opcode1, IA32GeneralRegister16 destination, SegmentRegister source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
    }

    private void assemble0372(byte opcode1, IA32GeneralRegister16 destination, short imm16) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0373(byte opcode1, IA32GeneralRegister16 register, short imm16) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (opcode1 + register.value())); // opcode1
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0374(byte opcode2, IA32GeneralRegister32 destination, ControlRegister source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
    }

    private void assemble0375(byte opcode2, IA32GeneralRegister32 destination, DebugRegister source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
    }

    private void assemble0376(byte opcode1, IA32GeneralRegister32 destination, int imm32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0377(byte opcode1, IA32GeneralRegister32 register, int imm32) {
        emitByte((byte) (opcode1 + register.value())); // opcode1
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0378(byte opcode1, IA32GeneralRegister8 register, byte imm8) {
        emitByte((byte) (opcode1 + register.value())); // opcode1
        emitByte(imm8); // appended
    }

    private void assemble0379(byte opcode1, IA32GeneralRegister8 destination, byte imm8) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        emitByte(imm8); // appended
    }

    private void assemble0380(byte opcode1, IA32IndirectRegister32 destination, byte imm8) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0363(opcode1, (byte) 0, destination, imm8);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(imm8); // appended
    }

    private void assemble0381(byte opcode1, IA32IndirectRegister32 destination, SegmentRegister source) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0364(opcode1, (byte) 0, destination, source);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0382(byte opcode1, IA32IndirectRegister32 destination, int imm32) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0365(opcode1, (byte) 0, destination, imm32);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0383(byte opcode1, IA32IndirectRegister32 destination, short imm16) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0366(opcode1, (byte) 0, destination, imm16);
            return;
        }
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0384(byte opcode2, ControlRegister destination, IA32GeneralRegister32 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0385(byte opcode2, DebugRegister destination, IA32GeneralRegister32 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0386(byte opcode1, SegmentRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0387(byte opcode1, SegmentRegister destination, byte disp8, IA32IndirectRegister32 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0388(byte opcode1, SegmentRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0386(opcode1, destination, (byte) 0, base, index, scale);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0389(byte opcode1, SegmentRegister destination, IA32GeneralRegister16 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0390(byte opcode1, SegmentRegister destination, IA32IndirectRegister32 source) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0387(opcode1, destination, (byte) 0, source);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0391(byte opcode1, SegmentRegister destination, int m32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0392(byte opcode1, SegmentRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0393(byte opcode1, SegmentRegister destination, int m32, IA32IndexRegister32 index, Scale scale) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0394(byte opcode1, SegmentRegister destination, int disp32, IA32IndirectRegister32 source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0395(byte opcode1, int m32) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0396(byte opcode1, int m32, byte imm8) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0397(byte opcode1, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0398(byte opcode1, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, SegmentRegister source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0399(byte opcode1, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, int imm32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0400(byte opcode1, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, short imm16) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0401(byte opcode1, int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0402(byte opcode1, int m32, IA32IndexRegister32 index, Scale scale, SegmentRegister source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0403(byte opcode1, int m32, IA32IndexRegister32 index, Scale scale, int imm32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0404(byte opcode1, int m32, IA32IndexRegister32 index, Scale scale, short imm16) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0405(byte opcode1, int disp32, IA32IndirectRegister32 destination, byte imm8) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0406(byte opcode1, int disp32, IA32IndirectRegister32 destination, SegmentRegister source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0407(byte opcode1, int disp32, IA32IndirectRegister32 destination, int imm32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0408(byte opcode1, int disp32, IA32IndirectRegister32 destination, short imm16) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0409(byte opcode1, int m32, SegmentRegister source) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0410(byte opcode1, int m32, int imm32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        // appended:
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
        imm32 >>= 8;
        emitByte((byte) (imm32 & 0xff));
    }

    private void assemble0411(byte opcode1, int m32, short imm16) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0412(byte opcode2, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0413(byte opcode2, byte disp8, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0414(byte opcode2, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0412(opcode2, (byte) 0, base, index, scale, source);
            return;
        }
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0415(byte opcode2, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0413(opcode2, (byte) 0, destination, source);
            return;
        }
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0416(byte opcode2, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0417(byte opcode2, int m32, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0418(byte opcode2, int disp32, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0419(byte opcode2, int m32, IA32XMMRegister source) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0420(byte opcode2, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0421(byte opcode2, byte disp8, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0422(byte opcode2, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0420(opcode2, (byte) 0, base, index, scale, source);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0423(byte opcode2, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0421(opcode2, (byte) 0, destination, source);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0424(byte opcode2, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0425(byte opcode2, int m32, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0426(byte opcode2, int disp32, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0427(byte opcode2, int m32, IA32XMMRegister source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0428(byte opcode2, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, MMXRegister source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0429(byte opcode2, byte disp8, IA32IndirectRegister32 destination, MMXRegister source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0430(byte opcode2, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, MMXRegister source) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0428(opcode2, (byte) 0, base, index, scale, source);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0431(byte opcode2, IA32GeneralRegister32 destination, IA32XMMRegister source) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
    }

    private void assemble0432(byte opcode2, IA32GeneralRegister32 destination, MMXRegister source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
    }

    private void assemble0433(byte opcode2, IA32IndirectRegister32 destination, MMXRegister source) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0429(opcode2, (byte) 0, destination, source);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0434(byte opcode2, IA32XMMRegister destination, IA32GeneralRegister32 source) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0435(byte opcode2, MMXRegister destination, IA32GeneralRegister32 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0436(byte opcode2, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, MMXRegister source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0437(byte opcode2, int m32, IA32IndexRegister32 index, Scale scale, MMXRegister source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0438(byte opcode2, int disp32, IA32IndirectRegister32 destination, MMXRegister source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0439(byte opcode2, int m32, MMXRegister source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0440(byte opcode2, MMXRegister destination, IA32XMMRegister source) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0441(byte opcode2, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0442(byte opcode2, byte disp8, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0443(byte opcode2, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0441(opcode2, (byte) 0, base, index, scale, source);
            return;
        }
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0444(byte opcode2, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0442(opcode2, (byte) 0, destination, source);
            return;
        }
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0445(byte opcode2, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0446(byte opcode2, int m32, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0447(byte opcode2, int disp32, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0448(byte opcode2, int m32, IA32XMMRegister source) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0449(byte opcode2, IA32GeneralRegister32 destination, IA32XMMRegister source) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0450(byte opcode2, IA32GeneralRegister32 destination, IA32XMMRegister source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0451(byte opcode2, IA32XMMRegister destination, IA32XMMRegister source) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
    }

    private void assemble0452(byte opcode2, MMXRegister destination, MMXRegister source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0453(byte opcode2, IA32XMMRegister destination, MMXRegister source) {
        emitByte(((byte) 0xF3)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0454(byte opcode2, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0455(byte opcode2, byte disp8, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0456(byte opcode2, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0454(opcode2, (byte) 0, base, index, scale, source);
            return;
        }
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0457(byte opcode2, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0455(opcode2, (byte) 0, destination, source);
            return;
        }
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0458(byte opcode2, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0459(byte opcode2, int m32, IA32IndexRegister32 index, Scale scale, IA32XMMRegister source) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0460(byte opcode2, int disp32, IA32IndirectRegister32 destination, IA32XMMRegister source) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0461(byte opcode2, int m32, IA32XMMRegister source) {
        emitByte(((byte) 0xF2)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0462(byte opcode2, IA32GeneralRegister16 destination, IA32GeneralRegister8 source) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0463(byte opcode2, IA32GeneralRegister32 destination, IA32GeneralRegister8 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0464(byte opcode2, IA32GeneralRegister32 destination, IA32GeneralRegister16 source) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0465(byte opcode2, IA32GeneralRegister32 destination, IA32XMMRegister source, byte imm8) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        emitByte(imm8); // appended
    }

    private void assemble0466(byte opcode2, IA32GeneralRegister32 destination, MMXRegister source, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        emitByte(imm8); // appended
    }

    private void assemble0467(byte opcode2, IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
        emitByte(imm8); // appended
    }

    private void assemble0468(byte opcode2, IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source, byte imm8) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
        emitByte(imm8); // appended
    }

    private void assemble0469(byte opcode2, IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0467(opcode2, destination, (byte) 0, base, index, scale, imm8);
            return;
        }
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(imm8); // appended
    }

    private void assemble0470(byte opcode2, IA32XMMRegister destination, IA32GeneralRegister32 source, byte imm8) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        emitByte(imm8); // appended
    }

    private void assemble0471(byte opcode2, IA32XMMRegister destination, IA32IndirectRegister32 source, byte imm8) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0468(opcode2, destination, (byte) 0, source, imm8);
            return;
        }
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(imm8); // appended
    }

    private void assemble0472(byte opcode2, IA32XMMRegister destination, int m32, byte imm8) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0473(byte opcode2, IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0474(byte opcode2, IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0475(byte opcode2, IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source, byte imm8) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0476(byte opcode2, MMXRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
        emitByte(imm8); // appended
    }

    private void assemble0477(byte opcode2, MMXRegister destination, byte disp8, IA32IndirectRegister32 source, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
        emitByte(imm8); // appended
    }

    private void assemble0478(byte opcode2, MMXRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0476(opcode2, destination, (byte) 0, base, index, scale, imm8);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(imm8); // appended
    }

    private void assemble0479(byte opcode2, MMXRegister destination, IA32GeneralRegister32 source, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        emitByte(imm8); // appended
    }

    private void assemble0480(byte opcode2, MMXRegister destination, IA32IndirectRegister32 source, byte imm8) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0477(opcode2, destination, (byte) 0, source, imm8);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(imm8); // appended
    }

    private void assemble0481(byte opcode2, MMXRegister destination, int m32, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0482(byte opcode2, MMXRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0483(byte opcode2, MMXRegister destination, int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0484(byte opcode2, MMXRegister destination, int disp32, IA32IndirectRegister32 source, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0485(byte opcode1, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0486(byte opcode1, byte disp8, IA32IndirectRegister32 destination) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0487(byte opcode1, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0485(opcode1, (byte) 0, base, index, scale);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0488(byte opcode1, IA32IndirectRegister32 destination) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0486(opcode1, (byte) 0, destination);
            return;
        }
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0489(byte opcode1, int m32) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0490(byte opcode1, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0491(byte opcode1, int m32, IA32IndexRegister32 index, Scale scale) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0492(byte opcode1, int disp32, IA32IndirectRegister32 destination) {
        emitByte(opcode1);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0493(byte opcode2, MMXRegister destination, MMXRegister source, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        emitByte(imm8); // appended
    }

    private void assemble0494(byte opcode2, byte modRmOpcode, MMXRegister destination, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        emitByte(imm8); // appended
    }

    private void assemble0495(byte opcode2, byte modRmOpcode, IA32XMMRegister destination, byte imm8) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        emitByte(imm8); // appended
    }

    private void assemble0496(byte opcode1, short imm16) {
        emitByte(opcode1);
        // appended:
        emitByte((byte) (imm16 & 0xff));
        imm16 >>= 8;
        emitByte((byte) (imm16 & 0xff));
    }

    private void assemble0497(byte opcode2, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
    }

    private void assemble0498(byte opcode2, byte disp8, IA32IndirectRegister32 destination) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
    }

    private void assemble0499(byte opcode2, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0497(opcode2, (byte) 0, base, index, scale);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
    }

    private void assemble0500(byte opcode2, IA32GeneralRegister8 destination) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0501(byte opcode2, IA32IndirectRegister32 destination) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0498(opcode2, (byte) 0, destination);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
    }

    private void assemble0502(byte opcode2, int m32) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0503(byte opcode2, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0504(byte opcode2, int m32, IA32IndexRegister32 index, Scale scale) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
    }

    private void assemble0505(byte opcode2, int disp32, IA32IndirectRegister32 destination) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
    }

    private void assemble0506(byte opcode2, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source, byte imm8) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
        emitByte(imm8); // appended
    }

    private void assemble0507(byte opcode2, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
        emitByte(imm8); // appended
    }

    private void assemble0508(byte opcode2, byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister16 source, byte imm8) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
        emitByte(imm8); // appended
    }

    private void assemble0509(byte opcode2, byte disp8, IA32IndirectRegister32 destination, IA32GeneralRegister32 source, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
        emitByte(imm8); // appended
    }

    private void assemble0510(byte opcode2, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source, byte imm8) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0506(opcode2, (byte) 0, base, index, scale, source, imm8);
            return;
        }
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(imm8); // appended
    }

    private void assemble0511(byte opcode2, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source, byte imm8) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0507(opcode2, (byte) 0, base, index, scale, source, imm8);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(imm8); // appended
    }

    private void assemble0512(byte opcode2, IA32GeneralRegister16 destination, IA32GeneralRegister16 source, byte imm8) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        emitByte(imm8); // appended
    }

    private void assemble0513(byte opcode2, IA32GeneralRegister32 destination, IA32GeneralRegister32 source, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        emitByte(imm8); // appended
    }

    private void assemble0514(byte opcode2, IA32IndirectRegister32 destination, IA32GeneralRegister16 source, byte imm8) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0508(opcode2, (byte) 0, destination, source, imm8);
            return;
        }
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(imm8); // appended
    }

    private void assemble0515(byte opcode2, IA32IndirectRegister32 destination, IA32GeneralRegister32 source, byte imm8) {
        if (destination == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0509(opcode2, (byte) 0, destination, source, imm8);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(imm8); // appended
    }

    private void assemble0516(byte opcode2, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source, byte imm8) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0517(byte opcode2, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0518(byte opcode2, int m32, IA32GeneralRegister16 source, byte imm8) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0519(byte opcode2, int m32, IA32GeneralRegister32 source, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0520(byte opcode2, int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister16 source, byte imm8) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0521(byte opcode2, int m32, IA32IndexRegister32 index, Scale scale, IA32GeneralRegister32 source, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0522(byte opcode2, int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister16 source, byte imm8) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0523(byte opcode2, int disp32, IA32IndirectRegister32 destination, IA32GeneralRegister32 source, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        modRMByte |= (source.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        if (destination == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0524(byte opcode2, IA32XMMRegister destination, IA32XMMRegister source, byte imm8) {
        emitByte(((byte) 0x66)); // instruction selection prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        emitByte(imm8); // appended
    }

    private void assemble0525(byte opcode2, IA32XMMRegister destination, byte disp8, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(disp8); // appended
        emitByte(imm8); // appended
    }

    private void assemble0526(byte opcode2, IA32XMMRegister destination, byte disp8, IA32IndirectRegister32 source, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((1 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(disp8); // appended
        emitByte(imm8); // appended
    }

    private void assemble0527(byte opcode2, IA32XMMRegister destination, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        if (base == IA32BaseRegister32.EBP_BASE) {
            assemble0525(opcode2, destination, (byte) 0, base, index, scale, imm8);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        emitByte(imm8); // appended
    }

    private void assemble0528(byte opcode2, IA32XMMRegister destination, IA32IndirectRegister32 source, byte imm8) {
        if (source == IA32IndirectRegister32.EBP_INDIRECT) {
            assemble0526(opcode2, destination, (byte) 0, source, imm8);
            return;
        }
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        emitByte(imm8); // appended
    }

    private void assemble0529(byte opcode2, IA32XMMRegister destination, IA32XMMRegister source, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        emitByte(imm8); // appended
    }

    private void assemble0530(byte opcode2, IA32XMMRegister destination, int m32, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 5 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0531(byte opcode2, IA32XMMRegister destination, int disp32, IA32BaseRegister32 base, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) 0;
        sibByte |= (base.value() & 7) << 0; // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0532(byte opcode2, IA32XMMRegister destination, int m32, IA32IndexRegister32 index, Scale scale, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((0 << 6)); // mod field
        modRMByte |= 4 << 0; // rm field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        emitByte(modRMByte);
        byte sibByte = (byte) (5 << 0); // base field
        sibByte |= (index.value() & 7) << 3; // index field
        sibByte |= scale.value() << 6; // scale field
        emitByte(sibByte);
        // appended:
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        m32 >>= 8;
        emitByte((byte) (m32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0533(byte opcode2, IA32XMMRegister destination, int disp32, IA32IndirectRegister32 source, byte imm8) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((2 << 6)); // mod field
        modRMByte |= (destination.value() & 7) << 3; // reg field
        modRMByte |= (source.value() & 7) << 0; // rm field
        emitByte(modRMByte);
        if (source == IA32IndirectRegister32.ESP_INDIRECT) {
            emitByte(((byte) 0x24)); // SIB byte
        }
        // appended:
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        disp32 >>= 8;
        emitByte((byte) (disp32 & 0xff));
        emitByte(imm8); // appended
    }

    private void assemble0534(byte opcode2, byte modRmOpcode, IA32GeneralRegister16 destination) {
        emitByte(((byte) 0x66)); // operand size prefix
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

    private void assemble0535(byte opcode2, byte modRmOpcode, IA32GeneralRegister32 destination) {
        emitByte((byte) (0x0F)); // opcode1
        emitByte(opcode2);
        byte modRMByte = (byte) ((3 << 6) | (modRmOpcode << 3)); // mod field, group opcode in reg field
        modRMByte |= (destination.value() & 7) << 0; // rm field
        emitByte(modRMByte);
    }

// END GENERATED RAW ASSEMBLER METHODS
}
